/*
 * FXVis3.c
 *
 * FX example module for Visual3.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "Visual3.h"
#include "FX.h"


/* these 2 parameters can be tuned to remove spurious shock regions */

#define MINSHOCKNODES 4			/* minimum # of nodes for inclusion */
#define MINSHOCKDPRES 0.1		/* minimum delta pressure for region */


#ifdef PREFIX
#undef PREFIX
#endif


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define V3PEVENTS    v3pevents_
#define V3DRAW3D     v3draw3d_
#define XFTNSDIALFUN xftnsdialfun_

#else

#define V3PEVENTS    v3pevents
#define V3DRAW3D     v3draw3d
#define XFTNSDIALFUN xftnsdialfun

#endif
#endif


  /* globals */

  static int    FXvc, FXsf, FXsl, FXt = 0;
  static int    FXnpts = 0;
  static int   *FXsegs, FXnseg = 0;
  static int   *FXsepind, *FXattind, *FXshowSurf, FXnbc;
  static float *FXvcoor,  *FXstren,  *FXcoors;
  static float *FXsepxyz = NULL, *FXattxyz = NULL;


  /* externals */
  
  extern void PREFIX V3SCAL(int *type, float *s);
  extern void PREFIX XFTNSDIALFUN(long *win, int *x, int *y, int *active,
                                  int *outline);


  /*  take the shock regions and remove those that are small or do not
   *    display a large enough pressure gradient
   */

static void
mark_tris(int nclust, int *clust, int *cell, int *cnode, int *cnodes,
          float *press, int ntris, int *cells)
{
  char  *cellTable;
  int   i, j, k, l, clen, nlen, max;
  float minp, maxp;

  clen = clust[nclust-1];
  for (max = i = 0; i < clen; i++) if (cell[i] > max) max = cell[i];

  cellTable = (char *) malloc(max*sizeof(char));
  if (cellTable == NULL) {
    printf(" Cannot Allocate cellTable (mark_tris)\n");
    return;
  }

  for (i = 0; i < max; i++) cellTable[i] = 0;

  /* mark the cells that are OK */

  for (k = j = i = 0; i < nclust; i++) {
    clen = clust[i] - j;
    nlen = cnode[i] - k;
    if (nlen > MINSHOCKNODES) {
      minp = maxp = press[cnodes[k]-1];
      for (l = 1; l < nlen; l++) {
        if (minp > press[cnodes[k+l]-1]) minp = press[cnodes[k+l]-1];
        if (maxp < press[cnodes[k+l]-1]) maxp = press[cnodes[k+l]-1];
      }
      /* maxp-minp measures the approximate shock strength for the region */
      if ((maxp-minp) > MINSHOCKDPRES) 
        for (l = j; l < clust[i]; l++) cellTable[cell[l]-1] = 1;
    }
    j = clust[i];
    k = cnode[i];
  }

  /* mark the not-OK triangles */

  for (i = 0; i < ntris; i++)
    if (cellTable[cells[i]-1] == 0) cells[i] = -cells[i];

  free(cellTable);
}


/* Visual3 programmed events routine */

void PREFIX
V3PEVENTS(long *win, int *type, int *xe, int *ye, int *state)
{
  int   i, j, k, nnode, ntets, npyra, nprism, nhexa, nblocks, *blocks, nsurf;
  int   istat, len, *tets, *pyra, *prism, *hexa, *blk, *iblk, zero = 0;
  int   *bc, *scel, nclust, *clust, *cell, *cnode, *cnodes;
  int   ntris, npts, *tris, *cells;
  long  ivec[5];
  float gamma, rvec[7], *xyz, *test, *vols, *press, *coors;
  char  cstr[80];
  static long  winid[8];
  static int   iopt, lineout = 1;
  static float ltime;

  /* initialization */

  if (FXt == 0) {
    FXvc = FXsf = FXsl = 0;
    i = 0;
    V3_GETSTATE(&i, winid, rvec, cstr, 80);
    FXt     = 1;
    FXnpts  = 0;

    i = 315;
    V3_GETSTRUC(&i, (void **) &tets,  &ntets);
    if (ntets != 0) {
      printf(" This Case has %d PolyTet Strips \n\n", ntets);
      return;
    }
    i = -2;
    V3_GETSTATE(&i, ivec, rvec, cstr, 80);
    iopt  = ivec[0];
    ltime = rvec[0];
    istat = 45;

    i = 301;
    V3_GETSTRUC(&i, (void **) &xyz,   &nnode);

    i = 311;
    V3_GETSTRUC(&i, (void **) &tets,  &ntets);
    i = 312;
    V3_GETSTRUC(&i, (void **) &pyra,  &npyra);
    i = 313;
    V3_GETSTRUC(&i, (void **) &prism, &nprism);
    i = 314;
    V3_GETSTRUC(&i, (void **) &hexa,  &nhexa);
    i = 317;
    V3_GETSTRUC(&i, (void **) &blk,   &nblocks);
    if (nblocks != 0) {
      blocks = (int *) malloc(3*nblocks*sizeof(int));
      if (blocks == NULL) {
        printf(" Error in Memory Allocation for %d Blocks\n\n", nblocks);
        return;
      }
      for (i = 0; i < nblocks; i++) {
        blocks[3*i  ] = blk[6*i  ];
        blocks[3*i+1] = blk[6*i+1];
        blocks[3*i+2] = blk[6*i+2];
        nnode -= blocks[3*i]*blocks[3*i+1]*blocks[3*i+2];
      }
      i = 306;
      V3_GETSTRUC(&i, (void **) &iblk,  &len);
      if (len != 0) istat += 16;
    }

    i = 321;
    V3_GETSTRUC(&i, (void **) &bc,   &FXnbc);
    i = 324;
    V3_GETSTRUC(&i, (void **) &scel, &nsurf);
    if (FXnbc > 0) {
      FXshowSurf = malloc(FXnbc*sizeof(int));
      FXattind   = malloc(FXnbc*sizeof(int));
      FXsepind   = malloc(FXnbc*sizeof(int));
      if ((FXshowSurf == NULL) || (FXattind == NULL) || (FXsepind == NULL)) {
        if (FXshowSurf != NULL) free(FXshowSurf);
        if (FXattind   != NULL) free(FXattind);
        if (FXsepind   != NULL) free(FXsepind);
        printf(" Error in Memory Allocation for %d Surface Storage\n\n", FXnbc);
        return;
      }
    }

    gamma = 1.4;
    FX_INIT(&gamma, &iopt, &nnode, &zero, &ntets, &npyra, &nprism, &nhexa,
            &nblocks, blocks, &zero, &nsurf, &FXnbc, &istat);
    printf(" FX_Init return = %d\n\n", istat);
    if (nblocks != 0) free(blocks);
    if (iopt != 0) FX_UPDATE(&ltime);
  }

  /* unsteady options */

  if (iopt != 0) {
    i = -2;
    V3_GETSTATE(&i, ivec, rvec, cstr, 80);
    if (ltime != rvec[0]) FX_UPDATE(rvec);
    ltime = rvec[0];
  }

  /* 2D window additions */

  if (*win == winid[1]) {
    if ((*type == 4) && (*state == 2)) {
      XFTNSDIALFUN(win, xe, ye, state, &lineout);
      *win = 0;
    }
    if ((*type == 5) && (*state == 2)) {
      XFTNSDIALFUN(win, xe, ye, &zero, &lineout);
      *win = 0;
    }
    /* replace middle mouse with "s" */
    if ((*type == 2) && (*state == 's')) {
      *type  = 4;
      *state = 2;
    }
  }

  /* 3D window additions */

  if (*win == winid[2]) {
    if ((*type == 4) && (*state == 2)) {
      i = 3;
      XFTNSDIALFUN(win, xe, ye, &i, &lineout);
      *win = 0;
    }
    if ((*type == 5) && (*state == 2)) {
      XFTNSDIALFUN(win, xe, ye, &zero, &lineout);
      *win = 0;
    }

    /* Votrex core finder -- V */

    if ((*type == 2) && (*state == 'V')) {
      if (FXvc == 1) {
        printf(" Votrex Core Off!\n\n");
        FXvc = 0;
        return;
      }
      FXvc = 1;
      if ((iopt == 0) && (FXnseg != 0)) {
        printf(" Votrex Core On!\n\n");
        return;
      }
      if (FXnseg != 0) {
        free(FXvcoor);
        free(FXsegs);
        free(FXstren);
      }

      /* use vorticity */
      if (iopt == 0) printf(" Computing Vortex Cores...\n");
      i = 1;
      FX_VORTEXCORE(&i, &FXnseg, &FXsegs, &FXvcoor, &FXstren);
      if (iopt == 0) printf(" ...Votrex Core On!\n\n");
    }

    /* Separation and Attachment Line finder -- L  */

    if ((*type == 2) && (*state == 'L')) {
      if (FXsl == 1) {
        FXsl = 0;
        printf(" Separation and Attachment Lines Off!\n\n");
        return;
      }
      FXsl = 1;
      /* set to current surface */
      i = 10;
      V3_GETSTATE(&i, ivec, rvec, cstr, 80);
      if (ivec[0] > FXnbc) ivec[0] = 1;
      for (i = 0; i < FXnbc; i++) FXshowSurf[i] = 0;
      FXshowSurf[ivec[0]-1] = 1;

      if ((iopt == 0) && (FXsepxyz != NULL)) {
        printf(" Separation and Attachment Lines On!\n\n");
        return;
      }
      if (FXsepxyz != NULL) {
        free(FXsepxyz);
        free(FXattxyz);
      }

      if (iopt == 0) printf(" Computing Separation and Attachment Lines...\n");
      FX_SEPNLINE(FXsepind, &FXsepxyz, FXattind, &FXattxyz);
      if (iopt == 0) printf(" ...Separation and Attachment Lines On!\n\n");
    }

    /* shock finder -- S */

    if ((*type == 2) && (*state == 'S')) {
      if (FXsf == 1) {
        printf(" Shocks Off!\n\n");
        FXsf = 0;
        return;
      }
      FXsf = 1;
      if ((iopt == 0) && (FXnpts != 0)) {
        printf(" Shocks On!\n\n");
        return;
      }
      if (FXnpts != 0) {
        free(FXcoors);
        FXnpts = 0;
      }

      if (iopt == 0) printf(" Computing Shocks...\n");
      i = 301;
      V3_GETSTRUC(&i, (void **) &xyz, &nnode);
      press = (float *) malloc(nnode*sizeof(float));
      if (press == NULL) {
        printf(" Malloc error for pressure\n");
        return;
      }
      i = 2;
      V3SCAL(&i, press);
      
      FX_SHOCKFIND(&test);
      if (test != NULL) {
        FX_SHOCKSURFACE(test, &npts, &coors, &ntris, &tris, &cells);

        /* get individual volumes */

        FX_SHOCKVOLUMES(test, &nclust, &vols, &clust, &cell, &cnode, &cnodes);
        if (nclust > 0) {
          mark_tris(nclust, clust, cell, cnode, cnodes, press, ntris, cells);
          if (vols != NULL) free(vols);
          free(cnode);
          free(cnodes);
          free(clust);
          free(cell);
        }
        for (j = i = 0; i < ntris; i++) if (cells[i] > 0) j++;
        FXnpts = 3*j;
        if (FXnpts != 0) {
          FXcoors = (float *) malloc(3*FXnpts*sizeof(float));
          if (FXcoors == NULL) {
            FXnpts = 0;
          } else {
            for (j = i = 0; i < ntris; i++) {
              if (cells[i] < 0) continue;
              k = tris[3*i  ]-1;
              FXcoors[j] = coors[3*k  ];
              j++;
              FXcoors[j] = coors[3*k+1];
              j++;
              FXcoors[j] = coors[3*k+2];
              j++;
              k = tris[3*i+1]-1;
              FXcoors[j] = coors[3*k  ];
              j++;
              FXcoors[j] = coors[3*k+1];
              j++;
              FXcoors[j] = coors[3*k+2];
              j++;
              k = tris[3*i+2]-1;
              FXcoors[j] = coors[3*k  ];
              j++;
              FXcoors[j] = coors[3*k+1];
              j++;
              FXcoors[j] = coors[3*k+2];
              j++;
            }
          }
        }
        free(test);
        free(coors);
        free(tris);
        free(cells);
      }
      free(press);
      if (iopt == 0) printf(" ...Shocks On!\n\n");
    }
  }

  /* Dial window additions */

  if (*win == winid[3]) {
    if ((*type == 2) && (*state == 'O')) {
      lineout++;
      if (lineout >  1) lineout = 0;
      if (lineout == 0) {
        printf("Outline mode: Off!\n");
      } else {
        printf("Outline mode: On!\n");
      }
      printf("\n");
    }
  }

}


/* FX system call-backs */

void PREFIX
FXCELLPTR(int **tets, int **pyras, int **prisms, int **hexas, int *halo)
{
  int i, len;

  i = 311;
  V3_GETSTRUC(&i, (void **) tets,   &len);
  i = 312;
  V3_GETSTRUC(&i, (void **) pyras,  &len);
  i = 313;
  V3_GETSTRUC(&i, (void **) prisms, &len);
  i = 314;
  V3_GETSTRUC(&i, (void **) hexas,  &len);
}


void PREFIX
FXSURFACE(int *nsurf, int *cell, int *facet)
{
  int i, len, nbc, *bc, *scel, *surf;

  i = 324;
  V3_GETSTRUC(&i, (void **) &scel, &len);
  i = 323;
  V3_GETSTRUC(&i, (void **) &surf, &len);
  for (i = 0; i <   len; i++) cell[i]  = surf[i]/8;
  for (i = 0; i < 4*len; i++) facet[i] = scel[i];

  i = 321;
  V3_GETSTRUC(&i, (void **) &bc,   &nbc);
  for (i = 0; i < nbc; i++) {
    nsurf[2*i  ] = bc[2*i];
    nsurf[2*i+1] = 6;
  }
}


void PREFIX
FXGRIDPTR(float **xyz, float *hxyz)
{
  int i, len;

  i = 301;
  V3_GETSTRUC(&i, (void **) xyz, &len);
}


void PREFIX
FXVELPTR(float **vel, float *hvel)
{
  int i, len;

  i = 304;
  V3_GETSTRUC(&i, (void **) vel, &len);
}


void PREFIX
FXSCAL(int *type, float *s, float *hs)
{
  int   i, n;
  float *xyz;

  if ((*type <= 5) && (*type != 4)) {
    V3SCAL(type, s);
  } else {
    printf(" FXSCAL Error: type = %d\n", *type);
    i = 301;
    V3_GETSTRUC(&i, (void **) &xyz, &n);
    for (i = 0; i < n; i++) s[i] = 0.0;
  }
}


/* Visual3 drawing routine for additional objects */

void PREFIX
V3DRAW3D(int *istat)
{
  int   i, j, m, n, attIndex, sepIndex, nb, djseg;
  int   itype = 1, icolor = 0, itlnseg = 3, itline = 4, white = -1, grey = 0;
  float dummy[2];

  if (*istat != 0) return;

  /* drawing for the vortex cores */

  if (FXvc == 1)
    for (m = i = 0; i < FXnseg; i++) {
      n = FXsegs[i] - m;
      V3_OBJECT3D(&itline, &white, &FXvcoor[3*m], dummy, dummy, &j, &n);
      m = FXsegs[i];
    }

  /* drawing for the separation and attachment lines */

  if ((FXsl == 1) && (FXsepxyz != NULL)) {
    for (nb = 0; nb < FXnbc; nb++) {
      if (nb == 0) {
	attIndex = 0;
        sepIndex = 0;
      } else {
	attIndex = FXattind[nb-1];
        sepIndex = FXsepind[nb-1];
      }

      if (FXshowSurf[nb] == 1) {

        /* Draw the separation lines */
        djseg = (FXsepind[nb]-sepIndex);
        if (djseg > 0)
          V3_OBJECT3D(&itlnseg, &white, &FXsepxyz[3*sepIndex], dummy, dummy,
                      &j, &djseg);

        /* Draw the attachment lines */
        djseg = (FXattind[nb]-attIndex);
        if (djseg > 0)
          V3_OBJECT3D(&itlnseg, &grey,  &FXattxyz[3*attIndex], dummy, dummy,
                      &j, &djseg);
      }

    }
  }

  /* draw the shock regions */

  if ((FXsf == 1) && (FXnpts != 0))
    V3_OBJECT3D(&itype, &icolor, FXcoors, dummy, dummy, &j, &FXnpts);

}
