/*
 * FXpV3.c
 *
 * FX example module for pV3  Rev 2
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "pV3.h"
#include "FX.h"



#ifdef PREFIX
#undef PREFIX
#endif


#ifdef WIN32

#define PREFIX __stdcall
#define pVSSINIT PVSSINIT

#else

#define PREFIX

#ifdef UNDERSCORE
#define pVSSINIT pvssinit_
#else
#define pVSSINIT pvssinit
#endif

#endif


  /* globals */

  static int   xlate[5] = {1, 4, 3, 0, 12};	/* translate pV3 Scalars to FX Scalars */

  static int   *FXsepind, *FXattind, *FXshowSurf, FXnbc;
  static int   case_opt;


  /* externals */

  extern void PREFIX pVSCAL(int *type, float *s);



  /*  take the shock regions and remove those that are small or do not
   *    display a large enough pressure gradient
   */

static void
mark_tris(int nclust, int *clust, int *cell, int *cnode, int *cnodes,
          float *press, int ntris, int *cells, int minode, float mindp)
{
  char  *cellTable;
  int   i, j, k, l, clen, nlen, max;
  float minp, maxp;

  clen = clust[nclust-1];
  for (max = i = 0; i < clen; i++) if (cell[i] > max) max = cell[i];

  cellTable = (char *) malloc(max*sizeof(char));
  if (cellTable == NULL) {
    printf(" Cannot Allocate cellTable (mark_tris)\n");
    return;
  }

  for (i = 0; i < max; i++) cellTable[i] = 0;

  /* mark the cells that are OK */

  for (k = j = i = 0; i < nclust; i++) {
    clen = clust[i] - j;
    nlen = cnode[i] - k;
    if (nlen > minode) {
      minp = maxp = press[cnodes[k]-1];
      for (l = 1; l < nlen; l++) {
        if (minp > press[cnodes[k+l]-1]) minp = press[cnodes[k+l]-1];
        if (maxp < press[cnodes[k+l]-1]) maxp = press[cnodes[k+l]-1];
      }
       /* maxp-minp measures the approximate shock strength for the region */
      if ((maxp-minp) > mindp) 
        for (l = j; l < clust[i]; l++) cellTable[cell[l]-1] = 1;
    }
    j = clust[i];
    k = cnode[i];
  }

  /* mark the not-OK triangles */

  for (i = 0; i < ntris; i++)
    if (cellTable[cells[i]-1] == 0) cells[i] = -cells[i];

  free(cellTable);
}


/* Initialization */

static void
init_fx(int iopt)
{
  int   i, nnode, ntets, npyra, nprism, nhexa, nblocks, *blocks, nsurf;
  int   istat, len, *tets, *pyra, *prism, *hexa, *blk, *iblk, zero = 0;
  int   *bc, *scel;
  float gamma, *xyz;

  case_opt = iopt;

  i = 315;
  pV_GETSTRUC(&i, (void **) &tets,  &len);
  if (len != 0) {
    printf(" This Case has %d PolyTet Strips \n\n", len);
    return;
  }
  i = 318;
  pV_GETSTRUC(&i, (void **) &tets,  &len);
  if (len != 0) {
    printf(" This Case has %d Complex Polyhedra \n\n", len);
    return;
  }

  istat = 47;
  i = 301;
  pV_GETSTRUC(&i, (void **) &xyz,   &nnode);

  i = 311;
  pV_GETSTRUC(&i, (void **) &tets,  &ntets);
  i = 312;
  pV_GETSTRUC(&i, (void **) &pyra,  &npyra);
  i = 313;
  pV_GETSTRUC(&i, (void **) &prism, &nprism);
  i = 314;
  pV_GETSTRUC(&i, (void **) &hexa,  &nhexa);
  i = 317;
  pV_GETSTRUC(&i, (void **) &blk,   &nblocks);
  if (nblocks != 0) {
    blocks = (int *) malloc(3*nblocks*sizeof(int));
    if (blocks == NULL) {
      printf(" Error in Memory Allocation for %d Blocks\n\n", nblocks);
      return;
    }
    for (i = 0; i < nblocks; i++) {
      blocks[3*i  ] = blk[6*i  ];
      blocks[3*i+1] = blk[6*i+1];
      blocks[3*i+2] = blk[6*i+2];
      nnode -= blocks[3*i]*blocks[3*i+1]*blocks[3*i+2];
    }
    i = 306;
    pV_GETSTRUC(&i, (void **) &iblk,  &len);
    if (len != 0) istat += 16;
  }

  i = 321;
  pV_GETSTRUC(&i, (void **) &bc,   &FXnbc);
  i = 323;
  pV_GETSTRUC(&i, (void **) &scel, &nsurf);
  if (FXnbc > 0) {
    FXshowSurf = malloc(FXnbc*sizeof(int));
    FXattind   = malloc(FXnbc*sizeof(int));
    FXsepind   = malloc(FXnbc*sizeof(int));
    if ((FXshowSurf == NULL) || (FXattind == NULL) || (FXsepind == NULL)) {
      if (FXshowSurf != NULL) free(FXshowSurf);
      if (FXattind   != NULL) free(FXattind);
      if (FXsepind   != NULL) free(FXsepind);
      printf(" Error in Memory Allocation for %d Surface Storage\n\n", FXnbc);
      if (nblocks != 0) free(blocks);
      return;
    }
  }

  gamma = 1.4;
  FX_INIT(&gamma, &iopt, &nnode, &zero, &ntets, &npyra, &nprism, &nhexa,
          &nblocks, blocks, &zero, &nsurf, &FXnbc, &istat);
  printf(" FX_Init return = %d\n\n", istat);
  if (nblocks != 0) free(blocks);
}


void PREFIX
pVSSINIT()
{
  init_fx(0);
}


void PREFIX
pVUPDATE(int *status)
{
  int    i, iopt;
  float  *time;
  static int count = 0;

  /* use undocumented method to get iopt and time */
  i = 0;
  pV_GETSTRUC(&i, (void **) &time,  &iopt);
  if (iopt == 3) {
    printf(" This Structure Unsteady Case!\n\n");
    return;
  }
  if ((count == 0) && (*status == 0)) init_fx(iopt);
  count++;

  FX_UPDATE(time);
}


static void
FXpV3Shock(int index, int exnum, int minode, float mindp)
{
  int   i, j, k, nnode, nclust, *clust, *cell, *cnode, *cnodes;
  int   npts, ntris, *tris, *cells, zero = 0, three = 3;
  float *xyz, *coors, *press, *test, *vols, *pts;

  i = 301;
  pV_GETSTRUC(&i, (void **) &xyz, &nnode);
  press = (float *) malloc(nnode*sizeof(float));
  if (press == NULL) {
    printf(" Malloc error for pressure\n");
    i = 1;
    pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
    return;
  }
  i = xlate[1];
  pVSCAL(&i, press);
  
  npts = 0;
  FX_SHOCKFIND(&test);
  if (test != NULL) {
    FX_SHOCKSURFACE(test, &npts, &coors, &ntris, &tris, &cells);

    /* get individual volumes */

    FX_SHOCKVOLUMES(test, &nclust, &vols, &clust, &cell, &cnode, &cnodes);
    if (nclust > 0) {
      mark_tris(nclust, clust, cell, cnode, cnodes, press, ntris, cells,
                minode, mindp);
      if (vols != NULL) free(vols);
      free(cnode);
      free(cnodes);
      free(clust);
      free(cell);
    }
    /* save away only the surfaces for OK regions */
    for (j = i = 0; i < ntris; i++) if (cells[i] > 0) j++;
    npts = 3*j;
    if (npts != 0) {
      pts = (float *) malloc(3*npts*sizeof(float));
      if (pts == NULL) {
        npts = 0;
      } else {
        for (j = i = 0; i < ntris; i++) {
          if (cells[i] < 0) continue;
          k = tris[3*i  ]-1;
          pts[j] = coors[3*k  ];
          j++;
          pts[j] = coors[3*k+1];
          j++;
          pts[j] = coors[3*k+2];
          j++;
          k = tris[3*i+1]-1;
          pts[j] = coors[3*k  ];
          j++;
          pts[j] = coors[3*k+1];
          j++;
          pts[j] = coors[3*k+2];
          j++;
          k = tris[3*i+2]-1;
          pts[j] = coors[3*k  ];
          j++;
          pts[j] = coors[3*k+1];
          j++;
          pts[j] = coors[3*k+2];
          j++;
        }
      }
    }
    free(test);
    free(coors);
    free(tris);
    free(cells);
  }
  free(press);
  if (npts != 0) {
    pV_SENDXR(&index, &exnum, &zero, &three, &npts, pts);
    free(pts);
  } else {
    i = 1;
    pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
  }
}


/* FX system call-backs */

void PREFIX
FXCELLPTR(int **tets, int **pyras, int **prisms, int **hexas, int *halo)
{
  int i, len;

  i = 311;
  pV_GETSTRUC(&i, (void **) tets,   &len);
  i = 312;
  pV_GETSTRUC(&i, (void **) pyras,  &len);
  i = 313;
  pV_GETSTRUC(&i, (void **) prisms, &len);
  i = 314;
  pV_GETSTRUC(&i, (void **) hexas,  &len);
}


void PREFIX
FXSURFACEPTR(int *nsurf, int **cell, int **facet)
{
  int i, len, *bc;

  i = 324;
  pV_GETSTRUC(&i, (void **) facet, &len);
  i = 323;
  pV_GETSTRUC(&i, (void **) cell, &len);
  i = 321;
  pV_GETSTRUC(&i, (void **) &bc,  &len);

  for (i = 0; i < len; i++) {
    nsurf[2*i  ] = bc[3*i];
    nsurf[2*i+1] = 6;
  }
}


void PREFIX
FXGRIDPTR(float **xyz, float *hxyz)
{
  int i, len;

  i = 301;
  pV_GETSTRUC(&i, (void **) xyz, &len);
}


void PREFIX
FXVELPTR(float **vel, float *hvel)
{
  int i, len;

  i = 304;
  pV_GETSTRUC(&i, (void **) vel, &len);
}


void PREFIX
FXSCAL(int *type, float *s, float *hs)
{
  int i;

  if ((*type <= 5) && (*type != 4)) {
    i = xlate[*type-1];
    pVSCAL(&i, s);
  } else {
    printf(" FXSCAL Error: type = %d\n", *type);
  }
}


/* pV3 Programmed Extract communication */

void PREFIX
pVEXTRACT(int *index, int *exnum, int *reqmask, int *ival, float *rvals)
{
  int   i, ibc, *bc, nseg, *segs, zero = 0, one = 1, three = 3;
  float *sepxyz, *attxyz, *vcoor, *stren, fone[3] = {1.0,1.0,1.0};

  if (*index == 101) {
    FX_VORTEXCORE(ival, &nseg, &segs, &vcoor, &stren);
    if (nseg != 0) {
      pV_SENDXR(index, exnum, &zero, &three, &segs[nseg-1], vcoor);
      pV_SENDXI(index, exnum, &one,  &one,   &nseg,         segs);
      free(segs);
      free(vcoor);
      free(stren);
    } else {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXI(index, exnum, &one,  &one,   &one, &one);
    }
    return;
  }

  if (*index == 102) {
    FXpV3Shock(*index, *exnum, *ival, rvals[0]);
    return;
  }

  if (*index == 103) {
    FX_SEPNLINE(FXsepind, &sepxyz, FXattind, &attxyz);

    if (sepxyz == NULL) {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
      return;
    }
    i = 321;
    pV_GETSTRUC(&i, (void **) &bc, &ibc);
    ibc = -1;
    for (i = 0; i < FXnbc; i++)
      if (bc[3*i+2] == *ival) ibc = i;

    if (ibc == -1) {
      printf(" pVExtract: Global Surface # %d NOT Found!\n", *ival);
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
      return;
    }

    i = 0;
    if (ibc > 0) i = FXsepind[ibc-1];
    nseg = FXsepind[ibc] - i;
    if (nseg > 1) {
      pV_SENDXR(index, exnum, &zero, &three, &nseg, &sepxyz[3*i]);
    } else {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
    }

    i = 0;
    if (ibc > 0) i = FXattind[ibc-1];
    nseg = FXattind[ibc] - i;
    if (nseg > 1) {
      pV_SENDXR(index, exnum, &one,  &three, &nseg, &attxyz[3*i]);
    } else {
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
    }
    free(sepxyz);
    free(attxyz);
      
    return;
  }

  printf(" Unknown Programmed Extract request = %d\n", *index);
}
