/*
 * FX_DivSubs.c
 *
 * FX modules that calculate divergence.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>


#include "FX_types.h"


#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXCELLGET fxcellget_

#else

#define FXCELLGET fxcellget

#endif
#endif

  extern void PREFIX FXCELLGET (int *, int *, int *);



/*
 *  Computes and returns the gradient of src
 */

static void
FX_TetDivPtr(int *node, FXcoor *coor, float *src, float *divg, float *vol)
{
  int    i1, i2, i3, i4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, cx, cy, cz, div;
  double u1, v1, w1, u2, v2, w2, u3, v3, w3, u4, v4, w4;

  i1 = node[0]-1;
  x1 = coor[i1].x;
  y1 = coor[i1].y;
  z1 = coor[i1].z;
  u1 = src[3*i1  ];
  v1 = src[3*i1+1];
  w1 = src[3*i1+2];

  i2 = node[1]-1;
  x2 = coor[i2].x;
  y2 = coor[i2].y;
  z2 = coor[i2].z;
  u2 = src[3*i2  ];
  v2 = src[3*i2+1];
  w2 = src[3*i2+2];

  i3 = node[2]-1;
  x3 = coor[i3].x;
  y3 = coor[i3].y;
  z3 = coor[i3].z;
  u3 = src[3*i3  ];
  v3 = src[3*i3+1];
  w3 = src[3*i3+2];

  i4 = node[3]-1;
  x4 = coor[i4].x;
  y4 = coor[i4].y;
  z4 = coor[i4].z;
  u4 = src[3*i4  ];
  v4 = src[3*i4+1];
  w4 = src[3*i4+2];

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  vol[i1] += V;
  vol[i2] += V;
  vol[i3] += V;
  vol[i4] += V;

  cx   = (y1-y2)*(z3-z2)-(z1-z2)*(y3-y2);
  cy   = (z1-z2)*(x3-x2)-(x1-x2)*(z3-z2);
  cz   = (x1-x2)*(y3-y2)-(y1-y2)*(x3-x2);
  div  = (u2+u1+u3)*cx + (v2+v1+v3)*cy + (w2+w1+w3)*cz;

  cx   = (y3-y2)*(z4-z2)-(z3-z2)*(y4-y2);
  cy   = (z3-z2)*(x4-x2)-(x3-x2)*(z4-z2);
  cz   = (x3-x2)*(y4-y2)-(y3-y2)*(x4-x2);
  div += (u2+u3+u4)*cx + (v2+v3+v4)*cy + (w2+w3+w4)*cz;

  cx   = (y1-y3)*(z4-z3)-(z1-z3)*(y4-y3);
  cy   = (z1-z3)*(x4-x3)-(x1-x3)*(z4-z3);
  cz   = (x1-x3)*(y4-y3)-(y1-y3)*(x4-x3);
  div += (u3+u1+u4)*cx + (v3+v1+v4)*cy + (w3+w1+w4)*cz;

  cx   = (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1);
  cy   = (z2-z1)*(x4-x1)-(x2-x1)*(z4-z1);
  cz   = (x2-x1)*(y4-y1)-(y2-y1)*(x4-x1);
  div += (u1+u2+u4)*cx + (v1+v2+v4)*cy + (w1+w2+w4)*cz;

  divg[i1] += div;
  divg[i2] += div;
  divg[i3] += div;
  divg[i4] += div;
}


static void
FX_TetDivP3d(int *node, float *xcoor, float *ycoor, float *zcoor,
             float *src, float *divg, float *vol)
{
  int    i1, i2, i3, i4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, cx, cy, cz, div;
  double u1, v1, w1, u2, v2, w2, u3, v3, w3, u4, v4, w4;

  i1 = node[0]-1;
  x1 = xcoor[i1];
  y1 = ycoor[i1];
  z1 = zcoor[i1];
  u1 = src[3*i1  ];
  v1 = src[3*i1+1];
  w1 = src[3*i1+2];

  i2 = node[1]-1;
  x2 = xcoor[i2];
  y2 = ycoor[i2];
  z2 = zcoor[i2];
  u2 = src[3*i2  ];
  v2 = src[3*i2+1];
  w2 = src[3*i2+2];

  i3 = node[2]-1;
  x3 = xcoor[i3];
  y3 = ycoor[i3];
  z3 = zcoor[i3];
  u3 = src[3*i3  ];
  v3 = src[3*i3+1];
  w3 = src[3*i3+2];

  i4 = node[3]-1;
  x4 = xcoor[i4];
  y4 = ycoor[i4];
  z4 = zcoor[i4];
  u4 = src[3*i4  ];
  v4 = src[3*i4+1];
  w4 = src[3*i4+2];

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  vol[i1] += V;
  vol[i2] += V;
  vol[i3] += V;
  vol[i4] += V;

  cx   = (y1-y2)*(z3-z2)-(z1-z2)*(y3-y2);
  cy   = (z1-z2)*(x3-x2)-(x1-x2)*(z3-z2);
  cz   = (x1-x2)*(y3-y2)-(y1-y2)*(x3-x2);
  div  = (u2+u1+u3)*cx + (v2+v1+v3)*cy + (w2+w1+w3)*cz;

  cx   = (y3-y2)*(z4-z2)-(z3-z2)*(y4-y2);
  cy   = (z3-z2)*(x4-x2)-(x3-x2)*(z4-z2);
  cz   = (x3-x2)*(y4-y2)-(y3-y2)*(x4-x2);
  div += (u2+u3+u4)*cx + (v2+v3+v4)*cy + (w2+w3+w4)*cz;

  cx   = (y1-y3)*(z4-z3)-(z1-z3)*(y4-y3);
  cy   = (z1-z3)*(x4-x3)-(x1-x3)*(z4-z3);
  cz   = (x1-x3)*(y4-y3)-(y1-y3)*(x4-x3);
  div += (u3+u1+u4)*cx + (v3+v1+v4)*cy + (w3+w1+w4)*cz;

  cx   = (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1);
  cy   = (z2-z1)*(x4-x1)-(x2-x1)*(z4-z1);
  cz   = (x2-x1)*(y4-y1)-(y2-y1)*(x4-x1);
  div += (u1+u2+u4)*cx + (v1+v2+v4)*cy + (w1+w2+w4)*cz;

  divg[i1] += div;
  divg[i2] += div;
  divg[i3] += div;
  divg[i4] += div;
}


static void
FX_TetDiv(FXglobal *globals, int *node, float *src, float *divg, float *vol)
{

  if ((globals->cmask & 64) == 0) {
    FX_TetDivPtr(node, globals->coor, src, divg, vol);
  } else {
    FX_TetDivP3d(node, globals->xcoor, globals->ycoor, globals->zcoor, src,
                 divg, vol);
  }

}


/*
 *  Computes and returns the divergence of src with halo nodes
 */

static void
FX_HTetDivPtr(int *node, int n, FXcoor *coor, FXcoor *hcoor,
              float *src, float *hsrc, float *divg, float *vol)
{
  int    i1, i2, i3, i4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, cx, cy, cz, div;
  double u1, v1, w1, u2, v2, w2, u3, v3, w3, u4, v4, w4;

  i1 = node[0]-1;
  if (i1 < n) {
    x1 = coor[i1].x;
    y1 = coor[i1].y;
    z1 = coor[i1].z;
    u1 = src[3*i1  ];
    v1 = src[3*i1+1];
    w1 = src[3*i1+2];
  } else {
    x1 = hcoor[i1-n].x;
    y1 = hcoor[i1-n].y;
    z1 = hcoor[i1-n].z;
    u1 = hsrc[3*(i1-n)  ];
    v1 = hsrc[3*(i1-n)+1];
    w1 = hsrc[3*(i1-n)+2];
  }
  i2 = node[1]-1;
  if (i2 < n) {
    x2 = coor[i2].x;
    y2 = coor[i2].y;
    z2 = coor[i2].z;
    u2 = src[3*i2  ];
    v2 = src[3*i2+1];
    w2 = src[3*i2+2];
  } else {
    x2 = hcoor[i2-n].x;
    y2 = hcoor[i2-n].y;
    z2 = hcoor[i2-n].z;
    u2 = hsrc[3*(i2-n)  ];
    v2 = hsrc[3*(i2-n)+1];
    w2 = hsrc[3*(i2-n)+2];
  }
  i3 = node[2]-1;
  if (i3 < n) {
    x3 = coor[i3].x;
    y3 = coor[i3].y;
    z3 = coor[i3].z;
    u3 = src[3*i3  ];
    v3 = src[3*i3+1];
    w3 = src[3*i3+2];
  } else {
    x3 = hcoor[i3-n].x;
    y3 = hcoor[i3-n].y;
    z3 = hcoor[i3-n].z;
    u3 = hsrc[3*(i3-n)  ];
    v3 = hsrc[3*(i3-n)+1];
    w3 = hsrc[3*(i3-n)+2];
  }
  i4 = node[3]-1;
  if (i4 < n) {
    x4 = coor[i4].x;
    y4 = coor[i4].y;
    z4 = coor[i4].z;
    u4 = src[3*i4  ];
    v4 = src[3*i4+1];
    w4 = src[3*i4+2];
  } else {
    x4 = hcoor[i4-n].x;
    y4 = hcoor[i4-n].y;
    z4 = hcoor[i4-n].z;
    u4 = hsrc[3*(i4-n)  ];
    v4 = hsrc[3*(i4-n)+1];
    w4 = hsrc[3*(i4-n)+2];
  }

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  if (i1 < n) vol[i1] += V;
  if (i2 < n) vol[i2] += V;
  if (i3 < n) vol[i3] += V;
  if (i4 < n) vol[i4] += V;

  cx   = (y1-y2)*(z3-z2)-(z1-z2)*(y3-y2);
  cy   = (z1-z2)*(x3-x2)-(x1-x2)*(z3-z2);
  cz   = (x1-x2)*(y3-y2)-(y1-y2)*(x3-x2);
  div  = (u2+u1+u3)*cx + (v2+v1+v3)*cy + (w2+w1+w3)*cz;

  cx   = (y3-y2)*(z4-z2)-(z3-z2)*(y4-y2);
  cy   = (z3-z2)*(x4-x2)-(x3-x2)*(z4-z2);
  cz   = (x3-x2)*(y4-y2)-(y3-y2)*(x4-x2);
  div += (u2+u3+u4)*cx + (v2+v3+v4)*cy + (w2+w3+w4)*cz;

  cx   = (y1-y3)*(z4-z3)-(z1-z3)*(y4-y3);
  cy   = (z1-z3)*(x4-x3)-(x1-x3)*(z4-z3);
  cz   = (x1-x3)*(y4-y3)-(y1-y3)*(x4-x3);
  div += (u3+u1+u4)*cx + (v3+v1+v4)*cy + (w3+w1+w4)*cz;

  cx   = (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1);
  cy   = (z2-z1)*(x4-x1)-(x2-x1)*(z4-z1);
  cz   = (x2-x1)*(y4-y1)-(y2-y1)*(x4-x1);
  div += (u1+u2+u4)*cx + (v1+v2+v4)*cy + (w1+w2+w4)*cz;

  if (i1 < n) divg[i1] += div;
  if (i2 < n) divg[i2] += div;
  if (i3 < n) divg[i3] += div;
  if (i4 < n) divg[i4] += div;
}


static void
FX_HTetDivP3d(int *node, int n, float *xcoor, float *ycoor, float *zcoor,
              FXcoor *hcoor, float *src, float *hsrc, float *divg, float *vol)
{
  int    i1, i2, i3, i4;
  double x1, y1, z1, x2, y2, z2, x3, y3, z3, x4, y4, z4, V, cx, cy, cz, div;
  double u1, v1, w1, u2, v2, w2, u3, v3, w3, u4, v4, w4;

  i1 = node[0]-1;
  if (i1 < n) {
    x1 = xcoor[i1];
    y1 = ycoor[i1];
    z1 = zcoor[i1];
    u1 = src[3*i1  ];
    v1 = src[3*i1+1];
    w1 = src[3*i1+2];
  } else {
    x1 = hcoor[i1-n].x;
    y1 = hcoor[i1-n].y;
    z1 = hcoor[i1-n].z;
    u1 = hsrc[3*(i1-n)  ];
    v1 = hsrc[3*(i1-n)+1];
    w1 = hsrc[3*(i1-n)+2];
  }
  i2 = node[1]-1;
  if (i2 < n) {
    x2 = xcoor[i2];
    y2 = ycoor[i2];
    z2 = zcoor[i2];
    u2 = src[3*i2  ];
    v2 = src[3*i2+1];
    w2 = src[3*i2+2];
  } else {
    x2 = hcoor[i2-n].x;
    y2 = hcoor[i2-n].y;
    z2 = hcoor[i2-n].z;
    u2 = hsrc[3*(i2-n)  ];
    v2 = hsrc[3*(i2-n)+1];
    w2 = hsrc[3*(i2-n)+2];
  }
  i3 = node[2]-1;
  if (i3 < n) {
    x3 = xcoor[i3];
    y3 = ycoor[i3];
    z3 = zcoor[i3];
    u3 = src[3*i3  ];
    v3 = src[3*i3+1];
    w3 = src[3*i3+2];
  } else {
    x3 = hcoor[i3-n].x;
    y3 = hcoor[i3-n].y;
    z3 = hcoor[i3-n].z;
    u3 = hsrc[3*(i3-n)  ];
    v3 = hsrc[3*(i3-n)+1];
    w3 = hsrc[3*(i3-n)+2];
  }
  i4 = node[3]-1;
  if (i4 < n) {
    x4 = xcoor[i4];
    y4 = ycoor[i4];
    z4 = zcoor[i4];
    u4 = src[3*i4  ];
    v4 = src[3*i4+1];
    w4 = src[3*i4+2];
  } else {
    x4 = hcoor[i4-n].x;
    y4 = hcoor[i4-n].y;
    z4 = hcoor[i4-n].z;
    u4 = hsrc[3*(i4-n)  ];
    v4 = hsrc[3*(i4-n)+1];
    w4 = hsrc[3*(i4-n)+2];
  }

  V = (x2-x1)*( (y3-y1)*(z4-z1)-(z3-z1)*(y4-y1) ) -
      (x3-x1)*( (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1) ) +
      (x4-x1)*( (y2-y1)*(z3-z1)-(z2-z1)*(y3-y1) );
  if (i1 < n) vol[i1] += V;
  if (i2 < n) vol[i2] += V;
  if (i3 < n) vol[i3] += V;
  if (i4 < n) vol[i4] += V;

  cx   = (y1-y2)*(z3-z2)-(z1-z2)*(y3-y2);
  cy   = (z1-z2)*(x3-x2)-(x1-x2)*(z3-z2);
  cz   = (x1-x2)*(y3-y2)-(y1-y2)*(x3-x2);
  div  = (u2+u1+u3)*cx + (v2+v1+v3)*cy + (w2+w1+w3)*cz;

  cx   = (y3-y2)*(z4-z2)-(z3-z2)*(y4-y2);
  cy   = (z3-z2)*(x4-x2)-(x3-x2)*(z4-z2);
  cz   = (x3-x2)*(y4-y2)-(y3-y2)*(x4-x2);
  div += (u2+u3+u4)*cx + (v2+v3+v4)*cy + (w2+w3+w4)*cz;

  cx   = (y1-y3)*(z4-z3)-(z1-z3)*(y4-y3);
  cy   = (z1-z3)*(x4-x3)-(x1-x3)*(z4-z3);
  cz   = (x1-x3)*(y4-y3)-(y1-y3)*(x4-x3);
  div += (u3+u1+u4)*cx + (v3+v1+v4)*cy + (w3+w1+w4)*cz;

  cx   = (y2-y1)*(z4-z1)-(z2-z1)*(y4-y1);
  cy   = (z2-z1)*(x4-x1)-(x2-x1)*(z4-z1);
  cz   = (x2-x1)*(y4-y1)-(y2-y1)*(x4-x1);
  div += (u1+u2+u4)*cx + (v1+v2+v4)*cy + (w1+w2+w4)*cz;

  if (i1 < n) divg[i1] += div;
  if (i2 < n) divg[i2] += div;
  if (i3 < n) divg[i3] += div;
  if (i4 < n) divg[i4] += div;
}


static void
FX_HTetDiv(FXglobal *globals, int *node, float *src, float *hsrc, 
           float *divg, float *vol)
{

  if ((globals->cmask & 64) == 0) {
    FX_HTetDivPtr(node, globals->total_nodes, globals->coor, globals->hcoor,
                  src, hsrc, divg, vol);
  } else {
    FX_HTetDivP3d(node, globals->total_nodes, globals->xcoor, globals->ycoor,
                  globals->zcoor, globals->hcoor, src, hsrc, divg, vol);
  }

}


/*
 *  calculates the divergence
 */

void
FX_Div(FXglobal *globals, float *vec, float *hvec, float **divg)
{
  int    kc, i, j, k, ib, ni, nj, nk, nc, kns, kz, bnode, type;
  int    cell[8], node[8], tet[4];
  float  *vol, *sdivg;
  static int pyr2tet[2][4] = { 1,2,4,5, 2,3,4,5 };
  static int pri2tet[3][4] = { 1,2,4,5, 2,3,4,5, 1,4,5,6 };
  static int hex2tet[6][4] = { 2,4,5,6, 1,2,4,5, 2,3,4,6,
                               3,4,6,8, 4,5,6,8, 3,6,7,8 };

  *divg = NULL;
  vol = (float *) malloc(globals->total_nodes*sizeof(float));
  if (vol == NULL) {
    printf(" FX Error: MALLOC for Volume storage (FX_Div)!\n");
    return;
  }
  sdivg = (float *) malloc(globals->total_nodes*sizeof(float));
  if (sdivg == NULL) {
    free(vol);
    printf(" FX Error: MALLOC for Divergence storage (FX_Div)!\n");
    return;
  }

  /* initialize counters */
  for (i = 0; i < globals->total_nodes; i++) vol[i] = sdivg[i] = 0.0;

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetDiv(globals, tet, vec, sdivg, vol);
        }
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (globals->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = node[k];
           }
           FX_TetDiv(globals, tet, vec, sdivg, vol);
        }
      }
    }
  }

  /* disjoint and halo elements */

  if ((globals->cmask & 128) == 0) {
  
    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++)
      FX_TetDiv(globals, globals->tets[kc].nodes, vec, sdivg, vol);

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++)
      for (j = 0; j < 2; j++) {
        for (i = 0; i < 4; i++) {
          k = pyr2tet[j][i]-1;
          node[i] = globals->pyra[kc].nodes[k];
        }
        FX_TetDiv(globals, node, vec, sdivg, vol);
      }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++)
      for (j = 0; j < 3; j++) {
        for (i = 0; i < 4; i++) {
          k = pri2tet[j][i]-1;
          node[i] = globals->prism[kc].nodes[k];
        }
        FX_TetDiv(globals, node, vec, sdivg, vol);
      }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++)
      for (j = 0; j < 6; j++) {
        for (i = 0; i < 4; i++) {
          k = hex2tet[j][i]-1;
          node[i] = globals->hexa[kc].nodes[k];
        }
        FX_TetDiv(globals, node, vec, sdivg, vol);
      }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++)
      switch (globals->hcell[kc].type) {

      case 1:
        /* tetrahedra */
        FX_HTetDiv(globals, globals->hcell[kc].nodes, vec, hvec, sdivg, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = globals->hcell[kc].nodes[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
      }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, cell);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_TetDiv(globals, cell, vec, sdivg, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            node[i] = cell[k];
          }
          FX_TetDiv(globals, node, vec, sdivg, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            node[i] = cell[k];
          }
          FX_TetDiv(globals, node, vec, sdivg, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            node[i] = cell[k];
          }
          FX_TetDiv(globals, node, vec, sdivg, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d, gives TYPE = %d (FX_Div)!\n",
               kc+1, type);
      }
    }

    /* halos */
    for (kc = 0; kc < globals->nhcell; kc++) {
      i = -(kc+1);
      FXCELLGET(&i, &type, cell);

      switch (type) {

      case 1:
        /* tetrahedra */
        FX_HTetDiv(globals, cell, vec, hvec, sdivg, vol);
        break;

      case 2:
        /* pyramids */
        for (j = 0; j < 2; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pyr2tet[j][i]-1;
            tet[i] = cell[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
        break;

      case 3:
        /* prisms */
        for (j = 0; j < 3; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = pri2tet[j][i]-1;
            tet[i] = cell[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
        break;

      case 4:
        /* hexahedra */
        for (j = 0; j < 6; j++) {
          for (kz = i = 0; i < 4; i++) {
            k = hex2tet[j][i]-1;
            tet[i] = cell[k];
            if (tet[i] <= globals->total_nodes) kz++;
          }
          if (kz != 0) FX_HTetDiv(globals, tet, vec, hvec, sdivg, vol);
        }
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d, gives TYPE = %d (FX_Div)!\n",
               -(kc+1), type);
      }
    }

  }

  /* finish vol weighting */

  for (i = 0; i < globals->total_nodes; i++)
    if (vol[i] != 0.0) sdivg[i] /= vol[i];

  free(vol);
  *divg = sdivg;
}
