/*
 * FX_GenSurf.c
 *
 * FX module that generates an iso-surface from the input scalar field.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>


#include "FX_types.h"

#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXCELLGET fxcellget_

#else

#define FXCELLGET fxcellget

#endif
#endif


  extern void PREFIX FXCELLGET (int *, int *, int *);

  extern void FX_CutTet(int, float, float *, int *, int *, FXfacet *, int *, 
                        FXinter *, int *);
  extern void FX_CutPyr(int, float, float *, int *, int *, FXfacet *, int *, 
                        FXinter *, int *);
  extern void FX_CutPri(int, float, float *, int *, int *, FXfacet *, int *, 
                        FXinter *, int *);
  extern void FX_CutHex(int, float, float *, int *, int *, FXfacet *, int *, 
                        FXinter *, int *);

void
FX_GenSurf(FXglobal *globals, float val, float *scalar,
           int *npts, FXcoor **xyz, int *ntris, int **tris, int **cell)
{
  int     i, j, k, kc, kcr, ib, medgeX, nedgeX, mtri, ntri, type, node[8];
  int     n1, n2, ni, nj, nk, nc, kns, kz, bnode, *ptris, *pcell, *nodeTable;
  float   d, scal[8];
  FXcoor  *coor;
  FXfacet *tri, *tmpt;
  FXinter *edgeX, *tmpe;

  *npts = *ntris = 0;
  *tris = *cell  = NULL;
  *xyz  = NULL;

  nodeTable = (int *) malloc(globals->total_nodes*sizeof(int));
  if (nodeTable == NULL) {
    printf(" FX Error: MALLOC for nodeTable (FX_GenSurf)!\n");
    return;
  }

  /* make a fist guess on what we will need for the intersection table */
  medgeX = (int) pow(2.5*globals->total_nodes, 0.6666667);
  if (medgeX < 1000) medgeX = 1000;
  edgeX = (FXinter *) malloc(medgeX*sizeof(FXinter));
  if (edgeX == NULL) {
    printf(" FX Error: MALLOC for edgeX (FX_GenSurf)!\n");
    free(nodeTable);
    return;
  }
  /* and number of triangles */
  mtri = medgeX;
  tri  = (FXfacet *) malloc(mtri*sizeof(FXfacet));
  if (tri == NULL) {
    printf(" FX Error: MALLOC for tris (FX_GenSurf)!\n");
    free(edgeX);
    free(nodeTable);
    return;
  }

  /* clear out counters */
  ntri = nedgeX = 0;
  for (i = 0; i < globals->total_nodes; i++) nodeTable[i] = 0;

  /* structured blocks */

  if ((globals->nblock > 0) && (globals->iblank != NULL))
    bnode = globals->block[0].last_node - 
            globals->block[0].ni*globals->block[0].nj*globals->block[0].nk;
  for (ib = 0; ib < globals->nblock; ib++) {
    ni  = globals->block[ib].ni;
    nj  = globals->block[ib].nj;
    nk  = globals->block[ib].nk;
    nc  = (ni-1)*(nj-1)*(nk-1);
    kns = globals->block[ib].last_node - ni*nj*nk + 1;
    kcr = globals->block[ib].last_cell - (ni-1)*(nj-1)*(nk-1) + 1;
    if (globals->iblank == NULL) {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (i = 0; i < 8; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+12 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- strblk (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+5 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- strblk (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutHex(kc+kcr, val, scal, node, &ntri, tri, &nedgeX, edgeX,
                  nodeTable);
      }
    } else {
      for (kc = 0; kc < nc; kc++) {
        i  =  kc%(ni-1);
        j  = (kc/(ni-1))%(nj-1);
        k  =  kc/((ni-1)*(nj-1));
        kz =  kns + i + j*ni + k*ni*nj;
        node[0] = kz;
        node[1] = kz + 1;
        node[2] = kz + 1 + ni;
        node[3] = kz     + ni;
        node[4] = kz          + ni*nj;
        node[5] = kz + 1      + ni*nj;
        node[6] = kz + 1 + ni + ni*nj;
        node[7] = kz     + ni + ni*nj;
        for (j = i = 0; i < 8; i++)
          if (globals->iblank[node[i]-bnode-1] == 0) j++;
        if (j != 0) continue;

        for (i = 0; i < 8; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+12 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- striblk (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+5 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- striblk (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutHex(kc+kcr, val, scal, node, &ntri, tri, &nedgeX, edgeX,
                  nodeTable);
      }
    }
  }

  /* disjoint elements */

  if ((globals->cmask & 128) == 0) {

    /* tetrahedra */
    for (kc = 0; kc < globals->ntets; kc++) {
      for (i = 0; i < 4; i++) {
        node[i] = globals->tets[kc].nodes[i];
        scal[i] = scalar[node[i]-1];
      }
      if (nedgeX+4 > medgeX) {
        i    = 2*medgeX;
        tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
        if (tmpe == NULL) {
          printf(" FX Error: REALLOC for edgeX -- tets (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        medgeX = i;
        edgeX  = tmpe;
      }
      if (ntri+2 > mtri) {
        i    = 2*mtri;
        tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
        if (tmpt == NULL) {
          printf(" FX Error: REALLOC for tris -- tets (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        mtri = i;
        tri  = tmpt;
      }
      FX_CutTet(kc+1, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
    }

    /* pyramids */
    for (kc = 0; kc < globals->npyra; kc++) {
      for (i = 0; i < 5; i++) {
        node[i] = globals->pyra[kc].nodes[i];
        scal[i] = scalar[node[i]-1];
      }
      if (nedgeX+6 > medgeX) {
        i    = 2*medgeX;
        tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
        if (tmpe == NULL) {
          printf(" FX Error: REALLOC for edgeX -- pyras (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        medgeX = i;
        edgeX  = tmpe;
      }
      if (ntri+4 > mtri) {
        i    = 2*mtri;
        tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
        if (tmpt == NULL) {
          printf(" FX Error: REALLOC for tris -- pyras (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        mtri = i;
        tri  = tmpt;
      }
      kcr = kc + globals->ntets + 1;
      FX_CutPyr(kcr, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
    }

    /* prisms */
    for (kc = 0; kc < globals->nprism; kc++) {
      for (i = 0; i < 6; i++) {
        node[i] = globals->prism[kc].nodes[i];
        scal[i] = scalar[node[i]-1];
      }
      if (nedgeX+7 > medgeX) {
        i    = 2*medgeX;
        tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
        if (tmpe == NULL) {
          printf(" FX Error: REALLOC for edgeX -- prisms (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        medgeX = i;
        edgeX  = tmpe;
      }
      if (ntri+4 > mtri) {
        i    = 2*mtri;
        tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
        if (tmpt == NULL) {
          printf(" FX Error: REALLOC for tris -- prisms (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        mtri = i;
        tri  = tmpt;
      }
      kcr = kc + globals->ntets + globals->npyra + 1;
      FX_CutPri(kcr, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
    }

    /* hexahedra */
    for (kc = 0; kc < globals->nhexa; kc++) {
      for (i = 0; i < 8; i++) {
        node[i] = globals->hexa[kc].nodes[i];
        scal[i] = scalar[node[i]-1];
      }
      if (nedgeX+12 > medgeX) {
        i    = 2*medgeX;
        tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
        if (tmpe == NULL) {
          printf(" FX Error: REALLOC for edgeX -- hexas (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        medgeX = i;
        edgeX  = tmpe;
      }
      if (ntri+5 > mtri) {
        i    = 2*mtri;
        tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
        if (tmpt == NULL) {
          printf(" FX Error: REALLOC for tris -- hexas (FX_GenSurf)!\n");
          free(tri);
          free(edgeX);
          free(nodeTable);
          return;
        }
        mtri = i;
        tri  = tmpt;
      }
      kcr = kc + globals->ntets + globals->npyra + globals->nprism + 1;
      FX_CutHex(kcr, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
    }

  } else {

    /* construct elements on-the-fly */

    for (kc = 0; kc < globals->total_dis; kc++) {
      i = kc+1;
      FXCELLGET(&i, &type, node);

      switch (type) {
      case 1:
        /* tetrahedra */
        for (i = 0; i < 4; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+4 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- tets (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+2 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- tets (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutTet(kc+1, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
        break;

      case 2:
        /* pyramids */
        for (i = 0; i < 5; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+6 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- pyras (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+4 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- pyras (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutPyr(kc+1, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
        break;

      case 3:
        /* prisms */
        for (i = 0; i < 6; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+7 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- prisms (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+4 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- prisms (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutPri(kc+1, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
        break;

      case 4:
        /* hexahedra */
        for (i = 0; i < 8; i++) scal[i] = scalar[node[i]-1];
        if (nedgeX+12 > medgeX) {
          i    = 2*medgeX;
          tmpe = (FXinter *) realloc((void *) edgeX, i*sizeof(FXinter));
          if (tmpe == NULL) {
            printf(" FX Error: REALLOC for edgeX -- hexas (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          medgeX = i;
          edgeX  = tmpe;
        }
        if (ntri+5 > mtri) {
          i    = 2*mtri;
          tmpt = (FXfacet *) realloc((void *) tri, i*sizeof(FXfacet));
          if (tmpt == NULL) {
            printf(" FX Error: REALLOC for tris -- hexas (FX_GenSurf)!\n");
            free(tri);
            free(edgeX);
            free(nodeTable);
            return;
          }
          mtri = i;
          tri  = tmpt;
        }
        FX_CutHex(kc+1, val, scal, node, &ntri, tri, &nedgeX, edgeX, nodeTable);
        break;

      default:
        printf(" FX Warning: FXcellGet INDEX = %d -> TYPE = %d (FX_GenSurf)!\n",
               kc+1, type);
      }

    }
  }

  /* ok, now put the data in the primitive arrays */

  if ((ntri == 0) || (nedgeX == 0)) {
    free(tri);
    free(edgeX);
    free(nodeTable);
    return;
  }

  coor = (FXcoor *) malloc(nedgeX*sizeof(FXcoor));
  if (coor == NULL) {
    printf(" FX Error: MALLOC for coors (FX_GenSurf)!\n");
    free(tri);
    free(edgeX);
    free(nodeTable);
    return;
  }
  ptris = (int *) malloc(ntri*3*sizeof(int));
  if (ptris == NULL) {
    printf(" FX Error: MALLOC for ptris (FX_GenSurf)!\n");
    free(coor);
    free(tri);
    free(edgeX);
    free(nodeTable);
    return;
  }
  pcell = (int *) malloc(ntri*sizeof(int));
  if (pcell == NULL) {
    printf(" FX Error: MALLOC for pcell (FX_GenSurf)!\n");
    free(ptris);
    free(coor);
    free(tri);
    free(edgeX);
    free(nodeTable);
    return;
  }

  *npts = nedgeX;
  if ((globals->cmask & 64) == 0) {
    for (i = 0; i < nedgeX; i++) {
      n1 = edgeX[i].n1-1;
      n2 = edgeX[i].n2-1;
      d  = edgeX[i].interp;
      coor[i].x = globals->coor[n1].x + d*(globals->coor[n2].x-
                  globals->coor[n1].x);
      coor[i].y = globals->coor[n1].y + d*(globals->coor[n2].y-
                  globals->coor[n1].y);
      coor[i].z = globals->coor[n1].z + d*(globals->coor[n2].z-
                  globals->coor[n1].z);
    }
  } else {
    for (i = 0; i < nedgeX; i++) {
      n1 = edgeX[i].n1-1;
      n2 = edgeX[i].n2-1;
      d  = edgeX[i].interp;
      coor[i].x = globals->xcoor[n1] + d*(globals->xcoor[n2]-
                  globals->xcoor[n1]);
      coor[i].y = globals->ycoor[n1] + d*(globals->ycoor[n2]-
                  globals->ycoor[n1]);
      coor[i].z = globals->zcoor[n1] + d*(globals->zcoor[n2]-
                  globals->zcoor[n1]);
    }
  }
  *xyz = coor;

  *ntris = ntri;
  for (i = 0; i < ntri; i++) {
    ptris[3*i  ] = tri[i].nodes[0];
    ptris[3*i+1] = tri[i].nodes[1];
    ptris[3*i+2] = tri[i].nodes[2];
    pcell[i]     = tri[i].nodes[3];
  }
  *tris = ptris;
  *cell = pcell;

  free(tri);
  free(edgeX);
  free(nodeTable);
}
