C      FORTRAN Utility routines for Residence Time
C
C      Copyright 1999 - 2000, Massachusetts Institute of Technology.
C
C******************************************************************
       SUBROUTINE FX_Smooth2
C******************************************************************
C
C      2nd DIFFERENCE SMOOTHING USING WEIGHT FACTORS
C      (PSEUDO-LAPLACIAN)
C
C..................................................................
C
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
c     local variables
c
      integer KN, KI, KS
      real    SDEN
C
C---- 2ND DIFFERENCE FOR ALL NODES
C 
      DO KN = 1, NNODE
C
         D2(KN) = 0.0
         SDEN   = SIG0*ABS(Q(KN)) + 0.0001
         DO KI = 1, INDSMO(1,KN)
            KS = INDSMO(2,KN) + KI
            D2(KN) = D2(KN) +          WGHSMO(KS)*(Q(KNSMO(KS))-Q(KN))
            SDEN   = SDEN   + SIG1*ABS(WGHSMO(KS)*(Q(KNSMO(KS))-Q(KN))) 
     &                      + SIG0*ABS(Q(KNSMO(KS)))
         ENDDO
         SWITCH(KN) = MIN(1.0, ABS(D2(KN)/SDEN))
      ENDDO
C
      RETURN
      END
c
c
C********************************************************************
       SUBROUTINE FX_Weights
C********************************************************************
C
C      WEIGHTS FOR SECOND-DIFFERENCE NUMERICAL SMOOTHING
C      Use of Lagrange Multiplier
C      Linear functions have zero Laplacian
C
C....................................................................
C
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
c     local variables
c
      integer KN, KI, KS, KO
      real    DX, DY, DZ, DN, DTX, DTY, DTZ, DTT
      real    RX, RY, RZ, IXX, IYY, IZZ, IXY, IXZ, IYZ, LX, LY, LZ
      real    DET, A1, A2, A3, B1, B2, B3, C1, C2, C3
C
      DET(A1,A2,A3,B1,B2,B3,C1,C2,C3) = A1*(B2*C3-B3*C2) -
     &                                  A2*(B1*C3-B3*C1) +
     &                                  A3*(B1*C2-B2*C1)
C
C---- Sweep over all nodes, get all neighbors
C
      DO KN = 1, NNODE
C
         RX  = 0.0
         RY  = 0.0
         RZ  = 0.0
         IXX = 0.0
         IYY = 0.0
         IZZ = 0.0
         IXY = 0.0
         IXZ = 0.0
         IYZ = 0.0
C
         DO KI = 1, INDSMO(1,KN)
C
            KS = INDSMO(2,KN) + KI
            KO = KNSMO(KS)
            DX = XYZ(1,KO) - XYZ(1,KN)
            DY = XYZ(2,KO) - XYZ(2,KN)
            DZ = XYZ(3,KO) - XYZ(3,KN)
C
            RX  = RX  + DX
            RY  = RY  + DY
            RZ  = RZ  + DZ
            IXX = IXX + DX*DX
            IYY = IYY + DY*DY
            IZZ = IZZ + DZ*DZ
            IXY = IXY + DX*DY
            IXZ = IXZ + DX*DZ
            IYZ = IYZ + DY*DZ
         ENDDO
C
         DN = 0.
         LX = 0.
         LY = 0.
         LZ = 0.
C
         DTX = DET(-RX,-RY,-RZ,IXY,IYY,IYZ,IXZ,IYZ,IZZ)
         DTY = DET( RX, RY, RZ,IXX,IXY,IXZ,IXZ,IYZ,IZZ)
         DTZ = DET( RX, RY, RZ,IXY,IYY,IYZ,IXX,IXY,IXZ)  
         DTT = DET(IXX,IXY,IXZ,IXY,IYY,IYZ,IXZ,IYZ,IZZ)
C
         IF (ABS(DTT).GE.1.E-15) DN = 1./DTT
C
         IF (ABS(DTX).GT.1.E-15) LX = DN*DTX
         IF (ABS(DTY).GT.1.E-15) LY = DN*DTY
         IF (ABS(DTZ).GT.1.E-15) LZ = DN*DTZ
C
         DO KI = 1, INDSMO(1,KN)
            KS = INDSMO(2,KN) + KI
            KO = KNSMO(KS)
            WGHSMO(KS) = 1.0 + LX*(XYZ(1,KO)-XYZ(1,KN)) +
     &                         LY*(XYZ(2,KO)-XYZ(2,KN)) +
     &                         LZ*(XYZ(3,KO)-XYZ(3,KN))
         ENDDO
C
      ENDDO
C
      RETURN
      END
c
c
C*****************************************************************************
      SUBROUTINE FX_RTTimeStep(DTIME)
C*****************************************************************************
C
C     Returns the allowable timestep
C
C.............................................................................
c
      implicit none
      real DTIME
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
C
C---- Local variables
C
      real    PHI, PHIMIN, PHIMAX, UM, VM, WM, US
      integer KC, KT, KN, KI, KNODE(4)
c
      data KNODE/4, 5, 6, 8/
C
      DTIME = 1.E30
      DO 1 KC = 1, NCELLI
C
         KT = FX_RTCELL(KC, PCELL)
C
C------- Calculate Velocity at Cell Center
C
         UM = 0.0
         VM = 0.0
         WM = 0.0
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            UM = UM + VEL(1,KN)
            VM = VM + VEL(2,KN)
            WM = WM + VEL(3,KN)
         ENDDO
         UM = UM/FLOAT(KNODE(KT))
         VM = VM/FLOAT(KNODE(KT))
         WM = WM/FLOAT(KNODE(KT))
         US = UM*UM + VM*VM + WM*WM
         IF(US .EQ. 0.0) GO TO 1
C
C------- Dot With Coordinates
C
         KN = PCELL(1)
         PHIMIN = UM*XYZ(1,KN) + VM*XYZ(2,KN) + WM*XYZ(3,KN)
         PHIMAX = PHIMIN
         DO KI  = 2, KNODE(KT)
            KN  = PCELL(KI)
            PHI = UM*XYZ(1,KN) + VM*XYZ(2,KN) + WM*XYZ(3,KN)
            IF(PHI .GT. PHIMAX) PHIMAX = PHI
            IF(PHI .LT. PHIMIN) PHIMIN = PHI
         ENDDO
C
         PHI = (PHIMAX - PHIMIN)/US
         IF (PHI .LT. DTIME) DTIME = PHI
C
 1    CONTINUE
C
      RETURN
      END
c
c
C*****************************************************************************
       SUBROUTINE FX_BCsRT
C*****************************************************************************
C
C      CORRECTION FOR BOUNDARY FLUXES (ONLY FIRST-ORDER TERMS) -- NOT INFLOW
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTFACE, FX_RTCELL, PCELL(8)
C
C---- Local variables
C
      POINTER (pNSurf, NSurf),     (pSurf, Surf),   (pSCel, SCel)
      INTEGER          NSurf(2,1),         Surf(1),         SCel(4,1)
      INTEGER nBC, kBC, IST, KC, KT, K1, K2, K3, K4, KN, KI, KF
      INTEGER KNODE(4)
      REAL    DQC, DFC, DGC, DHC
C
      DATA KNODE/4, 5, 6, 8/
C
      CALL FX_GetBC(nBC, pNSurf, pSurf, pSCel)
C
      IST = 1
      DO kBC = 1, nBC
         IF(NSurf(2,kBC) .EQ. 2) THEN
C
C---------- OutFlow -- Close 2nd order terms
C
            DO 1 KN = IST, NSurf(1,kBC)
              KC = Surf(KN)
              KT = FX_RTCELL(KC, PCELL)
              KF = FX_RTFACE(KT, KC, SCEL(1,KN), PCELL)
              IF(KF .EQ. 0) GO TO 1
              CALL FX_RT2Flux(KC, KT, PCELL, DFC, DGC, DHC)
              DQC = -0.125*DT*( DFC*FM(1,KF,KC) + DGC*FM(2,KF,KC) +
     &                          DHC*FM(3,KF,KC) )
C------------ Correct deltas
              K1  = SCel(1,KN)
              K2  = SCel(2,KN)
              K3  = SCel(3,KN)
              K4  = SCel(4,KN)
              DQ(K1) = DQ(K1) + DQC/VN(K1)
              DQ(K2) = DQ(K2) + DQC/VN(K2)
              DQ(K3) = DQ(K3) + DQC/VN(K3)
              IF(K4 .NE. 0) 
     &        DQ(K4) = DQ(K4) + DQC/VN(K4)
 1          CONTINUE
C
         ELSE IF(NSurf(2,kBC) .EQ. 3 .OR. NSurf(2,kBC) .EQ. 4) THEN
C
C---------- Walls -- no flux
C
            DO 2 KN = IST, NSurf(1,kBC)
              KC = Surf(KN)
              KT = FX_RTCELL(KC, PCELL)
              KF = FX_RTFACE(KT, KC, SCEL(1,KN), PCELL)
              IF(KF .EQ. 0) GO TO 2
              K1 = SCel(1,KN)
              K2 = SCel(2,KN)
              K3 = SCel(3,KN)
              K4 = SCel(4,KN)
C------------ Calculate flux error 
              IF(K4 .NE. 0) THEN
                DQC = 0.25*( (F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC) +
     &                       (G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC) +
     &                       (H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC) )
              ELSE
                DQC = ( (F(K1)+F(K2)+F(K3))*FM(1,KF,KC) +
     &                  (G(K1)+G(K2)+G(K3))*FM(2,KF,KC) +
     &                  (H(K1)+H(K2)+H(K3))*FM(3,KF,KC) ) / 3.0
              ENDIF
              DQC = DQC*DT/FLOAT(KNODE(KT))
C------------ Correct deltas
              DO KI = 1, KNODE(KT)
                K1 = PCELL(KI)
                DQ(K1) = DQ(K1) + DQC/VN(K1)
              ENDDO
 2          CONTINUE
C
         ELSE IF(NSurf(2,kBC) .EQ. 5) THEN
C
C---------- Symmerty -- no flux
C
            DO 3 KN = IST, NSurf(1,kBC)
              KC = Surf(KN)
              KT = FX_RTCELL(KC, PCELL)
              KF = FX_RTFACE(KT, KC, SCEL(1,KN), PCELL)
              IF(KF .EQ. 0) GO TO 3
              K1 = SCel(1,KN)
              K2 = SCel(2,KN)
              K3 = SCel(3,KN)
              K4 = SCel(4,KN)
C------------ Calculate flux error
              IF (K4 .NE. 0) THEN
                DQC = 0.25*( (F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC) +
     &                       (G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC) +
     &                       (H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC) )
              ELSE
                DQC = ( (F(K1)+F(K2)+F(K3))*FM(1,KF,KC) +
     &                  (G(K1)+G(K2)+G(K3))*FM(2,KF,KC) +
     &                  (H(K1)+H(K2)+H(K3))*FM(3,KF,KC) ) / 3.0
              ENDIF
              DQC = DQC*DT/FLOAT(KNODE(KT))
C------------ Correct deltas
              DO KI = 1, KNODE(KT)
                K1 = PCELL(KI)
                DQ(K1) = DQ(K1) + DQC/VN(K1)
              ENDDO
 3          CONTINUE
C
         ENDIF
C
         IST = NSurf(1,kBC) + 1
C
      ENDDO
C
      RETURN
      END
c
c
C*****************************************************************************
      INTEGER FUNCTION FX_RTFACE(KT, KC, FNODE, CNODE)
C*****************************************************************************
C
C     return the face index for the face
C.............................................................................
c
      implicit none
      INTEGER KT, KC, FNODE(4), CNODE(8)
c
c     include files
c
      include 'ResTime.inc'
c
c     local variables
c
      INTEGER KF, KN
      INTEGER NFACE(4), KFACE(4,6,4), SFACE(4), NNODES(4)
c
      data NFACE/4, 5, 5, 6/, NNODES/4, 5, 6, 8/
      data KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     &           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     &           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     &           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
      FX_RTFACE = 0
      DO KF = 1, NFACE(KT)
         SFACE(1) = FNODE(1)
         SFACE(2) = FNODE(2)
         SFACE(3) = FNODE(3)
         SFACE(4) = FNODE(4)
         DO KN = 1, 4
            IF(CNODE(KFACE(1,KF,KT)) .EQ. SFACE(KN)) SFACE(KN) = 0
            IF(CNODE(KFACE(2,KF,KT)) .EQ. SFACE(KN)) SFACE(KN) = 0
            IF(CNODE(KFACE(3,KF,KT)) .EQ. SFACE(KN)) SFACE(KN) = 0
            IF(CNODE(KFACE(4,KF,KT)) .EQ. SFACE(KN)) SFACE(KN) = 0
         ENDDO
         IF(SFACE(1)+SFACE(2)+SFACE(3)+SFACE(4) .EQ. 0) THEN
            FX_RTFACE = KF
            RETURN
         ENDIF
      ENDDO
      WRITE(*,*) 'Warning: Face not found!', KT, KC
      WRITE(*,*) '      Face =', FNODE
      WRITE(*,*) '      Cell =', (CNODE(KN), KN = 1, NNODES(KT))
C
      RETURN
      END
c
c
C*****************************************************************************
       SUBROUTINE FX_GradTau
C*****************************************************************************
C
C     COMPUTE GRADIENTS OF TAU 
C
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      real    DTDX, DTDY, DTDZ, CON1
      integer KN, KC, KT, KF, K1, K2, K3, K4
      integer KFACE(4,6,4), NFACE(4), KNODE(4)
c
      data NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      data KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     &           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     &           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     &           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
       DO KN = 1, NNODE
          DTAUDX(KN) = 0.0
          DTAUDY(KN) = 0.0
          DTAUDZ(KN) = 0.0
       ENDDO
C
C----- MAIN LOOP: GRADIENT EVALUATION AND DISTRIBUTION TO NODES
C
       DO KC = 1, NCELL
C
          DTDX = 0.0
          DTDY = 0.0
          DTDZ = 0.0
          KT   = FX_RTCELL(KC, PCELL)
C
          DO KF = 1, NFACE(KT)
             K1 = PCELL(KFACE(1,KF,KT))
             K2 = PCELL(KFACE(2,KF,KT))
             K3 = PCELL(KFACE(3,KF,KT))
             K4 = PCELL(KFACE(4,KF,KT))
C
C----------- Calculate Gradient based on cell
C            average node Q values for cell faces
C
             IF (K1 .EQ. K4) THEN
               CON1 = 0.33333333
               DTDX = DTDX + CON1*(Q(K1)+Q(K2)+Q(K3))*FM(1,KF,KC)
               DTDY = DTDY + CON1*(Q(K1)+Q(K2)+Q(K3))*FM(2,KF,KC)
               DTDZ = DTDZ + CON1*(Q(K1)+Q(K2)+Q(K3))*FM(3,KF,KC)
             ELSE
               CON1 = 0.25
               DTDX = DTDX + CON1*(Q(K1)+Q(K2)+Q(K3)+Q(K4))*FM(1,KF,KC)
               DTDY = DTDY + CON1*(Q(K1)+Q(K2)+Q(K3)+Q(K4))*FM(2,KF,KC)
               DTDZ = DTDZ + CON1*(Q(K1)+Q(K2)+Q(K3)+Q(K4))*FM(3,KF,KC)
             ENDIF
          ENDDO
C
C-------- Distribute to nodes
C
          CON1  = 1.0/FLOAT(KNODE(KT))
          DO K1 = 1, KNODE(KT)
             KN = PCELL(K1)
             DTAUDX(KN) = DTAUDX(KN) + CON1*DTDX
             DTAUDY(KN) = DTAUDY(KN) + CON1*DTDY 
             DTAUDZ(KN) = DTAUDZ(KN) + CON1*DTDZ 
          ENDDO
C
       ENDDO
C
C----- Divide update by node volume 
C
       DO KN = 1, NNODE
          DTAUDX(KN) = DTAUDX(KN) / VN(KN)
          DTAUDY(KN) = DTAUDY(KN) / VN(KN)
          DTAUDZ(KN) = DTAUDZ(KN) / VN(KN)
       ENDDO
C
       RETURN
       END
