C      FORTRAN Routines that perform the flux balance for Residence Time
C
C      Copyright 1999 - 2000, Massachusetts Institute of Technology.
C
C*****************************************************************************
       SUBROUTINE FX_RTFlux
C*****************************************************************************
C
C      FLUX EVALUATION (FIRST AND SECOND ORDER LAX-WENDROFF TERMS) 
C      DISTRIBUTE TO CONSIDERED CELL NODES
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL    QC, FSUM, GSUM, HSUM
      REAL    U, V, W, SW, UO, VO, WO, DU, DV, DW
      REAL    DQC, DFC, DGC, DHC
      REAL    CON1, CON2
      REAL    SIX, SIY, SIZ, SJX, SJY, SJZ, SKX, SKY, SKZ
      REAL    D2C1, DQN(8)
      INTEGER KT, KN, KC, KI, KF, K1, K2, K3, K4
      INTEGER KFACE(4,6,4), NFACE(4), KNODE(4)
c
      DATA NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      DATA KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
C---- CALCULATE F, G AND H TERMS FOR ALL NODES
C
      DO KN = 1, NNODE
         F(KN)  = Q(KN) * VEL(1,KN)
         G(KN)  = Q(KN) * VEL(2,KN)
         H(KN)  = Q(KN) * VEL(3,KN)
         DQ(KN) = 0.
      ENDDO
C
C---- MAIN LOOP: FLUXES EVALUATION AND DISTRIBUTION TO NODES
C
      DO KC = 1, NCELL
C
         KT = FX_RTCELL(KC, PCELL)
C
C------- Calculate average cell state
C
         QC   = 0.0
         D2C1 = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         SW   = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            D2C1 = D2C1 + D2(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            SW   = SW   + SWITCH(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         D2C1 = CON1*D2C1
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         SW   = CON1*SW
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
C
         DU = U - UO
         DV = V - VO
         DW = W - WO
C
C------- Calculate cell residual (first-order Lax-Wendroff terms)
C        average node F, G and H values for cell face fluxes evaluation
C
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
C
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT
         ENDIF
C
C------- Calculate second order Lax-Wendroff terms & distribute to nodes
C
         DFC = DQC*UO + QC*DU
         DGC = DQC*VO + QC*DV
         DHC = DQC*WO + QC*DW
C
         CON2 = -0.125*DT
         IF (KT .EQ. 1) THEN
C
C--------- Tets
           SIX = 0.25*(FM(1,1,KC)+FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC))
           SIY = 0.25*(FM(2,1,KC)+FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC))
           SIZ = 0.25*(FM(3,1,KC)+FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*(FM(1,2,KC)-SIX) +
     .                     DGC*(FM(2,2,KC)-SIY) +
     .                     DHC*(FM(3,2,KC)-SIZ) )
           DQN(2) = CON2*( DFC*(FM(1,3,KC)-SIX) +
     .                     DGC*(FM(2,3,KC)-SIY) +
     .                     DHC*(FM(3,3,KC)-SIZ) )
           DQN(3) = CON2*( DFC*(FM(1,4,KC)-SIX) +
     .                     DGC*(FM(2,4,KC)-SIY) +
     .                     DHC*(FM(3,4,KC)-SIZ) )
           DQN(4) = CON2*( DFC*(FM(1,1,KC)-SIX) +
     .                     DGC*(FM(2,1,KC)-SIY) +
     .                     DHC*(FM(3,1,KC)-SIZ) )
C
         ELSE IF (KT .EQ. 2) THEN
C
C--------- Pyrmids
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,5,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,5,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,5,KC))
           SKX = 0.5*FM(1,1,KC)
           SKY = 0.5*FM(2,1,KC)
           SKZ = 0.5*FM(3,1,KC)
           DQN(1) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
           SIX = 0.25*(FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC)+FM(1,5,KC))
           SIY = 0.25*(FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC)+FM(2,5,KC))
           SIZ = 0.25*(FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC)+FM(3,5,KC))
           DQN(5) = CON2*( DFC*(-SIX+SKX) +
     .                     DGC*(-SIY+SKY) +
     .                     DHC*(-SIZ+SKZ) )
C
         ELSE IF (KT .EQ. 3) THEN
C
C--------- Prisms
           SJX =     (FM(1,1,KC) + FM(1,2,KC) + FM(1,3,KC))/3.0
           SJY =     (FM(2,1,KC) + FM(2,2,KC) + FM(2,3,KC))/3.0
           SJZ =     (FM(3,1,KC) + FM(3,2,KC) + FM(3,3,KC))/3.0
           SIX = 0.5*(FM(1,5,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,5,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,5,KC) - FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*( SIX+FM(1,3,KC)-SJX) +
     .                     DGC*( SIY+FM(2,3,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,3,KC)-SJZ) )
           DQN(2) = CON2*( DFC*(-SIX+FM(1,3,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,3,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,3,KC)-SJZ) )
           DQN(3) = CON2*( DFC*(-SIX+FM(1,2,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,2,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,2,KC)-SJZ) )
           DQN(4) = CON2*( DFC*( SIX+FM(1,2,KC)-SJX) +
     .                     DGC*( SIY+FM(2,2,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,2,KC)-SJZ) )
           DQN(5) = CON2*( DFC*(-SIX+FM(1,1,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,1,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,1,KC)-SJZ) )
           DQN(6) = CON2*( DFC*( SIX+FM(1,1,KC)-SJX) +
     .                     DGC*( SIY+FM(2,1,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,1,KC)-SJZ) )
C
         ELSE
C
C--------- Hexas
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,6,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,6,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,6,KC))
           SKX = 0.5*(FM(1,5,KC) - FM(1,1,KC))
           SKY = 0.5*(FM(2,5,KC) - FM(2,1,KC))
           SKZ = 0.5*(FM(3,5,KC) - FM(3,1,KC))
C
           DQN(1) = CON2*( DFC*( SIX+SJX+SKX) +
     .                     DGC*( SIY+SJY+SKY) +
     .                     DHC*( SIZ+SJZ+SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX+SKX) +
     .                     DGC*(-SIY+SJY+SKY) +
     .                     DHC*(-SIZ+SJZ+SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX+SKX) +
     .                     DGC*(-SIY-SJY+SKY) +
     .                     DHC*(-SIZ-SJZ+SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX+SKX) +
     .                     DGC*( SIY-SJY+SKY) +
     .                     DHC*( SIZ-SJZ+SKZ) )
           DQN(5) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(6) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(7) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(8) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
         ENDIF
C
C------- Total change: residence time with smoothing
C
         CON1 = (1.0-SW) * SIG4 * VC(KC)
         CON2 =      SW  * SIG2 * VC(KC)
         DQC  = DQC * VC(KC) / FLOAT(KNODE(KT))
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            DQ(KN) = DQ(KN) + DQC + DQN(KI) + CON2*(QC-Q(KN)) -
     .                                        CON1*(D2C1-D2(KN))
         ENDDO
C
       ENDDO
C
C----- Divide update by node volume
C
       DO KN = 1, NNODE
          DQ(KN) = DQ(KN) / VN(KN)
       ENDDO
C
       RETURN
       END

C
C*****************************************************************************
       SUBROUTINE FX_RTFluxInvComp
C*****************************************************************************
C
C      FLUX EVALUATION FOR ALL CELLS (FIRST AND SECOND ORDER LAX-WENDROFF
C      TERMS, AND VISCOUS TERM), INVISCID COMPRESSIBLE EQUATION
C
C      d(rho tau)/d(t) + del . (rho tau U) = rho
C
C..............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL    RTAU, QC, U, V, W, SW, R
      REAL    UO, VO, WO, RO, DU, DV, DW, DR
      REAL    DQC, DFC, DGC, DHC, DQN(8)
      REAL    FSUM, GSUM, HSUM, DSS, D2C1, CON1, CON2
      REAL    SIX, SIY, SIZ, SJX, SJY, SJZ, SKX, SKY, SKZ
      INTEGER KN, KC, KT, KI, KF, K1, K2, K3, K4
      INTEGER KFACE(4,6,4), NFACE(4), KNODE(4)
c
      DATA NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      DATA KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
C---- CALCULATE F, G AND H TERMS FOR ALL NODES
C
      DO KN = 1, NNODE
         RTAU   = Q(KN)
         U      = VEL(1,KN)
         V      = VEL(2,KN)
         W      = VEL(3,KN)
         F(KN)  = RTAU * U
         G(KN)  = RTAU * V
         H(KN)  = RTAU * W
         DQ(KN) = 0.
      ENDDO
C
C---- MAIN LOOP: FLUXES EVALUATION AND DISTRIBUTION TO NODES
C
      DO KC = 1, NCELL
C
         KT = FX_RTCELL(KC, PCELL)
C
C------- Calculate average cell state
C
         QC   = 0.0
         D2C1 = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         SW   = 0.0
         R    = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
         RO   = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            D2C1 = D2C1 + D2(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            SW   = SW   + SWITCH(KN)
            R    = R    + RHO(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
            RO   = RO   + RHOO(KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         D2C1 = CON1*D2C1
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         SW   = CON1*SW
         R    = CON1*R
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
         RO   = CON1*RO
C
         DU = U - UO
         DV = V - VO
         DW = W - WO
         DR = R - RO
C
C------- Calculate cell residual (first-order Lax-Wendroff terms)
C        average node F, G and H values for cell face fluxes evaluation
C
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
C
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R
         ENDIF
C
C--- Calculate second order Lax-Wendroff terms
C
         DFC = DQC * UO + QC * DU
         DGC = DQC * VO + QC * DV
         DHC = DQC * WO + QC * DW
C
         CON2 = -0.125*DT
         IF (KT .EQ. 1) THEN
C
C--------- Tets
           SIX = 0.25*(FM(1,1,KC)+FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC))
           SIY = 0.25*(FM(2,1,KC)+FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC))
           SIZ = 0.25*(FM(3,1,KC)+FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*(FM(1,2,KC)-SIX) +
     .                     DGC*(FM(2,2,KC)-SIY) +
     .                     DHC*(FM(3,2,KC)-SIZ) )
           DQN(2) = CON2*( DFC*(FM(1,3,KC)-SIX) +
     .                     DGC*(FM(2,3,KC)-SIY) +
     .                     DHC*(FM(3,3,KC)-SIZ) )
           DQN(3) = CON2*( DFC*(FM(1,4,KC)-SIX) +
     .                     DGC*(FM(2,4,KC)-SIY) +
     .                     DHC*(FM(3,4,KC)-SIZ) )
           DQN(4) = CON2*( DFC*(FM(1,1,KC)-SIX) +
     .                     DGC*(FM(2,1,KC)-SIY) +
     .                     DHC*(FM(3,1,KC)-SIZ) )
C
         ELSE IF (KT .EQ. 2) THEN
C
C--------- Pyrmids
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,5,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,5,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,5,KC))
           SKX = 0.5*FM(1,1,KC)
           SKY = 0.5*FM(2,1,KC)
           SKZ = 0.5*FM(3,1,KC)
           DQN(1) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
           SIX = 0.25*(FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC)+FM(1,5,KC))
           SIY = 0.25*(FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC)+FM(2,5,KC))
           SIZ = 0.25*(FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC)+FM(3,5,KC))
           DQN(5) = CON2*( DFC*(-SIX+SKX) +
     .                     DGC*(-SIY+SKY) +
     .                     DHC*(-SIZ+SKZ) )
C
         ELSE IF (KT .EQ. 3) THEN
C
C--------- Prisms
           SJX =     (FM(1,1,KC) + FM(1,2,KC) + FM(1,3,KC))/3.0
           SJY =     (FM(2,1,KC) + FM(2,2,KC) + FM(2,3,KC))/3.0
           SJZ =     (FM(3,1,KC) + FM(3,2,KC) + FM(3,3,KC))/3.0
           SIX = 0.5*(FM(1,5,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,5,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,5,KC) - FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*( SIX+FM(1,3,KC)-SJX) +
     .                     DGC*( SIY+FM(2,3,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,3,KC)-SJZ) )
           DQN(2) = CON2*( DFC*(-SIX+FM(1,3,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,3,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,3,KC)-SJZ) )
           DQN(3) = CON2*( DFC*(-SIX+FM(1,2,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,2,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,2,KC)-SJZ) )
           DQN(4) = CON2*( DFC*( SIX+FM(1,2,KC)-SJX) +
     .                     DGC*( SIY+FM(2,2,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,2,KC)-SJZ) )
           DQN(5) = CON2*( DFC*(-SIX+FM(1,1,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,1,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,1,KC)-SJZ) )
           DQN(6) = CON2*( DFC*( SIX+FM(1,1,KC)-SJX) +
     .                     DGC*( SIY+FM(2,1,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,1,KC)-SJZ) )
C
         ELSE
C
C--------- Hexas
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,6,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,6,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,6,KC))
           SKX = 0.5*(FM(1,5,KC) - FM(1,1,KC))
           SKY = 0.5*(FM(2,5,KC) - FM(2,1,KC))
           SKZ = 0.5*(FM(3,5,KC) - FM(3,1,KC))
C
           DQN(1) = CON2*( DFC*( SIX+SJX+SKX) +
     .                     DGC*( SIY+SJY+SKY) +
     .                     DHC*( SIZ+SJZ+SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX+SKX) +
     .                     DGC*(-SIY+SJY+SKY) +
     .                     DHC*(-SIZ+SJZ+SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX+SKX) +
     .                     DGC*(-SIY-SJY+SKY) +
     .                     DHC*(-SIZ-SJZ+SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX+SKX) +
     .                     DGC*( SIY-SJY+SKY) +
     .                     DHC*( SIZ-SJZ+SKZ) )
           DQN(5) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(6) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(7) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(8) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
         ENDIF
C
C------- Total change: residence time with smoothing
C
         CON1 = (1.0-SW) * SIG4 * VC(KC)
         CON2 =      SW  * SIG2 * VC(KC)
         DQC  = DQC  * VC(KC)         / FLOAT(KNODE(KT))
         DSS  = 0.5  * VC(KC) * DR*DT / FLOAT(KNODE(KT))
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            DQ(KN) = DQ(KN) + DQC + DSS + DQN(KI) + CON2*(QC-Q(KN)) -
     .                                              CON1*(D2C1-D2(KN))
         ENDDO
C
      ENDDO
C
C---- Divide update by node volume
      DO KN = 1, NNODE
         DQ(KN) = DQ(KN) / VN(KN)
      ENDDO
C
      RETURN
      END

C
C*****************************************************************************
       SUBROUTINE FX_RTFluxVisComp
C*****************************************************************************
C
C      FLUX EVALUATION FOR ALL CELLS (FIRST AND SECOND ORDER LAX-WENDROFF 
C      TERMS, AND VISCOUS TERM), VISCOUS COMPRESSIBLE EQUATION 
C      (SPATIALLY-VARYING VISCOSITY)
C
C      d(rho tau)/d(t) + del . (rho tau U) = rho + del . mu del(tau)
C
C..............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL    RTAU, QC, U, V, W, R, M, SW
      REAL    UO, VO, WO, RO, MO, DU, DV, DW, DR, DM
      REAL    DTXO, DTYO, DTZO, DDTX, DDTY, DDTZ
      REAL    FSUM, GSUM, HSUM, DTX, DTY, DTZ, DSS
      REAL    DQC, DFC, DGC, DHC, DQN(8)
      REAL    D2C1, CON1, CON2
      REAL    SIX, SIY, SIZ, SJX, SJY, SJZ, SKX, SKY, SKZ
      INTEGER KN, KC, KT, KI, KF, K1, K2, K3, K4
      INTEGER KFACE(4,6,4), NFACE(4), KNODE(4)
c
      DATA NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      DATA KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
C---- CALCULATE F, G AND H TERMS FOR ALL NODES
C
      DO KN = 1, NNODE
         RTAU   = Q(KN)
         U      = VEL(1,KN)
         V      = VEL(2,KN)
         W      = VEL(3,KN)
         R      = RHO(KN)
         M      = MU(KN)
         DTX    = DTAUDX(KN)
         DTY    = DTAUDY(KN)
         DTZ    = DTAUDZ(KN)
         F(KN)  = RTAU * U - M * DTX
         G(KN)  = RTAU * V - M * DTY
         H(KN)  = RTAU * W - M * DTZ
         DQ(KN) = 0.
      ENDDO
C
C---- MAIN LOOP: FLUXES EVALUATION AND DISTRIBUTION TO NODES
C
      DO KC = 1, NCELL
C
         KT = FX_RTCELL(KC, PCELL)
C
C------- Calculate average cell state
C
         QC   = 0.0
         D2C1 = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         SW   = 0.0
         R    = 0.0
         M    = 0.0
         DTX  = 0.0
         DTY  = 0.0
         DTZ  = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
         RO   = 0.0
         MO   = 0.0
         DTXO = 0.0
         DTYO = 0.0
         DTZO = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            D2C1 = D2C1 + D2(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            SW   = SW   + SWITCH(KN)
            R    = R    + RHO(KN)
            M    = M    + MU(KN)
            DTX  = DTX  + DTAUDX(KN)
            DTY  = DTY  + DTAUDY(KN)
            DTZ  = DTZ  + DTAUDZ(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
            RO   = RO   + RHOO(KN)
            MO   = MO   + MUO(KN)
            DTXO = DTXO + DTAUDXO(KN)
            DTYO = DTYO + DTAUDYO(KN)
            DTZO = DTZO + DTAUDZO(KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         D2C1 = CON1*D2C1
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         SW   = CON1*SW
         R    = CON1*R
         M    = CON1*M
         DTX  = CON1*DTX
         DTY  = CON1*DTY
         DTZ  = CON1*DTZ
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
         RO   = CON1*RO
         MO   = CON1*MO
         DTXO = CON1*DTXO
         DTYO = CON1*DTYO
         DTZO = CON1*DTZO
C
         DU   = U - UO
         DV   = V - VO
         DW   = W - WO
         DR   = R - RO
         DM   = M - MO
         DDTX = DTX - DTXO
         DDTY = DTY - DTYO
         DDTZ = DTZ - DTZO
C
C------- Calculate cell residual (first-order Lax-Wendroff terms)
C        average node F, G and H values for cell face fluxes evaluation
C
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
C
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R
         ENDIF
C
C------- Calculate second order Lax-Wendroff terms
C
         DFC = DQC * UO + QC * DU - MO * DDTX - DTX * DM
         DGC = DQC * VO + QC * DV - MO * DDTY - DTY * DM
         DHC = DQC * WO + QC * DW - MO * DDTZ - DTZ * DM
C
         CON2 = -0.125*DT
         IF (KT .EQ. 1) THEN
C
C--------- Tets
           SIX = 0.25*(FM(1,1,KC)+FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC))
           SIY = 0.25*(FM(2,1,KC)+FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC))
           SIZ = 0.25*(FM(3,1,KC)+FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*(FM(1,2,KC)-SIX) +
     .                     DGC*(FM(2,2,KC)-SIY) +
     .                     DHC*(FM(3,2,KC)-SIZ) )
           DQN(2) = CON2*( DFC*(FM(1,3,KC)-SIX) +
     .                     DGC*(FM(2,3,KC)-SIY) +
     .                     DHC*(FM(3,3,KC)-SIZ) )
           DQN(3) = CON2*( DFC*(FM(1,4,KC)-SIX) +
     .                     DGC*(FM(2,4,KC)-SIY) +
     .                     DHC*(FM(3,4,KC)-SIZ) )
           DQN(4) = CON2*( DFC*(FM(1,1,KC)-SIX) +
     .                     DGC*(FM(2,1,KC)-SIY) +
     .                     DHC*(FM(3,1,KC)-SIZ) )
C
         ELSE IF (KT .EQ. 2) THEN
C
C--------- Pyrmids
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,5,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,5,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,5,KC))
           SKX = 0.5*FM(1,1,KC)
           SKY = 0.5*FM(2,1,KC)
           SKZ = 0.5*FM(3,1,KC)
           DQN(1) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
           SIX = 0.25*(FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC)+FM(1,5,KC))
           SIY = 0.25*(FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC)+FM(2,5,KC))
           SIZ = 0.25*(FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC)+FM(3,5,KC))
           DQN(5) = CON2*( DFC*(-SIX+SKX) +
     .                     DGC*(-SIY+SKY) +
     .                     DHC*(-SIZ+SKZ) )
C
         ELSE IF (KT .EQ. 3) THEN
C
C--------- Prisms
           SJX =     (FM(1,1,KC) + FM(1,2,KC) + FM(1,3,KC))/3.0
           SJY =     (FM(2,1,KC) + FM(2,2,KC) + FM(2,3,KC))/3.0
           SJZ =     (FM(3,1,KC) + FM(3,2,KC) + FM(3,3,KC))/3.0
           SIX = 0.5*(FM(1,5,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,5,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,5,KC) - FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*( SIX+FM(1,3,KC)-SJX) +
     .                     DGC*( SIY+FM(2,3,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,3,KC)-SJZ) )
           DQN(2) = CON2*( DFC*(-SIX+FM(1,3,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,3,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,3,KC)-SJZ) )
           DQN(3) = CON2*( DFC*(-SIX+FM(1,2,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,2,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,2,KC)-SJZ) )
           DQN(4) = CON2*( DFC*( SIX+FM(1,2,KC)-SJX) +
     .                     DGC*( SIY+FM(2,2,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,2,KC)-SJZ) )
           DQN(5) = CON2*( DFC*(-SIX+FM(1,1,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,1,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,1,KC)-SJZ) )
           DQN(6) = CON2*( DFC*( SIX+FM(1,1,KC)-SJX) +
     .                     DGC*( SIY+FM(2,1,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,1,KC)-SJZ) )
C
         ELSE
C
C--------- Hexas
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,6,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,6,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,6,KC))
           SKX = 0.5*(FM(1,5,KC) - FM(1,1,KC))
           SKY = 0.5*(FM(2,5,KC) - FM(2,1,KC))
           SKZ = 0.5*(FM(3,5,KC) - FM(3,1,KC))
C
           DQN(1) = CON2*( DFC*( SIX+SJX+SKX) +
     .                     DGC*( SIY+SJY+SKY) +
     .                     DHC*( SIZ+SJZ+SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX+SKX) +
     .                     DGC*(-SIY+SJY+SKY) +
     .                     DHC*(-SIZ+SJZ+SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX+SKX) +
     .                     DGC*(-SIY-SJY+SKY) +
     .                     DHC*(-SIZ-SJZ+SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX+SKX) +
     .                     DGC*( SIY-SJY+SKY) +
     .                     DHC*( SIZ-SJZ+SKZ) )
           DQN(5) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(6) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(7) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(8) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
         ENDIF
C
C------- Total change: residence time with smoothing
C
         CON1 = (1.0-SW) * SIG4 * VC(KC)
         CON2 =      SW  * SIG2 * VC(KC)
         DQC  = DQC  * VC(KC)         / FLOAT(KNODE(KT))
         DSS  = 0.5  * VC(KC) * DR*DT / FLOAT(KNODE(KT))
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            DQ(KN) = DQ(KN) + DQC + DSS + DQN(KI) + CON2*(QC-Q(KN)) - 
     .                                              CON1*(D2C1-D2(KN))
         ENDDO
C
      ENDDO
C
C---- Divide update by node volume
C
      DO KN = 1, NNODE
         DQ(KN) = DQ(KN) / VN(KN)
      ENDDO
C
      RETURN
      END
C
C*****************************************************************************
       SUBROUTINE FX_RTFluxConVisDens
C*****************************************************************************
C
C      FLUX EVALUATION FOR ALL CELLS (FIRST AND SECOND ORDER LAX-WENDROFF 
C      TERMS, AND VISCOUS TERM), CONSTANT VISCOSITY AND DENSITY EQUATION:
C
C      d(tau)/d(t) + del . (tau U) = 1 + (mu / rho) del2 (tau)
C
C..............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL    TAU, QC, U, V, W, SW
      REAL    UO, VO, WO, DU, DV, DW
      REAL    DTXO, DTYO, DTZO, DDTX, DDTY, DDTZ
      REAL    FSUM, GSUM, HSUM, DTX, DTY, DTZ
      REAL    DQC, DFC, DGC, DHC, DQN(8)
      REAL    D2C1, CON1, CON2
      REAL    SIX, SIY, SIZ, SJX, SJY, SJZ, SKX, SKY, SKZ
      INTEGER KN, KC, KT, KI, KF, K1, K2, K3, K4
      INTEGER KFACE(4,6,4), NFACE(4), KNODE(4)
c
      DATA NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      DATA KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
C---- CALCULATE F, G AND H TERMS FOR ALL NODES
C
      DO KN = 1, NNODE
         TAU    = Q(KN)
         U      = VEL(1,KN)
         V      = VEL(2,KN)
         W      = VEL(3,KN)
         DTX    = DTAUDX(KN)
         DTY    = DTAUDY(KN)
         DTZ    = DTAUDZ(KN)
         F(KN)  = TAU * U - NU * DTX
         G(KN)  = TAU * V - NU * DTY
         H(KN)  = TAU * W - NU * DTZ
         DQ(KN) = 0.
      ENDDO
C
C---- MAIN LOOP: FLUXES EVALUATION AND DISTRIBUTION TO NODES
C
      DO KC = 1, NCELL
C
         KT = FX_RTCELL(KC, PCELL)
C
C------- Calculate average cell state
C
         QC   = 0.0
         D2C1 = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         SW   = 0.0
         DTX  = 0.0
         DTY  = 0.0
         DTZ  = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
         DTXO = 0.0
         DTYO = 0.0
         DTZO = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            D2C1 = D2C1 + D2(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            SW   = SW   + SWITCH(KN)
            DTX  = DTX  + DTAUDX(KN)
            DTY  = DTY  + DTAUDY(KN)
            DTZ  = DTZ  + DTAUDZ(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
            DTXO = DTXO + DTAUDXO(KN)
            DTYO = DTYO + DTAUDYO(KN)
            DTZO = DTZO + DTAUDZO(KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         D2C1 = CON1*D2C1
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         SW   = CON1*SW
         DTX  = CON1*DTX
         DTY  = CON1*DTY
         DTZ  = CON1*DTZ
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
         DTXO = CON1*DTXO
         DTYO = CON1*DTYO
         DTZO = CON1*DTZO
C
         DU = U - UO
         DV = V - VO
         DW = W - WO
         DDTX = DTX - DTXO
         DDTY = DTY - DTYO
         DDTZ = DTZ - DTZO
C
C------- Calculate cell residual (first-order Lax-Wendroff terms)
C        average node F, G and H values for cell face fluxes evaluation
C
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
C
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT
         ENDIF
C
C------- Calculate second order Lax-Wendroff terms
C
         DFC = DQC * UO + QC * DU - NU * DDTX
         DGC = DQC * VO + QC * DV - NU * DDTY
         DHC = DQC * WO + QC * DW - NU * DDTZ
C
         CON2 = -0.125*DT
         IF (KT .EQ. 1) THEN
C
C--------- Tets
           SIX = 0.25*(FM(1,1,KC)+FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC))
           SIY = 0.25*(FM(2,1,KC)+FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC))
           SIZ = 0.25*(FM(3,1,KC)+FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*(FM(1,2,KC)-SIX) +
     .                     DGC*(FM(2,2,KC)-SIY) +
     .                     DHC*(FM(3,2,KC)-SIZ) )
           DQN(2) = CON2*( DFC*(FM(1,3,KC)-SIX) +
     .                     DGC*(FM(2,3,KC)-SIY) +
     .                     DHC*(FM(3,3,KC)-SIZ) )
           DQN(3) = CON2*( DFC*(FM(1,4,KC)-SIX) +
     .                     DGC*(FM(2,4,KC)-SIY) +
     .                     DHC*(FM(3,4,KC)-SIZ) )
           DQN(4) = CON2*( DFC*(FM(1,1,KC)-SIX) +
     .                     DGC*(FM(2,1,KC)-SIY) +
     .                     DHC*(FM(3,1,KC)-SIZ) )
C
         ELSE IF (KT .EQ. 2) THEN
C
C--------- Pyrmids
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,5,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,5,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,5,KC))
           SKX = 0.5*FM(1,1,KC)
           SKY = 0.5*FM(2,1,KC)
           SKZ = 0.5*FM(3,1,KC)
           DQN(1) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
           SIX = 0.25*(FM(1,2,KC)+FM(1,3,KC)+FM(1,4,KC)+FM(1,5,KC))
           SIY = 0.25*(FM(2,2,KC)+FM(2,3,KC)+FM(2,4,KC)+FM(2,5,KC))
           SIZ = 0.25*(FM(3,2,KC)+FM(3,3,KC)+FM(3,4,KC)+FM(3,5,KC))
           DQN(5) = CON2*( DFC*(-SIX+SKX) +
     .                     DGC*(-SIY+SKY) +
     .                     DHC*(-SIZ+SKZ) )
C
         ELSE IF (KT .EQ. 3) THEN
C
C--------- Prisms
           SJX =     (FM(1,1,KC) + FM(1,2,KC) + FM(1,3,KC))/3.0
           SJY =     (FM(2,1,KC) + FM(2,2,KC) + FM(2,3,KC))/3.0
           SJZ =     (FM(3,1,KC) + FM(3,2,KC) + FM(3,3,KC))/3.0
           SIX = 0.5*(FM(1,5,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,5,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,5,KC) - FM(3,4,KC))
C
           DQN(1) = CON2*( DFC*( SIX+FM(1,3,KC)-SJX) +
     .                     DGC*( SIY+FM(2,3,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,3,KC)-SJZ) )
           DQN(2) = CON2*( DFC*(-SIX+FM(1,3,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,3,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,3,KC)-SJZ) )
           DQN(3) = CON2*( DFC*(-SIX+FM(1,2,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,2,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,2,KC)-SJZ) )
           DQN(4) = CON2*( DFC*( SIX+FM(1,2,KC)-SJX) +
     .                     DGC*( SIY+FM(2,2,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,2,KC)-SJZ) )
           DQN(5) = CON2*( DFC*(-SIX+FM(1,1,KC)-SJX) +
     .                     DGC*(-SIY+FM(2,1,KC)-SJY) +
     .                     DHC*(-SIZ+FM(3,1,KC)-SJZ) )
           DQN(6) = CON2*( DFC*( SIX+FM(1,1,KC)-SJX) +
     .                     DGC*( SIY+FM(2,1,KC)-SJY) +
     .                     DHC*( SIZ+FM(3,1,KC)-SJZ) )
C
         ELSE
C
C--------- Hexas
           SIX = 0.5*(FM(1,2,KC) - FM(1,4,KC))
           SIY = 0.5*(FM(2,2,KC) - FM(2,4,KC))
           SIZ = 0.5*(FM(3,2,KC) - FM(3,4,KC))
           SJX = 0.5*(FM(1,3,KC) - FM(1,6,KC))
           SJY = 0.5*(FM(2,3,KC) - FM(2,6,KC))
           SJZ = 0.5*(FM(3,3,KC) - FM(3,6,KC))
           SKX = 0.5*(FM(1,5,KC) - FM(1,1,KC))
           SKY = 0.5*(FM(2,5,KC) - FM(2,1,KC))
           SKZ = 0.5*(FM(3,5,KC) - FM(3,1,KC))
C
           DQN(1) = CON2*( DFC*( SIX+SJX+SKX) +
     .                     DGC*( SIY+SJY+SKY) +
     .                     DHC*( SIZ+SJZ+SKZ) )
           DQN(2) = CON2*( DFC*(-SIX+SJX+SKX) +
     .                     DGC*(-SIY+SJY+SKY) +
     .                     DHC*(-SIZ+SJZ+SKZ) )
           DQN(3) = CON2*( DFC*(-SIX-SJX+SKX) +
     .                     DGC*(-SIY-SJY+SKY) +
     .                     DHC*(-SIZ-SJZ+SKZ) )
           DQN(4) = CON2*( DFC*( SIX-SJX+SKX) +
     .                     DGC*( SIY-SJY+SKY) +
     .                     DHC*( SIZ-SJZ+SKZ) )
           DQN(5) = CON2*( DFC*( SIX+SJX-SKX) +
     .                     DGC*( SIY+SJY-SKY) +
     .                     DHC*( SIZ+SJZ-SKZ) )
           DQN(6) = CON2*( DFC*(-SIX+SJX-SKX) +
     .                     DGC*(-SIY+SJY-SKY) +
     .                     DHC*(-SIZ+SJZ-SKZ) )
           DQN(7) = CON2*( DFC*(-SIX-SJX-SKX) +
     .                     DGC*(-SIY-SJY-SKY) +
     .                     DHC*(-SIZ-SJZ-SKZ) )
           DQN(8) = CON2*( DFC*( SIX-SJX-SKX) +
     .                     DGC*( SIY-SJY-SKY) +
     .                     DHC*( SIZ-SJZ-SKZ) )
         ENDIF
C
C------- Total change: residence time with smoothing
C
         CON1 = (1.0-SW) * SIG4 * VC(KC)
         CON2 =      SW  * SIG2 * VC(KC)
         DQC  = DQC  * VC(KC) / FLOAT(KNODE(KT))
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            DQ(KN) = DQ(KN) + DQC + DQN(KI) + CON2*(QC-Q(KN)) - 
     .                                        CON1*(D2C1-D2(KN))
         ENDDO
C
      ENDDO
C
C---- Divide update by node
C
      DO KN = 1, NNODE
         DQ(KN) = DQ(KN) / VN(KN)
      ENDDO
C
      RETURN
      END
C
C*****************************************************************************
       SUBROUTINE FX_RT2Flux(KC, KT, PCELL, DFC, DGC, DHC)
C*****************************************************************************
C
C      SECOND ORDER LAX-WENDROFF TERMS FOR BCs
C
C..............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      INTEGER KC, KT, PCELL(8)
      REAL    DFC, DGC, DHC
c
c     local variables
c
      REAL    QC, U, V, W, R, M
      REAL    UO, VO, WO, RO, MO, DU, DV, DW, DM
      REAL    DTXO, DTYO, DTZO, DDTX, DDTY, DDTZ
      REAL    FSUM, GSUM, HSUM, DTX, DTY, DTZ
      REAL    DQC, CON1
      INTEGER KN, KI, KF, K1, K2, K3, K4
      INTEGER KFACE(4,6,4), NFACE(4), KNODE(4)
c
      DATA NFACE/4, 5, 5, 6/, KNODE/4, 5, 6, 8/
      DATA KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
C
      IF(MODE .EQ. 0 .OR. MODE .EQ. 3) THEN
         QC = 0.0
         U  = 0.0
         V  = 0.0
         W  = 0.0
         UO = 0.0
         VO = 0.0
         WO = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC = QC   + Q(KN)
            U  = U    + VEL(1,KN)
            V  = V    + VEL(2,KN)
            W  = W    + VEL(3,KN)
            UO = UO   + VELO(1,KN)
            VO = VO   + VELO(2,KN)
            WO = WO   + VELO(3,KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
C
         DU = U - UO
         DV = V - VO
         DW = W - WO
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT
         ENDIF
         DFC = DQC*UO + QC*DU
         DGC = DQC*VO + QC*DV
         DHC = DQC*WO + QC*DW
      ELSEIF(MODE .EQ. 1) THEN
         QC   = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         R    = 0.0
         M    = 0.0
         DTX  = 0.0
         DTY  = 0.0
         DTZ  = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
         MO   = 0.0
         DTXO = 0.0
         DTYO = 0.0
         DTZO = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            R    = R    + RHO(KN)
            M    = M    + MU(KN)
            DTX  = DTX  + DTAUDX(KN)
            DTY  = DTY  + DTAUDY(KN)
            DTZ  = DTZ  + DTAUDZ(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
            MO   = MO   + MUO(KN)
            DTXO = DTXO + DTAUDXO(KN)
            DTYO = DTYO + DTAUDYO(KN)
            DTZO = DTZO + DTAUDZO(KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         R    = CON1*R
         M    = CON1*M
         DTX  = CON1*DTX
         DTY  = CON1*DTY
         DTZ  = CON1*DTZ
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
         MO   = CON1*MO
         DTXO = CON1*DTXO
         DTYO = CON1*DTYO
         DTZO = CON1*DTZO
C
         DU   = U - UO
         DV   = V - VO
         DW   = W - WO
         DM   = M - MO
         DDTX = DTX - DTXO
         DDTY = DTY - DTYO
         DDTZ = DTZ - DTZO
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*R
         ENDIF
         DFC = DQC*UO + QC*DU - MO*DDTX - DTX*DM
         DGC = DQC*VO + QC*DV - MO*DDTY - DTY*DM
         DHC = DQC*WO + QC*DW - MO*DDTZ - DTZ*DM
      ELSE
         QC   = 0.0
         U    = 0.0
         V    = 0.0
         W    = 0.0
         DTX  = 0.0
         DTY  = 0.0
         DTZ  = 0.0
         UO   = 0.0
         VO   = 0.0
         WO   = 0.0
         DTXO = 0.0
         DTYO = 0.0
         DTZO = 0.0
C
         DO KI = 1, KNODE(KT)
            KN = PCELL(KI)
            QC   = QC   + Q(KN)
            U    = U    + VEL(1,KN)
            V    = V    + VEL(2,KN)
            W    = W    + VEL(3,KN)
            DTX  = DTX  + DTAUDX(KN)
            DTY  = DTY  + DTAUDY(KN)
            DTZ  = DTZ  + DTAUDZ(KN)
            UO   = UO   + VELO(1,KN)
            VO   = VO   + VELO(2,KN)
            WO   = WO   + VELO(3,KN)
            DTXO = DTXO + DTAUDXO(KN)
            DTYO = DTYO + DTAUDYO(KN)
            DTZO = DTZO + DTAUDZO(KN)
         ENDDO
         CON1 = 1.0/FLOAT(KNODE(KT))
         QC   = CON1*QC
         U    = CON1*U
         V    = CON1*V
         W    = CON1*W
         DTX  = CON1*DTX
         DTY  = CON1*DTY
         DTZ  = CON1*DTZ
         UO   = CON1*UO
         VO   = CON1*VO
         WO   = CON1*WO
         DTXO = CON1*DTXO
         DTYO = CON1*DTYO
         DTZO = CON1*DTZO
C
         DU = U - UO
         DV = V - VO
         DW = W - WO
         DDTX = DTX - DTXO
         DDTY = DTY - DTYO
         DDTZ = DTZ - DTZO
         FSUM = 0.0
         GSUM = 0.0
         HSUM = 0.0
         DO KF = 1, NFACE(KT)
            K1 = PCELL(KFACE(1,KF,KT))
            K2 = PCELL(KFACE(2,KF,KT))
            K3 = PCELL(KFACE(3,KF,KT))
            K4 = PCELL(KFACE(4,KF,KT))
            IF (K1 .EQ. K4) THEN
              CON1 = 0.33333333
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3))*FM(3,KF,KC)
            ELSE
              CON1 = 0.25
              FSUM = FSUM + CON1*(F(K1)+F(K2)+F(K3)+F(K4))*FM(1,KF,KC)
              GSUM = GSUM + CON1*(G(K1)+G(K2)+G(K3)+G(K4))*FM(2,KF,KC)
              HSUM = HSUM + CON1*(H(K1)+H(K2)+H(K3)+H(K4))*FM(3,KF,KC)
            ENDIF
         ENDDO
         IF (RTLIM .GT. 0.0) THEN
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT*(1-QC/RTLIM)
         ELSE
           DQC = -( FSUM + GSUM + HSUM )*DT/VC(KC) + DT
         ENDIF
         DFC = DQC*UO + QC*DU - NU*DDTX
         DGC = DQC*VO + QC*DV - NU*DDTY
         DHC = DQC*WO + QC*DW - NU*DDTZ
      ENDIF
C
      RETURN
      END
