c	FORTRAN Residence Time Init routine
c
c	Copyright 1999, Massachusetts Institute of Technology.
c
c*****************************************************************
      subroutine FX_RTParams (itype, sm2, sm4, nux)
c*****************************************************************
c
c     where: itype  = 0 -> inviscid incompressible case
c                     1 -> viscous compressible case (non-constant viscosity)
c                     2 -> constant viscosity and density case
c                     3 -> inviscid compressible
c
      integer itype, nux
      real    sm2, sm4
c
c     include file
c
      include 'ResTime.inc'
c
c     local variables
c
      integer lints, mints, lptrs, KN, on
c
      call FX_SetRT(on, iopt, cmask, nNode, nCell, nNodeI, nCellI)
      if(on .eq. 0) then
        write(*,*) 'Residence Time Calculation Terminated'
        call FX_RTClean
        itype = -3
        return
      endif
c
c     start-up
c
      if(iopt .lt. 0 .or. iopt .gt. 2) then
        write(*,*) 'Error in Residence Time Initialization ',iopt
        call FX_SetRT(on, iopt, cmask, nNode, nCell, nNodeI, nCellI)
        itype = -2
        return
      endif
      if(itype .lt. 0 .or. itype .gt. 3) then
        write(*,*) 'Error in RT Initialization: RTTYPE =',itype
        call FX_SetRT(on, iopt, cmask, nNode, nCell, nNodeI, nCellI)
        itype = -2
        return
      endif
c
      mode  = itype
      NU    = nux
      SIG0  = 0.05
      SIG1  = 1.0
      SIG2  = sm2
      SIG4  = sm4
      RTLIM = 0.0
      ITER  = 0
c
      pINDSMO = 0
      pVn     = 0
      pQ      = 0
      pDQ     = 0
      pVELO   = 0
      pD2     = 0
      pSWITCH = 0
      pF      = 0
      pG      = 0
      pH      = 0
      pFM     = 0
      pVC     = 0
c
      pVEL    = 0
      pXYZ    = 0
      pRHO    = 0
      pRHOO   = 0
      pMU     = 0
      pMUO    = 0
      pDTAUDX = 0
      pDTAUDY = 0
      pDTAUDZ = 0
      pDTXO   = 0
      pDTYO   = 0
      pDTZO   = 0
      pKNSMO  = 0
      pWGHSMO = 0
c
c---- Allocate Arrays
c
      call FX_Sizes(lints, mints, lptrs)
c
      call FX_Malloc(2*lints*nNode, pINDSMO)
      if(pINDSMO .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on INDSMO:', nNode
        go to 99
      endif
c
      if(nNode .ne. nNodeI .or. iand(cmask,64).ne. 0) then
        call FX_Malloc(3*lints*nNode, pVEL)
        if(pVEL .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on VEL:', nNode
          go to 99
        endif
        call FX_Malloc(3*lints*nNode, pXYZ)
        if(pXYZ .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on XYZ:', nNode
          go to 99
        endif
      endif
c
      call FX_Malloc(lints*nNode, pVn)
      if(pVn .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on Vn:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pQ)
      if(pQ .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on Q:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pDQ)
      if(pDQ .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on DQ:', nNode
        go to 99
      endif
c
      call FX_Malloc(3*lints*nNode, pVELO)
      if(pVELO .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on VELO:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pD2)
      if(pD2 .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on D2:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pSWITCH)
      if(pSWITCH .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on SWITCH:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pF)
      if(pF .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on F:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pG)
      if(pG .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on G:', nNode
        go to 99
      endif
c
      call FX_Malloc(lints*nNode, pH)
      if(pH .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on H:', nNode
        go to 99
      endif
c
      call FX_Malloc(18*lints*nCell, pFM)
      if(pFM .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on FM:', nCell
        go to 99
      endif
c
      call FX_Malloc(lints*nCell, pVC)
      if(pVC .eq. 0) then
        write(*,*) ' FX_RTParams: Malloc ERROR on VC:', nCell
        go to 99
      endif
c
c
      if(mode .eq. 1 .or. mode .eq. 3) then
        call FX_Malloc(lints*nNode, pRHO)
        if(pRHO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on RHO:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pRHOO)
        if(pRHOO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on RHOO:', nNode
          go to 99
        endif
      endif
c
      if(mode .eq. 1) then
        call FX_Malloc(lints*nNode, pMU)
        if(pMU .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on MU:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pMUO)
        if(pMUO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on MUO:', nNode
          go to 99
        endif
      endif
c
      if(mode .eq. 1 .or. mode .eq. 2) then
        call FX_Malloc(lints*nNode, pDTAUDX)
        if(pDTAUDX .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDX:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pDTAUDY)
        if(pDTAUDY .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDY:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pDTAUDZ)
        if(pDTAUDZ .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDZ:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pDTXO)
        if(pDTXO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDXO:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pDTYO)
        if(pDTYO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDYO:', nNode
          go to 99
        endif
        call FX_Malloc(lints*nNode, pDTZO)
        if(pDTZO .eq. 0) then
          write(*,*) ' FX_RTParams: Malloc ERROR on DTAUDZO:', nNode
          go to 99
        endif
      endif
c
c---- INITIALIZE FIELD
c
      DO KN = 1, NNODE
         Q(KN) = 0.0
      ENDDO
c
      if (mode .eq. 1 .or. mode .eq. 2) then
        DO KN = 1, NNODE
           DTAUDXO(KN) = 0.
           DTAUDYO(KN) = 0.
           DTAUDZO(KN) = 0.
        ENDDO
      endif
c
c---- GET GRID/CELL STRUCTURE AND SETUP METRICS
c
      if(iopt .lt. 2) call FX_RTGetStruc
c
      return
c
c     error out -- cleanup and turn off
c
 99   call FX_RTClean
      call FX_SetRT(on, iopt, cmask, nNode, nCell, nNodeI, nCellI)
      itype = -1
      return
      end
c
c
C*****************************************************************************
       SUBROUTINE FX_RTGetStruc
C*****************************************************************************
C
C      GET GRID/CELL STRUCTURE AND INFO COMPUTED ON THIS DATA
C
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      call FX_GetXYZ(pXYZ)
c
c---- GET NODE STENCIL FROM CELLS
c
      if(pKNSMO  .ne. 0) call FX_Free(pKNSMO)
      if(pWGHSMO .ne. 0) call FX_Free(pWGHSMO)
      call FX_Stencil(INDSMO, pKNSMO, pWGHSMO)
c
c---- Calculate Weights for Smoothing
c
      call FX_Weights
c
c---- CELL AND NODE VOLUME
c
      call FX_Volume
c
c---- CELL SURFACE TERMS
c
      call FX_Metric
c
      return
      end
c
c
C*****************************************************************************
       SUBROUTINE FX_Volume
C*****************************************************************************
C
C      EVALUATE CELL AND NODE VOLUMES
C
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL    VOL, DIV
      REAL*8  XYZ1(3), XYZ2(3), XYZ3(3), XYZ4(3)
      INTEGER KN, KC, KT, ITC, NCN, NTC
      INTEGER K1, K2, K3, K4, NTET(2,4), TET(4,5,4)
      DATA    NTET/4,1, 5,2, 6,3, 8,5/
      DATA    TET /1,2,3,4, 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0,
     .             1,4,2,5, 4,3,2,5, 0,0,0,0, 0,0,0,0, 0,0,0,0,
     .             1,4,2,5, 2,4,3,5, 1,4,5,6, 0,0,0,0, 0,0,0,0,
     .             5,2,6,7, 5,1,2,4, 5,7,8,4, 2,3,7,4, 5,2,7,4/
C
C---- ZERO OUT NODE VOLUME
C
      DO KN = 1, NNODE
         VN(KN) = 0.
      ENDDO
C 
C---- CALCULATE CELL VOLUMES
C
      DO KC = 1, NCELL
         ITC    = FX_RTCELL(KC, PCELL)
         NCN    = NTET(1,ITC)
         NTC    = NTET(2,ITC)
         DIV    = 1.0/FLOAT(NCN)
         VC(KC) = 0.0
         DO KT = 1, NTC
            K1      = PCELL(TET(1,KT,ITC))
	    XYZ1(1) = XYZ(1,K1)
	    XYZ1(2) = XYZ(2,K1)
	    XYZ1(3) = XYZ(3,K1)
            K2      = PCELL(TET(2,KT,ITC))
	    XYZ2(1) = XYZ(1,K2)
	    XYZ2(2) = XYZ(2,K2)
	    XYZ2(3) = XYZ(3,K2)
            K3      = PCELL(TET(3,KT,ITC))
	    XYZ3(1) = XYZ(1,K3)
	    XYZ3(2) = XYZ(2,K3)
	    XYZ3(3) = XYZ(3,K3)
            K4      = PCELL(TET(4,KT,ITC))
	    XYZ4(1) = XYZ(1,K4)
	    XYZ4(2) = XYZ(2,K4)
	    XYZ4(3) = XYZ(3,K4)
            VOL     = XYZ1(3)*( XYZ2(1)*(XYZ3(2)-XYZ4(2)) -
     .                          XYZ3(1)*(XYZ2(2)-XYZ4(2)) +
     .                          XYZ4(1)*(XYZ2(2)-XYZ3(2)) ) -
     .                XYZ2(3)*( XYZ1(1)*(XYZ3(2)-XYZ4(2)) -
     .                          XYZ3(1)*(XYZ1(2)-XYZ4(2)) +
     .                          XYZ4(1)*(XYZ1(2)-XYZ3(2)) ) +
     .                XYZ3(3)*( XYZ1(1)*(XYZ2(2)-XYZ4(2)) -
     .                          XYZ2(1)*(XYZ1(2)-XYZ4(2)) +
     .                          XYZ4(1)*(XYZ1(2)-XYZ2(2)) ) -
     .                XYZ4(3)*( XYZ1(1)*(XYZ2(2)-XYZ3(2)) -
     .                          XYZ2(1)*(XYZ1(2)-XYZ3(2)) +
     .                          XYZ3(1)*(XYZ1(2)-XYZ2(2)) )
            IF (VOL .LT. 0.0) THEN
              WRITE(*,*) 'Volume < 0.0 for Cell #:',KC,'   Tet #:', KT,
     .                   VOL
            ENDIF
            VC(KC) = VC(KC) + 0.1666667*VOL
         ENDDO
         IF(VC(KC) .LT. 0.0) THEN
           WRITE(*,*) 'Volume < 0.0 for Cell #:',KC, VC(KC)
           STOP
         ENDIF
C
C------- Nodes volumes
C
         DO KT = 1, NCN
            KN     = PCELL(KT)
            VN(KN) = VN(KN) + VC(KC)*DIV
         ENDDO
C
      ENDDO
C
      RETURN
      END
c
c
C*****************************************************************************
       subroutine FX_Metric
C*****************************************************************************
C
C      EVALUATE CELL SURFACE TERMS
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      integer FX_RTCELL, PCELL(8)
c
c     local variables
c
      REAL*8  XYZ1(3), XYZ2(3), XYZ3(3), XYZ4(3)
      INTEGER KC, KT, KF, K1, K2, K3, K4, KFACE(4,6,4), NFACE(4)
c
      data NFACE/4, 5, 5, 6/
      data KFACE/1,3,2,1, 2,3,4,2, 1,4,3,1, 4,1,2,4, 0,0,0,0, 0,0,0,0,
     .           1,2,3,4, 2,5,3,2, 3,5,4,3, 4,5,1,4, 1,5,2,1, 0,0,0,0,
     .           1,2,3,4, 1,6,5,2, 3,5,6,4, 4,6,1,4, 2,5,3,2, 0,0,0,0,
     .           1,2,3,4, 6,7,3,2, 4,3,7,8, 1,4,8,5, 5,8,7,6, 1,5,6,2/
c
      DO KC = 1, NCELL
        KT = FX_RTCELL(KC, PCELL)
C
        DO KF = 1, NFACE(KT)
          K1      = PCELL(KFACE(1,KF,KT))
	  XYZ1(1) = XYZ(1,K1)
	  XYZ1(2) = XYZ(2,K1)
	  XYZ1(3) = XYZ(3,K1)
          K2      = PCELL(KFACE(2,KF,KT))
	  XYZ2(1) = XYZ(1,K2)
	  XYZ2(2) = XYZ(2,K2)
	  XYZ2(3) = XYZ(3,K2)
          K3      = PCELL(KFACE(3,KF,KT))
	  XYZ3(1) = XYZ(1,K3)
	  XYZ3(2) = XYZ(2,K3)
	  XYZ3(3) = XYZ(3,K3)
          K4      = PCELL(KFACE(4,KF,KT))
	  XYZ4(1) = XYZ(1,K4)
	  XYZ4(2) = XYZ(2,K4)
	  XYZ4(3) = XYZ(3,K4)
C
          FM(1,KF,KC) = 0.5*( (XYZ3(2)-XYZ1(2))*(XYZ2(3)-XYZ4(3))-
     .                        (XYZ2(2)-XYZ4(2))*(XYZ3(3)-XYZ1(3)) )
          FM(2,KF,KC) = 0.5*( (XYZ3(3)-XYZ1(3))*(XYZ2(1)-XYZ4(1))-
     .                        (XYZ2(3)-XYZ4(3))*(XYZ3(1)-XYZ1(1)) )
          FM(3,KF,KC) = 0.5*( (XYZ3(1)-XYZ1(1))*(XYZ2(2)-XYZ4(2))-
     .                        (XYZ2(1)-XYZ4(1))*(XYZ3(2)-XYZ1(2)) )
         ENDDO
C
       ENDDO
C
       RETURN
       END
c
c
C*****************************************************************************
       subroutine FX_RTClean
C*****************************************************************************
C
C      Clean up memory allocations
C.............................................................................
c
      implicit none
c
c     include files
c
      include 'ResTime.inc'
c
      if(nNode .ne. nNodeI .or. iand(cmask,64).ne. 0) then
        if(pVEL  .ne. 0) call FX_Free(pVEL)
        if(pXYZ  .ne. 0) call FX_Free(pXYZ)
      endif
      if(pVn     .ne. 0) call FX_Free(pVn)
      if(pQ      .ne. 0) call FX_Free(pQ)
      if(pDQ     .ne. 0) call FX_Free(pDQ)
      if(pVELO   .ne. 0) call FX_Free(pVELO)
      if(pD2     .ne. 0) call FX_Free(pD2)
      if(pF      .ne. 0) call FX_Free(pF)
      if(pG      .ne. 0) call FX_Free(pG)
      if(pH      .ne. 0) call FX_Free(pH)
      if(pFM     .ne. 0) call FX_Free(pFM)
      if(pVC     .ne. 0) call FX_Free(pVC)
c
      if(pRHO    .ne. 0) call FX_Free(pRHO)
      if(pRHOO   .ne. 0) call FX_Free(pRHOO)
      if(pMU     .ne. 0) call FX_Free(pMU)
      if(pMUO    .ne. 0) call FX_Free(pMUO)
      if(pDTAUDX .ne. 0) call FX_Free(pDTAUDX)
      if(pDTAUDY .ne. 0) call FX_Free(pDTAUDY)
      if(pDTAUDZ .ne. 0) call FX_Free(pDTAUDZ)
      if(pDTXO   .ne. 0) call FX_Free(pDTXO)
      if(pDTYO   .ne. 0) call FX_Free(pDTYO)
      if(pDTZO   .ne. 0) call FX_Free(pDTZO)
      if(pSWITCH .ne. 0) call FX_Free(pSWITCH)
      if(pINDSMO .ne. 0) call FX_Free(pINDSMO)
      if(pKNSMO  .ne. 0) call FX_Free(pKNSMO)
      if(pWGHSMO .ne. 0) call FX_Free(pWGHSMO)
c
      return
      end
