c
c     Residence time iteration module
c*****************************************************************
      subroutine FX_RTUpdate (timeNew)
c*****************************************************************
c
c
      implicit none
c
      real timeNew
c
c     include file
c
      include 'ResTime.inc'
c
c     local variables
c
      POINTER (pNSurf, NSurf),     (pSurf, Surf),   (pSCel, SCel)
      INTEGER          NSurf(2,1),         Surf(1),         SCel(4,1)
      INTEGER KN, nBC, kBC, IST
c
      timeNow = timeNew
      iter    = iter + 1
c
      if(iter .eq. 1) then
         timePrev = timeNow
         call FX_GetVel(pVel)
         do kn = 1, nnode
           velo(1,kn) = vel(1,kn)
           velo(2,kn) = vel(2,kn)
           velo(3,kn) = vel(3,kn)
         enddo
         if (mode .eq. 1 .or. mode .eq. 3) call FX_GetScal(1, rhoo)
         if (mode .eq. 1) call FX_GetScal(4, muo)
         return
      endif
C
      if(iopt .ge. 2) call FX_RTGetStruc
C
C---- EVALUATE TIME STEP 
C
      DT = timeNow - timePrev
      timePrev = timeNow
C
C--- 2nd DIFFERENCE SMOOTHING
C
      call FX_Smooth2
C
C---- OBTAIN VELOCITY DATA
C
      call FX_GetVel(pVel)
C
C---- FLUX EVALUATION
C
      if (mode .eq. 0) then
        call FX_RTFlux
C
      elseif (mode .eq. 1) then
C
C------ OBTAIN RHO AND MU, AND COMPUTE GRADIENTS OF TAU
        call FX_GetScal(1, rho)
        call FX_GetScal(4, mu)
        call FX_GradTau
C
        call FX_RTFluxVisComp
C
      elseif (mode .eq. 2) then
C
C------ COMPUTE GRADIENT OF TAU
        call FX_GradTau
C
        call FX_RTFluxConVisDens
C
      elseif (mode .eq. 3) then
C
C------ OBTAIN RHO 
        call FX_GetScal(1, rho)
C
        call FX_RTFluxInvComp
C
      endif
C
C---- BOUNDARY CONDITIONS
C
      call FX_BCsRT
C
C---- MODIFICATION OF RESIDUALS (i.e. DQ) BY USER
C
      call FXmodifyRT(q, dq)
C
C---- UPDATE STATE VECTOR FOR NEXT ITERATION
C
      DO KN = 1, NNODE
         Q(KN) = Q(KN) + DQ(KN)
         VELO(1,KN) = VEL(1,KN)
         VELO(2,KN) = VEL(2,KN)
         VELO(3,KN) = VEL(3,KN)
      ENDDO
C
      if (mode .eq. 1) then
        DO KN = 1, NNODE
           RHOO(KN)    = RHO(KN)
           MUO(KN)     = MU(KN)
           DTAUDXO(KN) = DTAUDX(KN)
           DTAUDYO(KN) = DTAUDY(KN)
           DTAUDZO(KN) = DTAUDZ(KN)
        ENDDO
C
      elseif (mode .eq. 2) then
        DO KN = 1, NNODE
           DTAUDXO(KN) = DTAUDX(KN)
           DTAUDYO(KN) = DTAUDY(KN)
           DTAUDZO(KN) = DTAUDZ(KN)
        ENDDO
C
      elseif (mode .eq. 3) then
        DO KN = 1, NNODE
           RHOO(KN) = RHO(KN)
        ENDDO
c
      endif
C
C---- IMPOSE INLET BOUNDARY CONDITIONS
C
      CALL FX_GetBC(nBC, pNSurf, pSurf, pSCel)
C
      IST = 1
      DO kBC = 1, nBC
         IF(NSurf(2,kBC) .EQ. 1) THEN
            DO KN = IST, NSurf(1,kBC)
              Q(SCel(1,KN)) = 0.0
              Q(SCel(2,KN)) = 0.0
              Q(SCel(3,KN)) = 0.0
              if (SCel(4,KN) .ne. 0) Q(SCel(4,KN)) = 0.0
            ENDDO
         ENDIF
         IST = NSurf(1,kBC) + 1
      ENDDO
C
      return
      end
c
c
c**********************************************************
      subroutine FX_RTGet(resTime)
c**********************************************************
c     
c     Returns the residence time for all nodes. 
c     resTime has length NNODE.
c----------------------------------------------------------
c
      implicit none
c
      real resTime(*)
c
c     include files
c
      include 'ResTime.inc'
c
c     local variables
c
      integer i
c
      if (mode .eq. 1 .or. mode .eq. 3) then
         do i = 1, nNodeI
            resTime(i) = Q(i)/rho(i)
         enddo
      else
         do i = 1, nNodeI
            resTime(i) = Q(i)
         enddo
      endif
c
      return
      end
