/*
 * FX_Update.c
 *
 * FX module that Updates the FX system data for transient cases.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "FX_types.h"



#ifdef WIN32

#define PREFIX __stdcall

#else

#define PREFIX

#ifdef UNDERSCORE

#define FXSTRUC      fxstruc_
#define FXCELL       fxcell_
#define FXCELLPTR    fxcellptr_
#define FXSURFACE    fxsurface_
#define FXSURFACEPTR fxsurfaceptr_
#define FXGRID       fxgrid_
#define FXGRIDPTR    fxgridptr_
#define FXGRIDP3D    fxgridp3d_
#define FXBLANK      fxblank_
#define FXBLANKPTR   fxblankptr_

#define FX_RTUPDATE  fx_rtupdate_
#define FX_UPDATE    fx_update_
#define FX_SETRT     fx_setrt_

#else

#define FXSTRUC      fxstruc
#define FXCELL       fxcell
#define FXCELLPTR    fxcellptr
#define FXSURFACE    fxsurface
#define FXSURFACEPTR fxsurfaceptr
#define FXGRID       fxgrid
#define FXGRIDPTR    fxgridptr
#define FXGRIDP3D    fxgridp3d
#define FXBLANK      fxblank
#define FXBLANKPTR   fxblankptr

#define FX_RTUPDATE  fx_rtupdate
#define FX_UPDATE    fx_update
#define FX_SETRT     fx_setrt

#endif
#endif


/* globals */

  extern FXglobal *FX_global;


/* externals */

  extern void PREFIX
         FXSTRUC(int *, int *, int *, int *, int *, int *, int *, int *, int *,
                 int *, int *);
  extern void PREFIX
         FXCELL   (FXtet *,  FXpyra *,  FXprism *,  FXhexa *,  FXhalo *);
  extern void PREFIX
         FXCELLPTR(FXtet **, FXpyra **, FXprism **, FXhexa **, FXhalo *);
  extern void PREFIX
         FXSURFACE   (FXbc *, int *,  FXfacet *);
  extern void PREFIX
         FXSURFACEPTR(FXbc *, int **, FXfacet **);
  extern void PREFIX
         FXGRID   (FXcoor *,  FXcoor *);
  extern void PREFIX
         FXGRIDPTR(FXcoor **, FXcoor *);
  extern void PREFIX
         FXGRIDP3D(float **, float **, float **, FXcoor *);
  extern void PREFIX
         FXBLANK   (int *);
  extern void PREFIX
         FXBLANKPTR(int **);

  extern void PREFIX
         FX_RTUPDATE(float *);

  extern int FX_Check();



void PREFIX
FX_UPDATE(float *time)
{
  int i, ni, nj, nk, lnode, lcell, *blks;

  if (FX_global == NULL) {
    printf(" FX Error: NOT Initialized (FX_Update)!\n");
    return;
  }

  switch(FX_global->iopt) {
  case 0:
    printf(" FX Error: Steady-State Case (FX_Update)!\n");
    return;

  case 3:

    /* get new sizes */

    blks = (int *) FX_global->block;
    FXSTRUC(&lnode,             &FX_global->nhalo,  &FX_global->ntets,
            &FX_global->npyra,  &FX_global->nprism, &FX_global->nhexa,
            &FX_global->nblock,  blks,              &FX_global->nhcell,
            &FX_global->nfacet, &FX_global->nbc);

    /* are we static? */

    if (lnode == -1) {
      if ((FX_global->cmask & 4) == 0) {
        FXGRID   ( FX_global->coor, FX_global->hcoor);
      } else {
        if ((FX_global->cmask & 64) == 0) {
          FXGRIDPTR(&FX_global->coor, FX_global->hcoor);
        } else {
          FXGRIDP3D(&FX_global->xcoor, &FX_global->ycoor, &FX_global->zcoor,
                     FX_global->hcoor);
        }
      }
      if (FX_global->nblock > 0)
        if ((FX_global->cmask & 16) != 0)
          if ((FX_global->cmask & 32) == 0) {
            FXBLANK   ( FX_global->iblank);
          } else {
            FXBLANKPTR(&FX_global->iblank);
          }
      FX_global->time = *time;
      return;
    }

    FX_global->nnode = lnode;
    if (FX_global->nblock > FX_global->mblock) {
      printf(" FX Error: # Blocks = %d > %d (FX_Update)!\n", FX_global->nblock,
             FX_global->mblock);
      FX_global->status = -1;
      return;
    }

    /* free up all allocated memory */

    if (((FX_global->cmask & 1) == 0) && ((FX_global->cmask & 128) == 0)) {
      if (FX_global->tets  != NULL) free(FX_global->tets);
      if (FX_global->pyra  != NULL) free(FX_global->pyra);
      if (FX_global->prism != NULL) free(FX_global->prism);
      if (FX_global->hexa  != NULL) free(FX_global->hexa);
    }
    if (FX_global->bc    != NULL) free(FX_global->bc);
    if ((FX_global->cmask & 2) == 0) {
      if (FX_global->facet != NULL) free(FX_global->facet);
      if (FX_global->scel  != NULL) free(FX_global->scel);
    }
    if ((FX_global->cmask & 4) == 0)
      if (FX_global->coor != NULL) free(FX_global->coor);
    if (FX_global->scalar != NULL) free(FX_global->scalar);
    if ((FX_global->cmask & 8) == 0)
      if (FX_global->vel != NULL) free(FX_global->vel);
    if ((FX_global->cmask & 32) == 0)
      if (FX_global->iblank != NULL) free(FX_global->iblank);
    if (FX_global->hcoor   != NULL) free(FX_global->hcoor);
    if (FX_global->hscalar != NULL) free(FX_global->hscalar);
    if (FX_global->hvel    != NULL) free(FX_global->hvel);
    if (FX_global->hcell   != NULL) free(FX_global->hcell);
    FX_global->tets    = NULL;
    FX_global->pyra    = NULL;
    FX_global->prism   = NULL;
    FX_global->hexa    = NULL;
    FX_global->bc      = NULL;
    FX_global->facet   = NULL;
    FX_global->scel    = NULL;
    FX_global->coor    = NULL;
    FX_global->xcoor   = NULL;
    FX_global->ycoor   = NULL;
    FX_global->zcoor   = NULL;
    FX_global->scalar  = NULL;
    FX_global->vel     = NULL;
    FX_global->velx    = NULL;
    FX_global->vely    = NULL;
    FX_global->velz    = NULL;
    FX_global->iblank  = NULL;
    FX_global->hcoor   = NULL;
    FX_global->hscalar = NULL;
    FX_global->hvel    = NULL;
    FX_global->hcell   = NULL;
    FX_global->status  = 0;

    /* expand the block storage */

    for (i = FX_global->nblock-1; i >= 0; i--) {
      FX_global->block[i].ni = blks[3*i  ];
      FX_global->block[i].nj = blks[3*i+1];
      FX_global->block[i].nk = blks[3*i+2];
    }
    lnode = FX_global->nnode;
    lcell = FX_global->ntets + FX_global->npyra + FX_global->nprism +
            FX_global->nhexa;
    FX_global->total_dis = lcell;
    for (i = 0; i < FX_global->nblock; i++) {
      ni = FX_global->block[i].ni;
      nj = FX_global->block[i].nj;
      nk = FX_global->block[i].nk;
      lcell += (ni-1)*(nj-1)*(nk-1);
      lnode += ni*nj*nk;
      FX_global->block[i].last_cell = lcell;
      FX_global->block[i].last_node = lnode;
    }
    FX_global->total_cells = lcell;
    FX_global->total_nodes = lnode;

    if ((FX_global->cmask & 128) == 0) {

      /* disjoint elements */

      if (FX_global->nhcell > 0) {
        FX_global->hcell = (FXhalo *) malloc(FX_global->nhcell*sizeof(FXhalo));
        if (FX_global->hcell == NULL) {
          printf(" FX Error: Allocation of Halo Cells (FX_Update)!\n");
          FX_global->status = -3;
          return;
        }
      }

      if ((FX_global->cmask & 1) == 0) {
        if (FX_global->ntets > 0) {
          FX_global->tets = (FXtet *) malloc(FX_global->ntets*sizeof(FXtet));
          if (FX_global->tets == NULL) {
            printf(" FX Error: Allocation of Tetrahedra (FX_Update)!\n");
            FX_global->status = -3;
            return;
          }
        }
        if (FX_global->npyra > 0) {
          FX_global->pyra = (FXpyra *) malloc(FX_global->npyra*sizeof(FXpyra));
          if (FX_global->pyra == NULL) {
            printf(" FX Error: Allocation of Pyramids (FX_Update)!\n");
            FX_global->status = -3;
            return;
          }
        }
        if (FX_global->nprism > 0) {
          FX_global->prism = (FXprism *)
                             malloc(FX_global->nprism*sizeof(FXprism));
          if (FX_global->prism == NULL) {
            printf(" FX Error: Allocation of Prisms (FX_Update)!\n");
            FX_global->status = -3;
            return;
          }
        }
        if (FX_global->nhexa > 0) {
          FX_global->hexa = (FXhexa *) malloc(FX_global->nhexa*sizeof(FXhexa));
          if (FX_global->hexa == NULL) {
            printf(" FX Error: Allocation of Hexahedra (FX_Update)!\n");
            FX_global->status = -3;
            return;
          }
        }
        if (FX_global->total_dis + FX_global->nhcell > 0)
          FXCELL   ( FX_global->tets,  FX_global->pyra,  FX_global->prism,
                     FX_global->hexa,  FX_global->hcell);
      } else {
        if (FX_global->total_dis + FX_global->nhcell > 0)
          FXCELLPTR(&FX_global->tets, &FX_global->pyra, &FX_global->prism,
                    &FX_global->hexa,  FX_global->hcell);
      }
    }

    /* surface data */

    if (FX_global->nbc > 0) {
      FX_global->bc = (FXbc *) malloc(FX_global->nbc*sizeof(FXbc));
      if (FX_global->bc == NULL) {
        printf(" FX Error: Allocation of Boundary Groups (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
      if ((FX_global->cmask & 2) == 0) {
        FX_global->facet = (FXfacet *)malloc(FX_global->nfacet*sizeof(FXfacet));        if (FX_global->facet == NULL) {
          printf(" FX Error: Allocation of Surface Facets (FX_Update)!\n");
          FX_global->status = -3;
          return;
        }
        FX_global->scel = (int *) malloc(FX_global->nfacet*sizeof(int));
        if (FX_global->scel == NULL) {
          printf(" FX Error: Allocation of Surface Cells (FX_Update)!\n");
          FX_global->status = -3;
          return;
        }
        FXSURFACE   (FX_global->bc,  FX_global->scel,  FX_global->facet);
      } else {
        FXSURFACEPTR(FX_global->bc, &FX_global->scel, &FX_global->facet);
      }
    }

    /* halo data */

    if (FX_global->nhalo > 0) {
      FX_global->hcoor = (FXcoor *) malloc(FX_global->nhalo*sizeof(FXcoor));
      if (FX_global->hcoor == NULL) {
        printf(" FX Error: Allocation of Halo Coordinates (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
      FX_global->hvel = (FXcoor *) malloc(FX_global->nhalo*sizeof(FXcoor));
      if (FX_global->hvel == NULL) {
        printf(" FX Error: Allocation of Halo Velocities (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
      FX_global->hscalar = (float *) malloc(FX_global->nhalo*sizeof(float));
      if (FX_global->hscalar == NULL) {
        printf(" FX Error: Allocation of Halo Scalars (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
    }

    /* node data */

    if ((FX_global->cmask & 4) == 0) {
      FX_global->coor = (FXcoor *)malloc(FX_global->total_nodes*sizeof(FXcoor));
      if (FX_global->coor == NULL) {
        printf(" FX Error: Allocation of Coordinates (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
    }
    if ((FX_global->cmask & 8) == 0) {
      FX_global->vel = (FXcoor *) malloc(FX_global->total_nodes*sizeof(FXcoor));
      if (FX_global->vel == NULL) {
        printf(" FX Error: Allocation of Velocity Field (FX_Update)!\n");
        FX_global->status = -3;
        return;
      }
    }
    FX_global->scalar = (float *) malloc(FX_global->total_nodes*sizeof(float));
    if (FX_global->scalar == NULL) {
      printf(" FX Error: Allocation of Scalar Field (FX_Update)!\n");
      FX_global->status = -3;
      return;
    }

    /* Iblanking */

    if (FX_global->nblock > 0)
      if ((FX_global->cmask & 16) != 0)
        if ((FX_global->cmask & 32) == 0) {
          lnode = FX_global->total_nodes - FX_global->nnode;
          FX_global->iblank = (int *) malloc(lnode*sizeof(int));
          if (FX_global->iblank == NULL) {
            printf(" FX Error: Allocation of IBlank (FX_Update)!\n");
            FX_global->status = -3;
            return;
          }
        }

    /* Structure Check */

    if (FX_Check() != 0) {
      FX_global->status = -5;
      return;
    }

  case 2:
    if ((FX_global->cmask & 4) == 0) {
      FXGRID(FX_global->coor, FX_global->hcoor);
    } else {
      if ((FX_global->cmask & 64) == 0) {
        FXGRIDPTR(&FX_global->coor, FX_global->hcoor);
      } else {
        FXGRIDP3D(&FX_global->xcoor, &FX_global->ycoor, &FX_global->zcoor,
                   FX_global->hcoor);
      }
    }
    if (FX_global->nblock > 0)
      if ((FX_global->cmask & 16) != 0)
        if ((FX_global->cmask & 32) == 0) {
          FXBLANK   ( FX_global->iblank);
        } else {
          FXBLANKPTR(&FX_global->iblank);
        }

  case 1:
    FX_global->time  = *time;
    FX_global->sload = 0;
    FX_global->vload = 0;
  }

  if (FX_global->rt != 0) FX_RTUPDATE(time);
}


void PREFIX
FX_SETRT(int *on, int *opt, int *cmask, int *nodes, int *cells, int *inodes,
         int *icells)
{
  if (FX_global == NULL) return;

  if (FX_global->rt == 0) {
    FX_global->rt = 1;
  } else {
    FX_global->rt = 0;
  }
  *on     = FX_global->rt;
  *opt    = FX_global->iopt;
  *cmask  = FX_global->cmask;
  *nodes  = FX_global->total_nodes + FX_global->nhalo;
  *cells  = FX_global->total_cells + FX_global->nhcell;
  *inodes = FX_global->total_nodes;
  *icells = FX_global->total_cells;
}
