/*
 * FX_VortexSubs.c
 *
 * FX module that supports connecting Vortex Core segments.
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "FX_VC.h"


static int
FX_AddSegVC(int face1, int face2, int term1, int term2, segs *segment)
{
  int stat, *tmp, j;

  /* first in segment */
  if (segment->mface == 0) {
    segment->faces = (int *) malloc(VCFSEG*sizeof(int));
    if (segment->faces == NULL) return -2;
    segment->mface = VCFSEG;
    segment->nface = 2;
    segment->faces[0] = face1;
    segment->faces[1] = face2;
    segment->start    = term1;
    segment->finish   = term2;
    return 0;
  }

  /* are we in this segment ? */
  stat = -1;
  j    = segment->nface-1;
  if (segment->faces[0] == face1) {
    if (segment->start  == NOTFILLED) stat = 0;
  } else if (segment->faces[j] == face1) {
    if (segment->finish == NOTFILLED) stat = 1;
  } else if (segment->faces[0] == face2) {
    if (segment->start  == NOTFILLED) stat = 2;
  } else if (segment->faces[j] == face2) {
    if (segment->finish == NOTFILLED) stat = 3;
  }

  if(stat == -1) return -1;

  if (segment->nface >= segment->mface) {
    j = segment->nface + VCFSEG;
    tmp = (int *) realloc(segment->faces, j*sizeof(int));
    if (tmp == NULL) return -2;
    segment->faces = tmp;
    segment->mface = j;
  }

  if ((stat == 0) || (stat == 2)) {
    for (j = segment->nface; j >= 1; j--) 
      segment->faces[j] = segment->faces[j-1];
    if (stat == 0) {
      segment->faces[0] = face2;
      segment->start    = term2;
    } else {
      segment->faces[0] = face1;
      segment->start    = term1;
    }
  } else {
    j = segment->nface;
    if (stat == 1) {
      segment->faces[j] = face2;
      segment->finish   = term2;
    } else {
      segment->faces[j] = face1;
      segment->finish   = term1;
    }
  }
  segment->nface++;

  return 0;
}


static int
FX_CatSegVC(segs *seg1, segs *seg2)
{
  int i, nf, nf1, nf2, *tmp;

  nf1 = seg1->nface-1;
  nf2 = seg2->nface-1;
  nf  = nf1 + nf2;
  if (nf+1 > seg1->mface) {
    tmp = (int *) realloc(seg1->faces, (nf+1)*sizeof(int));
    if (tmp == NULL) return -2;
    seg1->faces = tmp;
    seg1->mface = nf+1;
  }
  if ((seg1->faces[0] == seg2->faces[0]) && 
      (seg1->start    == seg2->start)) {
    if (seg1->start < 0) nf++;
    for (i = 0; i < seg1->nface; i++) seg1->faces[nf-i-1] = seg1->faces[nf1-i];
    for (i = 0; i < nf2; i++)         seg1->faces[i] = seg2->faces[nf2-i];
    seg1->start = seg2->finish;
  } else if ((seg1->faces[0] == seg2->faces[nf2]) &&
             (seg1->start    == seg2->finish)) {
    if (seg1->start < 0) nf++;
    for (i = 0; i < seg1->nface; i++) seg1->faces[nf-i-1] = seg1->faces[nf1-i];
    for (i = 0; i < nf2; i++)         seg1->faces[i] = seg2->faces[i];
    seg1->start = seg2->start;
  } else if ((seg1->faces[nf1] == seg2->faces[0]) &&
             (seg1->finish     == seg2->start)) {
    if (seg1->finish < 0) nf++;
    for (i = 0; i < nf2; i++) seg1->faces[nf-i-1] = seg2->faces[nf2-i];
    seg1->finish = seg2->finish;
  } else if ((seg1->faces[nf1] == seg2->faces[nf2]) &&
             (seg1->finish     == seg2->finish)) {
    if (seg1->finish < 0) nf++;
    for (i = 0; i < nf2; i++) seg1->faces[nf-i-1] = seg2->faces[i];
    seg1->finish = seg2->start;
  } else {
    printf(" FX_CatSegVC Error: No Match!\n");
    return -1;
  }
  seg1->nface = nf;
  seg2->nface = 0;
  return 0;
}


/*
 *  make the core segments 
 */

void
FX_SegmentVC(int nnodes, vcf *pack)
  /* note: reuse node array for tet storage */
{
  int  i, j, k, m, n, iend, iface, itet, ntet, ntetp, nhit, flag, nsegs;
  int  msegs, seg[4][3], iface1, iface2, iterm1, iterm2, norder, iorder;
  int  leaf[VCFHUNK][3][2], up[VCFHUNK][2], ileaf, nleaf, nleafp, nl;
  segs *segments, *tmp;

  pack->nsegs = 0;

start:
  /* find first tet */
  for (ntet = i = 0; i < pack->ntets;  i++)
    if (pack->tets[i].nface > 0) {

      /* collect all tets that touch the first */
      itet = i;
      if (ntet == nnodes) return;
      pack->tets[itet].nface = -pack->tets[itet].nface;
      if (pack->tets[itet].nface == -1) continue;
      pack->nodes[ntet] = itet;
      ntet++;
      ntetp = ntet;
next:
      for (j = 0; j < -pack->tets[itet].nface; j++) {
        iface = pack->tets[itet].face[j];
        nhit  = 4;
        if (pack->table[iface].itet[3] == NOTFILLED) nhit = 3;
        if (pack->table[iface].itet[2] == NOTFILLED) nhit = 2;
        if (pack->table[iface].itet[1] == NOTFILLED) nhit = 1;
        for (k = 0; k < nhit; k++) {
          for (flag = -1, m = 0; m < ntet; m++) 
            if (pack->table[iface].itet[k] == pack->nodes[m]) {
              flag = m;
              break;
            }
          if (flag == -1) {
            if (ntet == nnodes) return;
            m = pack->table[iface].itet[k];
            if (pack->tets[m].nface > 0) {
              pack->tets[m].nface = -pack->tets[m].nface;
              if (pack->tets[m].nface != -1) {
                pack->nodes[ntet] = m;
                ntet++;
              }
            }
          }
        }
      }
      if (ntetp != ntet) {
        itet = pack->nodes[ntetp];
        ntetp++;
        goto next;
      }

      /* tet list is now complete in pack->nodes */

      if(ntet <= MINFACE) goto start;

      segments = (segs *) malloc(VCFSEG*sizeof(segs));
      if (segments == NULL) return;
      nsegs = 0;
      msegs = VCFSEG;
      for (j = 0; j < ntet; j++) {
        itet   = pack->nodes[j];
        iface  = -pack->tets[itet].nface;
        if (iface == 2) {

          /* normal tet with 2 faces hit */

          iface1 = pack->tets[itet].face[0];
          iface2 = pack->tets[itet].face[1];
          iterm1 = iterm2 = NOTFILLED;
          if (pack->table[iface1].itet[2] != NOTFILLED) iterm1 = -2;
          if (pack->table[iface2].itet[2] != NOTFILLED) iterm2 = -2;
          for (flag = -1, k = 0; k < nsegs; k++) {
            nhit = FX_AddSegVC(iface1,iface2,iterm1,iterm2, &segments[k]);
            if (nhit == 0) {
              flag = k;
              break;
            }
            if (nhit == -2) printf(" FX_AddSegVC Memory Error!\n");
          }
          /* add a new segment */
          if (flag == -1) {
            if (nsegs >= msegs) {
              k = msegs + VCFSEG;
              tmp = (segs *) realloc(segments, k*sizeof(segs));
              if (tmp == NULL) return;
              segments = tmp;
              msegs    = k;
            }
            segments[nsegs].mface = 0;
            nhit = FX_AddSegVC(iface1,iface2,iterm1,iterm2, &segments[nsegs]);
            if (nhit == -2) printf(" FX_AddSegVC Memory Error (init)!\n");
            nsegs++;
          }
        } else if (iface > 2) {

          /* tet with 3 or 4 faces */

          iface2 = NOTFILLED;
          iterm2 = itet;
          for (m = 0; m < iface; m++) {
            iface1 = pack->tets[itet].face[m];
            iterm1 = NOTFILLED;
            if (pack->table[iface1].itet[2] != NOTFILLED) iterm1 = -2;
            for (flag = -1, k = 0; k < nsegs; k++) {
              nhit = FX_AddSegVC(iface1,iface2,iterm1,iterm2, &segments[k]);
              if (nhit == 0) {
                flag = k;
                break;
              }
              if (nhit == -2) printf(" FX_AddSegVC Memory Error!\n");
            }
            /* add a new segment */
            if (flag == -1) {
              if (nsegs >= msegs) {
                k = msegs + VCFSEG;
                tmp = (segs *) realloc(segments, k*sizeof(segs));
                if (tmp == NULL) return;
                segments = tmp;
                msegs    = k;
              }
              segments[nsegs].mface = 0;
              nhit = FX_AddSegVC(iface1,iface2,iterm1,iterm2, &segments[nsegs]);
              if (nhit == -2) printf(" FX_AddSegVC Memory Error (init)!\n");
              nsegs++;
            }
          }
        }
      }

loop:
      for (n = flag = j = 0; j < nsegs; j++) {
        if (segments[j].nface == 0) continue;
        n++;
        if (segments[j].nface > flag) flag = segments[j].nface;
/*
        printf(" Segment: %d (%d), start = %d, finish = %d\n", n-1, j,
                 segments[j].start, segments[j].finish);
        for (k = 0; k < segments[j].nface; k+=14) {
          printf("        %4d", segments[j].faces[k]);
          iface = k+14;
          if(iface > segments[j].nface) iface = segments[j].nface;
          for (m = k+1; m < iface; m++) printf(" %4d", segments[j].faces[m]);
          printf("\n");
        }
*/
      }

      if (n != 1) {

        /* remove segments that have same internal endings */

        for (j = 0; j < nsegs; j++) {
          if (segments[j].nface == 0) continue;
          iterm1 = segments[j].start;
          iterm2 = segments[j].finish;
          if ((iterm1 == NOTFILLED) || (iterm2 == NOTFILLED)) continue;
          iface1 = segments[j].faces[0];
          iface2 = segments[j].faces[segments[j].nface-1];
          for (k = j+1; k < nsegs; k++) {
            if (segments[k].nface  == 0) continue;
            if (segments[k].start  == NOTFILLED) continue;
            if (segments[k].finish == NOTFILLED) continue;
            if (segments[k].start  == iterm1)
              if ((iface1 == segments[k].faces[0]) &&
                  (iterm2 == segments[k].finish) &&
                  (iface2 == segments[k].faces[segments[k].nface-1])) {
                segments[k].nface = 0;
                continue;
              }
            if (segments[k].start == iterm2)
              if ((iface2 == segments[k].faces[0]) &&
                  (iterm1 == segments[k].finish) &&
                  (iface1 == segments[k].faces[segments[k].nface-1])) {
                segments[k].nface = 0;
              }
          }
        }

        /* concatinate segments based on 2 hits */

        for (j = 0; j < nsegs; j++) {
          if (segments[j].nface == 0) continue;

          /* look on start */
          iterm1 = segments[j].start;
          iface1 = segments[j].faces[0];
          if (iterm1 != NOTFILLED) {
            for (flag = -1, k = 0; k < nsegs; k++) {
              if (k == j) continue;
              if (segments[k].nface == 0) continue;
              if ((iterm1 == segments[k].start) &&
                  (iface1 == segments[k].faces[0]))
                if (flag == -1) {
                  flag = k;
                } else {
                  flag = -2;
                }
              if ((iterm1 == segments[k].finish) &&
                  (iface1 == segments[k].faces[segments[k].nface-1]))
                if (flag == -1) {
                  flag = k;
                } else {
                  flag = -2;
                }
            }
            if (flag == -1) {
              segments[j].start = NOTFILLED;
              if (iterm1 >= 0) {
                segments[j].nface--;
                for (k = 0; k < segments[j].nface; k++) 
                  segments[j].faces[k] = segments[j].faces[k+1];
              }
            }
            if (flag >= 0) {
              FX_CatSegVC(&segments[j], &segments[flag]);
              goto loop;
            }
          }

          /* look at finish */
          iterm1 = segments[j].finish;
          iface1 = segments[j].faces[segments[j].nface-1];
          if (iterm1 != NOTFILLED) {
            for (flag = -1, k = 0; k < nsegs; k++) {
              if (k == j) continue;
              if (segments[k].nface == 0) continue;
              if ((iterm1 == segments[k].start) &&
                  (iface1 == segments[k].faces[0]))
                if (flag == -1) {
                  flag = k;
                } else {
                  flag = -2;
                }
              if ((iterm1 == segments[k].finish) &&
                  (iface1 == segments[k].faces[segments[k].nface-1]))
                if (flag == -1) {
                  flag = k;
                } else {
                  flag = -2;
                }
            }
            if (flag == -1) {
              segments[j].finish = NOTFILLED;
              if (iterm1 >= 0) segments[j].nface--;
            }
            if (flag >= 0) {
              FX_CatSegVC(&segments[j], &segments[flag]);
              goto loop;
            }
          }
        }

        /* concatinate segments based on 3 or more hits */

        /* order segments based on length */
        for (norder = j = 0; j < nsegs; j++) 
          if (segments[j].nface > 0) {
            for (flag = m = 0; m < norder; m++) {
              k = segments[m].order;
              if (segments[k].nface > segments[j].nface) {
                for (iend = norder; iend > m; iend--)
                  segments[iend].order = segments[iend-1].order;
                segments[m].order = j;
                norder++;
                flag = 1;
                break;
              }
            }
            if (flag == 0) {
              segments[norder].order = j;
              norder++;
            }
          }
/*
	printf(" order (len = %d):", norder);
	for (iorder = 0; iorder < norder; iorder++) {
	  j = segments[iorder].order;
	  printf(" %d (%d) ", segments[j].nface, j);
	}
	printf("\n");
*/
        for (iend = 2; iend >= 0; iend--)
          for (iorder = norder-1; iorder >= 0; iorder--) {
            j = segments[iorder].order;

            if (iend == 0) {
              /* look on start */
              iterm1 = segments[j].start;
              iface1 = segments[j].faces[0];
              seg[0][0] = j;
              seg[0][1] = iend;
            } else if (iend == 1) {
              /* look at finish */
              iterm1 = segments[j].finish;
              iface1 = segments[j].faces[segments[j].nface-1];
              seg[0][0] = j;
              seg[0][1] = iend;
            } else {
              if ((segments[j].start  != NOTFILLED) &&
                  (segments[j].finish != NOTFILLED)) continue;
              if (segments[j].start == NOTFILLED) {
                iterm1 = segments[j].finish;
                iface1 = segments[j].faces[segments[j].nface-1];
                seg[0][0] = j;
                seg[0][1] = 1;
              } else {
                iterm1 = segments[j].start;
                iface1 = segments[j].faces[0];
                seg[0][0] = j;
                seg[0][1] = 0;
              }
            }
            if (iterm1 == NOTFILLED) continue;

            for (flag = 1, k = 0; k < nsegs; k++) {
              if (k == j) continue;
              if (segments[k].nface == 0) continue;
              if ((iterm1 == segments[k].start) &&
                  (iface1 == segments[k].faces[0])) {
                seg[flag][0] = k;
                seg[flag][1] = 0;
                flag++;
              }
              if ((iterm1 == segments[k].finish) &&
                  (iface1 == segments[k].faces[segments[k].nface-1])) {
                seg[flag][0] = k;
                seg[flag][1] = 1;
                flag++;
              }
            }

            if (flag > 2) {
              /* get potential lengths */
              for (j = 0; j < flag; j++) {
                k = seg[j][0];
                seg[j][2] = segments[k].nface;
                if (seg[j][1] == 0) {
                  iterm2 = segments[k].finish;
                  iface2 = segments[k].faces[segments[k].nface-1];
                } else {
                  iterm2 = segments[k].start;
                  iface2 = segments[k].faces[0];
                }
                seg[j][1] = 0;
                if (iterm2 == NOTFILLED) continue;

                nl = segments[k].nface-1;
                if (iterm2 > 0) nl--;
                for (m = 0; m < 3; m++) leaf[0][m][1] = nl;
                leaf[0][0][0] = iface2;
                leaf[0][1][0] = iterm2;
                segments[k].nface = -segments[k].nface;
                nleaf  = 1;
                nleafp = 0;
              
                while ((nleaf != nleafp) && (nleaf < VCFHUNK)) {
                  iface2 = leaf[nleafp][0][0];
                  iterm2 = leaf[nleafp][1][0];
                  for (m = 0; m < 3; m++) leaf[nleafp][m][0] = NOTFILLED;
                  for (m = k = 0; k < nsegs; k++) {
                    if (segments[k].nface <= 0) continue;
                    if ((iterm2 == segments[k].start) &&
                        (iface2 == segments[k].faces[0])) {
                      if (segments[k].finish == NOTFILLED) {
                        leaf[nleafp][m][1] += segments[k].nface;
                      } else {
                        leaf[nleafp][m][0] = nleaf;
                        leaf[nleaf][0][0]  = 
                          segments[k].faces[segments[k].nface-1];
                        leaf[nleaf][1][0]  = segments[k].finish;
                        ileaf = segments[k].nface-1;
                        if (segments[k].finish > 0) ileaf--;
                        for (nl = 0; nl < 3; nl++) leaf[nleaf][nl][1] = ileaf;
                        up[nleaf][0] = nleafp;
                        up[nleaf][1] = m;
                        nleaf++;
                        if (nleaf >= VCFHUNK) break;
                      }
                      m++;
                      segments[k].nface = -segments[k].nface;
                    }
                    if ((iterm2 == segments[k].finish) &&
                        (iface2 == segments[k].faces[segments[k].nface-1])) {
                      if (segments[k].start == NOTFILLED) {
                        leaf[nleafp][m][1] += segments[k].nface;
                      } else {
                        leaf[nleafp][m][0] = nleaf;
                        leaf[nleaf][0][0]  = segments[k].faces[0];
                        leaf[nleaf][1][0]  = segments[k].start;
                        ileaf = segments[k].nface-1;
                        if (segments[k].start > 0) ileaf--;
                        for (nl = 0; nl < 3; nl++) leaf[nleaf][nl][1] = ileaf;
                        up[nleaf][0] = nleafp;
                        up[nleaf][1] = m;
                        nleaf++;
                        if (nleaf >= VCFHUNK) break;
                      }
                      m++;
                      segments[k].nface = -segments[k].nface;
                    }
                  }
                  nleafp++;
                }

                if (nleaf >= VCFHUNK) printf(" Warning: Tree too large!\n");

                for (nl = nleaf-1; nl >= 1; nl--) {
                  m = leaf[nl][0][1];
                  if (m < leaf[nl][1][1]) m = leaf[nl][1][1];
                  if (m < leaf[nl][2][1]) m = leaf[nl][2][1];
                  leaf[up[nl][0]][up[nl][1]][1] += m;
                }
                m = leaf[0][0][1];
                if (m < leaf[0][1][1]) m = leaf[0][1][1];
                if (m < leaf[0][2][1]) m = leaf[0][2][1];
                seg[j][1] = 1;
                seg[j][2] = m;
/*
		printf(" Tree for %d (%d),  len = %d - nleaf = %d\n",
                       seg[j][0], seg[j][1], m, nleaf);
		for (m = 0; m < nleaf; m++)
		  printf("%d: %d %d    %d %d    %d %d\n", m, leaf[m][0][0],
			leaf[m][0][1], leaf[m][1][0], leaf[m][1][1],
			leaf[m][2][0], leaf[m][2][1]);
*/
                for (k = 0; k < nsegs; k++) 
                  if (segments[k].nface < 0) 
                    segments[k].nface = -segments[k].nface;
              }

              /* find the 2 longest segments */

              for (iface1 = -1, j = 0; j < flag; j++)
                if (seg[j][1] == 0)
                  if (iface1 == -1) {
                    iface1 = seg[j][0];
                    iterm1 = seg[j][2];
                  } else {
                    if (seg[j][2] > iterm1) {
                      iface1 = seg[j][0];
                      iterm1 = seg[j][2];
                    }
                  }

              /* start from the longest terminated segment */

              if (iface1 != -1) {
                for (iterm2 = j = 0; j < flag; j++) {
                  if (iface1 == seg[j][0]) continue;
                  if (iterm2 == 0) {
                    iface2 = seg[j][0];
                    iterm2 = seg[j][2];
                  } else if (seg[j][2] > iterm2) {
                    segments[iface2].nface = 0;
                    iface2 = seg[j][0];
                    iterm2 = seg[j][2];
                  } else {
                    segments[seg[j][0]].nface = 0;
                  }
                }
                FX_CatSegVC(&segments[iface1], &segments[iface2]);
                goto loop;
              }

              /* all unterminated */

              iface1 = seg[0][0];
              iterm1 = seg[0][2];
              iface2 = seg[1][0];
              iterm2 = seg[1][2];
              if (iterm1 < iterm2) {
                m = iterm2;
                iface2 = seg[0][0];
                iterm2 = iterm1;
                iface1 = seg[1][0];
                iterm1 = m;
              }
              for (j = 2; j < flag; j++) {
                m = seg[j][2];
                if (m > iterm1) {
                  segments[iface2].nface = 0;
                  iface2 = iface1;
                  iterm2 = iterm1;
                  iface1 = seg[j][0];
                  iterm1 = m;
                } else if (m > iterm2) {
                  segments[iface2].nface = 0;
                  iface2 = seg[j][0];
                  iterm2 = m;
                } else {
                  segments[seg[j][0]].nface = 0;
                }
              }
              FX_CatSegVC(&segments[iface1], &segments[iface2]);
              goto loop;
            }
          }

      }

      /* store away the segments we have so far */

      for (m = n = j = 0; j < nsegs; j++) 
        if (segments[j].nface != 0) {
          segments[n] = segments[j];
          n++;
        } else {
          free(segments[j].faces);
        }

      if (n != 0) {
        if (pack->nsegs == 0) {
          pack->msegs = msegs;
          pack->nsegs = n;
          pack->segs  = segments;
          m = 1;
        } else {
          if (n+pack->nsegs > pack->msegs) {
            k = pack->msegs + VCFSEG;
            if (k < n+pack->nsegs) k = n+pack->nsegs;
            tmp = (segs *) realloc(pack->segs, k*sizeof(segs));
            if (tmp == NULL) {
              printf(" FX_SegmentVC: Error Reallocating Segs!\n");
              return;
            }
            pack->segs  = tmp;
            pack->msegs = k;
          }
          for (j = 0; j < n; j++, pack->nsegs++)
            pack->segs[pack->nsegs] = segments[j];
        }
      }
      if (m == 0) free(segments);
      goto start;

    }

}
