
  /* structures */

  typedef struct {
    int   n1;                   /* first 3D node index */
    int   n2;                   /* second 3D node index */
    float interp;               /* normalized distance from n1 */
    int   thread;               /* thread for next point with n1 */
  } FXinter;


  typedef struct
  {
    double re;
    double im;
  } complex;


  typedef struct
  {
    double xyz[3][3];
    double xyzRot[3][3];
    double uvw[3][3];
    double uvwRot[3][3];
    double q[3][3];
    double qinv[3][3];
    double tnsr[3][3][3];
    double tnsrRot[3][3][3];
  } facet;


  typedef struct {
    int nodes[4];
  } FXtet;


  typedef struct {
    int nodes[5];
  } FXpyra;


  typedef struct {
    int nodes[6];
  } FXprism;


  typedef struct {
    int nodes[8];
  } FXhexa;


  typedef struct {
    int ni;			/* Imax */
    int nj;			/* Jmax */
    int nk;			/* kmax */
    int last_cell;		/* last cell number of the block */
    int last_node;		/* last node number in the block */
  } FXblock;


  typedef struct {
    int nodes[8];
    int type;			/* 1 - tet, 2 - pyra, 3 - prism, 4 - hexa */
  } FXhalo;


  typedef struct {
    int last_facet;		/* last facet for the grouping */
    int bc_type;		/* boundary type */
  } FXbc;


  typedef struct {
    int nodes[4];
  } FXfacet;


  typedef struct {
    float x;
    float y;
    float z;
  } FXcoor;


  typedef struct {
    float   gamma;		/* ratio of specific heats */
    int     iopt;		/* unsteady control (0 - 3) */
    int     cmask;		/* calling mask */
    float   time;		/* simulation time */
    int     status;		/* status for structure unsteady */
    int     rt;			/* residence time flag */

    int     total_cells;	/* total number of cells (non-halo) */
    int     total_dis;		/* total number of disjoint cells */
    int     ntets;		/* number of tets */
    FXtet   *tets;		/* tetrahedral definitions */
    int     npyra;		/* number of pyramids */
    FXpyra  *pyra;		/* pyramid definitions */
    int     nprism;		/* number of prisms */
    FXprism *prism;		/* prism definitions */
    int     nhexa;		/* number of disjoint hexahedra */
    FXhexa  *hexa;		/* hexahedral definitions */
    int     mblock;		/* allocated size -- struc unsteady */
    int     nblock;		/* number of blocks */
    FXblock *block;		/* structured block definitions */

    int     nbc;		/* number of domain surface (BC) groups */
    FXbc    *bc;		/* BC definitions */
    int     nfacet;		/* number of domain surface facets */
    FXfacet *facet;		/* facet definitions */
    int     *scel;		/* cell indices for facets */

    int     nnode;		/* non block nodes */
    int     total_nodes;	/* total number of nodes (non-halo) */
    FXcoor  *coor;		/* coordinates -- or -- */
    float   *xcoor;		/* x coordinates */
    float   *ycoor;		/* y coordinates */
    float   *zcoor;		/* z coordinates */
    float   *scalar;		/* scalar field */
    FXcoor  *vel;		/* velocity coordinates -- or -- */
    float   *velx;		/* velocity in x direction */
    float   *vely;		/* velocity in y direction */
    float   *velz;		/* velocity in z direction */
    int     *iblank;		/* Iblanking */

    int     nhalo;		/* halo nodes */
    FXcoor  *hcoor;		/* halo coordinates */
    float   *hscalar;		/* halo scalar field */
    FXcoor  *hvel;		/* halo velocity coordinates */
    int     nhcell;		/* number of halo cells */
    FXhalo  *hcell;		/* halo cell definitions */

    int     vload;		/* are the velocity fields loaded? */
    int     sload;		/* the last scalar field loaded */
  } FXglobal;


