#include <stdio.h>
#include <stdlib.h>
#include "pvm3.h"

/*
  task that mimics the PVM group function for Concentrators
    on systems that do not support PVM groups
    for pV3 Revision 2.0

  Copyright 1996 - 1999, Massachusetts Institute of Technology.

*/


main()
{
  int  tid, pid, i, num, sid;
  FILE *fp;
  char sname[80], cname[80];
#ifndef PVM33
  int  new, info, buf;
#endif

  fp = fopen("/tmp/PseudoGroup.out", "w");
  fprintf(fp, "PseudoGroup log:\n");

  tid = pvm_mytid();
  if (tid < 0) {
    fprintf(fp, " pvmd Not running!\n");
    fclose(fp);
    exit(1);
  }

#ifndef PVM33
  new = pvm_newcontext();
  pvm_initsend(PvmDataDefault);
  pvm_pkint(&new, 1, 1);
  info = pvm_putinfo("pV3context", pvm_getsbuf(), PvmMboxPersistent);
  if (info < 0) {
    pvm_freecontext(new);
    new = -1;
  }
  if ((info == PvmExists) || (info == PvmDenied)) {
    /* must be registered already -- get context */
    buf = pvm_recvinfo("pV3context", 0, PvmMboxFirstAvail);
    if (buf >= 0) {
      pvm_setrbuf(buf);
      pvm_upkint(&new, 1, 1);
    }
  }
  info = -1000;
  if (new >= 0) info = pvm_setcontext(new);
  if (info < 0) new  = -1;
  if (new == -1) {
    fprintf(fp, " Error: Can not set Context %d!\n", info);
    fclose(fp);
    pvm_exit();
    exit(1);
  }
  fprintf(fp, " pV3 ctx = %x\n", new);
#endif

  pid = pvm_parent();
  if (pid <= 0) {
    fprintf(fp, " Error: Parent TID = %d!\n", pid);
    fclose(fp);
    pvm_exit();
    exit(1);
  } else {
    fprintf(fp, " Parent tid = %x\n", pid);
  }

  pvm_recv(pid, 1);
  pvm_upkstr(cname);
  pvm_upkstr(sname);
  fprintf(fp, " Client Group Name = %s\n", cname);
  fprintf(fp, " Server Group Name = %s\n", sname);

  if (pvm_joingroup(cname) < 0) {
    fprintf(fp, " Error joining group!\n");
    fclose(fp);
    pvm_exit();
    exit(1);
  }

  while (pvm_pstat(pid) >= 0) {
/*  fprintf(fp, " Waiting...\n");  */
    pvm_initsend(PvmDataDefault);
    num = pvm_gsize(sname);
    pvm_pkint(&num, 1, 1);
    for (i = 0; i < num; i++) {
      sid = pvm_gettid(sname, i);
      pvm_pkint(&sid, 1, 1);
    }
    pvm_send(pid, 1);
    sleep(10);
  }
  fprintf(fp, " pV3 Concentrator has exited!\n");

  pvm_lvgroup(cname);
  fclose(fp);

  pvm_exit();

}
