#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#if defined (WIN32) || defined (ABSOFT)
#include <WINSOCK.H>
#else
#include <unistd.h>
#endif
#include <time.h>
#include <sys/types.h>
#include "mpi.h"
#include "pvm3.h"

#ifdef ABSOFT
#undef WIN32
#endif


/*
  PVM/MPI interface for pV3 clients and concentrator
          pV3 Revision 2.10 beta

  Copyright 1993 - 2005, Massachusetts Institute of Technology.

*/


#ifdef CRAY
#define pV_MPISTART    PV_MPISTART
#define pV_MPISTOP     PV_MPISTOP
#define pV_SERVER      PV_SERVER
#define pV_NETCONTROL  PV_NETCONTROL
#define pV_SETCID      PV_SETCID
#define pV_SERVERSTAT  PV_SERVERSTAT
#define pV_SETCSERVER  PV_SETCSERVER
#define pV_SETMATCH    PV_SETMATCH
#define pV_SENDINIT    PV_SENDINIT
#define pV_SENDMES     PV_SENDMES
#define pV_SENDXI      PV_SENDXI
#define pV_SENDXR      PV_SENDXR
#define pV_SENDXV      PV_SENDXV
#define pV_SENDSLXI    PV_SENDSLXI
#define pV_SENDSLXR    PV_SENDSLXR
#define pV_SENDSV      PV_SENDSV
#define pV_SENDTIME    PV_SENDTIME
#define pV_SENDSTRING  PV_SENDSTRING
#define pV_SENDXFER    PV_SENDXFER
#define pV_SENDXFERACK PV_SENDXFERACK
#define pV_GETSERDATA  PV_GETSERVERDATA
#define pV_GETREQUEST  PV_GETREQUEST
#define pV_GETTREQUEST PV_GETTREQUEST

#define pVSTATE        PVSTATE
#define pVPSTATE       PVPSTATE
#define pV_CELL2NODES  PV_CELL2NODES
#define pV_FILLVCLOUD  PV_FILLVCLOUD
#endif

#ifdef WIN32
#define pV_MPISTART    __stdcall PV_MPISTART
#define pV_MPISTOP     __stdcall PV_MPISTOP
#define pV_SERVER      __stdcall PV_SERVER
#define pV_NETCONTROL  __stdcall PV_NETCONTROL
#define pV_SETCID      __stdcall PV_SETCID
#define pV_SERVERSTAT  __stdcall PV_SERVERSTAT
#define pV_SETCSERVER  __stdcall PV_SETCSERVER
#define pV_SETMATCH    __stdcall PV_SETMATCH
#define pV_SENDINIT    __stdcall PV_SENDINIT
#define pV_SENDMES     __stdcall PV_SENDMES
#define pV_SENDXI      __stdcall PV_SENDXI
#define pV_SENDXR      __stdcall PV_SENDXR
#define pV_SENDXV      __stdcall PV_SENDXV
#define pV_SENDSLXI    __stdcall PV_SENDSLXI
#define pV_SENDSLXR    __stdcall PV_SENDSLXR
#define pV_SENDSV      __stdcall PV_SENDSV
#define pV_SENDTIME    __stdcall PV_SENDTIME
#define pV_SENDSTRING  __stdcall PV_SENDSTRING
#define pV_SENDXFER    __stdcall PV_SENDXFER
#define pV_SENDXFERACK __stdcall PV_SENDXFERACK
#define pV_GETSERDATA  __stdcall PV_GETSERVERDATA
#define pV_GETREQUEST  __stdcall PV_GETREQUEST
#define pV_GETTREQUEST __stdcall PV_GETTREQUEST

#define pVSTATE        PVSTATE
#define pVPSTATE       PVPSTATE
#define pV_CELL2NODES  PV_CELL2NODES
#define pV_FILLVCLOUD  PV_FILLVCLOUD
#endif

#ifdef ABSOFT
#define pV_MPISTART    __cdecl pv_mpistart
#define pV_MPISTOP     __cdecl pv_mpistop
#define pV_SERVER      __cdecl pv_server
#define pV_NETCONTROL  __cdecl pv_netcontrol
#define pV_SETCID      __cdecl pv_setcid
#define pV_SERVERSTAT  __cdecl pv_serverstat
#define pV_SETCSERVER  __cdecl pv_setcserver
#define pV_SETMATCH    __cdecl pv_setmatch
#define pV_SENDINIT    __cdecl pv_sendinit
#define pV_SENDMES     __cdecl pv_sendmes
#define pV_SENDXI      __cdecl pv_sendxi
#define pV_SENDXR      __cdecl pv_sendxr
#define pV_SENDXV      __cdecl pv_sendxv
#define pV_SENDSLXI    __cdecl pv_sendslxi
#define pV_SENDSLXR    __cdecl pv_sendslxr
#define pV_SENDSV      __cdecl pv_sendsv
#define pV_SENDTIME    __cdecl pv_sendtime
#define pV_SENDSTRING  __cdecl pv_sendstring
#define pV_SENDXFER    __cdecl pv_sendxfer
#define pV_SENDXFERACK __cdecl pv_sendxferack
#define pV_GETSERDATA  __cdecl pv_getserverdata
#define pV_GETREQUEST  __cdecl pv_getrequest
#define pV_GETTREQUEST __cdecl pv_gettrequest

#define pVSTATE        pvstate
#define pVPSTATE       pvpstate
#define pV_CELL2NODES  pv_cell2nodes
#define pV_FILLVCLOUD  pv_fillvcloud
#endif


#ifdef UNDERSCORE
#define pV_MPISTART    pv_mpistart_
#define pV_MPISTOP     pv_mpistop_
#define pV_SERVER      pv_server_
#define pV_NETCONTROL  pv_netcontrol_
#define pV_SETCID      pv_setcid_
#define pV_SERVERSTAT  pv_serverstat_
#define pV_SETCSERVER  pv_setcserver_
#define pV_SETMATCH    pv_setmatch_
#define pV_SENDINIT    pv_sendinit_
#define pV_SENDMES     pv_sendmes_
#define pV_SENDXI      pv_sendxi_
#define pV_SENDXR      pv_sendxr_
#define pV_SENDXV      pv_sendxv_
#define pV_SENDSLXI    pv_sendslxi_
#define pV_SENDSLXR    pv_sendslxr_
#define pV_SENDSV      pv_sendsv_
#define pV_SENDTIME    pv_sendtime_
#define pV_SENDSTRING  pv_sendstring_
#define pV_SENDXFER    pv_sendxfer_
#define pV_SENDXFERACK pv_sendxferack_
#define pV_GETSERDATA  pv_getserverdata_
#define pV_GETREQUEST  pv_getrequest_
#define pV_GETTREQUEST pv_gettrequest_

#define pVSTATE        pvstate_
#define pVPSTATE       pvpstate_
#define pV_CELL2NODES  pv_cell2nodes_
#define pV_FILLVCLOUD  pv_fillvcloud_
#endif

#ifdef NOUNDERSCORE
#define pV_MPISTART    pv_mpistart
#define pV_MPISTOP     pv_mpistop
#define pV_SERVER      pv_server
#define pV_NETCONTROL  pv_netcontrol
#define pV_SETCID      pv_setcid
#define pV_SERVERSTAT  pv_serverstat
#define pV_SETCSERVER  pv_setcserver
#define pV_SETMATCH    pv_setmatch
#define pV_SENDINIT    pv_sendinit
#define pV_SENDMES     pv_sendmes
#define pV_SENDXI      pv_sendxi
#define pV_SENDXR      pv_sendxr
#define pV_SENDXV      pv_sendxv
#define pV_SENDSLXI    pv_sendslxi
#define pV_SENDSLXR    pv_sendslxr
#define pV_SENDSV      pv_sendsv
#define pV_SENDTIME    pv_sendtime
#define pV_SENDSTRING  pv_sendstring
#define pV_SENDXFER    pv_sendxfer
#define pV_SENDXFERACK pv_sendxferack
#define pV_GETSERDATA  pv_getserverdata
#define pV_GETREQUEST  pv_getrequest
#define pV_GETTREQUEST pv_gettrequest

#define pVSTATE        pvstate
#define pVPSTATE       pvpstate
#define pV_CELL2NODES  pv_cell2nodes
#define pV_FILLVCLOUD  pv_fillvcloud
#endif


#ifdef FLOATDOUBLE
#define FLOAT double
#else
#define FLOAT float
#define pv_pkFLOAT pv_pkfloat
#define pv_upkFLOAT pv_upkfloat
#endif


#define MAXLEN 262144		/* Maximum buffer size in words */
#define MAXREQ   4096		/* Maximum request buffer size */
#define MAXSER      8		/* Maximum number of servers */
#define MAXSERINS 100		/* number of server instances to check */

#define HIPRIO  26729		/* High priority message tag */
#define LOPRIO  27759		/* Low priority message tag */
#define INPRIO  26990		/* Init message tag */
#define PVUP        1		/* pV3 service up message */
#define PVDOWN      2		/* pV3 service down message */


/* structures */

typedef struct {
  int  *id2r;			/* client id translation to MPI rank */
  int  mrank;			/* size of id2r */
  int  did;			/* discipline id -- assigned via server */
  char title[20];		/* discipline title */
} Discstr;

typedef struct {
  int tid;                      /* server task ID */
  int kvpa;                     /* index for vector fields (particles) */
  int status;                   /* 0-exist, 1-startup, 2-running, 3-eof,
                                   4-close-up, 5-dead (but not gone) */
  int hwbuf;			/* concentrator hi prio save buffer */
  int lwbuf;			/* concentrator lo prio save buffer */
  int ltag;			/* concentrator lo prio mesage tag */
  int wrdcnt;			/* concentrator lo prio mesage word count */
} Server;

typedef struct ServerData {
  int   handle;			/* handle index */
  int   mid;			/* message id */
  int   nint;			/* number of ints */
  int   *ints;			/* pointer to ints */
  int   nfloat;			/* number of floats */
  FLOAT *floats;		/* pointer to floats */
  int   nchar;			/* number of chars */
  char  *chars;			/* pointer to chars */
  struct ServerData *next;	/* pointer to next set of server data */
} ServerData;


/* used by these routines */

       MPI_Comm    pv3_comm;		/* application's communicator */

static int         pv3_ctid, pv3_stid, pv3_did, pv3_id, pv3_nser, pv3_sindx;
static char       *pv3_SGroup, *pv3_CGroup;
static MPI_Comm    pv3_hicomm, pv3_locomm;
static MPI_Group   pv3_group,  pv3_groupfull;
static char       *pv3_hibuf, *pv3_lobuf, *pv3_buffer;
static char       *pv3_hrbuf, *pv3_lrbuf;
static int         pv3_hipos,  pv3_lopos, pv3_total, pv3_prio,  pv3_bsize;
static int         pv3_hrpos,  pv3_lrpos, pv3_psize, pv3_rsize, pv3_srbuf;
static int         pv3_myrank, pv3_srank, pv3_ncl,   pv3_conc,  pv3_sdh = 0;
static Server      pv3_ser[MAXSER];
static ServerData *pv3_sd = NULL;


#ifndef NO_PROTO
#ifdef WIN32
extern void __stdcall pVSTATE ( int *rank, int *kn, FLOAT *sv );
extern void __stdcall pVPSTATE ( int *rank, int *index, int *kn, FLOAT *sv );
extern void __stdcall pV_CELL2NODES ( int *kcr, int *ktet, int *n, int *kn );
extern void __stdcall pV_FILLVCLOUD ( int *isub, int *rank, FLOAT *cut,
                                      FLOAT *work, int *kn, FLOAT *out );
#else
#ifdef ABSOFT
extern void __cdecl pVSTATE ( int *rank, int *kn, FLOAT *sv );
extern void __cdecl pVPSTATE ( int *rank, int *index, int *kn, FLOAT *sv );
extern void __cdecl pV_CELL2NODES ( int *kcr, int *ktet, int *n, int *kn );
extern void __cdecl pV_FILLVCLOUD ( int *isub, int *rank, FLOAT *cut,
                                    FLOAT *work, int *kn, FLOAT *out );

#else
extern void pVSTATE ( int *rank, int *kn, FLOAT *sv );
extern void pVPSTATE ( int *rank, int *index, int *kn, FLOAT *sv );
extern void pV_CELL2NODES ( int *kcr, int *ktet, int *n, int *kn );
extern void pV_FILLVCLOUD ( int *isub, int *rank, FLOAT *cut,
                            FLOAT *work, int *kn, FLOAT *out );
#endif
#endif
#endif


/* ********* Utility Routines Used by Clients & Concentrator ********* */

/* pack bytes into buffer */

static void
pv_pkbyte(char *bytes, int len)
{

  if (pv3_prio == 1) {
    MPI_Pack(bytes, len, MPI_BYTE, pv3_lobuf, pv3_psize, &pv3_lopos,
             pv3_locomm);
  } else {
    MPI_Pack(bytes, len, MPI_BYTE, pv3_hibuf, pv3_psize, &pv3_hipos, 
             pv3_hicomm);
  }
}


/* pack ints into buffer */

static void
pv_pkint(int *ints, int len)
{

  if (pv3_prio == 1) {
    MPI_Pack(ints, len, MPI_INT, pv3_lobuf, pv3_psize, &pv3_lopos, pv3_locomm);
  } else {
    MPI_Pack(ints, len, MPI_INT, pv3_hibuf, pv3_psize, &pv3_hipos, pv3_hicomm);
  }
}


/* pack floats into buffer */

static void
pv_pkfloat(float *floats, int len)
{

  if (pv3_prio == 1) {
    MPI_Pack(floats, len, MPI_FLOAT, pv3_lobuf, pv3_psize, &pv3_lopos,
             pv3_locomm);
  } else {
    MPI_Pack(floats, len, MPI_FLOAT, pv3_hibuf, pv3_psize, &pv3_hipos, 
             pv3_hicomm);
  }
}

#ifdef FLOATDOUBLE
static void
pv_pkFLOAT(FLOAT *floats, int len)
{
  int   i;
  float j;

  if (pv3_prio == 1) {
    for (i = 0; i < len; i++) {
      j = floats[i];
      MPI_Pack(&j, 1, MPI_FLOAT, pv3_lobuf, pv3_psize, &pv3_lopos, pv3_locomm);
    }
  } else {
    for (i = 0; i < len; i++) {
      j = floats[i];
      MPI_Pack(&j, 1, MPI_FLOAT, pv3_hibuf, pv3_psize, &pv3_hipos, pv3_hicomm);
    }
  }
}
#endif


/* unpack bytes from buffer */

static void
pv_upkbyte(char *bytes, int len)
{

  if (pv3_srbuf == 1) {
    MPI_Unpack(pv3_lrbuf, pv3_rsize, &pv3_lrpos, bytes, len, MPI_BYTE,
               pv3_locomm);
  } else {
    MPI_Unpack(pv3_hrbuf, pv3_rsize, &pv3_hrpos, bytes, len, MPI_BYTE,
               pv3_hicomm);
  }
}


/* unpack ints from buffer */

static void
pv_upkint(int *ints, int len)
{

  if (pv3_srbuf == 1) {
    MPI_Unpack(pv3_lrbuf, pv3_rsize, &pv3_lrpos, ints, len, MPI_INT,
               pv3_locomm);
  } else {
    MPI_Unpack(pv3_hrbuf, pv3_rsize, &pv3_hrpos, ints, len, MPI_INT,
               pv3_hicomm);
  }
}


/* unpack floats from buffer */

static void
pv_upkfloat(float *floats, int len)
{

  if (pv3_srbuf == 1) {
    MPI_Unpack(pv3_lrbuf, pv3_rsize, &pv3_lrpos, floats, len, MPI_FLOAT,
               pv3_locomm);
  } else {
    MPI_Unpack(pv3_hrbuf, pv3_rsize, &pv3_hrpos, floats, len, MPI_FLOAT,
               pv3_hicomm);
  }
}

#ifdef FLOATDOUBLE
static void
pv_upkFLOAT(FLOAT *floats, int len)
{
  int   i;
  float j;

  if (pv3_srbuf == 1) {
    for (i = 0; i < len; i++) {
      MPI_Unpack(pv3_lrbuf, pv3_rsize, &pv3_lrpos, &j, 1, MPI_FLOAT,
                 pv3_locomm);
      floats[i] = j;
    }
  } else {
    for (i = 0; i < len; i++) {
      MPI_Unpack(pv3_hrbuf, pv3_rsize, &pv3_hrpos, &j, 1, MPI_FLOAT,
                 pv3_hicomm);
      floats[i] = j;
    }
  }
}
#endif


/* sets the appropriate send buffer or flushes both */

static void
pv_MesBuffer(int prio, int bytes)
{
  int mid;

  if (prio == 2) {
    if (pv3_prio == 2) {
      if (pv3_hipos+bytes < pv3_psize-sizeof(int)) return;
      mid = -1;
      pv_pkint(&mid, 1);
      MPI_Bsend(pv3_hibuf, pv3_hipos, MPI_PACKED, pv3_srank, HIPRIO,
                pv3_hicomm);
      pv3_hipos = 0;
    }
    pv3_prio = 2;
  }

  if (prio == 1) {
    if (pv3_prio == 2) {
      mid = -1;
      pv_pkint(&mid, 1);
      MPI_Bsend(pv3_hibuf, pv3_hipos, MPI_PACKED, pv3_srank, HIPRIO,
                pv3_hicomm);
      pv3_hipos = 0;
      if (pv3_lopos != 0) pv3_prio = 1;
    }
    if (pv3_prio == 1) {
      if (pv3_lopos+bytes < pv3_psize-sizeof(int)) return;
      mid = -1;
      pv_pkint(&mid, 1);
      MPI_Bsend(pv3_lobuf, pv3_lopos, MPI_PACKED, pv3_srank, LOPRIO, 
                pv3_locomm);
      pv3_lopos = 0;
    }
    pv3_prio = 1;
  }

  if (prio == 0) {
    if (pv3_prio == 2) {
      mid = -1;
      pv_pkint(&mid, 1);
      MPI_Bsend(pv3_hibuf, pv3_hipos, MPI_PACKED, pv3_srank, HIPRIO, 
                pv3_hicomm);
      if (pv3_lopos != 0) pv3_prio = 1;
    }
    if (pv3_prio == 1) {
      mid = -1;
      pv_pkint(&mid, 1);
      mid = LOPRIO;
      if (bytes != 0) mid = bytes;
      MPI_Bsend(pv3_lobuf, pv3_lopos, MPI_PACKED, pv3_srank, mid, pv3_locomm);
    }
    pv3_lopos = 0;
    pv3_hipos = 0;
    pv3_prio  = 0;
  }

}

#if defined (WIN32) || defined (ABSOFT)

static void
pv_Sleep(int secs)
{
  Sleep(1000*secs);
}

static void
pv_mSleep(int msecs)
{
  Sleep(msecs);
}

#else

/* sleep -- unix sleep kick's out at any signal! */

static void
pv_Sleep(int nsec)
{
  time_t start;
  int    sec;

  start = time(NULL);
  sec   = nsec;

  while (sec > 0) {
    sleep(sec);
    sec = nsec - (time(NULL) - start);
  }
}

static void
pv_mSleep(int msec)
{
  usleep(msec*1000);
}
#endif


/* ************** Routines Used by the Concentrator ************** */


/* clear out any messages comming to the concentrator */

static void
pv_mesClear(int tid)
{

  while (pvm_probe(tid, -1) > 0) pvm_recv(tid, -1);
  if (tid == -1) pvm_exit();
}


/* check for client termination messages */

static void
pv_clscan()
{
  int        i;
  MPI_Status status;

  MPI_Iprobe(MPI_ANY_SOURCE, PVDOWN, pv3_locomm, &i, &status);
  if (i == 0) return;

  printf("\n pV3 Concentrator: Shutting Down...\n");
#ifndef NOGROUPS
  pvm_lvgroup(pv3_CGroup);
#endif
  pv_Sleep(1);
  pv_mesClear(-1);
  MPI_Barrier(pv3_locomm);

  MPI_Comm_free(&pv3_locomm);
  MPI_Comm_free(&pv3_hicomm);
  MPI_Finalize();
  exit(0);
}


/* do the concentrator thing */

static void
pv_concentrator()
{
  int        i, j, k, len, did, gsize, tid, finit, hit, TimeOut, stat, mlookup;
  int        wrds, mid, lrank, mdisc, ndisc, ins;
  int        *buf, *tmp, *lopos, *dlookup, ints[10], mstid, bufid;
  char       **lobuf, *cbuf, dtit[20], *env;
  Discstr    *discipl;
  MPI_Status status;
  time_t     last;
#ifdef NOGROUPS
  int        pgid, pgsids[2*MAXSER];
#endif
#ifndef PVM33
  int        new, info, buff;
#endif
  static char Server[7] = "Server";
  static char Client[7] = "Client";

  /* set group names */

  pv3_nser = 0;
  env = getenv("pV3_Group");
  if (env == NULL) {
    env = "pV3";
  } else {
    printf("\n pV3 Concentrator: using group prefix = %s\n", env);
  }
  len = strlen(env);
  pv3_CGroup = malloc(len + 7);
  pv3_SGroup = malloc(len + 7);
  for (i = 0; i < len; i++) {
    pv3_CGroup[i] = env[i];
    pv3_SGroup[i] = env[i];
  }
  for (i = 0; i < 7; i++) pv3_CGroup[i+len] = Client[i];
  for (i = 0; i < 7; i++) pv3_SGroup[i+len] = Server[i];

  /* allocate the required structures */

  mlookup = 8;
  dlookup = (int *) malloc(mlookup*sizeof(int));
  mdisc   = 8;
  discipl = (Discstr *) malloc(mdisc*sizeof(Discstr));
  if ((dlookup == NULL) || (discipl == NULL)) {
    printf(" pV3 Concentrator Error in memory allocation!\n");
    exit(1);
  }
  for (i = 0; i < mdisc; i++) discipl[i].id2r = NULL;

  buf = (int *) malloc(MAXLEN*sizeof(int));
  if (buf == NULL) {
    printf(" pV3 Concentrator Error in memory allocation (buf)!\n");
    exit(1);
  }

  pv3_psize = MAXREQ*sizeof(int);
  pv3_hibuf = (char *) malloc(pv3_psize);
  if (pv3_hibuf == NULL) {
    printf(" pV3/MPI Failure in allocating buffer!\n");
    exit(1);
  }

  lobuf = (char **) malloc(pv3_total*sizeof(char *));
  lopos = (int *)   malloc(pv3_total*sizeof(int));
  if ((lobuf == NULL) || (lopos == NULL)) {
    printf(" pV3 Concentrator Error in memory allocation (bufs)!\n");
    exit(1);
  }
  for (i = 0; i < pv3_total; i++)
    if (i != pv3_conc) {
      lobuf[i] = (char *) malloc(pv3_psize);
      if (lobuf[i] == NULL) {
        printf(" pV3/MPI Failure in allocating buffer!\n");
        exit(1);
      }
    }

  pv3_rsize = pv3_bsize;
  pv3_hrbuf = (char *) malloc(pv3_rsize);
  if (pv3_hrbuf == NULL) {
    printf(" pV3/MPI Failure in allocating buffer (hi)!\n");
    exit(1);
  }
  pv3_lrbuf = (char *) malloc(pv3_rsize);
  if (pv3_lrbuf == NULL) {
    printf(" pV3/MPI Failure in allocating buffer (lo)!\n");
    exit(1);
  }

  printf(" pV3 Concentrator: Initialized!\n");

  pv3_ctid = pvm_mytid();

  /* wait for pvm to start up */
  while (pv3_ctid < 0) {
    pv_clscan();
    pv_Sleep(5);
    pv3_ctid = pvm_mytid();
  }

#ifndef PVM33
  /* set/get context */
  new = pvm_newcontext();
  pvm_initsend(PvmDataDefault);
  pvm_pkint(&new, 1, 1);
  info = pvm_putinfo("pV3context", pvm_getsbuf(), PvmMboxPersistent);
  if (info < 0) {
    pvm_freecontext(new);
    new = -1;
  }
  if ((info == PvmExists) || (info == PvmDenied)) {
    /* must be registered already  -- get context */
    buff = pvm_recvinfo("pV3context", 0, PvmMboxFirstAvail);
    if (buff >= 0) {
      pvm_setrbuf(buff);
      pvm_upkint(&new, 1, 1);
    }
  }
  info = -1000;
  if (new >= 0) info = pvm_setcontext(new);
  if (info < 0) new  = -1;
  if (new == -1) {
    printf("pV3 Concentrator Error: Can not set PVM Context %d!\n", info);
    exit(1);
  }
#endif

#ifdef NOGROUPS
pgstart:
  i = pvm_spawn("pseudogroup", 0, PvmTaskArch|PvmHostCompl, "CRAY", 1, &pgid);
  printf(" pV3 Concentrator: pvm_spawn = %d, tid = %x\n", i, pgid);
  if (i == 0) {
    pv_clscan();
    pv_Sleep(10);
    goto pgstart;
  }
  pvm_initsend(PvmDataDefault);
  pvm_pkstr(pv3_CGroup);
  pvm_pkstr(pv3_SGroup);
  pvm_send(pgid, 1);
#else
  pvm_joingroup(pv3_CGroup);
#endif

  /* wait for server(s) to start */

  printf(" pV3 Concentrator: Wait for Server/Batch to startup 0...\n");

sstartup:
  pv3_stid = 0;
#ifdef NOGROUPS
  pvm_recv(pgid, 1);
  pvm_upkint(&gsize, 1, 1);
#else
  gsize = pvm_gsize(pv3_SGroup);
#endif
  while(gsize <= 0) {
    pv_clscan();
    pv_Sleep(5);
#ifdef NOGROUPS
    if (pvm_pstat(pgid) < 0) goto pgstart;
    pvm_recv(pgid, 1);
    pvm_upkint(&gsize, 1, 1);
#else
    gsize = pvm_gsize(pv3_SGroup);
#endif
  }
#ifdef NOGROUPS
  if (gsize > 2*MAXSER) gsize = 2*MAXSER;
  pvm_upkint(pgsids, gsize, 1);
#endif

  ins = -1;
  for (i = 0; i < gsize; i++) {
#ifdef NOGROUPS
    tid = pgsids[i];
#else
    tid = PvmNoInst;
    while ((tid == PvmNoInst) && (ins < MAXSERINS)) {
      ins++;
      tid = pvm_gettid(pv3_SGroup, ins);
    }
#endif
    if (tid > 0) {
      if ((pvm_pstat(tid) >= 0) && (pv3_nser < MAXSER)) {
        pv3_stid  = tid;
        pv3_sindx = pv3_nser;
        pv3_ser[pv3_nser].tid   = tid;
        pv3_ser[pv3_nser].hwbuf = 0;
        pv3_ser[pv3_nser].lwbuf = 0;
        pv3_ser[pv3_nser].ltag  = -1;
        pv3_nser++;
      }
    }
  }
  if (pv3_stid == 0) goto sstartup;

  /* OK we have at least one server */

  for (j = 0; j < pv3_nser; j++) {
    printf(" pV3 Concentrator: Server %X has started!\n", pv3_ser[j].tid);
    for (i = 0; i < pv3_total; i++)
      if (i != pv3_conc) MPI_Bsend(&pv3_ser[j].tid, 1, MPI_INT, i,
                                   PVUP, pv3_hicomm);
    pv3_ser[j].status = 1;
  }

  for (i = 0; i < mlookup; i++) dlookup[i] = -1;
  for (i = 0; i < mdisc; i++) {
    if (discipl[i].id2r != NULL) free(discipl[i].id2r);
    discipl[i].id2r = NULL;
  }
  ndisc =  0;
  finit =  0;
  pv3_srbuf = 0;
  pv3_hrpos = 0;
  pv3_lrpos = 0;
  pv3_prio  = 0;
  pv3_hipos = 0;
  TimeOut   = 0;
  for (i = 0; i < pv3_total; i++) lopos[i] = 0;
  last = time(NULL);

loop:

  /* hi prio messages from server */

  if (pvm_probe(-1, HIPRIO) > 0) {
    bufid = pvm_recv(-1, HIPRIO);
    pvm_bufinfo(bufid, &len, &i, &mstid);
    for (k = j = 0; j < pv3_nser; j++) if (mstid == pv3_ser[j].tid) k++;
    if (k == 0) {
      printf(" pV3 Concentrator: 1. Server not found: %X!\n", mstid);
      goto loop;
    }
#ifdef PVDEBUG
    printf("Get hi prio message from server: %X\n", mstid);
#endif
    pv3_lopos = 0;
    lrank     = -1;
    ints[0]   = -1;
    pvm_upkint(ints, 1, 1);
    while (ints[0] != -1) {
      pvm_upkint(&ints[1], 9, 1);
      j = ints[1];
      i = ints[2];
      if (ints[0] == 35) {
        j = ints[4];
        if (j >= mlookup) {
          tmp = (int *) realloc(dlookup, (j+1)*sizeof(int));
          if (tmp != NULL) {
            dlookup = tmp;
            for (k = mlookup; k < j+1; k++) dlookup[k] = -1;
            mlookup = j+1;
          }
        }
        if (j < mlookup) {
          if (dlookup[j] != -1) goto normal;
          if ((ints[7] != 2) || (ints[8] != 9) || (ints[9] != 20)) {
            printf(" pV3 Concentrator: Illegal Init HandShake!\n");
            exit(1);
          }
          pvm_upkint(buf, ints[7], 1);
          if ((TimeOut != 0) && (buf[1] != 0)) {
            TimeOut = buf[1];
            printf(" pV3 Concentrator: Time Out = %d\n", TimeOut);
          }
          pvm_upkfloat((float *) &buf[2], ints[8], 1);
          pvm_upkbyte(dtit, ints[9], 1);
          for (did = 0; did < ndisc; did++) {
            for (hit = k = 0; k < 20; k++) 
              if (dtit[k] != discipl[did].title[k]) hit++;
            if (hit == 0) {
              dlookup[j] = did;
              discipl[did].did = j;
              printf(" pV3 Concentrator: ");
              for (hit = 0; hit < 19; hit++) printf("%c", dtit[hit]);
              printf(" assigned to DID = %d\n", j);
              break;
            }
          }
          did = dlookup[j];
          if (did == -1) {
            printf(" pV3 Concentrator: 2. Error DID = %d,  MID = %d\n",
                   j, ints[0]);
          } else {
            if ((i <= 0) || (i > discipl[did].mrank)) {
              printf(" pV3 Concentrator: 2. Error DID = %d,  CID = %d,  MID = %d\n",
                     did, i, ints[0]);
              goto next;
            }
            if ((lrank != -1) && (lrank != discipl[did].id2r[i-1]) &&
                (discipl[did].id2r[i-1] != -1)) pv_MesBuffer(0, 0);
            pv3_srank = discipl[did].id2r[i-1];
            if (pv3_srank != -1) lrank = pv3_srank;
#ifdef PVDEBUG
            printf("MID = %d, DID = %d, CID = %d, rank = %d\n", ints[0], did,
                   ints[2], pv3_srank);
#endif
            if (pv3_srank != -1) {
              pv_MesBuffer(2, (int) ((11+ints[7]+ints[8])*sizeof(int)+ints[9]));
              pv_pkint(&mstid,  1);
              pv_pkint(ints,   10);
              pv_pkint(buf, ints[7]);
              pv_pkfloat((float *) &buf[2], ints[8]);
              pv_pkbyte(dtit, ints[9]);
            }
          }
          goto next;
        }
      }
normal:
      if ((j < 0) || (j >= mlookup)) {
        pv3_srank = -1;
        printf(" pV3 Concentrator: 1. Error DID = %d,  MID = %d\n", j, ints[0]);
      } else {
        did = dlookup[j];
        if (did == -1) {
          pv3_srank = -1;
          printf(" pV3 Concentrator: 1. Bad DID -- CID = %d,  MID = %d\n", i,
                 ints[0]);
        } else {
          if ((i <= 0) || (i > discipl[did].mrank)) {
            pv3_srank = -1;
            printf(" pV3 Concentrator: 1. Error DID = %d,  CID = %d,  MID = %d\n", 
                   did, i, ints[0]);
          } else {
            if ((lrank != -1) && (lrank != discipl[did].id2r[i-1]) &&
                (discipl[did].id2r[i-1] != -1)) pv_MesBuffer(0, 0);
            pv3_srank = discipl[did].id2r[i-1];
            if (pv3_srank != -1) lrank = pv3_srank;
          }
        }
      }
#ifdef PVDEBUG
      printf("MID = %d, DID = %d, CID = %d, rank = %d\n", ints[0], did,
              ints[2], pv3_srank);
#endif
      if (pv3_srank != -1) {
        pv_MesBuffer(2, (int) ((11+ints[7]+ints[8])*sizeof(int)+ints[9]));
        pv_pkint(&mstid,  1);
        pv_pkint(ints,   10);
      }
      if (ints[7] != 0) {
        pvm_upkint(buf, ints[7], 1);
        if (pv3_srank != -1) pv_pkint(buf, ints[7]);
      }
      if (ints[8] != 0) {
        pvm_upkfloat((float *)buf, ints[8], 1);
        if (pv3_srank != -1) pv_pkfloat((float *)buf, ints[8]);
      }
      if (ints[9] != 0) {
        pvm_upkbyte((char *)buf, ints[9], 1);
        if (pv3_srank != -1) pv_pkbyte((char *)buf, ints[9]);
      }
      if (pv3_srank == -1) pv3_srank = lrank;
next:
      ints[0] = -1;
      pvm_upkint(ints, 1, 1);
    }
    if (lrank != -1) pv_MesBuffer(0, 0);
    last = time(NULL);
    goto loop;
  }

  /* hi prio messages from clients */

  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, pv3_hicomm, &i, &status);
  if (i != 0) {
    pv3_srbuf = 2;
    MPI_Recv(pv3_hrbuf, pv3_rsize, MPI_PACKED, MPI_ANY_SOURCE, MPI_ANY_TAG,
             pv3_hicomm, &status);
    pv3_hrpos = 0;
    pv_upkint(&mstid, 1);
    pv3_sindx = -1;
    for (j = 0; j < pv3_nser; j++) if (mstid == pv3_ser[j].tid) pv3_sindx = j;
    while (mstid != -1) {
      if (pv3_sindx == -1) {
        printf(" pV3 Concentrator: 2. Server not found: %X!\n", mstid);
      } else {
        if (pv3_ser[pv3_sindx].hwbuf != 0) {
          pvm_setsbuf(pv3_ser[pv3_sindx].hwbuf);
        } else {
          pvm_initsend(PvmDataDefault);
          pvm_pkbyte("pV3#", 4, 1);
        }
      }
      pv_upkint(ints, 9);
      if (pv3_sindx != -1) pvm_pkint(ints, 9, 1);
      if (ints[6] != 0) {
        pv_upkint(buf, ints[6]);
        if (pv3_sindx != -1) pvm_pkint(buf, ints[6], 1);
      }
      if (ints[7] != 0) {
        pv_upkfloat((float *)buf, ints[7]);
        if (pv3_sindx != -1) pvm_pkfloat((float *)buf, ints[7], 1);
      }
      if (ints[8] != 0) {
        pv_upkbyte((char *)buf, ints[8]);
        if (pv3_sindx != -1) pvm_pkbyte((char *)buf, ints[8], 1);
      }
      if (pv3_sindx != -1) pv3_ser[pv3_sindx].hwbuf = pvm_setsbuf(0);
      mstid = -1;
      pv_upkint(&mstid, 1);
      if (mstid != -1) {
        pv3_sindx = -1;
        for (j = 0; j < pv3_nser; j++) 
          if (mstid == pv3_ser[j].tid) pv3_sindx = j;
      }
    }
    last = time(NULL);
    goto loop;
  }

  /* clear hi prio messages from clients */

  for (i = j = 0; j < pv3_nser; j++)
    if (pv3_ser[j].hwbuf != 0) {
      pvm_setsbuf(pv3_ser[j].hwbuf);
      ints[0] = -1;
      pvm_pkint(ints, 1, 1);
      stat = pvm_send(pv3_ser[j].tid, HIPRIO);
      if (stat != 0) printf(" pV3 Concentrator: 1. Send to %X status = %d\n",
                              pv3_ser[pv3_sindx].tid, stat);
      pvm_freebuf(pv3_ser[j].hwbuf);
      pv3_ser[j].hwbuf = 0;
      i++;
    }
  if (i != 0) goto loop;

  /* lo prio messages from server */

  if (pvm_probe(-1, LOPRIO) > 0) {
    bufid = pvm_recv(-1, LOPRIO);
    pvm_bufinfo(bufid, &len, &i, &mstid);
    for (k = j = 0; j < pv3_nser; j++) if (mstid == pv3_ser[j].tid) k++;
    if (k == 0) {
      printf(" pV3 Concentrator: 3. Server not found: %X!\n", mstid);
      goto loop;
    }
#ifdef PVDEBUG
    printf("Get lo prio message from server: %X\n", mstid);
#endif
    pv3_srank = -1;
    lrank     = -1;
    ints[0]   = -1;
    pvm_upkint(ints, 1, 1);
    while (ints[0] != -1) {
      pvm_upkint(&ints[1], 9, 1);
      j = ints[1];
      i = ints[2];
      if (j >= mlookup) {
        pv3_srank = -1;
        printf(" pV3 Concentrator: 3. Error DID = %d,  MID = %d\n", j, ints[0]);
      } else {
        did = dlookup[j];
        if (did == -1) {
          pv3_srank = -1;
          printf(" pV3 Concentrator: 2. Bad DID -- CID = %d,  MID = %d\n", i,
                 ints[0]);
        } else {
          if ((i <= 0) || (i > discipl[did].mrank)) {
            pv3_srank = -1;
            printf(" pV3 Concentrator: 3. Error DID = %d,  CID = %d,  MID = %d\n",
                   did, i, ints[0]);
          } else {
            if (lrank != discipl[did].id2r[i-1]) {
              if (lrank != -1) lopos[lrank] = pv3_lopos;
              pv3_srank = discipl[did].id2r[i-1];
              if (pv3_srank != -1) {
                pv3_lopos = lopos[pv3_srank];
                pv3_lobuf = lobuf[pv3_srank];
                lrank = pv3_srank;
              }
            }
          }
        }
      }
#ifdef PVDEBUG
      printf("MID = %d, DID = %d, CID = %d, rank = %d\n", ints[0], did,
             ints[2], pv3_srank);
#endif
      if (pv3_srank != -1) {
        pv_MesBuffer(1, (int) ((11+ints[7]+ints[8])*sizeof(int)+ints[9]));
        pv_pkint(&mstid,  1);
        pv_pkint(ints,   10);
      }
      if (ints[7] != 0) {
        pvm_upkint(buf, ints[7], 1);
        if (pv3_srank != -1) pv_pkint(buf, ints[7]);
      }
      if (ints[8] != 0) {
        pvm_upkfloat((float *)buf, ints[8], 1);
        if (pv3_srank != -1) pv_pkfloat((float *)buf, ints[8]);
      }
      if (ints[9] != 0) {
        pvm_upkbyte((char *)buf, ints[9], 1);
        if (pv3_srank != -1) pv_pkbyte((char *)buf, ints[9]);
      }
      /* clear buffers with framing messages */
      if (pv3_srank != -1)
        if ((ints[0] == 34) || (ints[0] == 36)) {
#ifdef PVDEBUG
          printf("Send lo prio message: rank = %d,  len = %d\n", pv3_srank,
                  pv3_lopos);
#endif
          pv_MesBuffer(0, 0);
          lopos[pv3_srank] = 0;
        }
      ints[0] = -1;
      pvm_upkint(ints, 1, 1);
    }
    if (lrank != -1) lopos[lrank] = pv3_lopos;
    last = time(NULL);
    goto loop;
  }

  /* lo prio messages from clients */

  MPI_Iprobe(MPI_ANY_SOURCE, MPI_ANY_TAG, pv3_locomm, &i, &status);
  if (i != 0) {
    if (status.MPI_TAG == PVDOWN) pv_clscan();
    pv3_srbuf = 1;
    MPI_Recv(pv3_lrbuf, pv3_rsize, MPI_PACKED, MPI_ANY_SOURCE, MPI_ANY_TAG,
             pv3_locomm, &status);
    pv3_lrpos = 0;
    pv_upkint(&mstid, 1);
    pv3_sindx = -1;
    for (j = 0; j < pv3_nser; j++) if (mstid == pv3_ser[j].tid) pv3_sindx = j;
    while (mstid != -1) {
      if (pv3_sindx == -1) {
        printf(" pV3 Concentrator: 4. Server not found: %X!\n", mstid);
      } else {
        if (pv3_ser[pv3_sindx].hwbuf != 0) {
          pvm_setsbuf(pv3_ser[pv3_sindx].hwbuf);
          ints[0] = -1;
          pvm_pkint(ints, 1, 1);
          stat = pvm_send(pv3_ser[pv3_sindx].tid, HIPRIO);
          if (stat != 0) 
            printf(" pV3 Concentrator: 2. Send to %X status = %d\n",
                     pv3_ser[pv3_sindx].tid, stat);
          pvm_freebuf(pv3_ser[pv3_sindx].hwbuf);
          pv3_ser[pv3_sindx].hwbuf = 0;
        }
        if (pv3_ser[pv3_sindx].lwbuf != 0) {
          pvm_setsbuf(pv3_ser[pv3_sindx].lwbuf);
          if (status.MPI_TAG != pv3_ser[pv3_sindx].ltag) {
            ints[0] = -1;
            pvm_pkint(ints, 1, 1);
            stat = pvm_send(pv3_ser[pv3_sindx].tid, pv3_ser[pv3_sindx].ltag);
            if (stat != 0) 
               printf(" pV3 Concentrator: 3. Send to %X status = %d\n",
                        pv3_ser[pv3_sindx].tid, stat);
            pvm_freebuf(pv3_ser[pv3_sindx].lwbuf);
            pv3_ser[pv3_sindx].lwbuf = 0;
          }
        }
        if (pv3_ser[pv3_sindx].lwbuf == 0) {
          pvm_initsend(PvmDataDefault);
          pvm_pkbyte("pV3#", 4, 1);
          pv3_ser[pv3_sindx].wrdcnt = 0;
        }
        pv3_ser[pv3_sindx].ltag = status.MPI_TAG;
      }

      pv_upkint(ints, 9);
      wrds = 9 + ints[6] + ints[7] + ints[8]/sizeof(int);

      if (pv3_sindx != -1) {
        if ((pv3_ser[pv3_sindx].wrdcnt+wrds >= MAXLEN) &&
            (pv3_ser[pv3_sindx].wrdcnt != 0)) {
          mid = -1;
          pvm_pkint(&mid, 1, 1);
          stat = pvm_send(pv3_ser[pv3_sindx].tid, pv3_ser[pv3_sindx].ltag);
          if (stat != 0) 
            printf(" pV3 Concentrator: 4. Send to %X status = %d\n",
                     pv3_ser[pv3_sindx].tid, stat);
          pvm_initsend(PvmDataDefault);
          pvm_pkbyte("pV3#", 4, 1);
          pv3_ser[pv3_sindx].wrdcnt = 0;
        }
        pv3_ser[pv3_sindx].wrdcnt += wrds;
        pvm_pkint(ints, 9, 1);
      }

      if (ints[6] != 0) {
        pv_upkint(buf, ints[6]);
        if ((status.MPI_TAG == INPRIO) && (ints[0] == 32)) {
          if (finit == 0) buf[0] = pv3_ncl;
          finit++;
        }
        if (pv3_sindx != -1) pvm_pkint(buf, ints[6], 1);
      }
      if (ints[7] != 0) {
        pv_upkfloat((float *)buf, ints[7]);
        if (pv3_sindx != -1) pvm_pkfloat((float *)buf, ints[7], 1);
      }
      if (ints[8] != 0) {
        pv_upkbyte((char *)buf, ints[8]);
        if (pv3_sindx != -1) pvm_pkbyte((char *)buf, ints[8], 1);
      }
      if ((status.MPI_TAG == INPRIO) && (ints[0] == 32)) {
        cbuf = (char *)buf;
        for (j = 0; j < 20; j++) dtit[j] = cbuf[j+80];
        hit = -1;
        for (did = 0; did < ndisc; did++) {
          for (i=j=0; j < 20; j++) if (dtit[j] != discipl[did].title[j]) i++;
          if (i == 0) hit = did;
        }
        if (hit == -1) {
          if (ndisc == mdisc) {
            mdisc += 2;
            discipl = (Discstr *) realloc(discipl, mdisc*sizeof(Discstr));
            if (discipl == NULL) {
              printf(" pV3 Concentrator Error in memory re-allocation!\n");
              exit(1);
            }
            discipl[ndisc+1].id2r = NULL;
          }
          hit = ndisc;
          discipl[ndisc].mrank = pv3_total;
          if (ints[2] > pv3_total) discipl[ndisc].mrank = ints[2];
          discipl[ndisc].id2r = (int *)malloc(discipl[ndisc].mrank*sizeof(int));
          if (discipl[ndisc].id2r == NULL) {
            printf(" pV3 Concentrator Error in memory allocation (id2r)!\n");
            exit(1);
          }
          for (i = 0; i < discipl[ndisc].mrank; i++) discipl[ndisc].id2r[i]=-1;
          for (j = 0; j < 20; j++) discipl[ndisc].title[j] = dtit[j];
#ifdef PVDEBUG
          dtit[19] = 0;
          printf(" New Discipline: idisc = %d,  title = %s\n", ndisc, dtit);
#endif
          discipl[ndisc].did = -1;
          ndisc++;
        } else {
          if (ints[2] > discipl[hit].mrank) {
            tmp = (int *) realloc(discipl[hit].id2r, ints[2]*sizeof(int));
            if (tmp != NULL) {
              discipl[hit].id2r  = tmp;
              for (i = discipl[hit].mrank; i < ints[2]; i++) 
                discipl[hit].id2r[i] = -1;
              discipl[hit].mrank = ints[2];
            }
          }
        }
        if (ints[2] <= discipl[hit].mrank) 
          discipl[hit].id2r[ints[2]-1] = status.MPI_SOURCE;
        if (discipl[hit].did == -1) {
          for (j = 0; j < 19; j++) dtit[j] = discipl[hit].title[j];
          dtit[19] = 0;
          printf(" pV3 Concentrator: %s CID = %d,  MPI rank = %d (of %d)\n",
                 dtit, ints[2], status.MPI_SOURCE, pv3_total);
        } else {
          printf(" pV3 Concentrator: DID = %d,  CID = %d,  MPI rank = %d (of %d)\n",
                 discipl[hit].did, ints[2], status.MPI_SOURCE, pv3_total);
        }
      }
      if (pv3_sindx != -1) pv3_ser[pv3_sindx].lwbuf = pvm_setsbuf(0);
      mstid = -1;
      pv_upkint(&mstid, 1);
      if (mstid != -1) {
        pv3_sindx = -1;
        for (j = 0; j < pv3_nser; j++) 
          if (mstid == pv3_ser[j].tid) pv3_sindx = j;
      }
    }
    last = time(NULL);
    goto loop;
  }

  /* clear lo prio pvm and mpi send buffers */

  for (i = pv3_srank = 0; pv3_srank < pv3_total; pv3_srank++)
    if ((pv3_srank != pv3_conc) && (lopos[pv3_srank] != 0)) {
      pv3_prio  = 1;
      pv3_lopos = lopos[pv3_srank];
      pv3_lobuf = lobuf[pv3_srank];
#ifdef PVDEBUG
      printf("Send lo prio message: rank = %d,  len = %d\n", pv3_srank,
              pv3_lopos);
#endif
      pv_MesBuffer(0, 0);
      lopos[pv3_srank] = 0;
      i++;
    }
  if (i != 0) goto loop;
  
  for (j = 0; j < pv3_nser; j++)
    if (pv3_ser[j].lwbuf != 0) {
      pvm_setsbuf(pv3_ser[j].lwbuf);
      ints[0] = -1;
      pvm_pkint(ints, 1, 1);
      stat = pvm_send(pv3_ser[j].tid, pv3_ser[j].ltag);
      if (stat != 0) printf(" pV3 Concentrator: 5. Send to %X status = %d\n",
                              pv3_ser[pv3_sindx].tid, stat);
      pvm_freebuf(pv3_ser[j].lwbuf);
      pv3_ser[j].lwbuf =  0;
      pv3_ser[j].ltag  = -1;
    }

  /* check for timeout */

  if (TimeOut != 0)
    if (time(NULL) - last > TimeOut) {

      /* we do no know where the problem is (i.e., which server)
           so we let the clients take care of this */

      printf(" pV3 Concentrator: Server(s) have Timed Out!\n");

      for (j = i = 0; i < pv3_total; i++)
        if (i != pv3_conc) MPI_Bsend(&j, 1, MPI_INT, i, PVDOWN, pv3_hicomm);
      last = time(NULL);
    }

  /* any new servers? */

  pv_mSleep(1);
#ifdef NOGROUPS
  while (pvm_probe(pgid, 1) > 0) {
    pvm_recv(pgid, 1);
    pvm_upkint(&gsize, 1, 1);
    if (gsize > 2*MAXSER) gsize = 2*MAXSER;
    if (gsize > 0) pvm_upkint(pgsids, gsize, 1);
  }
#else
  gsize = pvm_gsize(pv3_SGroup);
#endif
  ins = -1;
  for (i = 0; i < gsize; i++) {
#ifdef NOGROUPS
    tid = pgsids[i];
#else
    tid = PvmNoInst;
    while ((tid == PvmNoInst) && (ins < MAXSERINS)) {
      ins++;
      tid = pvm_gettid(pv3_SGroup, ins);
    }
#endif
    if (tid > 0) {
      for (hit = j = 0; j < pv3_nser; j++) if (pv3_ser[j].tid == tid) hit++;
      if (hit == 0) {
        if ((pvm_pstat(tid) >= 0) && (pv3_nser < MAXSER)) {
          printf(" pV3 Concentrator: Server %X has started!\n", tid);
          pv3_ser[pv3_nser].tid    = tid;
          pv3_ser[pv3_nser].status = 1;
          pv3_ser[pv3_nser].hwbuf  = 0;
          pv3_ser[pv3_nser].lwbuf  = 0;
          pv3_ser[pv3_nser].ltag   = -1;
          pv3_nser++;
          for (j = 0; j < pv3_total; j++)
            if (j != pv3_conc) MPI_Bsend(&tid, 1, MPI_INT, j,
                                         PVUP, pv3_hicomm);
        }
      }
    }
  }

  /* any of our servers die? */

  for (i = 0; i < pv3_nser; i++)
    if (pvm_pstat(pv3_ser[i].tid) < 0) {
      tid = pv3_ser[i].tid;
      printf(" pV3 Concentrator: Server %X has exitted!\n", tid);
      for (j = 0; j < pv3_total; j++)
        if (j != pv3_conc) MPI_Bsend(&tid, 1, MPI_INT, j, PVDOWN, pv3_hicomm);
      pv3_ser[i].status = 5;
      pv_mesClear(tid);
    }
  j = pv3_nser;
  pv3_nser = 0;
  for (i = 0; i < j; i++)
    if (pv3_ser[i].status == 5) {
      if (pv3_ser[i].hwbuf != 0) pvm_freebuf(pv3_ser[i].hwbuf);
      if (pv3_ser[i].lwbuf != 0) pvm_freebuf(pv3_ser[i].lwbuf);
    } else {
      pv3_ser[pv3_nser].tid    = pv3_ser[i].tid;
      pv3_ser[pv3_nser].status = pv3_ser[i].status;
      pv3_ser[pv3_nser].hwbuf  = pv3_ser[i].hwbuf;
      pv3_ser[pv3_nser].lwbuf  = pv3_ser[i].lwbuf;
      pv3_ser[pv3_nser].ltag   = pv3_ser[i].ltag;
      pv3_ser[pv3_nser].wrdcnt = pv3_ser[i].wrdcnt;
      pv3_nser++;
    }

  if (pv3_nser == 0) {
    printf(" pV3 Concentrator: Wait for Server/Batch to startup 1...\n");
    goto sstartup;
  }

  goto loop;

}


/* **************** Routines Used by the Clients **************** */


/* open the MPI concentrator and set the new communicator 
   returns  0 = OK
           -1 = bsize too small
           -2 = rank range error
           -3 = alloc error

   NOTE: This is a C only binding
 */

int
pV_MPIStart (MPI_Comm commin, int rank, int ncl, int bsize, MPI_Comm *commout)
{

  if ((bsize != 0) && ((MAXLEN+32)*sizeof(int) > bsize)) return -1;

  MPI_Comm_size(commin, &pv3_total);
  MPI_Comm_rank(commin, &pv3_myrank);
  pv3_ncl = ncl;
  if (ncl == 0) pv3_ncl = pv3_total-1;
  if ((rank < 0) || (rank >= pv3_total)) {
    MPI_Comm_dup(commin, commout);
    pv3_conc = -1;
    pv3_comm = *commout;
  } else {
    MPI_Comm_dup(commin, &pv3_hicomm);
    MPI_Comm_dup(commin, &pv3_locomm);
    MPI_Comm_group(commin, &pv3_groupfull);
    MPI_Group_excl(pv3_groupfull, 1, &rank, &pv3_group);
    MPI_Comm_create(commin, pv3_group, commout);
    pv3_comm  = *commout;
    pv3_conc  = rank;
    pv3_srank = rank;
    pv3_bsize = (MAXLEN+32)*sizeof(int);
    if (bsize == 0) {
      pv3_buffer = (char *) malloc(pv3_bsize);
      if (pv3_buffer == NULL) {
        printf(" pV3/MPI Failure in allocating buffer!\n");
        if (rank != pv3_myrank) {
          return -3;
        } else {
          exit(1);
        }
      }
      MPI_Buffer_attach(pv3_buffer, pv3_bsize);
    }
  }

  if (rank == pv3_myrank) pv_concentrator();

  if (pv3_conc != -1) {
    pv3_psize = pv3_bsize;
    pv3_hibuf = (char *) malloc(pv3_psize);
    if (pv3_hibuf == NULL) {
      printf(" pV3/MPI Failure in allocating buffer!\n");
      return -3;
    }
    pv3_lobuf = (char *) malloc(pv3_psize);
    if (pv3_lobuf == NULL) {
      printf(" pV3/MPI Failure in allocating buffer!\n");
      return -3;
    }

    pv3_rsize = MAXREQ*sizeof(int);
    pv3_hrbuf = (char *) malloc(pv3_rsize);
    if (pv3_hrbuf == NULL) {
      printf(" pV3/MPI Failure in allocating buffer!\n");
      return -3;
    }
    pv3_lrbuf = (char *) malloc(pv3_rsize);
    if (pv3_lrbuf == NULL) {
      printf(" pV3/MPI Failure in allocating buffer!\n");
      return -3;
    }
    return 0;
  }
  return -2;
}


void
pV_MPISTART (int *commhin,  int *rank, int *ncl, int *bsize,
             int *commhout, int *ierror)
{
  MPI_Comm commin, commout;

  commin    = MPI_Comm_f2c(*commhin);
  *ierror   = pV_MPIStart(commin, *rank, *ncl, *bsize, &commout);
  *commhout = MPI_Comm_c2f(commout);
}


/* close the MPI subsystem */

void
pV_MPISTOP ()
{
  if (pv3_conc != -1) {
    MPI_Bsend(pv3_lobuf, 1, MPI_INT, pv3_conc, PVDOWN, pv3_locomm);
    MPI_Barrier(pv3_locomm);
  }

  /* free up stuff */
  MPI_Comm_free(&pv3_comm);

  if (pv3_conc != -1) {
    MPI_Comm_free(&pv3_locomm);
    MPI_Comm_free(&pv3_hicomm);
    MPI_Group_free(&pv3_groupfull);
    MPI_Group_free(&pv3_group);
  }
}


/* check for the server through the concentrator */

void
pV_SERVER (int *type)
{
  int        i;
  MPI_Status status;

  pv3_did  = -1;
  pv3_nser =  0;

  if (*type == 0) {
    /* init and wait for startup */

    if (pv3_conc == -1) exit(1);
    MPI_Recv(&i, 1, MPI_INT, pv3_conc, PVUP, pv3_hicomm, &status);
    pv3_ser[pv3_nser].tid    = i;
    pv3_ser[pv3_nser].kvpa   = 0;
    pv3_ser[pv3_nser].status = 0;
    pv3_ser[pv3_nser].ltag   = 0;
    pv3_nser++;

  } else {
    /* init and don't wait for startup */

    if (pv3_conc == -1) {
      *type = -1;
      return;
    }
    MPI_Iprobe(pv3_conc, PVUP, pv3_hicomm, &i, &status);
    if (i == 0) {
      *type = -1;
      return;
    }
    MPI_Recv(&i, 1, MPI_INT, pv3_conc, PVUP, pv3_hicomm, &status);
    pv3_ser[pv3_nser].tid    = i;
    pv3_ser[pv3_nser].kvpa   = 0;
    pv3_ser[pv3_nser].status = 0;
    pv3_ser[pv3_nser].ltag   = 0;
    pv3_nser++;
  }
}


/* init/terminate the network system for the client */

void
pV_NETCONTROL (int *type)
{

  pv3_did   = -1;

  if (*type == 0) {

    /* start */
    if (pv3_conc == -1) {
      printf(" pV3 Error: Concentrator Not running!\n");
      *type = -2;
      return;
    }
    *type = MAXSER;
    
  } else {

    if(*type == 1) {
#ifdef WIN32
      PV_MPISTOP();
#else
      pV_MPISTOP();
#endif
      MPI_Finalize();
      exit(0);
    }

  }

}


/* set client id */

void
pV_SETCID (int *cid)
{
  pv3_id = *cid;
}


/* return server data */

void
pV_SERVERSTAT (int *indx, int *nser, int *stid, int *kvpa, int *stat)
{

  *nser = pv3_nser;
  *indx = pv3_sindx + 1;
  *stid = pv3_ser[pv3_sindx].tid;
  *kvpa = pv3_ser[pv3_sindx].kvpa;
  *stat = pv3_ser[pv3_sindx].status;
}


/* set server information */

void
pV_SETCSERVER (int *indx, int *kvpa, int *stat)
{
  int        i, j, tid;
  MPI_Status status;

  /* indx = -2 -- get new servers */
  if (*indx == -2) {
    i = 1;
    while (i != 0) {
      MPI_Iprobe(pv3_conc, PVUP, pv3_hicomm, &i, &status);
      if (i != 0) {
        MPI_Recv(&tid, 1, MPI_INT, pv3_conc, PVUP, pv3_hicomm, &status);
        if (pv3_nser < MAXSER) {
          pv3_ser[pv3_nser].tid    = tid;
          pv3_ser[pv3_nser].kvpa   = 0;
          pv3_ser[pv3_nser].status = 0;
          pv3_ser[pv3_nser].ltag   = 0;
          pv3_nser++;
        }
      }
    }
    return;
  }

  /* indx = -1 -- clear up terminated servers */
  if (*indx == -1) {
    j = pv3_nser;
    pv3_nser = 0;
    for (i = 0; i < j; i++)
      if (pv3_ser[i].status != 5) {
        pv3_ser[pv3_nser].tid    = pv3_ser[i].tid;
        pv3_ser[pv3_nser].kvpa   = pv3_ser[i].kvpa;
        pv3_ser[pv3_nser].status = pv3_ser[i].status;
        pv3_ser[pv3_nser].ltag   = pv3_ser[i].ltag;
        pv3_nser++;
      }
    if ((pv3_nser == 0) && (*stat == -1))
      printf(" pV3 Warning: No servers after cleanup!\n");

    i = 1;
    while (i != 0) {
      MPI_Iprobe(pv3_conc, PVUP, pv3_hicomm, &i, &status);
      if (i != 0) {
        MPI_Recv(&tid, 1, MPI_INT, pv3_conc, PVUP, pv3_hicomm, &status);
        if (pv3_nser < MAXSER) {
          pv3_ser[pv3_nser].tid    = tid;
          pv3_ser[pv3_nser].kvpa   = 0;
          pv3_ser[pv3_nser].status = 0;
          pv3_ser[pv3_nser].ltag   = 0;
          pv3_nser++;
        }
      }
    }
    pv3_sindx = 0;
    pv3_stid  = pv3_ser[pv3_sindx].tid;
    return;
  }

  if ((*indx > 0) && (*indx <= pv3_nser)) {
    pv3_sindx = *indx-1;
    pv3_stid  = pv3_ser[pv3_sindx].tid;
    if (*kvpa != -1) pv3_ser[pv3_sindx].kvpa   = *kvpa;
    if (*stat != -1) pv3_ser[pv3_sindx].status = *stat;
  } else {
    printf(" pV3 Warning: Index out of range (1-%d): %d\n", pv3_nser, *indx);
  }

}


/* change the message matching critera */

void
pV_SETMATCH (int *type)
{

  if (*type == 0) {

    /* start message handling */
    pv3_prio  = 0;
    pv3_hipos = 0;
    pv3_lopos = 0;
    pv3_srbuf = 0;
    pv3_hrpos = 0;
    pv3_lrpos = 0;

  } else {

    /* terminate message handling */
    pv_MesBuffer(0, 0);
  }
}


/* pack and send init handshake to server */

void
#ifdef T3E
pV_SENDINIT (int *iopt, int *npc, char *tc, int tcLEN, int *mirror,
             FLOAT *repmat, FLOAT *rev, int *nk, int *ik, int *fk, 
             FLOAT fl[][2], char *tk, char *tit, int titLEN, char *dtit, 
             int dtitLEN, char *ptit, int ptitLEN, FLOAT *xmx, int *kns, 
             int *ns, char *nt, int *mblk)
#else
#ifdef WIN32
pV_SENDINIT (int *iopt, int *npc, char *tc, int tcLEN, int *mirror, 
             FLOAT *repmat, FLOAT *rev, int *nk, int *ik, int *fk, 
             FLOAT fl[][2], char *tk, int tkLEN, char *tit, int titLEN, 
             char *dtit, int dtitLEN, char *ptit, int ptitLEN, FLOAT *xmx, 
             int *kns, int *ns, char *nt, int ntLEN, int *mblk)
#else
pV_SENDINIT (int *iopt, int *npc, char *tc, int *mirror, FLOAT *repmat, 
             FLOAT *rev, int *nk, int *ik, int *fk, FLOAT fl[][2], char *tk, 
             char *tit, char *dtit, char *ptit, FLOAT *xmx, int *kns, int *ns,
             char *nt, int *mblk)
#endif
#endif
{
  int size, ier, i, l;

  ier = pv3_ncl;

  size = ( 22 + abs(*nk)*4  + *kns*3)*sizeof(int) +
          120 + abs(*nk)*32 + *kns*20 + *npc*32;
  if (*mirror < 0) size += 16*sizeof(int);
  pv_MesBuffer(1, size);
  pv_pkint(&pv3_stid, 1);
  i = 32;
  pv_pkint(&i, 1);
  pv_pkint(&pv3_did, 1);
  pv_pkint(&pv3_id, 1);
  pv_pkint(&ier, 1);
  pv_pkint(kns, 1);
  pv_pkint(iopt, 1);

  l = 5 + abs(*nk)*2 + *kns*3;
  pv_pkint(&l, 1);
  l = 7 + abs(*nk)*2;
  if (*mirror < 0) l += 16;
  pv_pkint(&l, 1);
  l = 120 + abs(*nk)*32 + *kns*20 + *npc*32;
  pv_pkint(&l, 1);

  i = 0;
  pv_pkint(&i, 1);
  pv_pkint(npc, 1);
  pv_pkint(mirror, 1);
  pv_pkint(nk, 1);
  for (i = 0; i < abs(*nk); i++) {
    pv_pkint(&ik[i], 1);
    pv_pkint(&fk[i], 1);
  }
  if (*kns != 0) pv_pkint(ns, *kns*3);
  pv_pkint(mblk, 1);

  pv_pkFLOAT(rev, 1);
  pv_pkFLOAT(xmx, 6);
  if (*mirror < 0) pv_pkFLOAT(repmat, 16);
  for (i = 0; i < abs(*nk); i++) pv_pkFLOAT(fl[i], 2);

  pv_pkbyte(tit, 80);
  pv_pkbyte(dtit, 20);
  pv_pkbyte(ptit, 20);
  for (i = 0; i < abs(*nk); i++) pv_pkbyte(&tk[i*32], 32);
  if (*kns != 0) pv_pkbyte(nt, *kns*20);
  if (*npc != 0) pv_pkbyte(tc, *npc*32);

  pv_MesBuffer(0, INPRIO);
}


/* pack ack type message */

void
pV_SENDMES (int *id, int *ints, FLOAT *reals)
{
  int l;

  /* used for mid = 35 & 37 */

  pv_MesBuffer(1, (int) (14*sizeof(int)));
  pv_pkint(&pv3_stid, 1);
  pv_pkint(id, 1);
  pv_pkint(&pv3_did, 1);
  pv_pkint(&pv3_id, 1);
  pv_pkint(ints, 3);

  l = 0;
  pv_pkint(&l, 1);
  l = 5;
  pv_pkint(&l, 1);
  l = 0;
  pv_pkint(&l, 1);

  pv_pkFLOAT(reals, 5);
}


/* pack int extract */

void
pV_SENDXI (int *id, int *num, int *type, int *nd, int *len, int *ints)
{
  int size, i, np, l;

  size = *nd;
  size = *len*size;
  if (size <= 0) return;

  i = 0;
  while (size > 0) {
    np = MAXLEN;
    if (np > size) np = size;
    pv_MesBuffer(1, (int) ((np+11)*sizeof(int)));
    pv_pkint(&pv3_stid, 1);
    pv_pkint(id, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(type, 1);
    pv_pkint(len, 1);
    pv_pkint(num, 1);

    l = 2 + np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    pv_pkint(&ints[i], np);
    size -= np;
    i += np;
  }
}


/* pack float extract */

void
pV_SENDXR (int *id, int *num, int *type, int *nd, int *len, FLOAT *reals)
{
  int size, i, np, l;

  size = *nd;
  size = *len*size;
  if (size <= 0) return;

  i = 0;
  while (size > 0) {
    np = MAXLEN;
    if (np > size) np = size;
    pv_MesBuffer(1, (int) ((np+11)*sizeof(int)));
    pv_pkint(&pv3_stid, 1);
    pv_pkint(id, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(type, 1);
    pv_pkint(len, 1);
    pv_pkint(num, 1);

    l = 2;
    pv_pkint(&l, 1);
    l = np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    pv_pkFLOAT(&reals[i], np);
    size -= np;
    i += np;
  }
}


/* pack vector cloud extracts */

void
pV_SENDXV (int *type, int *nd, FLOAT *reals, FLOAT *out, FLOAT *t, int *n)
{
  int i, np, mid, num, j, it, remain, l;

  mid = 20;
  num = *n;
  remain = *nd*num;
  if (remain <= 0) return;

  it = 0;
  i  = 0;
  while (remain > 0) {
    np = MAXLEN;
    j = *nd;
    np = np/j;
    np *= j;
    if (np > remain) np = remain;
    pv_MesBuffer(1, (int) ((np+11)*sizeof(int)));
    j = 0;
    pv_pkint(&pv3_stid, 1);
    pv_pkint(&mid, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(type, 1);
    pv_pkint(&num, 1);
    pv_pkint(&j, 1);

    l = 2;
    pv_pkint(&l, 1);
    l = np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    while (j < np) {
      pV_FILLVCLOUD(type, nd, t, reals, &it, out);
      pv_pkFLOAT(out, *nd);
      j  += *nd;
    }
    remain -= np;
    i += np;
  }
}


/* pack streamline int extract */

void
pV_SENDSLXI (int *id, int *num, int *type, int *nd, int *len, int *nseg,
             int *ints)
{
  int size, i, np, l;

  size = *nd;
  size = *len*size;

  i = 0;
  while (size > 0) {
    np = MAXLEN;
    if (np > size) np = size;
    pv_MesBuffer(1, (int) ((np+12)*sizeof(int)));
    pv_pkint(&pv3_stid, 1);
    pv_pkint(id, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(type, 1);
    pv_pkint(len, 1);
    pv_pkint(num, 1);

    l = 3 + np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(nseg, 1);
    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    pv_pkint(&ints[i], np);
    size -= np;
    i += np;
  }
}


/* pack streamline float extract */

void
pV_SENDSLXR (int *id, int *num, int *type, int *nd, int *len, int *nseg,
             FLOAT *reals)
{
  int size, i, np, l;

  size = *nd;
  size = *len*size;

  i = 0;
  while (size > 0) {
    np = MAXLEN;
    if (np > size) np = size;
    pv_MesBuffer(1, (int) ((np+12)*sizeof(int)));
    pv_pkint(&pv3_stid, 1);
    pv_pkint(id, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(type, 1);
    pv_pkint(len, 1);
    pv_pkint(num, 1);

    l = 3;
    pv_pkint(&l, 1);
    l = np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(nseg, 1);
    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    pv_pkFLOAT(&reals[i], np);
    size -= np;
    i += np;
  }
}


/* pack state-vector extract */

void
pV_SENDSV (int *id, int *num, int *nd, int *len, int *index, FLOAT *weight,
           FLOAT *reals, int *kcellt, int *jsurf, int *istream)
{
  int size, i, k, np, type, off, rank, length, i19;
  int kcr, j, l, m, kn[8], kc;

  length = *len;
  if (*id == 19) {
    length = 0;
    for (k = 0; k < *len; k++) if (index[10*k+7] == pv3_stid) length++;
  }

  rank = *nd;
  size = length*rank;
  if (size <= 0) return;
  type = 12;

  i19 = 0;
  i   = 0;
  off = 0;

  while (size > 0) {
    np = MAXLEN;
    np = np/rank;
    np *= rank;
    if (np > size) np = size;
    k = np + 11;
    if (*id == 18) k++;
    pv_MesBuffer(1, (int) (k*sizeof(int)));
    pv_pkint(&pv3_stid, 1);
    pv_pkint(id, 1);
    pv_pkint(&pv3_did, 1);
    pv_pkint(&pv3_id, 1);
    pv_pkint(&type, 1);
    pv_pkint(&length, 1);
    pv_pkint(num, 1);

    l = 2;
    if (*id == 18) l++;
    pv_pkint(&l, 1);
    l = np;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    if (*id == 18) pv_pkint(&num[6], 1);
    pv_pkint(&i, 1);
    pv_pkint(&np, 1);
    if (*id == 18) {
      /* StreamLines */
      for (k = 0; k < np/rank; k++) {
        kcr = index[k+off];
        if ((*jsurf != 0) && (istream[(k+off)*2+1] == 0)) kcr = -kcr;
        pV_CELL2NODES (&kcr, &istream[(k+off)*2+1], &j, kn);
        for (m = 0; m < j; m++)
          if (kn[m] > 0) {
            pVSTATE (nd, &kn[m], &reals[*nd*(m+1)]);
          } else {
            l  = -kn[m];
            kc = kcr - *kcellt;
            pVPSTATE (nd, &kc, &l, &reals[*nd*(m+1)]);
          }
        for (l = 0; l < *nd; l++) {
          reals[l] = 0.0;
          for (m = 0; m < j; m++)
            reals[l] += weight[(k+off)*8+m]*reals[*nd*(m+1)+l];
        }
        pv_pkFLOAT(reals, *nd);
      }
    } else if (*id == 19) {
      /* Bubbles */
      for (k = 0; k < np/rank; k++) {
        while (index[10*i19+7] != pv3_stid) i19++;
        kcr = index[10*i19];
        if ((index[10*i19+2] <= 0) && (index[10*i19+9] == 0)) kcr = -kcr;
        pV_CELL2NODES (&kcr, &index[10*i19+9], &j, kn);
        for (m = 0; m < j; m++)
          if (kn[m] > 0) {
            pVSTATE (nd, &kn[m], &reals[*nd*(m+1)]);
          } else {
            l  = -kn[m];
            kc = kcr - *kcellt;
            pVPSTATE (nd, &kc, &l, &reals[*nd*(m+1)]);
          }
        for (l = 0; l < *nd; l++) {
          reals[l] = 0.0;
          for (m = 0; m < j; m++)
            reals[l] += weight[(k+off)*8+m]*reals[*nd*(m+1)+l];
        }
        pv_pkFLOAT(reals, *nd);
        i19++;
      }
    } else if (*id == 21) {
      /* Points */
      for (k = 0; k < np/rank; k++) {
        kcr = index[4*i19];
        pV_CELL2NODES (&kcr, &index[4*i19+2], &j, kn);
        for (m = 0; m < j; m++)
          if (kn[m] > 0) {
            pVSTATE (nd, &kn[m], &reals[*nd*(m+1)]);
          } else {
            l  = -kn[m];
            kc = kcr - *kcellt;
            pVPSTATE (nd, &kc, &l, &reals[*nd*(m+1)]);
          }
        for (l = 0; l < *nd; l++) {
          reals[l] = 0.0;
          for (m = 0; m < j; m++)
            reals[l] += weight[11*i19+m]*reals[*nd*(m+1)+l];
        }
        pv_pkFLOAT(reals, *nd);
        i19++;
      }
    } else {
      /* surfaces */
      j = 2;
      if (*jsurf != 0) j = 1;
      for (k = 0; k < np/rank; k++) {
        if (index[3*(k+off)+2] == 0) {
          pVSTATE (nd, &index[3*(k+off)], reals);
          if (j != 1) {
            pVSTATE (nd, &index[3*(k+off)+1], &reals[*nd]);
            for (l = 0; l < *nd; l++)
              reals[l] += weight[k+off]*(reals[*nd+l]-reals[l]);
          }
        } else {
          m = index[3*(k+off)];
          if (m > 0) {
            pVSTATE (nd, &m, reals);
          } else {
            m = -m;
            pVPSTATE (nd, &index[3*(k+off)+2], &m, reals);
          }
          if (j == 2) {
            m = index[3*(k+off)+1];
            if (m > 0) {
              pVSTATE (nd, &m, &reals[*nd]);
            } else {
              m = -m;
              pVPSTATE (nd, &index[3*(k+off)+2], &m, &reals[*nd]);
            }
            for (l = 0; l < *nd; l++)
              reals[l] += weight[k+off]*(reals[*nd+l]-reals[l]);
          }
        }
        pv_pkFLOAT(reals, *nd);
      }
    }
    size -= np;
    i    += np;
    off  += np/rank;
  }
}


/* pack transfer sl/part */

void
pV_SENDXFER (int *id, int *ints, FLOAT *reals)
{
  int l;

  l = 15;
  if (*id == 16) l = 48;
  if (*id == 17) l = 37;
  pv_MesBuffer(2, (int) (l*sizeof(int)));
  pv_pkint(&pv3_stid, 1);
  pv_pkint(id, 1);
  pv_pkint(&pv3_did, 1);
  pv_pkint(&pv3_id, 1);
  pv_pkint(ints, 3);

  switch (*id) {
  case 16:
    l = 9;
    pv_pkint(&l, 1);
    l = 30;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&ints[3], 9);
    pv_pkFLOAT(reals, 30);
    break;
  case 17:
    l = 4;
    pv_pkint(&l, 1);
    l = 24;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&ints[3], 4);
    pv_pkFLOAT(reals, 24);
    break;
  case 45:
    l = 3;
    pv_pkint(&l, 1);
    l = 3;
    pv_pkint(&l, 1);
    l = 0;
    pv_pkint(&l, 1);

    pv_pkint(&ints[3], 3);
    pv_pkFLOAT(reals, 3);
    break;
  }
}


/* pack time type message */

void
pV_SENDTIME (int *istat, FLOAT *time, FLOAT *fmm, int *nk,
             int *sizes, int *blocks, char *str, int strLEN)
{
  int  i, j, l, n;

  n = 3*sizes[7];
  l = (20 + 2*(*nk+1) + n)*sizeof(int) + 80;
  pv_MesBuffer(1, l);
  pv_pkint(&pv3_stid, 1);
  j = 33;
  pv_pkint(&j, 1);
  pv_pkint(&pv3_did, 1);
  pv_pkint(&pv3_id, 1);
  pv_pkint(istat, 1);
  j = 0;
  pv_pkint(&j, 1);
  pv_pkint(&j, 1);
  j = *nk;

  l = 10 + n;
  pv_pkint(&l, 1);
  l = 1 + 2*(j+1);
  pv_pkint(&l, 1);
  l = 80;
  pv_pkint(&l, 1);

  pv_pkint(nk, 1);
  pv_pkint(sizes, 9);
  for (i = 0; i < sizes[7]; i++) pv_pkint(&blocks[6*i], 3);
  pv_pkFLOAT(time, 1);
  pv_pkFLOAT(fmm, 2*(j+1));
  pv_pkbyte(str, 80);
}


/* pack string message */

void
pV_SENDSTRING (char *str, int strLEN)
{
  int  i, j, l;

  for(i = 0; i < pv3_nser; i++)
    if ((pv3_ser[i].status > 1) && (pv3_ser[i].status < 4)) {

      l = 9*sizeof(int) + 80;
      pv_MesBuffer(1, l);
      pv_pkint(&pv3_ser[i].tid, 1);
      j = 31;
      pv_pkint(&j, 1);
      pv_pkint(&pv3_did, 1);
      pv_pkint(&pv3_id, 1);
      j = 0;
      pv_pkint(&j, 1);
      pv_pkint(&j, 1);
      pv_pkint(&j, 1);

      l = 0;
      pv_pkint(&l, 1);
      l = 0;
      pv_pkint(&l, 1);
      l = 80;
      pv_pkint(&l, 1);

      pv_pkbyte(str, 80);
    }

}


/* send an acknowledgement that a transfered particle/sl has
   been dealt with by this client */

void
pV_SENDXFERACK (int *mid, int *ID, int *IDO)
{
  int  j, l;

  j = 2;
  if (*mid == 44) j = 1;
  pv_MesBuffer(j, (int) (9*sizeof(int)));

  pv_pkint(&pv3_stid, 1);
  pv_pkint(mid,  1);
  pv_pkint(&pv3_did, 1);
  pv_pkint(&pv3_id, 1);
  pv_pkint(ID,  1);
  pv_pkint(IDO, 1);
  j = 0;
  pv_pkint(&j, 1);

  l = 0;
  pv_pkint(&l, 1);
  l = 0;
  pv_pkint(&l, 1);
  l = 0;
  pv_pkint(&l, 1);

}


/* check on server down messages */

static void
pV_CheckServer()
{
  int        i, j, tid, tot;
  MPI_Status status;

  i = 1;
  while (i != 0) {
    MPI_Iprobe(pv3_conc, PVDOWN, pv3_hicomm, &i, &status);
    if (i != 0) {
      MPI_Recv(&tid, 1, MPI_INT, pv3_conc, PVDOWN, pv3_hicomm, &status);
      if (tid == 0) {
        pv3_ser[0].ltag = -1;
      } else {
        for (tot = j = 0; j < pv3_nser; j++) 
          if (pv3_ser[j].tid == tid) {
            pv3_ser[j].ltag = 1;
            tot++;
          }
        if (tot == 0) 
          printf("pV3 Error: 0. Server not found: %X!\n", tid);
      }
    }
  }
}


/* return server data to client */

void
pV_GETSERDATA (int *handle, int *mid, int *nint, int **ints, int *nfloat,
               FLOAT **floats, int *nchar, char **chars)
{
  ServerData *sd, *lsd;

  sd  = pv3_sd;
  lsd = NULL;

  while (sd != NULL) {
    if (sd->handle == *handle) {
      *mid    = sd->mid;
      *nint   = sd->nint;
      *ints   = sd->ints;
      *nfloat = sd->nfloat;
      *floats = sd->floats;
      *nchar  = sd->nchar;
      *chars  = sd->chars;
      if (lsd == NULL) {
        pv3_sd = sd->next;
      } else {
        lsd->next = sd->next;
      }
      free(sd);
      return;
    }
    lsd = sd;
    sd  = sd->next;
  }
  printf(" pV3 Warning: Handle not Found! %d  (pV_GetServerData)\n", *handle);
}


/* get server to client data stream -- mid = 50 only */

static int
pv_readData(int mid)
{
  int        i, nint, nfloat, nchar;
  int        idum;
  FLOAT      fdum;
  char       cdum;
  ServerData *sd, *lsd;

  pv_upkint(&nint, 1);
  pv_upkint(&nfloat, 1);
  pv_upkint(&nchar, 1);

  lsd = NULL;
  sd  = pv3_sd;
  while (sd != NULL) {
    lsd = sd;
    sd  = sd->next;
  }
  sd = (ServerData *) malloc(sizeof(ServerData));
  if (sd == NULL) {
    for (i = 0; i < nint;   i++) pv_upkint(&idum, 1);
    for (i = 0; i < nfloat; i++) pv_upkFLOAT(&fdum, 1);
    for (i = 0; i < nchar;  i++) pv_upkbyte(&cdum, 1);
    return -1;
  }
  sd->handle = pv3_sdh;
  sd->mid    = mid;
  sd->nint   = nint;
  sd->ints   = NULL;
  sd->nfloat = nfloat;
  sd->floats = NULL;
  sd->nchar  = nchar;
  sd->chars  = NULL;
  sd->next   = NULL;
  if (lsd == NULL) {
    pv3_sd = sd;
  } else {
    lsd->next = sd;
  }

  if (nint > 0) {
    sd->ints = (int *) malloc(nint*sizeof(int));
    if (sd->ints == NULL) {
      sd->nint = 0;
      for (i = 0; i < nint; i++) pv_upkint(&idum, 1);
    } else {
      pv_upkint(sd->ints, nint);
    }
  }
  if (nfloat > 0) {
    sd->floats = (FLOAT *) malloc(nfloat*sizeof(FLOAT));
    if (sd->floats == NULL) {
      sd->nfloat = 0;
      for (i = 0; i < nfloat; i++) pv_upkFLOAT(&fdum, 1);
    } else {
      pv_upkFLOAT(sd->floats, nfloat);
    }
  }
  if (nchar > 0) {
    sd->chars = (char *) malloc(nchar*sizeof(char));
    if (sd->chars == NULL) {
      sd->nchar = 0;
      for (i = 0; i < nchar; i++) pv_upkbyte(&cdum, 1);
    } else {
      pv_upkbyte(sd->chars, nchar);
    }
  }
  pv3_sdh++;
  return sd->handle;
}


/* get extract request */

void
pV_GETREQUEST (int *mid, int *is, int *iv, int *it, int *mask, int *ints,
               FLOAT *reals, char *cstr, int cstrLEN)
{
  int        did, cl, i, j, k, tid;
  int        nint, nfloat, nchar;
  MPI_Status status;

  *mid = -1;

  /* in high prio read buffer */

  if (pv3_srbuf == 2) {
    *mid = -1;
    tid  = -1;
    pv_upkint(&tid, 1);
    if (tid != -1) {
      for (k = j = 0; j < pv3_nser; j++)
        if (tid == pv3_ser[j].tid) {
          pv3_stid  = tid;
          pv3_sindx = j;
          k++;
        }
      if (k == 0) {
        printf(" pV3 Warning: 1. Server not found: %X!\n", tid);
      } else {
        pv_upkint(mid, 1);
      }
    }
    if (*mid == -1) {
      pv3_hrpos = 0;
      pv3_srbuf = 1;
      if (pv3_lrpos == 0) pv3_srbuf = 0;
    }
  }

  /* in lo prio read buffer */

  if (pv3_srbuf == 1) {
    MPI_Iprobe(pv3_conc, HIPRIO, pv3_hicomm, &i, &status);
    if (i != 0) {
      MPI_Recv(pv3_hrbuf, pv3_rsize, MPI_PACKED, pv3_conc, HIPRIO,
               pv3_hicomm, &status);
      pv3_hrpos = 0;
      pv3_srbuf = 2;
      pv_upkint(&tid, 1);
      for (k = j = 0; j < pv3_nser; j++)
        if (tid == pv3_ser[j].tid) {
          pv3_stid  = tid;
          pv3_sindx = j;
          k++;
        }
      if (k == 0) {
        printf(" pV3 Warning: 2. Server not found: %X!\n", tid);
        pv3_srbuf = 1;
        pv_upkint(mid, 1);
        if (*mid == -1) pv3_srbuf = pv3_lrpos = 0;
      } else {
        pv_upkint(mid, 1);
      }
    } else {
      *mid = -1;
      tid  = -1;
      pv_upkint(&tid, 1);
      if (tid != -1) {
        for (k = j = 0; j < pv3_nser; j++)
          if (tid == pv3_ser[j].tid) {
            pv3_stid  = tid;
            pv3_sindx = j;
            k++;
          }
        if (k == 0) {
          printf(" pV3 Warning: 3. Server not found: %X!\n", tid);
        } else {
          pv_upkint(mid, 1);
        }
      }
      if (*mid == -1) pv3_srbuf = pv3_lrpos = 0;
    }
  }

  /* no current read buffer */

  if (pv3_srbuf == 0) {

    pV_CheckServer();

    /* tell client routines about any concentrator time-outs */
    if (pv3_ser[0].ltag == -1) {
      pv3_ser[0].ltag = 0;
      *mid = 36;
      *is  = -1;
      *iv  =  0;
      *it  =  0;
      return;
    }

    /* tell client routines about any terminated server */
    for (j = 0; j < pv3_nser; j++)
      if (pv3_ser[j].ltag != 0) {
        pv3_ser[j].ltag = 0;
        *mid = 36;
        *is  = -1;
        *iv  =  j+1;
        *it  = pv3_ser[j].tid;
        return;
      }

    /* any hi prio messages? */
    MPI_Iprobe(pv3_conc, HIPRIO, pv3_hicomm, &i, &status);
    if (i != 0) {
      pv3_srbuf = 2;
      MPI_Recv(pv3_hrbuf, pv3_rsize, MPI_PACKED, pv3_conc, HIPRIO,
               pv3_hicomm, &status);
      pv3_hrpos = 0;
      pv3_srbuf = 2;
      pv_upkint(&tid, 1);
      for (k = j = 0; j < pv3_nser; j++)
        if (tid == pv3_ser[j].tid) {
          pv3_stid  = tid;
          pv3_sindx = j;
          k++;
        }
      if (k == 0) {
        printf(" pV3 Warning: 4. Server not found: %X!\n", tid);
        pv3_srbuf = 0;
      } else {
        pv_upkint(mid,  1);
      }
    }
    if (pv3_srbuf == 0) {
      /* any lo prio messages? */
      MPI_Iprobe(pv3_conc, MPI_ANY_TAG, pv3_locomm, &i, &status);
      if (i != 0) {
        MPI_Recv(pv3_lrbuf, pv3_rsize, MPI_PACKED, pv3_conc, MPI_ANY_TAG,
                 pv3_locomm, &status);
        pv3_lrpos = 0;
        pv3_srbuf = 1;
        pv_upkint(&tid, 1);
        for (k = j = 0; j < pv3_nser; j++)
          if (tid == pv3_ser[j].tid) {
            pv3_stid  = tid;
            pv3_sindx = j;
            k++;
          }
        if (k == 0) {
          printf(" pV3 Warning: 5. Server not found: %X!\n", tid);
          pv3_srbuf = 0;
        } else {
          pv_upkint(mid,  1);
        }
      }
    }
  }

  if (pv3_srbuf == 0) {
    pv_MesBuffer(0, 0);
    return;
  }

  pv_upkint(&did, 1);
  pv_upkint(&cl, 1);
  pv_upkint(is, 1);
  pv_upkint(iv, 1);
  pv_upkint(it, 1);
  if ((*mid == 35) && (pv3_did == -1)) pv3_did = *iv;
  if ((pv3_did != did) || (pv3_id != cl)) {
    printf(" Warning: DID/CID MisMatch-> DID = %d (%d),  CID = %d (%d)\n",
           pv3_did, did, pv3_id, cl);
    i = pv3_lrpos;
    if (pv3_srbuf == 2) i = pv3_hrpos;
    printf("          MID = %d, MPI buffer = %d, %d\n", *mid, pv3_srbuf, i);
    if (pv3_srbuf == 2) {
      pv3_hrpos = 0;
      pv3_srbuf = 1;
      if (pv3_lrpos == 0) pv3_srbuf = 0;
    } else {
      pv3_srbuf = 0;
    }
    *mid = -1;
    return;
  }

  nint   = 0;
  nfloat = 0;
  nchar  = 0;
  pv_upkint(mask, 1);
  if (*mid != 50) {
    pv_upkint(&nint, 1);
    pv_upkint(&nfloat, 1);
    pv_upkint(&nchar, 1);
    if (nint   != 0) pv_upkint(ints, nint);
    if (nfloat != 0) pv_upkFLOAT(reals, nfloat);
    if (nchar  != 0) pv_upkbyte(cstr, nchar);
  } else {
    *is = pv_readData(*mask);
  }
  if (*mid == 35) ints[1] = 0;

}


/* blocked get extract request (with timeout) */

void
pV_GETTREQUEST (int *mid, int *is, int *iv, int *it, int *mask, int *to,
                int *ints, FLOAT *reals, char *cstr, int cstrLEN)
{
  printf(" pV3 ERROR: pV_GetTRequest should not be called!\n");
  exit(1);
}
