#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>

/*
 * FORTRAN/C routines for Visual3 post-processing
 *
 * Copyright 1990 - 1998, Massachusetts Institute of Technology.
 *
 */


#ifdef UNDERSCORE

#define MALLOCV3 mallocv3_
#define FREEV3 freev3_
#define PPDATE ppdate_
#define FIXTXT fixtxt_
#define IMGREAD imgread_
#define IOOPEN ioopen_
#define IOWRIT iowrit_
#define IOCLOS ioclos_

#else

#define MALLOCV3 mallocv3
#define FREEV3 freev3
#define PPDATE ppdate
#define FIXTXT fixtxt
#define IMGREAD imgread
#define IOOPEN ioopen
#define IOWRIT iowrit
#define IOCLOS ioclos

#endif


#ifdef POINTERLONG

#define PTR long

#else

#define PTR int

#endif


#define False 0
#define True  1


PTR
MALLOCV3(nbytes)
  int *nbytes;

{
  PTR addr;

  addr = (PTR) malloc(*nbytes);
  return addr;
}



void
FREEV3(addr)
  char **addr;

{
  free(*addr);
}


void
revbytes(bytes)
    char *bytes;
{
    char sav;

    sav      = bytes[0];
    bytes[0] = bytes[3];
    bytes[3] = sav;
    sav      = bytes[1];
    bytes[1] = bytes[2];
    bytes[2] = sav;
}


void
revbyte2(bytes)
    char *bytes;
{
    char sav;

    sav      = bytes[0];
    bytes[0] = bytes[1];
    bytes[1] = sav;
}


void
PPDATE(string)
   char    *string;
{
   char    *ctime();
   struct  timeval  tp;
   struct  timezone tzp;
   long    seconds;

   gettimeofday(&tp, &tzp);
   seconds = tp.tv_sec;

   string[ 0] = ctime(&seconds)[ 8];
   string[ 1] = ctime(&seconds)[ 9];
   string[ 2] = ctime(&seconds)[ 3];
   string[ 3] = ctime(&seconds)[ 4];
   string[ 4] = ctime(&seconds)[ 5];
   string[ 5] = ctime(&seconds)[ 6];
   string[ 6] = ctime(&seconds)[ 7];
   string[ 7] = ctime(&seconds)[22];
   string[ 8] = ctime(&seconds)[23];
   string[ 9] = ctime(&seconds)[10];
   string[10] = ctime(&seconds)[11];
   string[11] = ctime(&seconds)[12];
   string[12] = ctime(&seconds)[13];
   string[13] = ctime(&seconds)[14];
   string[14] = ctime(&seconds)[15];
   string[15] = ctime(&seconds)[16];
   string[16] = ctime(&seconds)[17];
   string[17] = ctime(&seconds)[18];
}


void
FIXTXT(text, textLEN)
  char *text;
  int  textLEN;
{
  int i;

  for (i = 0; i < textLEN; i++) {
    if (text[i] == '(') text[i] = '{';
    if (text[i] == ')') text[i] = '}';
  }
}


int
IMGREAD(rev, filename, text, factor, cnpx, cnpy, cadr, fnLEN)
  char rev[4], *filename, text[][80];
  int  *factor, *cnpx, *cnpy, **cadr, fnLEN;
{
  FILE  *fp;
  char  pix1, *name, stext[17], *bytes;
  short pix2;
  int   *pix4, nbytes, dum, iend, byterev, cbgr, bpp, cbo;
  int   col, r, g, b, ctcolor, cdepth, *clut, mask, mbo;
  int   i, j, npixx, npixy;

  mbo = 1;
  if (rev[0] != 0) mbo = 0;

  name  = (char *) malloc((fnLEN+1)*sizeof(char));
  for (i = 0; i < fnLEN; i++) if (filename[i] != ' ') break;
  for (iend = fnLEN; iend > 1; iend--) if (filename[iend-1] != ' ') break;
  for (dum = i; dum < iend; dum++) name[dum-i] = filename[dum];
  name[iend-i] = '\0';

  fp = fopen(name,"r");
  if (fp == NULL) {
    fprintf(stderr, " Error opening file: %s\n\n", name);
    free(name);
    return 1;
  }

  fread(stext, sizeof(char), 16, fp);
  stext[16] = 0;
  if (strcmp("Visual ImageFile",stext) != 0) {
    fprintf(stderr," Error Not a Rev. 2 Visual ImageFile: %s\n\n", name);
    free(name);
    fclose(fp);
    return 2;
  }
  fread(&dum, sizeof(int), 1, fp);
  if ((dum != 2) && (dum != 33554432)) {
    fprintf(stderr," Error Not a Rev. 2 Visual ImageFile: %s\n\n", name);
    free(name);
    fclose(fp);
    return 3;
  }
  byterev = False;
  if (dum == 33554432) byterev = True;

  fread(&ctcolor,  sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &ctcolor);
  fread(&cbgr, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &cbgr);

  fread(&npixx, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &npixx);
  fread(&npixy, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &npixy);
  fread(&cdepth, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &cdepth);
  fread(&bpp, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &bpp);
  if ((bpp != 32) && (bpp != 24) && (bpp != 16) && (bpp != 8)) {
    fprintf(stderr," Error Bits per Pixel Illegal (%d)!\n\n", bpp);
    free(name);
    fclose(fp);
    return 4;
  }
  fread(&cbo, sizeof(int), 1, fp);
  if (byterev) revbytes((char *) &cbo);

  dum   = *factor;
  *cnpx = npixx/dum;
  if ((npixx % dum) != 0) *cnpx += 1;
  *cnpy = npixy/dum;
  if ((npixy % dum) != 0) *cnpy += 1;
  dum   = *cnpx;
  dum  *= *cnpy;
  *cadr = (int *) malloc(dum*sizeof(int));
  if (*cadr == NULL) {
    fprintf(stderr," Error in Image Allocation! (%d x %d)\n\n", *cnpx, *cnpy);
    free(name);
    fclose(fp);
    return 5;
  }

  nbytes = 80;
  fread(text[0], sizeof(char), nbytes, fp);
  fread(text[1], sizeof(char), nbytes, fp);
  fread(text[2], sizeof(char), nbytes, fp);
  fread(text[3], sizeof(char), nbytes, fp);

  dum = 1 << cdepth;
  mask = dum-1;

  /* get color look-up table (PseudoColor) */

  if (!ctcolor) {
    clut = (int *) malloc(dum*sizeof(int));
    fread(clut, sizeof(int), dum, fp);
    if (byterev)
      for (iend = 0; iend < dum; iend++) revbytes((char *) &clut[iend]);
  }

  /* get image based on bits/pixel of file */

  pix4 = *cadr;
  if (bpp == 32) {
    for (j = 0; j < npixy; j++)
    for (i = 0; i < npixx; i++) {
      fread(&col, 1, sizeof(int), fp);
      if ((j % *factor) != 0) continue;
      if ((i % *factor) != 0) continue;
      *pix4 = col;
      if (ctcolor) {
        if (cbo != mbo) revbytes((char *) pix4);
        *pix4 &= mask;
        if (cdepth == 30) {
          col = *pix4;
          r = (col >> 22) & 255;
          g = (col >> 12) & 255;
          b = (col >>  2) & 255;
          *pix4 = r << 16 | g << 8 | b;
        } else if (cdepth == 15) {
          col = *pix4;
          r = ((col >> 10) & 31) << 3;
          if (r != 0) r |= 7;
          g = ((col >>  5) & 31) << 3;
          if (g != 0) g |= 7;
          b = ( col        & 31) << 3;
          if (b != 0) b |= 7;
          *pix4 = r << 16 | g << 8 | b;
        } else if (cdepth == 12) {
          col = *pix4;
          r = ((col >> 8) & 15) << 4;
          if (r != 0)  r |= 15;
          g = ((col >> 4) & 15) << 4;
          if (g != 0)  g |= 15;
          b = ( col       & 15) << 4;
          if (b != 0)  b |= 15;
          *pix4 = r << 16 | g << 8 | b;
        }
        if (cbgr) {
          col = *pix4;
          r = (col >> 16) & 255;
          g = (col >>  8) & 255;
          b =  col        & 255;
          *pix4 = b << 16 | g << 8 | r;
        }
      } else {
        if (byterev) revbytes((char *) pix4);
        *pix4 = clut[*pix4 & mask];
      }
      pix4++;
    }

  } else if (bpp == 24) {
    bytes = (char *) &col;
    for (j = 0; j < npixy; j++)
    for (i = 0; i < npixx; i++) {
      col = 0;
      fread(&bytes[mbo], 1, 3, fp);
      if ((j % *factor) != 0) continue;
      if ((i % *factor) != 0) continue;
      *pix4 = col;
      if (ctcolor) {
        if (cbo != mbo) {
          pix1         = bytes[mbo];
          bytes[mbo]   = bytes[mbo+2];
          bytes[mbo+2] = pix1;
          *pix4 = col;
        }
        *pix4 &= mask;
        if (cdepth == 15) {
          col = *pix4;
          r = ((col >> 10) & 31) << 3;
          if (r != 0) r |= 7;
          g = ((col >>  5) & 31) << 3;
          if (g != 0) g |= 7;
          b = ( col        & 31) << 3;
          if (b != 0) b |= 7;
          *pix4 = r << 16 | g << 8 | b;
        } else if (cdepth == 12) {
          col = *pix4;
          r = ((col >> 8) & 15) << 4;
          if (r != 0)  r |= 15;
          g = ((col >> 4) & 15) << 4;
          if (g != 0)  g |= 15;
          b = ( col       & 15) << 4;
          if (b != 0)  b |= 15;
          *pix4 = r << 16 | g << 8 | b;
        }
        if (cbgr) {
          col = *pix4;
          r = (col >> 16) & 255;
          g = (col >>  8) & 255;
          b =  col        & 255;
          *pix4 = b << 16 | g << 8 | r;
        }
      } else {
        if (byterev) revbytes((char *) pix4);
        *pix4 = clut[*pix4 & mask];
      }
      pix4++;
    }

  } else if (bpp == 16) {
    for (j = 0; j < npixy; j++)
    for (i = 0; i < npixx; i++) {
      fread(&pix2, 1, sizeof(short), fp);
      if ((j % *factor) != 0) continue;
      if ((i % *factor) != 0) continue;
      if (ctcolor) {
        if (cbo != mbo) revbyte2((char *) &pix2);
        col = pix2;
        if (cdepth == 15) {
          r = ((col >> 10) & 31) << 3;
          if (r != 0) r |= 7;
          g = ((col >>  5) & 31) << 3;
          if (g != 0) g |= 7;
          b = ( col        & 31) << 3;
          if (b != 0) b |= 7;
        } else {
          r = ((col >> 8) & 15) << 4;
          if (r != 0)  r |= 15;
          g = ((col >> 4) & 15) << 4;
          if (g != 0)  g |= 15;
          b = ( col       & 15) << 4;
          if (b != 0)  b |= 15;
        }
        if (cbgr) {
          *pix4 = b << 16 | g << 8 | r;
        } else {
          *pix4 = r << 16 | g << 8 | b;
        }
      } else {
        if (byterev) revbyte2((char *) &pix2);
        *pix4 = clut[pix2 & mask];
      }
      pix4++;
    }

  } else {
    /* 8 bit -- PseudoColor File */
    for (j = 0; j < npixy; j++)
    for (i = 0; i < npixx; i++) {
      fread(&pix1, 1, sizeof(char), fp);
      if ((j % *factor) != 0) continue;
      if ((i % *factor) != 0) continue;
      *pix4 = clut[pix1 & mask];
      pix4++;
    }
  }

  if (!ctcolor) free(clut);
  fclose(fp);

  free(name);
  return 0;
}


void
IOOPEN(fname, mode, fp, fnLEN) 
  char *fname;
  int  *mode, fnLEN;
  FILE **fp;
{
  int  i, iend, dum;
  char *cname;

  cname  = (char *) malloc((fnLEN+1)*sizeof(char));
  for (i = 0; i < fnLEN; i++) if (fname[i] != ' ') break;
  for (iend = fnLEN; iend > 1; iend--) if (fname[iend-1] != ' ') break;
  for (dum = i; dum < iend; dum++) cname[dum-i] = fname[dum];
  cname[iend-i] = '\0';
  switch(*mode) {
  case 0:
    *fp = fopen(cname,"rb");
    break;
  case 1:
    *fp = fopen(cname,"wb");
    break;
  }
  free(cname);
}


void
IOCLOS(fp)
  FILE **fp;
{
  fclose(*fp);
}


void
IOWRIT(buf, size, fp)
  char *buf;
  int  *size;
  FILE **fp;
{
  fwrite(buf,*size,1,*fp);
}
