/*
 * xsubs.c
 *
 * Xlib/FORTRAN interface for img2X.
 *
 * Copyright 1990 - 1996, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/cursorfont.h>

#ifdef UNDERSCORE
#define XFTNINITS xftninits_
#define XFTNCOLORWINDOWS xftncolorwindows_
#define XFTNGCS xftngcs_
#define XFTNNUMEVENTS xftnnumevents_
#define XFTNEVENT xftnevent_
#define XFTNLOWER xftnlower_
#define XFTNSYNCHRONIZE xftnsynchronize_
#define XFTNFLUSH xftnflush_
#define XFTNDESTROYWINDOW xftndestroywindow_
#define XFTNCLOSEDISPLAY xftnclosedisplay_
#define XFTNIMGFILL xftnimgfill_
#define XFTNIMGPUT xftnimgput_
#define XFTNIMGFREE xftnimgfree_
#define XFTNDUMP xftndump_
#else
#define XFTNINITS xftninits
#define XFTNCOLORWINDOWS xftncolorwindows
#define XFTNGCS xftngcs
#define XFTNNUMEVENTS xftnnumevents
#define XFTNEVENT xftnevent
#define XFTNLOWER xftnlower
#define XFTNSYNCHRONIZE xftnsynchronize
#define XFTNFLUSH xftnflush
#define XFTNDESTROYWINDOW xftndestroywindow
#define XFTNCLOSEDISPLAY xftnclosedisplay
#define XFTNIMGFILL xftnimgfill
#define XFTNIMGPUT xftnimgput
#define XFTNIMGFREE xftnimgfree
#define XFTNDUMP xftndump
#endif

#ifdef POINTERLONG
#define PTR long
#else
#define PTR int
#endif

#define argc 0
#define argv (char **) NULL
#define MATTEWIDTH 2

Cursor		cursor;
XVisualInfo	vinfo;
Visual		*vsl;
Display		*dpy;
Colormap	colormap;
int		vdepth, white, black, bgr;
XColor		wforeground, wbackground, wborder;



Display *
XFTNINITS (xsize, ysize, bsize)
    int *xsize, *ysize, *bsize;
{
    XWindowAttributes xwa;

    if ((dpy = XOpenDisplay(NULL)) == NULL) {
        fprintf(stderr, "Oops! Cannot open display %s.\n",
                XDisplayName(NULL));
        exit(1);
    }
    if (XMatchVisualInfo(dpy, DefaultScreen(dpy), 24, TrueColor,
                         &vinfo) == 0) 
      if (XMatchVisualInfo(dpy, DefaultScreen(dpy), 15, TrueColor,
                           &vinfo) == 0)
        if (XMatchVisualInfo(dpy, DefaultScreen(dpy), 12, TrueColor,
                             &vinfo) == 0) {
          fprintf(stderr, "Oops! TrueColor Visual Not Found: %s!\n",
                  XDisplayName(NULL));
          exit(1);
        }

    vsl    = vinfo.visual;
    vdepth = vinfo.depth;
    bgr    = False;
    if (vinfo.red_mask < vinfo.blue_mask) bgr = True;
    if (vdepth == 24) {
      if (vinfo.green_mask != 0xff00) {
        fprintf(stderr, "Oops! Improper X-visual mask for Display %s.\n",
                XDisplayName(NULL));
        exit(1);
      }
    } else if (vdepth == 15) {
      if (vinfo.green_mask != 0x3e0) {
        fprintf(stderr, "Oops! Improper X-visual mask for Display %s.\n",
                XDisplayName(NULL));
        exit(1);
      }
    } else {
      if (vinfo.green_mask != 0xf0) {
        fprintf(stderr, "Oops! Improper X-visual mask for Display %s.\n",
                XDisplayName(NULL));
        exit(1);
      }
    }
    XGetWindowAttributes(dpy, DefaultRootWindow(dpy), &xwa);
    *xsize = xwa.width;
    *ysize = xwa.height;
    *bsize = MATTEWIDTH;

    cursor = XCreateFontCursor(dpy, XC_draft_small);

    colormap = XCreateColormap(dpy, DefaultRootWindow(dpy),
                               vsl, AllocNone);

    wforeground.red   = (1 << 16) - 1;
    wforeground.green = (1 << 16) - 1;
    wforeground.blue  = (1 << 16) - 1;
    wforeground.flags = DoRed | DoGreen | DoBlue;
    XAllocColor(dpy, colormap, &wforeground);
    white = wforeground.pixel;

    wbackground.red   = 0;
    wbackground.green = 0;
    wbackground.blue  = 0;
    wbackground.flags = DoRed | DoGreen | DoBlue;
    XAllocColor(dpy, colormap, &wbackground);
    black = wbackground.pixel;

    wborder.red   = (1 << 15) - 1;
    wborder.green = (1 << 15) - 1;
    wborder.blue  = (1 << 15) - 1;
    wborder.flags = DoRed | DoGreen | DoBlue;
    XAllocColor(dpy, colormap, &wborder);

    XRecolorCursor(dpy, cursor, &wforeground, &wbackground);

    return dpy;
}


Window
XFTNCOLORWINDOWS (display, STRING, Class, xsize, ysize, xstart, ystart, 
                           stLEN,  clLEN)
    Display **display;
    char *STRING, *Class;
    int *xsize, *ysize, *xstart, *ystart, stLEN, clLEN;
{
    Window window;
    XSizeHints xsh;
    XSetWindowAttributes xswa;
    int swa_mask;
    XClassHint chint;
    int iend, i;
    char *name;

    xswa.colormap = colormap;
    xswa.background_pixel = wbackground.pixel;
    xswa.border_pixel = wborder.pixel;
    xswa.event_mask = ExposureMask | ButtonPressMask | ButtonReleaseMask
                    | KeyPressMask | KeyReleaseMask;
    swa_mask = CWColormap | CWEventMask | CWBackPixel | CWBorderPixel;

    window = XCreateWindow(*display, DefaultRootWindow(*display),
                           *xstart, *ystart, *xsize, *ysize, MATTEWIDTH,
                           vdepth, InputOutput, vsl, swa_mask, &xswa);

    xsh.flags = USPosition | USSize;
    xsh.flags |= PMinSize | PMaxSize;
    xsh.width = *xsize;
    xsh.height = *ysize;
    xsh.x = *xstart;
    xsh.y = *ystart;
    xsh.min_width = xsh.width;
    xsh.max_width = xsh.width;
    xsh.min_height = xsh.height;
    xsh.max_height = xsh.height;
    XSetNormalHints(*display, window, &xsh);

    name  = (char *) malloc((stLEN+1)*sizeof(char));
    for (iend = stLEN; iend > 1; iend--) if (STRING[iend-1] != ' ') break;
    for (i = 0; i < iend; i++) name[i] = STRING[i];
    name[iend] = '\0';
    XSetStandardProperties(*display, window, name, name, None, argv, argc,
                           &xsh);
    free(name);

    name  = (char *) malloc((clLEN+1)*sizeof(char));
    for (iend = clLEN; iend > 1; iend--) if (Class[iend-1] != ' ') break;
    for (i = 0; i < iend; i++) name[i] = Class[i];
    name[iend] = '\0';
    chint.res_name = name;
    chint.res_class = name;
    XSetClassHint(*display, window, &chint);
    free(name);

    XDefineCursor(*display, window, cursor);

    XMapWindow(*display, window);

    return window;
}


GC 
XFTNGCS (display, window)
    Display **display;
    Window *window;
{
    XGCValues gcv;
    int gcv_mask;

    gcv.foreground = white;
    gcv.background = black;
    gcv_mask = GCForeground | GCBackground;
    return XCreateGC(*display, *window, gcv_mask, &gcv);
}


int 
XFTNNUMEVENTS (display)
    Display **display;
{
    return XEventsQueued(*display, QueuedAfterFlush);
}


Window 
XFTNEVENT (display, type, x, y, state)
    Display **display;
    int *type, *x, *y, *state;
{
    Window window;
    XEvent report;
    KeySym key;
    char buffer[2];

    XNextEvent(*display, &report);
    *type = report.type;
    switch(report.type) {
    case Expose:
      *state = report.xexpose.count;
      *x = 0;
      *y = 0;
      window = report.xexpose.window;
      break;
    case ButtonPress:
      *state = report.xbutton.button;
      *x = report.xbutton.x;
      *y = report.xbutton.y;
      window = report.xbutton.window;
      break;
    case ButtonRelease:
      *state = report.xbutton.button;
      *x = report.xbutton.x;
      *y = report.xbutton.y;
      window = report.xbutton.window;
      break;
    case KeyPress:
      *x = report.xkey.x;
      *y = report.xkey.y;
      window = report.xkey.window;
      XLookupString((XKeyEvent *) &report, buffer, 1, &key, NULL);
      *state = key;
      break;
    case KeyRelease:
      *x = report.xkey.x;
      *y = report.xkey.y;
      window = report.xkey.window;
      XLookupString((XKeyEvent *) &report, buffer, 1, &key, NULL);
      *state = key;
      break;
    }

    return window;
}


void
XFTNLOWER (display, window)
    Display **display;
    Window *window;
{
    XLowerWindow(*display, *window);
}


void
XFTNSYNCHRONIZE (display, onoff)
    Display **display;
    int     *onoff;
{
    XSynchronize(*display, *onoff);
}


void
XFTNFLUSH (display)
    Display **display;
{
    XFlush(*display);
}


void
XFTNDESTROYWINDOW (display, window)
    Display **display;
    Window *window;
{
    XDestroyWindow(*display, *window);
}


void
XFTNCLOSEDISPLAY (display)
    Display **display;
{
    XCloseDisplay(*display);
}


XImage *
XFTNIMGFILL (display, window, width, height, addr, rev)
    Display  **display;
    Drawable *window;
    unsigned int *width, *height;
    int **addr;
    char rev[4];
{
    Drawable win;
    XImage   *xim;
    char     *data, *pix1;
    short    *pix2;
    int      *pix4, *ptr, col, r, g, b, i, j, mbo;

    mbo = 1;
    if (rev[0] != 0) mbo = 0;

    win = *window;
    if (win == 0) win = DefaultRootWindow(*display);
    xim = XGetImage(*display, win, 0, 0, *width, *height, AllPlanes, ZPixmap);

    ptr  = *addr;
    data = xim->data;
    for (j = 0; j < *height; j++, data += xim->bytes_per_line) {
      pix1 =           data;
      pix2 = (short *) data;
      pix4 = (int *)   data;
      for (i = 0; i < *width; i++, ptr++) {
        if (xim->bits_per_pixel == 32) {
          *pix4 = *ptr;
          if (bgr) {
            col = *pix4;
            r = (col >> 16) & 255;
            g = (col >>  8) & 255;
            b =  col        & 255;
            *pix4 = b << 16 | g << 8 | r;
          }
          if (xim->byte_order != mbo) revbytes(pix4);
          pix4++;
        } else if (xim->bits_per_pixel == 24) {
	  col = *ptr;
          if (bgr) {
            *pix1 = (col >> 16) & 255;
            pix1++;
            *pix1 = (col >>  8) & 255;
            pix1++;
            *pix1 =  col        & 255;
            pix1++;
          } else {
            *pix1 =  col        & 255;
            pix1++;
            *pix1 = (col >>  8) & 255;
            pix1++;
            *pix1 = (col >> 16) & 255;
            pix1++;
          }
        } else if (vdepth == 15) {
          col = *ptr;
          r = (col >> 19) & 31;
          g = (col >> 11) & 31;
          b = (col >>  3) & 31;
          if (bgr) {
            *pix2 = b << 10 | g << 5 | r;
          } else {
            *pix2 = r << 10 | g << 5 | b;
          }
          if (xim->byte_order != mbo) revbyte2(pix2);
          pix2++;
        } else {
          col = *ptr;
          r = (col >> 20) & 15;
          g = (col >> 12) & 15;
          b = (col >>  4) & 15;
          if (bgr) {
            *pix2 = b << 8 | g << 4 | r;
          } else {
            *pix2 = r << 8 | g << 4 | b;
          }
          if (xim->byte_order != mbo) revbyte2(pix2);
          pix2++;
        }
      }
    }

    return xim;
}


void
XFTNIMGPUT (display, window, gc, xim, width, height)
    Display **display;
    Drawable *window;
    GC *gc;
    XImage **xim;
    unsigned int *width, *height;
{
    XPutImage(*display, *window, *gc, *xim, 0, 0, 0, 0, *width, *height);
}


void
XFTNIMGFREE (display, xim)
    Display **display;
    XImage **xim;
{
    XDestroyImage(*xim);
}


void
XFTNDUMP(window, filename, text, fnLEN)
    Window *window;
    char   *filename, *text;
    int    fnLEN;
{
    XImage            *xim;
    XWindowAttributes xwa;
    FILE              *fp;
    char              *ptr, *name;
    int               nbytes, dum, i, iend;

    name  = (char *) malloc((fnLEN+1)*sizeof(char));
    for (i = 0; i < fnLEN; i++) if (filename[i] != ' ') break;
    for (iend = fnLEN; iend > 1; iend--) if (filename[iend-1] != ' ') break;
    for (dum = i; dum < iend; dum++) name[dum-i] = filename[dum];
    name[iend-i] = '\0';

    fp = fopen(name,"w");
    if (fp == NULL) {
      printf(" Error opening file: %s\n\n", name);
      free(name);
      return;
    }

    fwrite("Visual ImageFile", sizeof(char), 16, fp);
    dum = 2;
    fwrite(&dum, sizeof(int), 1, fp);

    XGetWindowAttributes(dpy, *window, &xwa);
    xim = XGetImage(dpy, *window, 0, 0, xwa.width, xwa.height,
                    AllPlanes, ZPixmap);

    dum = True;
    fwrite(&dum, sizeof(int), 1, fp);
    fwrite(&bgr, sizeof(int), 1, fp);
    dum = xwa.width;
    fwrite(&dum, sizeof(int), 1, fp);
    dum = xwa.height;
    fwrite(&dum, sizeof(int), 1, fp);
    dum = vdepth;
    fwrite(&dum, sizeof(int), 1, fp);
    dum = xim->bits_per_pixel;
    fwrite(&dum, sizeof(int), 1, fp);
    dum = xim->byte_order;
    fwrite(&dum, sizeof(int), 1, fp);

    fwrite(text, sizeof(char), 4*80, fp);

    dum = 4;
    if (xim->bits_per_pixel == 16) dum = 2;
    nbytes = dum*xwa.width;
    ptr = xim->data;
    for (i = 0; i < xwa.height; i++) {
      fwrite(ptr, 1, nbytes, fp);
      ptr = ptr + xim->bytes_per_line;
    }

    XDestroyImage(xim);
    fclose(fp);
    printf(" Visual Image file created: %s\n\n", name);
    free(name);
}
