/*
 *      clients.h
 *
 *      pV3 server-side client structures
 *        for pV3 Revision 2.10 beta
 *
 *	Copyright 1995 - 2003, Massachusetts Institute of Technology.
 *
 */

#ifndef _CLIENTS_H_

#define _CLIENTS_H_

/*
 * defines
 */

#define MDISCIPLINE 24
#define MCLIENTS   512


/*
 * structures
 */


  typedef struct {
    int sizes[9];		/* number of nodes and ther sizes */
    int maxblk;			/* maximum number of blocks */
    int *blksz;			/* block sizes */
  } Mesh;


  typedef struct {
    int   nClient;		/* number of clients */
    int   mClient;		/* max client index */
    int   tid[MCLIENTS];	/* PVM tid */
    int   opt[MCLIENTS];	/* status - unsteady option */
    Mesh  size[MCLIENTS];	/* current discretization sizes */
    int   mark[MCLIENTS];	/* mark for timing frame */
    float time[MCLIENTS];	/* time from client */
    char  title[MCLIENTS][20];	/* partition name */
    int   on[MCLIENTS];		/* render flag for sub-setting */
    int   mirror[MCLIENTS];	/* mirror / instancing flag (- = number) */
    float *repm[MCLIENTS];	/* pointer to replication matrix */
    int   *repi[MCLIENTS];	/* pointer to replication masks */
    int   mrep[MCLIENTS];	/* maximum number of replicants */
    int   ctid[MCLIENTS];	/* concen PVM tid, !=-1 points to 1st client */
    int   sbuf[MCLIENTS];	/* send buffers */
  } Clients;


  typedef struct {
    int   type;                 /* type index */
    int   recalc;		/* recalculate flag */
    float fmin;			/* min of field */
    float fmax;			/* max of field */
    float time;			/* time stamp for min and max */
  } Fields;


  typedef struct {
    int nDiscipline;			/* number of displines */
    int scal[MDISCIPLINE];              /* current scalar */
    int vect[MDISCIPLINE];              /* current vector */
    int srfs[MDISCIPLINE];              /* current surface scalar */
    int srfv[MDISCIPLINE];              /* current surface vector */
    int thrs[MDISCIPLINE];              /* current threshold */
    int nWords;				/* FORTRAN common length */
    int *ptr;				/* FORTRAN common pointer */
    int *common[MDISCIPLINE];           /* FORTRAN common save area */
    Clients *clients[MDISCIPLINE];	/* discipline pointers */
    char title[MDISCIPLINE][20];	/* discipline name */
    int nFields[MDISCIPLINE];		/* number of fileds */
    Fields *fields[MDISCIPLINE];	/* field pointers */
    float scaling[MDISCIPLINE][4];	/* discipline level scaling, xlate */
  } Discipline;

#endif  /*_CLIENTS_H_*/
