/*
 * dials.c
 *
 * Xlib/dial-box interface for the AIX 4.1 (up) and Visual3/pV3
 *
 * Copyright 1995 - 1999, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XI.h>
#include <X11/extensions/XInput.h>


#ifdef UNDERSCORE

#define XFTNINITDIAL xftninitdial_
#define XFTNGETDIALS xftngetdials_

#else

#define XFTNINITDIAL xftninitdial
#define XFTNGETDIALS xftngetdials

#endif


  extern Display *XFtnDpy;
  extern XEvent   XFtnXEvent;


#define NUMEVENTS 1

  int XFtnDVals[8] = {0,0,0,0,0,0,0,0};
  int XFtnDeviceMotionNotify;
  XEventClass XFtnEventClass[NUMEVENTS];
  XDevice *XFtnDDevice = NULL;

int
XFTNINITDIAL()
{
  int                      i, stat, ndevices, res[8];
  XDeviceInfoPtr           list, slist;
  XDeviceResolutionControl value;

  if (XFtnDDevice != NULL) return 1;
  list = (XDeviceInfoPtr) XListInputDevices (XFtnDpy, &ndevices);
  stat  = 0;
  slist = list;
  for (i=0; i<ndevices; i++, list++) {
    if (strcmp(XI_KNOB_BOX, list->name) == 0) {
      stat = 1;
      XFtnDDevice = XOpenDevice(XFtnDpy, list->id);

      for (i = 0; i < 8; i++) res[i] = 8;
      value.control = DEVICE_RESOLUTION;
      value.length = sizeof(XDeviceResolutionControl);
      value.first_valuator = 0;
      value.num_valuators = 8;
      value.resolutions = res;
      XChangeDeviceControl(XFtnDpy, XFtnDDevice, DEVICE_RESOLUTION, &value);

      DeviceMotionNotify (XFtnDDevice, XFtnDeviceMotionNotify,
                          XFtnEventClass[0]);
    }
  }
  XFreeDeviceList(slist);
  return stat*200;
}


int
XFTNGETDIALS(drev)
  int drev[];

{
  int k, ret;

  ret = 0;
  if (XFtnDDevice == NULL) return ret;
  for (k = 0; k < 8; k++) {
    drev[k] = XFtnDVals[k];
    if (drev[k] != 0) ret = 1;
    XFtnDVals[k] = 0;
  }
  return ret;
}


void
xftnDialEvent()

{
  XDeviceMotionEvent *ereport;
  unsigned char first_axis;
  int i, k, *axis_data;

  if (XFtnDDevice == NULL) return;
  if (XFtnXEvent.type == XFtnDeviceMotionNotify) {
    ereport = (XDeviceMotionEvent *) &XFtnXEvent;
    i = ereport->first_axis;
    axis_data = ereport->axis_data;
    for (k = 0; k < ereport->axes_count; k++, i++)
      XFtnDVals[i] += axis_data[k];
  }
}


void
xftnSetDialEvent(window)
  Window window;

{
  if (XFtnDDevice == NULL) return;
  XSelectExtensionEvent(XFtnDpy, window, XFtnEventClass, 1);
}
