/*
 * dials.c
 *
 * Xlib/dial-box interface for Visual3 for HPs
 *
 * Copyright 1990 - 1998, Massachusetts Institute of Technology.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <X11R6/X11/Xlib.h>
#include <X11R6/X11/extensions/XInput.h>
#include <X11R6/X11/extensions/XI.h>

#ifdef UNDERSCORE

#define XFTNINITDIAL xftninitdial_
#define XFTNGETDIALS xftngetdials_

#else

#define XFTNINITDIAL xftninitdial
#define XFTNGETDIALS xftngetdials

#endif

#define MAXKNOBDEVS 3

  extern Display *XFtnDpy;
  extern XEvent   XFtnXEvent;

  XID          XFtnDID[MAXKNOBDEVS];
  XDevice     *XFtnDDevice[MAXKNOBDEVS];
  XEventClass  XFtnDMotionMask[MAXKNOBDEVS];
  int          XFtnDMotion, XFtnDDevs, XFtnDVals[9];

int
XFTNINITDIAL()
{
    XDeviceInfo *list;
    int         i, j, numDevices;

    XFtnDDevs = 0;
    if ((list = XListInputDevices(XFtnDpy,&numDevices)) != NULL ) {
      for (i = 0; i < numDevices; ++i) 
        if (strcmp(list[i].name,XI_NINE_KNOB) == 0) 
          if (XFtnDDevs < MAXKNOBDEVS) XFtnDID[XFtnDDevs++] = list[i].id;
      XFreeDeviceList(list);
    }
    if (XFtnDDevs != 3) XFtnDDevs = 0;
    if (XFtnDDevs != 0) printf("Using 9 Knob DialBox!\n");
    for (j = 0; j < XFtnDDevs; ++j) {
      if ((XFtnDDevice[j] = XOpenDevice(XFtnDpy,XFtnDID[j])) == NULL) {
        printf("Could not open knob box: %d\n",j);
        XFtnDDevs = 0;
      }
      DeviceMotionNotify(XFtnDDevice[j],XFtnDMotion,XFtnDMotionMask[j]);
    }

    for (i = 0; i < 9; i++) XFtnDVals[i] = 0;
    if (XFtnDDevs != 0) return 700;
    return XFtnDDevs;
}

int
XFTNGETDIALS(drev)
    int drev[];

{
    int k, ret;
    static int klate[8] = {6, 3, 0, 8, 7, 4, 1, 5};

    ret = 0;
    if (XFtnDDevs == 0) return ret;
    for (k = 0; k < 8; k++) {
      drev[k] = XFtnDVals[klate[k]];
      if (drev[k] != 0) ret = 1;
    }
    for (k = 0; k < 9; k++) XFtnDVals[k] = 0;
    return ret;
}


void
xftnDialEvent()

{
    int                count;
    XDeviceMotionEvent *pMotionEvent;

    if (XFtnDDevs == 0) return;
    if (XFtnXEvent.type == XFtnDMotion) {
      pMotionEvent = (XDeviceMotionEvent *) &XFtnXEvent;
      count = (pMotionEvent->deviceid-1)*3;
      XFtnDVals[count  ] += pMotionEvent->axis_data[0];
      XFtnDVals[count+1] += pMotionEvent->axis_data[1];
      XFtnDVals[count+2] += pMotionEvent->axis_data[2];
    }
}


void
xftnSetDialEvent(window)
  Window window;
{
    if (XFtnDDevs == 0) return;
    XSelectExtensionEvent(XFtnDpy, window, XFtnDMotionMask, XFtnDDevs);
}
