/*
 * sballSGI.c
 *
 * Xlib/spaceball interface for the SGI and pV3
 *
 * M. Behr (behr@ahpcrc.umn.edu) July 1, 1997
 *
 * 6DOF 3D and 4DOF 2D control for Spaceball or Magellan
 * button 9 (right) resets the default 3D view (almost)
 * based on dialsHPPA.c from the Visual3 distribution
 * compile with -DUNDERSCORE and optional -DDEBUG
 * link with pV3Server.o rev 1.25 or higher
 *
 * to do:
 * sensitivity control by user
 * soft buttons (1-8) support
 *
 */

#include <memory.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/extensions/XInput.h>

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

#if defined(UNDERSCORE)

#define XFTNINITDIAL  xftninitdial_
#define XFTNGETDIALS  xftngetdials_
#define PV_GETPOINTER pv_getpointer_
#define PV_GETSTATE   pv_getstate_
#define PV_SETSTATE   pv_setstate_

#else

#define XFTNINITDIAL  xftninitdial
#define XFTNGETDIALS  xftngetdials
#define PV_GETPOINTER pv_getpointer
#define PV_GETSTATE   pv_getstate
#define PV_SETSTATE   pv_setstate

#endif

extern void PV_GETPOINTER(int *opt, int *ptr);
extern void PV_GETSTATE  (int *opt, int *ivec, float *rvec, char *str, int len);
extern void PV_SETSTATE  (int *opt, int *ivec, float *rvec, char *str, int len);

extern Display  *XFtnDpy;
extern XEvent    XFtnXEvent;

XDevice     *XFtnDDev = NULL;
XEventClass  XFtnDEventMask[3];
int          XFtnDMotion, XFtnDPress, XFtnDRelease, XFtnDDevs, XFtnDVals[9];
int          ipass = 0, firstcall = TRUE;
Window       window1d = NULL, window2d = NULL, window3d = NULL;

int
XFTNINITDIAL()
{
  int              i, stat, ndevices, sensitivity;
  XDeviceInfoPtr   list;
  char             *id;

  if (XFtnDDev != NULL) return 1;
  list = (XDeviceInfoPtr) XListInputDevices(XFtnDpy, &ndevices);
  stat = 0;
  for (i=0; i<ndevices; i++) {
    if (strcmp(list[i].name,"spaceball") == NULL) {
      stat = 1;
      XFtnDDev = XOpenDevice(XFtnDpy, list[i].id);
	id = XGetAtomName(XFtnDpy, list[i].type);
    }
  }
  XFreeDeviceList(list);

  if (stat == 0) {
    XFtnDDevs = 0;
  } else {
    DeviceMotionNotify (XFtnDDev, XFtnDMotion , XFtnDEventMask[0]);
    DeviceButtonPress  (XFtnDDev, XFtnDPress  , XFtnDEventMask[1]);
    DeviceButtonRelease(XFtnDDev, XFtnDRelease, XFtnDEventMask[2]);
    XFtnDDevs = 3;
  }

  for (i=0; i<9; i++) XFtnDVals[i] = 0;

  /* std setting is for Logitech  Magellan, id == "SPACE MOUSE" */
  /* alt setting is for IMC Spaceball 3003, id ==   "SPACEBALL" */
  sensitivity = 8000;
  if (strcmp(id,"SPACEBALL") == NULL) sensitivity = 2000;
  XFree(id);
  return - stat * sensitivity;
}

int
XFTNGETDIALS(drev)
  int drev[];

{
  int k, ret, *dummy; static int dzero[9] = {0};

  /* if no controller present return immediately */
  ret = 0; if (XFtnDDevs == 0) return ret;

  /* to prevent overflows in looong pV3 sessions */
  if ( ipass < 8 ) ipass++;

  /* copy valuators and check for non-zero input */
  dummy = memcpy(drev, XFtnDVals, 9 * sizeof(int));
  if ( memcmp(dzero, drev, 9 * sizeof(int)) != 0 ) ret = 1;
  dummy = memcpy(XFtnDVals, dzero, 9 * sizeof(int));

  /* check for pending motion events and call event handler */
  if (XCheckTypedEvent(XFtnDpy, XFtnDMotion, &XFtnXEvent)) {
	xftnDialEvent(); ret = 1;
  }

  /* return control only after N consecutive idle passes */
  if ( ret == 1 ) ipass = 0; else if ( ipass < 4 ) ret = 1;

  return ret;
}

xftnDialEvent()
{
  Window window;
  int iopt;
  if (XFtnDDevs == 0) return;
  if (firstcall) {
    iopt = 1; PV_GETPOINTER(&iopt, (int *) &window1d);
    iopt = 2; PV_GETPOINTER(&iopt, (int *) &window2d);
    iopt = 3; PV_GETPOINTER(&iopt, (int *) &window3d);
#if defined(DEBUG)
    printf("xftnDialEvent: 1D window id is %d\n", window1d);
    printf("xftnDialEvent: 2D window id is %d\n", window2d);
    printf("xftnDialEvent: 3D window id is %d\n", window3d);
#endif
    firstcall = FALSE;
  }
  if (XFtnXEvent.type == XFtnDMotion) {
    XDeviceMotionEvent *pMotionEvent = (XDeviceMotionEvent *) &XFtnXEvent;
    if (pMotionEvent->deviceid == XFtnDDev->device_id) {
      int i;
      static short sbdata[7] = {0};
      static long lasttime = 0, guesstime = TRUE;
      for (i=0;i<pMotionEvent->axes_count;i++)
        *(sbdata+pMotionEvent->first_axis+i) = pMotionEvent->axis_data[i];
      sbdata[6] = 16 * (guesstime ? 50 : pMotionEvent->time - lasttime);
      lasttime = pMotionEvent->time;
      guesstime = (sbdata[0]==0 && sbdata[1]==0 && sbdata[2]==0
                && sbdata[3]==0 && sbdata[4]==0 && sbdata[5]==0);
      window = XFtnXEvent.xany.window;
#if defined(DEBUG)
      printf("xftnDialEvent: sball motion %4d%4d%4d%4d%4d%4d%4d in window %d\n",
      sbdata[0],sbdata[1],sbdata[2],sbdata[3],sbdata[4],sbdata[5],sbdata[6],
	window);
#endif
	if (window == window3d) {
        XFtnDVals[0] += sbdata[3]; /*RX*/   XFtnDVals[4] -= sbdata[5]; /*RZ*/
        XFtnDVals[1] += sbdata[4]; /*RY*/ /*XFtnDVals[5] +=         0;     */
        XFtnDVals[2] += sbdata[0]; /*TX*/   XFtnDVals[6] -= sbdata[1]; /*TY*/
        XFtnDVals[3] += sbdata[2]; /*TZ*/ /*XFtnDVals[7] +=         0;     */
	  XFtnDVals[8]  =         3;
	} else if (window == window2d) {
      /*XFtnDVals[0] +=         0;     */   XFtnDVals[4] -= sbdata[4]; /*RY*/
      /*XFtnDVals[1] +=         0;     */ /*XFtnDVals[5] +=         0;     */
        XFtnDVals[2] -= sbdata[0]; /*TX*/   XFtnDVals[6] += sbdata[2]; /*TZ*/
        XFtnDVals[3] += sbdata[1]; /*TY*/ /*XFtnDVals[7] +=         0;     */
	  XFtnDVals[8]  =         2;
	}
    }
  } else if (XFtnXEvent.type == XFtnDPress) {
    XDeviceButtonEvent *pButtonEvent = (XDeviceButtonEvent *) &XFtnXEvent;
    if (pButtonEvent->deviceid == XFtnDDev->device_id) {
      int button = pButtonEvent->button;
#if defined(DEBUG)
      printf("xftnDialEvent: sball button %d pressed\n", button);
#endif
    }
  } else if (XFtnXEvent.type == XFtnDRelease) {
    XDeviceButtonEvent *pButtonEvent = (XDeviceButtonEvent *) &XFtnXEvent;
    if (pButtonEvent->deviceid == XFtnDDev->device_id) {
      int button = pButtonEvent->button;
      if (button == 9) {
        int ivec[1]; float rad, rvec[12];
        iopt = 14;
        PV_GETSTATE(&iopt, ivec, rvec, NULL, 0);
        rad = 1.0 / rvec[3];
        iopt = 1;
        rvec[ 0] = rad; rvec[ 1] = 0.0; rvec[ 2] = 0.0; rvec[ 3] = 0.0;
        rvec[ 4] = 0.0; rvec[ 5] = rad; rvec[ 6] = 0.0; rvec[ 7] = 0.0;
        rvec[ 8] = 0.0; rvec[ 9] = 0.0; rvec[10] = rad; rvec[11] = 0.0;
        PV_SETSTATE(&iopt, NULL, rvec, NULL, 0);
      }
#if defined(DEBUG)
      printf("xftnDialEvent: sball button %d released\n", button);
#endif
    }
  }
}

xftnSetDialEvent(window)
  Window window;
{
  if (XFtnDDevs == 0) return;
#if defined(DEBUG)
  printf("xftnSetDialEvent: called for %d events in window %d\n",
  XFtnDDevs, window);
#endif
  XSelectExtensionEvent(XFtnDpy, window, XFtnDEventMask, 3);
}
