#include <stdio.h>
#include <stdlib.h>
#if defined (WIN32) || defined (ABSOFT)
#include <WINSOCK.H>
#include <direct.h>
#else
#include <unistd.h>
#endif
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "pvm3.h"
#include "clients.h"

#ifdef ABSOFT
#undef WIN32
#endif


/*
  routines to deal with PVM network traffic for the pV3 server
    for pV3 Revision 2.10 beta

  Copyright 1993 - 2003, Massachusetts Institute of Technology.

*/

#ifdef WIN32

#define PREFIX __stdcall

#else

#ifdef ABSOFT
#define PREFIX __cdecl
#else
#define PREFIX
#endif

#ifdef UNDERSCORE

#define PV_GETEXTRACT pv_getextract_
#define PV_GETSTATE pv_getstate_

#define PV_PVMINIT pv_pvminit_
#define PV_RECVINIT pv_recvinit_
#define PV_PVMEND pv_pvmend_
#define PV_CHECKCLIENTS pv_checkclients_
#define PV_CHECKSERVERS pv_checkservers_
#define PV_SENDSERVERS pv_sendservers_
#define PV_RECVMES pv_recvmes_
#define PV_RECVTIME pv_recvtime_
#define PV_RECVSTRING pv_recvstring_
#define PV_SENDSTRING pv_sendstring_
#define PV_SENDREQUEST pv_sendrequest_
#define PV_SENDDATA pv_senddata_
#define PV_RECV pv_recv_
#define PV_GETMESS pv_getmess_
#define PV_SENDPARTXFER pv_sendpartxfer_
#define PV_SENDSLXFER pv_sendslxfer_
#define PV_SENDPARTBOUNCE pv_sendpartbounce_
#define PV_GETXFER pv_getxfer_
#define PV_GETINPUT pv_getinput_
#define PV_NUMSERVERS pv_numservers_
#define PV_SERVERINFO pv_serverinfo_
#define PV_SETSERVER pv_setserver_
#define PV_GETSERVER pv_getserver_
#define PV_SERVERUPDATE pv_serverupdate_
#define PV_QUERYSEREX pv_queryserex_
#define PV_MARKSEREX pv_markserex_
#define PV_MESSERVER pv_messerver_

#else

#define PV_GETEXTRACT pv_getextract
#define PV_GETSTATE pv_getstate

#define PV_PVMINIT pv_pvminit
#define PV_RECVINIT pv_recvinit
#define PV_PVMEND pv_pvmend
#define PV_CHECKCLIENTS pv_checkclients
#define PV_CHECKSERVERS pv_checkservers
#define PV_SENDSERVERS pv_sendservers
#define PV_RECVMES pv_recvmes
#define PV_RECVTIME pv_recvtime
#define PV_RECVSTRING pv_recvstring
#define PV_SENDSTRING pv_sendstring
#define PV_SENDREQUEST pv_sendrequest
#define PV_SENDDATA pv_senddata
#define PV_RECV pv_recv
#define PV_GETMESS pv_getmess
#define PV_SENDPARTXFER pv_sendpartxfer
#define PV_SENDSLXFER pv_sendslxfer
#define PV_SENDPARTBOUNCE pv_sendpartbounce
#define PV_GETXFER pv_getxfer
#define PV_GETINPUT pv_getinput
#define PV_NUMSERVERS pv_numservers
#define PV_SERVERINFO pv_serverinfo
#define PV_SETSERVER pv_setserver
#define PV_GETSERVER pv_getserver
#define PV_SERVERUPDATE pv_serverupdate
#define PV_QUERYSEREX pv_queryserex
#define PV_MARKSEREX pv_markserex
#define PV_MESSERVER pv_messerver

#endif
#endif

#define HIPRIO 26729		/* High priority message tag */
#define LOPRIO 27759		/* Low priority message tag */
#define INPRIO 26990		/* Init message tag */
#define SERSER 29555		/* server-server message tag */

#define MAXSERINS 100		/* number of server instances to check */


typedef struct {
  int   did;			/* dicipline ID */
  int   type;			/* extract type */
  int   num;			/* instance number */
  int   ivec[11];		/* integer extract data */
  float rvec[9];		/* float extract data */
  char  title[20];		/* extract title */
  int   mark;			/* do we want this extract: 0 - No, 1 - Yes */
} SExtr;


typedef struct {
  int   tid;			/* server tid */
  int   type;			/* 0 - interactive, 1 - batch or 3 - built */
  char  *host;			/* host name */
  int   view;			/* 0 - no view, 1 - view extracts */
  int   vtid;			/* lock view tid */
  int   cntl;			/* control-> 0 - no, 1 - yes */
  float xform[12];		/* transformation matrix */
  int   cursor[5];		/* window, x and y, xsize and ysize */
  int   nExtr[2];		/* number of extracts */
  int   mExtr[2];		/* allocated number of extracts */
  SExtr *Extr[2];		/* pointer to allocated extracts */
  int   smes;			/* send message flag */
  char  mess[80];		/* the messaage */
} SPool;


/* globals */

  static int   isnum, last_tid, (*omatch)(), pv3_ifd, pv3_nser, pv3_mser;
  static int   pv3_to, lo_rbuf, st_rbuf, ser_update = 0, pv3_rev = 2;
  static char  *pv3_SGroup = NULL, *pv3_CGroup = NULL;
  static SPool *servers = NULL;

  extern Discipline pV3_mdc;
  extern int        pV3_cDisc, pV3_plot, pV3_gath;


  extern int  pv_LookupID(int did, int cid, int num);
  extern void PREFIX PV_GETSTATE(int *, int *, float *, char *, int);

#ifdef WIN32
  extern void __stdcall PV_GETEXTRACT(void **, int *, int *, int *, float *,
                                      char *, int, void **);
static void sleep(int secs)
{
  Sleep(1000*secs);
}

#else

  extern void PREFIX PV_GETEXTRACT(void **, int *, int *, int *, float *, 
                                   char *, void **, int);
#ifdef ABSOFT
static void sleep(int secs)
{
  Sleep(1000*secs);
}
#endif

#endif


/* recv comparison function - get all hi prio messages first */

static int
pv_match(int bid, int tid, int mid)
{
  int t, m, cc, len;

  if ((cc = pvm_bufinfo(bid, &len, &m, &t)) < 0) return cc;

  if (mid == -1) {
    if (m == HIPRIO)  return 1;
    if (m == LOPRIO)  return 2;
    if (m == INPRIO)  return 2;
    if (m == SERSER)  return 2;
  } else {
    if (m == mid) return 1;
  }

  return 0;
}


/* interrupt handler */

#ifdef SYS5
void
pv_handler(int sig)
#else
void
pv_handler(int sig, int code, struct sigcontext sc)
#endif

{
  pvm_recvf(omatch);
  pvm_lvgroup(pv3_SGroup);
  pvm_exit();
  exit(sig);
}


/* init and check for clients */

void PREFIX
PV_PVMINIT (int *num, int *nserver, int *timeout)
{
  int    gsize, i, j, k, tid, active, len, flags, stat, nhost, narch, hid, ins;
  int    normal;
  char   *env;
  struct pvmhostinfo *hostp;
#ifndef PVM33
  int    new, info, buf;
#endif
#if defined (WIN32) || defined (ABSOFT)
  char   buffer[_MAX_PATH], cdok;
#endif
  static char Server[7] = "Server";
  static char Client[7] = "Client";

  normal = *num;

  if (normal == 0) {
    signal(SIGINT , pv_handler);
#if defined (WIN32) || defined (ABSOFT)
#else
    signal(SIGBUS , pv_handler);
#endif
    signal(SIGSEGV, pv_handler);
  }

#if defined (WIN32) || defined (ABSOFT)
   /* Get the current working directory: */
  cdok = 0;
  if(_getcwd(buffer,_MAX_PATH ) == NULL) cdok = 1;
#endif

  isnum = pvm_mytid();
  if (isnum < 0) {
    printf("pV3 Server Error: pvmd Not running!\n");
    if (normal == 0) {
      exit(1);
    } else {
      *num = -1;
      return;
    }
  }
#if defined (WIN32) || defined (ABSOFT)
#else
  if (normal == 0) {
    if ((pv3_ifd = open("/dev/tty",O_RDONLY)) == -1) {
       printf("pV3 Server Warning: Error Opening input terminal port!\n");
    } else {
      flags = fcntl(pv3_ifd, F_GETFL, 0);
      stat  = fcntl(pv3_ifd, F_SETFL, flags|O_NDELAY);
      if (stat != 0) 
        printf("pV3 Server Warning: Input open status = %d\n", stat);
    }
  }
#endif

  env = getenv("pV3_Rev");
  if (env != NULL) 
    if (strcmp(env, "1") == 0) {
      printf(" Assuming pV3 Clients @ Rev 1\n");
      pv3_rev = 1;
    }

#ifndef PVM33
  /* set/get context */

  if (pv3_rev > 1) {
    new = pvm_newcontext();
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&new, 1, 1);
    info = pvm_putinfo("pV3context", pvm_getsbuf(), PvmMboxPersistent);
    if (info < 0) {
      pvm_freecontext(new);
      new = -1;
    }
    if ((info == PvmExists) || (info == PvmDenied)) {
      /* must be already registered -- get context */
      buf = pvm_recvinfo("pV3context", 0, PvmMboxFirstAvail);
      if (buf >= 0) {
        pvm_setrbuf(buf);
        pvm_upkint(&new, 1, 1);
      }
    }
    info = -1000;
    if (new >= 0) info = pvm_setcontext(new);
    if (info < 0) new  = -1;
    if (new == -1) {
      printf("pV3 Server Error: Can not set Context %d!\n", info);
      if (normal == 0) {
        exit(1);
      } else {
        *num = -1;
        return;
      }
    }
  }
#endif

  st_rbuf = 0;
  lo_rbuf = 0;

  /* set group names */

  env = getenv("pV3_Group");
  if (env == NULL) {
    env = "pV3";
  } else {
    printf(" Using group prefix = %s\n", env);
  }
  len = strlen(env);
  pv3_CGroup = malloc(len + 7);
  pv3_SGroup = malloc(len + 7);
  for (i = 0; i < len; i++) {
    pv3_CGroup[i] = env[i];
    pv3_SGroup[i] = env[i];
  }
  for (i = 0; i < 7; i++) pv3_CGroup[i+len] = Client[i];
  for (i = 0; i < 7; i++) pv3_SGroup[i+len] = Server[i];

  active = 0;
  gsize  = pvm_gsize(pv3_CGroup);
  if (gsize > 0)
    for (i = 0; i < gsize; i++) {
      tid = pvm_gettid(pv3_CGroup, i);
      if (tid > 0) if (pvm_pstat(tid) >= 0) active++;
    }
  if (active == 0) {
    sleep(15);
    gsize = pvm_gsize(pv3_CGroup);
    if (gsize > 0)
      for (i = 0; i < gsize; i++) {
        tid = pvm_gettid(pv3_CGroup, i);
        if (tid > 0) if (pvm_pstat(tid) >= 0) active++;
      }
  }
  if (active == 0) {
    printf("pV3 Server Error: No pV3 Clients running (group = %s)!\n",
            pv3_CGroup);
    if (normal == 0) {
      exit(1);
    } else {
      *num = -2;
      return;
    }
  }
  *num = active;
  if ((gsize > 0) && (gsize != active))
    printf(" pV3 Server Warning: %s size = %d, only %d task(s) active!\n\n",
           pv3_CGroup, gsize, active);

  pvm_config(&nhost, &narch, &hostp);
  active = 1;
  gsize  = pvm_gsize(pv3_SGroup);
  ins    = -1;
  i      = 0;
  while ((i < gsize) && (ins < MAXSERINS)) {
    ins++;
    tid = pvm_gettid(pv3_SGroup, ins);
    if (tid == PvmNoInst) continue;
    i++;
    if (tid > 0) {
      active++;
      if (active == 2) 
        printf(" pV3 Information: Server(s) Already running (group = %s):\n",
               pv3_SGroup);
      hid = pvm_tidtohost(tid);
      for (j = 0; j < nhost; j++) if (hostp[j].hi_tid == hid)
        printf("   %d.  Server TID = %X,  Group Ins = %d,  Host = %s\n", 
               active, tid, ins, hostp[j].hi_name);
    }
  }
  if (active != 1)  printf("\n");

  pv3_nser = active;
  *nserver = pv3_nser;
  pv3_mser = 8;
  if (pv3_mser < pv3_nser) pv3_mser = pv3_nser;
  servers = (SPool *) malloc(pv3_mser*sizeof(SPool));
  if (servers == NULL) {
    printf("pV3 Server Error: Can't Allocate server pool!\n");
    if (normal == 0) {
      exit(1);
    } else {
      *num = -3;
      return;
    }
  }
  servers[0].tid  = isnum;
  servers[0].type = -1;
  servers[0].host = NULL;
  servers[0].view = 0;
  servers[0].vtid = 0;
  servers[0].cntl = 1;
  servers[0].smes = 0;
  hid = pvm_tidtohost(isnum);
  for (j = 0; j < nhost; j++) 
    if (hostp[j].hi_tid == hid) servers[0].host = strdup(hostp[j].hi_name);
  ins = -1;
  i   =  0;
  k   =  0;
  while ((i < gsize) && (ins < MAXSERINS)) {
    ins++;
    tid = pvm_gettid(pv3_SGroup, ins);
    if (tid == PvmNoInst) continue;
    i++;
    if (tid > 0) {
      k++;
      servers[0].cntl = 0;
      servers[k].tid  = tid;
      servers[k].type = -1;
      servers[k].host = NULL;
      servers[k].view = 0;
      servers[k].vtid = 0;
      servers[k].cntl = 0;
      servers[k].smes = 0;
      for (j = 0; j < 12; j++) servers[k].xform[j] = 0.0;
      servers[k].cursor[0] = 0;
      servers[k].nExtr[0]  = 0;
      servers[k].nExtr[1]  = 0;
      servers[k].mExtr[0]  = 0;
      servers[k].mExtr[1]  = 0;
      servers[k].Extr[0]   = NULL;
      servers[k].Extr[1]   = NULL;
      hid = pvm_tidtohost(tid);
      for (j = 0; j < nhost; j++) 
        if (hostp[j].hi_tid == hid) servers[k].host = strdup(hostp[j].hi_name);
    }
  }

  pvm_joingroup(pv3_SGroup);
  omatch = pvm_recvf(pv_match);
  pv3_to = *timeout;
#if defined (WIN32) || defined (ABSOFT)
  if (cdok == 0) _chdir(buffer);
#endif
}


/* recv init handshake from a client */

#ifdef WIN32
void __stdcall
PV_RECVINIT (int *ncon, int *tid, int *kns, int *npc, char *tc, int tcLEN, 
             int *mirror, float *rev, int *nk, int **ik, int **fk, float **fl,
             char **tk, char *tit, int titLEN, char *dtit, int dtitLEN, 
             char *ptit, int ptitLEN, float *xmx, int *ns, char *nt, int ntLEN,
             float *repmat, int *mblk)
#else
void PREFIX
PV_RECVINIT (int *ncon, int *tid, int *kns, int *npc, char *tc, int *mirror,
             float *rev, int *nk, int **ik, int **fk, float **fl, char **tk,
             char *tit, char *dtit, char *ptit, float *xmx, int *ns, char *nt,
             float *repmat, int *mblk, int tcLEN, int titLEN, int dtitLEN,
             int ptitLEN, int ntLEN)
#endif
{
  int   i, nkeys, *ikey, *fkey;
  float *flim;
  char  *tkey;

  /* unpack the init packet */
  *tid = last_tid;
  pvm_upkint(ncon, 1, 1);
  pvm_upkint(npc, 1, 1);
  pvm_upkint(mirror, 1, 1);
  pvm_upkint(nk, 1, 1);
  nkeys = abs(*nk);
  *ik = (int *) malloc(nkeys*sizeof(int));
  *fk = (int *) malloc(nkeys*sizeof(int));
  *fl = (float *) malloc(2*nkeys*sizeof(int));
  *tk = (char *) malloc(32*nkeys*sizeof(char));
  if ((*ik == NULL) || (*fk == NULL) || (*fl == NULL) || (*tk == NULL)) {
    printf("ERROR - Memory Allocation for Key Bindings (%d)\n", nkeys);
    exit(1);
  }
  ikey = *ik;
  fkey = *fk;
  flim = *fl;
  tkey = *tk;
  for (i = 0; i < nkeys; i++, ikey++, fkey++) {
    pvm_upkint(ikey, 1, 1);
    pvm_upkint(fkey, 1, 1);
  }
  if (*kns != 0)    pvm_upkint(ns, *kns*3, 1);
  if (pv3_rev != 1) pvm_upkint(mblk, 1, 1);
  pvm_upkfloat(rev, 1, 1);
  if ((*rev < 1.31) || (*rev > 2.90)) {
    printf(" pV3 Error: Client signalling bad Rev number -> %.2f\n\n", *rev);
    exit(1);
  }
  if ((*rev > 1.90) && (pv3_rev == 1)) {
    printf(" pV3 Error: expecting Rev 1 clients -> Got %.2f\n\n", *rev);
    exit(1);
  }
  if ((*rev < 1.90) && (pv3_rev == 2)) {
    printf(" pV3 Error: expecting Rev 2 clients -> Got %.2f\n\n", *rev);
    exit(1);
  }
  pvm_upkfloat(xmx, 6, 1);
  for (i = 0; i < nkeys; i++, flim+=2) pvm_upkfloat(flim, 2, 1);
  if (*mirror < 0) pvm_upkfloat(repmat, 16, 1);
  pvm_upkbyte(tit, 80, 1);
  pvm_upkbyte(dtit, 20, 1);
  pvm_upkbyte(ptit, 20, 1);
  for (i = 0; i < nkeys; i++, tkey+=32) pvm_upkbyte(tkey, 32, 1);
  if (*kns != 0) pvm_upkbyte(nt, *kns*20, 1);
  if (*npc != 0) pvm_upkbyte(tc, *npc*32, 1);
}


/* terminate server */

void PREFIX
PV_PVMEND (int *stat)
{
  int j;

  pvm_recvf(omatch);
  pvm_lvgroup(pv3_SGroup);
  pvm_exit();
  if (servers != NULL) {
    for (j = 1; j < pv3_nser; j++) {
      if (servers[j].host    != NULL) free(servers[j].host);
      if (servers[j].Extr[0] != NULL) free(servers[j].Extr[0]);
      if (servers[j].Extr[1] != NULL) free(servers[j].Extr[1]);
    }
    free(servers);
    servers = NULL;
  }
  if (pv3_CGroup != NULL) free(pv3_CGroup);
  if (pv3_SGroup != NULL) free(pv3_SGroup);
  pv3_CGroup = NULL;
  pv3_SGroup = NULL;
  if (*stat == 0) exit(0);
}


/* check clients */

int PREFIX
PV_CHECKCLIENTS ()
{
  int      stat, i, ier, id;
  Clients *clients;

  stat = 0;
  for (id = 0; id < pV3_mdc.nDiscipline; id++) {
    clients = pV3_mdc.clients[id];
    if (clients != NULL)
      for (i = 0; i < clients->mClient; i++) 
        if (clients->opt[i] != -1) {
          ier = pvm_pstat(clients->tid[i]);
          if (ier != 0) stat = ier;
        }
  }
  return stat;
}


/* check the server pool */

void PREFIX
PV_CHECKSERVERS (int *nserver, int *pvtype)
{
  int    i, j, k, tid, active, nhost, narch, hid, update, gsize, ins, nser;
  FILE   *fp;
  struct pvmhostinfo *hostp;

  servers[0].type = *pvtype;

  /* stop? */

  if (*pvtype == 1) {
    servers[0].cntl = 1;
    if ((fp = fopen("pV3Batch.stop","r")) != NULL) {
      fclose(fp);
      unlink("pV3Batch.stop");
      *nserver = -1;
      return;
    }
  }

  /* do we have a big enough structure? */

  active = pvm_gsize(pv3_SGroup);
  if (active > pv3_mser) {
    pv3_mser *= 2;
    if (active > pv3_mser) pv3_mser = active;
    servers = (SPool *) realloc(servers, pv3_mser*sizeof(SPool));
    if (servers == NULL) {
      printf("pV3 Server Error: Can't ReAllocate server pool!\n");
      exit(1);
    }
  }

  /* remove dead servers */

  update = 0;
  if (pv3_nser != 1) {
    for (nser = j = 1; j < pv3_nser; j++)
      if (pvm_pstat(servers[j].tid) >= 0) {
        servers[nser].tid  = servers[j].tid;
        servers[nser].type = servers[j].type;
        servers[nser].host = servers[j].host;
        servers[nser].view = servers[j].view;
        servers[nser].vtid = servers[j].vtid;
        servers[nser].cntl = servers[j].cntl;
        servers[nser].smes = servers[j].smes;
        for (k = 0; k < 80; k++) servers[nser].mess[k]   = servers[j].mess[k];
        for (k = 0; k < 12; k++) servers[nser].xform[k]  = servers[j].xform[k];
        for (k = 0; k <  5; k++) servers[nser].cursor[k] = servers[j].cursor[k];
        servers[nser].nExtr[0] = servers[j].nExtr[0];
        servers[nser].nExtr[1] = servers[j].nExtr[1];
        servers[nser].mExtr[0] = servers[j].mExtr[0];
        servers[nser].mExtr[1] = servers[j].mExtr[1];
        servers[nser].Extr[0]  = servers[j].Extr[0];
        servers[nser].Extr[1]  = servers[j].Extr[1];
        nser++;
      } else {
        if (servers[j].host    != NULL) free(servers[j].host);
        if (servers[j].Extr[0] != NULL) free(servers[j].Extr[0]);
        if (servers[j].Extr[1] != NULL) free(servers[j].Extr[1]);
        update++;
      }
    pv3_nser = nser;
  }

  /* add new servers */

  gsize = pvm_gsize(pv3_SGroup);
  ins   = -1;
  i     = 0;
  while ((i < gsize) && (ins < MAXSERINS)) {
    ins++;
    tid = pvm_gettid(pv3_SGroup, ins);
    if (tid == PvmNoInst) continue;
    i++;
    if (tid > 0) {
      active = 0;
      for (j = 0; j < pv3_nser; j++) if (servers[j].tid == tid) active++;
      if (active == 0) {
        servers[pv3_nser].tid  = tid;
        servers[pv3_nser].type = -1;
        servers[pv3_nser].host = NULL;
        servers[pv3_nser].view =  0;
        servers[pv3_nser].vtid =  0;
        servers[pv3_nser].cntl =  0;
        servers[pv3_nser].smes =  0;
        for (k = 0; k < 12; k++) servers[pv3_nser].xform[k] = 0.0;
        servers[pv3_nser].cursor[0] = 0;
        servers[pv3_nser].nExtr[0]  = 0;
        servers[pv3_nser].nExtr[1]  = 0;
        servers[pv3_nser].mExtr[0]  = 0;
        servers[pv3_nser].mExtr[1]  = 0;
        servers[pv3_nser].Extr[0]   = NULL;
        servers[pv3_nser].Extr[1]   = NULL;
        pv3_nser++;
        update++;
      }
    }
  }

  *nserver = pv3_nser;
  if (pv3_nser == 1) servers[0].cntl = 1;
  if (pv3_nser == 0) printf(" pV3 Warning: # server == 0!\n");
  if (update == 0) return;

  pvm_config(&nhost, &narch, &hostp);

  printf(" pV3 Information: Current Server Pool (group = %s):\n", pv3_SGroup);
  for (i = 0; i < pv3_nser; i++) {
    hid = pvm_tidtohost(servers[i].tid);
    for (j = 0; j < nhost; j++) if (hostp[j].hi_tid == hid) {
      if (servers[i].host == NULL) servers[i].host = strdup(hostp[j].hi_name);
      printf("   %d.  Server TID = %X,  Host = %s\n", i+1, servers[i].tid,
             hostp[j].hi_name);
      }
  }
  printf("\n");
}


/* send all active servers our state */

void PREFIX
PV_SENDSERVERS (int *cursor)
{
  int   i, did, nex, stat, type, num, ivec[11], dsave;
  float rvec[12], xform[12];
  char  name[20];
  void  *ex, *nextex;

  if (pv3_nser == 1) return;
  dsave = pV3_cDisc;

  for (nex = did = 0; did < pV3_mdc.nDiscipline; did++) {
    pV3_cDisc = did;
    ex = NULL;
    nex++;
#ifdef WIN32
    PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, 20, &nextex);
#else
    PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, &nextex, 20);
#endif
    while(nextex != NULL) {
      ex = nextex;
      nex++;
#ifdef WIN32
      PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, 20, &nextex);
#else
      PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, &nextex, 20);
#endif
    }
  }

  i = 1;
  PV_GETSTATE(&i, ivec, xform, name, 20);

  pvm_initsend(PvmDataDefault);
  pvm_pkbyte("pV3s", 4, 1);
  pvm_pkint(&servers[0].type, 1, 1);
  pvm_pkint(&servers[0].vtid, 1, 1);
  pvm_pkint(&servers[0].cntl, 1, 1);
  pvm_pkfloat(xform, 12, 1);
  pvm_pkint(cursor, 5, 1);
  pvm_pkint(&nex, 1, 1);
  for (did = 0; did < pV3_mdc.nDiscipline; did++) {
    pV3_cDisc = did;
    ex = NULL;
#ifdef WIN32
    PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, 20, &nextex);
#else
    PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, &nextex, 20);
#endif
    pvm_pkint(&did, 1, 1);
    pvm_pkint(&type, 1, 1);
    pvm_pkint(&num, 1, 1);
    pvm_pkint(ivec, 11, 1);
    pvm_pkfloat(rvec, 9, 1);
    pvm_pkbyte(name, 20, 1);
    ex = nextex;
    while(ex != NULL) {
#ifdef WIN32
      PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, 20, &nextex);
#else
      PV_GETEXTRACT(&ex, &type, &num, ivec, rvec, name, &nextex, 20);
#endif
      pvm_pkint(&did, 1, 1);
      pvm_pkint(&type, 1, 1);
      pvm_pkint(&num, 1, 1);
      pvm_pkint(ivec, 11, 1);
      pvm_pkfloat(rvec, 9, 1);
      pvm_pkbyte(name, 20, 1);
      ex = nextex;
    }
  }

  for (i = 1; i < pv3_nser; i++) {
    stat = pvm_send(servers[i].tid, SERSER);
    if (stat != 0) printf(" pV3 Server Warning: Server Send Stat-> %d to %X\n",
                          servers[i].tid, stat);
  }

  pV3_cDisc = dsave;

  /* send out any server-to-server communication */

  for (i = 1; i < pv3_nser; i++)
    if (servers[i].smes != 0) {
      pvm_initsend(PvmDataDefault);
      pvm_pkbyte("pV3S", 4, 1);
      pvm_pkint(&servers[i].smes,  1, 1);
      pvm_pkbyte(servers[i].mess, 80, 1);
      stat = pvm_send(servers[i].tid, SERSER);
      if (stat != 0) printf(" pV3 Server Warning: Server Send Stat-> %d to %X\n",
                            servers[i].tid, stat);
      servers[i].smes = 0;
    }
}


/* unpack the ack message */

void PREFIX
PV_RECVMES (float *reals)
{
  pvm_upkfloat(reals, 5, 1);
}


/* unpack single int */

void
pv_recvi(int *ints)
{
  pvm_upkint(ints, 1, 1);
}


/* unpack int extract */

void
pv_recvxi(int *ints)
{
  int offset, np;

  pvm_upkint(&offset, 1, 1);
  pvm_upkint(&np, 1, 1);
  pvm_upkint(&ints[offset], np, 1);
}


/* unpack float extract */

void
pv_recvxr(float *reals)
{
  int offset, np;

  pvm_upkint(&offset, 1, 1);
  pvm_upkint(&np, 1, 1);
  pvm_upkfloat(&reals[offset], np, 1);
}


/* skip the extract -- needs sizeof(float) == sizeof(int)! */

void
pv_recvxd()
{
  int i, offset, np;

  pvm_upkint(&offset, 1, 1);
  pvm_upkint(&np, 1, 1);
  for (i = 0; i < np; i++) pvm_upkint(&offset, 1, 1);
}


/* unpack the time type message */

void PREFIX
PV_RECVTIME (int *did, int *cl, float *time, char *str, int strLEN)
{
  Fields  *fields;
  Clients *clients;
  int     nks, i, ier, block[3], sizes[9] = { 0, 0, 0, 0, 0, 0, 0, 0, 0 };
  float   mm[2];

  ier = 0;
  if ((*did < 0) || (*did >= pV3_mdc.nDiscipline)) {
    printf("pV3 Server Error: Discipline ID Out of Range! (%d)\n", *did);
    ier = 1;
  }
  if (pV3_mdc.clients[*did] == NULL) {
    printf("pV3 Server Error: Discipline ID not used! (%d)\n", *did);
    ier = 1;
  }
  clients = pV3_mdc.clients[*did];
  pvm_upkint(&nks, 1, 1);
  if (ier == 0) 
    if (nks != pV3_mdc.nFields[*did]) {
      printf("pV3 Server Error: Number of Fields Mis-Match! (%d != %d)\n",
             nks, pV3_mdc.nFields[*did]);
      ier = 2;
    }
  if (ier == 0) 
    if ((*cl < 1) || (*cl > clients->mClient)) {
      printf("pV3 Server Error: Client ID Out of Range -> %d (%d)\n",
             *cl, clients->mClient);
      ier = 3;
    }
  if (pv3_rev > 1) pvm_upkint(sizes, 9, 1);
  if (ier == 0) 
    for (i = 0; i < 9; i++) clients->size[*cl-1].sizes[i] = sizes[i];
  for (i = 0; i < sizes[7]; i++) {
    pvm_upkint(block, 3, 1);
    if (ier == 0) {
      clients->size[*cl-1].blksz[3*i  ] = block[0];
      clients->size[*cl-1].blksz[3*i+1] = block[1];
      clients->size[*cl-1].blksz[3*i+2] = block[2];
    }
  }
  pvm_upkfloat(time, 1, 1);
  if (ier == 0) fields = pV3_mdc.fields[*did];
  for (i = 0; i <= nks; i++) {
    pvm_upkfloat(mm, 2, 1);
    if (ier == 0) {
      if ((mm[0] < mm[1]) || (mm[1] != 0.0)) {
        /* NOTE: time must be exactly the same for this to work! */
        if (*time != fields[i].time) {
          fields[i].time = *time;
          fields[i].fmin = mm[0];
          fields[i].fmax = mm[1];
        } else {
          if (fields[i].fmin > mm[0]) fields[i].fmin = mm[0];
          if (fields[i].fmax < mm[1]) fields[i].fmax = mm[1];
        }
      }
    }
  }
  pvm_upkbyte(str, 80, 1);
}


/* unpack the time type message */

void PREFIX
PV_RECVSTRING (char *str, int strLEN)
{
  pvm_upkbyte(str, 80, 1);
}


/* recieve from everyone */

void PREFIX
PV_RECV (int *did,  int *cl,  int *mid,  int *i1,  int *i2,  int *i3)
{
  int   i, j, k, stat, bufid, len, tid, id, sizes[3], ivec[5], nmess;
  float rvec[12];
  char  pw[4], mess[80];
  SExtr *tmp;
  struct timeval tmout;

  /* in high prio read buffer */

  if (st_rbuf == 2) {
    *mid = -1;
    pvm_upkint(mid, 1, 1);
    if (*mid == -1) {
      if (lo_rbuf == 0) {
        st_rbuf = 0;
      } else {
        pvm_freebuf(pvm_setrbuf(lo_rbuf));
        st_rbuf = 1;
      }
    }
  }

  /* in low prio read buffer */

  if (st_rbuf == 1) {
    if (pvm_probe(-1, HIPRIO) > 0) {
      stat = pvm_setrbuf(0);
      if (stat != lo_rbuf) printf("Buffer ERROR: %d, %d\n", stat, lo_rbuf);
      bufid = pvm_recv(-1, HIPRIO);
      stat = pvm_bufinfo(bufid, &len, &id, &tid);
      if (stat < 0) {
        printf("pV3 Server Warning: bufinfo status on bufid %d is %d\n",
                bufid, stat);
        *mid = -3;
        stat = pvm_setrbuf(lo_rbuf);
        return;
      }
      pvm_upkbyte(pw, 4, 1);
      if ((pw[0]!='p') || (pw[1]!='V') || (pw[2]!='3') || (pw[3]!='#')) {
        *i1 = id;
        *i2 = *i3 = 0;
        *mid = -2;
        stat = pvm_setrbuf(lo_rbuf);
        return;
      }
      st_rbuf = 2;
      pvm_upkint(mid, 1, 1);
    } else {
      *mid = -1;
      pvm_upkint(mid, 1, 1);
      if (*mid == -1) {
        pvm_freebuf(pvm_setrbuf(lo_rbuf));
        st_rbuf = 0;
        lo_rbuf = 0;
      }
    }
  }

  /* no current read buffer */

  if (st_rbuf == 0) {
    *mid = -1;
    if (pv3_to == 0) {
      bufid = pvm_nrecv(-1, -1);
    } else {
      tmout.tv_sec  = pv3_to;
      tmout.tv_usec = 0;
      bufid = pvm_trecv(-1, -1, &tmout);
    }
    if (bufid <= 0) return;

    stat = pvm_bufinfo(bufid, &len, &id, &tid);
    if (stat < 0) {
      printf("pV3 Server Warning: bufinfo status on bufid %d is %d\n",
              bufid, stat);
      *mid = -3;
      return;
    }
    pvm_upkbyte(pw, 4, 1);

    /* Server-to-Server Message */
    if ((pw[0]=='p') && (pw[1]=='V') && (pw[2]=='3') && (pw[3]=='S')) {
      for (i = 1; i < pv3_nser; i++) if (servers[i].tid == tid) {
        pvm_upkint(&nmess, 1, 1);
        if (nmess != 0) {
          printf(" pV3 Server-to-Server Message from %s:\n", servers[i].host);
          for (k = 0; k < nmess; k++) {
            pvm_upkbyte(mess, 80, 1);
            for (len = 79; len > 0; len--) if (mess[len] != ' ') break;
            for (j = 0; j <= len; j++) printf("%c", mess[j]);
            printf("\n");
          }
          printf("\n");
        }
      }
      *mid = -3;
      return;
    }

    /* Server-to-Server Status */
    if ((pw[0]=='p') && (pw[1]=='V') && (pw[2]=='3') && (pw[3]=='s')) {
      for (i = 1; i < pv3_nser; i++)
        if (servers[i].tid == tid) {
          for (j = 0; j < 12; j++) rvec[j] = servers[i].xform[j];
          for (j = 0; j <  5; j++) ivec[j] = servers[i].cursor[j];
          j = servers[i].type;
          pvm_upkint(&servers[i].type, 1, 1);
          if (j != servers[i].type) ser_update++;
          j = servers[i].vtid;
          pvm_upkint(&servers[i].vtid, 1, 1);
          if (j != servers[i].vtid) ser_update++;
          j = servers[i].cntl;
          pvm_upkint(&servers[i].cntl, 1, 1);
          if (j != servers[i].cntl) ser_update++;
          pvm_upkfloat(servers[i].xform, 12, 1);
          if (servers[0].vtid == tid)
            for (j = 0; j < 12; j++)
              if (rvec[j] != servers[i].xform[j]) ser_update++;
          pvm_upkint(servers[i].cursor, 5, 1);
          if ((servers[i].cursor != 0) || (ivec[0] != 0))
            for (j = 0; j < 5; j++)
              if (ivec[j] != servers[i].cursor[j]) ser_update++;
          j = servers[i].nExtr[pV3_gath];
          pvm_upkint(&servers[i].nExtr[pV3_gath], 1, 1);
          if (j != servers[i].nExtr[pV3_gath]) ser_update++;
          if (servers[i].nExtr[pV3_gath] > servers[i].mExtr[pV3_gath]) {
            if (servers[i].Extr[pV3_gath] == NULL) {
              tmp = (SExtr *) malloc(servers[i].nExtr[pV3_gath]*sizeof(SExtr));
            } else {
              tmp = (SExtr *) realloc(servers[i].Extr[pV3_gath],
                                      servers[i].nExtr[pV3_gath]*sizeof(SExtr));
            }
            if (tmp == NULL) {
              servers[i].nExtr[pV3_gath] = 0;
              printf("pV3 Server Warning: MemAlloc for Server Comm!\n");
              *mid = -3;
              return;
            } else {
              servers[i].mExtr[pV3_gath] = servers[i].nExtr[pV3_gath];
              servers[i].Extr[pV3_gath]  = tmp;
            }
          }
          for (k = 0; k < servers[i].nExtr[pV3_gath]; k++) {
            pvm_upkint(&servers[i].Extr[pV3_gath][k].did, 1, 1);
            pvm_upkint(&servers[i].Extr[pV3_gath][k].type, 1, 1);
            pvm_upkint(&servers[i].Extr[pV3_gath][k].num, 1, 1);
            pvm_upkint(servers[i].Extr[pV3_gath][k].ivec, 11, 1);
            pvm_upkfloat(servers[i].Extr[pV3_gath][k].rvec, 9, 1);
            pvm_upkbyte(servers[i].Extr[pV3_gath][k].title, 20, 1);
            servers[i].Extr[pV3_gath][k].mark = 0;
          }
          if ((servers[i].type == 1) && (servers[0].cntl == 1))
            servers[0].cntl = 0;
          *mid = -3;
          return;
        }

      printf("pV3 Server Warning: MessID = %d from UnRegerstered Server %X!\n",
             id, tid);
      *mid = -3;
      return;
    }

    /* Client-to-Server Message */
    if ((pw[0]!='p') || (pw[1]!='V') || (pw[2]!='3') || (pw[3]!='#')) {
      *i1 = id;
      *i2 = *i3 = 0;
      *mid = -2;
      return;
    }
    st_rbuf = 2;
    if ((id == LOPRIO) || (id == INPRIO)) {
      st_rbuf = 1;
      lo_rbuf = bufid;
    }
    if (id == INPRIO) last_tid = tid;
    pvm_upkint(mid, 1, 1);
  }

  pvm_upkint(did, 1, 1);
  pvm_upkint(cl, 1, 1);
  pvm_upkint(i1, 1, 1);
  pvm_upkint(i2, 1, 1);
  pvm_upkint(i3, 1, 1);

  /* get message sizes and toss */
  pvm_upkint(sizes, 3, 1);
}


/* get a message from anyone - no concatination */

void PREFIX
PV_GETMESS (int *did,  int *cl,  int *mid,  int *i1,  int *i2,  int *i3)
{
  int  stat, bufid, len, tid, id;
  char pw[4];

  *mid = -1;
  bufid = pvm_nrecv(-1, -1);
  if (bufid <= 0) return;

  stat = pvm_bufinfo(bufid, &len, &id, &tid);
  if (stat < 0) {
    printf("pV3 Server warning: bufinfo status on bufid %d is %d\n",
           bufid, stat);
    *mid = -1;
    return;
  }
  pvm_upkbyte(pw, 4, 1);
  if ((pw[0]=='p') && (pw[1]=='V') && (pw[2]=='3') && (pw[3]=='S')) {
    *mid = -1;
    return;
  }
  if ((pw[0]=='p') && (pw[1]=='V') && (pw[2]=='3') && (pw[3]=='s')) {
    *mid = -1;
    return;
  }
  if ((pw[0]!='p') || (pw[1]!='V') || (pw[2]!='3') || (pw[3]!='#')) {
    *i1 = *mid;
    *i2 = *i3 = 0;
    *mid = -2;
    return;
  }
  pvm_upkint(mid, 1, 1);
  pvm_upkint(did, 1, 1);
  pvm_upkint(cl, 1, 1);
  pvm_upkint(i1, 1, 1);
  pvm_upkint(i2, 1, 1);
  pvm_upkint(i3, 1, 1);
}


/* send string */

void PREFIX
PV_SENDSTRING (int *did, int *cl, char *str, int strLEN)
{
  int     i, j, mid, len;
  Clients *clients;

  if (pv3_nser > 1)
    if (servers[0].cntl != 1) {
      printf(" pV3 Server Warning: Message NOT sent -- No Control!\n\n");
      return;
    }

  i = 0;
  len = 80;

  if (*cl == -1) {
    if ((*did < 0) || (*did >= pV3_mdc.nDiscipline)) {
      printf("pV3 Server Error: Discipline ID Out of Range! (%d)\n", *did);
      return;
    }
    clients = pV3_mdc.clients[*did];
    if (clients == NULL) {
      printf("pV3 Server Error: Discipline ID not used! (%d)\n", *did);
      return;
    }
    for (j = 1; j <= clients->mClient; j++) {
      if (clients->opt[j-1] != -1) {
        pvm_initsend(PvmDataDefault);
        mid = 31;
        pvm_pkint(&mid, 1, 1);
        pvm_pkint(did, 1, 1);
        pvm_pkint(&j, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&len, 1, 1);
        pvm_pkbyte(str, len, 1);
        mid = -1;
        pvm_pkint(&mid, 1, 1);
        pvm_send(clients->tid[j-1], LOPRIO);
      }
    }
  } else {
    pvm_initsend(PvmDataDefault);
    mid = 31;
    pvm_pkint(&mid, 1, 1);
    pvm_pkint(did, 1, 1);
    pvm_pkint(cl, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&i, 1, 1);
    pvm_pkint(&len, 1, 1);
    pvm_pkbyte(str, len, 1);
    mid = -1;
    pvm_pkint(&mid, 1, 1);
    i = pv_LookupID(*did, *cl, 1);
    if (i > 0) pvm_send(i, LOPRIO);
  }
}


/* clear send buffers */

void
pv_ClearBuffers(int did)
{
  int     j;
  Clients *clients;

  clients = pV3_mdc.clients[did];
  for (j = 0; j <= clients->mClient; j++) clients->sbuf[j] = 0;
  j = pvm_getsbuf();
  if (j != 0) pvm_freebuf(j);
}


/* send packed requests */

void
pv_SendBuffers(int did, int prio)
{
  int     i, j;
  Clients *clients;

  clients = pV3_mdc.clients[did];
  for (j = 0; j <= clients->mClient; j++) 
    if (clients->sbuf[j] != 0) {
      if (pvm_getsbuf() != clients->sbuf[j]) pvm_setsbuf(clients->sbuf[j]);
      i = -1;
      pvm_pkint(&i, 1, 1);
      i = LOPRIO;
      if (prio == 2) i = HIPRIO;
      pvm_send(clients->tid[j], i);
      pvm_freebuf(clients->sbuf[j]);
    }
}


/* pack request */

void
pv_PackRequest(int did, int cl, int mid, int is, int iv, int it, int mask,
               int *ints, float *reals)
{
  int     nint, nfloat, nchar, ci;
  Clients *clients;

  clients = pV3_mdc.clients[did];
  ci = cl-1;
  if (clients->ctid[ci] != -1) ci = clients->ctid[ci];
  if (clients->sbuf[ci] ==  0) clients->sbuf[ci] = pvm_mkbuf(PvmDataDefault);
  if (clients->sbuf[ci] <=  0) {
    clients->sbuf[ci] = 0;
    return;
  }

  if (pvm_getsbuf() != clients->sbuf[ci]) pvm_setsbuf(clients->sbuf[ci]);

  nint   = 2;
  nfloat = 9;
  nchar  = 0;
  if (mid == 18) {
    nint   = 8;
    nfloat = 30;
  }

  pvm_pkint(&mid, 1, 1);
  pvm_pkint(&did, 1, 1);
  pvm_pkint(&cl, 1, 1);
  pvm_pkint(&is, 1, 1);
  pvm_pkint(&iv, 1, 1);
  pvm_pkint(&it, 1, 1);
  pvm_pkint(&mask, 1, 1);
  pvm_pkint(&nint, 1, 1);
  pvm_pkint(&nfloat, 1, 1);
  pvm_pkint(&nchar, 1, 1);
  pvm_pkint(ints, nint, 1);
  pvm_pkfloat(reals, nfloat, 1);

}


/* send request */

void PREFIX
PV_SENDREQUEST (int *did, int *cl, int *mid, int *is, int *iv, int *it,
                int *mask, int *ints, float *reals)
{
  int     i, nint, nfloat, nchar, prio, j;
  Clients *clients;
  Fields  *fields;

  nint   = 2;
  nfloat = 9;
  nchar  = 0;
  prio   = LOPRIO;
  if (*mid == 18) {
    nint   = 8;
    nfloat = 30;
    prio   = HIPRIO;
  }
  if (*mid == 38) prio  = HIPRIO;
  if (*mid == 40) prio  = HIPRIO;
  if (*mid == 35) {
    nchar = 20;
    prio  = HIPRIO;
  }
  if (*mid == 34) {
    fields  = pV3_mdc.fields[*did];
    fields[0].time = -1.e30;
  }

  if ((*did < 0) || (*did >= pV3_mdc.nDiscipline)) {
    printf("pV3 Server Error: Discipline ID Out of Range! (%d)\n", *did);
    return;
  } 
  if (pV3_mdc.clients[*did] == NULL) {
    printf("pV3 Server Error: Discipline ID not used! (%d)\n", *did);
    return;
  } 

  if (*cl == -1) {
    clients = pV3_mdc.clients[*did];
    for (i = 1; i <= clients->mClient; i++) {
      if (clients->opt[i-1] != -1) {
        pvm_initsend(PvmDataDefault);
        pvm_pkint(mid, 1, 1);
        pvm_pkint(did, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(is, 1, 1);
        pvm_pkint(iv, 1, 1);
        pvm_pkint(it, 1, 1);
        pvm_pkint(mask, 1, 1);
        pvm_pkint(&nint, 1, 1);
        pvm_pkint(&nfloat, 1, 1);
        pvm_pkint(&nchar, 1, 1);
        pvm_pkint(ints, nint, 1);
        pvm_pkfloat(reals, nfloat, 1);
        if (*mid == 35) pvm_pkbyte(pV3_mdc.title[*did], 20, 1);
        j = -1;
        pvm_pkint(&j, 1, 1);
        pvm_send(clients->tid[i-1], prio);
      }
    }
  } else {
    pvm_initsend(PvmDataDefault);
    pvm_pkint(mid, 1, 1);
    pvm_pkint(did, 1, 1);
    pvm_pkint(cl, 1, 1);
    pvm_pkint(is, 1, 1);
    pvm_pkint(iv, 1, 1);
    pvm_pkint(it, 1, 1);
    pvm_pkint(mask, 1, 1);
    pvm_pkint(&nint, 1, 1);
    pvm_pkint(&nfloat, 1, 1);
    pvm_pkint(&nchar, 1, 1);
    pvm_pkint(ints, nint, 1);
    pvm_pkfloat(reals, nfloat, 1);
    if (*mid == 35) pvm_pkbyte(pV3_mdc.title[*did], 20, 1);
    j = -1;
    pvm_pkint(&j, 1, 1);
    i = pv_LookupID(*did, *cl, 2);
    if (i > 0) pvm_send(i, prio);
  }
}


/* send server data */

void PREFIX
PV_SENDDATA (int *did, int *cl, int *mid, int *nint, int *ints, int *nfloat,
             float *floats, int *nchar, char *chars)
{
  int     i, j, messid = 50;
  Clients *clients;

  if ((*nint <= 0) && (*nfloat <= 0) && (*nchar <= 0)) return;

  if ((*did < 0) || (*did >= pV3_mdc.nDiscipline)) {
    printf("pV_SendData Error: Discipline ID Out of Range! (%d)\n", *did);
    return;
  } 
  if (pV3_mdc.clients[*did] == NULL) {
    printf("pV_SendData Error: Discipline ID not used! (%d)\n", *did);
    return;
  } 

  if (*cl == -1) {
    clients = pV3_mdc.clients[*did];
    for (i = 1; i <= clients->mClient; i++) {
      if (clients->opt[i-1] != -1) {
        j = 0;
        pvm_initsend(PvmDataDefault);
        pvm_pkint(&messid, 1, 1);
        pvm_pkint(did, 1, 1);
        pvm_pkint(&i, 1, 1);
        pvm_pkint(&j, 1, 1);
        pvm_pkint(&j, 1, 1);
        pvm_pkint(&j, 1, 1);
        pvm_pkint(mid, 1, 1);
        pvm_pkint(nint, 1, 1);
        pvm_pkint(nfloat, 1, 1);
        pvm_pkint(nchar, 1, 1);
        if (*nint > 0)   pvm_pkint(ints, *nint, 1);
        if (*nfloat > 0) pvm_pkfloat(floats, *nfloat, 1);
        if (*nchar > 0)  pvm_pkbyte(chars, *nchar, 1);
        j = -1;
        pvm_pkint(&j, 1, 1);
        pvm_send(clients->tid[i-1], LOPRIO);
      }
    }
  } else {
    j = 0;
    pvm_initsend(PvmDataDefault);
    pvm_pkint(&messid, 1, 1);
    pvm_pkint(did, 1, 1);
    pvm_pkint(cl, 1, 1);
    pvm_pkint(&j, 1, 1);
    pvm_pkint(&j, 1, 1);
    pvm_pkint(&j, 1, 1);
    pvm_pkint(mid, 1, 1);
    pvm_pkint(nint, 1, 1);
    pvm_pkint(nfloat, 1, 1);
    pvm_pkint(nchar, 1, 1);
    if (*nint > 0)   pvm_pkint(ints, *nint, 1);
    if (*nfloat > 0) pvm_pkfloat(floats, *nfloat, 1);
    if (*nchar > 0)  pvm_pkbyte(chars, *nchar, 1);
    j = -1;
    pvm_pkint(&j, 1, 1);
    i = pv_LookupID(*did, *cl, 2);
    if (i > 0) pvm_send(i, LOPRIO);
  }
}


/* bounce a particle transfer to the proper client(s) */

void PREFIX
PV_SENDPARTXFER (int *did, int *toCl, int *intData, float *realData)
{
  int i, j;

  j = 0;
  pvm_initsend(PvmDataDefault);
  i = 16;
  pvm_pkint(&i, 1, 1);
  pvm_pkint(did, 1, 1);
  pvm_pkint(toCl, 1, 1);
  pvm_pkint(&j, 1, 1);
  pvm_pkint(&pV3_mdc.vect[*did], 1, 1);
  pvm_pkint(&j, 1, 1);
  pvm_pkint(&j, 1, 1);
  i = 8;
  pvm_pkint(&i, 1, 1);
  i = 30;
  pvm_pkint(&i, 1, 1);
  i = 0;
  pvm_pkint(&i, 1, 1);
  pvm_pkint(intData, 8, 1);
  pvm_pkfloat(realData, 30, 1);
  i = -1;
  pvm_pkint(&i, 1, 1);

  j = pv_LookupID(*did, *toCl, 4);
  if (j > 0) pvm_send(j, HIPRIO);
}


/*  bounce a streamline transfer to the proper client(s) */

void PREFIX
PV_SENDSLXFER (int *did, int *toCl, int *intData, float *realData)
{
  int j;

  pvm_initsend(PvmDataDefault);
  j = 18;
  pvm_pkint(&j, 1, 1);
  pvm_pkint(did, 1, 1);
  pvm_pkint(toCl, 1, 1);
  pvm_pkint(&intData[9], 1, 1);
  pvm_pkint(&intData[10], 1, 1);
  pvm_pkint(&intData[11], 1, 1);
  pvm_pkint(&intData[8], 1, 1);
  j = 8;
  pvm_pkint(&j, 1, 1);
  j = 30;
  pvm_pkint(&j, 1, 1);
  j = 0;
  pvm_pkint(&j, 1, 1);
  pvm_pkint(intData, 8, 1);
  pvm_pkfloat(realData, 30, 1);
  j = -1;
  pvm_pkint(&j, 1, 1);

  j = pv_LookupID(*did, *toCl, 5);
  if (j > 0) pvm_send(j, HIPRIO);
}


/*  bounce a particle start to the proper client(s) */

void PREFIX
PV_SENDPARTBOUNCE (int *did, int *toCl, int *intData, float *realData)
{
  int i, j;

  j = 0;
  pvm_initsend(PvmDataDefault);
  i = 17;
  pvm_pkint(&i, 1, 1);
  pvm_pkint(did, 1, 1);
  pvm_pkint(toCl, 1, 1);
  pvm_pkint(&intData[4], 1, 1);
  if (intData[4] == 0) {
    pvm_pkint(&pV3_mdc.vect[*did], 1, 1);
  } else {
    pvm_pkint(&intData[5], 1, 1);
  }
  pvm_pkint(&intData[2], 1, 1);
  pvm_pkint(&j, 1, 1);
  i = 2;
  pvm_pkint(&i, 1, 1);
  i = 9;
  pvm_pkint(&i, 1, 1);
  i = 0;
  pvm_pkint(&i, 1, 1);
  pvm_pkint(intData, 2, 1);
  pvm_pkfloat(realData, 9, 1);
  i = -1;
  pvm_pkint(&i, 1, 1);

  j = pv_LookupID(*did, *toCl, 6);
  if (j > 0) pvm_send(j, HIPRIO);
}


/* Unpack the rest of a transfer message */

void PREFIX
PV_GETXFER (int *mid, int *intData, float *realData)
{

  switch (*mid) {
  case 16:
    pvm_upkint(intData, 9, 1);
    pvm_upkfloat(realData, 30, 1);
    break;
  case 17:
    pvm_upkint(intData, 4, 1);
    pvm_upkfloat(realData, 24, 1);
    break;
  case 45:
    pvm_upkint(intData, 3, 1);
    pvm_upkfloat(realData, 3, 1);
    break;
  }
}


/* get input line -- non blocked */

int PREFIX
PV_GETINPUT (char *line, int lineLEN)
{
  int  i;
  char fill;

  if (pv3_ifd == -1) {
    printf("pV3 Server Error: Using input terminal port!\n");
    exit(1);
  }

  for (i = 0; i < lineLEN; i++) line[i] = ' ';
  if ((i = read(pv3_ifd, line, lineLEN)) < 1) return -1;
  if (line[i-1] != 10) {
    fill = 0;
    while (fill != 10) read(pv3_ifd, &fill, 1);
  } else {
    line[i-1] = ' ';
  }
  return i-1;
}


/* return the current number of active servers */

void PREFIX
PV_NUMSERVERS (int *num)
{
  *num = pv3_nser;
}


/* return the details about each server for collaboration */

#ifdef WIN32
void __stdcall
PV_SERVERINFO (int *tids, char *host, int hostLEN, int *type, int *view, 
               int *vindex, int *ctrl)
#else
void PREFIX
PV_SERVERINFO (int *tids, char *host, int *type, int *view, int *vindex,
               int *ctrl, int hostLEN)
#endif
{
  int    i, j;
  char   *hfill;

  /* initialize */

  for (i = 0; i < pv3_nser*hostLEN; i++) host[i] = ' ';
  for (i = 0; i < pv3_nser; i++) {
    type[i] = -1;
    tids[i] =  0;
  }

  /* get host name and fill in type, view and ctrl for valid servers */

  for (i = 0, hfill = host; i < pv3_nser; i++, hfill += hostLEN) {
    tids[i] = servers[i].tid;
    type[i] = servers[i].type;
    view[i] = servers[i].view;
    ctrl[i] = servers[i].cntl;
    for (j = 0; j < hostLEN; j++) {
      if (servers[i].host[j] == 0) break;
      hfill[j] = servers[i].host[j];
    }
  }

  /* translate tids to indices - (bias 1), 0 indicates not set */
  
  for (i = 0; i < pv3_nser; i++) {
    vindex[i] = 0;
    for (j = 0; j < pv3_nser; j++) 
      if (servers[i].vtid == servers[j].tid) vindex[i] = j+1;
  }

}


/* set info for this server in a collaborative setting */

void PREFIX
PV_SETSERVER (int *view, int *vindex, int *ctrl, int *stat)
{
  int i, vtid, cntl;

  *stat = -1;
  for (i = 1; i < pv3_nser; i++) 
    if ((view[i] < 0) || (view[i] > 1)) return;

  *stat = -2;
  vtid  = 0;
  if (*vindex < 0) return;
  if (*vindex > pv3_nser) return;
  if (*vindex != 0) vtid = servers[*vindex-1].tid;

  *stat = -3;
  if (*ctrl < 0) return;
  if (*ctrl > 1) return;
  cntl = *ctrl;

  /* check for control settings */

  *stat = 0;
  if (cntl == 1)
    for (i = 1; i < pv3_nser; i++) 
      if (servers[i].cntl == 1) {
        *stat = 1;
        cntl  = 0;
      }

  servers[0].vtid = vtid;
  servers[0].cntl = cntl;
  for (i = 1; i < pv3_nser; i++) servers[i].view = view[i];
}


/* get display info about the server pool */

void PREFIX
PV_GETSERVER (int *index, int *cursor, float *xform, int *stat)
{
  int i;

  *stat = -1;
  if ((*index < 2) || (*index > pv3_nser)) return;

  *stat = 0;
  for (i = 0; i <  5; i++) cursor[i] = servers[*index-1].cursor[i];
  for (i = 0; i < 12; i++) xform[i]  = servers[*index-1].xform[i];
}


/* has the server data been changed? */

int PREFIX
PV_SERVERUPDATE ()
{
  int i;

  i = ser_update;
  ser_update = 0;
  return i;
}


/* return the details about the server's extracts for collaboration */

int PREFIX
PV_QUERYSEREX (int *ser, int *index, int *did, int *type, int *num, int *ivec,
              float *rvec, char *title, int titleLEN)
{
  int i, iser, indx;

  if ((*ser < 2) || (*ser > pv3_nser)) {
     printf("pV3 Warning: Illegal SerIndex-> %d (pV_QuerySerEx)\n", *ser);
    return 0;
  }
  iser = *ser - 1;

  if (*index < 1) {
    *index = servers[iser].nExtr[pV3_plot];
    return 0;
  }

  if (*index > servers[iser].nExtr[pV3_plot]) {
     printf("pV3 Warning: Extract Index-> %d,  SerIndex = %d (pV_QuerySerEx)\n",
            *index, *ser);
    return 0;
  }
  indx = *index - 1;

  *did  = servers[iser].Extr[pV3_plot][indx].did;
  *type = servers[iser].Extr[pV3_plot][indx].type;
  *num  = servers[iser].Extr[pV3_plot][indx].num;
  for (i = 0; i < 11; i++) ivec[i]  = servers[iser].Extr[pV3_plot][indx].ivec[i];
  for (i = 0; i <  9; i++) rvec[i]  = servers[iser].Extr[pV3_plot][indx].rvec[i];
  for (i = 0; i < 20; i++) title[i] = servers[iser].Extr[pV3_plot][indx].title[i];
  return servers[iser].Extr[pV3_plot][indx].mark;
}


/* mark the server's extracts for inclusion */

void PREFIX
PV_MARKSEREX (int *ser, int *index)
{
  int iser, indx;

  if ((*ser < 2) || (*ser > pv3_nser)) {
     printf("pV3 Warning: Illegal SerIndex-> %d (pV_MarkSerEx)\n", *ser);
    return;
  }
  iser = *ser - 1;

  if (*index < 1) {
    *index = servers[iser].nExtr[pV3_plot];
    return;
  }

  if (*index > servers[iser].nExtr[pV3_plot]) {
     printf("pV3 Warning: Extract Index-> %d,  SerIndex = %d (pV_MarkSerEx)\n",
            *index, *ser);
    return;
  }
  indx = *index - 1;
  
  if (servers[iser].Extr[pV3_plot][indx].type <= 100) 
      servers[iser].Extr[pV3_plot][indx].mark = 1;
}


/* set-up for sending the message to the specified server */

void PREFIX
PV_MESSERVER (int *ser, char *mess, int messLEN)
{
  int i, iser, len;

  if ((*ser < 2) || (*ser > pv3_nser)) {
     printf("pV3 Warning: Illegal SerIndex-> %d (pV_MesServer)\n", *ser);
    return;
  }
  iser = *ser - 1;

  if (servers[iser].smes != 0) {
     printf("pV3 Warning: Message Pending -- This Message will Not be Sent!\n");
    return;
  }

  for (i = 0; i < 80; i++) servers[iser].mess[i] = ' ';
  len = messLEN;
  if (len > 80) len = 80;
  for (i = 0; i < len; i++) servers[iser].mess[i] = mess[i];
  servers[iser].smes = 1;
}
