/*
 * FXpV3.c
 *
 * FX example module for pV3  Rev 2
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "pV3.h"
#include "FX.h"



#ifdef PREFIX
#undef PREFIX
#endif


#ifdef WIN32

#define PREFIX __stdcall
#define pVSSINIT PVSSINIT

#else

#define PREFIX

#ifdef UNDERSCORE
#define pVSSINIT pvssinit_
#define FXPV3SHOCK fxpv3shock_
#else
#define pVSSINIT pvssinit
#define FXPV3SHOCK fxpv3shock
#endif

#endif

  /* globals */

  static int   xlate[5] = {0, 10000, 10001, 10002, 10003}; /* translate pV3 Scalars to FX Scalars */


  /* externals */

  extern void PREFIX pVSCAL(int *type, float *s);


/* FX system call-backs */

void PREFIX
FXCELLPTR(int **tets, int **pyras, int **prisms, int **hexas, int *halo)
{
  int i, len;

  i = 311;
  pV_GETSTRUC(&i, (void **) tets,   &len);
  i = 312;
  pV_GETSTRUC(&i, (void **) pyras,  &len);
  i = 313;
  pV_GETSTRUC(&i, (void **) prisms, &len);
  i = 314;
  pV_GETSTRUC(&i, (void **) hexas,  &len);
}

void PREFIX
FXSURFACEPTR(int *nsurf, int **cell, int **facet)
{
  int i, len, *bc;

  i = 324;
  pV_GETSTRUC(&i, (void **) facet, &len);
  i = 323;
  pV_GETSTRUC(&i, (void **) cell, &len);
  i = 321;
  pV_GETSTRUC(&i, (void **) &bc,  &len);

  for (i = 0; i < len; i++) {
    nsurf[2*i  ] = bc[3*i];
    nsurf[2*i+1] = 6;
  }
}

void PREFIX
FXGRIDPTR(float **xyz, float *hxyz)
{
  int i, len;

  i = 301;
  pV_GETSTRUC(&i, (void **) xyz, &len);
}


void PREFIX
FXVELPTR(float **vel, float *hvel)
{
  int i, len;

  i = 304;
  pV_GETSTRUC(&i, (void **) vel, &len);
}


void PREFIX
FXSCAL(int *type, float *s, float *hs)
{
  int i;

  if ((*type <= 5) && (*type != 4)) {
    i = xlate[*type-1];
    pVSCAL(&i, s);
  } else {
    printf(" FXSCAL Error: type = %d\n", *type);
  }
}

  /*  take the shock regions and remove those that are small or do not
   *    display a large enough pressure gradient
   */

static void
mark_tris(int nclust, int *clust, int *cell, int *cnode, int *cnodes,
          float *press, int ntris, int *cells, int minode, float mindp)
{
  char  *cellTable;
  int   i, j, k, l, clen, nlen, max;
  float minp, maxp;

  clen = clust[nclust-1];
  for (max = i = 0; i < clen; i++) if (cell[i] > max) max = cell[i];

  cellTable = (char *) malloc(max*sizeof(char));
  if (cellTable == NULL) {
    printf(" Cannot Allocate cellTable (mark_tris)\n");
    return;
  }

  for (i = 0; i < max; i++) cellTable[i] = 0;

  /* mark the cells that are OK */

  for (k = j = i = 0; i < nclust; i++) {
    clen = clust[i] - j;
    nlen = cnode[i] - k;
    if (nlen > minode) {
      minp = maxp = press[cnodes[k]-1];
      for (l = 1; l < nlen; l++) {
        if (minp > press[cnodes[k+l]-1]) minp = press[cnodes[k+l]-1];
        if (maxp < press[cnodes[k+l]-1]) maxp = press[cnodes[k+l]-1];
      }
       /* maxp-minp measures the approximate shock strength for the region */
      if ((maxp-minp) > mindp) 
        for (l = j; l < clust[i]; l++) cellTable[cell[l]-1] = 1;
    }
    j = clust[i];
    k = cnode[i];
  }

  /* mark the not-OK triangles */

  for (i = 0; i < ntris; i++)
    if (cellTable[cells[i]-1] == 0) cells[i] = -cells[i];

  free(cellTable);
}




/*
static void
FXpV3Shock(int index, int exnum, int minode, float mindp)
*/
void PREFIX
FXPV3SHOCK(int index, int exnum, int minode, float mindp)
{
  int   i, j, k, nnode, nclust, *clust, *cell, *cnode, *cnodes;
  int   npts, ntris, *tris, *cells, zero = 0, three = 3;
  float *xyz, *coors, *press, *test, *vols, *pts;

  i = 301;
  pV_GETSTRUC(&i, (void **) &xyz, &nnode);
//  if (tstoreFX.mglFXsv[citime].npts == -1) {
    printf("Computing shock surfaces\n");
    press = (float *) malloc(nnode*sizeof(float));
    if (press == NULL) {
      printf(" Malloc error for pressure\n");
      i = 1;
      pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
  
//      tstoreFX.mglFXsv[citime].npts = 1;
//      tstoreFX.mglFXsv[citime].pts = xyz;
  
      return;
    }
    i = xlate[1];
printf(" Getting pressure\n");
    pVSCAL(&i, press);
  
    npts = 0;
printf(" Finding Shocks\n");
    FX_SHOCKFIND(&test);
    if (test != NULL) {
printf(" Finding Shock surfaces\n");
      FX_SHOCKSURFACE(test, &npts, &coors, &ntris, &tris, &cells);
  
      /* get individual volumes */
  
printf(" \\Finding Shock volumes\n");
      FX_SHOCKVOLUMES(test, &nclust, &vols, &clust, &cell, &cnode, &cnodes);
printf(" /Done Finding Shock volumes\n");
      if (nclust > 0) {
        mark_tris(nclust, clust, cell, cnode, cnodes, press, ntris, cells,
                  minode, mindp);
printf(" Freeing vols\n");
        if (vols != NULL) free(vols);
        free(cnode);
        free(cnodes);
        free(clust);
        free(cell);
      }
      /* save away only the surfaces for OK regions */
      for (j = i = 0; i < ntris; i++) if (cells[i] > 0) j++;
      npts = 3*j;
      if (npts != 0) {
        pts = (float *) malloc(3*npts*sizeof(float));
        if (pts == NULL) {
          npts = 0;
        } else {
          for (j = i = 0; i < ntris; i++) {
            if (cells[i] < 0) continue;
            k = tris[3*i  ]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
            k = tris[3*i+1]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
            k = tris[3*i+2]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
          }
        }
      }
printf(" Freeing volume tries\n");
      free(test);
      free(coors);
      free(tris);
      free(cells);
    }
printf(" Freeing pressure\n");
    free(press);

//printf(" Storing shock points\n");
//    tstoreFX.mglFXsv[citime].npts = npts;
//    tstoreFX.mglFXsv[citime].pts = pts;

//  } else {

    //printf("Using stored shock surfaces\n");
//    npts = tstoreFX.mglFXsv[citime].npts;
//    pts = tstoreFX.mglFXsv[citime].pts;

//  }

  if (npts != 0) {
    pV_SENDXR(&index, &exnum, &zero, &three, &npts, pts);
  } else {
    i = 1;
    pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
  }
}
