/*
 * FXpV3.c
 *
 * FX example module for pV3  Rev 2
 *
 * Copyright 1999 - 2000, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "pV3.h"
#include "FX.h"



#ifdef PREFIX
#undef PREFIX
#endif


#ifdef WIN32

#define PREFIX __stdcall
#define pVSSINIT PVSSINIT

#else

#define PREFIX

#ifdef UNDERSCORE
#define pVSSINIT pvssinit_
#else
#define pVSSINIT pvssinit
#endif

#endif

/* Structure definitions for storing data */
typedef struct myFXvc_t {
  int nseg;
  int *segs;
  float *vcoor;
  float *stren;
} myFXvc;

typedef struct myFXsa_t {
  int calc;
  int *FXsepind;
  int *FXattind;
  float *sepxyz;
  float *attxyz;
} myFXsa;

typedef struct myFXsv_t {
  int npts;
  float *pts;
} myFXsv;

typedef struct myFXstorage_t {
   myFXvc *mglFXvc;
   myFXsv *mglFXsv;
   myFXsa *mglFXsa;
} myFXstorage;


  /* globals */

  static int   xlate[5] = {0, 10000, 10001, 10002, 10003}; /* translate pV3 Scalars to FX Scalars */

  static int   *FXsepind, *FXattind, *FXshowSurf, FXnbc;
  static int   case_opt;

  static int   citime; /* This is to store the simulation time */
  static int   cntime; /* This is to store the simulation max time */

  myFXstorage  tstoreFX;


  /* externals */

  extern void PREFIX pVSCAL(int *type, float *s);


/*
 * These functions were added to allow the storage of FX data in time
 * such that recalculations could be avoided.
 */
void
myfxsetcntime_(int *setcntime)
{
  int i;

  cntime = *setcntime;
  tstoreFX.mglFXvc = (myFXvc *)malloc(cntime*sizeof(myFXvc));
  tstoreFX.mglFXsa = (myFXsa *)malloc(cntime*sizeof(myFXsa));
  tstoreFX.mglFXsv = (myFXsv *)malloc(cntime*sizeof(myFXsv));
  for ( i=0; i < cntime; i++ ) {

     tstoreFX.mglFXvc[i].nseg = -1;
     tstoreFX.mglFXvc[i].segs = NULL;
     tstoreFX.mglFXvc[i].vcoor = NULL;
     tstoreFX.mglFXvc[i].stren = NULL;

     tstoreFX.mglFXsa[i].calc = 0;
     tstoreFX.mglFXsa[i].sepxyz = NULL;
     tstoreFX.mglFXsa[i].attxyz = NULL;

     tstoreFX.mglFXsv[i].npts = -1;
     tstoreFX.mglFXsv[i].pts = NULL;

  }
}

void
myfxsetcitime_(int *setcitime)
{
  citime = *setcitime - 1;
}

void
myfxclearstorevc_()
{
  int i;

  for ( i=0; i < cntime; i++ ) {
     printf("  \\ Clearing Vortex Cores for citime = %d \n",i);
     tstoreFX.mglFXvc[i].nseg = -1;
     if (tstoreFX.mglFXvc[i].segs != NULL) {
       printf("   \\ Clearing segs\n");
       free(tstoreFX.mglFXvc[i].segs);
       tstoreFX.mglFXvc[i].segs = NULL;
       printf("   / Done\n");
     }
     if (tstoreFX.mglFXvc[i].vcoor != NULL) {
       printf("   \\ Clearing vcoor\n");
       free(tstoreFX.mglFXvc[i].vcoor);
       tstoreFX.mglFXvc[i].vcoor = NULL;
       printf("   / Done\n");
     }
     if (tstoreFX.mglFXvc[i].stren != NULL) {
       printf("   \\ Clearing stren\n");
       free(tstoreFX.mglFXvc[i].stren);
       tstoreFX.mglFXvc[i].stren = NULL;
       printf("   / Done\n");
     }
  }
  printf("  / Done clearing vortex cores \n");

}

void
myfxclearstoresa_()
{
  int i;

  for ( i=0; i < cntime; i++ ) {
     printf("  \\ Clearing Sep & Att lines for citime = %d \n",i);
     tstoreFX.mglFXsa[i].calc = 0;
     if (tstoreFX.mglFXsa[i].sepxyz != NULL) {
       printf("   \\ Clearing sepxyz\n");
       free(tstoreFX.mglFXsa[i].sepxyz);
       tstoreFX.mglFXsa[i].sepxyz = NULL;
       printf("   / Done\n");
     }
     if (tstoreFX.mglFXsa[i].attxyz != NULL) {
       printf("   \\ Clearing attxyz\n");
       free(tstoreFX.mglFXsa[i].attxyz);
       tstoreFX.mglFXsa[i].attxyz = NULL;
       printf("   / Done\n");
     }
  }
  printf("  / Done clearing sep/att lines \n");

}

void
myfxclearstoresv_()
{
  int i;

  for ( i=0; i < cntime; i++ ) {
     printf("  \\ Clearing Shock Volumes for citime = %d \n",i);
     tstoreFX.mglFXsv[i].npts = -1;
     if (tstoreFX.mglFXsv[i].pts != NULL) {
       printf("   \\ Clearing pts\n");
       free(tstoreFX.mglFXsv[i].pts);
       tstoreFX.mglFXsv[i].pts = NULL;
       printf("   / Done\n");
     }
  }
  printf("  / Done clearing shock vols \n");

}

void
myfxclearstore_(int *dumnum)
{
  printf("\\ Clearing storage %d\n",*dumnum);
  printf(" | Clearing VC\n");
  myfxclearstorevc_();
  printf(" | Clearing SA\n");
  myfxclearstoresa_();
  printf(" | Clearing SV\n");
  myfxclearstoresv_();
  printf("/ Done %d\n\n",*dumnum);
}


  /*  take the shock regions and remove those that are small or do not
   *    display a large enough pressure gradient
   */

static void
mark_tris(int nclust, int *clust, int *cell, int *cnode, int *cnodes,
          float *press, int ntris, int *cells, int minode, float mindp)
{
  char  *cellTable;
  int   i, j, k, l, clen, nlen, max;
  float minp, maxp;

  clen = clust[nclust-1];
  for (max = i = 0; i < clen; i++) if (cell[i] > max) max = cell[i];

  cellTable = (char *) malloc(max*sizeof(char));
  if (cellTable == NULL) {
    printf(" Cannot Allocate cellTable (mark_tris)\n");
    return;
  }

  for (i = 0; i < max; i++) cellTable[i] = 0;

  /* mark the cells that are OK */

  for (k = j = i = 0; i < nclust; i++) {
    clen = clust[i] - j;
    nlen = cnode[i] - k;
    if (nlen > minode) {
      minp = maxp = press[cnodes[k]-1];
      for (l = 1; l < nlen; l++) {
        if (minp > press[cnodes[k+l]-1]) minp = press[cnodes[k+l]-1];
        if (maxp < press[cnodes[k+l]-1]) maxp = press[cnodes[k+l]-1];
      }
       /* maxp-minp measures the approximate shock strength for the region */
      if ((maxp-minp) > mindp) 
        for (l = j; l < clust[i]; l++) cellTable[cell[l]-1] = 1;
    }
    j = clust[i];
    k = cnode[i];
  }

  /* mark the not-OK triangles */

  for (i = 0; i < ntris; i++)
    if (cellTable[cells[i]-1] == 0) cells[i] = -cells[i];

  free(cellTable);
}


/* Initialization */

static void
init_fx(int iopt)
{
  int   i, nnode, ntets, npyra, nprism, nhexa, nblocks, *blocks, nsurf;
  int   istat, len, *tets, *pyra, *prism, *hexa, *blk, *iblk, zero = 0;
  int   *bc, *scel;
  float gamma, *xyz;

  case_opt = iopt;

  i = 315;
  pV_GETSTRUC(&i, (void **) &tets,  &len);
  if (len != 0) {
    printf(" This Case has %d PolyTet Strips \n\n", len);
    return;
  }
  i = 318;
  pV_GETSTRUC(&i, (void **) &tets,  &len);
  if (len != 0) {
    printf(" This Case has %d Complex Polyhedra \n\n", len);
    return;
  }

  istat = 47;
  i = 301;
  pV_GETSTRUC(&i, (void **) &xyz,   &nnode);

  i = 311;
  pV_GETSTRUC(&i, (void **) &tets,  &ntets);
  i = 312;
  pV_GETSTRUC(&i, (void **) &pyra,  &npyra);
  i = 313;
  pV_GETSTRUC(&i, (void **) &prism, &nprism);
  i = 314;
  pV_GETSTRUC(&i, (void **) &hexa,  &nhexa);
  i = 317;
  pV_GETSTRUC(&i, (void **) &blk,   &nblocks);
  if (nblocks != 0) {
    blocks = (int *) malloc(3*nblocks*sizeof(int));
    if (blocks == NULL) {
      printf(" Error in Memory Allocation for %d Blocks\n\n", nblocks);
      return;
    }
    for (i = 0; i < nblocks; i++) {
      blocks[3*i  ] = blk[6*i  ];
      blocks[3*i+1] = blk[6*i+1];
      blocks[3*i+2] = blk[6*i+2];
      nnode -= blocks[3*i]*blocks[3*i+1]*blocks[3*i+2];
    }
    i = 306;
    pV_GETSTRUC(&i, (void **) &iblk,  &len);
    if (len != 0) istat += 16;
  }

  i = 321;
  pV_GETSTRUC(&i, (void **) &bc,   &FXnbc);
  i = 323;
  pV_GETSTRUC(&i, (void **) &scel, &nsurf);
  if (FXnbc > 0) {
    FXshowSurf = malloc(FXnbc*sizeof(int));
    FXattind   = malloc(FXnbc*sizeof(int));
    FXsepind   = malloc(FXnbc*sizeof(int));
    if ((FXshowSurf == NULL) || (FXattind == NULL) || (FXsepind == NULL)) {
      if (FXshowSurf != NULL) free(FXshowSurf);
      if (FXattind   != NULL) free(FXattind);
      if (FXsepind   != NULL) free(FXsepind);
      printf(" Error in Memory Allocation for %d Surface Storage\n\n", FXnbc);
      if (nblocks != 0) free(blocks);
      return;
    }
  }

  gamma = 1.4;
  FX_INIT(&gamma, &iopt, &nnode, &zero, &ntets, &npyra, &nprism, &nhexa,
          &nblocks, blocks, &zero, &nsurf, &FXnbc, &istat);
  printf(" FX_Init return = %d\n\n", istat);
  if (nblocks != 0) free(blocks);
}


void PREFIX
pVSSINIT()
{
  init_fx(0);
}

void PREFIX
pVUPDATE(int *status)
{
  int    i, iopt;
  float  *time;
  static int count = 0;

  /* use undocumented method to get iopt and time */
  i = 0;
  pV_GETSTRUC(&i, (void **) &time,  &iopt);
  if (iopt == 3) {
    printf(" This is a Structure Unsteady Case!\n\n");
    return;
  }
  if ((count == 0) && (*status == 0)) init_fx(iopt);
  count++;

  FX_UPDATE(time);
}


static void
FXpV3Shock(int index, int exnum, int minode, float mindp)
{
  int   i, j, k, nnode, nclust, *clust, *cell, *cnode, *cnodes;
  int   npts, ntris, *tris, *cells, zero = 0, three = 3;
  float *xyz, *coors, *press, *test, *vols, *pts;

  i = 301;
  pV_GETSTRUC(&i, (void **) &xyz, &nnode);
  if (tstoreFX.mglFXsv[citime].npts == -1) {
    printf("Computing shock surfaces\n");
    press = (float *) malloc(nnode*sizeof(float));
    if (press == NULL) {
      printf(" Malloc error for pressure\n");
      i = 1;
      pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
  
      tstoreFX.mglFXsv[citime].npts = 1;
      tstoreFX.mglFXsv[citime].pts = xyz;
  
      return;
    }
    i = xlate[1];
//printf(" Getting pressure\n");
    pVSCAL(&i, press);
  
    npts = 0;
//printf(" Finding Shocks\n");
    FX_SHOCKFIND(&test);
    if (test != NULL) {
//printf(" Finding Shock surfaces\n");
      FX_SHOCKSURFACE(test, &npts, &coors, &ntris, &tris, &cells);
  
      /* get individual volumes */
  
//printf(" \\Finding Shock volumes\n");
      FX_SHOCKVOLUMES(test, &nclust, &vols, &clust, &cell, &cnode, &cnodes);
//printf(" /Done Finding Shock volumes\n");
      if (nclust > 0) {
        mark_tris(nclust, clust, cell, cnode, cnodes, press, ntris, cells,
                  minode, mindp);
//printf(" Freeing vols\n");
        if (vols != NULL) free(vols);
        free(cnode);
        free(cnodes);
        free(clust);
        free(cell);
      }
      /* save away only the surfaces for OK regions */
      for (j = i = 0; i < ntris; i++) if (cells[i] > 0) j++;
      npts = 3*j;
      if (npts != 0) {
        pts = (float *) malloc(3*npts*sizeof(float));
        if (pts == NULL) {
          npts = 0;
        } else {
          for (j = i = 0; i < ntris; i++) {
            if (cells[i] < 0) continue;
            k = tris[3*i  ]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
            k = tris[3*i+1]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
            k = tris[3*i+2]-1;
            pts[j] = coors[3*k  ];
            j++;
            pts[j] = coors[3*k+1];
            j++;
            pts[j] = coors[3*k+2];
            j++;
          }
        }
      }
//printf(" Freeing volume tries\n");
      free(test);
      free(coors);
      free(tris);
      free(cells);
    }
//printf(" Freeing pressure\n");
    free(press);

//printf(" Storing shock points\n");
    tstoreFX.mglFXsv[citime].npts = npts;
    tstoreFX.mglFXsv[citime].pts = pts;

  } else {

    //printf("Using stored shock surfaces\n");
    npts = tstoreFX.mglFXsv[citime].npts;
    pts = tstoreFX.mglFXsv[citime].pts;

  }

  if (npts != 0) {
    pV_SENDXR(&index, &exnum, &zero, &three, &npts, pts);
  } else {
    i = 1;
    pV_SENDXR(&index, &exnum, &zero, &three, &i, xyz);
  }
}


/* FX system call-backs */

void PREFIX
FXCELLPTR(int **tets, int **pyras, int **prisms, int **hexas, int *halo)
{
  int i, len;

  i = 311;
  pV_GETSTRUC(&i, (void **) tets,   &len);
  i = 312;
  pV_GETSTRUC(&i, (void **) pyras,  &len);
  i = 313;
  pV_GETSTRUC(&i, (void **) prisms, &len);
  i = 314;
  pV_GETSTRUC(&i, (void **) hexas,  &len);
}


void PREFIX
FXSURFACEPTR(int *nsurf, int **cell, int **facet)
{
  int i, len, *bc;

  i = 324;
  pV_GETSTRUC(&i, (void **) facet, &len);
  i = 323;
  pV_GETSTRUC(&i, (void **) cell, &len);
  i = 321;
  pV_GETSTRUC(&i, (void **) &bc,  &len);

  for (i = 0; i < len; i++) {
    nsurf[2*i  ] = bc[3*i];
    nsurf[2*i+1] = 6;
  }
}


void PREFIX
FXGRIDPTR(float **xyz, float *hxyz)
{
  int i, len;

  i = 301;
  pV_GETSTRUC(&i, (void **) xyz, &len);
}


void PREFIX
FXVELPTR(float **vel, float *hvel)
{
  int i, len;

  i = 304;
  pV_GETSTRUC(&i, (void **) vel, &len);
}


void PREFIX
FXSCAL(int *type, float *s, float *hs)
{
  int i;

  if ((*type <= 5) && (*type != 4)) {
    i = xlate[*type-1];
    pVSCAL(&i, s);
  } else {
    printf(" FXSCAL Error: type = %d\n", *type);
  }
}


/* pV3 Programmed Extract communication */

void PREFIX
pVEXTRACT(int *index, int *exnum, int *reqmask, int *ival, float *rvals)
{
  int   i, ibc, *bc, nseg, *segs, zero = 0, one = 1, three = 3;
  float *sepxyz, *attxyz, *vcoor, *stren, fone[3] = {1.0,1.0,1.0};

  if (*index == 101) {
    if (tstoreFX.mglFXvc[citime].nseg == -1) {
      //printf("Computing Vortex Cores\n");
      FX_VORTEXCORE(ival, &nseg, &segs, &vcoor, &stren);
      tstoreFX.mglFXvc[citime].nseg = nseg;
      tstoreFX.mglFXvc[citime].segs = segs;
      tstoreFX.mglFXvc[citime].vcoor = vcoor;
      tstoreFX.mglFXvc[citime].stren = stren;
    } else {
      //printf("Using stored Vortex Cores\n");
      nseg = tstoreFX.mglFXvc[citime].nseg;
      segs = tstoreFX.mglFXvc[citime].segs;
      vcoor = tstoreFX.mglFXvc[citime].vcoor;
      stren = tstoreFX.mglFXvc[citime].stren;
    }
    if (nseg != 0) {
      pV_SENDXR(index, exnum, &zero, &three, &segs[nseg-1], vcoor);
      pV_SENDXI(index, exnum, &one,  &one,   &nseg,         segs);
    } else {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXI(index, exnum, &one,  &one,   &one, &one);
    }
    //printf("***DONE WITH VORTEX CORES\n");
    return;
  }

  if (*index == 102) {
    FXpV3Shock(*index, *exnum, *ival, rvals[0]);
    //printf("***DONE WITH SHOCK SURFACES\n");
    return;
  }

  if (*index == 103) {
    if (tstoreFX.mglFXsa[citime].calc == 0) {
      //printf("Computing Sep & Att lines\n");
      FX_SEPNLINE(FXsepind, &sepxyz, FXattind, &attxyz);
      tstoreFX.mglFXsa[citime].calc = 1;
      tstoreFX.mglFXsa[citime].FXsepind = FXsepind;
      tstoreFX.mglFXsa[citime].sepxyz = sepxyz;
      tstoreFX.mglFXsa[citime].FXattind = FXattind;
      tstoreFX.mglFXsa[citime].attxyz = attxyz;
    } else {
      //printf("Using stored Sep & Att lines\n");
      FXsepind = tstoreFX.mglFXsa[citime].FXsepind;
      sepxyz = tstoreFX.mglFXsa[citime].sepxyz;
      FXattind = tstoreFX.mglFXsa[citime].FXattind;
      attxyz = tstoreFX.mglFXsa[citime].attxyz;
    }

    if (sepxyz == NULL) {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
      return;
    }
    i = 321;
    pV_GETSTRUC(&i, (void **) &bc, &ibc);
    ibc = -1;
    for (i = 0; i < FXnbc; i++)
      if (bc[3*i+2] == *ival) ibc = i;

    if (ibc == -1) {
      printf(" pVExtract: Global Surface # %d NOT Found!\n", *ival);
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
      return;
    }

    i = 0;
    if (ibc > 0) i = FXsepind[ibc-1];
    nseg = FXsepind[ibc] - i;
    if (nseg > 1) {
      pV_SENDXR(index, exnum, &zero, &three, &nseg, &sepxyz[3*i]);
    } else {
      pV_SENDXR(index, exnum, &zero, &three, &one, fone);
    }

    i = 0;
    if (ibc > 0) i = FXattind[ibc-1];
    nseg = FXattind[ibc] - i;
    if (nseg > 1) {
      pV_SENDXR(index, exnum, &one,  &three, &nseg, &attxyz[3*i]);
    } else {
      pV_SENDXR(index, exnum, &one,  &three, &one, fone);
    }
/*
    free(sepxyz);
    free(attxyz);
*/
      
    //printf("***DONE WITH SEP & ATT LINE\n");
    return;
  }

  if (*index == 104) {
    return;
  }

  if (*index == 105) {
    return;
  }

  printf(" Unknown Programmed Extract request = %d\n", *index);
}
