/*
 * NOdials.c
 *
 * Xlib interface for pV3 so that the dialbox can be replaced
 *
 * Copyright 1995 - 1999, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <time.h>
#ifdef WIIMOTE
#  include "wiiv3.h"
#endif

#define MOVEMENT 256

#ifdef UNDERSCORE

#define PV_GETSTATE  pv_getstate_

#define XFTNSDIALFUN xftnsdialfun_
#define XFTNINITDIAL xftninitdial_
#define XFTNGETDIALS xftngetdials_

#else

#define PV_GETSTATE  pv_getstate

#define XFTNSDIALFUN xftnsdialfun
#define XFTNINITDIAL xftninitdial
#define XFTNGETDIALS xftngetdials

#endif

extern void PV_GETSTATE( int *opt,  int *ivec,  float *rvec,
                         char *str, int str_len );

/* extrenal */

  extern Display *XFtnDpy;                 /* X Display */


/* globals */

  Window XFtnDWin = 0;
  int XFtnDWinT   = 0;
  int XFtnDialX   = 0;
  int XFtnDialY   = 0;
  int XFtnDOutlin = 0;
  clock_t last_tyme  = 0;

/*
 * This routine allow control of the 2D and 3D window from cursor movements
 *
 *   This should be called from V3PEvents after a middle mouse click and
 *     the event should be cancled (IWIN = 0).  active should be set to 1.
 *
 *   outline controls the action:
 *     if 0 the windows are allowed to re-render between movements
 *       and this routine must be called (active = 0) to terminate this
 *       mode.
 *     if 1 then outline mode is enforced and no re-rendering is allowed
 *       until after the middle mouse button is released.
 */

void
XFTNSDIALFUN(Window *win, int *x, int *y, int *active, int *outline)
{
  if (*active != 0) {
    XFtnDWin    = *win;
    XFtnDWinT   = *active;
    XFtnDialX   = *x;
    XFtnDialY   = *y;
    XFtnDOutlin = *outline;
  } else {
    XFtnDWin  = 0;
    XFtnDWinT = 0;
  }
}


int
XFTNINITDIAL()
{
  int k, ivec[2];
  float rvec[2];
  char cstr[80];

  k = 0;
  PV_GETSTATE(&k, ivec, rvec, cstr, 80);
  if (rvec[0] > 1.20) return -MOVEMENT;
  return 0;
}


int
XFTNGETDIALS(int drev[])
{
  Window root, child;
  XWindowAttributes xwa;
  int k, ret, x, y, root_x, root_y, ivec[2];
  unsigned int key_buttons;
  float rvec[8], theta1, theta2, xf, yf;
  char cstr[80];
  clock_t	    tmp_tyme;

  ret = 0;
  for (k = 0; k < 8; k++) drev[k] = 0;

#ifdef WIIMOTE
/* Process Wiimote first so that mousemode for wiimote can use Bob's stuff...*/
struct WIIStatus WV3Status;

  //flipflop = flipflop + 1;
  WV3Status.update = 0;

  /*if (flipflop / 3 == flipflop/3.0) {
     WV3Status = WMGetState();
  }*/
  WV3Status = WMGetState();
  

  tmp_tyme=clock();

  //printf("WV3Status.update %d\n",WV3Status.update);

  if (WV3Status.update != 0) {
      //printf("mode is: %d\n",WV3Status.mode);
      drev[0] = WV3Status.dails[0];
      drev[1] = WV3Status.dails[1];
      drev[2] = WV3Status.dails[2];
      drev[3] = WV3Status.dails[3];
      drev[4] = WV3Status.dails[4];
      drev[5] = WV3Status.dails[5];
      drev[6] = WV3Status.dails[6];
      drev[7] = WV3Status.dails[7];

    for (k = 0; k < 8; k++) if (drev[k] != 0) ret = 1;
    if (XFtnDOutlin != 0) ret = 1;
    if (XFtnDOutlin == 0) ret = -ret;
    /*if (flipflop / 3 != flipflop/3.0) ret = 0;
    if (flipflop >= 250) flipflop = 1;*/
    if (tmp_tyme - last_tyme <= 50 ) ret = 0;
    drev[8] = XFtnDWinT;

    if (ret != 0) {
	last_tyme = tmp_tyme;
	//printf("dial5: %d\n",WV3Status.dails[5]);
    }

  }

#endif

/* process Bob's stuff */

  /* we have been activated */
  if (XFtnDWin != 0) {

    /* get current state of cursor and special keys */
    XQueryPointer(XFtnDpy, XFtnDWin, &root, &child, &root_x, &root_y,
                  &x, &y, &key_buttons);

    /* has the middle mouse been released? */
    if ((key_buttons&0x200) == 0) {
      XFtnDWin  = 0;
      XFtnDWinT = 0;
      return ret;
    }

    /* get the current size of the window and the active tool */
    XGetWindowAttributes(XFtnDpy, XFtnDWin, &xwa);
    k = 3;
    PV_GETSTATE(&k, ivec, rvec, cstr, 80);

    /* 2D Window */
    if (XFtnDWinT == 2) {
      k = 4;
      if (ivec[0] != 2) k = 8;
      if ((key_buttons&0xf) == 0) {
        /* no special keys */
        drev[2] = k*(XFtnDialX-x)*MOVEMENT/xwa.width;
        drev[6] = k*(XFtnDialY-y)*MOVEMENT/xwa.height;
        if ((ivec[0] != 2) && (rvec[3] < 0.0)) drev[6] = -drev[6];
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(XFtnDialY-y)*MOVEMENT/xwa.height;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/xwa.height;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - xwa.width/2;
        yf = XFtnDialY - xwa.height/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - xwa.width/2;
          yf = y - xwa.height/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            k = 2*MOVEMENT;
            if ((ivec[0] != 2) && (rvec[3] > 0.0)) k = -k;
            drev[4] = k*(theta2-theta1)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* 3D Window */
    if (XFtnDWinT == 3) {
      if ((key_buttons&0xf) == 0) {
        /* no special keys */
        drev[2] = 8*(x-XFtnDialX)*MOVEMENT/xwa.width;
        drev[6] = 8*(y-XFtnDialY)*MOVEMENT/xwa.height;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(y-XFtnDialY)*MOVEMENT/xwa.height;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 4) {
        /* ctrl is active */
        drev[1] = 4*(x-XFtnDialX)*MOVEMENT/xwa.width;
        drev[0] = 4*(y-XFtnDialY)*MOVEMENT/xwa.height;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/xwa.height;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - xwa.width/2;
        yf = XFtnDialY - xwa.height/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - xwa.width/2;
          yf = y - xwa.height/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            drev[4] = 2*MOVEMENT*(theta1-theta2)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* set return code:
     *   outline on -- stay in loop
     *   outline off -- allow to fall through if no input
     */

    for (k = 0; k < 8; k++) if (drev[k] != 0) ret = 1;
    if (XFtnDOutlin != 0) ret = 1;
    if (XFtnDOutlin == 0) ret = -ret;
    drev[8] = XFtnDWinT;
  }

  return ret;
}


void
xftnDialEvent()
{
}


void
xftnSetDialEvent(Window window)
{
}
