/*
 * NOdials.c
 *
 * Xlib interface for Visual3 so that the dialbox can be replaced
 *
 * Copyright 1995 - 2006, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include <unistd.h>
#include "wiiv3.h"


#define MOVEMENT 256

#ifdef UNDERSCORE

#ifdef G77
#define V3_GETSTATE  v3_getstate__
#define V3_SLEEP     v3_sleep__
#else
#define V3_GETSTATE  v3_getstate_
#define V3_SLEEP     v3_sleep_
#endif

#define XFTNSDIALFUN  xftnsdialfun_
#define XFTNINITDIAL  xftninitdial_
#define XFTNGETDIALS  xftngetdials_
#define XFTNGETMOTION xftngetmotion_

#else

#define V3_GETSTATE  v3_getstate
#define V3_SLEEP     v3_sleep

#define XFTNSDIALFUN  xftnsdialfun
#define XFTNINITDIAL  xftninitdial
#define XFTNGETDIALS  xftngetdials
#define XFTNGETMOTION xftngetmotion

#endif

extern int  xftnGetMotion( );
extern void V3_GETSTATE( int *opt,  long *ivec,  float *rvec,
                         char *str, int str_len );

/* extrenal */

  extern Display *XFtnDpy;                 /* X Display */
  extern int      XFtn3Dtype;              /* handedness of the display */


/* globals */

  Window XFtnDWin = 0;
  int XFtnDWinT   = 0;
  int XFtnDialX   = 0;
  int XFtnDialY   = 0;
  int XFtnDOutlin = 0;
  int flipflop    = 1;
  int motpause    = 1;

/*
 * This routine sleeps for the specified seconds
 */

void
V3_SLEEP(float *time)
{
  int sec;

  sec = *time + 0.000001;
  if (sec > 0) {
    sleep(sec);
  } else {
    sec = *time*1000000;
    usleep(sec);
  }
}


/*
 * This routine allow control of the 2D and 3D window from cursor movements
 *
 *   This should be called from V3PEvents after a middle mouse click and
 *     the event should be cancled (IWIN = 0).  active should be set to 1.
 *
 *   outline controls the action:
 *     if 0 the windows are allowed to re-render between movements
 *       and this routine must be called (active = 0) to terminate this
 *       mode.
 *     if 1 then outline mode is enforced and no re-rendering is allowed
 *       until after the middle mouse button is released.
 */

void
XFTNSDIALFUN(Window *win, int *x, int *y, int *active, int *outline)
{
  if (*active != 0) {
    XFtnDWin    = *win;
    XFtnDWinT   = *active;
    XFtnDialX   = *x;
    XFtnDialY   = *y;
    XFtnDOutlin = *outline;
  } else {
    XFtnDWin  = 0;
    XFtnDWinT = 0;
  }
}


int
XFTNINITDIAL()
{
  int k;
  long ivec[8];
  float rvec[2];
  char cstr[80];

  k = 0;
  V3_GETSTATE(&k, ivec, rvec, cstr, 80);
  if (rvec[0] > 2.285) return -MOVEMENT;
  return 0;
}


int
XFTNGETDIALS(int drev[])
{
  Window            root, child;
  XWindowAttributes xwa;
  int               k, ret, x, y, root_x, root_y, min_size;
  unsigned int      key_buttons;
  long              ivec[2];
  float             rvec[7], theta1, theta2, xf, yf;
  char              cstr[80];

  ret = 0;
  for (k = 0; k < 8; k++) drev[k] = 0;

/* Process Wiimote first so that mousemode for wiimote can use Bob's stuff...*/

struct WIIStatus WV3Status;
  WV3Status = WMGetState();  

  if (WV3Status.update != 0) {
      //printf("mode is: %d\n",WV3Status.mode);
      drev[0] = WV3Status.dails[0];
      drev[1] = WV3Status.dails[1];
      drev[2] = WV3Status.dails[2];
      drev[3] = WV3Status.dails[3];
      drev[4] = WV3Status.dails[4];
      drev[5] = WV3Status.dails[5];
      drev[6] = WV3Status.dails[6];
      drev[7] = WV3Status.dails[7];

      flipflop = flipflop + 1;

    for (k = 0; k < 8; k++) if (drev[k] != 0) ret = 1;
    if (XFtnDOutlin != 0) ret = 1;
    if (XFtnDOutlin == 0) ret = -ret;
    if (flipflop / 5 != flipflop/5.0) ret = 0;
    drev[8] = XFtnDWinT;

  }

/* process Bob's stuff */

  /* we have been activated */
  if (XFtnDWin != 0) {

    /* get current state of cursor and special keys */
    XQueryPointer(XFtnDpy, XFtnDWin, &root, &child, &root_x, &root_y,
                  &x, &y, &key_buttons);

    /* has the middle mouse been released? */
    if ((key_buttons&0x200) == 0) {
      XFtnDWin  = 0;
      XFtnDWinT = 0;
      return ret;
    }

    /* get the current size of the window and the active tool */
    XGetWindowAttributes(XFtnDpy, XFtnDWin, &xwa);
    min_size = xwa.width;
    if (xwa.width > xwa.height) min_size = xwa.height;
    k = 3;
    ivec[0] = -1;
    V3_GETSTATE(&k, ivec, rvec, cstr, 80);

    /* 2D Window */
    if (XFtnDWinT == 2) {
      k = 4;
      if (ivec[0] != 2) k = 8;
      if ((key_buttons&0xf) == 0) {
        /* no special keys */
        drev[2] = k*(XFtnDialX-x)*MOVEMENT/min_size;
        drev[6] = k*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - xwa.width/2;
        yf = XFtnDialY - xwa.height/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - xwa.width/2;
          yf = y - xwa.height/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            k = 2*MOVEMENT;
            if (ivec[0] != 2) k = -k;
            drev[4] = k*(theta2-theta1)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* 3D Window */
    if (XFtnDWinT == 3) {
      if ((key_buttons&0xf) == 0) {
        /* no special keys */
        drev[2] = 8*(x-XFtnDialX)*MOVEMENT/min_size;
        drev[6] = 8*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 4) {
        /* ctrl is active */
        drev[1] = 4*(x-XFtnDialX)*MOVEMENT/min_size;
        drev[0] = 4*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
        if (XFtn3Dtype <= -3) {
          drev[0] = -drev[0];
          drev[1] = -drev[1];
        }
      } else if ((key_buttons&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key_buttons&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - xwa.width/2;
        yf = XFtnDialY - xwa.height/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - xwa.width/2;
          yf = y - xwa.height/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            drev[4] = 2*MOVEMENT*(theta1-theta2)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* set return code:
     *   outline on -- stay in loop
     *   outline off -- allow to fall through if no input
     */

    for (k = 0; k < 8; k++) if (drev[k] != 0) ret = 1;
    if (XFtnDOutlin != 0) ret = 1;
    if (XFtnDOutlin == 0) ret = -ret;
    drev[8] = XFtnDWinT;
  }

  return ret;
}


int
XFTNGETMOTION()
{
  return xftnGetMotion();
}


void
xftnDialEvent()
{
}


void
xftnSetDialEvent(Window window)
{
}
