/*
 * NOdials.c
 *
 * WindowsNT interface for Visual3 so that the dialbox can be replaced
 *
 * Copyright 1995 - 2005, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <windows.h>


#ifdef ABSOFT

#define PREFIX __cdecl
#define V3_GETSTATE   v3_getstate
#define V3_SLEEP      v3_sleep
#define XFTNINITDIAL  xftninitdial
#define XFTNGETDIALS  xftngetdials
#define XFTNSDIALFUN  xftnsdialfun
#define XFTNGETMOTION xftngetmotion

#else

#define PREFIX __stdcall

#endif

#define MOVEMENT 256
#define NOGLW 5


/* extrenals */

  extern HWND XFtnGLWin[NOGLW];
  extern int  XFtnNGLWin;
  extern int  XFtnSize[NOGLW][5];

  extern int  xftnGetMotion( );
  extern void PREFIX V3_GETSTATE( int *opt,  int *ivec,  float *rvec,
                                  char *str, int str_len );


/* globals */

  HWND XFtnDWin    = NULL;
  int  XFtnDWinT   = 0;
  int  XFtnDialX   = 0;
  int  XFtnDialY   = 0;
  int  XFtnDOutlin = 0;



int 
xftnCheckMessage(MSG msg)
{
  return 1;
}


/*
 * This routine sleeps for the specified seconds
 */

void PREFIX
V3_SLEEP(float *time)
{
  int msec;

  msec = *time*1000. + 0.45;
  Sleep(msec);
}


/*
 * This routine allow control of the 2D and 3D window from cursor movements
 *
 *   This should be called from V3PEvents after a middle mouse click and
 *     the event should be cancled (IWIN = 0).  active should be set to 1.
 *
 *   outline controls the action:
 *     if 0 the windows are allowed to re-render between movements
 *       and this routine must be called (active = 0) to terminate this
 *       mode.
 *     if 1 then outline mode is enforced and no re-rendering is allowed
 *       until after the middle mouse button is released.
 */

void PREFIX
XFTNSDIALFUN(HWND *win, int *x, int *y, int *active, int *outline)
{
  if (*active != 0) {
    XFtnDWin    = *win;
    XFtnDWinT   = *active;
    XFtnDialX   = *x;
    XFtnDialY   = *y;
    XFtnDOutlin = *outline;
  } else {
    XFtnDWin  = NULL;
    XFtnDWinT = 0;
  }
}


int PREFIX
XFTNINITDIAL()
{
  int   k;
  int   ivec[8];
  float rvec[2];
  char  cstr[80];

  k = 0;
  V3_GETSTATE(&k, ivec, rvec, cstr, 80);
  if (rvec[0] > 2.285) return -MOVEMENT;
  return 0;
}


int PREFIX
XFTNGETDIALS(int *drev)
{
  int   k, ret, x, y, min_size, key, iwin;
  int   ivec[2];
  float rvec[7], theta1, theta2, xf, yf;
  char  cstr[80];
  POINT point;

  ret = 0;
  for (k = 0; k < 8; k++) drev[k] = 0;

  /* we have been activated */
  if (XFtnDWin != NULL) {

    for (iwin = k = 0; k < XFtnNGLWin; k++) 
      if (XFtnDWin == XFtnGLWin[k]) iwin = k;

    /* get current state of cursor and special keys */
    GetCursorPos(&point);
    key = 0;
    if  (GetAsyncKeyState(VK_SHIFT)   < 0)  key += 1;
    if  (GetAsyncKeyState(VK_CONTROL) < 0)  key += 4;
    if  (GetAsyncKeyState(VK_MENU)    < 0)  key += 8;
    if  (GetAsyncKeyState(VK_LBUTTON) < 0)  key += 256;
    if  (GetAsyncKeyState(VK_MBUTTON) < 0)  key += 512;
    if  (GetAsyncKeyState(VK_RBUTTON) < 0)  key += 1024;
    if ((GetAsyncKeyState(VK_LBUTTON) < 0) &&
        (GetAsyncKeyState(VK_RBUTTON) < 0)) key -= 768;
    x = point.x - XFtnSize[iwin][2];
    y = point.y - XFtnSize[iwin][3];

    /* has the middle mouse been released? */
    if ((key&0x200) == 0) {
      XFtnDWin  = NULL;
      XFtnDWinT = 0;
      return ret;
    }

    /* get the current size of the window and the active tool */
    min_size = XFtnSize[iwin][0];
    if (XFtnSize[iwin][0] > XFtnSize[iwin][1]) min_size = XFtnSize[iwin][1];
    k = 3;
    ivec[0] = -1;
    V3_GETSTATE(&k, ivec, rvec, cstr, 80);

    /* 2D Window */
    if (XFtnDWinT == 2) {
      k = 4;
      if (ivec[0] != 2) k = 8;
      if ((key&0xf) == 0) {
        /* no special keys */
        drev[2] = k*(XFtnDialX-x)*MOVEMENT/min_size;
        drev[6] = k*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - XFtnSize[iwin][0]/2;
        yf = XFtnDialY - XFtnSize[iwin][1]/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - XFtnSize[iwin][0]/2;
          yf = y - XFtnSize[iwin][1]/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            k = 2*MOVEMENT;
            if (ivec[0] != 2) k = -k;
            drev[4] = k*(theta2-theta1)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* 3D Window */
    if (XFtnDWinT == 3) {
      if ((key&0xf) == 0) {
        /* no special keys */
        drev[2] = 8*(x-XFtnDialX)*MOVEMENT/min_size;
        drev[6] = 8*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key&0xf) == 1) {
        /* shift is down */
        drev[3] = 2*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key&0xf) == 4) {
        /* ctrl is active */
        drev[1] = 4*(x-XFtnDialX)*MOVEMENT/min_size;
        drev[0] = 4*(y-XFtnDialY)*MOVEMENT/min_size;
        XFtnDialX = x;
        XFtnDialY = y;
      } else if ((key&0xf) == 5) {
        /* shift&ctrl is down */
        drev[5] = 10*(XFtnDialY-y)*MOVEMENT/min_size;
        XFtnDialY = y;
      } else if ((key&0xf) == 8) {
        /* alt is pressed */
        xf = XFtnDialX - XFtnSize[iwin][0]/2;
        yf = XFtnDialY - XFtnSize[iwin][1]/2;
        if ((xf != 0.0) || (yf != 0.0)) {
          theta1 = atan2(yf, xf);
          xf = x - XFtnSize[iwin][0]/2;
          yf = y - XFtnSize[iwin][1]/2;
          if ((xf != 0.0) || (yf != 0.0)) {
            theta2 = atan2(yf, xf);
            drev[4] = 2*MOVEMENT*(theta1-theta2)/3.1415926;
          }
        }
        XFtnDialX = x;
        XFtnDialY = y;
      }
    }

    /* set return code:
     *   outline on -- stay in loop
     *   outline off -- allow to fall through if no input
     */

    for (k = 0; k < 8; k++) if (drev[k] != 0) ret = 1;
    if (XFtnDOutlin != 0) ret = 1;
    if (XFtnDOutlin == 0) ret = -ret;
    drev[8] = XFtnDWinT;
  }

  return ret;
}


int PREFIX
XFTNGETMOTION()
{
  return xftnGetMotion();
}
