#include "parameters.h"
#define GRID pgr(igrid)%
!#######################################################################
INTEGER FUNCTION ijk2nod(i,j,k,mbb,msp)
!#######################################################################
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: i,j,k,mbb,msp

  ijk2nod=i+(j-1)*mbb+(k-1)*mbb*msp

END FUNCTION ijk2nod

!#######################################################################
INTEGER FUNCTION ijk2nodm(i,j,k,il,jl,bn)
!#######################################################################
  USE common_area, ONLY: pgr, igrid
  IMPLICIT NONE
  INTEGER,INTENT(IN) :: i,j,k,il,jl,bn
  INTEGER :: s,q

  s = 0
  DO q = 1, bn-1, 1
    s = s + GRID mvmax(q)
  ENDDO
  ijk2nodm = i + (j-1)*il + (k-1)*il*jl + s

END FUNCTION ijk2nodm

!#######################################################################
INTEGER FUNCTION ijk2cell(i,j,k,bn)
!#######################################################################
  USE common_area, ONLY: pgr, igrid
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: i,j,k, bn
  INTEGER :: mfl,mbb,msp

  mbb = GRID li(bn)
  msp = GRID lj(bn)
  mfl = GRID lk(bn)
  ijk2cell=i+(j-1)*(mbb-1)+(k-1)*(mbb-1)*(msp-1)

END FUNCTION ijk2cell

!#######################################################################
INTEGER FUNCTION ijk2cellm(i,j,k,bn)
!#######################################################################
  USE common_area, ONLY: pgr, igrid
  IMPLICIT NONE
  INTEGER, INTENT(IN) :: i,j,k, bn
  INTEGER :: mfl,mbb,msp
  INTEGER :: base, q

  base = 0
  DO q = 1, bn-1, 1
    base = base + (GRID li(q)-1)*(GRID lj(q)-1)*(GRID lk(q)-1)
  ENDDO

  mbb = GRID li(bn)
  msp = GRID lj(bn)
  mfl = GRID lk(bn)
!  IF(i.GE.GRID li(bn))PRINT*,'ERROR in ijk2cellm -i- ',i,j,k,bn
!  IF(j.GE.GRID lj(bn))PRINT*,'ERROR in ijk2cellm -j- ',i,j,k,bn
!  IF(k.GE.GRID lk(bn))PRINT*,'ERROR in ijk2cellm -k- ',i,j,k,bn
  ijk2cellm=base + i+(j-1)*(mbb-1)+(k-1)*(mbb-1)*(msp-1)

END FUNCTION ijk2cellm

!#######################################################################
SUBROUTINE IJKFaceMatching(i1,j1,k1,b1,p1, i2,j2,k2)
!#######################################################################
! Minimum distance point matching for a given face.
  USE common_area, ONLY: pgr, igrid, coordsys
  USE solver_turbo, ONLY: patches
  IMPLICIT NONE
  INTEGER :: ijk2nodm

  INTEGER,INTENT(IN) :: i1,j1,k1,b1,p1
  INTEGER,INTENT(OUT) :: i2,j2,k2

  INTEGER :: b2,p2, i,j,k, nn
  REAL :: xcur,ycur,zcur, del, mindel
  REAL,POINTER,DIMENSION(:) :: xp, yp, zp

  nn = ijk2nodm(i1, j1, k1, GRID li(b1), GRID lj(b1), b1)
  SELECT CASE(coordsys)
    CASE(CYLINDRICAL,RHCYLINDRICAL)
      xp => GRID r
      yp => GRID t
      zp => GRID z
    CASE DEFAULT
      xp => GRID x
      yp => GRID y
      zp => GRID z
  END SELECT

  xcur = xp(nn)
  ycur = yp(nn)
  zcur = zp(nn)

  mindel = 1.
  i2 = 1
  j2 = 1
  k2 = 1
  b2 = patches(b1)%ptch(p1,9)
  p2 = patches(b1)%ptch(p1,10)
  DO i = patches(b2)%ptch(p2,1), patches(b2)%ptch(p2,4)
    DO j = patches(b2)%ptch(p2,2), patches(b2)%ptch(p2,5)
      DO k = patches(b2)%ptch(p2,3), patches(b2)%ptch(p2,6)
        nn = ijk2nodm(i, j, k, GRID li(b2), GRID lj(b2), b2)
        del = SQRT((xcur-xp(nn))**2 + (ycur-yp(nn))**2 + (zcur-zp(nn))**2)
        IF(del.LT.mindel)THEN
          i2 = i
          j2 = j
          k2 = k
          mindel = del
        ENDIF
        IF(del.EQ.0.)RETURN
      ENDDO
    ENDDO
  ENDDO
  
END SUBROUTINE IJKFaceMatching
