#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_angles
  IMPLICIT NONE

CONTAINS
!#######################################################################
SUBROUTINE ScalAngles(v, key, mvmax)
!#######################################################################
  USE common_area, ONLY: pfw, itime
  USE mod_events, ONLY: event_get_secondary

  REAL,INTENT(INOUT) :: v(*)
  INTEGER,INTENT(IN) :: key,mvmax
  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(1)
      CALL Calc_AlphaZ()
      ptr => FLOW alphaz
    CASE(2)
      CALL Calc_AlphaM()
      ptr => FLOW alpham
    CASE(3)
      CALL Calc_BetaZ()
      ptr => FLOW betaz
    CASE(4)
      CALL Calc_BetaM()
      ptr => FLOW betam
    CASE(5)
      CALL Calc_Phi()
      ptr => FLOW phi
    CASE DEFAULT
      RETURN
  END SELECT

  DO k = 1, mvmax
    v(k) = ptr(k)
  ENDDO

END SUBROUTINE ScalAngles

!#######################################################################
SUBROUTINE Calc_AlphaZ()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL todeg
  DATA todeg/57.29578/
  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts

  IF(.NOT. ASSOCIATED(FLOW alphaz))THEN
    ALLOCATE(FLOW alphaz(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          IF(FLOW vx(k).EQ.0 .AND. FLOW vy(k).EQ.0)THEN
            FLOW alphaz(k) = 0.
          ELSE
            FLOW alphaz(k) = todeg*ATAN2(FLOW vy(k),FLOW vx(k))
          ENDIF
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          IF(FLOW vz(k).EQ.0 .AND. FLOW vt(k).EQ.0)THEN
            FLOW alphaz(k) = 0.
          ELSE
            FLOW alphaz(k) = todeg*ATAN2(FLOW vt(k),FLOW vz(k))
          ENDIF
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW alphaz

  RETURN

END SUBROUTINE Calc_AlphaZ

!#######################################################################
SUBROUTINE Calc_AlphaM()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL todeg,vm
  DATA todeg/57.29578/
  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts

  IF(.NOT. ASSOCIATED(FLOW alpham))THEN
    ALLOCATE(FLOW alpham(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          vm = SQRT(FLOW vy(k)**2 + FLOW vx(k)**2)
          IF(vm.EQ.0 .AND. FLOW vz(k).EQ.0)THEN
            FLOW alpham(k) = 0.
          ELSE
            FLOW alpham(k) = todeg*ATAN2(FLOW vz(k),vm)
          ENDIF
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          vm = SQRT(FLOW vr(k)**2 + FLOW vz(k)**2)
          IF(vm.EQ.0 .AND. FLOW vt(k).EQ.0)THEN
            FLOW alpham(k) = 0.
          ELSE
            FLOW alpham(k) = todeg*ATAN2(FLOW vt(k),vm)
          ENDIF
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW alpham

  RETURN

END SUBROUTINE Calc_AlphaM

!#######################################################################
SUBROUTINE Calc_BetaZ()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL todeg,wt
  DATA todeg/57.29578/
  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts

  IF(.NOT. ASSOCIATED(FLOW betaz))THEN
    ALLOCATE(FLOW betaz(mvmax))
    SELECT CASE(coordsys)
      CASE(0)
        CALL Calc_AlphaZ
        FLOW betaz => FLOW alphaz
        NULLIFY(FLOW alphaz)
      CASE DEFAULT
        DO k=1,mvmax
          wt = FLOW vt(k) - omega*pgr(igrid)%r(k)
          IF(FLOW vz(k).EQ.0 .AND. wt.EQ.0)THEN
            FLOW betaz(k) = 0.
          ELSE
            FLOW betaz(k) = todeg*ATAN2(wt,FLOW vz(k))
          ENDIF
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW betaz

  RETURN

END SUBROUTINE Calc_BetaZ

!#######################################################################
  SUBROUTINE Calc_BetaM()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL todeg,wt,vm
  DATA todeg/57.29578/
  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts

  IF(.NOT. ASSOCIATED(FLOW betam))THEN
    ALLOCATE(FLOW betam(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_AlphaM
        FLOW betam => FLOW alpham
        NULLIFY(FLOW alpham)
      CASE DEFAULT
        DO k=1,mvmax
          vm = SQRT(FLOW vr(k)**2 + FLOW vz(k)**2)
          wt = FLOW vt(k) - omega*pgr(igrid)%r(k)
          IF(vm.EQ.0 .AND. wt.EQ.0)THEN
            FLOW betam(k) = 0.
          ELSE
            FLOW betam(k) = todeg*ATAN2(wt,vm)
          ENDIF
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW betam

  END SUBROUTINE Calc_BetaM

!#######################################################################
  SUBROUTINE Calc_Phi()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL todeg
  DATA todeg/57.29578/
  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts

  IF(.NOT. ASSOCIATED(FLOW phi))THEN
    ALLOCATE(FLOW phi(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          IF(FLOW vx(k).EQ.0 .AND. FLOW vz(k).EQ.0)THEN
            FLOW phi(k) = 0.
          ELSE
            FLOW phi(k) = todeg*ATAN2(FLOW vz(k),FLOW vx(k))
          ENDIF
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          IF(FLOW vz(k).EQ.0 .AND. FLOW vr(k).EQ.0)THEN
            FLOW phi(k) = 0.
          ELSE
            FLOW phi(k) = todeg*ATAN2(FLOW vr(k),FLOW vz(k))
          ENDIF
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW phi

  END SUBROUTINE Calc_Phi
END MODULE calc_angles
