#define PI 3.141592653589793
MODULE calc_coord
  IMPLICIT NONE

CONTAINS
!#######################################################################
SUBROUTINE ScalCoord(v, key, mvmax)
!#######################################################################
  USE common_area, ONLY: pgr, igrid
  USE mod_events, ONLY: event_get_secondary

  REAL,INTENT(INOUT) :: v(*)
  INTEGER,INTENT(IN) :: key, mvmax

  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(1)
      CALL Calc_CoordI()
      ptr => pgr(igrid)%i
    CASE(2)
      CALL Calc_CoordJ()
      ptr => pgr(igrid)%j
    CASE(3)
      CALL Calc_CoordK()
      ptr => pgr(igrid)%k
    CASE(4)
      CALL Calc_CoordR()
      ptr => pgr(igrid)%r
    CASE(5)
      CALL Calc_CoordT()
      ptr => pgr(igrid)%t
    CASE(6)
      CALL Calc_CoordX()
      ptr => pgr(igrid)%x
    CASE(7)
      CALL Calc_CoordY()
      ptr => pgr(igrid)%y
    CASE(8)
      CALL Calc_CoordZ()
      ptr => pgr(igrid)%z
  END SELECT

  DO k=1,mvmax
      v(k) = ptr(k)
  ENDDO

END SUBROUTINE ScalCoord

!#######################################################################
SUBROUTINE Calc_CoordI()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, cur, itime
  IMPLICIT NONE

  INTEGER i,j,k,m,n,s,mbb,msp,mfl,mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%i))THEN
    ALLOCATE(pgr(igrid)%i(mvmax))
    s = 0
    DO n = 1, pgr(igrid)%nblks
    mbb=pgr(igrid)%li(n)
    msp=pgr(igrid)%lj(n)
    mfl=pgr(igrid)%lk(n)
    DO k=1,mfl
      DO j=1,msp
        DO i=1,mbb
          m=i+(j-1)*mbb+(k-1)*msp*mbb + s
          pgr(igrid)%i(m)=i+pgr(igrid)%gis(n)
        ENDDO
      ENDDO
    ENDDO
    s = s + pgr(igrid)%mvmax(n)
    ENDDO
  ENDIF

  cur(itime)%v => pgr(igrid)%i

  RETURN

END SUBROUTINE Calc_CoordI

!#######################################################################
SUBROUTINE Calc_CoordJ()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER i,j,k,m,n,s,mbb,msp,mfl,mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%j))THEN
    ALLOCATE(pgr(igrid)%j(mvmax))
    s = 0
    DO n = 1, pgr(igrid)%nblks
    mbb=pgr(igrid)%li(n)
    msp=pgr(igrid)%lj(n)
    mfl=pgr(igrid)%lk(n)
    DO k=1,mfl
      DO j=1,msp
        DO i=1,mbb
          m=i+(j-1)*mbb+(k-1)*msp*mbb + s
          pgr(igrid)%j(m)=j+pgr(igrid)%gjs(n)
        ENDDO
      ENDDO
    ENDDO
    s = s + pgr(igrid)%mvmax(n)
    ENDDO
  ENDIF

  cur(itime)%v => pgr(igrid)%j

  RETURN

END SUBROUTINE Calc_CoordJ

!#######################################################################
SUBROUTINE Calc_CoordK()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER i,j,k,m,n,s,mbb,msp,mfl,mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%k))THEN
    ALLOCATE(pgr(igrid)%k(mvmax))
    s = 0
    DO n = 1, pgr(igrid)%nblks
      mbb=pgr(igrid)%li(n)
      msp=pgr(igrid)%lj(n)
      mfl=pgr(igrid)%lk(n)
      DO k=1,mfl
        DO j=1,msp
          DO i=1,mbb
          m=i+(j-1)*mbb+(k-1)*msp*mbb + s
            pgr(igrid)%k(m)=k+pgr(igrid)%gks(n)
          ENDDO
        ENDDO
      ENDDO
      s = s + pgr(igrid)%mvmax(n)
    ENDDO
  ENDIF

  cur(itime)%v => pgr(igrid)%k

  RETURN

END SUBROUTINE Calc_CoordK

!#######################################################################
SUBROUTINE Calc_CoordR()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER m, mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%r))THEN
    ALLOCATE(pgr(igrid)%r(mvmax))
    DO m = 1, mvmax
      pgr(igrid)%r(m) = SQRT(pgr(igrid)%y(m)**2 + pgr(igrid)%z(m)**2)
    ENDDO 
  ENDIF

  cur(itime)%v => pgr(igrid)%r

  RETURN

END SUBROUTINE Calc_CoordR

!#######################################################################
SUBROUTINE Calc_CoordT()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER m, mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%t0))THEN
    ALLOCATE(pgr(igrid)%t0(mvmax))
    DO m = 1, mvmax
      pgr(igrid)%t0(m) = ATAN2(pgr(igrid)%y(m),-pgr(igrid)%z(m))
      DO WHILE(pgr(igrid)%t0(m).LT.0)
        pgr(igrid)%t0(m)=pgr(igrid)%t0(m) + 2* PI
      ENDDO
    ENDDO 
  ENDIF

  cur(itime)%v => pgr(igrid)%t0

  RETURN

END SUBROUTINE Calc_CoordT

!#######################################################################
SUBROUTINE Calc_CoordX()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%x))THEN
    ALLOCATE(pgr(igrid)%x(mvmax))
    DO k=1,mvmax
      pgr(igrid)%x(k)=pgr(igrid)%r(k)*SIN(pgr(igrid)%t(k))
    ENDDO
  ENDIF

  cur(itime)%v => pgr(igrid)%x

  RETURN

END SUBROUTINE Calc_CoordX

!#######################################################################
SUBROUTINE Calc_CoordY()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, cur, itime
  IMPLICIT NONE

  INTEGER k, mvmax

  mvmax=pgr(igrid)%tpts
  IF(.NOT. ASSOCIATED(pgr(igrid)%y))THEN
    ALLOCATE(pgr(igrid)%y(mvmax))
    DO k=1,mvmax
      pgr(igrid)%y(k)=pgr(igrid)%r(k)*COS(pgr(igrid)%t(k))
    ENDDO
  ENDIF

  cur(itime)%v => pgr(igrid)%y

  RETURN

END SUBROUTINE Calc_CoordY

!#######################################################################
SUBROUTINE Calc_CoordZ()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, cur, itime
  IMPLICIT NONE

  IF(.NOT. ASSOCIATED(pgr(igrid)%z))THEN
  ENDIF

  cur(itime)%v => pgr(igrid)%z

  RETURN

END SUBROUTINE Calc_Coordz
END MODULE calc_coord
