#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_enth
  IMPLICIT NONE
CONTAINS
!#######################################################################
  SUBROUTINE ScalEnthalpy(v, key, mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime
    USE mod_events, ONLY: event_get_secondary

    REAL,INTENT(INOUT) :: v(*)
    INTEGER,INTENT(IN) :: key, mvmax
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(2)
        CALL Calc_EnthalpyTAbs
        ptr => FLOW htabs
      CASE(3)
        CALL Calc_EnthalpyTRel
        ptr => FLOW htrel
      CASE(1)
        CALL Calc_Enthalpy
        ptr => FLOW h
      CASE DEFAULT
        RETURN
    END SELECT
    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)

  END SUBROUTINE ScalEnthalpy

!#######################################################################
  SUBROUTINE Calc_Enthalpy()
!#######################################################################
    USE common_area, ONLY: pfw, itime, pgr, igrid, cur, coordsys, vref
    USE calc_ener, ONLY: Calc_Energy
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_dens, ONLY: Calc_Density

    INTEGER k,mvmax
    REAL vsq

    IF(.NOT. ASSOCIATED(FLOW h))THEN
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density

      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW h(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            vsq=FLOW vx(k)**2+FLOW vy(k)**2+FLOW vz(k)**2
            FLOW h(k)=FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
            vsq=FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            FLOW h(k)=(FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)) * vref**2
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW h

  END SUBROUTINE Calc_Enthalpy

!#######################################################################
  SUBROUTINE Calc_EnthalpyTAbs()
!#######################################################################
    USE common_area
    USE calc_ener, ONLY: Calc_Energy
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_dens, ONLY: Calc_Density

    INTEGER k,mvmax

    IF(.NOT. ASSOCIATED(FLOW htabs))THEN
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density

      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW htabs(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            FLOW htabs(k)=FLOW e(k)+FLOW p(k)/FLOW r(k)
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
            FLOW htabs(k)=FLOW e(k)+FLOW p(k)/FLOW r(k)+ &
                            TKEinE*FLOW k(k)
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW htabs

  END SUBROUTINE Calc_EnthalpyTAbs

!#######################################################################
  SUBROUTINE Calc_EnthalpyTRel()
!#######################################################################
    USE common_area

    INTEGER k,mvmax
    REAL wt,vsq,wsq

    IF(.NOT. ASSOCIATED(FLOW htrel))THEN
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          CALL Calc_EnthalpyTAbs
          FLOW htrel => FLOW htabs
          NULLIFY(FLOW htabs)
        CASE DEFAULT
          mvmax=pgr(igrid)%tpts
          ALLOCATE(FLOW htrel(mvmax))
          DO k=1,mvmax
            wt=FLOW vt(k)-omega*pgr(igrid)%r(k)
            wsq=wt**2+FLOW vr(k)**2+FLOW vz(k)**2
            vsq=FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            FLOW htrel(k)=FLOW e(k)-vsq/2.+ &
                            FLOW p(k)/FLOW r(k)-vsq/2.-wsq/2.
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW htrel

  END SUBROUTINE Calc_EnthalpyTRel

END MODULE calc_enth
