#include "parameters.h"
#define FLOW pfw(itime)%
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE calc_entr
  IMPLICIT NONE
CONTAINS
!#######################################################################
  SUBROUTINE ScalEntropy(v, key, mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime
    USE mod_events, ONLY: event_get_secondary

    INTEGER,INTENT(IN) :: key,mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(2)
        CALL Calc_EntropyRatio()
        ptr => FLOW srat
      CASE(1)
        CALL Calc_EntropyAbs()
        ptr => FLOW sabs
      CASE DEFAULT
        RETURN
    END SELECT

    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO

  END SUBROUTINE ScalEntropy

!#######################################################################
  SUBROUTINE Calc_EntropyAbs()
!#######################################################################
! Only works for vstage rgas=constant
    USE common_area
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_gas,  ONLY: Calc_Gamma
    USE calc_ener, ONLY: Calc_Energy

    INTEGER :: k, mvmax
    REAL vsq,hstat,cp,cpb,cp1

    mvmax=pgr(igrid)%tpts

    IF(.NOT. ASSOCIATED(FLOW sabs))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy
      cp1=gamtable(1)/(gamtable(1)-1.)*rgasref
      ALLOCATE(FLOW sabs(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            vsq=FLOW vx(k)**2+FLOW vy(k)**2+FLOW vz(k)**2
            hstat=FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            cpb=0.5*(cp+cp1)
            FLOW sabs(k)=cpb*ALOG(hstat/href)-rgasref* &
                             ALOG(FLOW p(k)/pref)
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
            vsq=FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            hstat=(FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)) * vref**2
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            cpb=.5*(cp+cp1)
!            FLOW sabs(k)=cpb*ALOG(hstat/href)-rgasref* &
!                             ALOG(FLOW p(k)/pref)
! Tstate and Pstate must be used because this is truly to a reference
! state and not a reference quantity....
            FLOW sabs(k)=cpb*ALOG(hstat/(cpref*tstate*tref))-rgasref* &
                             ALOG(FLOW p(k)/pstate)
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW sabs

  END SUBROUTINE Calc_EntropyAbs

!#######################################################################
  SUBROUTINE Calc_EntropyRatio()
!#######################################################################
    USE common_area
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_gas,  ONLY: Calc_Gamma
    USE calc_ener, ONLY: Calc_Energy

    INTEGER k, mvmax
    REAL vsq,hstat,gm3

    IF(.NOT. ASSOCIATED(FLOW srat))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy
  
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW srat(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k = 1, mvmax
            vsq = FLOW vx(k)**2+FLOW vy(k)**2+FLOW vz(k)**2
            hstat = FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)
            gm3 = FLOW g(k)/(FLOW g(k)-1.)
            FLOW srat(k) = pref/FLOW p(k)*(hstat/href)**gm3
            FLOW srat(k) = 0.0
          ENDDO
        CASE DEFAULT
          DO k = 1, mvmax
            vsq = FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            hstat = FLOW e(k)-vsq/2.+FLOW p(k)/FLOW r(k)
            gm3 = FLOW g(k)/(FLOW g(k)-1.)
            FLOW srat(k) = pref/FLOW p(k)*(hstat/href)**gm3
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW srat

  END SUBROUTINE Calc_EntropyRatio
END MODULE calc_entr
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
