#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_mach
  IMPLICIT NONE

CONTAINS
!#######################################################################
SUBROUTINE ScalMach(v, key, mvmax)
!#######################################################################
  USE common_area, ONLY: pfw, itime
  USE mod_events, ONLY: event_get_secondary

  INTEGER,INTENT(IN) :: key, mvmax
  REAL,INTENT(INOUT) :: v(*)
  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(1)
      CALL Calc_MachThAbs()
      ptr => FLOW machthabs
    CASE(2)
      CALL Calc_MachThRel()
      ptr => FLOW machthrel
    CASE(3)
      CALL Calc_MachR()
      ptr => FLOW machr
    CASE(4)
      CALL Calc_MachTAbs()
      ptr => FLOW machtabs
    CASE(5)
      CALL Calc_MachTRel()
      ptr => FLOW machtrel
    CASE(6)
      CALL Calc_MachXAbs()
      ptr => FLOW machxabs
    CASE(7)
      CALL Calc_MachXRel()
      ptr => FLOW machxrel
    CASE(8)
      CALL Calc_MachYAbs()
      ptr => FLOW machyabs
    CASE(9)
      CALL Calc_MachYRel()
      ptr => FLOW machyrel
    CASE(10)
      CALL Calc_MachZ()
      ptr => FLOW machz
    CASE(11)
      CALL Calc_MachIdeal()
      ptr => FLOW machideal
    CASE DEFAULT
      RETURN
  END SELECT
  DO k = 1, mvmax
    v(k) = ptr(k)
  ENDDO
  NULLIFY(ptr)

END SUBROUTINE ScalMach

!#######################################################################
SUBROUTINE Calc_MachThAbs()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax

  IF(.NOT. ASSOCIATED(FLOW machthabs))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machthabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          FLOW machthabs(m)=0.0
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          FLOW machthabs(m)=ABS(FLOW vt(m))/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machthabs

END SUBROUTINE Calc_MachThAbs

!#######################################################################
SUBROUTINE Calc_MachThRel()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machthrel))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machthrel(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          FLOW machthrel(m)=0.0
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          vel=FLOW vt(m)-omega*pgr(igrid)%r(m)
          FLOW machthrel(m)=ABS(vel)/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machthrel

END SUBROUTINE Calc_MachThRel

!#######################################################################
SUBROUTINE Calc_MachR()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax

  IF(.NOT. ASSOCIATED(FLOW machr))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machr(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          FLOW machr(m)=0.0
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          FLOW machr(m)=ABS(FLOW vr(m))/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machr

END SUBROUTINE Calc_MachR

!#######################################################################
SUBROUTINE Calc_MachTAbs()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machtabs))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machtabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          vel=SQRT(FLOW vx(m)**2+FLOW vy(m)**2+FLOW vz(m)**2)
          FLOW machtabs(m)=ABS(vel)/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          vel=SQRT(FLOW vt(m)**2+FLOW vr(m)**2+FLOW vz(m)**2)
          FLOW machtabs(m)=ABS(vel)/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machtabs

END SUBROUTINE Calc_MachTAbs

!#######################################################################
SUBROUTINE Calc_MachTRel()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel,wt

  IF(.NOT. ASSOCIATED(FLOW machtrel))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machtrel(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          FLOW machtrel(m)=0.0
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          wt=FLOW vt(m)-omega*pgr(igrid)%r(m)
          vel=SQRT(wt**2+FLOW vr(m)**2+FLOW vz(m)**2)
          FLOW machtrel(m)=ABS(vel)/SQRT(FLOW g(m) &
                              *FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machtrel

END SUBROUTINE Calc_MachTRel

!#######################################################################
SUBROUTINE Calc_MachXAbs()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machxabs))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machxabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          vel=FLOW vx(m)
          FLOW machxabs(m)=ABS(vel)/ &
                 SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          vel=FLOW vr(m)*COS(pgr(igrid)%t(m))- &
              FLOW vt(m)*SIN(pgr(igrid)%t(m))
          FLOW machxabs(m)=ABS(vel)/ &
                 SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machxabs

END SUBROUTINE Calc_MachXAbs

!#######################################################################
SUBROUTINE Calc_MachXRel()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machxrel))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_MachXAbs
        FLOW machxrel => FLOW machxabs
        NULLIFY(FLOW machxabs)
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW machxrel(mvmax))
        DO m=1,mvmax
          vel=FLOW vr(m)*COS(pgr(igrid)%t(m))- &
              (FLOW vt(m)-omega*pgr(igrid)%r(m))*SIN(pgr(igrid)%t(m))
          FLOW machxrel(m)=ABS(vel)/ &
               SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
    ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machxrel

END SUBROUTINE Calc_MachXRel

!#######################################################################
SUBROUTINE Calc_MachYAbs()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machyabs))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machyabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO m=1,mvmax
          vel=FLOW vy(m)
          FLOW machyabs(m)=ABS(vel)/ &
                 SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
        ENDDO
      CASE DEFAULT
        DO m=1,mvmax
          vel=FLOW vr(m)*SIN(pgr(igrid)%t(m))+ &
              FLOW vt(m)*COS(pgr(igrid)%t(m))
          FLOW machyabs(m)=ABS(vel)/ &
                 SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machyabs

END SUBROUTINE Calc_MachYAbs

!#######################################################################
SUBROUTINE Calc_MachYRel()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW machyrel))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_MachYAbs
        FLOW machyrel => FLOW machyabs
        NULLIFY(FLOW machyabs)
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW machyrel(mvmax))
        DO m=1,mvmax
          vel=FLOW vr(m)*SIN(pgr(igrid)%t(m))+ &
              (FLOW vt(m)-omega*pgr(igrid)%r(m))*COS(pgr(igrid)%t(m))
          FLOW machyrel(m)=ABS(vel)/ &
                 SQRT(FLOW g(m)*FLOW p(m)/FLOW r(m))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW machyrel

END SUBROUTINE Calc_MachYRel

!#######################################################################
SUBROUTINE Calc_MachZ()
!#######################################################################
  USE common_area
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure
  USE calc_gas,  ONLY: Calc_Gamma

  INTEGER m,mvmax

  IF(.NOT. ASSOCIATED(FLOW machz))THEN
    IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW machz(mvmax))
    DO m=1,mvmax
      FLOW machz(m)=ABS(FLOW vz(m))/SQRT(FLOW g(m) &
                          *FLOW p(m)/FLOW r(m))
    ENDDO
  ENDIF

  cur(itime)%v => FLOW machz

END SUBROUTINE Calc_MachZ

!#######################################################################
  SUBROUTINE Calc_MachIdeal()
!#######################################################################
    USE common_area
    USE calc_dens, ONLY: Calc_Density
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_gas,  ONLY: Calc_Gamma

    INTEGER m,mvmax
    REAL gm1,gm3,gm4,rothin,ptr
    REAL hidl,ptidl,amsq

    IF(.NOT. ASSOCIATED(FLOW machideal))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()

      rothin=1.
      ptr=1.
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW machideal(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO m=1,mvmax
            FLOW machideal(m) = 0.0
          ENDDO
        CASE DEFAULT
          DO m=1,mvmax
            gm1 = FLOW g(m) - 1
            gm3 = FLOW g(m)/(FLOW g(m)-1.)
            gm4 = 1./gm3
            hidl = rothin + (omega*pgr(igrid)%r(m))**2/2.
            ptidl = ptr*(hidl/rothin)**gm3
            amsq = ((MAX(FLOW p(m),ptidl)/FLOW p(m))**gm4-1)/(gm1/2.)
            FLOW machideal(m) = SQRT(amsq)
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW machideal

  END SUBROUTINE Calc_MachIdeal
END MODULE calc_mach
