#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_othr
  IMPLICIT NONE

CONTAINS
!#######################################################################
  SUBROUTINE ScalOther(v,key,mvmax)
!#######################################################################
    USE common_area, ONLY: pgr, igrid, pfw, itime, cid, omega
    USE mod_events, ONLY: event_get_secondary
    USE calc_gas, ONLY: Calc_Gamma

    INTEGER,INTENT(IN) :: key,mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k, s, j
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(1)
        IF(ASSOCIATED(pgr(igrid)%iblank))THEN
          DO k=1,mvmax
            v(k) = pgr(igrid)%iblank(k)
          ENDDO
        ELSE
          DO k = 1, mvmax
            v(k) = 0.0
          ENDDO
        ENDIF
        RETURN
      CASE(2)
        CALL Calc_Gamma()
        ptr => FLOW g
      CASE(3)
        DO k = 1, mvmax
          v(k) = cid
        ENDDO
        RETURN
      CASE(4)
        s = 0
        DO j = 1, pgr(igrid)%nblks
          DO k = 1, pgr(igrid)%mvmax(j)
            v(s+k) = FLOAT(j)
          ENDDO
          s = s + pgr(igrid)%mvmax(j)
        ENDDO
        RETURN
      CASE(5)
        DO k=1,mvmax
          v(k)=omega
        ENDDO
        RETURN
      CASE(6)
        DO k = 1, mvmax
          v(k) = 0.
        ENDDO
        RETURN
      CASE(7)
        DO k = 1, mvmax
          v(k) = 0.
        ENDDO
        RETURN
      CASE(8)
        CALL Calc_Reynolds
        ptr => FLOW reyn
!#ifdef FXPV3CLIENT
!      CASE(9)
!        CALL FX_RTGet(v)
!        RETURN
!#endif
      CASE DEFAULT
        RETURN
    END SELECT
    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)

  END SUBROUTINE ScalOther
    

!#######################################################################
SUBROUTINE Calc_Reynolds()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, cur, itime, vref, pref, tref, &
                         rgasref, visref, suthc1, suthc2, coordsys, diam
  IMPLICIT NONE

  INTEGER k, mvmax
  REAL mu, vmag, stemp

  IF(.NOT. ASSOCIATED(FLOW reyn))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW reyn(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k = 1, mvmax
          vmag = SQRT(FLOW vx(k)**2 + &
                      FLOW vy(k)**2 + FLOW vz(k)**2)
          stemp = FLOW p(k) / (rgasref * FLOW r(k))
          mu = suthc1 * (stemp**1.5) / (stemp * suthc2)
          FLOW reyn(k) = (FLOW r(k)*pref/rgasref/tref)* &
                               (vmag*vref)*diam/(mu*visref)
        ENDDO
      CASE DEFAULT
        DO k = 1, mvmax
          vmag = SQRT(FLOW vr(k)**2 + &
                      FLOW vt(k)**2 + FLOW vz(k)**2)
          stemp = tref * FLOW p(k) / (rgasref * FLOW r(k))
          mu = suthc1 * (stemp**1.5) / (stemp * suthc2)
          FLOW reyn(k) = (FLOW r(k)*pref/rgasref/tref)* &
                               (vmag*vref)*diam/(mu*visref)
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW reyn

END SUBROUTINE Calc_Reynolds
END MODULE calc_othr
