#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_pres
  IMPLICIT NONE

CONTAINS
!#######################################################################
SUBROUTINE ScalPressure(v, key, mvmax)
!#######################################################################
  USE common_area, ONLY: pfw, itime, pref
  USE mod_events, ONLY: event_get_secondary

  INTEGER,INTENT(IN) :: key, mvmax
  REAL,INTENT(INOUT) :: v(*)
  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(2)
      CALL Calc_PressureTAbs()
      ptr => FLOW ptabs
    CASE(3)
      CALL Calc_PressureTRel()
      ptr => FLOW ptrel
    CASE(1)
      CALL Calc_Pressure()
      ptr => FLOW p
    CASE DEFAULT
      RETURN
  END SELECT

  DO k = 1, mvmax
    v(k) = ptr(k) * pref
  ENDDO

END SUBROUTINE ScalPressure

!#######################################################################
  SUBROUTINE Calc_Pressure()
!#######################################################################
    USE common_area
    USE calc_gas, ONLY: Calc_Gamma

    INTEGER k,mvmax

    IF(.NOT. ASSOCIATED(FLOW p))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma

      mvmax = pgr(igrid)%tpts
      ALLOCATE(FLOW p(mvmax))
      DO k = 1, mvmax
        FLOW p(k) = (FLOW g(k)-1.) * FLOW r(k) * FLOW e(k)
      ENDDO

    ENDIF

    cur(itime)%v => FLOW p

  END SUBROUTINE Calc_Pressure

!#######################################################################
  SUBROUTINE Calc_PressureTAbs()
!#######################################################################
    USE common_area, ONLY: pgr, igrid, pfw, itime, coordsys, TKEinE, cur
    USE calc_gas,  ONLY: Calc_Gamma
    USE calc_ener, ONLY: Calc_Energy
    USE calc_vel,  ONLY: Calc_VelocityTheta, Calc_VelocityRadial

    INTEGER m,mvmax
    REAL gm3,vsq,hstat,htot,bs
    REAL, DIMENSION(:), POINTER :: v1, v2, v3, k

    IF(.NOT. ASSOCIATED(FLOW ptabs))THEN
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy()
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
      IF(.NOT. ASSOCIATED(FLOW vt))CALL Calc_VelocityTheta()
      IF(.NOT. ASSOCIATED(FLOW vr))CALL Calc_VelocityRadial()

      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW ptabs(mvmax))

      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          v1 => FLOW vx
          v2 => FLOW vy
          v3 => FLOW vz
          k  => FLOW p !This is really really wrong, but works because TKEinE=0
        CASE DEFAULT
          v1 => FLOW vt
          v2 => FLOW vr
          v3 => FLOW vz
          k  => FLOW k
      END SELECT

      DO m=1,mvmax
        gm3=FLOW g(m)/(FLOW g(m)-1.)
        vsq = v1(m)**2 + v2(m)**2 + v3(m)**2
        bs = FLOW e(m) + FLOW p(m) / FLOW r(m)
        hstat = bs - vsq/2.
        htot = bs + TKEinE*k(m)
        FLOW ptabs(m) = FLOW p(m)*(htot/hstat)**gm3
      ENDDO
    ENDIF

    cur(itime)%v => FLOW ptabs

  END SUBROUTINE Calc_PressureTAbs

!#######################################################################
  SUBROUTINE Calc_PressureTRel()
!#######################################################################
    USE common_area, ONLY: pgr, igrid, pfw, itime, omega, cur, coordsys, &
                           TKEinE
    USE calc_gas,  ONLY: Calc_Gamma
    USE calc_vel,  ONLY: Calc_VelocityTheta, Calc_VelocityRadial

    INTEGER m,mvmax
    REAL wt,gm3,wsq,vsq,hstat,htrel,htot

    IF(.NOT. ASSOCIATED(FLOW ptrel))THEN
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
      IF(.NOT. ASSOCIATED(FLOW vt))CALL Calc_VelocityTheta()
      IF(.NOT. ASSOCIATED(FLOW vr))CALL Calc_VelocityRadial()

      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          CALL Calc_PressureTAbs()
          FLOW ptrel =>  FLOW ptabs
          NULLIFY(FLOW ptabs)
        CASE DEFAULT
          mvmax=pgr(igrid)%tpts
          ALLOCATE(FLOW ptrel(mvmax))
          DO m=1,mvmax
            gm3=FLOW g(m)/(FLOW g(m)-1.)
            wt = FLOW vt(m) - omega*pgr(igrid)%r(m)
            wsq = wt**2 + FLOW vr(m)**2 + FLOW vz(m)**2
            vsq = FLOW vt(m)**2 + FLOW vr(m)**2 + FLOW vz(m)**2
            htot = FLOW e(m)+FLOW p(m)/FLOW r(m)+TKEinE*FLOW k(m)
            htrel = htot - vsq/2. + wsq/2.
            hstat = FLOW e(m) - vsq/2. + FLOW p(m)/FLOW r(m)
            FLOW ptrel(m) = FLOW p(m)*(htrel/hstat)**gm3
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW ptrel

  END SUBROUTINE Calc_PressureTRel
END MODULE calc_pres
