#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_rhov
  IMPLICIT NONE

CONTAINS
!#######################################################################
SUBROUTINE ScalRhov(v,key,mvmax)
!#######################################################################
  USE common_area, ONLY: pfw, itime
  USE mod_events, ONLY: event_get_secondary

  INTEGER,INTENT(IN) :: key, mvmax
  REAL,INTENT(INOUT) :: v(*)
  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(1)
      CALL Calc_RhoVThAbs()
      ptr => FLOW rhovthabs
    CASE(2)
      CALL Calc_RhoVThRel()
      ptr => FLOW rhovthrel
    CASE(3)
      CALL Calc_RhoVR()
      ptr => FLOW rhovr
    CASE(4)
      CALL Calc_RhoVTAbs()
      ptr => FLOW rhovtabs
    CASE(5)
      CALL Calc_RhoVTRel()
      ptr => FLOW rhovtrel
    CASE(6)
      CALL Calc_RhoVXAbs()
      ptr => FLOW rhovxabs
    CASE(7)
      CALL Calc_RhoVXRel()
      ptr => FLOW rhovxrel
    CASE(8)
      CALL Calc_RhoVYAbs()
      ptr => FLOW rhovyabs
    CASE(9)
      CALL Calc_RhoVYRel()
      ptr => FLOW rhovyrel
    CASE(10)
      CALL Calc_RhoVZ()
      ptr => FLOW rhovz
    CASE DEFAULT
      RETURN
  END SELECT
  DO k = 1, mvmax
    v(k) = ptr(k)
  ENDDO
  NULLIFY(ptr)

END SUBROUTINE ScalRhov

!#######################################################################
SUBROUTINE Calc_RhoVThAbs()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovthabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovthabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW rhovthabs(k)=0.0
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW rhovthabs(k)=FLOW r(k)*FLOW vt(k)
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovthabs

END SUBROUTINE Calc_RhoVThAbs

!#######################################################################
SUBROUTINE Calc_RhoVThRel()
!#######################################################################
  USE common_area

  INTEGER k,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW rhovthrel))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovthrel(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW rhovthrel(k)=0.0
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          vel=FLOW vt(k)-omega*pgr(igrid)%r(k)
          FLOW rhovthrel(k)=FLOW r(k)*vel
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovthrel

END SUBROUTINE Calc_RhoVThRel

!#######################################################################
SUBROUTINE Calc_RhoVR()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovr))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovr(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW rhovr(k)=0.0
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW rhovr(k)=FLOW r(k)*FLOW vr(k)
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovr

END SUBROUTINE Calc_RhoVR

!#######################################################################
SUBROUTINE Calc_RhoVTAbs()
!#######################################################################
  USE common_area

  INTEGER k,mvmax
  REAL vel

  IF(.NOT. ASSOCIATED(FLOW rhovtabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovtabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          vel=SQRT(FLOW vx(k)**2+FLOW vy(k)**2+FLOW vz(k)**2)
          FLOW rhovtabs(k)=FLOW r(k)*vel
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          vel=SQRT(FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2)
          FLOW rhovtabs(k)=FLOW r(k)*vel
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovtabs

END SUBROUTINE Calc_RhoVTAbs

!#######################################################################
SUBROUTINE Calc_RhoVTRel()
!#######################################################################
  USE common_area

  INTEGER k,mvmax
  REAL vel,wt

  IF(.NOT. ASSOCIATED(FLOW rhovtrel))THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_RhoVTAbs
        FLOW rhovtrel => FLOW rhovtabs
        NULLIFY(FLOW rhovtabs)
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW rhovtrel(mvmax))
        DO k=1,mvmax
          wt=FLOW vt(k)-omega*pgr(igrid)%r(k)
          vel=SQRT(wt**2+FLOW vr(k)**2+FLOW vz(k)**2)
          FLOW rhovtrel(k)=FLOW r(k)*vel
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovtrel

END SUBROUTINE Calc_RhoVTRel

!#######################################################################
SUBROUTINE Calc_RhoVXAbs()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovxabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovxabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW rhovxabs(k)=FLOW r(k)*FLOW vx(k)
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW rhovxabs(k)=FLOW r(k)* &
             (FLOW vr(k)*SIN(pgr(igrid)%t(k)) + &
              FLOW vt(k)*COS(pgr(igrid)%t(k)))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovxabs

END SUBROUTINE Calc_RhoVXAbs

!#######################################################################
SUBROUTINE Calc_RhoVXRel()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovxrel))THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_RhoVXAbs
        FLOW rhovxrel => FLOW rhovxabs
        NULLIFY(FLOW rhovxabs)
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW rhovxrel(mvmax))
        DO k=1,mvmax
          FLOW rhovxrel(k)=FLOW r(k)* &
             (FLOW vr(k)*SIN(pgr(igrid)%t(k)) + &
             (FLOW vt(k)-omega*pgr(igrid)%r(k))*COS(pgr(igrid)%t(k)))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovxrel

END SUBROUTINE Calc_RhoVXRel

!#######################################################################
SUBROUTINE Calc_RhoVYAbs()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovyabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW rhovyabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW rhovyabs(k)=FLOW r(k)*FLOW vy(k)
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW rhovyabs(k)=FLOW r(k)* &
             (FLOW vr(k)*COS(pgr(igrid)%t(k)) - &
              FLOW vt(k)*SIN(pgr(igrid)%t(k)))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovyabs

END SUBROUTINE Calc_RhoVYAbs

!#######################################################################
SUBROUTINE Calc_RhoVYRel()
!#######################################################################
  USE common_area

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW rhovyrel))THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_RhoVYAbs
        FLOW rhovyrel => FLOW rhovyabs
        NULLIFY(FLOW rhovyabs)
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW rhovyrel(mvmax))
        DO k=1,mvmax
          FLOW rhovyrel(k)=FLOW r(k)* &
             (FLOW vr(k)*COS(pgr(igrid)%t(k)) - &
             (FLOW vt(k)-omega*pgr(igrid)%r(k))*SIN(pgr(igrid)%t(k)))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW rhovyrel

END SUBROUTINE Calc_RhoVYRel

!#######################################################################
  SUBROUTINE Calc_RhoVZ()
!#######################################################################
    USE common_area

    INTEGER k,mvmax

    IF(.NOT. ASSOCIATED(FLOW rhovz))THEN
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW rhovz(mvmax))
      DO k=1,mvmax
        FLOW rhovz(k)=FLOW r(k)*FLOW vz(k)
      ENDDO
    ENDIF

    cur(itime)%v => FLOW rhovz

  END SUBROUTINE Calc_RhoVZ
END MODULE calc_rhov
