#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_roth
  IMPLICIT NONE

CONTAINS
!#######################################################################
  SUBROUTINE ScalRothalpy(v,key,mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime
    USE mod_events, ONLY: event_get_secondary

    REAL,INTENT(INOUT) :: v(*)
    INTEGER,INTENT(IN) :: key, mvmax
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(2)
        CALL Calc_RothalpyCP()
        ptr => FLOW rothcp
      CASE(1)
        CALL Calc_Rothalpy()
        ptr => FLOW roth
      CASE DEFAULT
        RETURN
    END SELECT
    DO k=1,mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)
  

  END SUBROUTINE ScalRothalpy

!#######################################################################
  SUBROUTINE Calc_Rothalpy()
!#######################################################################
    USE common_area, ONLY: omega, pgr, igrid, pfw, itime, coordsys, cur, &
                           TKEinE
    USE calc_pres, ONLY: Calc_Pressure

    INTEGER k,mvmax
    REAL wt,wsq,vsq,htot,hstat,omgr

    IF(.NOT. ASSOCIATED(FLOW roth))THEN
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW roth(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            FLOW roth(k)=0.0
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
            omgr=omega*pgr(igrid)%r(k)
            wt=FLOW vt(k)-omgr
    wsq=wt**2+FLOW vr(k)**2+FLOW vz(k)**2
    vsq=FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            htot=FLOW e(k)+FLOW p(k)/FLOW r(k)+ &
                 TKEinE*FLOW k(k)
    hstat=htot-vsq/2.
    FLOW roth(k)=hstat+wsq/2.-omgr**2/2.
!            FLOW roth(k)=htot+(wt**2-FLOW vt(k)**2)/2.-omgr**2/2.
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW roth

  END SUBROUTINE Calc_Rothalpy

!#######################################################################
  SUBROUTINE Calc_RothalpyCP()
!#######################################################################
    USE common_area
    USE calc_gas, ONLY: Calc_Gamma

    INTEGER k,mvmax
    REAL wt,wsq,vsq,htot,hstat,omgr,cp
  
    IF(.NOT. ASSOCIATED(FLOW rothcp))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
  
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW rothcp(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            FLOW rothcp(k)=0.0
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
            omgr=omega*pgr(igrid)%r(k)
            wt=FLOW vt(k)-omgr
            wsq=wt**2+FLOW vr(k)**2+FLOW vz(k)**2
            vsq=FLOW vt(k)**2+FLOW vr(k)**2+FLOW vz(k)**2
            htot=FLOW e(k)+FLOW p(k)/FLOW r(k)+ &
                 TKEinE*FLOW k(k)
            hstat=htot-vsq/2.
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            FLOW rothcp(k)=(hstat+wsq/2.-omgr**2/2.)/cp
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW rothcp

  END SUBROUTINE Calc_RothalpyCP
END MODULE calc_roth
