#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_temp
  IMPLICIT NONE

CONTAINS
!#######################################################################
  SUBROUTINE ScalTemperature(v, key, mvmax)
!#######################################################################
    USE common_area
    USE mod_events, ONLY: event_get_secondary

    INTEGER,INTENT(IN) :: key, mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(2)
        CALL Calc_TemperatureTAbs()
        ptr => FLOW temptabs
      CASE(3)
        CALL Calc_TemperatureTRel()
        ptr => FLOW temptrel
      CASE(1)
        CALL Calc_Temperature()
        ptr => FLOW temp
      CASE DEFAULT
        RETURN
    END SELECT
    DO k=1,mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)

END SUBROUTINE ScalTemperature

!#######################################################################
  SUBROUTINE Calc_Temperature()
!#######################################################################
    USE common_area, ONLY: pgr, igrid, pfw, itime, cur, pref, rhoref, &
                           rgasref, refl
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_dens, ONLY: Calc_Density

    INTEGER k,mvmax

    IF(.NOT. ASSOCIATED(FLOW temp))THEN
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
      IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()

      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW temp(mvmax))
      DO k=1,mvmax
        FLOW temp(k) = FLOW p(k)/(rgasref*FLOW r(k)) * (pref/rhoref) / refl**3
      ENDDO
    ENDIF

! I promise I'll think about this...
! Basically, the rgasref for Marshall's case needs to be set to 1/(g*Minf**2)
! g * Minf**2 * p / rho

    cur(itime)%v => FLOW temp

  END SUBROUTINE Calc_Temperature

!#######################################################################
  SUBROUTINE Calc_TemperatureTAbs()
!#######################################################################
! This is only true for vstage with rgas=constant
    USE common_area
    USE calc_pres, ONLY: Calc_Pressure
    USE calc_ener, ONLY: Calc_Energy
    USE calc_gas, ONLY: Calc_Gamma

    INTEGER k,mvmax
    REAL htot,cp,cpb,cp1

    IF(.NOT. ASSOCIATED(FLOW temptabs))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma
      IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure
      IF(.NOT. ASSOCIATED(FLOW e))CALL Calc_Energy
      cp1=gamtable(1)/(gamtable(1)-1.)*rgasref
      mvmax=pgr(igrid)%tpts
      ALLOCATE(FLOW temptabs(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k=1,mvmax
            htot=FLOW e(k)+FLOW p(k)/FLOW r(k) !+TKEinE*FLOW k(k)
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            cpb=0.5*(cp+cp1)
            FLOW temptabs(k)=tref*(htot-tref*(cpref-cpb))/cpb
          ENDDO
        CASE DEFAULT
          DO k=1,mvmax
!          htot=FLOW e(k) + FLOW p(k)/FLOW r(k) + TKEinE*FLOW k(k)
            htot=(FLOW e(k) + FLOW p(k)/(rgasref*FLOW r(k))/refl**3 + &
                 TKEinE*FLOW k(k))*vref**2
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            cpb=0.5*(cp+cp1)
!          FLOW temptabs(k)=tref*(htot-tref*(cpref-cpb))/cpb
            FLOW temptabs(k)=(htot-tref*(cpref-cpb))/cpb
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW temptabs

  END SUBROUTINE Calc_TemperatureTAbs

!#######################################################################
  SUBROUTINE Calc_TemperatureTRel()
!#######################################################################
! This is only true for vstage with rgas=constant
    USE common_area
    USE calc_gas, ONLY: Calc_Gamma

    INTEGER k,mvmax
    REAL htot,cp,cpb,cp1,wt,wsq,vsq,htrel

    IF(.NOT. ASSOCIATED(FLOW temptrel))THEN
      IF(.NOT. ASSOCIATED(FLOW g))CALL Calc_Gamma()
      cp1=gamtable(1)/(gamtable(1)-1.)*rgasref
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          CALL Calc_TemperatureTAbs
          FLOW temptrel => FLOW temptabs
          NULLIFY(FLOW temptabs)
        CASE DEFAULT
          mvmax=pgr(igrid)%tpts
          ALLOCATE(FLOW temptrel(mvmax))
          DO k=1,mvmax
            wt=FLOW vt(k)-omega*pgr(igrid)%r(k)
            wsq=wt**2 + FLOW vr(k)**2 + FLOW vz(k)**2
            vsq=FLOW vt(k)**2 + FLOW vr(k)**2 + FLOW vz(k)**2
            htot=FLOW e(k) + FLOW p(k)/(rgasref*FLOW r(k))/refl**3 + &
                 TKEinE*FLOW k(k)
            htrel=(htot-vsq/2.+wsq/2.)*vref**2
            cp=FLOW g(k)/(FLOW g(k)-1.)*rgasref
            cpb=0.5*(cp+cp1)
            FLOW temptrel(k)=(htrel-tref*(cpref-cpb))/cpb
          ENDDO
      END SELECT
    ENDIF

    cur(itime)%v => FLOW temptrel

  END SUBROUTINE Calc_TemperatureTRel
END MODULE calc_temp
