#define FLOW pfw(itime)%
#define PI 3.141592653589793
MODULE calc_tavg
  IMPLICIT NONE
CONTAINS
!#######################################################################
  SUBROUTINE ScalTimeaverage(v,key,mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime
    USE mod_events, ONLY: event_get_secondary

    INTEGER,INTENT(IN) :: key,mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(1)
        CALL Calc_TmAvg()
        ptr => FLOW tmavg
      CASE(2)
        CALL Calc_TmDiff()
        ptr => FLOW tmdif
      CASE(3)
        CALL Calc_TmMin()
        ptr => FLOW tmmin
      CASE(4)
        CALL Calc_TmMax()
        ptr => FLOW tmmax
      CASE(5)
        CALL Calc_TmRms()
        ptr => FLOW tmrms
      CASE(6)
        CALL Calc_TmFFT()
        ptr => FLOW tmfft
      CASE DEFAULT
        RETURN
    END SELECT
    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)

  END SUBROUTINE ScalTimeaverage

!#######################################################################
SUBROUTINE Calc_TmAvg()
!#######################################################################
! This routine calculates the time average of a scalar quantity.  Mike
! you really should read some of your comments, you would make lots
! fewer mistakes
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,m,mvmax,itmp,rkey
  REAL,POINTER,DIMENSION(:) :: tmparr

  IF(ASSOCIATED(pfw(1)%tmavg))RETURN

  mvmax=pgr(igrid)%tpts
  ALLOCATE(tmparr(mvmax))
! Generate a sum of all the flow files for the variable specified
! by the user in key2_2 (primary scalar) and key2_1 (option)
! Key2_3 is the time-averaging option
  itmp=itime
  rkey=1
  DO WHILE(key2_2.NE.ikeys(rkey))
    rkey=rkey+1
    IF(rkey.GT.nkeys)GOTO 9001
  ENDDO
  DO itime=1,ntime
    CALL pVScal(rkey,tmparr)
  ENDDO
  DEALLOCATE(tmparr)
  NULLIFY(tmparr)
  itime=itmp

!=======================================================================
! Complete the average by dividing by the number of steps
  ALLOCATE(pfw(1)%tmavg(mvmax))
  DO m=1,mvmax
    pfw(1)%tmavg(m) = 0.
    DO k=1,ntime
      pfw(1)%tmavg(m) = pfw(1)%tmavg(m) + cur(k)%v(m)
    ENDDO
    pfw(1)%tmavg(m) = pfw(1)%tmavg(m)/FLOAT(ntime)
  ENDDO

!=======================================================================
! Assign all time steps to the average
  DO k=2,ntime
    pfw(k)%tmavg => pfw(1)%tmavg
  ENDDO

  RETURN

9001 CONTINUE
  WRITE(*,*) ' ERROR, key not found in list'

  RETURN

END SUBROUTINE Calc_TmAvg

!#######################################################################
SUBROUTINE Calc_TmDiff()
!#######################################################################
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,i,mvmax,rkey
!  REAL,POINTER,DIMENSION(:) :: tmparr

  IF(ASSOCIATED(pfw(1)%tmdif))RETURN
  IF(.NOT. ASSOCIATED(pfw(1)%tmavg))CALL Calc_TmAvg()
! The above call sets the pointers for CUR(ITIME)

  mvmax=pgr(igrid)%tpts
  DO i=1,ntime
    ALLOCATE(pfw(i)%tmdif(mvmax))
    DO k=1,mvmax
      pfw(i)%tmdif(k)=cur(i)%v(k)-pfw(1)%tmavg(k)
    ENDDO
  ENDDO

  RETURN

END SUBROUTINE Calc_TmDiff

!#######################################################################
SUBROUTINE Calc_TmRms()
!#######################################################################
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,m,mvmax

  IF(ASSOCIATED(FLOW tmrms))RETURN
  IF(.NOT. ASSOCIATED(FLOW tmrms))CALL Calc_TmAvg()
  mvmax=pgr(igrid)%tpts
  ALLOCATE(pfw(1)%tmrms(mvmax))
  DO m=1,mvmax
    pfw(1)%tmrms(m)=0.0
  ENDDO
  DO m=1,mvmax
    DO k=1,ntime
      pfw(1)%tmrms(m) = pfw(1)%tmrms(m) + (cur(k)%v(m) - pfw(k)%tmavg(m))**2
    ENDDO
    pfw(1)%tmrms(m)=SQRT(pfw(1)%tmrms(m)/FLOAT(ntime))
  ENDDO

!=======================================================================
! Assign all time steps to the RMS
  DO k=2,ntime
    pfw(k)%tmrms => pfw(1)%tmrms
  ENDDO

  RETURN

END SUBROUTINE Calc_TmRms

!#######################################################################
SUBROUTINE Calc_TmMin()
!#######################################################################
! This routine calculates the time average of a scalar quantity.  Mike
! you really should read some of your comments, you would make lots
! fewer mistakes
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,m,mvmax,itmp,rkey
  REAL,POINTER,DIMENSION(:) :: tmparr

!=======================================================================
! Check to see whether or not calculations have occurred
  IF(ASSOCIATED(FLOW tmmin))RETURN

!=======================================================================
! Check to see whether or not scalars have been calculated
  mvmax=pgr(igrid)%tpts
  ALLOCATE(tmparr(mvmax))
! Generate a sum of all the flow files for the variable specified
! by the user in key2_2 (primary scalar) and key2_1 (option)
! Key2_3 is the time-averaging option
  itmp=itime
  rkey=1
  DO WHILE(key2_2.NE.ikeys(rkey))
    rkey=rkey+1
    IF(rkey.GT.nkeys)GOTO 9001
  ENDDO
  DO itime=1,ntime
    CALL pVScal(rkey,tmparr)
  ENDDO
  DEALLOCATE(tmparr)
  NULLIFY(tmparr)
  itime=itmp

!=======================================================================
! Assign all time steps to the min
  ALLOCATE(pfw(1)%tmmin(mvmax))
  DO m=1,mvmax
    pfw(1)%tmmin(m) = cur(1)%v(m)
    DO k=2,ntime
      pfw(1)%tmmin(m) = MIN(pfw(1)%tmmin(m),cur(k)%v(m))
    ENDDO
  ENDDO
  DO k=2,ntime
    pfw(k)%tmmin => pfw(1)%tmmin
  ENDDO

  RETURN

9001 CONTINUE
  WRITE(*,*) ' ERROR, key not found in list'

  RETURN

END SUBROUTINE Calc_TmMin

!#######################################################################
SUBROUTINE Calc_TmMax()
!#######################################################################
! This routine calculates the time average of a scalar quantity.  Mike
! you really should read some of your comments, you would make lots
! fewer mistakes
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,m,mvmax,itmp,rkey
  REAL,POINTER,DIMENSION(:) :: tmparr

!=======================================================================
! Check to see whether or not calculations have occurred
  IF(ASSOCIATED(FLOW tmmax))RETURN

!=======================================================================
! Check to see whether or not scalars have been calculated
  mvmax=pgr(igrid)%tpts
  ALLOCATE(tmparr(mvmax))
! Generate a sum of all the flow files for the variable specified
! by the user in key2_2 (primary scalar) and key2_1 (option)
! Key2_3 is the time-averaging option
  itmp=itime
  rkey=1
  DO WHILE(key2_2.NE.ikeys(rkey))
    rkey=rkey+1
    IF(rkey.GT.nkeys)GOTO 9001
  ENDDO
  DO itime=1,ntime
    CALL pVScal(rkey,tmparr)
  ENDDO
  DEALLOCATE(tmparr)
  NULLIFY(tmparr)
  itime=itmp

!=======================================================================
! Assign all time steps to the min
  ALLOCATE(pfw(1)%tmmax(mvmax))
  DO m=1,mvmax
    pfw(1)%tmmax(m) = cur(1)%v(m)
    DO k=2,ntime
      pfw(1)%tmmax(m) = MAX(pfw(1)%tmmax(m),cur(k)%v(m))
    ENDDO
  ENDDO
  DO k=2,ntime
    pfw(k)%tmmax => pfw(1)%tmmax
  ENDDO

  RETURN

9001 CONTINUE
  WRITE(*,*) ' ERROR, key not found in list'

  RETURN

END SUBROUTINE Calc_TmMax

!#######################################################################
SUBROUTINE Calc_TmFFT()
!#######################################################################
! The FFT utility uses the cur(itime) pointers to avoid extraneous
! memory allocation.  The only folly is that this MUST be filled fully
! prior to the start of the fft.  In addition, it MUST be filled with
! the correct data, so it is imperative that the calls to pVScal be
! made first!!!!
  USE general
  USE common_area
  USE keydata
  IMPLICIT NONE

  INTEGER k,m,mvmax,itmp,rkey,dir,mode

  REAL arg
  REAL,POINTER,DIMENSION(:) :: tmparr
  REAL,ALLOCATABLE,DIMENSION(:,:) :: sinarg,cosarg
  REAL,ALLOCATABLE,DIMENSION(:) :: x1,x2,y1,y2

  IF(ASSOCIATED(pfw(1)%tmfft))RETURN

  mvmax=pgr(igrid)%tpts
  ALLOCATE(tmparr(mvmax))

! Find the right key and get the scalar data stored in CUR(ITIME)
  itmp=itime
  rkey=1
  DO WHILE(key2_2.NE.ikeys(rkey))
    rkey=rkey+1
    IF(rkey.GT.nkeys)GOTO 9001
  ENDDO
  DO itime=1,ntime
    CALL pVScal(rkey,tmparr)
  ENDDO
  DEALLOCATE(tmparr)
  NULLIFY(tmparr)
  itime=itmp

! Calculate the DFT
  ALLOCATE(x1(ntime),x2(ntime),y1(ntime),y2(ntime))
  ALLOCATE(sinarg(ntime,ntime),cosarg(ntime,ntime))

  dir = 1 ! Forward transformation

  DO m=1,ntime
    y1(m) = 0.
    arg = -dir*2.* PI *FLOAT(m-1)/FLOAT(ntime)
    DO k=1,ntime
      cosarg(m,k) = COS(FLOAT(k-1)*arg)
      sinarg(m,k) = SIN(FLOAT(k-1)*arg)
    ENDDO
  ENDDO

  mode=key2_mode+1
  ALLOCATE(pfw(1)%tmfft(mvmax))
  DO m=1,mvmax
    DO k=1,ntime
      x1(k) = cur(k)%v(m)
    ENDDO
    CALL dft(dir,ntime,sinarg,cosarg,x1,y1,x2,y2)
    SELECT CASE(key2_fft)
      CASE(ICHAR('p'))
        pfw(1)%tmfft(m) = ATAN2(y2(mode),x2(mode))
      CASE(ICHAR('m'))
        pfw(1)%tmfft(m) = SQRT(x2(mode)**2 + y2(mode)**2)
      CASE(ICHAR('r'))
        pfw(1)%tmfft(m) = x2(mode)
      CASE(ICHAR('i'))
        pfw(1)%tmfft(m) = y2(mode)
    END SELECT
  ENDDO
  DEALLOCATE(x1,y1,x2,y2,sinarg,cosarg)

! Store back the values to all time steps
  DO k=2,ntime
    pfw(k)%tmfft => pfw(1)%tmfft
  ENDDO

  RETURN

9001 CONTINUE
  WRITE(*,*) ' ERROR, key not found in list'

  RETURN

END SUBROUTINE Calc_TmFFT

!#######################################################################
SUBROUTINE dft(dir,m,sinarg,cosarg,x1,y1,x2,y2)
!#######################################################################
  IMPLICIT NONE

  INTEGER i,k
  INTEGER,INTENT(IN) :: dir,m
  REAL,INTENT(IN) :: x1(m),y1(m)
  REAL,INTENT(IN) :: cosarg(m,m),sinarg(m,m)
  REAL,INTENT(OUT) :: x2(m),y2(m)

  DO i=1,m
    x2(i) = 0.
    y2(i) = 0.
    DO k=1,m
      x2(i) = x2(i) + (x1(k)*cosarg(i,k) - y1(k)*sinarg(i,k))
      y2(i) = y2(i) + (x1(k)*sinarg(i,k) + y1(k)*cosarg(i,k))
    ENDDO
  ENDDO

  IF(dir.EQ.1)THEN
    DO i=1,m
      x2(i) = x2(i)/FLOAT(m)
      y2(i) = y2(i)/FLOAT(m)
    ENDDO
  ENDIF

  RETURN

END SUBROUTINE dft
END MODULE calc_tavg
