#include "parameters.h"
#define FLOW pfw(itime)%
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE calc_turb
  IMPLICIT NONE

CONTAINS

!#######################################################################
  SUBROUTINE ScalTurbulence(v,key,mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime
    USE mod_events, ONLY: event_get_secondary

    INTEGER,INTENT(IN) :: key, mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k
    REAL,POINTER,DIMENSION(:) :: ptr

    SELECT CASE(event_get_secondary(key))
      CASE(1)
        CALL Calc_TurbMu()
        ptr => FLOW mut
      CASE(2)
        CALL Calc_TurbK()
        ptr => FLOW k
      CASE(3)
        CALL Calc_TurbEps()
        ptr => FLOW eps
      CASE(4)
        CALL Calc_TurbInt()
        ptr => FLOW turbint
      CASE(5)
        CALL Calc_QCrit()
        ptr => FLOW qcrit
      CASE(6)
        CALL Calc_TurbRat()
        ptr => FLOW turbrat
      CASE(7)
        CALL Calc_TurbTVisc()
        ptr => FLOW turbtvisc
      CASE DEFAULT
        RETURN
    END SELECT
    DO k=1,mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)

  END SUBROUTINE ScalTurbulence

!#######################################################################
  SUBROUTINE Calc_TurbMu()
!#######################################################################
    USE common_area, ONLY: cur, pfw, itime, pgr, igrid

    INTEGER :: k,mvmax

    IF(.NOT. ASSOCIATED( FLOW mut ))THEN
      mvmax = pgr(igrid)%tpts
      ALLOCATE(FLOW mut(mvmax))
      DO k = 1, mvmax
        FLOW mut(k) = 0.
      ENDDO
    ENDIF

    cur(itime)%v => FLOW mut

  END SUBROUTINE Calc_TurbMu

!#######################################################################
  SUBROUTINE Calc_TurbK()
!#######################################################################
    USE common_area, ONLY: cur, pfw, itime, pgr, igrid

    INTEGER :: k,mvmax

    IF(.NOT. ASSOCIATED(FLOW k))THEN
      mvmax = pgr(igrid)%tpts
      ALLOCATE(FLOW k(mvmax))
      DO k = 1, mvmax
        FLOW k(k) = 0.
      ENDDO
    ENDIF

    cur(itime)%v => FLOW k

  END SUBROUTINE Calc_TurbK

!#######################################################################
  SUBROUTINE Calc_TurbEps()
!#######################################################################
    USE common_area, ONLY: cur, pfw, itime, pgr, igrid

    INTEGER :: k,mvmax

    IF(.NOT. ASSOCIATED(FLOW eps))THEN
      mvmax = pgr(igrid)%tpts
      ALLOCATE(FLOW eps(mvmax))
      DO k = 1, mvmax
        FLOW eps(k) = 0.
      ENDDO
    ENDIF

    cur(itime)%v => FLOW eps

  END SUBROUTINE Calc_TurbEps

!#######################################################################
  SUBROUTINE Calc_TurbInt()
!#######################################################################
    USE common_area, ONLY: cur, itime, pfw, pgr, igrid, omega, coordsys

    INTEGER :: k, mvmax
    REAL :: wt, vmag

    IF(.NOT. ASSOCIATED(FLOW turbint))THEN
      mvmax = pgr(igrid)%tpts
      ALLOCATE(FLOW turbint(mvmax))
      SELECT CASE(coordsys)
        CASE(CARTESIAN)
          DO k = 1, mvmax
            FLOW turbint(k) = 0.
          ENDDO
        CASE DEFAULT
          DO k = 1, mvmax
            wt = FLOW vt(k) - omega*pgr(igrid)%r(k)
            FLOW turbint(k) = SQRT(2./3. * FLOW k(k)) / &
                              SQRT(FLOW vr(k)**2 + wt**2 + FLOW vz(k)**2)
          ENDDO
      END SELECT
    ENDIF
  
    cur(itime)%v => FLOW turbint

  END SUBROUTINE Calc_TurbInt

!#######################################################################
SUBROUTINE Calc_TurbRat()
!#######################################################################
  USE common_area, ONLY: pfw, pgr, igrid, itime, suthc1, suthc2, cur, &
                         tref, visref, rgasref, pref, rhoref
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure

  INTEGER k,mvmax
  REAL stemp,dmol

  IF(.NOT. ASSOCIATED(FLOW turbrat))THEN
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()
    IF(.NOT. ASSOCIATED(FLOW mut))CALL Calc_TurbMu()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW turbrat(mvmax))
    DO k=1,mvmax
      stemp= FLOW p(k) / (rgasref * FLOW r(k)) * (pref/rhoref)
      dmol=suthc1*(stemp**1.5)/(stemp + suthc2)
      FLOW turbrat(k)= visref * FLOW mut(k) / dmol
    ENDDO
  ENDIF

  cur(itime)%v => FLOW turbrat

END SUBROUTINE Calc_TurbRat

!#######################################################################
SUBROUTINE Calc_TurbTVisc()
!#######################################################################
  USE common_area, ONLY: pfw, itime, rgasref, suthc1, suthc2, pref, &
                         rhoref, visref, cur, igrid, pgr
  USE calc_dens, ONLY: Calc_Density
  USE calc_pres, ONLY: Calc_Pressure

  INTEGER k,mvmax
  REAL stemp,dmol

  IF(.NOT. ASSOCIATED(FLOW turbtvisc))THEN
    IF(.NOT. ASSOCIATED(FLOW r))CALL Calc_Density()
    IF(.NOT. ASSOCIATED(FLOW p))CALL Calc_Pressure()
    IF(.NOT. ASSOCIATED(FLOW mut))CALL Calc_TurbMu()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW turbtvisc(mvmax))
    DO k=1,mvmax
      stemp = FLOW p(k) / (rgasref * FLOW r(k)) * pref/rhoref
      dmol = suthc1 * (stemp**1.5) / (stemp + suthc2)
      FLOW turbtvisc(k) =  visref * FLOW mut(k) + dmol
    ENDDO
  ENDIF

  cur(itime)%v => FLOW turbtvisc

END SUBROUTINE Calc_TurbTVisc

!#######################################################################
  SUBROUTINE Calc_QCrit()
!#######################################################################
  USE common_area, ONLY: liblank, pfw, itime, cur
  USE calc_dens, ONLY: Calc_Density

  INTEGER :: cel1(1)   ! Node pointers for tetrahedral cells
  INTEGER :: cel2(1)   ! Node pointers for pyramid cells
  INTEGER :: cel3(1)   ! Node pointers for prism cells
  INTEGER :: cel4(1)   ! Node pointers for hexahedral cells
  INTEGER :: nptet(1)  ! Node pointers for poly-tet strips
  INTEGER :: ptet(1)   ! Poly-tet strips
  INTEGER :: blocks(1) ! Node pointers for structured block definition
  INTEGER :: iblank(1) ! iblank values for structured blocks

  INTEGER :: kcel1  ! Number of tetrahedral cells
  INTEGER :: kcel2  ! Number of pyramid cells
  INTEGER :: kcel3  ! Number of prism cells
  INTEGER :: kcel4  ! Number of hexahedral cells
  INTEGER :: knptet ! Number of poly-tet strips
  INTEGER :: kptet  ! Number of poly-tet strips
  INTEGER :: knblck ! Number of structured blocks
  INTEGER :: iblnkf ! 0 - No IBlanking, 1 - IBlanking
  INTEGER :: nnode  ! number of nodes
  INTEGER :: ctr,k    ! Dummy counter

  REAL :: xyz(1)   ! coordinates of grid
  REAL :: v(1)     ! vector values
  REAL,ALLOCATABLE,DIMENSION(:) :: swork ! Scalar field work array
  REAL,ALLOCATABLE,DIMENSION(:,:) :: vwork ! Vector field work array

  POINTER (pcel1,cel1),(pcel2,cel2),(pcel3,cel3),(pcel4,cel4)
  POINTER (pnptet,nptet),(pptet,ptet)
  POINTER (pblocks,blocks),(piblank,iblank)
  POINTER (pxyz,xyz),(pv,v)

  IF(.NOT. ASSOCIATED(FLOW qcrit))THEN
    IF(liblank)THEN
      iblnkf = 1
    ELSE
      iblnkf = 0
    ENDIF
    CALL GETSTRUC(301, pxyz, nnode)
    CALL GETSTRUC(304, pv, nnode)
    CALL GETSTRUC(306, piblank, ctr)
    CALL GETSTRUC(311, pcel1, kcel1)
    CALL GETSTRUC(312, pcel2, kcel2)
    CALL GETSTRUC(313, pcel3, kcel3)
    CALL GETSTRUC(314, pcel4, kcel4)
    CALL GETSTRUC(315, pnptet, knptet)
    CALL GETSTRUC(316, pptet, kptet)
    CALL GETSTRUC(317, pblocks, knblck)
    ALLOCATE(vwork(3,nnode), FLOW qcrit(nnode), swork(nnode))
    CALL V3Grad(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&           knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,FLOW p,    &
&           nnode,vwork,swork)
    CALL V3Lapl(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&           knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,FLOW p,    &
&           vwork,nnode,FLOW qcrit,swork)
    DEALLOCATE(vwork,swork)

    CALL Calc_Density()
    DO k = 1, nnode
      FLOW qcrit(k) = FLOW qcrit(k) * .5 / FLOW r(k)
    ENDDO
  ENDIF

    cur(itime)%v => FLOW qcrit

  END SUBROUTINE Calc_QCrit

!#######################################################################
  SUBROUTINE Calc_YPlus(kn,n)
!#######################################################################
! This is a surface calculation only.
    USE common_area, ONLY: runmode, pfw, itime, suthc1, suthc2, rgasref

    INTEGER, INTENT(IN) :: n, kn(n)

    REAL :: kcmu, dmol, stemp
    INTEGER :: i,m

    IF(.NOT. ASSOCIATED(FLOW yplus))THEN
      SELECT CASE(runmode)
        CASE(-1,1,2)
          ALLOCATE(FLOW yplus(n))
          DO i = 1, n
            m = kn(i)
            kcmu = SQRT(FLOW mut(m)*FLOW eps(m)/FLOW k(m))
            stemp = FLOW p(m) / (rgasref * FLOW r(m))
            dmol = suthc1 * (stemp**1.5) / (stemp + suthc2)
            FLOW yplus(i) = SQRT(FLOW r(m)/dmol*kcmu)
          ENDDO
      END SELECT
    ENDIF

  END SUBROUTINE Calc_YPlus

END MODULE calc_turb
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
