#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_vel
  IMPLICIT NONE
CONTAINS
!#######################################################################
SUBROUTINE ScalVelocity(v,key,mvmax)
!#######################################################################
  USE common_area, ONLY: pfw, itime, vref
  USE mod_events, ONLY: event_get_secondary

  INTEGER,INTENT(IN) :: key,mvmax
  REAL,INTENT(INOUT) :: v(*)

  INTEGER :: k
  REAL,POINTER,DIMENSION(:) :: ptr

  SELECT CASE(event_get_secondary(key))
    CASE(1)
      CALL Calc_VelocityTheta()
      ptr => FLOW vt
    CASE(2)
      CALL Calc_VelocityThetaRel()
      ptr => FLOW vthrel
    CASE(3)
      CALL Calc_VelocityRadial()
      ptr => FLOW vr
    CASE(4)
      CALL Calc_VelocityTAbs()
      ptr => FLOW vtabs
    CASE(5)
      CALL Calc_VelocityTRel()
      ptr => FLOW vtrel
    CASE(6)
      CALL Calc_VelocityXAbs()
      ptr => FLOW vxabs
    CASE(7)
      CALL Calc_VelocityXRel()
      ptr => FLOW vx
    CASE(8)
      CALL Calc_VelocityYAbs()
      ptr => FLOW vyabs
    CASE(9)
      CALL Calc_VelocityYRel()
      ptr => FLOW vy
    CASE(10)
      CALL Calc_VelocityZ()
      ptr => FLOW vz
    CASE DEFAULT
      RETURN
  END SELECT

  DO k = 1, mvmax
    v(k) = vref * ptr(k)
  ENDDO
  NULLIFY(ptr)

END SUBROUTINE ScalVelocity

!#######################################################################
SUBROUTINE Calc_VelocityTheta()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur
  USE calc_coord, ONLY: Calc_CoordT
  IMPLICIT NONE

  INTEGER m,mvmax

  IF(.NOT. ASSOCIATED(FLOW vt))THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%t0))CALL Calc_CoordT()

    mvmax = pgr(igrid)%tpts
    ALLOCATE(FLOW vt(mvmax))
    DO m = 1, mvmax
      FLOW vt(m) = FLOW vx(m)*SIN(pgr(igrid)%t0(m)) + &
                         FLOW vy(m)*COS(pgr(igrid)%t0(m))
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vt

  RETURN

END SUBROUTINE Calc_VelocityTheta

!#######################################################################
SUBROUTINE Calc_VelocityThetaRel()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur, omega
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vthrel))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vthrel(mvmax))
    DO k=1,mvmax
      FLOW vthrel(k)=FLOW vt(k)-omega*pgr(igrid)%r(k)
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vthrel

  RETURN

END SUBROUTINE Calc_VelocityThetaRel

!#######################################################################
SUBROUTINE Calc_VelocityRadial()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur
  USE calc_coord, ONLY: Calc_CoordT
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vr))THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%t0))CALL Calc_CoordT()

    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vr(mvmax))
    DO k=1,mvmax
      FLOW vr(k) = FLOW vy(k)*SIN(pgr(igrid)%t0(k)) - &
                         FLOW vx(k)*COS(pgr(igrid)%t0(k))
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vr

  RETURN

END SUBROUTINE Calc_VelocityRadial

!#######################################################################
SUBROUTINE Calc_VelocityTAbs()
!#######################################################################
  USE common_area, ONLY: runmode, pfw, pgr, igrid, itime, cur

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vtabs))THEN
    SELECT CASE(runmode)
      CASE(3,4,5,6)
        CALL Calc_VelocityTRel
        FLOW vtabs => FLOW vtrel
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW vtabs(mvmax))
        DO k=1,mvmax
          FLOW vtabs(k)=SQRT(FLOW vt(k)**2+FLOW vr(k)**2+ &
                                   FLOW vz(k)**2)
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vtabs

  RETURN

END SUBROUTINE Calc_VelocityTAbs

!#######################################################################
SUBROUTINE Calc_VelocityTRel()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, coordsys, omega, cur

  INTEGER k,mvmax
  REAL wt

  IF(.NOT. ASSOCIATED(FLOW vtrel))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vtrel(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW vtrel(k)=SQRT(FLOW vx(k)**2+FLOW vy(k)**2+FLOW vz(k)**2)
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          wt=FLOW vt(k)-omega*pgr(igrid)%r(k)
          FLOW vtrel(k)=SQRT(wt**2+FLOW vr(k)**2+FLOW vz(k)**2)
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vtrel

END SUBROUTINE Calc_VelocityTRel

!#######################################################################
SUBROUTINE Calc_VelocityXAbs()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur, coordsys

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vxabs))THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_VelocityXRel
        FLOW vxabs => FLOW vx
      CASE(RHCYLINDRICAL)
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW vxabs(mvmax))
        DO k=1,mvmax
          FLOW vxabs(k)=FLOW vr(k)*COS(pgr(igrid)%t(k)) -  &
                        FLOW vt(k)*SIN(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW vxabs(mvmax))
        DO k=1,mvmax
          FLOW vxabs(k)=FLOW vr(k)*SIN(pgr(igrid)%t(k)) +  &
                        FLOW vt(k)*COS(pgr(igrid)%t(k))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vxabs

END SUBROUTINE Calc_VelocityXAbs

!#######################################################################
SUBROUTINE Calc_VelocityXRel()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur, omega, coordsys

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vx))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vx(mvmax))
    SELECT CASE(coordsys)
      CASE(RHCYLINDRICAL)
        DO k=1,mvmax
          FLOW vx(k)=FLOW vr(k)*COS(pgr(igrid)%t(k)) -   &
                     (FLOW vt(k)-omega*pgr(igrid)%r(k))* &
                     SIN(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW vx(k)=FLOW vr(k)*SIN(pgr(igrid)%t(k)) +   &
                     (FLOW vt(k)-omega*pgr(igrid)%r(k))* &
                     COS(pgr(igrid)%t(k))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vx

END SUBROUTINE Calc_VelocityXRel

!#######################################################################
SUBROUTINE Calc_VelocityYAbs()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur, coordsys

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vyabs))THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        CALL Calc_VelocityYRel
        FLOW vyabs => FLOW vy
      CASE(RHCYLINDRICAL)
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW vyabs(mvmax))
        DO k=1,mvmax
          FLOW vyabs(k)=-FLOW vr(k)*SIN(pgr(igrid)%t(k)) -  &
                        FLOW vt(k)*COS(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        mvmax=pgr(igrid)%tpts
        ALLOCATE(FLOW vyabs(mvmax))
        DO k=1,mvmax
          FLOW vyabs(k)=FLOW vr(k)*COS(pgr(igrid)%t(k)) -  &
                        FLOW vt(k)*SIN(pgr(igrid)%t(k))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vyabs

END SUBROUTINE Calc_VelocityYAbs

!#######################################################################
SUBROUTINE Calc_VelocityYRel()
!#######################################################################
  USE common_area, ONLY: pgr, igrid, pfw, itime, cur, omega, coordsys

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vy))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vy(mvmax))
    SELECT CASE(coordsys)
      CASE(RHCYLINDRICAL)
        DO k=1,mvmax
          FLOW vy(k)=-FLOW vr(k)*SIN(pgr(igrid)%t(k)) -  &
                     (FLOW vt(k)-omega*pgr(igrid)%r(k))* &
                     COS(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW vy(k)=FLOW vr(k)*COS(pgr(igrid)%t(k)) -   &
                     (FLOW vt(k)-omega*pgr(igrid)%r(k))* &
                     SIN(pgr(igrid)%t(k))
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vy

END SUBROUTINE Calc_VelocityYRel

!#######################################################################
  SUBROUTINE Calc_VelocityZ()
!#######################################################################
    USE common_area, ONLY: pfw, itime, cur

    IF(.NOT. ASSOCIATED(FLOW vz))THEN
    ENDIF

    cur(itime)%v => FLOW vz

  END SUBROUTINE Calc_VelocityZ
END MODULE calc_vel
