#include "parameters.h"
#define FLOW pfw(itime)%
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE calc_vmth
  IMPLICIT NONE

CONTAINS

!#######################################################################
  SUBROUTINE ScalVizmath(v, key, mvmax)
!#######################################################################
    USE keydata, ONLY: uvmstat, KeyAllocStatVM
    USE common_area, ONLY: pfw, itime, coordsys, Clear_Mem_VM, Clear_Mem_Flow
    USE mod_events, ONLY: event_get_primary, event_get_secondary

    INTEGER,INTENT(IN) :: key, mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k, ctr
    REAL,POINTER,DIMENSION(:) :: ptr

    ctr = event_get_primary(key) - 110

    CALL Clear_Mem_Flow
    CALL KeyAllocStatVM(ctr)
    IF(.NOT. uvmstat(ctr))THEN
      uvmstat(ctr)=.TRUE.
      CALL Clear_Mem_VM
    ENDIF

    CALL Calc_VM(ctr)
    SELECT CASE(ctr)
      CASE(1,3)
        SELECT CASE(event_get_secondary(key))
          CASE(1)
            CALL Calc_VMTheta
            ptr => FLOW vmatht
          CASE(2)
            CALL Calc_VMRadial
            ptr => FLOW vmathr
          CASE(3)
            CALL Calc_VMStreamAbs
            ptr => FLOW vmathsabs
          CASE(4)
            CALL Calc_VMStreamRel
            ptr => FLOW vmathsrel
          CASE(5)
            CALL Calc_VMTotalAbs
            ptr => FLOW vmathtabs
          CASE(6)
            CALL Calc_VMTotalRel
            ptr => FLOW vmathtrel
          CASE(10)
            CALL Calc_VMZAbs
            ptr => FLOW vmathzabs
          CASE(7)
            SELECT CASE(coordsys)
              CASE(CYLINDRICAL)
                DO k = 1, mvmax
                  v(k) = FLOW vmath(2,k)
                ENDDO
              CASE(RHCYLINDRICAL)
                DO k = 1, mvmax
                  v(k) = -FLOW vmath(1,k)
                ENDDO
              CASE DEFAULT
                DO k = 1, mvmax
                  v(k) = FLOW vmath(1,k)
                ENDDO
            END SELECT
            RETURN
          CASE(8)
            SELECT CASE(coordsys)
              CASE(CYLINDRICAL)
                DO k = 1, mvmax
                  v(k) = FLOW vmath(1,k)
                ENDDO
              CASE DEFAULT
                DO k = 1, mvmax
                  v(k) = FLOW vmath(2,k)
                ENDDO
            END SELECT
            RETURN
          CASE(9)
            SELECT CASE(coordsys)
              CASE(CARTESIAN,RHCYLINDRICAL)
                DO k = 1, mvmax
                  v(k) = FLOW vmath(3,k)
                ENDDO
              CASE DEFAULT
                DO k = 1, mvmax
                  v(k) = -FLOW vmath(3,k) ! Changes handedness
                ENDDO
            END SELECT
            RETURN
          CASE DEFAULT
            RETURN
        END SELECT
      CASE(2,4)
        ptr => FLOW vmaths

    END SELECT
    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO

  END SUBROUTINE ScalVizmath


!#######################################################################
SUBROUTINE Calc_VM(iopt)
!#######################################################################
  USE common_area, ONLY: liblank, pfw, itime, cur
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: iopt

  INTEGER :: cel1(1)   ! Node pointers for tetrahedral cells
  INTEGER :: cel2(1)   ! Node pointers for pyramid cells
  INTEGER :: cel3(1)   ! Node pointers for prism cells
  INTEGER :: cel4(1)   ! Node pointers for hexahedral cells
  INTEGER :: nptet(1)  ! Node pointers for poly-tet strips
  INTEGER :: ptet(1)   ! Poly-tet strips
  INTEGER :: blocks(1) ! Node pointers for structured block definition
  INTEGER :: iblank(1) ! iblank values for structured blocks

  INTEGER :: kcel1  ! Number of tetrahedral cells
  INTEGER :: kcel2  ! Number of pyramid cells
  INTEGER :: kcel3  ! Number of prism cells
  INTEGER :: kcel4  ! Number of hexahedral cells
  INTEGER :: knptet ! Number of poly-tet strips
  INTEGER :: kptet  ! Number of poly-tet strips
  INTEGER :: knblck ! Number of structured blocks
  INTEGER :: iblnkf ! 0 - No IBlanking, 1 - IBlanking
  INTEGER :: nnode  ! number of nodes
  INTEGER :: ctr    ! Dummy counter

  REAL :: xyz(1)   ! coordinates of grid
  REAL :: v(1)     ! vector values
  REAL,ALLOCATABLE,DIMENSION(:) :: swork ! Scalar field work array

  POINTER (pcel1,cel1),(pcel2,cel2),(pcel3,cel3),(pcel4,cel4)
  POINTER (pnptet,nptet),(pptet,ptet)
  POINTER (pblocks,blocks),(piblank,iblank)
  POINTER (pxyz,xyz),(pv,v)

  SELECT CASE(iopt)
    CASE(1,3)
      IF(ASSOCIATED(FLOW vmath))RETURN
    CASE(2,4)
      IF(ASSOCIATED(FLOW vmaths))RETURN
  END SELECT

  IF(liblank)THEN
    iblnkf = 1
  ELSE
    iblnkf = 0
  ENDIF

  CALL GETSTRUC(301, pxyz, nnode)
  CALL GETSTRUC(304, pv, nnode)
  CALL GETSTRUC(306, piblank, ctr)
  CALL GETSTRUC(311, pcel1, kcel1)
  CALL GETSTRUC(312, pcel2, kcel2)
  CALL GETSTRUC(313, pcel3, kcel3)
  CALL GETSTRUC(314, pcel4, kcel4)
  CALL GETSTRUC(315, pnptet, knptet)
  CALL GETSTRUC(316, pptet, kptet)
  CALL GETSTRUC(317, pblocks, knblck)

  SELECT CASE(iopt)
    CASE(1) ! Curl
      ALLOCATE(FLOW vmath(3,nnode), swork(nnode))
      CALL V3Curl(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&              knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,v,nnode,   &
&              FLOW vmath,swork)

    CASE(2) ! Divg
      ALLOCATE(FLOW vmaths(nnode), swork(nnode))
      CALL V3Divg(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&              knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,v,nnode,   &
&              FLOW vmaths,swork)

    CASE(3) ! Grad
      ALLOCATE(FLOW vmath(3,nnode), swork(nnode))
      CALL V3Grad(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&              knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,cur(itime)%v,nnode,   &
&              FLOW vmath,swork)
      
    CASE(4) ! Lapl
      ALLOCATE(FLOW vmath(3,nnode), FLOW vmaths(nnode), swork(nnode))
      CALL V3Grad(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&              knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,cur(itime)%v,nnode,   &
&              FLOW vmath,swork)
      CALL V3Lapl(cel1,kcel1,cel2,kcel2,cel3,kcel3,cel4,kcel4,nptet, &
&              knptet,ptet,blocks,knblck,iblank,iblnkf,xyz,cur(itime)%v,         &
&              FLOW vmath,nnode,FLOW vmaths,swork)

  END SELECT
  DEALLOCATE(swork)

END SUBROUTINE Calc_VM

!#######################################################################
SUBROUTINE Calc_VMTheta()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, coordsys
  IMPLICIT NONE

  INTEGER :: k, mvmax

  IF(.NOT. ASSOCIATED(FLOW vmatht))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmatht(mvmax))
    SELECT CASE(coordsys)
      CASE(CYLINDRICAL)
        DO k=1,mvmax
!       Zx cos(th) - Zy sin(th)
          FLOW vmatht(k) = FLOW vmath(2,k)*COS(pgr(igrid)%t(k)) - &
                                 FLOW vmath(1,k)*SIN(pgr(igrid)%t(k))
        ENDDO
      CASE(RHCYLINDRICAL)
!       -Zx sin(th) + Zy cos(th)
        DO k=1,mvmax
          FLOW vmatht(k) = FLOW vmath(1,k)*SIN(pgr(igrid)%t(k)) + &
                                 FLOW vmath(2,k)*COS(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW vmatht(k) = 0.
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vmatht

END SUBROUTINE Calc_VMTheta

!#######################################################################
SUBROUTINE Calc_VMRadial()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, coordsys
  IMPLICIT NONE

  INTEGER :: k, mvmax

  IF(.NOT. ASSOCIATED(FLOW vmathr))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathr(mvmax))
    SELECT CASE(coordsys)
      CASE(CYLINDRICAL)
        DO k=1,mvmax
!       Zx sin(th) + Zy cos(th)
          FLOW vmathr(k) = -FLOW vmath(2,k)*SIN(pgr(igrid)%t(k)) &
                           -FLOW vmath(1,k)*COS(pgr(igrid)%t(k))
        ENDDO
      CASE(RHCYLINDRICAL)
!       Zx cos(th) + Zy sin(th)
        DO k=1,mvmax
          FLOW vmathr(k) = -FLOW vmath(1,k)*COS(pgr(igrid)%t(k)) + &
                                 FLOW vmath(2,k)*SIN(pgr(igrid)%t(k))
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW vmathr(k) = 0.
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vmathr

END SUBROUTINE Calc_VMRadial

!#######################################################################
SUBROUTINE Calc_VMTotalAbs()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, omega
  IMPLICIT NONE

  INTEGER :: k, mvmax
  REAL :: zetaza

  IF(.NOT. ASSOCIATED(FLOW vmathtabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathtabs(mvmax))
    DO k=1,mvmax
      zetaza = -FLOW vmath(3,k)-2.*omega
      FLOW vmathtabs(k) = SQRT(FLOW vmath(1,k)**2 + &
         FLOW vmath(2,k)**2 + zetaza**2)
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vmathtabs

  RETURN

END SUBROUTINE Calc_VMTotalAbs

!#######################################################################
SUBROUTINE Calc_VMTotalRel()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur
  IMPLICIT NONE

  INTEGER :: k, mvmax

  IF(.NOT. ASSOCIATED(FLOW vmathtrel))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathtrel(mvmax))
    DO k=1,mvmax
      FLOW vmathtrel(k) = SQRT(FLOW vmath(1,k)**2 + &
         FLOW vmath(2,k)**2 + FLOW vmath(3,k)**2)
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vmathtrel

  RETURN

END SUBROUTINE Calc_VMTotalRel

!#######################################################################
SUBROUTINE Calc_VMZAbs()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, omega, coordsys
  IMPLICIT NONE

  INTEGER :: k, mvmax

  IF(.NOT. ASSOCIATED(FLOW vmathzabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathzabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW vmathzabs(k) = FLOW vmath(3,k)-2.*omega
        ENDDO
      CASE DEFAULT
        DO k=1,mvmax
          FLOW vmathzabs(k) = -FLOW vmath(3,k)+2.*omega
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vmathzabs

  RETURN

END SUBROUTINE Calc_VMZAbs

!#######################################################################
SUBROUTINE Calc_VMStreamAbs()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, omega, coordsys
  IMPLICIT NONE

  INTEGER :: k, mvmax
  REAL :: zetaza, vmag

  IF(.NOT. ASSOCIATED(FLOW vmathsabs))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathsabs(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          zetaza = -FLOW vmath(3,k)-2.*omega
          vmag = SQRT(FLOW vx(k)**2 + FLOW vy(k)**2 + FLOW vz(k)**2)
          FLOW vmathsabs(k) = 0.
        ENDDO
      CASE DEFAULT
        IF(.NOT. ASSOCIATED(FLOW vmathr))CALL Calc_VMRadial
        IF(.NOT. ASSOCIATED(FLOW vmatht))CALL Calc_VMTheta
        DO k=1,mvmax
          zetaza = -FLOW vmath(3,k)-2.*omega
          vmag = SQRT(FLOW vr(k)**2 + FLOW vt(k)**2 + FLOW vz(k)**2)
          FLOW vmathsabs(k) = (FLOW vmathr(k)*FLOW vr(k) + &
                                     FLOW vmatht(k)*FLOW vt(k) + &
                                     zetaza*FLOW vz(k))/vmag
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vmathsabs

END SUBROUTINE Calc_VMStreamAbs

!#######################################################################
  SUBROUTINE Calc_VMStreamRel()
!#######################################################################
  USE common_area, ONLY: pfw, itime, pgr, igrid, cur, coordsys, omega
  IMPLICIT NONE

  INTEGER :: k, mvmax
  REAL :: wt, vmag

  IF(.NOT. ASSOCIATED(FLOW vmathsrel))THEN
    mvmax=pgr(igrid)%tpts
    ALLOCATE(FLOW vmathsrel(mvmax))
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO k=1,mvmax
          FLOW vmathsrel(k) = 0.
        ENDDO
      CASE DEFAULT
        IF(.NOT. ASSOCIATED(FLOW vmathr))CALL Calc_VMRadial
        IF(.NOT. ASSOCIATED(FLOW vmatht))CALL Calc_VMTheta
        DO k=1,mvmax
          wt = FLOW vt(k) - omega*pgr(igrid)%r(k)
          vmag = SQRT(FLOW vr(k)**2 + FLOW vt(k)**2 + FLOW vz(k)**2)
          FLOW vmathsrel(k) = (FLOW vmathr(k)*FLOW vr(k) + &
                                     FLOW vmatht(k)*wt               - &
                                     FLOW vmath(3,k)*FLOW vz(k))/vmag
        ENDDO
    END SELECT
  ENDIF

  cur(itime)%v => FLOW vmathsrel

  END SUBROUTINE Calc_VMStreamRel

END MODULE calc_vmth
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
