#include "parameters.h"
#define FLOW pfw(itime)%
MODULE calc_vrtc
  IMPLICIT NONE
CONTAINS
!#######################################################################
  SUBROUTINE ScalVorticity(v, key, mvmax)
!#######################################################################
    USE common_area, ONLY: pfw, itime, runmode, ffmt
    USE calc_vmth, ONLY: ScalVizmath
    USE mod_events, ONLY: event_get_secondary

    INTEGER,INTENT(IN) :: key, mvmax
    REAL,INTENT(INOUT) :: v(*)
    INTEGER :: k, key2
    REAL,POINTER,DIMENSION(:) :: ptr

    key2 = event_get_secondary(key)

    DEBUG3('     \ ScalVorticity')
    IF((runmode.EQ.-1 .OR. runmode.EQ.1 .OR. runmode.EQ.2) .AND. ffmt.EQ.FFMTVSTAGE)THEN
      SELECT CASE(key2)
        CASE(1)
          CALL Calc_VorticityT()
          ptr => FLOW vortt
        CASE(2)
          CALL Calc_VorticityR()
          ptr => FLOW vortr
        CASE(3)
          CALL Calc_VorticitySAbs()
          ptr => FLOW vortsabs
        CASE(4)
          CALL Calc_VorticitySRel()
          ptr => FLOW vortsrel
        CASE(5)
          CALL Calc_VorticityTAbs()
          ptr => FLOW vorttabs
        CASE(6)
          CALL Calc_VorticityTRel()
          ptr => FLOW vorttrel
        CASE(7)
          CALL Calc_VorticityX()
          ptr => FLOW vortx
        CASE(8)
          CALL Calc_VorticityY()
          ptr => FLOW vorty
        CASE(9)
          CALL Calc_VorticityZAbs()
          ptr => FLOW vortzabs
        CASE(10)
          CALL Calc_VorticityZRel()
          ptr => FLOW vortz
        CASE DEFAULT
          RETURN
      END SELECT
    ELSE
      IF(key2.GT.0 .AND. key2.LT.11) CALL ScalVizmath(v, 11100+key2, mvmax)
      DEBUG3('     / ScalVorticity')
      RETURN
    ENDIF
    DO k = 1, mvmax
      v(k) = ptr(k)
    ENDDO
    NULLIFY(ptr)
    DEBUG3('     / ScalVorticity')

  END SUBROUTINE ScalVorticity

!#######################################################################
SUBROUTINE Calc_VorticityT()
!#######################################################################
  USE common_area

  IF(.NOT. ASSOCIATED(FLOW vortt))THEN
    CALL Calc_Vort()
  ENDIF

  cur(itime)%v => FLOW vortt

END SUBROUTINE Calc_VorticityT

!#######################################################################
SUBROUTINE Calc_VorticityR()
!#######################################################################
  USE common_area

  IF(.NOT. ASSOCIATED(FLOW vortr))THEN
    CALL Calc_Vort()
  ENDIF

  cur(itime)%v => FLOW vortr

END SUBROUTINE Calc_VorticityR

!#######################################################################
SUBROUTINE Calc_VorticitySAbs()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER mvmax,k
  REAL zetaza,vmag

  IF(.NOT. ASSOCIATED(FLOW vortsabs))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vortsabs(mvmax))
    DO k=1,mvmax
      zetaza=FLOW vortz(k)-2.*omega
      vmag=sqrt(FLOW vr(k)**2+FLOW vt(k)**2+FLOW vz(k)**2)
      FLOW vortsabs(k)=(FLOW vr(k)*FLOW vortr(k)+ &
         FLOW vt(k)*FLOW vortt(k)+FLOW vz(k)*zetaza)/vmag
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vortsabs

END SUBROUTINE Calc_VorticitySAbs

!#######################################################################
SUBROUTINE Calc_VorticitySRel()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER mvmax,k
  REAL wt,vmag

  IF(.NOT. ASSOCIATED(FLOW vortsrel))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vortsrel(mvmax))
    DO k=1,mvmax
      wt=FLOW vt(k)-omega*pgr(igrid)%r(k)
      vmag =SQRT(FLOW vr(k)**2+wt**2+FLOW vz(k)**2)
      FLOW vortsrel(k)=(FLOW vr(k)*FLOW vortr(k)+ &
         wt*FLOW vortt(k)+FLOW vz(k)*FLOW vortz(k))/vmag
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vortsrel

END SUBROUTINE Calc_VorticitySRel

!#######################################################################
SUBROUTINE Calc_VorticityTAbs()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  REAL zetaza
  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vorttabs))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vorttabs(mvmax))
    DO k=1,mvmax
      zetaza=FLOW vortz(k)-2.*omega
      FLOW vorttabs(k)=SQRT(FLOW vortr(k)**2 + &
                                  FLOW vortt(k)**2 + &
                                  zetaza**2)
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vorttabs

END SUBROUTINE Calc_VorticityTAbs

!#######################################################################
SUBROUTINE Calc_VorticityTRel()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vorttrel))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vorttrel(mvmax))
    DO k=1,mvmax
      FLOW vorttrel(k)=SQRT(FLOW vortr(k)**2 + &
                                  FLOW vortt(k)**2 + &
                                  FLOW vortz(k)**2)
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vorttrel

END SUBROUTINE Calc_VorticityTRel

!#######################################################################
SUBROUTINE Calc_VorticityX()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vortx))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vortx(mvmax))
    DO k=1,mvmax
      FLOW vortx(k)=FLOW vortr(k)*SIN(pgr(igrid)%t(k)) + &
                          FLOW vortt(k)*COS(pgr(igrid)%t(k))
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vortx

END SUBROUTINE Calc_VorticityX

!#######################################################################
SUBROUTINE Calc_VorticityY()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vorty))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vorty(mvmax))
    DO k=1,mvmax
      FLOW vorty(k)=FLOW vortr(k)*COS(pgr(igrid)%t(k)) - &
                          FLOW vortt(k)*SIN(pgr(igrid)%t(k))
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vorty

END SUBROUTINE Calc_VorticityY

!#######################################################################
SUBROUTINE Calc_VorticityZAbs()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  INTEGER k,mvmax

  IF(.NOT. ASSOCIATED(FLOW vortzabs))THEN
    IF(.NOT. ASSOCIATED(FLOW vortz))CALL Calc_Vort()

    mvmax=pgr(igrid)%mvmax(1) ! Just for single block vstage
    ALLOCATE(FLOW vortzabs(mvmax))
    DO k=1,mvmax
      FLOW vortzabs(k) = FLOW vortz(k) - 2.*omega
    ENDDO
  ENDIF

  cur(itime)%v => FLOW vortzabs

END SUBROUTINE Calc_VorticityZAbs

!#######################################################################
SUBROUTINE Calc_VorticityZRel()
!#######################################################################
  USE common_area
  IMPLICIT NONE

  IF(.NOT. ASSOCIATED(FLOW vortz))THEN
    CALL Calc_Vort()
  ENDIF

  cur(itime)%v => FLOW vortz

END SUBROUTINE Calc_VorticityZRel



!#######################################################################
SUBROUTINE Calc_Vort()
!#######################################################################
! Subroutine VORTCALC computes the vertex vorticities.
! All vorticity quantities are computed as relative quantities. It is
! assumed that if the absolute quantities are wanted, OMEGA will have
! been set to zero in the calling program.
!
! This routine only works for a single block, and assumes it is a
! standard vstage passage.  There will be a better way....
  USE common_area
  USE calc_coord, ONLY: Calc_CoordX, Calc_CoordY
  USE calc_vel, ONLY: Calc_VelocityXAbs, Calc_VelocityYAbs
  IMPLICIT NONE

  INTEGER j,ncr,nbb,nsp,nfl,nle,nte,nhb,ntp,mvmax,ctr
  INTEGER :: i1,i2,i3,i4,i5,i6,i7,i8
  INTEGER :: ifl,isp,ibb

  REAL costh,sinth
  REAL,POINTER,DIMENSION(:) :: x,y,z,vx,vy,vz,theta
  REAL :: ABBX1,ABBY1,ABBZ1,ABBX2,ABBY2,ABBZ2,aspx1,aspy1
  REAL :: vbbx1,vbby1,vbbz1,vbbx2,vbby2,vbbz2
  REAL :: aspz1,aspx2,aspy2,aspz2
  REAL :: areas,volumes
  REAL :: a1,a2,a3,a4,b1,b2,b3,b4,c1,c2,c3,c4
  REAL :: vspx1,vspy1,vspz1,vspx2,vspy2,vspz2
  REAL :: aflx1,afly1,aflz1,aflx2,afly2,aflz2
  REAL :: vflx1,vfly1,vflz1,vflx2,vfly2,vflz2
  REAL :: dvxdy,dvxdz,dvydx,dvydz,dvzdx,dvzdy
  REAL :: thetaa,vol,zetax,zetay

  LOGICAL :: keepxyv(2)

!**  statement function
!*  computes the projected face area*2.
  AREAS(A1,A2,A3,A4,B1,B2,B3,B4) = (A4-A1)*(B3-B2)-(A3-A2)*(B4-B1)
!*  computes the volumes
  VOLUMES(A1,A2,A3,A4,B1,B2,B3,B4,C1,C2,C3,C4) =           &
         (A1+A2+A3+A4)*((B4-B1)*(C3-C2)-(C4-C1)*(B3-B2)) + &
         (B1+B2+B3+B4)*((C4-C1)*(A3-A2)-(A4-A1)*(C3-C2)) + &
         (C1+C2+C3+C4)*((A4-A1)*(B3-B2)-(B4-B1)*(A3-A2))
!-----------------------------------------------------------------------
! Set some parameters
  nbb=pgr(igrid)%li(1)
  nsp=pgr(igrid)%lj(1)
  nfl=pgr(igrid)%lk(1)
  nle=pgr(igrid)%mle
  nte=pgr(igrid)%mte
  nhb=pgr(igrid)%mhb
  ntp=pgr(igrid)%mtp
  mvmax=pgr(igrid)%mvmax(1)
!** COMPUTE INDEXING CONSTANTS
  ncr = nbb*nsp
!=======================================================================
! COMPUTE  THE  CARTESIAN   COORINATES   AND  VELOCITIES
!.. Compute the Cartesian coordinates and (optionally) the vertex
!   velocities
  ctr=0
  NULLIFY(x,y,z,vx,vy,vz,theta)
  IF(.NOT. ASSOCIATED(pgr(igrid)%x))CALL Calc_CoordX()
  IF(.NOT. ASSOCIATED(pgr(igrid)%y))CALL Calc_CoordY()
  keepxyv(1)=.TRUE.
  IF(.NOT. ASSOCIATED(FLOW vxabs))THEN
    keepxyv(1)=.FALSE.
    CALL Calc_VelocityXAbs()
  ENDIF
  keepxyv(2)=.TRUE.
  IF(.NOT. ASSOCIATED(FLOW vyabs))THEN
    keepxyv(2)=.FALSE.
    CALL Calc_VelocityYAbs()
  ENDIF
  x=>pgr(igrid)%x
  y=>pgr(igrid)%y
  z=>pgr(igrid)%z
  vx=>FLOW vxabs
  vy=>FLOW vyabs
  vz=>FLOW vz
  theta=>pgr(igrid)%t

!=======================================================================
! COMPUTE  THE  RAW  CELL  VORTICITIES
  ALLOCATE(FLOW vortr(mvmax),FLOW vortt(mvmax), &
           FLOW vortz(mvmax))
  DO IFL=1,NFL-1
    DO ISP=1,NSP-1
      DO IBB=1,NBB-1
        J = IBB + (ISP-1)*NBB+(IFL-1)*NCR  !      6-------------8
!.. Set up 8 corner index                  !     /|            /|
        I1 = J                             !    / |           / |
        I2 = J+1                           !   5-------------7  |
        I3 = J+NCR                         !   |  |          |  |
        I4 = J+NCR+1                       !   |  2----------|--4
        I5 = J+NBB                         !   | /           | /
        I6 = J+NBB+1                       !   |/            |/
        I7 = J+NCR+NBB                     !   1-------------3
        I8 = J+NCR+NBB+1                   !
!.. Compute the blade-to-blade areas * 2 and face velocities * 4
!*      AREAS(A1,A2,A3,A4,B1,B2,B3,B4) = (A4-A1)*(B3-B2)-(A3-A2)*(B4-B1)
        ABBX1 = AREAS(Y(I1),Y(I3),Y(I5),Y(I7),Z(I1),Z(I3),Z(I5),Z(I7))
        ABBY1 = AREAS(Z(I1),Z(I3),Z(I5),Z(I7),X(I1),X(I3),X(I5),X(I7))
        ABBZ1 = AREAS(X(I1),X(I3),X(I5),X(I7),Y(I1),Y(I3),Y(I5),Y(I7))
        ABBX2 = AREAS(Y(I2),Y(I4),Y(I6),Y(I8),Z(I2),Z(I4),Z(I6),Z(I8))
        ABBY2 = AREAS(Z(I2),Z(I4),Z(I6),Z(I8),X(I2),X(I4),X(I6),X(I8))
        ABBZ2 = AREAS(X(I2),X(I4),X(I6),X(I8),Y(I2),Y(I4),Y(I6),Y(I8))
        VBBX1 = VX(I1)+VX(I3)+VX(I5)+VX(I7)
        VBBY1 = VY(I1)+VY(I3)+VY(I5)+VY(I7)
        VBBZ1 = VZ(I1)+VZ(I3)+VZ(I5)+VZ(I7)
        VBBX2 = VX(I2)+VX(I4)+VX(I6)+VX(I8)
        VBBY2 = VY(I2)+VY(I4)+VY(I6)+VY(I8)
        VBBZ2 = VZ(I2)+VZ(I4)+VZ(I6)+VZ(I8)
!.. Compute the spanwise areas * 2 and face velocities * 4
        ASPX1 = AREAS(Y(I1),Y(I2),Y(I3),Y(I4),Z(I1),Z(I2),Z(I3),Z(I4))
        ASPY1 = AREAS(Z(I1),Z(I2),Z(I3),Z(I4),X(I1),X(I2),X(I3),X(I4))
        ASPZ1 = AREAS(X(I1),X(I2),X(I3),X(I4),Y(I1),Y(I2),Y(I3),Y(I4))
        ASPX2 = AREAS(Y(I5),Y(I6),Y(I7),Y(I8),Z(I5),Z(I6),Z(I7),Z(I8))
        ASPY2 = AREAS(Z(I5),Z(I6),Z(I7),Z(I8),X(I5),X(I6),X(I7),X(I8))
        ASPZ2 = AREAS(X(I5),X(I6),X(I7),X(I8),Y(I5),Y(I6),Y(I7),Y(I8))
        VSPX1 = VX(I1)+VX(I2)+VX(I3)+VX(I4)
        VSPY1 = VY(I1)+VY(I2)+VY(I3)+VY(I4)
        VSPZ1 = VZ(I1)+VZ(I2)+VZ(I3)+VZ(I4)
        VSPX2 = VX(I5)+VX(I6)+VX(I7)+VX(I8)
        VSPY2 = VY(I5)+VY(I6)+VY(I7)+VY(I8)
        VSPZ2 = VZ(I5)+VZ(I6)+VZ(I7)+VZ(I8)
!.. Compute the streamwise areas * 2 and face velocities * 4
        AFLX1 = AREAS(Y(I1),Y(I5),Y(I2),Y(I6),Z(I1),Z(I5),Z(I2),Z(I6))
        AFLY1 = AREAS(Z(I1),Z(I5),Z(I2),Z(I6),X(I1),X(I5),X(I2),X(I6))
        AFLZ1 = AREAS(X(I1),X(I5),X(I2),X(I6),Y(I1),Y(I5),Y(I2),Y(I6))
        AFLX2 = AREAS(Y(I3),Y(I7),Y(I4),Y(I8),Z(I3),Z(I7),Z(I4),Z(I8))
        AFLY2 = AREAS(Z(I3),Z(I7),Z(I4),Z(I8),X(I3),X(I7),X(I4),X(I8))
        AFLZ2 = AREAS(X(I3),X(I7),X(I4),X(I8),Y(I3),Y(I7),Y(I4),Y(I8))
        VFLX1 = VX(I1)+VX(I5)+VX(I2)+VX(I6)
        VFLY1 = VY(I1)+VY(I5)+VY(I2)+VY(I6)
        VFLZ1 = VZ(I1)+VZ(I5)+VZ(I2)+VZ(I6)
        VFLX2 = VX(I3)+VX(I7)+VX(I4)+VX(I8)
        VFLY2 = VY(I3)+VY(I7)+VY(I4)+VY(I8)
        VFLZ2 = VZ(I3)+VZ(I7)+VZ(I4)+VZ(I8)
!.. Compute the velocity gradients * 8
        DVXDY = ABBY2*VBBX2-ABBY1*VBBX1 + ASPY2*VSPX2-ASPY1*VSPX1 + &
                AFLY2*VFLX2-AFLY1*VFLX1
        DVXDZ = ABBZ2*VBBX2-ABBZ1*VBBX1 + ASPZ2*VSPX2-ASPZ1*VSPX1 + &
                AFLZ2*VFLX2-AFLZ1*VFLX1
        DVYDX = ABBX2*VBBY2-ABBX1*VBBY1 + ASPX2*VSPY2-ASPX1*VSPY1 + &
                AFLX2*VFLY2-AFLX1*VFLY1
        DVYDZ = ABBZ2*VBBY2-ABBZ1*VBBY1 + ASPZ2*VSPY2-ASPZ1*VSPY1 + &
                AFLZ2*VFLY2-AFLZ1*VFLY1
        DVZDX = ABBX2*VBBZ2-ABBX1*VBBZ1 + ASPX2*VSPZ2-ASPX1*VSPZ1 + &
                AFLX2*VFLZ2-AFLX1*VFLZ1
        DVZDY = ABBY2*VBBZ2-ABBY1*VBBZ1 + ASPY2*VSPZ2-ASPY1*VSPZ1 + &
                AFLY2*VFLZ2-AFLY1*VFLZ1
!
!** COMPUTE THE VOLUME * 24
!
!*      VOLUMES(A1,A2,A3,A4,B1,B2,B3,B4,C1,C2,C3,C4) =
!*     &       (A1+A2+A3+A4)*((B4-B1)*(C3-C2)-(C4-C1)*(B3-B2)) +
!*     &       (B1+B2+B3+B4)*((C4-C1)*(A3-A2)-(A4-A1)*(C3-C2)) +
!*     &       (C1+C2+C3+C4)*((A4-A1)*(B3-B2)-(B4-B1)*(A3-A2))
        VOL = VOLUMES( X(I1)-X(I1),X(I2)-X(I1),X(I3)-X(I1),X(I4)-X(I1),  &
                       Y(I1)-Y(I1),Y(I2)-Y(I1),Y(I3)-Y(I1),Y(I4)-Y(I1),  &
                       Z(I1)-Z(I1),Z(I2)-Z(I1),Z(I3)-Z(I1),Z(I4)-Z(I1) ) &
            + VOLUMES( X(I2)-X(I1),X(I1)-X(I1),X(I6)-X(I1),X(I5)-X(I1),  &
                       Y(I2)-Y(I1),Y(I1)-Y(I1),Y(I6)-Y(I1),Y(I5)-Y(I1),  &
                       Z(I2)-Z(I1),Z(I1)-Z(I1),Z(I6)-Z(I1),Z(I5)-Z(I1) ) &
            + VOLUMES( X(I6)-X(I1),X(I5)-X(I1),X(I8)-X(I1),X(I7)-X(I1),  &
                       Y(I6)-Y(I1),Y(I5)-Y(I1),Y(I8)-Y(I1),Y(I7)-Y(I1),  &
                       Z(I6)-Z(I1),Z(I5)-Z(I1),Z(I8)-Z(I1),Z(I7)-Z(I1) ) &
            + VOLUMES( X(I3)-X(I1),X(I4)-X(I1),X(I7)-X(I1),X(I8)-X(I1),  &
                       Y(I3)-Y(I1),Y(I4)-Y(I1),Y(I7)-Y(I1),Y(I8)-Y(I1),  &
                       Z(I3)-Z(I1),Z(I4)-Z(I1),Z(I7)-Z(I1),Z(I8)-Z(I1) ) &
            + VOLUMES( X(I1)-X(I1),X(I3)-X(I1),X(I5)-X(I1),X(I7)-X(I1),  &
                       Y(I1)-Y(I1),Y(I3)-Y(I1),Y(I5)-Y(I1),Y(I7)-Y(I1),  &
                       Z(I1)-Z(I1),Z(I3)-Z(I1),Z(I5)-Z(I1),Z(I7)-Z(I1) ) &
            + VOLUMES( X(I4)-X(I1),X(I2)-X(I1),X(I8)-X(I1),X(I6)-X(I1),  &
                       Y(I4)-Y(I1),Y(I2)-Y(I1),Y(I8)-Y(I1),Y(I6)-Y(I1),  &
                       Z(I4)-Z(I1),Z(I2)-Z(I1),Z(I8)-Z(I1),Z(I6)-Z(I1) )
!.. Compute the raw vorticity and the average cell theta angle
        IF(VOL.NE.0.)THEN
          ZETAX = 3.0*(DVZDY-DVYDZ)/VOL
          ZETAY = 3.0*(DVXDZ-DVZDX)/VOL
          THETAA = (THETA(I1) + THETA(I2) + THETA(I3) + THETA(I4) + &
                    THETA(I5) + THETA(I6) + THETA(I7) + THETA(I8))/8. 
          COSTH = COS(THETAA)
          SINTH = SIN(THETAA)
          FLOW vortr(j)=costh*ZETAX+sinth*ZETAY
          FLOW vortt(j)=-sinth*ZETAX+costh*ZETAY
          FLOW vortz(J) = 3.0*(DVYDX-DVXDY)/VOL+2.0*OMEGA
        ELSE
          FLOW vortr(j) = 0.
          FLOW vortt(j) = 0.
          FLOW vortz(j) = 0.
        ENDIF
      ENDDO
    ENDDO
  ENDDO
!-----------------------------------------------------------------------
! Clean up the junk
  NULLIFY(x,y,z,vx,vy,vz,theta)
  IF(.NOT.keepxyv(1))THEN
    DEALLOCATE(FLOW vxabs)
    NULLIFY(FLOW vxabs)
  ENDIF
  IF(.NOT.keepxyv(2))THEN
    DEALLOCATE(FLOW vyabs)
    NULLIFY(FLOW vyabs)
  ENDIF

!=======================================================================
! COMPUTE  THE  VERTEX  VORTICITY
  CALL VERTVORT(NBB,NSP,NFL,NLE,NTE,NHB,NTP,mvmax,1)
  CALL VERTVORT(NBB,NSP,NFL,NLE,NTE,NHB,NTP,mvmax,2)
  CALL VERTVORT(NBB,NSP,NFL,NLE,NTE,NHB,NTP,mvmax,3)

END SUBROUTINE Calc_Vort

!#######################################################################
SUBROUTINE VertVort(nbb,nsp,nfl,nle,nte,nhb,ntp,mvmax,num)
!#######################################################################
  USE common_area
  IMPLICIT NONE
!
  INTEGER,INTENT(IN) :: nbb,nsp,nfl,nle,nte,nhb,ntp,mvmax,num
  INTEGER :: ifl, isp, ibb, ncr, j1, j2, j, jm, jp

  REAL, ALLOCATABLE, DIMENSION(:) :: dum1, dum2
  REAL, POINTER, DIMENSION(:) :: zeta

  ALLOCATE(dum1(mvmax),dum2(mvmax))
  IF(num.EQ.1)zeta=>FLOW vortr
  IF(num.EQ.2)zeta=>FLOW vortt
  IF(num.EQ.3)zeta=>FLOW vortz

  NCR = NBB*NSP
!
!.. Average in the blade-to-blade direction (DUM1 is a temporary storage array)
!
  DO IFL=1,NFL-1
    DO ISP=1,NSP-1
      DO IBB=1,NBB-1
        JM = MAX(1,    IBB-1)+(ISP-1)*NBB+(IFL-1)*NCR
        J  =     (     IBB  )+(ISP-1)*NBB+(IFL-1)*NCR
        JP = MIN(NBB-1,IBB  )+(ISP-1)*NBB+(IFL-1)*NCR
        DUM1(J) = 0.5*(ZETA(JM)+ZETA(JP))
      ENDDO
      IF(.NOT.(IFL.GE.NLE .AND. IFL.LT.NTE .AND. &
          ISP.GE.NHB .AND. ISP.LT.NTP))THEN
! There is a periodic boundary at ibb=1. 
        J1 = 1 + (ISP-1)*NBB+(IFL-1)*NCR
        J2 = NBB + (ISP-1)*NBB+(IFL-1)*NCR
        DUM1(J1) = 0.5*(DUM1(J1) + DUM1(J2))
        DUM1(J2) = DUM1(J1)
      END IF
    ENDDO
  ENDDO
  
!
!.. Average in the spanwise direction (DUM2 is a temporary storage array)
!
  DO IFL=1,NFL-1
    DO isp=1,nsp-1
      DO ibb=1,nbb-1
        JM = IBB+MAX(0,    ISP-2)*NBB+(IFL-1)*NCR
        J  = IBB+   (      ISP-1)*NBB+(IFL-1)*NCR
        JP = IBB+MIN(NSP-2,ISP-1)*NBB+(IFL-1)*NCR
        DUM2(J) = 0.5*(DUM1(JM)+DUM1(JP))
      ENDDO
    ENDDO
  ENDDO

!
!.. Average in the streamwise direction
!
    DO ifl=1,nfl-1
      DO isp=1,nsp-1
        DO ibb=1,nbb-1
          jm = ibb+(isp-1)*nbb+MAX(0,    ifl-2)*ncr
          j  = ibb+(isp-1)*nbb+   (      ifl-1)*ncr
          jp = ibb+(isp-1)*nbb+MIN(nfl-2,ifl-1)*ncr
          zeta(J) = 0.5*(dum2(jm)+dum2(jp))
        ENDDO
      ENDDO
    ENDDO

    DEALLOCATE(dum1,dum2) !Not zeta, or we lose vorticity
    NULLIFY(zeta)

  END SUBROUTINE VertVort

END MODULE calc_vrtc
