!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE mod_clear
  IMPLICIT NONE

  INTERFACE ClearVar
    MODULE PROCEDURE ClearVarR1, ClearVarR2, ClearVarI1, ClearVarI2, ClearVar2R1
  END INTERFACE

CONTAINS

!#######################################################################
  SUBROUTINE ClearVarR1(ptr)
!#######################################################################
    REAL,POINTER,DIMENSION(:) :: ptr

    IF(ASSOCIATED(ptr))THEN
      DEALLOCATE(ptr)
      NULLIFY(ptr)
    ENDIF

  END SUBROUTINE ClearVarR1

!#######################################################################
  SUBROUTINE ClearVarR2(ptr)
!#######################################################################
    REAL,POINTER,DIMENSION(:,:) :: ptr

    IF(ASSOCIATED(ptr))THEN
      DEALLOCATE(ptr)
      NULLIFY(ptr)
    ENDIF

  END SUBROUTINE ClearVarR2

!#######################################################################
  SUBROUTINE ClearVarI1(ptr)
!#######################################################################
    INTEGER,POINTER,DIMENSION(:) :: ptr

    IF(ASSOCIATED(ptr))THEN
      DEALLOCATE(ptr)
      NULLIFY(ptr)
    ENDIF

  END SUBROUTINE ClearVarI1

!#######################################################################
  SUBROUTINE ClearVarI2(ptr)
!#######################################################################
    INTEGER,POINTER,DIMENSION(:,:) :: ptr

    IF(ASSOCIATED(ptr))THEN
      DEALLOCATE(ptr)
      NULLIFY(ptr)
    ENDIF

  END SUBROUTINE ClearVarI2

!#######################################################################
  SUBROUTINE ClearVar2R1(ptr2, ptrb)
!#######################################################################
    REAL,POINTER,DIMENSION(:) :: ptrb,ptr2

    IF(ASSOCIATED(ptr2))THEN
      IF (.NOT.ASSOCIATED(ptr2,ptrb)) THEN
         DEALLOCATE(ptr2)
      ENDIF
      NULLIFY(ptr2)
    ENDIF

  END SUBROUTINE ClearVar2R1


END MODULE mod_clear
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE common_area
  USE mod_clear, ONLY: ClearVar
  USE typedefs
  IMPLICIT NONE

  INCLUDE 'parameters.f'

  INTEGER, PARAMETER :: nmx=100 ! Please note there is a hard limit on the
                                ! number of flow files that I am allowing
                                ! until I can dynamically allocate the
                                ! arrays that store the names.

!-----------------------------------------------------------------------
! General required parameters for normal unsteady operation
  INTEGER :: coordsys = 1 ! 0 cartesian, 1 cylindrical, 2 spherical
  INTEGER :: istat, maxblk, iopt, mirror ! Values for the pV_Init call
  INTEGER :: ngrid,igrid,ntime,itime ! Timing counters for grid/flow
  INTEGER :: runmode, & ! Solver operation mode
             ffmt,    & ! File format
             nclients   ! Total number of clients
  INTEGER :: gksurf ! Global storage of ksurf so I have it
  INTEGER :: cid ! Client id number for enrollment in PVM
                       ! and determination of files to read
  INTEGER :: dimen ! Dimensionalization flag
                   ! 0) Non-dimensional
                   ! 1) English
                   ! 2) Metric

  CHARACTER(LEN=NSTRLEN) :: usrinput ! The name of a user specified input file
  CHARACTER(LEN=NSTRLEN) :: gridarr(nmx) ! Storage of grid file names
  CHARACTER(LEN=NSTRLEN) :: flowarr(nmx) ! Storage of flow file names
  CHARACTER(LEN=NSTRLEN) :: avusbcfilename ! General bc filename
  CHARACTER(LEN=20) :: cname=' ',dname=' ' ! Client and discipline names

  LOGICAL :: lowmem ! For low-memory usage ??? Not yet implemented any-
                    ! where in the code ???
  LOGICAL :: novort ! For when the vorticity calcs crash the client
  LOGICAL :: griddouble ! Single/Double precision grids
  LOGICAL :: bcavail ! Is boundary data available?

  REAL :: rep(16) ! The replication matrix

!-----------------------------------------------------------------------
! Replication specific quantities
  INTEGER :: ntrep  ! Number of times to replicate objects
  REAL :: offsetx = 0., offsety = 0., offsetz = 0.
  LOGICAL :: trep=.TRUE.   ! Replication flag


!-----------------------------------------------------------------------
! TURBO specific quantities
  REAL :: omega = 0.0  ! This is used for the entire client!!!
                 ! not just the part that is rotating.  No mix'n'match!
  REAL :: dtdt = 0.0   ! Change in theta per change in time
  REAL :: dtheta = 0.0 ! Replication delta

  REAL :: suthc1 = 1.0, suthc1_m = 110 ! Input from TURBO's input00 that is
  REAL :: suthc2 = 1.0, suthc2_m = 110 ! used for normalization and reference
  REAL :: pref = 1.0, pref_m = 1.0
  REAL :: vref = 1.0, vref_m = 1.0
  REAL :: tref = 1.0, tref_m = 1.0
  REAL :: href = 1.0, href_m = 1.0
  REAL :: visref = 1.0, visref_m = .982143
  REAL :: rgasref = 1.0, rgasref_m = .287
  REAL :: diam = 1.0

  REAL :: tempgam(2),gamtable(2) ! Gamma interpolation values from the
  REAL :: gammabase = 1.4        ! TURBO input00 file
  REAL :: TKEinE = 0.0           ! Generally just set to 0.0

  CHARACTER*20 :: csnm  ! Case name for batch read of vstage files

  LOGICAL :: kpcell ! Determines whether or not cell-centered data is
                    ! kept.  This is useful for display.
  LOGICAL :: trot=.FALSE.  ! Turbomachinery rotation flag
  LOGICAL :: tdimen ! Sets pV3 to dimensionalize TURBO results

!-----------------------------------------------------------------------
! Plot3D specific quantities
  CHARACTER(LEN=NSTRLEN) :: gridindexfile

  LOGICAL :: liblank ! Flag for the existance of iblanking in grid files

!-----------------------------------------------------------------------
! AVUS variables
  INTEGER, POINTER, DIMENSION(:) :: npnts
  INTEGER nzones

!-----------------------------------------------------------------------
! FX Variables
  INTEGER :: FXnbc=0
  INTEGER,POINTER,DIMENSION(:) :: FXshowSurf, FXattind, FXsepind

!-----------------------------------------------------------------------
! The principle storage arrays
  TYPE(grid),POINTER,DIMENSION(:) :: pgr ! The grid
  TYPE(flow),POINTER,DIMENSION(:) :: pfw ! The flow
  TYPE(aop),POINTER,DIMENSION(:)  :: cur ! Array of pointers which are
            ! used to hit the current variable.  The primary purpose is
            ! to reduce the storage needed to compute DFT's on the data.
  TYPE(vortexcore),POINTER,DIMENSION(:) :: vcstor ! Vortex Cores
  TYPE(shocksurf),POINTER,DIMENSION(:)  :: sfstor ! Shock Surfaces
  TYPE(sepattline),POINTER,DIMENSION(:) :: sastor ! Sep & Att Lines

CONTAINS

!#######################################################################
  SUBROUTINE Clear_Mem_Flow()
!#######################################################################
! This routine can be used to clear out all of the stored variables
! that float around during allocations and such.  If a major key change
! occurs, this should happen.  Basically, if you are looking at
! one form of vorticity, we want to store it as long as possible.  The
! same is true with time averaging, and other large calculations.

    INTEGER i, j

!=======================================================================
! Clear out the variables from the grid
! These are the ones that are not always used, like i,j,k.  The xyz and
! rtz coordinates will be kept.
    DO i=1,ngrid

      CALL ClearVar(pgr(i)%i)
      CALL ClearVar(pgr(i)%j)
      CALL ClearVar(pgr(i)%k)

    ENDDO

! Clear out all non-primitive flow variables that have been stored.
! Primitives for vstage and for plot3d are not cleared.
    DO i=1,ntime
! Real Gas
      ! Not clearing g intentionally
      ! Not clearing rgas intentionally
      CALL ClearVar(pfw(i)%reyn)
      CALL ClearVar(pfw(i)%dens)

! Pressures
      CALL ClearVar(pfw(i)%ptrel)
      CALL ClearVar(pfw(i)%ptabs)

! Angles
      CALL ClearVar(pfw(i)%alphaz)
      CALL ClearVar(pfw(i)%alpham)
      CALL ClearVar(pfw(i)%betaz)
      CALL ClearVar(pfw(i)%betam)
      CALL ClearVar(pfw(i)%phi)

! Mach Numbers
      CALL ClearVar(pfw(i)%machthabs)
      CALL ClearVar(pfw(i)%machthrel)
      CALL ClearVar(pfw(i)%machr)
      CALL ClearVar(pfw(i)%machtabs)
      CALL ClearVar(pfw(i)%machtrel)
      CALL ClearVar(pfw(i)%machxabs)
      CALL ClearVar(pfw(i)%machxrel)
      CALL ClearVar(pfw(i)%machyabs)
      CALL ClearVar(pfw(i)%machyrel)
      CALL ClearVar(pfw(i)%machz)
      CALL ClearVar(pfw(i)%machideal)

! Rho-V
      CALL ClearVar(pfw(i)%rhovthabs)
      CALL ClearVar(pfw(i)%rhovthrel)
      CALL ClearVar(pfw(i)%rhovr)
      CALL ClearVar(pfw(i)%rhovtabs)
      CALL ClearVar(pfw(i)%rhovtrel)
      CALL ClearVar(pfw(i)%rhovxabs)
      CALL ClearVar(pfw(i)%rhovxrel)
      CALL ClearVar(pfw(i)%rhovyabs)
      CALL ClearVar(pfw(i)%rhovyrel)
      CALL ClearVar(pfw(i)%rhovz)

! Enthalpy
      CALL ClearVar(pfw(i)%h)
      CALL ClearVar(pfw(i)%htabs)
      CALL ClearVar(pfw(i)%htrel)

! Rothalpy
      CALL ClearVar(pfw(i)%roth)
      CALL ClearVar(pfw(i)%rothcp)

! Entropy
      CALL ClearVar(pfw(i)%sabs)
      CALL ClearVar(pfw(i)%srat)

! Temperature
      CALL ClearVar(pfw(i)%temp)
      CALL ClearVar(pfw(i)%temptabs)
      CALL ClearVar(pfw(i)%temptrel)

! Velocity
      ! Not clearing vx intentionally
      CALL ClearVar(pfw(i)%vxabs, pfw(i)%vx)
      ! Not clearing vy intentionally
      CALL ClearVar(pfw(i)%vyabs, pfw(i)%vy)
      CALL ClearVar(pfw(i)%vtabs)
      CALL ClearVar(pfw(i)%vtrel)
      CALL ClearVar(pfw(i)%vthrel)

! Vorticity
      ! Not clearing vortr intentionally
      ! Not clearing vortt intentionally
      ! Not clearing vortz intentionally
      CALL ClearVar(pfw(i)%vortsabs)
      CALL ClearVar(pfw(i)%vortsrel)
      CALL ClearVar(pfw(i)%vorttabs)
      CALL ClearVar(pfw(i)%vorttrel)
      CALL ClearVar(pfw(i)%vortx)
      CALL ClearVar(pfw(i)%vorty)
      CALL ClearVar(pfw(i)%vortzabs)

! Turbulence
      CALL ClearVar(pfw(i)%turbint)
      CALL ClearVar(pfw(i)%turbrat)
      CALL ClearVar(pfw(i)%turbtvisc)
      CALL ClearVar(pfw(i)%yplus)

! Time Average
      CALL ClearVar(pfw(i)%tmdif)

    ENDDO

! Time-Average
    CALL ClearVar(pfw(1)%tmavg)
    CALL ClearVar(pfw(1)%tmmin)
    CALL ClearVar(pfw(1)%tmmax)
    CALL ClearVar(pfw(1)%tmrms)
    CALL ClearVar(pfw(1)%tmfft)
    DO i = 1, ntime
      NULLIFY(pfw(i)%tmavg)
      NULLIFY(pfw(i)%tmmin)
      NULLIFY(pfw(i)%tmmax)
      NULLIFY(pfw(i)%tmrms)
      NULLIFY(pfw(i)%tmfft)
    ENDDO

  END SUBROUTINE Clear_Mem_Flow

!#######################################################################
  SUBROUTINE Clear_Mem_VM()
!#######################################################################
    INTEGER i

    DO i = 1, ntime

      CALL ClearVar(pfw(i)%vmath)
      CALL ClearVar(pfw(i)%vmaths)
      CALL ClearVar(pfw(i)%vmathr)
      CALL ClearVar(pfw(i)%vmatht)
      CALL ClearVar(pfw(i)%vmathsabs)
      CALL ClearVar(pfw(i)%vmathsrel)
      CALL ClearVar(pfw(i)%vmathtabs)
      CALL ClearVar(pfw(i)%vmathtrel)
      CALL ClearVar(pfw(i)%vmathzabs)

    ENDDO

  END SUBROUTINE Clear_Mem_VM

!#######################################################################
  SUBROUTINE Clear_Mem_FX()
!#######################################################################
    INTEGER i

    DO i = 1, ntime

      CALL ClearVar(vcstor(i)%segs)
      CALL ClearVar(vcstor(i)%vcoor)
      CALL ClearVar(vcstor(i)%stren)
      vcstor(i)%nseg = -1

      CALL ClearVar(sastor(i)%sepxyz)
      CALL ClearVar(sastor(i)%attxyz)
      CALL ClearVar(sastor(i)%sepind)
      CALL ClearVar(sastor(i)%attind)
      sastor(i)%calc = 0

      CALL ClearVar(sfstor(i)%pts)
      sfstor(i)%npts = -1

    ENDDO

  END SUBROUTINE Clear_Mem_FX

!#######################################################################
  SUBROUTINE Clear_Base_Mem_Flow()
!#######################################################################
! This routine clears the base variables for vstage, real gas, and
! the cartesian velocities
    INTEGER :: i

    DO i = 1, ntime

! Vstage
      CALL ClearVar(pfw(i)%p)
      CALL ClearVar(pfw(i)%r)
      CALL ClearVar(pfw(i)%vz)
      CALL ClearVar(pfw(i)%vr)
      CALL ClearVar(pfw(i)%vt)
      CALL ClearVar(pfw(i)%e)
      CALL ClearVar(pfw(i)%mut)
      CALL ClearVar(pfw(i)%k)
      CALL ClearVar(pfw(i)%eps)

! Real gas
      CALL ClearVar(pfw(i)%g)
      CALL ClearVar(pfw(i)%rgas)

! Cartesian
      CALL ClearVar(pfw(i)%vx)
      CALL ClearVar(pfw(i)%vy)

    ENDDO

  END SUBROUTINE Clear_Base_Mem_Flow

!#######################################################################
  SUBROUTINE ClearAllFlow()
!#######################################################################

    CALL Clear_Mem_Flow
    CALL Clear_Mem_VM
    CALL Clear_Mem_FX
    CALL Clear_Base_Mem_Flow
    DEALLOCATE(pfw)
    NULLIFY(pfw)

  END SUBROUTINE ClearAllFlow

!#######################################################################
  SUBROUTINE set_ref_quantities(state)
!#######################################################################
    IMPLICIT NONE

    INTEGER, INTENT(IN) :: state

    SELECT CASE(state)
! Non-dimensional
      CASE(0)
        pref = 1.0
        tref = 1.0
        vref = 1.0
        href = 1.0
        suthc1 = 1.0
        suthc2 = 1.0
        visref = .982143
        rgasref = 1.0

! Metric
      CASE(1)
        pref = pref_m
        tref = tref_m
        vref = vref_m
        href = 1005.0*tref_m
        suthc1 =  suthc1_m
        suthc2 =  suthc2_m
        visref = visref_m
        rgasref = rgasref_m

! English
      CASE(2)
        pref = 1.0
        tref = 1.0
        vref = 1.0
        href = 1.0
        suthc1 = 1.0
        suthc2 = 1.0
        visref = .982143
        rgasref = 1.0

    END SELECT

  END SUBROUTINE set_ref_quantities

END MODULE common_area
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
