MODULE keydata
  IMPLICIT NONE

  INTEGER :: RSTATE = 0

  INTEGER, PARAMETER :: nkeys = 19
  INTEGER, PARAMETER :: nscalkeys = 17
  INTEGER, PARAMETER :: nvectkeys = 1
  INTEGER, PARAMETER :: nthrekeys = 1
  INTEGER, PARAMETER :: ncuts = 6

  LOGICAL :: uistat(nscalkeys) ! Number of non-vectors
  LOGICAL :: uvstat(nvectkeys)  ! Number of vectors
  LOGICAL :: uvmstat(4)  ! Number of VizMath settings

  INTEGER :: keytyp(nkeys), ikeys(nkeys), kblk(3)

  REAL    :: flims(2,nkeys)

  CHARACTER*16 :: tkeys(nkeys), tcuts(ncuts)
  CHARACTER*80 :: ttl

#ifdef PV3SERIAL
  DATA ttl/'            V3 Post-Processor'/
#else
  DATA ttl/'            pV3 Client Post-Processor'/
#endif
  DATA tcuts / 'M Prime-Theta','R-Z','Constant R', &
               'Constant Z','Constant X','Constant Y'/
  DATA tkeys /    'Density         ', &
                  'Angles          ', &
                  'Coordinates     ', &
                  'Enthalpy        ', &
                  'Rothalpy        ', &
                  'Mach Number     ', &
                  'Pressure        ', &
                  'Rho V           ', &
                  'Entropy         ', &
                  'Temperature     ', &
                  'Velocity        ', &
                  'Vorticity       ', &
                  'Turbulence      ', &
                  'Energy          ', &
                  'Other Scalars   ', &
                  'Averaging       ', &
                  'Viz Math        ', &
                  'Vectors         ', &
                  'Thresholding    '/
  DATA keytyp/ nscalkeys*1, nvectkeys*2, nthrekeys*5 /
  DATA ikeys/100,97,99,104,105,109,112,114,115,116, &
             118,119,113,101,111,65,77,86,84/

  INTEGER :: key2_1 = 0, key2_2 = 0, key2_3 = 0, key2_fft = 0, key2_mode = 0
  INTEGER :: key2_v = ICHAR('W') ! Vector key
  INTEGER :: keycut = 1 ! Key for cut options
  INTEGER :: key2_t = 0 ! Thresholding key (default is not to calculate)
  INTEGER :: myfps = 999999999 ! Framerate
  INTEGER :: timedisp ! Time stepping mode

  INTEGER :: fwctr = 0 ! File-Write counter
  INTEGER :: reloadstatus = 0

  REAL :: dt=0., timesav=0.

  CHARACTER(LEN=80) :: keyfilename='', scriptfilename=''

  LOGICAL :: isprocess=.FALSE. ! Image script processing
  LOGICAL :: joprocess=.FALSE. ! Journal processing

CONTAINS

!#######################################################################
  SUBROUTINE KeyAllocStat(num)
!#######################################################################
    INTEGER num,i

    DO i=1,num-1,1
      uistat(i)=.FALSE.
    ENDDO
    DO i=num+1,nscalkeys,1
      uistat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocStat

!#######################################################################
  SUBROUTINE VKeyAllocStat(num)
!#######################################################################
    INTEGER num,i

    DO i=1,num-1,1
      uvstat(i)=.FALSE.
    ENDDO
    DO i=num+1,nvectkeys,1
      uvstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE VKeyAllocStat

!#######################################################################
  SUBROUTINE KeyAllocInit()
!#######################################################################
    INTEGER i

    DO i=1,nscalkeys
      uistat(i)=.FALSE.
    ENDDO
    DO i=1,nvectkeys
      uvstat(i)=.FALSE.
    ENDDO
    DO i=1,4
      uvmstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocInit

!#######################################################################
  SUBROUTINE KeyAllocStatVM(num)
!#######################################################################
    INTEGER num,i

    DO i = 1, num-1, 1
      uvmstat(i)=.FALSE.
    ENDDO
    DO i = num+1, 4, 1
      uvmstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocStatVM

END MODULE keydata
