!#######################################################################
SUBROUTINE FXpV3Shock(indx, exnum, minode, mindp)
!#######################################################################
  USE common_area, ONLY: sfstor, itime
  IMPLICIT NONE

  INTERFACE 
    SUBROUTINE pVScal(key,v)
      INTEGER :: key
      REAL,DIMENSION(*) :: v
    END SUBROUTINE pVScal
  END INTERFACE

  INTEGER,PARAMETER :: xlate(5)=(/0,10000,10001,10002,10003/)

  INTEGER, INTENT(IN) :: indx, exnum, minode
  REAL, INTENT(IN) :: mindp

  INTEGER :: i, j, k, nnode, npts
  INTEGER :: nclust, ntris
  INTEGER :: cells(1), tris(1), cell(1), cnode(1), cnodes(1), clust(1)
  REAL :: xyz(1), test(1), coors(1)
  REAL :: vols(1), pts(1)
!  REAL,ALLOCATABLE,DIMENSION(:) :: press
  REAL :: press(1)
  
  POINTER (ppress,press)
  POINTER (pxyz,xyz), (ptest,test), (pnpts,npts), (pcoors,coors)
  POINTER (pntris,ntris), (ptris,tris), (pcells,cells)
  POINTER (pnclust,nclust), (pvols,vols), (pclust,clust), (pcell,cell)
  POINTER (pcnode,cnode), (pcnodes,cnodes), (ppts,pts)

  INTEGER*BSIZE,EXTERNAL :: mallocpV

  CALL pV_GetStruc(301, pxyz, nnode)
  IF(sfstor(itime)%npts.EQ.-1)THEN
    WRITE(*,*) ' Computing shock surfaces'
    ppress = mallocpV(nnode*4)
!    ALLOCATE(press(nnode))
    IF(ppress.EQ.0)THEN
      WRITE(*,*) ' Malloc error for pressure'
      CALL pV_SendXR(indx, exnum, 0, 3, 1, xyz)
      RETURN
    ENDIF
    i = xlate(2)
    CALL pVScal(i,press)
!    CALL pVScal(i,press)

    npts = 0
    PRINT*,' CALLING FX_ShockFind'
    CALL FX_ShockFind(ptest)
    IF(ptest.NE.0)THEN
      PRINT*,' CALLING FX_ShockSurface'
      CALL FX_ShockSurface(test, npts, pcoors, ntris, ptris, pcells)
      PRINT*,' CALLING FX_ShockVolumes'
      CALL FX_ShockVolumes(test, nclust, pvols, pclust, pcell, pcnode, pcnodes)
      PRINT*,' DONE CALLING FX'
      IF(nclust.GT.0)THEN
        PRINT*,' Calling Mark_Tris'
        CALL Mark_Tris(nclust,clust,cell,cnode,cnodes,press,ntris,cells,minode,mindp)
        IF(pvols.NE.0) CALL FX_free(pvols)
        PRINT*,' FREEING'
        CALL FX_free(pcnode)
        CALL FX_free(pcnodes)
        CALL FX_free(pclust)
        CALL FX_free(pcell)
      ENDIF

      j = 0
      DO i = 1, ntris
        IF(cells(i).GT.0) j = j + 1
      ENDDO
      npts = 3*j
      IF(npts.NE.0)THEN
        ppts = mallocpV(3*npts*4)
        IF(ppts.EQ.0)THEN
          npts = 0
        ELSE
          j = 1
          DO i = 1, ntris
            IF(cells(i).LT.0)CYCLE
            k = tris(3*i+1)
            pts(j) = coors(3*k)
            j = j + 1
            pts(j) = coors(3*k+1)
            j = j + 1
            pts(j) = coors(3*k+2)
            j = j + 1
            k = tris(3*i+2)
            pts(j) = coors(3*k)
            j = j + 1
            pts(j) = coors(3*k+1)
            j = j + 1
            pts(j) = coors(3*k+2)
            j = j + 1
            k = tris(3*i+3)
            pts(j) = coors(3*k)
            j = j + 1
            pts(j) = coors(3*k+1)
            j = j + 1
            pts(j) = coors(3*k+2)
            j = j + 1
          ENDDO
        ENDIF
      ENDIF

      CALL FX_free(ptest)
      CALL FX_free(pcoors)
      CALL FX_free(ptris)
      CALL FX_free(pcells)
    ENDIF

    CALL FX_free(ppress)
!    DEALLOCATE(press)
    sfstor(itime)%npts = npts
    ALLOCATE(sfstor(itime)%pts(3,npts))
    DO i = 1, npts
      sfstor(itime)%pts(1,i) = pts(3*(i-1)+1)
      sfstor(itime)%pts(2,i) = pts(3*(i-1)+2)
      sfstor(itime)%pts(3,i) = pts(3*(i-1)+3)
    ENDDO

  ENDIF

  IF(sfstor(itime)%npts.NE.0)THEN
    CALL pV_SendXR(indx, exnum, 0, 3, sfstor(itime)%npts, sfstor(itime)%pts)
  ELSE
    CALL pV_SendXR(indx, exnum, 0, 3, 1, xyz)
  ENDIF

  RETURN

END SUBROUTINE FXpV3Shock

!!#######################################################################
!SUBROUTINE Mark_Tris(nclust, clust, cell, cnode, cnodes, press, ntris, cells, minode, mindp)
!!#######################################################################
!  INTEGER, INTENT(IN) :: nclust, minode
!  INTEGER,DIMENSION(:) :: clust
!  REAL, INTENT(IN) :: mindp
!
!  INTEGER :: i, clen, mx
!  REAL :: minp, maxp
!  CHARACTER :: cellTable(1)
!
!  POINTER (pcellTable,cellTable)
!
!  INTEGER*BSIZE :: mallocpV
!
!  clen = clust(nclust)
!  mx = 0
!  DO i = 1, clen
!    IF(cell(i).GT.mx) mx = cell(i)
!  ENDDO
!
!  pcelltable = mallocpV(mx*2)
!  IF(pcellTable.EQ.0)THEN
!    WRITE(*,*) ' Cannot Allocate cellTable (mark_tris)'
!    RETURN
!  ENDIF
!
!  DO i = 1, mx
!    cellTable(i) = CHAR(0)
!  ENDDO
!
!  RETURN
!
!END SUBROUTINE Mark_Tris
