#include "parameters.h"
!#######################################################################
  SUBROUTINE Get_PV3_Input()
!#######################################################################
    USE common_area
    USE solver_turbo, ONLY: bladerow, bladepass
    USE general, ONLY: int_to_string, Fatal_Error
    IMPLICIT NONE

    INTEGER :: i, j, k, nbpctr, ctr
    LOGICAL :: exists=.FALSE.

!---> Operations namelist
    INTEGER :: pv3_run_mode,pv3_file_form,pv3_num_flow_files, &
               pv3_num_clients,pv3_num_grid_files, pv3_dimen, &
               pv3_file_type, pv3_ssi(3), pv3_num_flow_max, &
               pv3_maxvars
    CHARACTER(LEN=NSTRLEN) :: pv3infile, &
                    pv3_grid_array(NMX),pv3_flow_array(NMX), &
                    pv3_func_array(NMX), pv3_bc_file_name, &
                    pv3_precalc_array(NMX)
    CHARACTER(LEN=20) :: pv3_case_name
    CHARACTER(LEN=80) :: pv3_format_string
    LOGICAL :: pv3_lowmem,pv3_double

    NAMELIST/OPERATION/ pv3_run_mode,   pv3_file_form,      pv3_lowmem,         &
&                       pv3_case_name,  pv3_grid_array,     pv3_num_clients,    &
&                       pv3_flow_array, pv3_num_flow_files, pv3_num_grid_files, &
&                       pv3_func_array, pv3_file_type, pv3_num_flow_max,        &
&                       novort,pv3_double,pv3_bc_file_name, pv3_dimen,          &
&                       pv3_format_string, pv3_ssi, pv3_precalc_array, pv3_maxvars

!---> Replication namelist
    INTEGER :: nrep
    LOGICAL :: replicate
    NAMELIST/REPLICATION/ replicate, nrep, offsetx, offsety, offsetz

!---> Reference condition namelist
    NAMELIST/REFERENCE/ dimen, pref_n, tref_n, vref_n, suthc1_n, rhoref_n, &
&                              suthc2_n, visref_n, rgasref_n, cpref_n,     &
&                              pref_m, tref_m, vref_m, suthc1_m, rhoref_m, &
&                              suthc2_m, visref_m, rgasref_m, cpref_m,     &
&                              diam, gammabase, gamtable, tempgam,         &
&                              pstate, tstate, refl_n, refl_m

!---> Turbo namelist
    REAL :: tur_omega
    INTEGER :: tur_nrows, tur_nbld
    INTEGER,DIMENSION(100) :: tur_nbps
    LOGICAL :: tur_keep_cell,tur_rotate

    NAMELIST/TURBO_INPUT/ tur_keep_cell, tur_rotate, tur_nbld, &
&                         tur_omega, tur_nrows, tur_nbps

!---> Plot3d namelist
    LOGICAL :: p3d_iblank
    CHARACTER(LEN=NSTRLEN) :: p3d_gridindexfile
    NAMELIST/PLOT3D_INPUT/ p3d_iblank, p3d_gridindexfile

!---> Grid Namelist
    REAL :: grid_offset_x, grid_offset_y, grid_offset_z
    REAL :: grid_offset_r, grid_offset_t
    REAL :: grid_transl_x, grid_transl_y, grid_transl_z
    REAL :: grid_transl_r, grid_transl_t
    LOGICAL :: grid_adjust_t, grid_in_degrees
    NAMELIST/GRIDMANIPULATION/ grid_offset_x, grid_offset_y, grid_offset_z, &
&                              grid_offset_r, grid_offset_t, grid_adjust_t, &
&                              grid_transl_x, grid_transl_y, grid_transl_z, &
&                              grid_transl_r, grid_transl_t, grid_in_degrees

    mirror = 0
!=======================================================================
! Set defaults
!=======================================================================
! Operations
    bcavail = .FALSE.
    novort = .FALSE.
    pv3infile = usrinput
    pv3_run_mode = -1
    pv3_file_form = 1 ! File format 1 = vstage
                      !             2 = plot3d
                      !             3 = "jixian" plot3d
    pv3_file_type = 1 ! File formatting 1 = unformatted, SP
    pv3_lowmem = .FALSE.
    pv3_case_name = 'test'
    pv3_grid_array(1) = ''
    pv3_flow_array(1) = 'xxx'
    pv3_func_array(1) = 'xxx'
    pv3_precalc_array(1) = 'xxx'
    pv3_bc_file_name = ''
    pv3_format_string = ''
    pv3_ssi = (/0,0,0/)
    pv3_num_flow_files = 0
    pv3_num_grid_files = 0
    pv3_num_flow_max = 99999
    pv3_maxvars = 0
    pv3_num_clients = 1
    pv3_double = .FALSE.
    pv3_dimen = 0
    DO i=2,NMX
      pv3_grid_array(i) = ''
      pv3_flow_array(i) = ''
      pv3_func_array(i) = ''
    ENDDO

! Replication
    replicate = trep
    nrep = -1
    offsetx = 0.
    offsety = 0.
    offsetz = 0.

! Turbo
    tur_keep_cell = .FALSE.
    tur_rotate = .FALSE.
    tur_omega = omega
    tur_nrows = 1
    tur_nbps = 0
    tur_nbld = -1

! Plot3D
    p3d_iblank = .FALSE.
    p3d_gridindexfile = 'none'

! AVUS

! Grid Manipulation
    grid_in_degrees = .FALSE.
    grid_transl_x = grd_translx
    grid_transl_y = grd_transly
    grid_transl_z = grd_translz
    grid_transl_r = grd_translr
    grid_transl_t = grd_translt
    grid_offset_x = grd_offsetx
    grid_offset_y = grd_offsety
    grid_offset_z = grd_offsetz
    grid_offset_r = grd_offsetr
    grid_offset_t = grd_offsett
    grid_adjust_t = grd_tadjust

!=======================================================================
! Read namelists
    INQUIRE(FILE=pv3infile,EXIST=exists)
    IF(exists)THEN
      OPEN(UNIT=INUNIT,FILE=pv3infile,FORM='formatted')

7000  CONTINUE
      READ(INUNIT,NML=OPERATION,ERR=7000,END=9000) ! The operations namelist is necessary
9000  REWIND(INUNIT)

7001  CONTINUE
      READ(INUNIT,NML=REPLICATION,ERR=7001,END=9001)
9001  REWIND(INUNIT)

7002  CONTINUE
      READ(INUNIT,NML=REFERENCE,ERR=7002,END=9002)
9002  REWIND(INUNIT)

7003  CONTINUE
      READ(INUNIT,NML=TURBO_INPUT,ERR=7003,END=9003)
9003  REWIND(INUNIT)

7004  CONTINUE
      READ(INUNIT,NML=PLOT3D_INPUT,ERR=7004,END=9004)
9004  REWIND(INUNIT)

7005  CONTINUE
      READ(INUNIT,NML=GRIDMANIPULATION,ERR=7005,END=9005)
9005  REWIND(INUNIT)

      CLOSE(INUNIT)
    ENDIF



!
!---> Overwrite default values
!
    runmode = pv3_run_mode
    ffmt = pv3_file_form
    fftype = pv3_file_type
    IF(runmode.EQ.6)fftype = pv3_file_form
    lowmem = pv3_lowmem
    csnm = pv3_case_name
    dimen = pv3_dimen

    flowarr = pv3_flow_array
    funcarr = pv3_func_array
    ntmax = pv3_num_flow_max
    nffile = pv3_num_flow_files
    IF(ANY(pv3_ssi.NE.0))THEN
      IF(nffile.GT.(pv3_ssi(2)-pv3_ssi(1))/pv3_ssi(3))THEN
        CALL Fatal_Error(.TRUE.,'Improper flow file input numbers')
      ELSE
        ctr =  0
        DO i = pv3_ssi(1), pv3_ssi(2), pv3_ssi(3)
          ctr = ctr + 1
          WRITE(flowarr(ctr),pv3_format_string) i
        ENDDO
      ENDIF
    ELSE
      IF(nffile.LE.0)THEN
        nffile = COUNT(flowarr.NE.'')
      ELSE
        nffile = MIN(COUNT(flowarr.NE.''),nffile)
      ENDIF
    ENDIF

    gridarr = pv3_grid_array
    ngrid = pv3_num_grid_files

    SELECT CASE(runmode)
      CASE(-1,2)
        ngrid = 1
      CASE DEFAULT
        IF(ngrid.LE.0)THEN
          ngrid = COUNT(gridarr.NE.'')
        ELSE
          ngrid = MIN(COUNT(gridarr.NE.''),ngrid)
        ENDIF
    END SELECT

    maxvars = pv3_maxvars
    IF (pv3_maxvars.NE.0) THEN
        ALLOCATE(vararr(maxvars))
    ENDIF

    nclients = pv3_num_clients
    griddouble = pv3_double
    avusbcfilename = pv3_bc_file_name

!-----------------------------------------------------------------------
! This call to get the finalized client ID used to be in pv3.f90.
! Due to the addition of specialized runmodes, it was necessary to move
! it to right before the processing for these events.
#ifndef PV3SERIAL
  CALL Get_ID
#endif
!-----------------------------------------------------------------------

! Replication
    trep = replicate
    ntrep = nrep

! Reference
    IF(cpref_n.EQ.0.)THEN
      cpref_n = gammabase/(gammabase-1.)
    ELSE
      rgasref_n = cpref_n*(gammabase-1.)/gammabase
    ENDIF
    href_n = cpref_n * tref_n
    visref_n = suthc1_n*tref_n**1.5/(tref_n+suthc2_n)
    rhoref_n = pref_n/vref_n**2

    IF(cpref_m.EQ.0.)THEN
      cpref_m = (gammabase*rgasref_m)/(gammabase-1.)
    ELSE
      rgasref_m = cpref_m*(gammabase-1.)/gammabase
    ENDIF
    href_m = (gammabase*rgasref_m)/(gammabase-1.)*tref_m
    visref_m = suthc1_m*tref_m**1.5/(tref_m+suthc2_m)
    rhoref_m = pref_m/vref_m**2
      

! Turbo
    kpcell = tur_keep_cell
    trot = tur_rotate
    omega = tur_omega
    nbpctr=0
    nbld = tur_nbld
    IF(runmode.EQ.1000)THEN
      DO i=1,tur_nrows
        DO j=1,tur_nbps(i)
          nbpctr=nbpctr+1
          IF(cid.EQ.nbpctr)THEN
            DO k=1,ngrid
              bladerow = i
              bladepass = j
              gridarr(k) = TRIM(ADJUSTL(csnm)) // '.r' // TRIM(ADJUSTL(int_to_string(i))) // 'b' // TRIM(ADJUSTL(int_to_string(j))) // pv3_grid_array(k)
              WRITE(*,*) 'New grid is ',gridarr(k)
            ENDDO
            DO k=1,ntime
              IF(flowarr(k).NE.'xxx')THEN
                flowarr(k) = TRIM(ADJUSTL(csnm)) // '.r' // TRIM(ADJUSTL(int_to_string(i))) // 'b' // TRIM(ADJUSTL(int_to_string(j))) // pv3_flow_array(k)
                WRITE(*,*) 'New flow is ',flowarr(k)
              ENDIF
            ENDDO
          ENDIF
        ENDDO
      ENDDO
      runmode=1
    ENDIF

! Plot3D
    liblank = p3d_iblank
    gridindexfile = p3d_gridindexfile

! Grid Manipulation
    grd_translx = grid_transl_x
    grd_transly = grid_transl_y
    grd_translz = grid_transl_z
    grd_translr = grid_transl_r
    IF(grid_in_degrees)THEN
      grd_translt = PI * grid_transl_t/180.
      grd_offsett = PI * grid_offset_t/180.
      PRINT*,' WILL CHANGE THETA BY:',grd_translt
    ELSE
      grd_translt = grid_transl_t
      grd_offsett = grid_offset_t
    ENDIF
    grd_offsetx = grid_offset_x
    grd_offsety = grid_offset_y
    grd_offsetz = grid_offset_z
    grd_offsetr = grid_offset_r
    grd_tadjust = grid_adjust_t

  END SUBROUTINE Get_PV3_Input

!#######################################################################
  SUBROUTINE Get_ID()
!#######################################################################
! Routine needs to return if already appropriately set
    USE common_area

    SELECT CASE(runmode)
!======================================================================
      CASE(2)
! Default batch AST mode for distributed disks???
!        DO i=1,nclients
!          WRITE(my_fnam,FORMAT('ast.b',i,'.x'))
!          CALL Check_File_Exist(my_fnam,ios)
!          IF(ios)GOTO 1000
!        ENDDO
!       CALL Fatal_Error(.TRUE.,'NO AST FILES FOUND')
!1000   CONTINUE
!      cid = i
!======================================================================
      CASE(-1)
! Default interactive mode
        WRITE(*,*) ' Enter client id==>'
        READ*,cid
!======================================================================
      CASE DEFAULT
! Default
        IF(cid.LE.0)THEN
          WRITE(*,*) ' Enter client id==>'
          READ*,cid
        ENDIF
    END SELECT

  END SUBROUTINE Get_ID
