#include "parameters.h"
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE journal_common
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  IMPLICIT NONE

!  INTEGER*BSIZE, SAVE :: ivec5(6), ivec6(6), ivec7(6) ! Scalar, vector, and threshold states
!  INTEGER*BSIZE, SAVE :: ivec12(6) ! Bubble state
!  INTEGER*BSIZE, SAVE :: ivec13(6) ! Probe state
!  INTEGER*BSIZE, SAVE :: ivecm4(6) ! Point probe/2D window pointer
!  INTEGER*BSIZE, SAVE :: ltool, lastw
!  INTEGER*BSIZE       :: iwini
!  INTEGER, SAVE       :: linenm
!  INTEGER             :: istati, len, iflag
!  REAL, SAVE :: rvecd(12), xform(12), cut(9), tool(7)
!  REAL, SAVE :: rvec5(9), rvec6(9), rvec7(9) ! Scalar, vector, and threshold values
!  REAL, SAVE :: rvec12(8), rvec13(8), rvecm4(8)
!  LOGICAL, SAVE :: scriptsent=.FALSE., rw=.FALSE., script=.FALSE.
!  LOGICAL, SAVE :: genpart=.FALSE.
!  LOGICAL flag
!  CHARACTER(LEN=80), SAVE :: scstr
!  CHARACTER(LEN=80) :: filenm

  TYPE statevar
#ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: ivec(8)=0
#else
    INTEGER :: ivec(8)=0
#endif
    REAL :: rvec(12)=0.
    CHARACTER(LEN=NSTRLEN) :: cstr=''
  END TYPE statevar

  TYPE(statevar) :: st(24)

#ifdef PV3SERIAL
  INTEGER(KIND=BSIZE) :: itool = 0
  INTEGER(KIND=BSIZE) :: ltool = 0
#else
  INTEGER :: itool = 0
  INTEGER :: ltool = 0
#endif
  INTEGER :: lastw = 0
  INTEGER :: linenm = 0

CONTAINS

!#######################################################################
  SUBROUTINE JournalGetAndRecord(iopt, unt)
!#######################################################################
! This routine first calls a GetState to query data and then records it
! to a specified unit.  This is currently used to record the current
! state into a script file or a journal.

    INTEGER, INTENT(IN) :: iopt, unt

    CALL pV_GetState(iopt, st(iopt)%ivec, st(iopt)%rvec, st(iopt)%cstr)
    CALL JournalDataFormat(iopt, st(iopt)%ivec, st(iopt)%rvec, &
                               st(iopt)%cstr, unt)

    RETURN

  END SUBROUTINE JournalGetAndRecord

!#######################################################################
  SUBROUTINE JournalCheckAndRecord(iopt, ni, nr, unt)
!#######################################################################
    INTEGER, INTENT(IN) :: iopt, unt
    INTEGER :: ni, nr

#   ifdef PV3SERIAL
      INTEGER(KIND=BSIZE) :: ivec(8)=0
#   else
      INTEGER :: ivec(8)=0
#   endif
    REAL                   :: rvec(12)=0.
    CHARACTER(LEN=NSTRLEN) :: str=''
    LOGICAL :: flag

    CALL pV_GetState(iopt, ivec, rvec, str)
    SELECT CASE(iopt)
      CASE( 2)
        IF((itool.EQ.2 .AND. ltool.EQ.2) .OR. itool.NE.2)THEN
          CALL JournalDiffUpdate(iopt, ivec, ni, rvec, nr, flag)
          IF(flag) CALL JournalDataFormat(iopt, ivec, rvec, str, unt)
        ENDIF

      CASE( 3) ! Scalar tool
        itool = ivec(1)
        IF(itool.EQ.ltool)THEN
          IF(itool.GE.4 .AND. itool.LE.7)THEN
            IF(itool.EQ.7) nr = 1
            CALL JournalDiffUpdate(iopt, ivec, ni, rvec, nr, flag)
            IF(flag) CALL JournalDataFormat(iopt, ivec, rvec, str, unt)
          ENDIF
        ENDIF
        RETURN

      CASE( 8) ! User response
        IF((ivec(1).NE.0 .OR. str.NE.' ') .AND. lastw.NE.-2)THEN
          CALL JournalResponseFormat(-2, ivec, str, -1, unt)
          lastw=-2
        ENDIF
        RETURN

    END SELECT
    CALL JournalDiffUpdate(iopt, ivec, ni, rvec, nr, flag)
    IF(flag) CALL JournalDataFormat(iopt, ivec, rvec, str, unt)

    RETURN

  END SUBROUTINE JournalCheckAndRecord

!#######################################################################
  SUBROUTINE JournalDiffUpdate(iopt, ivec, ni, rvec, nr, flag)
!#######################################################################
#   ifdef PV3SERIAL
      INTEGER(KIND=BSIZE), INTENT(IN) :: ivec(8)
#   else
      INTEGER, INTENT(IN) :: ivec(8)
#   endif
    INTEGER, INTENT(IN) :: iopt, ni, nr
    REAL, INTENT(IN) :: rvec(12)
    LOGICAL, INTENT(OUT) :: flag

    INTEGER :: i

    flag = .FALSE.

    DO i = 1, ni
      IF(ivec(i).NE.st(iopt)%ivec(i)) flag = .TRUE.
      st(iopt)%ivec(i) = ivec(i)
    ENDDO
    DO i = 1, nr
      IF(rvec(i).NE.st(iopt)%rvec(i)) flag = .TRUE.
      st(iopt)%rvec(i) = rvec(i)
    ENDDO

    RETURN

  END SUBROUTINE JournalDiffUpdate

!#######################################################################
  SUBROUTINE JournalDataFormat(iopt, ivecd, rvecd, scstr, unt)
!#######################################################################
! This routine formats the incoming data and places it in the assigned,
! pre-opened file.
    USE keydata, ONLY: key2_1, key2_2, key2_3, key2_fft, key2_mode, &
                       key2_v, key2_t, keycut

    INTEGER, INTENT(IN) :: iopt, unt
#   ifdef PV3SERIAL
      INTEGER(KIND=BSIZE), INTENT(IN) :: ivecd(8)
#   else
      INTEGER, INTENT(IN) :: ivecd(8)
#   endif
    REAL, INTENT(IN) :: rvecd(12)
    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: scstr

    INTEGER :: i

    WRITE(unt,6000) 0,iopt,0,0,0
    WRITE(unt,6000) ivecd
    WRITE(unt,6001) (rvecd(i),i=1,6)
    WRITE(unt,6001) (rvecd(i),i=7,12)
    WRITE(unt,6002) scstr
    WRITE(unt,6000) key2_1, key2_2, key2_3, key2_fft, key2_mode, &
                    key2_v, key2_t, keycut

    RETURN
  
 6000 FORMAT(1x,8i13)
 6001 FORMAT(1x,6e13.6)
 6002 FORMAT(a80)
  
  END SUBROUTINE JournalDataFormat

!#######################################################################
  SUBROUTINE JournalResponseFormat(val1, ivec, str, val2, unt)
!#######################################################################
! Format for the response from the user
#   ifdef PV3SERIAL
      INTEGER(KIND=BSIZE), INTENT(IN) :: ivec(8)
#   else
      INTEGER, INTENT(IN) :: ivec(8)
#   endif
    INTEGER, INTENT(IN) :: val1, val2, unt
    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: str

    WRITE(unt,6000) val1
    WRITE(unt,6000) ivec
    WRITE(unt,6002) str
    WRITE(unt,6000) val2

 6000 FORMAT(1x,8i13)
 6001 FORMAT(1x,6e13.6)
 6002 FORMAT(a80)
    
  END SUBROUTINE JournalResponseFormat

!#######################################################################
  SUBROUTINE OpenJournalFile()
!#######################################################################
! Opens the journal file and inserts the initial state (which may not be
! necessary)
    IMPLICIT NONE

    OPEN(UNIT=JOURNALOUNIT, FILE='pV3.journal')
  
    CALL JournalGetAndRecord( 2, JOURNALOUNIT )
    CALL JournalGetAndRecord( 3, JOURNALOUNIT )
    CALL JournalGetAndRecord( 5, JOURNALOUNIT )
    CALL JournalGetAndRecord( 6, JOURNALOUNIT )
    CALL JournalGetAndRecord( 7, JOURNALOUNIT )
    CALL JournalGetAndRecord( 8, JOURNALOUNIT )
    CALL JournalGetAndRecord(20, JOURNALOUNIT )
    CALL JournalGetAndRecord(21, JOURNALOUNIT )

  END SUBROUTINE OpenJournalFile

!#######################################################################
  SUBROUTINE FinalizeJournal(iopt, fnam)
!#######################################################################
! After a journal file has been copied over to a new file, this routine
! gets called to add a few final setstate commands

    INTEGER, INTENT(IN) :: iopt
    CHARACTER(LEN=80), INTENT(IN) :: fnam

    INTEGER :: unt

    IF(iopt.EQ.1)THEN
      unt = JOURNALSUNIT
      OPEN(UNIT=unt, FILE=fnam, ACCESS='append')
    ELSE
      unt = JOURNALOUNIT
    ENDIF

    CALL JournalGetAndRecord( 1, unt )
    CALL JournalGetAndRecord(14, unt )
    CALL JournalGetAndRecord(15, unt )
    CALL JournalGetAndRecord(16, unt )
    CALL JournalGetAndRecord(19, unt )

    CLOSE(unt)
    WRITE(*,*) ' Journal file created successfully.'

  END SUBROUTINE FinalizeJournal

!#######################################################################
  SUBROUTINE CopyJournal(fnam)
!#######################################################################

    CHARACTER(LEN=80), INTENT(IN) :: fnam
    CHARACTER(LEN=120) :: line

    REWIND(JOURNALOUNIT)
    OPEN(UNIT=JOURNALSUNIT, FILE=fnam)

    DO WHILE(1)
      READ(JOURNALOUNIT,"(A120)",ERR=9000,END=9999) line
      WRITE(JOURNALSUNIT,"(A120)") line
    ENDDO

 9000 CONTINUE
      WRITE(*,*) ' Caught error parsing journal.'
 9999 CONTINUE
      CLOSE(JOURNALSUNIT)

  END SUBROUTINE CopyJournal

!#######################################################################
  SUBROUTINE RecordJournalEntry(iwin, itype, ie, je, istate, winid)
!#######################################################################

    INTEGER*BSIZE, INTENT(IN) :: iwin
    INTEGER*BSIZE, INTENT(IN) :: winid(9)
    INTEGER, INTENT(IN) :: itype, ie, je, istate

    INTEGER :: iwini, i

    CALL JournalCheckAndRecord( 8, 0, 0, JOURNALOUNIT )
    CALL JournalCheckAndRecord( 2, 0, 9, JOURNALOUNIT )
    CALL JournalCheckAndRecord( 3, 2, 5, JOURNALOUNIT )
    CALL JournalCheckAndRecord( 5, 1, 2, JOURNALOUNIT )
    CALL JournalCheckAndRecord( 6, 1, 1, JOURNALOUNIT )
    CALL JournalCheckAndRecord( 7, 3, 2, JOURNALOUNIT )
    CALL JournalCheckAndRecord(20, 1, 2, JOURNALOUNIT )
    CALL JournalCheckAndRecord(21, 1, 1, JOURNALOUNIT )

    IF(iwin.EQ.-1) ltool = itool

    iwini = -1
    DO i = 1, 7
      IF(iwin.EQ.winid(i))iwini = i
    ENDDO
    IF(iwini.EQ.-1)RETURN ! Suppress all redraws
    WRITE(JOURNALOUNIT,6000) iwini, itype, ie, je, istate
    lastw = iwini

    RETURN

 6000 FORMAT(1x,8i13)

  END SUBROUTINE RecordJournalEntry

!#######################################################################
  SUBROUTINE ExecuteJournalEntry(iwin, itype, ie, je, istate, winid, keybin)
!#######################################################################
    USE keydata, ONLY: joprocess, key2_1, key2_2, key2_3, key2_fft, &
                       key2_mode, key2_v, key2_t, keycut
    INTEGER*BSIZE, INTENT(INOUT) :: iwin
    INTEGER*BSIZE, INTENT(IN) :: winid(9)
    INTEGER, INTENT(INOUT) :: itype, ie, je, istate
#   ifdef PV3SERIAL
      INTEGER(KIND=BSIZE), INTENT(IN) :: keybin(32)
      INTEGER(KIND=BSIZE) :: ivec(8)=0
#   else
      INTEGER, INTENT(IN) :: keybin(32)
      INTEGER :: ivec(8)=0
#   endif
    INTEGER :: iflag=0, i=0
    REAL :: rvecd(12)=0
    CHARACTER(LEN=80) :: scstr=''

 7003 linenm = linenm + 1
      READ(JOURNALIUNIT,6000,END=7002,ERR=7099) iwin, itype, ie, je, istate
      IF(iwin.GT.0)THEN
        IF(iwin.LE.7)THEN
          iwin = winid(iwin)
        ELSE
          iwin = 0 ! Continue in tight loop, do not redraw yet
        ENDIF

        IF(itype.EQ.2 .OR. itype.EQ.3)THEN
          IF(istate.LT.0 .AND. istate.GT.-32) istate = keybin(-istate)
        ENDIF

        linenm = linenm + 1
        READ(JOURNALIUNIT,6000,END=7002,ERR=7099) iflag
        IF(iflag.EQ.-2)THEN
          linenm = linenm + 1
          READ(JOURNALIUNIT,6000,END=7002,ERR=7099) ivec
          linenm = linenm + 1
          READ(JOURNALIUNIT,6002,END=7002,ERR=7099) scstr

          CALL pV_SetState(8, ivec, rvecd, ADJUSTL(scstr))

        ELSE
          linenm = linenm - 1
          BACKSPACE JOURNALIUNIT
        ENDIF

      ELSEIF(iwin.EQ.0)THEN
        linenm = linenm + 1
        READ(JOURNALIUNIT,6000,END=7002,ERR=7099) ivec
        linenm = linenm + 1
        READ(JOURNALIUNIT,6001,END=7002,ERR=7099) (rvecd(i),i=1,6)
        linenm = linenm + 1
        READ(JOURNALIUNIT,6001,END=7002,ERR=7099) (rvecd(i),i=7,12)
        linenm = linenm + 1
        READ(JOURNALIUNIT,6002,END=7002,ERR=7099) scstr
        linenm = linenm + 1
        READ(JOURNALIUNIT,6000,END=7002,ERR=7099) key2_1, key2_2, key2_3, &
           key2_fft, key2_mode, key2_v, key2_t, keycut

        CALL pV_SetState(itype, ivec, rvecd, scstr)
        PRINT*,' ITYPE: ',itype, ivec, rvecd, scstr

      ENDIF

      RETURN

 7099 WRITE(*,*) ' ERROR in journal file at line #',linenm
 7002 REWIND JOURNALIUNIT
      linenm = 0
      !GOTO 7003
      WRITE(*,*) ' '
      joprocess = .FALSE.
      linenm = 0
      CLOSE(JOURNALIUNIT)
      WRITE(*,*) ' Journal Processing off!'
      WRITE(*,*) ' '
      iwin = -1
      RETURN

 6000 FORMAT(1x,8i13)
 6001 FORMAT(1x,6e13.6)
 6002 FORMAT(a80)

  END SUBROUTINE ExecuteJournalEntry

END MODULE journal_common
