#include "parameters.h"
#define FLOW pfw(i)%
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE common_area
  USE mod_clear, ONLY: ClearVar
  USE typedefs
  IMPLICIT NONE

!-----------------------------------------------------------------------
! General required parameters for normal unsteady operation
  INTEGER :: coordsys = CYLINDRICAL
  INTEGER :: istat, maxblk, iopt, mirror ! Values for the pV_Init call
  INTEGER :: ngrid,igrid,ntime,itime,ntmax,nffile ! Timing counters for grid/flow
  INTEGER :: nfunc
  INTEGER :: runmode, & ! Solver operation mode
             ffmt,    & ! File format
             fftype,  & ! File formatting
             nclients,& ! Total number of clients
             maxvars    ! Total number of allocated variables
  INTEGER,ALLOCATABLE,DIMENSION(:) :: vararr
  INTEGER :: gksurf ! Global storage of knsurf so I have it
  INTEGER :: gsurf ! Global storage of ksurf so I have it
  INTEGER :: cid ! Client id number for enrollment in PVM
                       ! and determination of files to read
  INTEGER :: dimen ! Dimensionalization flag
                   ! 0) Non-dimensional
                   ! 1) English
                   ! 2) Metric

  CHARACTER(LEN=NSTRLEN) :: usrinput ! The name of a user specified input file
  CHARACTER(LEN=NSTRLEN) :: gridarr(NMX) ! Storage of grid file names
  CHARACTER(LEN=NSTRLEN) :: distarr(NMX) ! Storage of grid file names
  CHARACTER(LEN=NSTRLEN) :: flowarr(NMX) ! Storage of flow file names
  CHARACTER(LEN=NSTRLEN) :: funcarr(NMX) ! Storage of func file names
  CHARACTER(LEN=NSTRLEN) :: avusbcfilename ! General bc filename
  CHARACTER(LEN=NSTRLEN) :: pcalcarr(NMX) ! Storage of initial flow variables
  CHARACTER(LEN=20) :: cname='',dname='' ! Client and discipline names

  LOGICAL :: lowmem ! For low-memory usage ??? Not yet implemented any-
                    ! where in the code ???
  LOGICAL :: novort ! For when the vorticity calcs crash the client
  LOGICAL :: griddouble ! Single/Double precision grids
  LOGICAL :: bcavail ! Is boundary data available?

  REAL :: rep(16) ! The replication matrix

  INTEGER,ALLOCATABLE,DIMENSION(:,:) :: equivlist
  INTEGER :: gequiv

!-----------------------------------------------------------------------
! Replication specific quantities
  INTEGER :: ntrep  ! Number of times to replicate objects
  REAL :: offsetx = 0., offsety = 0., offsetz = 0.
  LOGICAL :: trep=.TRUE.   ! Replication flag


!-----------------------------------------------------------------------
! Reference quantities
  REAL :: tempgam(2)=(/1.,100./)
  REAL :: gamtable(2)=(/1.4,1.4/) ! Gamma interpolation values from the
  REAL :: gammabase = 1.4         ! TURBO input00 file
  REAL :: TKEinE = 0.0            ! Generally just set to 0.0

  REAL :: omega = 0.0  ! This is used for the entire client!!!
                 ! not just the part that is rotating.  No mix'n'match!
  REAL :: dtheta = 0.0 ! Replication delta

  REAL :: refl = 1.
  REAL :: refl_n = 1.
  REAL :: refl_m = 1.

  REAL :: cpref = 1.
  REAL :: cpref_n = 0.    ! Set as zero so that it gets overwritten
  REAL :: cpref_m = 6000.

  REAL :: rgasref = 1.
  REAL :: rgasref_n = 1.
  REAL :: rgasref_m = 287.

  REAL :: suthc1 = 1.
  REAL :: suthc1_n = 1.
  REAL :: suthc1_m = 1.45825E-6 ! Input from TURBO's input00 that is

  REAL :: suthc2 = 1.
  REAL :: suthc2_n = 1.
  REAL :: suthc2_m = 110.333    ! used for normalization and reference

  REAL :: pref = 1.
  REAL :: pref_n = 1.
  REAL :: pref_m = 101325.

  REAL :: vref = 1.
  REAL :: vref_n = 1.
  REAL :: vref_m = 287.5957

  REAL :: tref = 1.
  REAL :: tref_n = 1.
  REAL :: tref_m = 288.15

  REAL :: rhoref = 1.
  REAL :: rhoref_n = 1.
  REAL :: rhoref_m = 101325./287.5957**2

  REAL :: href = 1.
  REAL :: href_n = 1.
  REAL :: href_m = 1.4*287./.4*288.15

  REAL :: visref = 1.
  REAL :: visref_n = 1.
  REAL :: visref_m = 1.45825E-6*288.15**1.5/(288.15+110.333)

  REAL :: diam = 1.
  REAL :: pstate = 1.
  REAL :: tstate = 1.

  INTEGER :: nbld = 0

  CHARACTER*20 :: csnm  ! Case name for batch read of vstage files

  LOGICAL :: kpcell = .FALSE. ! Determines whether or not cell-centered data is
                              ! kept.  This is useful for display in V3.
  LOGICAL :: trot = .FALSE.  ! Turbomachinery rotation flag

!-----------------------------------------------------------------------
! Plot3D specific quantities
  CHARACTER(LEN=NSTRLEN) :: gridindexfile

  LOGICAL :: liblank ! Flag for the existance of iblanking in grid files

!-----------------------------------------------------------------------
! AVUS variables
  INTEGER, POINTER, DIMENSION(:) :: npnts
  INTEGER :: nzones = 0

!-----------------------------------------------------------------------
! Grid Manipulation
  REAL :: grd_offsetx = 0.
  REAL :: grd_offsety = 0.
  REAL :: grd_offsetz = 0.
  REAL :: grd_offsetr = 0.
  REAL :: grd_offsett = 0.
  REAL :: grd_translx = 0.
  REAL :: grd_transly = 0.
  REAL :: grd_translz = 0.
  REAL :: grd_translr = 0.
  REAL :: grd_translt = 0.
  LOGICAL :: grd_tadjust = .FALSE.  ! Adjust to current time theta location

!-----------------------------------------------------------------------
! FX Variables
  INTEGER :: FXnbc = 0
  INTEGER,POINTER,DIMENSION(:) :: FXshowSurf, FXattind, FXsepind

!-----------------------------------------------------------------------
! The principle storage arrays
  TYPE(grid),POINTER,DIMENSION(:) :: pgr ! The grid
  TYPE(flow),POINTER,DIMENSION(:) :: pfw ! The flow
  TYPE(aop),POINTER,DIMENSION(:)  :: cur ! Array of pointers which are
            ! used to hit the current variable.  The primary purpose is
            ! to reduce the storage needed to compute DFT's on the data.
  TYPE(vortexcore),POINTER,DIMENSION(:) :: vcstor ! Vortex Cores
  TYPE(shocksurf),POINTER,DIMENSION(:)  :: sfstor ! Shock Surfaces
  TYPE(sepattline),POINTER,DIMENSION(:) :: sastor ! Sep & Att Lines
  TYPE(residencetime),POINTER,DIMENSION(:) :: rtstor
  TYPE(boundarylayer),POINTER,DIMENSION(:) :: blstor

CONTAINS

!#######################################################################
  SUBROUTINE Clear_Mem_Flow()
!#######################################################################
! This routine can be used to clear out all of the stored variables
! that float around during allocations and such.  If a major key change
! occurs, this should happen.  Basically, if you are looking at
! one form of vorticity, we want to store it as long as possible.  The
! same is true with time averaging, and other large calculations.

    INTEGER i, j

!=======================================================================
! Clear out the variables from the grid
! These are the ones that are not always used, like i,j,k.  The xyz and
! rtz coordinates will be kept.
    DO i=1,ngrid

      CALL ClearVar(pgr(i)%i)
      CALL ClearVar(pgr(i)%j)
      CALL ClearVar(pgr(i)%k)

    ENDDO

! Clear out all non-primitive flow variables that have been stored.
! Primitives for vstage and for plot3d are not cleared.
    DO i=1,ntime
! Real Gas
      ! Not clearing g intentionally
      ! Not clearing rgas intentionally
      CALL ClearVar(FLOW reyn)

! Pressures
      CALL ClearVar(FLOW ptrel)
      CALL ClearVar(FLOW ptabs)

! Angles
      CALL ClearVar(FLOW alphaz)
      CALL ClearVar(FLOW alpham)
      CALL ClearVar(FLOW betaz)
      CALL ClearVar(FLOW betam)
      CALL ClearVar(FLOW phi)

! Mach Numbers
      CALL ClearVar(FLOW machthabs)
      CALL ClearVar(FLOW machthrel)
      CALL ClearVar(FLOW machr)
      CALL ClearVar(FLOW machtabs)
      CALL ClearVar(FLOW machtrel)
      CALL ClearVar(FLOW machxabs)
      CALL ClearVar(FLOW machxrel)
      CALL ClearVar(FLOW machyabs)
      CALL ClearVar(FLOW machyrel)
      CALL ClearVar(FLOW machz)
      CALL ClearVar(FLOW machideal)

! Rho-V
      CALL ClearVar(FLOW rhovthabs)
      CALL ClearVar(FLOW rhovthrel)
      CALL ClearVar(FLOW rhovr)
      CALL ClearVar(FLOW rhovtabs)
      CALL ClearVar(FLOW rhovtrel)
      CALL ClearVar(FLOW rhovxabs)
      CALL ClearVar(FLOW rhovxrel)
      CALL ClearVar(FLOW rhovyabs)
      CALL ClearVar(FLOW rhovyrel)
      CALL ClearVar(FLOW rhovz)

! Enthalpy
      CALL ClearVar(FLOW h)
      CALL ClearVar(FLOW htabs)
      CALL ClearVar(FLOW htrel)

! Rothalpy
      CALL ClearVar(FLOW roth)
      CALL ClearVar(FLOW rothcp)

! Entropy
      CALL ClearVar(FLOW sabs)
      CALL ClearVar(FLOW srat)

! Temperature
      CALL ClearVar(FLOW temp)
      CALL ClearVar(FLOW temptabs)
      CALL ClearVar(FLOW temptrel)

! Velocity
      ! Not clearing vx intentionally
      CALL ClearVar(FLOW vxabs, FLOW vx)
      ! Not clearing vy intentionally
      CALL ClearVar(FLOW vyabs, FLOW vy)
      CALL ClearVar(FLOW vtabs, FLOW vtrel)
      CALL ClearVar(FLOW vtrel)
      CALL ClearVar(FLOW vthrel)

! Vorticity
      ! Not clearing vortr intentionally
      ! Not clearing vortt intentionally
      ! Not clearing vortz intentionally
      CALL ClearVar(FLOW vortsabs)
      CALL ClearVar(FLOW vortsrel)
      CALL ClearVar(FLOW vorttabs)
      CALL ClearVar(FLOW vorttrel)
      CALL ClearVar(FLOW vortx)
      CALL ClearVar(FLOW vorty)
      CALL ClearVar(FLOW vortzabs)

! Turbulence
      CALL ClearVar(FLOW turbint)
      CALL ClearVar(FLOW turbrat)
      CALL ClearVar(FLOW turbtvisc)
      CALL ClearVar(FLOW qcrit)
      CALL ClearVar(FLOW yplus)

! Time Average
      CALL ClearVar(FLOW tmdif)


    ENDDO

! Time-Average
    i = 1
    CALL ClearVar(FLOW tmavg)
    CALL ClearVar(FLOW tmmin)
    CALL ClearVar(FLOW tmmax)
    CALL ClearVar(FLOW tmrms)
    CALL ClearVar(FLOW tmfft)
    DO i = 1, ntime
      NULLIFY(FLOW tmavg)
      NULLIFY(FLOW tmmin)
      NULLIFY(FLOW tmmax)
      NULLIFY(FLOW tmrms)
      NULLIFY(FLOW tmfft)
    ENDDO

! VizMath
    CALL Clear_Mem_VM()

  END SUBROUTINE Clear_Mem_Flow

!#######################################################################
  SUBROUTINE Clear_Mem_VM()
!#######################################################################
    INTEGER i

    DO i = 1, ntime

      CALL ClearVar(FLOW vmath)
      CALL ClearVar(FLOW vmaths)
      CALL ClearVar(FLOW vmathr)
      CALL ClearVar(FLOW vmatht)
      CALL ClearVar(FLOW vmathsabs)
      CALL ClearVar(FLOW vmathsrel)
      CALL ClearVar(FLOW vmathtabs)
      CALL ClearVar(FLOW vmathtrel)
      CALL ClearVar(FLOW vmathzabs)

    ENDDO

  END SUBROUTINE Clear_Mem_VM

!#######################################################################
  SUBROUTINE Clear_Mem_FX()
!#######################################################################
    INTEGER i

    DO i = 1, ntime

      CALL ClearVar(vcstor(i)%segs)
      CALL ClearVar(vcstor(i)%vcoor)
      CALL ClearVar(vcstor(i)%stren)
      vcstor(i)%nseg = -1

      CALL ClearVar(sastor(i)%sepxyz)
      CALL ClearVar(sastor(i)%attxyz)
      CALL ClearVar(sastor(i)%sepind)
      CALL ClearVar(sastor(i)%attind)
      sastor(i)%calc = 0

      CALL ClearVar(sfstor(i)%pts)
      sfstor(i)%npts = -1

    ENDDO

  END SUBROUTINE Clear_Mem_FX

!#######################################################################
  SUBROUTINE Clear_Base_Mem_Flow()
!#######################################################################
! This routine clears the base variables for vstage, real gas, and
! the cartesian velocities
    INTEGER :: i

    DO i = 1, ntime

! Vstage
      CALL ClearVar(FLOW p)
      CALL ClearVar(FLOW r)
      CALL ClearVar(FLOW vz)
      CALL ClearVar(FLOW vr)
      CALL ClearVar(FLOW vt)
      CALL ClearVar(FLOW e)
      CALL ClearVar(FLOW mut)
      CALL ClearVar(FLOW k)
      CALL ClearVar(FLOW eps)

! Real gas
      CALL ClearVar(FLOW g)
      CALL ClearVar(FLOW rgas)

! Cartesian
      CALL ClearVar(FLOW vx)
      CALL ClearVar(FLOW vy)

! Plot3D Functions
      CALL ClearVar(FLOW funk)

    ENDDO

  END SUBROUTINE Clear_Base_Mem_Flow

!#######################################################################
  SUBROUTINE ClearAllFlow()
!#######################################################################
    INTEGER i

    CALL Clear_Mem_Flow
    CALL Clear_Mem_VM
    CALL Clear_Mem_FX
    CALL Clear_Base_Mem_Flow

    DO i = 1, ntime
      CALL ClearVar(FLOW vortr)
      CALL ClearVar(FLOW vortt)
      CALL ClearVar(FLOW vortz)
    ENDDO

    DEALLOCATE(pfw)
    NULLIFY(pfw)

  END SUBROUTINE ClearAllFlow

!#######################################################################
  SUBROUTINE set_ref_quantities(state)
!#######################################################################
    IMPLICIT NONE

    INTEGER, INTENT(IN) :: state

    SELECT CASE(state)
! Non-dimensional
      CASE(0)
        cpref = cpref_n
        pref = pref_n
        tref = tref_n
        vref = vref_n
        href = href_n
        suthc1 = suthc1_n
        suthc2 = suthc2_n
        visref = visref_n
        rhoref = rhoref_n
        rgasref = rgasref_n
        refl = refl_n

! Metric
      CASE(1)
        cpref = cpref_m
        pref = pref_m
        tref = tref_m
        vref = vref_m
        href = href_m
        suthc1 = suthc1_m
        suthc2 = suthc2_m
        rhoref = rhoref_m
        visref = visref_m
        rgasref = rgasref_m
        refl = refl_m

! English
      CASE(2)
        pref = 1.
        tref = 1.
        vref = 1.
        href = 1.
        suthc1 = 1.
        suthc2 = 1.
        visref = .5
        rhoref = 1.
        rgasref = 1.
        refl = 1.

    END SELECT

  END SUBROUTINE set_ref_quantities

!#######################################################################
  SUBROUTINE PreCalcVars()
!#######################################################################
    IMPLICIT NONE

    INTEGER :: nvars = 0
    INTEGER :: i


    nvars = COUNT(pcalcarr.NE.'')
    IF (nvars.EQ.0) RETURN

    DO i=1,nvars
!        SELECT CASE (pcalcarr(i))
!          CASE ()
!          CASE ()
!          CASE ()
!          CASE ()
!        END SELECT
    ENDDO

  END SUBROUTINE PreCalcVars

END MODULE common_area
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
