#include "parameters.h"
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE mod_events
# ifdef PV3SERIAL
      USE server_common, ONLY: winid, extracton, cltitle
      USE common_area, ONLY: nfunc
# else
      USE server_common, ONLY: winid, nfunc, extracton, cltitle
# endif
  IMPLICIT NONE

#ifdef PV3SERIAL
  INTEGER(KIND=BSIZE) :: keybin(32)
  INTEGER(KIND=BSIZE) :: ivec(9)  ! ivec for normal calls
  INTEGER(KIND=BSIZE) :: ivect(8) ! ivec for tools
#else
  INTEGER :: keybin(32)
  INTEGER :: ivec(9)  ! ivec for normal calls
  INTEGER :: ivect(8) ! ivec for tools
#endif
  INTEGER :: lineout=0, aiwin=0, iopt=0
  INTEGER :: nintervals

  INTEGER :: window

  REAL :: rvec(13)
  REAL :: rvect(12)
  REAL :: uzmin, uzmax, ds

  LOGICAL :: first   = .TRUE.  ! First time through loop
  LOGICAL :: autoimg = .FALSE. ! Automatic image dumping
  LOGICAL :: imsent  = .FALSE. ! Image command sent
  LOGICAL :: ispbsent= .FALSE. ! Batch Image processing script
  LOGICAL :: auto    = .FALSE. ! Automatic plane scanning
  LOGICAL :: josent  = .FALSE. ! Journal command sent
  LOGICAL :: autotec = .FALSE. ! Automatic Tecplot/POV dumping
  LOGICAL :: tesent  = .FALSE. ! Tecplot command sent

  LOGICAL :: meta(8)

  REAL :: urate
  INTEGER :: udirec, usync

!--------------------------------------------------------
!  window type

CONTAINS

!########################################################
  INTEGER FUNCTION event_get_window()
!########################################################
    event_get_window = window
  END FUNCTION event_get_window


!########################################################
  LOGICAL FUNCTION event_set_window(iwin)
!########################################################
    INTEGER(KIND=BSIZE), INTENT(in) :: iwin
    INTEGER :: i

    event_set_window = .TRUE.
    DO i=1,9
        IF (iwin.EQ.winid(i)) window = i
    END DO
  END FUNCTION event_set_window

!--------------------------------------------------------
!  meta keys (left/right alt/ctrl/shift/meta)

!########################################################
  SUBROUTINE event_set_meta_false
!########################################################
    INTEGER :: i

    DO i=1,8
        meta(i) = .FALSE.
    ENDDO
  END SUBROUTINE event_set_meta_false

!########################################################
  LOGICAL FUNCTION event_set_meta(istate,iwin,itype)
!########################################################
    INTEGER, INTENT(in) :: istate, iwin, itype
    INTEGER :: etype

    event_set_meta = .FALSE.

    IF (itype.EQ.4 .OR. itype.EQ.5) THEN
        event_set_meta = .TRUE.
        RETURN
    END IF

    SELECT CASE(istate)
      CASE(65513,65511)
        IF (itype.EQ.2) THEN
            meta(1) = .TRUE.
        ELSE
            meta(1) = .FALSE.
        ENDIF
      CASE(65514,65512)
        IF (itype.EQ.2) THEN
            meta(2) = .TRUE.
        ELSE
            meta(2) = .FALSE.
        ENDIF
      CASE(65507)
        IF (itype.EQ.2) THEN
            meta(3) = .TRUE.
        ELSE
            meta(3) = .FALSE.
        ENDIF
      CASE(65508)
        IF (itype.EQ.2) THEN
            meta(4) = .TRUE.
        ELSE
            meta(4) = .FALSE.
        ENDIF
      CASE(65505)
        IF (itype.EQ.2) THEN
            meta(5) = .TRUE.
        ELSE
            meta(5) = .FALSE.
        ENDIF
      CASE(65506)
        IF (itype.EQ.2) THEN
            meta(6) = .TRUE.
        ELSE
            meta(6) = .FALSE.
        ENDIF
      CASE(65515)
        IF (itype.EQ.2) THEN
            meta(7) = .TRUE.
        ELSE
            meta(7) = .FALSE.
        ENDIF
      CASE(65516)
        IF (itype.EQ.2) THEN
            meta(8) = .TRUE.
        ELSE
            meta(8) = .FALSE.
        ENDIF
      CASE DEFAULT
        event_set_meta = .TRUE.
    END SELECT
  END FUNCTION event_set_meta

!########################################################
  LOGICAL FUNCTION event_get_meta_alt()
!########################################################
    event_get_meta_alt = meta(1).OR.meta(2)
  END FUNCTION event_get_meta_alt

!########################################################
  LOGICAL FUNCTION event_get_meta_altleft()
!########################################################
    event_get_meta_altleft = meta(1)
  END FUNCTION event_get_meta_altleft

!########################################################
  LOGICAL FUNCTION event_get_meta_altright()
!########################################################
    event_get_meta_altright = meta(2)
  END FUNCTION event_get_meta_altright

!########################################################
  LOGICAL FUNCTION event_get_meta_ctrl()
!########################################################
    event_get_meta_ctrl = meta(3).OR.meta(4)
  END FUNCTION event_get_meta_ctrl

!########################################################
  LOGICAL FUNCTION event_get_meta_ctrlleft()
!########################################################
    event_get_meta_ctrlleft = meta(3)
  END FUNCTION event_get_meta_ctrlleft

!########################################################
  LOGICAL FUNCTION event_get_meta_ctrlright()
!########################################################
    event_get_meta_ctrlright = meta(4)
  END FUNCTION event_get_meta_ctrlright

!########################################################
  LOGICAL FUNCTION event_get_meta_shift()
!########################################################
    event_get_meta_shift = meta(5).OR.meta(6)
  END FUNCTION event_get_meta_shift

!########################################################
  LOGICAL FUNCTION event_get_meta_shiftleft()
!########################################################
    event_get_meta_shiftleft = meta(5)
  END FUNCTION event_get_meta_shiftleft

!########################################################
  LOGICAL FUNCTION event_get_meta_shiftright()
!########################################################
    event_get_meta_shiftright = meta(6)
  END FUNCTION event_get_meta_shiftright

!########################################################
  LOGICAL FUNCTION event_get_meta_meta()
!########################################################
    event_get_meta_meta = meta(7).OR.meta(8)
  END FUNCTION event_get_meta_meta

!########################################################
  LOGICAL FUNCTION event_get_meta_metaleft()
!########################################################
    event_get_meta_metaleft = meta(7)
  END FUNCTION event_get_meta_metaleft


!########################################################
  LOGICAL FUNCTION event_get_meta_metaright()
!########################################################
    event_get_meta_metaright = meta(8)
  END FUNCTION event_get_meta_metaright



!--------------------------------------------------------
!  primary setting

!########################################################
  INTEGER FUNCTION event_get_primary(key)
!########################################################
    INTEGER, INTENT(in) :: key
    event_get_primary = key / 100
  END FUNCTION event_get_primary



!--------------------------------------------------------
!  secondary setting

!########################################################
  INTEGER FUNCTION event_get_secondary(key)
!########################################################
    INTEGER, INTENT(in) :: key
    event_get_secondary = key - MOD(key,100)
  END FUNCTION event_get_secondary




!--------------------------------------------------------
!  print help

!########################################################
  SUBROUTINE event_help_2d()
!########################################################
    WRITE(*,*) ' '
    WRITE(*,*) '  Other controls:'
    WRITE(*,*) ' I  Unsteady Image dumping'
    WRITE(*,*) ' '
  END SUBROUTINE event_help_2d


!########################################################
  SUBROUTINE event_help_3d()
!########################################################
    WRITE(*,*) '  Feature Extraction:'
    WRITE(*,*) ' X      Feature Extraction'
    WRITE(*,*) ' '
    WRITE(*,*) '  Other Controls:'
    WRITE(*,*) ' Alt-f  Modify frame rate (fps)'
    WRITE(*,*) ' F      Time step toggle'
    WRITE(*,*) ' I      Unsteady image dumping'
    WRITE(*,*) ' N      Modify replication states'
    WRITE(*,*) ' P      Dump surfaces to Tecplot/POV'
    WRITE(*,*) ' Alt-p  Unsteady dump surfaces to Tecplot/POV'
    WRITE(*,*) ' S      Automatic scanning of cut planes and iso-surfaces'
    WRITE(*,*) ' '
  END SUBROUTINE event_help_3d


!########################################################
  SUBROUTINE event_help_dials()
!########################################################

  END SUBROUTINE event_help_dials



!--------------------------------------------------------
!  read sub-options

!########################################################
  LOGICAL FUNCTION event_var_avg(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar
    event_var_avg = .TRUE.

    istate = 11000
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter averaging option:'
    WRITE(*,*) ' A  Time-average'
    WRITE(*,*) ' D  Time-difference'
    WRITE(*,*) ' M  Time-minimum'
    WRITE(*,*) ' X  Time-maximum'
    WRITE(*,*) ' R  RMS'
    WRITE(*,*) ' T  FFT'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('A'))
        istate = istate + 1
      CASE(ICHAR('D'))
        istate = istate + 2
      CASE(ICHAR('M'))
        istate = istate + 3
      CASE(ICHAR('X'))
        istate = istate + 4
      CASE(ICHAR('R'))
        istate = istate + 5
      CASE(ICHAR('T'))
        istate = istate + 6
      CASE DEFAULT
        event_var_avg = .FALSE.
    END SELECT
  END FUNCTION event_var_avg

!########################################################
  LOGICAL FUNCTION event_var_vizmath(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_vizmath = .TRUE.
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired Viz Math option:'
    WRITE(*,*) ' c  Curl'
    WRITE(*,*) ' d  Divergence'
    WRITE(*,*) ' g  Gradient'
    WRITE(*,*) ' l  Laplacian'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('c'))
        istate = 11100
        WRITE(*,*) '  Enter desired curl option:'
      CASE(ICHAR('d'))
        istate = 11200
        RETURN
      CASE(ICHAR('g'))
        istate = 11300
        WRITE(*,*) '  Enter desired gradient option:'
      CASE(ICHAR('l'))
        istate = 11400
        RETURN
      CASE DEFAULT
        event_var_vizmath = .FALSE.
        RETURN
    END SELECT

    CALL CURSOR(.TRUE.)
    WRITE(*,*) ' h  Theta-component'
    WRITE(*,*) ' r  r-component'
    WRITE(*,*) ' s  Streamwise component (absolute)'
    WRITE(*,*) ' S  Streamwise component (relative)'
    WRITE(*,*) ' t  Absolute total'
    WRITE(*,*) ' T  Relative total'
    WRITE(*,*) ' x  X-component'
    WRITE(*,*) ' y  Y-component'
    WRITE(*,*) ' z  Absolute Z-component'
    WRITE(*,*) ' Z  Relative Z-component'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('h'))
        istate = istate + 1
      CASE(ICHAR('r'))
        istate = istate + 2
      CASE(ICHAR('s'))
        istate = istate + 3
      CASE(ICHAR('S'))
        istate = istate + 4
      CASE(ICHAR('t'))
        istate = istate + 5
      CASE(ICHAR('T'))
        istate = istate + 6
      CASE(ICHAR('x'))
        istate = istate + 7
      CASE(ICHAR('y'))
        istate = istate + 8
      CASE(ICHAR('z'))
        istate = istate + 9
      CASE(ICHAR('Z'))
        istate = istate + 10
      CASE DEFAULT
        event_var_vizmath = .FALSE.
    END SELECT

  END FUNCTION event_var_vizmath

!########################################################
  LOGICAL FUNCTION event_var_vector(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_vector = .TRUE.
    istate = 13200
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter vector option:'
    WRITE(*,*) ' V  Absolute Velocity'
    WRITE(*,*) ' W  Relative Velocity'
    WRITE(*,*) ' Z  Absolute Vorticity'
    WRITE(*,*) ' O  Relative Vorticity'
    READ*,inchar

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('V'))
        istate = istate + 1
      CASE(ICHAR('W'))
        istate = istate + 2
      CASE(ICHAR('Z'))
        istate = istate + 3
      CASE(ICHAR('O'))
        istate = istate + 4
      CASE DEFAULT
        event_var_vector = .FALSE.
    END SELECT
  END FUNCTION event_var_vector

!########################################################
  LOGICAL FUNCTION event_var_angles(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_angles = .TRUE.
    istate = 11600
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired angle option:'
    WRITE(*,*) ' a  Alpha-Z'
    WRITE(*,*) ' A  Alpha-M'
    WRITE(*,*) ' b  Beta-Z'
    WRITE(*,*) ' B  Beta-M'
    WRITE(*,*) ' p  Phi'
    READ*,inchar
    CALL CURSOR(.FALSE.)
    
    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('a'))
        istate = istate + 1
      CASE(ICHAR('A'))
        istate = istate + 2
      CASE(ICHAR('b'))
        istate = istate + 3
      CASE(ICHAR('B'))
        istate = istate + 4
      CASE(ICHAR('p'))
        istate = istate + 5
      CASE DEFAULT
        event_var_angles = .FALSE.
    END SELECT
  END FUNCTION event_var_angles

!########################################################
  LOGICAL FUNCTION event_var_coord(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_coord = .TRUE.
    istate = 11700
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired coordinate option:'
    WRITE(*,*) ' i  I or Ibb value'
    WRITE(*,*) ' j  J or Isp value'
    WRITE(*,*) ' k  K or Ifl value'
    WRITE(*,*) ' r  R value'
    WRITE(*,*) ' t  Theta value'
    WRITE(*,*) ' x  X value'
    WRITE(*,*) ' y  Y value'
    WRITE(*,*) ' z  Z value'
    READ*,inchar
    CALL CURSOR(.FALSE.)
    
    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('i'))
        istate = istate + 1
      CASE(ICHAR('j'))
        istate = istate + 2
      CASE(ICHAR('k'))
        istate = istate + 3
      CASE(ICHAR('r'))
        istate = istate + 4
      CASE(ICHAR('t'))
        istate = istate + 5
      CASE(ICHAR('x'))
        istate = istate + 6
      CASE(ICHAR('y'))
        istate = istate + 7
      CASE(ICHAR('z'))
        istate = istate + 8
      CASE DEFAULT
        event_var_coord = .FALSE.
    END SELECT
  END FUNCTION event_var_coord

!########################################################
  LOGICAL FUNCTION event_var_intengy(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_intengy = .TRUE.
    istate = 11800
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired energy option:'
    WRITE(*,*) ' e  Internal Energy'
    WRITE(*,*) ' r  Rho*Internal Energy'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('e'))
        istate = istate + 1
      CASE(ICHAR('r'))
        istate = istate + 2
      CASE DEFAULT
        event_var_intengy = .FALSE.
    END SELECT
  END FUNCTION event_var_intengy

!########################################################
  LOGICAL FUNCTION event_var_enthalpy(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_enthalpy = .TRUE.
    istate = 11900
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired enthalpy option:'
    WRITE(*,*) ' s  Static Enthalpy'
    WRITE(*,*) ' a  Absolute Total Enthalpy'
    WRITE(*,*) ' r  Relative Total Enthalpy'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('s'))
        istate = istate + 1
      CASE(ICHAR('a'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE DEFAULT
        event_var_enthalpy = .FALSE.
    END SELECT
  END FUNCTION event_var_enthalpy

!########################################################
  LOGICAL FUNCTION event_var_rothalpy(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_rothalpy = .TRUE.
    istate = 12000
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter the desired rothalpy option:'
    WRITE(*,*) ' i  rothalpy'
    WRITE(*,*) ' t  rothalpy/cp (temp units)'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('i'))
        istate = istate + 1
      CASE(ICHAR('t'))
        istate = istate + 2
      CASE DEFAULT
        event_var_rothalpy = .FALSE.
    END SELECT
  END FUNCTION event_var_rothalpy

!########################################################
  LOGICAL FUNCTION event_var_mach(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_mach = .TRUE.
    istate = 12100
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter the desired rho-v option:'
    WRITE(*,*) ' h  Absolute theta-component'
    WRITE(*,*) ' H  Relative theta-component'
    WRITE(*,*) ' r  r-component'
    WRITE(*,*) ' t  Absolute total'
    WRITE(*,*) ' T  Relative total'
    WRITE(*,*) ' x  Absolute x-component'
    WRITE(*,*) ' X  Relative x-component'
    WRITE(*,*) ' y  Absolute y-component'
    WRITE(*,*) ' Y  Relative Y-component'
    WRITE(*,*) ' z  Z-component'
    WRITE(*,*) ' i  Ideal mach number from ps and ptideal'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('h'))
        istate = istate + 1
      CASE(ICHAR('H'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE(ICHAR('t'))
        istate = istate + 4
      CASE(ICHAR('T'))
        istate = istate + 5
      CASE(ICHAR('x'))
        istate = istate + 6
      CASE(ICHAR('X'))
        istate = istate + 7
      CASE(ICHAR('y'))
        istate = istate + 8
      CASE(ICHAR('Y'))
        istate = istate + 9
      CASE(ICHAR('z'))
        istate = istate + 10
      CASE(ICHAR('i'))
        istate = istate + 11
      CASE DEFAULT
        event_var_mach = .FALSE.
    END SELECT
  END FUNCTION event_var_mach

!########################################################
  LOGICAL FUNCTION event_var_other(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_other = .TRUE.
    istate = 12200
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired option for other scalars:'
    WRITE(*,*) ' b  IBlanking'
    WRITE(*,*) ' g  Gamma'
    WRITE(*,*) ' i  Processor ID number'
    WRITE(*,*) ' n  Block Number'
    WRITE(*,*) ' o  Omega scalar value'
    WRITE(*,*) ' v  Cell volume'
    WRITE(*,*) ' s  Cell skew'
    WRITE(*,*) ' r  Local Reynolds number'
    WRITE(*,*) ' t  Residence time'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('b'))
        istate = istate + 1
      CASE(ICHAR('g'))
        istate = istate + 2
      CASE(ICHAR('i'))
        istate = istate + 3
      CASE(ICHAR('n'))
        istate = istate + 4
      CASE(ICHAR('o'))
        istate = istate + 5
      CASE(ICHAR('v'))
        istate = istate + 6
      CASE(ICHAR('s'))
        istate = istate + 7
      CASE(ICHAR('r'))
        istate = istate + 8
      CASE(ICHAR('t'))
        istate = istate + 9
      CASE DEFAULT
        event_var_other = .FALSE.
    END SELECT
  END FUNCTION event_var_other

!########################################################
  LOGICAL FUNCTION event_var_pressure(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_pressure = .TRUE.
    istate = 12300
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired pressure option:'
    WRITE(*,*) ' s  Static Pressure'
    WRITE(*,*) ' a  Absolute Total Pressure'
    WRITE(*,*) ' r  Relative Total Pressure'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('s'))
        istate = istate + 1
      CASE(ICHAR('a'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE DEFAULT
        event_var_pressure = .FALSE.
    END SELECT
  END FUNCTION event_var_pressure

!########################################################
  LOGICAL FUNCTION event_var_turbulence(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_turbulence = .TRUE.
    istate = 12400
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired turbulence option:'
    WRITE(*,*) ' m  Turbulence mu_t'
    WRITE(*,*) ' k  Turbulence K'
    WRITE(*,*) ' e  Turbulence Epsilon'
    WRITE(*,*) ' i  Turbulence Intensity'
    WRITE(*,*) ' q  Q Criterion'
    WRITE(*,*) ' r  Ratio Turbulent/Molecular Viscosity'
    WRITE(*,*) ' t  Total Viscosity (Turbulent+Molecular)'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('m'))
        istate = istate + 1
      CASE(ICHAR('k'))
        istate = istate + 2
      CASE(ICHAR('e'))
        istate = istate + 3
      CASE(ICHAR('i'))
        istate = istate + 4
      CASE(ICHAR('q'))
        istate = istate + 5
      CASE(ICHAR('r'))
        istate = istate + 6
      CASE(ICHAR('t'))
        istate = istate + 7
      CASE DEFAULT
        event_var_turbulence = .FALSE.
    END SELECT
  END FUNCTION event_var_turbulence

!########################################################
  LOGICAL FUNCTION event_var_rhov(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_rhov = .TRUE.
    istate = 12500
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter the desired rho-v option:'
    WRITE(*,*) ' h  Absolute theta-component'
    WRITE(*,*) ' H  Relative theta-component'
    WRITE(*,*) ' r  r-component'
    WRITE(*,*) ' t  Absolute total'
    WRITE(*,*) ' T  Relative total'
    WRITE(*,*) ' x  Absolute x-component'
    WRITE(*,*) ' X  Relative x-component'
    WRITE(*,*) ' y  Absolute y-component'
    WRITE(*,*) ' Y  Relative Y-component'
    WRITE(*,*) ' z  Z-component'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('h'))
        istate = istate + 1
      CASE(ICHAR('H'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE(ICHAR('t'))
        istate = istate + 4
      CASE(ICHAR('T'))
        istate = istate + 5
      CASE(ICHAR('x'))
        istate = istate + 6
      CASE(ICHAR('X'))
        istate = istate + 7
      CASE(ICHAR('y'))
        istate = istate + 8
      CASE(ICHAR('Y'))
        istate = istate + 9
      CASE(ICHAR('z'))
        istate = istate + 10
      CASE DEFAULT
        event_var_rhov = .FALSE.
    END SELECT
  END FUNCTION event_var_rhov

!########################################################
  LOGICAL FUNCTION event_var_entropy(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_entropy = .TRUE.
    istate = 12600
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired entropy option:'
    WRITE(*,*) ' a  Absolute level'
    WRITE(*,*) ' e  Exp(s/R)'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('a'))
        istate = istate + 1
      CASE(ICHAR('e'))
        istate = istate + 2
      CASE DEFAULT
        event_var_entropy = .FALSE.
    END SELECT
  END FUNCTION event_var_entropy

!########################################################
  LOGICAL FUNCTION event_var_temperature(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_temperature = .TRUE.
    istate = 12700
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter the desired temperature option:'
    WRITE(*,*) ' s  Static temperature'
    WRITE(*,*) ' a  Absolute total temperature'
    WRITE(*,*) ' r  Relative total temperature'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('s'))
        istate = istate + 1
      CASE(ICHAR('a'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE DEFAULT
        event_var_temperature = .FALSE.
    END SELECT
  END FUNCTION event_var_temperature

!########################################################
  LOGICAL FUNCTION event_var_velocity(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_velocity = .TRUE.
    istate = 12800
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter the desired velocity option:'
    WRITE(*,*) ' h  Absolute theta-component'
    WRITE(*,*) ' H  Relative theta-component'
    WRITE(*,*) ' r  r-component'
    WRITE(*,*) ' t  Absolute total'
    WRITE(*,*) ' T  Relative total'
    WRITE(*,*) ' x  Absolute x-component'
    WRITE(*,*) ' X  Relative x-component'
    WRITE(*,*) ' y  Absolute y-component'
    WRITE(*,*) ' Y  Relative Y-component'
    WRITE(*,*) ' z  Z-component'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('h'))
        istate = istate + 1
      CASE(ICHAR('H'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE(ICHAR('t'))
        istate = istate + 4
      CASE(ICHAR('T'))
        istate = istate + 5
      CASE(ICHAR('x'))
        istate = istate + 6
      CASE(ICHAR('X'))
        istate = istate + 7
      CASE(ICHAR('y'))
        istate = istate + 8
      CASE(ICHAR('Y'))
        istate = istate + 9
      CASE(ICHAR('z'))
        istate = istate + 10
      CASE DEFAULT
        event_var_velocity = .FALSE.
    END SELECT
  END FUNCTION event_var_velocity

!########################################################
  LOGICAL FUNCTION event_var_vorticity(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_vorticity = .TRUE.
    istate = 12900
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter desired vorticity option:'
    WRITE(*,*) ' h  Theta-component'
    WRITE(*,*) ' r  r-component'
    WRITE(*,*) ' s  Streamwise component (absolute)'
    WRITE(*,*) ' S  Streamwise component (relative)'
    WRITE(*,*) ' t  Absolute total'
    WRITE(*,*) ' T  Relative total'
    WRITE(*,*) ' x  X-component'
    WRITE(*,*) ' y  Y-component'
    WRITE(*,*) ' z  Absolute Z-component'
    WRITE(*,*) ' Z  Relative Z-component'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('h'))
        istate = istate + 1
      CASE(ICHAR('H'))
        istate = istate + 2
      CASE(ICHAR('r'))
        istate = istate + 3
      CASE(ICHAR('t'))
        istate = istate + 4
      CASE(ICHAR('T'))
        istate = istate + 5
      CASE(ICHAR('x'))
        istate = istate + 6
      CASE(ICHAR('X'))
        istate = istate + 7
      CASE(ICHAR('y'))
        istate = istate + 8
      CASE(ICHAR('Y'))
        istate = istate + 9
      CASE(ICHAR('z'))
        istate = istate + 10
      CASE DEFAULT
        event_var_vorticity = .FALSE.
    END SELECT
  END FUNCTION event_var_vorticity

!########################################################
  LOGICAL FUNCTION event_var_func(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    INTEGER :: inint

    event_var_func = .TRUE.
    istate = 13000
    CALL CURSOR(.TRUE.)
    WRITE(*,"(A,I4,A)") '  Enter number of the function value to display [1-',nfunc,']'
    READ*,inint
    CALL CURSOR(.FALSE.)

    IF (inint.LT.nfunc .AND. inint.GE.1) THEN
        istate = istate + inint
    ELSE
        event_var_func = .FALSE.
    ENDIF
  END FUNCTION event_var_func

!########################################################
  LOGICAL FUNCTION event_var_threshold(istate)
!########################################################
    INTEGER, INTENT(inout) :: istate
    CHARACTER :: inchar

    event_var_threshold = .TRUE.
    istate = 13100
    CALL CURSOR(.TRUE.)
    WRITE(*,*) '  Enter thresholding option:'
    WRITE(*,*) ' i  I-Coordinate'
    WRITE(*,*) ' j  J-Coordinate'
    WRITE(*,*) ' k  K-Coordinate'
    WRITE(*,*) ' x  X-Coordinate'
    WRITE(*,*) ' y  Y-Coordinate'
    WRITE(*,*) ' z  Z-Coordinate'
    READ*,inchar
    CALL CURSOR(.FALSE.)

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('i'))
        istate = istate + 1
      CASE(ICHAR('j'))
        istate = istate + 2
      CASE(ICHAR('k'))
        istate = istate + 3
      CASE(ICHAR('x'))
        istate = istate + 4
      CASE(ICHAR('y'))
        istate = istate + 5
      CASE(ICHAR('z'))
        istate = istate + 6
      CASE DEFAULT
        event_var_threshold = .FALSE.
    END SELECT
  END FUNCTION event_var_threshold


!########################################################
  SUBROUTINE event_opt_domainmap(estr)
!########################################################
    INTEGER, INTENT(in) :: estr

!    CALL event_set_settype(estr,2)
!    CALL event_set_primary(estr,)
    CALL CURSOR(.TRUE.)
!    CALL event_set_secondary(event_opt_mapping('x-prime') + event_opt_mapping('y-prime')*10)
    CALL CURSOR(.FALSE.)

  END SUBROUTINE event_opt_domainmap

!########################################################
  INTEGER FUNCTION event_opt_mapping(str)
!########################################################
    CHARACTER(LEN=7), INTENT(in) :: str
    CHARACTER :: inchar

    event_opt_mapping = 0
    WRITE(*,*) '  Enter desired mapping for ',str
    WRITE(*,*) ' r  Radial'
    WRITE(*,*) ' t  Tangential'
    WRITE(*,*) ' z  Axial/Grid coordinate'
    WRITE(*,*) ' x  Grid coordinate'
    WRITE(*,*) ' y  Grid coordinate'
    WRITE(*,*) ' i  Block index'
    WRITE(*,*) ' j  Block index'
    WRITE(*,*) ' k  Block index'
    READ(*,*) inchar

    SELECT CASE (ICHAR(inchar))
      CASE(ICHAR('r'))
        event_opt_mapping = 1
      CASE(ICHAR('t'))
        event_opt_mapping = 2
      CASE(ICHAR('z'))
        event_opt_mapping = 3
      CASE(ICHAR('x'))
        event_opt_mapping = 4
      CASE(ICHAR('y'))
        event_opt_mapping = 5
      CASE(ICHAR('i'))
        event_opt_mapping = 6
      CASE(ICHAR('j'))
        event_opt_mapping = 7
      CASE(ICHAR('k'))
        event_opt_mapping = 8
    END SELECT
  END FUNCTION event_opt_mapping

!########################################################
  SUBROUTINE event_opt_featureext(estr)
!########################################################
    INTEGER, INTENT(in) :: estr
    CHARACTER :: inchar

!    CALL event_set_settype(estr,2)
!    CALL event_set_primary(estr,)
    IF (.NOT. ANY(extracton)) THEN
        CALL CURSOR(.TRUE.)
        WRITE(*,*) '  Enter desired FX extract:'
        WRITE(*,*) ' C  Vortex Cores'
        WRITE(*,*) ' S  Shock Surfaces'
        WRITE(*,*) ' L  Separation and Attachment Lines'
        WRITE(*,*) ' B  Boundary Layer/Wake Surfaces'
        READ*,inchar
        CALL CURSOR(.FALSE.)

        SELECT CASE (ICHAR(inchar))
          CASE(ICHAR('C'))
            CALL ToggleExtract(1, 'Vortex Cores        ')
          CASE(ICHAR('S'))
            CALL ToggleExtract(2, 'Shock Surfaces      ')
          CASE(ICHAR('L'))
            CALL ToggleExtract(3, 'Sep & Att Lines     ')
          CASE(ICHAR('B'))
            CALL ToggleExtract(5, 'Boundary Layer/Wake ')
        END SELECT
    ELSE
        CALL ToggleExtractAllOff
    ENDIF
  END SUBROUTINE event_opt_featureext

!########################################################
  SUBROUTINE event_opt_disprate(estr,istate)
!########################################################
    INTEGER, INTENT(in) :: estr, istate
    CHARACTER :: inchar

    REAL :: rtemp
    INTEGER :: myfps

!    CALL event_set_settype(estr,4)
    SELECT CASE (istate)
      CASE(65362)
        IF (auto) THEN
            ds = 1.5 * ds
        ELSE
            myfps = myfps + 1
        ENDIF
      CASE(65364)
        IF (auto) THEN
            ds = 0.75 * ds
        ELSE
            myfps = myfps - 1
        ENDIF
      CASE(ICHAR('f'))
        CALL CURSOR(.TRUE.)
        rtemp = 1./urate
1102    WRITE(*,*) ' Enter desired framerate (fps): [Currrent',rtemp,']'
        READ(*,*,ERR=1102) rtemp
        urate = 1./rtemp
        CALL CURSOR(.FALSE.)
      CASE(65451)
        urate = urate - urate*0.1
        IF(urate.EQ.0.0) urate = 0.05
        WRITE(*,*) '  Update Rate set to ',urate*FLOAT(udirec),' sec/step'
      CASE(65453)
        urate = urate + urate*0.1
        IF(urate.EQ.0.0) urate = 0.05
        WRITE(*,*) '  Update Rate set to ',urate*FLOAT(udirec),' sec/step'
      CASE(65299)
        IF (urate == 0.0) THEN
                  urate = 0.1
              ELSE
                  urate = 0.0
              END IF
      CASE(65450)
        udirec = -udirec
      CASE(65300)
        usync = 1 - usync
        IF (usync.EQ.1) THEN
            WRITE(*,*) '  Timestepping now synchronous with render cycle'
        ELSE
            WRITE(*,*) '  Timestepping now asynchronous'
            WRITE(*,*) '    Update Rate set to ',urate*FLOAT(udirec),' sec/step'
        ENDIF
    END SELECT
  END SUBROUTINE event_opt_disprate


END MODULE mod_events
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
