MODULE solver_uns
  IMPLICIT NONE

  INTEGER :: fv_magic, fv_faces, fv_elements, fv_variables, fv_bndry_vars
  INTEGER :: fv_elem_type_byte_shift = 18
  INTEGER :: nodesflag
  INTEGER, ALLOCATABLE, DIMENSION(:) :: facebased, clockcons
  
  INTEGER :: avus_ndm

  TYPE fvfaces
    INTEGER, POINTER, DIMENSION(:,:) :: f
  END TYPE

  TYPE fvstore
    CHARACTER*80, POINTER, DIMENSION(:) :: bcnam
    CHARACTER*80, POINTER, DIMENSION(:) :: varnames
    TYPE(fvfaces), POINTER, DIMENSION(:) :: f
    INTEGER, POINTER, DIMENSION(:) :: nbfaces
    INTEGER :: nbcs, nvars, nbvars
    INTEGER, POINTER, DIMENSION(:,:) :: c
    INTEGER, POINTER, DIMENSION(:) :: typ
  END TYPE fvstore

  TYPE(fvstore), POINTER, DIMENSION(:) :: fvd

CONTAINS
!#######################################################################
  SUBROUTINE Store_AVUS_2D(xyz,npatch,ipnts,npnts,nfaces,icell,ncells,numppf,n)
!#######################################################################
    USE avusconv
    USE avusconvvar
    USE common_area, ONLY: pgr, avusbcfilename

    REAL, POINTER, DIMENSION(:,:) :: xyz
    INTEGER, POINTER, DIMENSION(:,:) :: ipnts, icell
    INTEGER, POINTER, DIMENSION(:) :: npnts, nfaces, ncells, numppf
    INTEGER, INTENT(IN) :: npatch, n ! Current grid
    INTEGER, DIMENSION(npatch) :: bndcnt,bndloc
    CHARACTER*80, DIMENSION(npatch) :: bcname
    INTEGER :: i,j,k,sumcells,sumfaces,npts,cnt
    INTEGER :: num_tet,num_hex,num_prism,num_pyramid
    INTEGER istat



    CALL readavusbc(avusbcfilename,npatch,bcname)

    npts = SUM(npnts)
    pgr(n)%tpts = npts*2
    sumcells = SUM(ncells)
    sumfaces = SUM(nfaces)

    ALLOCATE(pgr(n)%x(npts*2),pgr(n)%y(npts*2),pgr(n)%z(npts*2))
    DO i=1,npts
      pgr(n)%x(i) = xyz(1,i)
      pgr(n)%y(i) = xyz(2,i)
      pgr(n)%z(i) = 0.0
    ENDDO
    DO i=1,npts
      pgr(n)%x(npts+i) = xyz(1,i)
      pgr(n)%y(npts+i) = xyz(2,i)
      pgr(n)%z(npts+i) = 1.0
    ENDDO

    CALL Set_AVUS_Cells(npatch,sumfaces,ipnts,icell,sumcells,numppf,bndcnt,bndloc,bcname)


     num_tet = 0
     num_hex = 0
     num_prism = 0
     num_pyramid = 0

     ALLOCATE(fvd(n)%c(8,sumcells),fvd(n)%typ(sumcells),fvd(n)%f(npatch),fvd(n)%nbfaces(npatch),fvd(n)%bcnam(npatch))
     fvd(n)%nbcs = npatch !bc count
     fvd(n)%bcnam = bcname
     DO i = 1, npatch
        ALLOCATE(fvd(n)%f(i)%f(4,bndcnt(i)))
        fvd(n)%nbfaces(i) = bndcnt(i)
     ENDDO


     do i=1,sumcells
! triangle extruded to prism
           if (cell(i)%nf == 3) then
              num_prism = num_prism + 1
              cell(i)%np = 6
              tmpfac(1:3) = face(cell(i)%faces(1:3))

              call user_tri(tmpfac,cell(i)%nodes,cell(i)%wall,npts)
! quadrilateral extruded to hexahedral
           elseif (cell(i)%nf == 4) then
              num_hex = num_hex + 1
              cell(i)%np = 8
              tmpfac(1:4) = face(cell(i)%faces(1:4))
              call user_quad(tmpfac,cell(i)%nodes,cell(i)%wall,npts)
           else
              write(*,*) 'Cell type unavailable...'
              istat = 1
           end if

! Copy cell(i) into fvd
      DO j=1,cell(i)%np
        fvd(n)%c(j,i) = cell(i)%nodes(j)
      ENDDO
      SELECT CASE(cell(i)%np)
        CASE(6)
          fvd(n)%typ(i) = 3
        CASE(8)
          fvd(n)%typ(i) = 4
      END SELECT
    enddo


     do k=1,npatch
        cnt = count(face(:)%bc == k)
        j = 0
        do i=1,sumfaces
           if (face(i)%bc == k) then
              j = j + 1
              fvd(n)%f(k)%f(1:2,j) = face(i)%nodes(1:2)
              fvd(n)%f(k)%f(3:4,j) = face(i)%nodes(2:1:-1) + npts
           end if
        end do
     end do


    pgr(n)%ntet = num_tet
    pgr(n)%nhex = num_hex
    pgr(n)%npris = num_prism
    pgr(n)%npyr = num_pyramid
    pgr(n)%tcells = num_tet + num_hex + num_prism + num_pyramid

    CALL CleanAvusConv()

    RETURN

  END SUBROUTINE Store_AVUS_2D

!#######################################################################
  SUBROUTINE Store_AVUS_3D(xyz,npatch,ipnts,npnts,nfaces,icell,ncells,numppf,n)
!#######################################################################
    USE avusconv
    USE avusconvvar
    USE common_area, ONLY: pgr, avusbcfilename

    REAL, POINTER, DIMENSION(:,:) :: xyz
    INTEGER, POINTER, DIMENSION(:,:) :: ipnts, icell
    INTEGER, POINTER, DIMENSION(:) :: npnts, nfaces, ncells, numppf
    INTEGER, INTENT(IN) :: npatch, n ! Current grid
    INTEGER, DIMENSION(npatch) :: bndcnt,bndloc
    CHARACTER*80, DIMENSION(npatch) :: bcname
    INTEGER :: i,j,k,sumcells,sumfaces,npts,idum,cnt,nid,fid
    INTEGER :: num_tet,num_hex,num_prism,num_pyramid
    INTEGER istat

    CALL readavusbc(avusbcfilename,npatch,bcname)

    npts = SUM(npnts)
    pgr(n)%tpts = npts
    sumcells = SUM(ncells)
    sumfaces = SUM(nfaces)

    ALLOCATE(pgr(n)%x(npts),pgr(n)%y(npts),pgr(n)%z(npts))
    DO i=1,npts
      pgr(n)%x(i) = xyz(1,i)
      pgr(n)%y(i) = xyz(2,i)
      pgr(n)%z(i) = xyz(3,i)
    ENDDO

    CALL Set_AVUS_Cells(npatch,sumfaces,ipnts,icell,sumcells,numppf,bndcnt,bndloc,bcname)

     num_tet = 0
     num_hex = 0
     num_prism = 0
     num_pyramid = 0

     ALLOCATE(fvd(n)%c(8,sumcells),fvd(n)%typ(sumcells),fvd(n)%f(npatch),fvd(n)%nbfaces(npatch),fvd(n)%bcnam(npatch))
     fvd(n)%nbcs = npatch !bc count
     fvd(n)%bcnam = bcname
     DO i = 1, npatch
        ALLOCATE(fvd(n)%f(i)%f(4,bndcnt(i)))
        fvd(n)%nbfaces(i) = bndcnt(i)
     ENDDO


     do i=1,sumcells

! tetrahedral

           if(cell(i)%nf == 4) then
              num_tet = num_tet + 1
              cell(i)%np = 4
              tmpfac(1:4) = face(cell(i)%faces(1:4))

              call user_tet(tmpfac,cell(i)%nodes,cell(i)%wall)

              fvd(n)%typ(i) = 1
! hexahedral

           elseif(cell(i)%nf == 6) then
              num_hex = num_hex + 1
              cell(i)%np = 8
              tmpfac = face(cell(i)%faces)

              call user_hex(tmpfac,cell(i)%nodes,cell(i)%wall)
         
              fvd(n)%typ(i) = 4


!-----------------------------------------------------------------------
! Flip the 3/4 and 7/8 values for the hexahedral elements because FV
! programmers are jerks and just wanted extra work to do!!!
            idum = cell(i)%nodes(4)
            cell(i)%nodes(4) = cell(i)%nodes(3)
            cell(i)%nodes(3) = idum
            idum = cell(i)%nodes(8)
            cell(i)%nodes(8) = cell(i)%nodes(7)
            cell(i)%nodes(7) = idum
!-----------------------------------------------------------------------




! pyramid or prism

           elseif(cell(i)%nf == 5) then

! check number of faces with 4 points - 1 face = pyramid, 3 faces = prism
              nid = 0
              do j=1,5
                 fid = cell(i)%faces(j)
                 tmpfac(j) = face(fid)
                 if(face(fid)%np == 4) then
                    nid = nid + 1
                    k = j
                 endif
              end do

! put four-sided face in tmpfac(1) for user_prism and user_pyr subroutines
              if (k /= 1) then
                 tmpfac(6) = tmpfac(1)
                 tmpfac(1) = tmpfac(k)
                 tmpfac(k) = tmpfac(6)
              end if

! prism
              if(nid == 3) then
                 num_prism = num_prism + 1
                 cell(i)%np = 6

                 call user_prism(tmpfac,cell(i)%nodes,cell(i)%wall)
          
                 fvd(n)%typ(i) = 3
! pyramid
              elseif(nid == 1) then
                 num_pyramid = num_pyramid + 1
                 cell(i)%np = 5

                 call user_pyr(tmpfac,cell(i)%nodes,cell(i)%wall)
          
                
                 fvd(n)%typ(i) = 2

! Unavailable cell type
              else
                 write(*,*) 'Cell type unavailable...'
                 stop
              endif

! none of the available types
           else
              write(*,*) 'Cell type unavailable...'
              stop
           endif

! Copy cell(i) into fvd
      DO j=1,cell(i)%np
        fvd(n)%c(j,i) = cell(i)%nodes(j)
      ENDDO
     end do     !end loop over cells


     do k=1,npatch
        cnt = count(face(:)%bc == bndloc(k))
        j = 0
        do i=1,sumfaces
           if (face(i)%bc == k) then
              j = j + 1
              fvd(n)%f(k)%f(1:4,j) = face(i)%nodes(1:4)
           end if
        end do
     end do

        
    pgr(n)%ntet = num_tet
    pgr(n)%nhex = num_hex
    pgr(n)%npris = num_prism
    pgr(n)%npyr = num_pyramid
    pgr(n)%tcells = num_tet + num_hex + num_prism + num_pyramid

    CALL CleanAvusConv()

    RETURN
        
  END SUBROUTINE Store_AVUS_3D

END MODULE solver_uns
