!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE general
  IMPLICIT NONE

  INTEGER :: iwinsizes(9)

CONTAINS

!#######################################################################
  SUBROUTINE Str2Int(str,val)
!#######################################################################
! Zero is ascii code 48, i.e. ichar('0') == 48
    CHARACTER(*),INTENT(IN) :: str
    INTEGER,INTENT(OUT) :: val
    INTEGER :: tmp
    INTEGER :: length, i, ctr

    length=LEN(TRIM(str))
    ctr = 0
    tmp = 0
    DO i=length,1,-1
      tmp = tmp + (10**ctr)*(ICHAR(str(i:i))-48)
      ctr=ctr+1
    ENDDO
    val = tmp

  END SUBROUTINE Str2Int

!#######################################################################
  SUBROUTINE Prompt_Get(var,msg)
!#######################################################################
    CHARACTER(*),INTENT(OUT) :: var
    CHARACTER(*),INTENT(IN) :: msg

    WRITE(*,*) msg
    READ*,var
    RETURN

  END SUBROUTINE Prompt_Get

!#######################################################################
  SUBROUTINE Check_File_Exist(fnam,ios)
!#######################################################################
    LOGICAL,INTENT(OUT) :: ios
    CHARACTER(*),INTENT(IN) :: fnam

    INQUIRE(FILE=fnam,EXIST=ios)
    RETURN

  END SUBROUTINE Check_File_Exist

!#######################################################################
  SUBROUTINE Fatal_Error(ios,msg)
!#######################################################################
    LOGICAL,INTENT(IN) :: ios
    CHARACTER(*),INTENT(IN) :: msg

    IF(ios)THEN
      RETURN
    ELSE
      WRITE(*,*) msg
      STOP
    ENDIF

  END SUBROUTINE Fatal_Error

!#######################################################################
  FUNCTION int_to_string(i)
!#######################################################################
    INTEGER :: i
    CHARACTER(LEN=50) :: int_to_string

    IF(i.LE.9)THEN
      WRITE(int_to_string,'(i1)') i
    ELSEIF(i.GE.10 .AND. i.LE.99)THEN
      WRITE(int_to_string,'(i2)') i
    ELSEIF(i.GE.100 .AND. i.LE.999)THEN
      WRITE(int_to_string,'(i3)') i
    ELSEIF(i.GE.1000 .AND. i.LE.9999)THEN
      WRITE(int_to_string,'(i4)') i
    ELSEIF(i.GE.10000 .AND. i.LE.99999)THEN
      WRITE(int_to_string,'(i5)') i
    ELSEIF(i.GE.100000 .AND. i.LE.999999)THEN
      WRITE(int_to_string,'(i6)') i
    ENDIF

    RETURN
  END FUNCTION int_to_string

!#######################################################################
  SUBROUTINE MakeUpper(str,slen)
!#######################################################################
    IMPLICIT NONE

    INTEGER, INTENT(IN) :: slen
    CHARACTER(LEN=slen),INTENT(INOUT) :: str

    INTEGER :: shift=0, ilen=0, i=0, ich=0

    shift = ICHAR('a')-ICHAR('A')
    ilen = LEN_TRIM(str)

    DO i = 1, ilen
      ich = ICHAR(str(i:i))
      IF(ich.GE.ICHAR('a') .AND. ich.LE.ICHAR('z')) str(i:i) = CHAR(ich-shift)
    ENDDO

    RETURN
 
  END SUBROUTINE MakeUpper

!#######################################################################
  SUBROUTINE KeywordSub(string,keyword,outint)
!#######################################################################
    CHARACTER(LEN=20),INTENT(IN) :: string, keyword
    INTEGER,INTENT(INOUT) :: outint

    INTEGER :: i, j, jrel,  strlen, keylen, ik
    INTEGER :: keyint

    CHARACTER(LEN=20) :: stringtemp, keywordtemp

    stringtemp=string
    keywordtemp=keyword
    stringtemp=TRIM(ADJUSTL(stringtemp))
    keywordtemp=TRIM(ADJUSTL(keywordtemp))
    strlen=LEN_TRIM(stringtemp)
    keylen=LEN_TRIM(keywordtemp)
    CALL MakeUpper(stringtemp,strlen)
    CALL MakeUpper(keywordtemp,keylen)
    keyint=0
    DO i=1,strlen
      IF(stringtemp(i:i).NE.keywordtemp(1:1)) then
        keyint=keyint+1
      ELSE
        ik=keyint+keylen
        keyint=keyint+1
        DO j=keyint,ik
          jrel=j-i+1
          IF(stringtemp(j:j).NE.keywordtemp(jrel:jrel))GOTO 10
        ENDDO
        EXIT
      ENDIF
    
10    CONTINUE
    ENDDO    

    IF(keyint.LT.strlen) then
      outint=keyint
    ELSE
      outint=-1
    ENDIF

  END SUBROUTINE KeywordSub

!#######################################################################
  INTEGER FUNCTION KeywordFunc(string,keyword)
!#######################################################################
    CHARACTER(LEN=20),INTENT(IN) :: string, keyword

    INTEGER :: i, j, jrel,  strlen, keylen, ik
    INTEGER :: keyint

    CHARACTER(LEN=20) :: stringtemp, keywordtemp

    stringtemp=string
    keywordtemp=keyword
    !stringtemp=TRIM(ADJUSTL(stringtemp))
    !keywordtemp=TRIM(ADJUSTL(keywordtemp))
    strlen=LEN_TRIM(stringtemp)
    keylen=LEN_TRIM(keywordtemp)
    CALL MakeUpper(stringtemp,strlen)
    CALL MakeUpper(keywordtemp,keylen)
    keyint=0
    DO i=1,strlen
      IF(stringtemp(i:i).NE.keywordtemp(1:1)) then
        keyint=keyint+1
      ELSE
        ik=keyint+keylen
        keyint=keyint+1
        DO j=keyint,ik
          jrel=j-i+1
          IF(stringtemp(j:j).NE.keywordtemp(jrel:jrel))GOTO 10
        ENDDO
        EXIT
      ENDIF
    
10    CONTINUE
    ENDDO    

    IF(keyint.LT.strlen) then
      KeywordFunc=keyint
    ELSE
      KeywordFunc=-1
    ENDIF

  END FUNCTION KeywordFunc

!#######################################################################
  SUBROUTINE BitsOn(larr,n,num)
!#######################################################################
    INTEGER,INTENT(IN) :: n, num
    LOGICAL,POINTER,DIMENSION(:) :: larr

    INTEGER :: i

    ALLOCATE(larr(0:n))
    DO i = 0, n
      larr(i) = BTEST(num,i)
    ENDDO

  END SUBROUTINE BitsOn

END MODULE general
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE averaging
  IMPLICIT NONE

CONTAINS

!#######################################################################
  REAL FUNCTION AVGV(X,imax,jmax,kmax,i,j,k)
!#######################################################################
! This function finds the average of the vertex array X for the 8 vertices
! defined by the coordinates (i,j,k)-(i+1,j+1,k+1).

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV = X(i,j,k)
  AVGV = AVGV + X(i+1,j  ,k)
  AVGV = AVGV + X(i+1,j+1,k)
  AVGV = AVGV + X(i  ,j+1,k)
  AVGV = AVGV + X(i  ,j  ,k+1)
  AVGV = AVGV + X(i+1,j  ,k+1)
  AVGV = AVGV + X(i+1,j+1,k+1)
  AVGV = AVGV + X(i  ,j+1,k+1)

  AVGV = AVGV/8.

  END FUNCTION AVGV

!#######################################################################
  REAL FUNCTION AVGV6_11a(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV6_11a = 0.
  AVGV6_11a = AVGV6_11a + X(i+1,  j,k)
  AVGV6_11a = AVGV6_11a + X(  i,j+1,k)
  AVGV6_11a = AVGV6_11a + X(i+1,j+1,k)
  AVGV6_11a = AVGV6_11a + X(i+1,  j,k+1)
  AVGV6_11a = AVGV6_11a + X(  i,j+1,k+1)
  AVGV6_11a = AVGV6_11a + X(i+1,j+1,k+1)

  AVGV6_11a = AVGV6_11a/6.

  END FUNCTION AVGV6_11a

!#######################################################################
  REAL FUNCTION AVGV6_1Ma(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV6_1Ma = 0.
  AVGV6_1Ma = AVGV6_1Ma + X(  i,  j,k)
  AVGV6_1Ma = AVGV6_1Ma + X(i+1,  j,k)
  AVGV6_1Ma = AVGV6_1Ma + X(i+1,j+1,k)
  AVGV6_1Ma = AVGV6_1Ma + X(  i,  j,k+1)
  AVGV6_1Ma = AVGV6_1Ma + X(i+1,  j,k+1)
  AVGV6_1Ma = AVGV6_1Ma + X(i+1,j+1,k+1)

  AVGV6_1Ma = AVGV6_1Ma/6.

  END FUNCTION AVGV6_1Ma

!#######################################################################
  REAL FUNCTION AVGV6_M1a(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV6_M1a = 0.
  AVGV6_M1a = AVGV6_M1a + X(  i,  j,k)
  AVGV6_M1a = AVGV6_M1a + X(  i,j+1,k)
  AVGV6_M1a = AVGV6_M1a + X(i+1,j+1,k)
  AVGV6_M1a = AVGV6_M1a + X(  i,  j,k+1)
  AVGV6_M1a = AVGV6_M1a + X(  i,j+1,k+1)
  AVGV6_M1a = AVGV6_M1a + X(i+1,j+1,k+1)

  AVGV6_M1a = AVGV6_M1a/6.

  END FUNCTION AVGV6_M1a

!#######################################################################
  REAL FUNCTION AVGV6_MMa(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV6_MMa = 0.
  AVGV6_MMa = AVGV6_MMa + X(  i,  j,k)
  AVGV6_MMa = AVGV6_MMa + X(  i,j+1,k)
  AVGV6_MMa = AVGV6_MMa + X(i+1,  j,k)
  AVGV6_MMa = AVGV6_MMa + X(  i,  j,k+1)
  AVGV6_MMa = AVGV6_MMa + X(  i,j+1,k+1)
  AVGV6_MMa = AVGV6_MMa + X(i+1,  j,k+1)

  AVGV6_MMa = AVGV6_MMa/6.

  END FUNCTION AVGV6_MMa

!#######################################################################
  REAL FUNCTION AVGV6_1a1(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_1a1 = 0.
    AVGV6_1a1 = AVGV6_1a1 + X(  i,  j,k+1)
    AVGV6_1a1 = AVGV6_1a1 + X(i+1,  j,  k)
    AVGV6_1a1 = AVGV6_1a1 + X(i+1,  j,k+1)
    AVGV6_1a1 = AVGV6_1a1 + X(  i,j+1,k+1)
    AVGV6_1a1 = AVGV6_1a1 + X(i+1,j+1,  k)
    AVGV6_1a1 = AVGV6_1a1 + X(i+1,j+1,k+1)
    AVGV6_1a1 = AVGV6_1a1/6.

  END FUNCTION AVGV6_1a1

!#######################################################################
  REAL FUNCTION AVGV6_1aM(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_1aM = 0.
    AVGV6_1aM = AVGV6_1aM + X(  i,  j,  k)
    AVGV6_1aM = AVGV6_1aM + X(i+1,  j,  k)
    AVGV6_1aM = AVGV6_1aM + X(i+1,  j,k+1)
    AVGV6_1aM = AVGV6_1aM + X(  i,j+1,  k)
    AVGV6_1aM = AVGV6_1aM + X(i+1,j+1,  k)
    AVGV6_1aM = AVGV6_1aM + X(i+1,j+1,k+1)
    AVGV6_1aM = AVGV6_1aM/6.

  END FUNCTION AVGV6_1aM

!#######################################################################
  REAL FUNCTION AVGV6_Ma1(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_Ma1 = 0.
    AVGV6_Ma1 = AVGV6_Ma1 + X(  i,  j,  k)
    AVGV6_Ma1 = AVGV6_Ma1 + X(  i,  j,k+1)
    AVGV6_Ma1 = AVGV6_Ma1 + X(i+1,  j,k+1)
    AVGV6_Ma1 = AVGV6_Ma1 + X(  i,j+1,  k)
    AVGV6_Ma1 = AVGV6_Ma1 + X(  i,j+1,k+1)
    AVGV6_Ma1 = AVGV6_Ma1 + X(i+1,j+1,k+1)
    AVGV6_Ma1 = AVGV6_Ma1/6.

  END FUNCTION AVGV6_Ma1

!#######################################################################
  REAL FUNCTION AVGV6_MaM(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_MaM = 0.
    AVGV6_MaM = AVGV6_MaM + X(  i,  j,  k)
    AVGV6_MaM = AVGV6_MaM + X(  i,  j,k+1)
    AVGV6_MaM = AVGV6_MaM + X(i+1,  j,  k)
    AVGV6_MaM = AVGV6_MaM + X(  i,j+1,  k)
    AVGV6_MaM = AVGV6_MaM + X(  i,j+1,k+1)
    AVGV6_MaM = AVGV6_MaM + X(i+1,j+1,  k)
    AVGV6_MaM = AVGV6_MaM/6.

  END FUNCTION AVGV6_MaM

!#######################################################################
  REAL FUNCTION AVGV6_a11(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_a11 = 0.
    AVGV6_a11 = AVGV6_a11 + X(  i,  j,k+1)
    AVGV6_a11 = AVGV6_a11 + X(  i,j+1,  k)
    AVGV6_a11 = AVGV6_a11 + X(  i,j+1,k+1)
    AVGV6_a11 = AVGV6_a11 + X(i+1,  j,k+1)
    AVGV6_a11 = AVGV6_a11 + X(i+1,j+1,  k)
    AVGV6_a11 = AVGV6_a11 + X(i+1,j+1,k+1)
    AVGV6_a11 = AVGV6_a11/6.

  END FUNCTION AVGV6_a11

!#######################################################################
  REAL FUNCTION AVGV6_a1M(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_a1M = 0.
    AVGV6_a1M = AVGV6_a1M + X(  i,  j,  k)
    AVGV6_a1M = AVGV6_a1M + X(  i,j+1,  k)
    AVGV6_a1M = AVGV6_a1M + X(  i,j+1,k+1)
    AVGV6_a1M = AVGV6_a1M + X(i+1,  j,  k)
    AVGV6_a1M = AVGV6_a1M + X(i+1,j+1,  k)
    AVGV6_a1M = AVGV6_a1M + X(i+1,j+1,k+1)
    AVGV6_a1M = AVGV6_a1M/6.

  END FUNCTION AVGV6_a1M

!#######################################################################
  REAL FUNCTION AVGV6_aM1(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_aM1 = 0.
    AVGV6_aM1 = AVGV6_aM1 + X(  i,  j,  k)
    AVGV6_aM1 = AVGV6_aM1 + X(  i,  j,k+1)
    AVGV6_aM1 = AVGV6_aM1 + X(  i,j+1,k+1)
    AVGV6_aM1 = AVGV6_aM1 + X(i+1,  j,  k)
    AVGV6_aM1 = AVGV6_aM1 + X(i+1,  j,k+1)
    AVGV6_aM1 = AVGV6_aM1 + X(i+1,j+1,k+1)
    AVGV6_aM1 = AVGV6_aM1/6.

  END FUNCTION AVGV6_aM1

!#######################################################################
  REAL FUNCTION AVGV6_aMM(X,imax,jmax,kmax,i,j,k)
!#######################################################################
    INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
    REAL, INTENT(IN) :: X(imax,jmax,kmax)

    AVGV6_aMM = 0.
    AVGV6_aMM = AVGV6_aMM + X(  i,  j,  k)
    AVGV6_aMM = AVGV6_aMM + X(  i,  j,k+1)
    AVGV6_aMM = AVGV6_aMM + X(  i,j+1,  k)
    AVGV6_aMM = AVGV6_aMM + X(i+1,  j,  k)
    AVGV6_aMM = AVGV6_aMM + X(i+1,  j,k+1)
    AVGV6_aMM = AVGV6_aMM + X(i+1,j+1,  k)
    AVGV6_aMM = AVGV6_aMM/6.

  END FUNCTION AVGV6_aMM

!#######################################################################
  REAL FUNCTION AVGV4_111(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_111 = 0.
  AVGV4_111 = AVGV4_111 + X(  i,j+1,k+1)
  AVGV4_111 = AVGV4_111 + X(i+1,j+1,k)
  AVGV4_111 = AVGV4_111 + X(i+1,  j,k+1)
  AVGV4_111 = AVGV4_111 + X(i+1,j+1,k+1)

  AVGV4_111 = AVGV4_111/4.

  END FUNCTION AVGV4_111

!#######################################################################
  REAL FUNCTION AVGV4_11M(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_11M = 0.
  AVGV4_11M = AVGV4_11M + X(  i,j+1,k)
  AVGV4_11M = AVGV4_11M + X(i+1,j+1,k)
  AVGV4_11M = AVGV4_11M + X(i+1,  j,k)
  AVGV4_11M = AVGV4_11M + X(i+1,j+1,k+1)

  AVGV4_11M = AVGV4_11M/4.

  END FUNCTION AVGV4_11M

!#######################################################################
  REAL FUNCTION AVGV4_1M1(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_1M1 = 0.
  AVGV4_1M1 = AVGV4_1M1 + X(  i,  j,k+1)
  AVGV4_1M1 = AVGV4_1M1 + X(i+1,  j,k)
  AVGV4_1M1 = AVGV4_1M1 + X(i+1,  j,k+1)
  AVGV4_1M1 = AVGV4_1M1 + X(i+1,j+1,k+1)

  AVGV4_1M1 = AVGV4_1M1/4.

  END FUNCTION AVGV4_1M1

!#######################################################################
  REAL FUNCTION AVGV4_1MM(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_1MM = 0.
  AVGV4_1MM = AVGV4_1MM + X(  i,  j,k)
  AVGV4_1MM = AVGV4_1MM + X(i+1,  j,k)
  AVGV4_1MM = AVGV4_1MM + X(i+1,  j,k+1)
  AVGV4_1MM = AVGV4_1MM + X(i+1,j+1,k)

  AVGV4_1MM = AVGV4_1MM/4.

  END FUNCTION AVGV4_1MM

!#######################################################################
  REAL FUNCTION AVGV4_M11(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_M11 = 0.
  AVGV4_M11 = AVGV4_M11 + X(  i,j+1,k)
  AVGV4_M11 = AVGV4_M11 + X(  i,  j,k+1)
  AVGV4_M11 = AVGV4_M11 + X(  i,j+1,k+1)
  AVGV4_M11 = AVGV4_M11 + X(i+1,j+1,k+1)

  AVGV4_M11 = AVGV4_M11/4.

  END FUNCTION AVGV4_M11

!#######################################################################
  REAL FUNCTION AVGV4_M1M(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_M1M = 0.
  AVGV4_M1M = AVGV4_M1M + X(  i,  j,  k)
  AVGV4_M1M = AVGV4_M1M + X(  i,j+1,  k)
  AVGV4_M1M = AVGV4_M1M + X(  i,j+1,k+1)
  AVGV4_M1M = AVGV4_M1M + X(i+1,j+1,  k)

  AVGV4_M1M = AVGV4_M1M/4.

  END FUNCTION AVGV4_M1M

!#######################################################################
  REAL FUNCTION AVGV4_MM1(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_MM1 = 0.
  AVGV4_MM1 = AVGV4_MM1 + X(  i,  j,  k)
  AVGV4_MM1 = AVGV4_MM1 + X(  i,  j,k+1)
  AVGV4_MM1 = AVGV4_MM1 + X(  i,j+1,k+1)
  AVGV4_MM1 = AVGV4_MM1 + X(i+1,  j,k+1)

  AVGV4_MM1 = AVGV4_MM1/4.

  END FUNCTION AVGV4_MM1

!#######################################################################
  REAL FUNCTION AVGV4_MMM(X,imax,jmax,kmax,i,j,k)
!#######################################################################

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV4_MMM = 0.
  AVGV4_MMM = AVGV4_MMM + X(  i,  j,  k)
  AVGV4_MMM = AVGV4_MMM + X(  i,j+1,  k)
  AVGV4_MMM = AVGV4_MMM + X(i+1,  j,  k)
  AVGV4_MMM = AVGV4_MMM + X(  i,  j,k+1)

  AVGV4_MMM = AVGV4_MMM/4.

  END FUNCTION AVGV4_MMM

END MODULE averaging
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
