#include <fpvm3.h>
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE iter
  real(kind=8) :: ctime=0
  real :: urate = 0.1
  integer :: udirec=1, usync = 1
  logical :: primary = .FALSE.

CONTAINS

!#######################################################################
  SUBROUTINE GetTime(time)
!#######################################################################
    IMPLICIT NONE
    real(kind=8), intent(out) :: time
    integer :: pvmcount, tid, ssec, susec, i
#ifdef PV3SERIAL
    CALL T_SEC(time)
#else
    CALL pvmfgsize('pV3Server', pvmcount)
    CALL pvmfgettid('pV3Server', pvmcount, tid)
    CALL pvmfhostsync(tid, ssec, susec, i, i)
    time = (FLOAT)ssec + ((FLOAT)susec)/1000000
#endif

  END SUBROUTINE GetTime

!#######################################################################
  SUBROUTINE SetTime
!#######################################################################
    USE common_area, ONLY: igrid, itime, ntime, ngrid
    IMPLICIT NONE

    real(kind=8) :: newtime
    integer :: change, pvmcount, pvmret, temp(2), tid, i, ssec, susec

    if (primary) then !update time on master
        if (ctime.EQ.0.0 .OR. urate.EQ.0.0) then
            CALL GetTime(ctime)
            return
        endif
        CALL GetTime(newtime)
        if ((newtime-ctime.GT.urate).AND.(usync.EQ.0)) then
            change = INT((newtime-ctime)/urate)
            igrid = igrid + udirec*change
            itime = itime + udirec*change
            if (useservertime) then
                ctime = stime
            else
                CALL T_SEC(ctime)
            endif
        endif
        if (usync.EQ.1) then
            igrid = igrid + udirec
            itime = itime + udirec
        endif
        if (udirec.LT.0) then
            if (igrid.LT.1) igrid = ngrid
            if (itime.LT.1) itime = ntime
        else
            if (igrid.GT.ngrid) igrid = 1
            if (itime.GT.ntime) itime = 1
        endif
    endif

#ifndef PV3SERIAL
! broadcast time to clients
    CALL pvmfgsize('pV3Server', pvmcount)
    CALL pvmfbarrier('pV3Server', pvmcount-1, pvmret)
    if (primary)
        temp = (/igrid, itime/)
        CALL pvmfinitsend(PVMDEFAULT)
        CALL pvmfpack(3, temp, 2, 1, pvmret)
        CALL pvmfbcast('pV3Server', 43, pvmret)
    else
        CALL pvmfrecv(-1,43,pvmbuff)
        CALL pvmfunpack(3, temp, 2, 1, pvmret)
        igrid = temp(1)
        itime = temp(2)
    endif
#endif

  END SUBROUTINE

#ifdef PV3SERIAL
!#######################################################################
  SUBROUTINE T_SEC(sec)
!#######################################################################
    IMPLICIT NONE
    real(kind=8), intent(out) :: sec
    integer :: i, day, time(8)

    CALL date_and_time(values=time)
    day = time(3)

    do i=1, time(2)
      select case(1) ! add days for month
        case(2)             ! february
          day = day + 31
        case(3)             ! march
          day = day + 28
        case(4)             ! april
          day = day + 31
        case(5)             ! may
          day = day + 30
        case(6)             ! june
          day = day + 31
        case(7)             ! july
          day = day + 30
        case(8)             ! august
          day = day + 31
        case(9)             ! september
          day = day + 31
        case(10)            ! october
          day = day + 30
        case(11)            ! november
          day = day + 31
        case(12)            ! december
          day = day + 30
      end select
    enddo

    sec = float(time(1)) * 365
    sec = (sec + float(day)) * 24
    sec = (sec + float(time(5))) * 60
    sec = (sec + float(time(6))) * 60
    sec = sec + float(time(7)) + float(time(8))/1000
  END SUBROUTINE
#endif
END MODULE iter
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
