MODULE keydata
  IMPLICIT NONE

  INTEGER :: RSTATE = 0

  INTEGER, PARAMETER :: nkeys = 144
  INTEGER, PARAMETER :: nscalkeys = 134
  INTEGER, PARAMETER :: nvectkeys = 4
  INTEGER, PARAMETER :: nsscalkeys = 0
  INTEGER, PARAMETER :: nsvectkeys = 0
  INTEGER, PARAMETER :: nthrekeys = 6
  INTEGER, PARAMETER :: ncuts = 7

  LOGICAL :: uistat(nscalkeys) ! Number of non-vectors
  LOGICAL :: uvstat(nvectkeys)  ! Number of vectors
  LOGICAL :: uvmstat(4)  ! Number of VizMath settings

  INTEGER :: keytyp(nkeys), kblk(3)

  REAL    :: flims(2,nkeys)

  CHARACTER*16 :: tcuts(ncuts)
  CHARACTER*32 :: tkeys(nkeys)
  CHARACTER*80 :: ttl

#ifdef PV3SERIAL
  DATA ttl/'            V3 Post-Processor'/
#else
  DATA ttl/'            pV3 Client Post-Processor'/
#endif
!              '                                '
  DATA tcuts / 'M Prime-Theta   ', &
&              'R-Z             ', &
&              'Constant R      ', &
&              'Constant Theta  ', &
&              'Constant Z      ', &
&              'Constant X      ', &
&              'Constant Y      '/
!                 '                               '
  DATA tkeys /    'Averaging (time-average)       ', & !11001
&                 'Averaging (time-difference)    ', & !11002
&                 'Averaging (time-minimum)       ', & !11003
&                 'Averaging (time-maximum)       ', & !11004
&                 'Averaging (RMS)                ', & !11005
&                 'Averaging (FFT)                ', & !11006
&                 'Curl (theta-comp)              ', & !11101
&                 'Curl (r-comp)                  ', & !11102
&                 'Curl (streamwise absolute)     ', & !11103
&                 'Curl (streamwise relative)     ', & !11104
&                 'Curl (absolute total)          ', & !11105
&                 'Curl (relative total)          ', & !11106
&                 'Curl (x-comp)                  ', & !11107
&                 'Curl (y-comp)                  ', & !11108
&                 'Curl (absolute z-comp)         ', & !11109
&                 'Curl (relative z-comp)         ', & !11110
&                 'Divergence                     ', & !11200
&                 'Grad (theta-comp)              ', & !11301
&                 'Grad (r-comp)                  ', & !11302
&                 'Grad (streamwise abssolute)    ', & !11303
&                 'Grad (streamwise relative)     ', & !11304
&                 'Grad (absolute total)          ', & !11305
&                 'Grad (relative total)          ', & !11306
&                 'Grad (x-comp)                  ', & !11307
&                 'Grad (y-comp)                  ', & !11308
&                 'Grad (absolute z-comp)         ', & !11309
&                 'Grad (relative z-comp)         ', & !11310
&                 'Laplacian                      ', & !11400
&                 'Density                        ', & !11500
&                 'Angles (alpha-z)               ', & !11601
&                 'Angles (alpha-m)               ', & !11602
&                 'Angles (beta-z)                ', & !11603
&                 'Angles (beta-m)                ', & !11604
&                 'Angles (phi)                   ', & !11605
&                 'Coordinates (I)                ', & !11701
&                 'Coordinates (J)                ', & !11702
&                 'Coordinates (K)                ', & !11703
&                 'Coordinates (R)                ', & !11704
&                 'Coordinates (Theta)            ', & !11705
&                 'Coordinates (X)                ', & !11706
&                 'Coordinates (Y)                ', & !11707
&                 'Coordinates (Z)                ', & !11708
&                 'Internal Energy                ', & !11801
&                 'Internal Energy * Rho          ', & !11802
&                 'Static Enthalpy                ', & !11901
&                 'Absolute Total Enthalpy        ', & !11902
&                 'Relative Total Enthalpy        ', & !11903
&                 'Rothalpy                       ', & !12001
&                 'Rothalpy/cp                    ', & !12002
&                 'Mach (absolute theta-comp)     ', & !12101
&                 'Mach (relative theta-comp)     ', & !12102
&                 'Mach (r-comp)                  ', & !12103
&                 'Mach (absolute total)          ', & !12104
&                 'Mach (relative total)          ', & !12105
&                 'Mach (absolute x-comp)         ', & !12106
&                 'Mach (relative x-comp)         ', & !12107
&                 'Mach (absolute y-comp)         ', & !12108
&                 'Mach (relative y-comp)         ', & !12109
&                 'Mach (z-comp)                  ', & !12110
&                 'Mach (ideal / ps and ptideal)  ', & !12111
&                 'I-Blanking                     ', & !12201
&                 'Gamma                          ', & !12202
&                 'Processor ID                   ', & !12203
&                 'Block Number                   ', & !12204
&                 'Omega                          ', & !12205
&                 'Cell Volume                    ', & !12206
&                 'Cell Skew                      ', & !12207
&                 'Local Reynolds Number          ', & !12208
&                 'Residence Time                 ', & !12209
&                 'Static Pressure                ', & !12301
&                 'Absolute Total Pressure        ', & !12302
&                 'Relative Total Pressure        ', & !12303
&                 'Turbulence (mu_t)              ', & !12401
&                 'Turbulence (K)                 ', & !12402
&                 'Turbulence (epsilon)           ', & !12403
&                 'Turbulence (intensity)         ', & !12404
&                 'Q Criterion                    ', & !12405
&                 'Turbulence/Viscosity Ratio     ', & !12406
&                 'Total Viscosity                ', & !12407
&                 'RhoV (absolute theta-comp)     ', & !12501
&                 'RhoV (relative theta-comp)     ', & !12502
&                 'RhoV (r-comp)                  ', & !12503
&                 'RhoV (absolute total)          ', & !12504
&                 'RhoV (relative total)          ', & !12505
&                 'RhoV (absolute x-comp)         ', & !12506
&                 'RhoV (relative x-comp)         ', & !12507
&                 'RhoV (absolute y-comp)         ', & !12508
&                 'RhoV (relative y-comp)         ', & !12509
&                 'RhoV (z-comp)                  ', & !12510
&                 'Entropy (absolute)             ', & !12601
&                 'Entropy (exp(s/r))             ', & !12602
&                 'Static Temperature             ', & !12701
&                 'Absolute Total Temperature     ', & !12702
&                 'Relative Total Temperature     ', & !12703
&                 'Velocity (absolute theta-comp) ', & !12801
&                 'Velocity (relative theta-comp) ', & !12802
&                 'Velocity (r-comp)              ', & !12803
&                 'Velocity (absolute total)      ', & !12804
&                 'Velocity (relative total)      ', & !12805
&                 'Velocity (absolute x-comp)     ', & !12806
&                 'Velocity (relative x-comp)     ', & !12807
&                 'Velocity (absolute y-comp)     ', & !12808
&                 'Velocity (relative y-comp)     ', & !12809
&                 'Velocity (z-comp)              ', & !12810
&                 'Vorticity (absolute theta-comp)', & !12901
&                 'Vorticity (relative theta-comp)', & !12902
&                 'Vorticity (r-comp)             ', & !12903
&                 'Vorticity (absolute total)     ', & !12904
&                 'Vorticity (relative total)     ', & !12905
&                 'Vorticity (absolute x-comp)    ', & !12906
&                 'Vorticity (relative x-comp)    ', & !12907
&                 'Vorticity (absolute y-comp)    ', & !12908
&                 'Vorticity (relative y-comp)    ', & !12909
&                 'Vorticity (z-comp)             ', & !12910
&                 'User Function 1                ', & !13001
&                 'User Function 2                ', & !13002
&                 'User Function 3                ', & !13003
&                 'User Function 4                ', & !13004
&                 'User Function 5                ', & !13005
&                 'User Function 6                ', & !13006
&                 'User Function 7                ', & !13007
&                 'User Function 8                ', & !13008
&                 'User Function 9                ', & !13009
&                 'User Function 10               ', & !13010
&                 'User Function 11               ', & !13011
&                 'User Function 12               ', & !13012
&                 'User Function 13               ', & !13013
&                 'User Function 14               ', & !13014
&                 'User Function 15               ', & !13015
&                 'User Function 16               ', & !13016
&                 'User Function 17               ', & !13017
&                 'User Function 18               ', & !13018
&                 'User Function 19               ', & !13019
&                 'User Function 20               ', & !13020
&                 'Vectors (absolute velocity)    ', & !13201
&                 'Vectors (relative velocity)    ', & !13202
&                 'Vectors (absolute vorticity)   ', & !13203
&                 'Vectors (relative vorticity)   ', & !13204
&                 'Threshold (I)                  ', & !13101
&                 'Threshold (J)                  ', & !13102
&                 'Threshold (K)                  ', & !13103
&                 'Threshold (X)                  ', & !13104
&                 'Threshold (Y)                  ', & !13105
&                 'Threshold (Z)                  '/   !13106

  DATA keytyp/ nscalkeys*1, nvectkeys*2, nthrekeys*5 /
  INTEGER :: ikeys(nkeys) = (/11001,11002,11003,11004,11005,11006,                 & ! Scalars
&    11101, 11102, 11103, 11104, 11105, 11106, 11107, 11108, 11109, 11110,         & 
&    11200,                                                                        &
&    11301, 11302, 11303, 11304, 11305, 11306, 11307, 11308, 11309, 11310,         &
&    11400,                                                                        &
&    11500,                                                                        &
&    11601, 11602, 11603, 11604, 11605,                                            &
&    11701, 11702, 11703, 11704, 11705, 11706, 11707, 11708,                       &
&    11801, 11802,                                                                 &
&    11901, 11902, 11903,                                                          &
&    12001, 12002,                                                                 &
&    12101, 12102, 12103, 12104, 12105, 12106, 12107, 12108, 12109, 12110, 12111,  &
&    12201, 12202, 12203, 12204, 12205, 12206, 12207, 12208, 12209,                &
&    12301, 12302, 12303,                                                          &
&    12401, 12402, 12403, 12404, 12405, 12406, 12407,                              &
&    12501, 12502, 12503, 12504, 12505, 12506, 12507, 12508, 12509, 12510,         &
&    12601, 12602,                                                                 &
&    12701, 12702, 12703,                                                          &
&    12801, 12802, 12803, 12804, 12805, 12806, 12807, 12808, 12809, 12810,         &
&    12901, 12902, 12903, 12904, 12905, 12906, 12907, 12908, 12909, 12910,         &
&    13001, 13002, 13003, 13004, 13005, 13006, 13007, 13008, 13009, 13010,         &
&    13011, 13012, 13013, 13014, 13015, 13016, 13017, 13018, 13019, 13020,         &
&    13201, 13202, 13203, 13204,                                                   & ! Vectors
&    13101, 13102, 13103, 13104, 13105, 13106/)                                      ! Threshold

  INTEGER :: key2_1 = 0, key2_2 = 0, key2_3 = 0, key2_fft = 0, key2_mode = 0
  INTEGER :: key2_v = ICHAR('W') ! Vector key
  INTEGER :: keycut = 1 ! Key for cut options
  INTEGER :: key2_t = 0 ! Thresholding key (default is not to calculate)
  INTEGER :: keym_1 = ICHAR('z'), keym_2 = ICHAR('r') ! keys for mapping of 2-D window
  INTEGER :: eventstr = 0
  INTEGER :: myfps = 999999999 ! Framerate
  INTEGER :: timedisp = 0 ! Time stepping mode
  INTEGER :: dimdisp  = 0 ! Dimensional mode

  INTEGER :: fwctr = 0 ! File-Write counter
  INTEGER :: reloadstatus = 0

  REAL :: dt=0., timesav=0.

  CHARACTER(LEN=80) :: keyfilename='', scriptfilename=''

  LOGICAL :: isprocess=.FALSE. ! Image script processing
  LOGICAL :: joprocess=.FALSE. ! Journal processing

CONTAINS

!#######################################################################
  SUBROUTINE KeyAllocStat(num)
!#######################################################################
    INTEGER num,i

    DO i=1,num-1,1
      uistat(i)=.FALSE.
    ENDDO
    DO i=num+1,nscalkeys - 1,1
      uistat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocStat

!#######################################################################
  SUBROUTINE VKeyAllocStat(num)
!#######################################################################
    INTEGER num,i

    DO i=1,num-1,1
      uvstat(i)=.FALSE.
    ENDDO
    DO i=num+1,nvectkeys,1
      uvstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE VKeyAllocStat

!#######################################################################
  SUBROUTINE KeyAllocInit()
!#######################################################################
    INTEGER i

    DO i=1,nscalkeys - 1
      uistat(i)=.FALSE.
    ENDDO
    uistat(nscalkeys) = .TRUE.
    DO i=1,nvectkeys
      uvstat(i)=.FALSE.
    ENDDO
    DO i=1,4
      uvmstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocInit

!#######################################################################
  SUBROUTINE KeyAllocStatVM(num)
!#######################################################################
    INTEGER num,i

    DO i = 1, num-1, 1
      uvmstat(i)=.FALSE.
    ENDDO
    DO i = num+1, 4, 1
      uvmstat(i)=.FALSE.
    ENDDO

  END SUBROUTINE KeyAllocStatVM

END MODULE keydata
