#include "parameters.h"
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE readbcs
! This module exists for the sole purpose of decreasing the number of
! modules which need to be included in other places in the code.  As
! solvers are added, there will be minimal changes in other code for
! the purpose of accessing boundary condition data.  If you don't like
! it, change it -- but don't give it back to me different; I may just
! have to pop ya.... :)
  USE general
  USE common_area
  USE solver_turbo
  USE solver_uns
  USE MOD_ReadPegFile, ONLY : ReadPegFile, nbc
  USE MOD_P3D_ParallelIO, ONLY : ReadDecompBCFile
  USE MOD_BCDefs

CONTAINS

!#######################################################################
  SUBROUTINE Get_Solver_Input()
!#######################################################################
    IMPLICIT NONE

    SELECT CASE(runmode)
      CASE(11,1,-1)
        CALL read_ref_nml()
        CALL read_bc()
        IF(bladerow.NE.0)CALL read_rowinput
        IF(runmode.EQ.11)THEN
          IF(avusbcfilename.NE.'')THEN
            CALL read_tacoma_p3d_bc(avusbcfilename)
          ELSE
            bcavail = .FALSE.
          ENDIF
        ENDIF

        IF(.NOT. trep)THEN
          mirror = 0
          dtheta = 0.
        ELSE
          IF(ksym.GT.0)THEN
            dtheta = 2.0* PI /FLOAT(ksym)
            mirror = -(ksym-1)
          ELSE
            dtheta = 2.0* PI /FLOAT(pgr(igrid)%nbld)
            IF(mirror.EQ.0 .AND. pgr(igrid)%nbld.GT.1)THEN
              SELECT CASE(ntrep)
                CASE(-4)
                  mirror = -(FLOOR(pgr(igrid)%nbld/4.)-1)
                CASE(-2)
                  mirror = -(FLOOR(pgr(igrid)%nbld/2.)-1)
                CASE(-1)
                  mirror = -(pgr(igrid)%nbld-1)
                CASE DEFAULT
                  mirror = -(ntrep-1) ! For now, just do the specified number
              END SELECT
            ENDIF
          ENDIF
        ENDIF

        IF(ffmt.EQ.1 .AND. runmode.NE.11)bcavail=.TRUE.
        PRINT*,' My Omega = ',omega

!=======================================================================
! TURBO ast files
! I need to read other namelists as well so I can get symmetry_factor
! for the case when we want replication...
      CASE(2)
        IF(ffmt.EQ.3)THEN
          bcavail = .FALSE.
          RETURN
        ENDIF
        CALL read_extract
        CALL read_ref_nml
        CALL read_rowinput
        CALL read_bc
        CALL read_dmap

        IF(nblds.NE.-1) pgr(igrid)%nbld = nblds
        WRITE(*,*) ' Turbo simulation has ',nblds,' blades'
        dtheta = 2.0* PI /FLOAT(pgr(igrid)%nbld)
        IF(.NOT. trep)THEN
          mirror = 0
        ELSE
          IF(ksym.GT.0)THEN
            mirror = -(ksym-1)
            dtheta = 2.0* PI /FLOAT(ksym)
          ELSE
            mirror = -(nblds-1)
            dtheta = 2.0* PI /FLOAT(nblds)
          ENDIF
        ENDIF

!=======================================================================
! FDLF90 boundary condition reads
      CASE(6)

         IF(ntrep.NE.-1)mirror = -ntrep

         CALL InitBCNames
         CALL ReadPegFile (pgr(igrid)%nblks, pgr(igrid)%li, pgr(igrid)%lj, pgr(igrid)%lk, avusbcfilename)
         IF (SUM(nbc) .eq. 0) CALL ReadDecompBCFile (cid,avusbcfilename) !Try reading a decomposed bc file
         IF (SUM(nbc) .gt. 0) bcavail = .true.

    END SELECT
  END SUBROUTINE Get_Solver_Input

END MODULE readbcs
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
