#include "parameters.h"
#define FLOW pfw(n)%
#define GRID pgr(c)%
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE readflow
  IMPLICIT NONE

CONTAINS

!#######################################################################
  SUBROUTINE Init_Pfw()
!#######################################################################
    USE common_area, ONLY: cur, ntime, pfw

    INTEGER n
    
    DO n=1,ntime
! Current value pointer
      NULLIFY(cur(n)%v)

! Vstage variables
      NULLIFY(FLOW p)
      NULLIFY(FLOW r)
      NULLIFY(FLOW vz)
      NULLIFY(FLOW vr)
      NULLIFY(FLOW vt)
      NULLIFY(FLOW e)
      NULLIFY(FLOW mut)
      NULLIFY(FLOW k)
      NULLIFY(FLOW eps)

! Real gas
      NULLIFY(FLOW g)
      NULLIFY(FLOW rgas)
      NULLIFY(FLOW reyn)

! Pressures
      NULLIFY(FLOW ptrel)
      NULLIFY(FLOW ptabs)

! Angles
      NULLIFY(FLOW alphaz)
      NULLIFY(FLOW alpham)
      NULLIFY(FLOW betaz)
      NULLIFY(FLOW betam)
      NULLIFY(FLOW phi)

! Mach numbers
      NULLIFY(FLOW machthabs)
      NULLIFY(FLOW machthrel)
      NULLIFY(FLOW machr)
      NULLIFY(FLOW machtabs)
      NULLIFY(FLOW machtrel)
      NULLIFY(FLOW machxabs)
      NULLIFY(FLOW machxrel)
      NULLIFY(FLOW machyabs)
      NULLIFY(FLOW machyrel)
      NULLIFY(FLOW machz)
      NULLIFY(FLOW machideal)

! Rho-V
      NULLIFY(FLOW rhovthabs)
      NULLIFY(FLOW rhovthrel)
      NULLIFY(FLOW rhovr)
      NULLIFY(FLOW rhovtabs)
      NULLIFY(FLOW rhovtrel)
      NULLIFY(FLOW rhovxabs)
      NULLIFY(FLOW rhovxrel)
      NULLIFY(FLOW rhovyabs)
      NULLIFY(FLOW rhovyrel)
      NULLIFY(FLOW rhovz)

! Enthalpy
      NULLIFY(FLOW h)
      NULLIFY(FLOW htabs)
      NULLIFY(FLOW htrel)

! Rothalpy
      NULLIFY(FLOW roth)
      NULLIFY(FLOW rothcp)

! Entropy
      NULLIFY(FLOW sabs)
      NULLIFY(FLOW srat)

! Temperature
      NULLIFY(FLOW temp)
      NULLIFY(FLOW temptabs)
      NULLIFY(FLOW temptrel)

! Velocities
      NULLIFY(FLOW vx)
      NULLIFY(FLOW vxabs)
      NULLIFY(FLOW vy)
      NULLIFY(FLOW vyabs)
      NULLIFY(FLOW vtabs)
      NULLIFY(FLOW vtrel)
      NULLIFY(FLOW vthrel)

! VizMath
      NULLIFY(FLOW vmath)
      NULLIFY(FLOW vmaths)
      NULLIFY(FLOW vmathr)
      NULLIFY(FLOW vmatht)
      NULLIFY(FLOW vmathsabs)
      NULLIFY(FLOW vmathsrel)
      NULLIFY(FLOW vmathtabs)
      NULLIFY(FLOW vmathtrel)
      NULLIFY(FLOW vmathzabs)

! Vorticity
      NULLIFY(FLOW vortr)
      NULLIFY(FLOW vortt)
      NULLIFY(FLOW vortz)
      NULLIFY(FLOW vortsabs)
      NULLIFY(FLOW vortsrel)
      NULLIFY(FLOW vorttabs)
      NULLIFY(FLOW vorttrel)
      NULLIFY(FLOW vortx)
      NULLIFY(FLOW vorty)
      NULLIFY(FLOW vortzabs)

! Turbulence
      NULLIFY(FLOW turbint)
      NULLIFY(FLOW turbrat)
      NULLIFY(FLOW turbtvisc)
      NULLIFY(FLOW qcrit)
      NULLIFY(FLOW yplus)

! Time-average
      NULLIFY(FLOW tmavg)
      NULLIFY(FLOW tmdif)
      NULLIFY(FLOW tmmin)
      NULLIFY(FLOW tmmax)
      NULLIFY(FLOW tmrms)
      NULLIFY(FLOW tmfft)

! Plot3D functions
      NULLIFY(FLOW funk)

    ENDDO

  END SUBROUTINE Init_Pfw

!#######################################################################
  SUBROUTINE Get_Flow()
!#######################################################################
    USE general, ONLY: int_to_string, Prompt_Get, Check_File_Exist, &
                       Fatal_Error
    USE common_area, ONLY: runmode, ntime, flowarr, ffmt, cid, pfw, cur, &
                           gammabase,gamtable,tempgam, omega, dimen, &
                           set_ref_quantities, ntmax, nffile

    CHARACTER(LEN=NSTRLEN) :: fnam,lstfile
    CHARACTER :: opt

    INTEGER i, j, rntime

    LOGICAL ios

    CALL set_ref_quantities(0)
!=======================================================================
! Get input for interactive case
    IF(runmode.EQ.-1)THEN
1000  CONTINUE
      WRITE(*,*) ' Single(s) or Multiple(m) Files?'
      READ(*,*,ERR=1000) opt
      IF(opt.EQ.'s')THEN
        ntime=1
        CALL Prompt_Get(fnam,' Enter flow file (xxx for none) ==>')
        SELECT CASE(ffmt)
          CASE(1,2)
            ntime=1
          CASE(3)
            CALL Rd_P3D_Flw_Ray(fnam,0,1,ntime)
        END SELECT
      ELSE
        WRITE(*,*) ' Enter list file ==>'
        READ*,lstfile
        CALL Check_File_Exist(lstfile,ios)
        CALL Fatal_Error(ios,'*** List file does not exist ***')
        OPEN(UNIT=21,FILE=lstfile,FORM='formatted')
        ntime=0
        nffile=0
        DO WHILE(1)
          READ(21,*,END=9000,ERR=9000) fnam
          nffile = nffile+1
          SELECT CASE(ffmt)
            CASE(1,2)
              ntime = ntime+1
            CASE(3)
              CALL Rd_P3D_Flw_Ray(fnam,0,1,rntime)
              ntime = ntime+rntime
          END SELECT
        ENDDO
        IF(ffmt.EQ.3) THEN
          IF(ntime.GT.ntmax) ntime = ntmax
        ENDIF
9000    CONTINUE
        CLOSE(21)
      ENDIF
    ENDIF

!=======================================================================
! Pre-scan flow files for time step count if nffile != ntime
    SELECT CASE(runmode)
      CASE(1,3,4,5,6,MODETACOMA)
        ntime = nffile
      CASE(2)
        SELECT CASE(ffmt)
          CASE(1,2)
            ntime = nffile
          CASE(3)
            ntime = 0
            DO i=1,nffile
              WRITE(fnam,*) 'ast.b' // TRIM(ADJUSTL(int_to_string(cid))) // '.' // TRIM(ADJUSTL(flowarr(i)))
              CALL Rd_P3D_Flw_Ray(fnam,0,1,rntime)
              ntime = ntime+rntime
            ENDDO
        END SELECT
    END SELECT
    IF(ntmax.LT.ntime) ntime=ntmax

    WRITE(*,*) ' ntime=',ntime

!=======================================================================
! Allocate flow storage
    ALLOCATE(pfw(ntime)) ! Allocate the flow storage
    ALLOCATE(cur(ntime)) ! Allocate the "current variable" pointers
    CALL Init_Pfw()

!=======================================================================
! Read flow files
    SELECT CASE(runmode)
!-----------------------------------------------------------------------
! Turbo files that have been merged or Tacoma files in p3d
      CASE(1,MODETACOMA)
        DO i=1,nffile
          SELECT CASE(ffmt)
            CASE(1)
              CALL Rd_Vstg_Flw(flowarr(i),i,1)
            CASE(2)
              CALL Rd_P3DT_Flw(flowarr(i),i,1)
            CASE(3)
              CALL Rd_P3D_Flw_Ray(flowarr(i),i,1,rntime)
          END SELECT
        ENDDO
!-----------------------------------------------------------------------
! Turbo ast files
      CASE(2)
        j = 1
        DO i=1,nffile
          WRITE(fnam,*) 'ast.b' // TRIM(ADJUSTL(int_to_string(cid))) // '.' // TRIM(ADJUSTL(flowarr(i)))
          SELECT CASE(ffmt)
            CASE(1)
              CALL Rd_Vstg_Flw(fnam,j,1)
              j = j+1
            CASE(2)
              CALL Rd_P3DT_Flw(fnam,j,1)
              j = j+1
            CASE(3)
              CALL Rd_P3D_Flw_Ray(fnam,j,1,rntime)
              j = j+rntime
          END SELECT
        ENDDO

!-----------------------------------------------------------------------
! P3D files
      CASE(3)
        DO i=1,ntime
          CALL Rd_P3D_Flow(flowarr(i),i,1)
        ENDDO

!-----------------------------------------------------------------------
! General FV files
      CASE(4)
        DO i=1,ntime
          SELECT CASE(ffmt)
            CASE(1) ! Formatted split
            CASE(2) ! Unformatted split
            CASE(3) ! Formatted combined
            CASE(4) ! Unformatted combined
              CALL Rd_FV_Unf_Comb(flowarr(i),i,i)
          END SELECT
        ENDDO

!-----------------------------------------------------------------------
! AVUS files
      CASE(5)
        DO i=1,ntime
          CALL Rd_AVUS_Flow(flowarr(i),i)
        ENDDO

!-----------------------------------------------------------------------
! FDL files
      CASE(6)
        IF(ntime.EQ.1 .AND. TRIM(ADJUSTL(flowarr(1))).EQ.'xxx')THEN
          CALL Rd_FDL_Flow_Dum(1,1)
        ELSE
          DO i=1,ntime
            CALL Rd_FDL_Flow(flowarr(i),i,1)
          ENDDO
        ENDIF

!-----------------------------------------------------------------------
! Default interactive mode for TURBO
! 1) We need to get the gamma interpolation data and omega
! 2) Get the flow file names from stdin or list file
! 3) Read the flow files
      CASE(-1)
        IF(fnam.NE.'xxx')THEN
1001      CONTINUE
          WRITE(*,*) ' Enter gamref, g1, g2, t1, t2:'
          READ(*,*,ERR=1001) gammabase,gamtable(1),gamtable(2), &
                           tempgam(1),tempgam(2)
1002      CONTINUE
          WRITE(*,*) ' Enter normalized omega:'
          READ(*,*,ERR=1002) omega
        ENDIF

! The default mode does not allow rotation because it's just one more
! input I have to wade through.  Change?  I think not.
        IF(opt.EQ.'s')THEN
          SELECT CASE(ffmt)
            CASE(1)
              CALL Rd_Vstg_Flw(fnam,1,1)
            CASE(2)
              CALL Rd_P3DT_Flw(fnam,1,1)
            CASE(3)
              CALL Rd_P3D_Flw_Ray(fnam,1,1,rntime)
          END SELECT
        ELSE
          OPEN(UNIT=21,FILE=lstfile,FORM='formatted')
          REWIND(21)
          j=1
          DO i=1,nffile
            READ(21,*) fnam
            SELECT CASE(ffmt)
              CASE(1)
                CALL Rd_Vstg_Flw(fnam,j,1)
                j = j+1
              CASE(2)
                CALL Rd_P3DT_Flw(fnam,j,1)
                j = j+1
              CASE(3)
                CALL Rd_P3D_Flw_Ray(fnam,j,1,rntime)
                j = j+rntime
            END SELECT
          ENDDO
          CLOSE(21)
        ENDIF
      CASE DEFAULT
        WRITE(*,*) 'No Flow Field Available for Runmode =', runmode
        STOP
    END SELECT

    CALL set_ref_quantities(dimen)

  END SUBROUTINE Get_Flow

!#######################################################################
  SUBROUTINE ReloadFlow(newflow)
!#######################################################################
    USE general, ONLY: int_to_string
    USE common_area, ONLY: runmode, pfw, ntime, dimen, cid, ffmt, &
                           set_ref_quantities, ClearAllFlow
    USE keydata, ONLY: RSTATE

    CHARACTER(LEN=NSTRLEN) :: newflow, fnam

    CALL ClearAllFlow

    ntime = 1
    ALLOCATE(pfw(ntime))
    CALL Init_Pfw

    SELECT CASE(runmode)
      CASE(-1,1,MODETACOMA)
        SELECT CASE(ffmt)
          CASE(1)
            CALL Rd_Vstg_Flw(newflow,1,1)
          CASE(2)
            CALL Rd_P3DT_Flw(newflow,1,1)
        END SELECT

      CASE(2)
        WRITE(fnam,*) 'ast.b' // TRIM(ADJUSTL(int_to_string(cid))) // '.q' // TRIM(ADJUSTL(newflow))
        SELECT CASE(ffmt)
          CASE(1)
            CALL Rd_Vstg_Flw(fnam,1,1)
          CASE(2)
            CALL Rd_P3DT_Flw(fnam,1,1)
        END SELECT

      CASE(3)
        CALL Rd_P3D_Flow(newflow,1,1)

      CASE(4)
        SELECT CASE(ffmt)
          CASE(1) ! Formatted split
          CASE(2) ! Unformatted split
          CASE(3) ! Formatted combined
          CASE(4) ! Unformatted combined
            CALL Rd_FV_Unf_Comb(newflow,1,1)
        END SELECT

      CASE(5)
        CALL Rd_AVUS_Flow(newflow,1)

      CASE(6)
        CALL Rd_FDL_Flow(newflow,1,1)

    END SELECT

    RSTATE = RSTATE_WAITING

    CALL set_ref_quantities(dimen)

  END SUBROUTINE ReloadFlow

!#######################################################################
  SUBROUTINE Rd_P3DT_Flw(fnam,n,c)
!#######################################################################
    USE general, ONLY: Fatal_Error, Check_File_Exist
    USE common_area, ONLY: pfw, pgr, igrid, omega, runmode, gammabase, &
                           cpref_n, nclients
    USE mod_clear, ONLY: ClearVar
    USE MOD_P3D_ParallelIO, ONLY : xm1, alfa, re, tau
    USE MOD_P3DGrid, ONLY: p3dxm1, p3dalfa, p3dre, p3dtau, SQVar, DQVar

    INTEGER,INTENT(IN) :: n, c
    CHARACTER(LEN=NSTRLEN),INTENT(IN) :: fnam
    LOGICAL :: ios
    INTEGER :: i, j, k, i2, j2, k2, id, jd, kd, flmax, mvmax, m, s
    INTEGER :: dumint

    CALL Rd_P3D_Flow(fnam, n, c)

    SELECT CASE(runmode)
      CASE(MODETACOMA)
        SELECT CASE(nclients)
          CASE(1)
            gammabase = p3dxm1
            cpref_n = p3dalfa
            omega = p3dre
            IF(ALLOCATED(SQVar))DEALLOCATE(SQVar)
            IF(ALLOCATED(DQVar))DEALLOCATE(DQVar)
          CASE DEFAULT
            gammabase = xm1
            cpref_n = alfa
            omega = re
        END SELECT
    END SELECT

    mvmax = pgr(igrid)%tpts
    ALLOCATE(FLOW vr(mvmax),FLOW vt(mvmax),FLOW k(mvmax))
! Momentum to velocities and cylindrical
    DO m = 1, mvmax
      FLOW k(m) = 0.
      FLOW vr(m) = FLOW vx(m)*COS(GRID t0(m)) + FLOW vy(m)*SIN(GRID t0(m))
      FLOW vt(m) = -FLOW vx(m)*SIN(GRID t0(m)) + FLOW vy(m)*COS(GRID t0(m))
      FLOW e(m)  = FLOW e(m) - &
                    (FLOW vr(m)**2 + FLOW vt(m)**2 + FLOW vz(m)**2)/2.
    ENDDO
    CALL ClearVar(FLOW vx)
    CALL ClearVar(FLOW vy)
    WRITE(*,*) '  Flow file converted to cylindrical successfully!'

  END SUBROUTINE Rd_P3DT_Flw

!#######################################################################
  SUBROUTINE Rd_P3DT_Dum(n,c)
!#######################################################################
    USE common_area, ONLY: pgr, pfw

    INTEGER,INTENT(IN) :: n,c
    INTEGER :: mvmax, i

    mvmax = 0
    DO i = 1, GRID nblks
      mvmax = GRID li(i) * GRID lj(i) * GRID lk(i) + mvmax
    ENDDO
    ALLOCATE(FLOW r(mvmax), FLOW vx(mvmax))
    DO i = 1, mvmax
      FLOW r(i) = 1.
      FLOW vx(i) = 0.
    ENDDO
    FLOW vy => FLOW vx
    FLOW vz => FLOW vx
    FLOW vr => FLOW vx
    FLOW vt => FLOW vx
    FLOW e => FLOW r
    FLOW k => FLOW vx
    FLOW eps => FLOW vx
    FLOW mut => FLOW vx

    RETURN
  END SUBROUTINE Rd_P3DT_Dum

!#######################################################################
  SUBROUTINE Rd_Vstg_Flw(fnam,n,c)
!#######################################################################
! fnam is the name of the flow file
! n is the number in the list of flow files
! c is the number of the associated grid (this will be 1 until we
!   implement structure unsteady grids)
    USE common_area, ONLY: pgr, pfw
    USE general, ONLY: Fatal_Error, Check_File_Exist
    USE averaging

    LOGICAL :: ios
    CHARACTER(LEN=NSTRLEN) :: fnam
    INTEGER :: i,j,k,i2,j2,k2,id,jd,kd,flmax,mvmax,n,c
    REAL,POINTER,DIMENSION(:) :: pcell,rcell,vzcell,vrcell,vtcell, &
                                 recell,dmutcell,akcell,epscell
!    REAL,ALLOCATABLE,TARGET,DIMENSION(:) :: p,r,vz,vr,vt,re,dmut,ak,eps

    IF(fnam.EQ.'xxx')THEN
      CALL Rd_Vstg_Dum(n,c)
      RETURN
    ENDIF

    WRITE(*,*) ' Reading flow file ',n,'=',fnam
    CALL Check_File_Exist(fnam,ios)
    CALL Fatal_Error(ios,'*** Flow file does not exist ***')

!=======================================================================
! Allocate temporary arrays and read the flow
    id = GRID li(1) 
    jd = GRID lj(1)
    kd = GRID lk(1)
    mvmax = GRID tpts
    i2 = id + 1
    j2 = jd + 1
    k2 = kd + 1
    flmax = i2*j2*k2
    ALLOCATE(pcell(flmax),rcell(flmax),vzcell(flmax),vrcell(flmax), &
             vtcell(flmax),recell(flmax),dmutcell(flmax),akcell(flmax), &
             epscell(flmax))

    OPEN(UNIT=QUNIT,FILE=fnam,FORM='unformatted')
    READ(QUNIT) FLOW ncyc, FLOW tyme, &
       ((( pcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( rcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( vzcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( vrcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( vtcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( recell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( dmutcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( akcell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2), &
       ((( epscell(i+(j-1)*i2+(k-1)*i2*j2),k=1,k2),j=1,j2),i=1,i2)
    CLOSE(QUNIT)
    WRITE(*,*) '  The flow file was read successfully time=',FLOW tyme

!=======================================================================
! Averaging flow node values to vertex
    ALLOCATE(FLOW p(mvmax), FLOW r(mvmax),FLOW vz(mvmax),FLOW vr(mvmax), &
             FLOW vt(mvmax),FLOW e(mvmax),FLOW mut(mvmax),                 &
             FLOW k(mvmax), FLOW eps(mvmax))
! Center of flowfield
    DO i=1,id
      DO j=1,jd
        DO k=1,kd
          FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV(pcell,i2,j2,k2,i,j,k)
          FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV(rcell,i2,j2,k2,i,j,k)
          FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV(vzcell,i2,j2,k2,i,j,k)
          FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV(vrcell,i2,j2,k2,i,j,k)
          FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV(vtcell,i2,j2,k2,i,j,k)
          FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV(recell,i2,j2,k2,i,j,k)
          FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV(dmutcell,i2,j2,k2,i,j,k)
          FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV(akcell,i2,j2,k2,i,j,k)
          FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV(epscell,i2,j2,k2,i,j,k)
        ENDDO
      ENDDO
    ENDDO
! **** Block edges ****
!    K-Edges
    DO k = 2, kd-1
      i = 1
      j = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_11a(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_11a(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_11a(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_11a(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_11a(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_11a(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_11a(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_11a(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_11a(epscell,i2,j2,k2,i,j,k)
      i = 1
      j = jd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1Ma(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1Ma(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1Ma(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1Ma(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1Ma(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1Ma(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1Ma(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1Ma(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1Ma(epscell,i2,j2,k2,i,j,k)
      i = id
      j = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_M1a(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_M1a(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_M1a(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_M1a(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_M1a(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_M1a(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_M1a(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_M1a(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_M1a(epscell,i2,j2,k2,i,j,k)
      i = id
      j = jd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MMa(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MMa(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MMa(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MMa(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MMa(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MMa(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_MMa(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MMa(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_MMa(epscell,i2,j2,k2,i,j,k)
    ENDDO
!    J-Edges
    DO j = 2, jd-1
      i = 1
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1a1(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1a1(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1a1(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1a1(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1a1(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1a1(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1a1(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1a1(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1a1(epscell,i2,j2,k2,i,j,k)
      i = 1
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1aM(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1aM(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1aM(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1aM(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_1aM(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1aM(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1aM(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_1aM(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_1aM(epscell,i2,j2,k2,i,j,k)
      i = id
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_Ma1(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_Ma1(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_Ma1(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_Ma1(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_Ma1(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_Ma1(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_Ma1(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_Ma1(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_Ma1(epscell,i2,j2,k2,i,j,k)
      i = id
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MaM(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MaM(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MaM(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MaM(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_MaM(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MaM(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_MaM(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_MaM(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_MaM(epscell,i2,j2,k2,i,j,k)
    ENDDO
!    I-Edges
    DO i = 2, id-1
      j = 1
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a11(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a11(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a11(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a11(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a11(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a11(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_a11(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a11(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_a11(epscell,i2,j2,k2,i,j,k)
      j = 1
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a1M(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a1M(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a1M(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a1M(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_a1M(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a1M(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_a1M(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_a1M(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_a1M(epscell,i2,j2,k2,i,j,k)
      j = jd
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aM1(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aM1(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aM1(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aM1(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aM1(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aM1(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_aM1(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aM1(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_aM1(epscell,i2,j2,k2,i,j,k)
      j = jd
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aMM(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aMM(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aMM(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aMM(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV6_aMM(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aMM(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV6_aMM(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV6_aMM(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV6_aMM(epscell,i2,j2,k2,i,j,k)
    ENDDO

! **** Corners ****
      i = 1
      j = 1
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_111(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_111(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_111(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_111(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_111(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_111(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_111(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_111(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_111(epscell,i2,j2,k2,i,j,k)
      i = 1
      j = 1
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_11M(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_11M(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_11M(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_11M(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_11M(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_11M(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_11M(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_11M(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_11M(epscell,i2,j2,k2,i,j,k)
      i = 1
      j = jd
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1M1(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1M1(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1M1(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1M1(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1M1(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1M1(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_1M1(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1M1(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_1M1(epscell,i2,j2,k2,i,j,k)
      i = 1
      j = jd
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1MM(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1MM(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1MM(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1MM(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_1MM(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1MM(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_1MM(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_1MM(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_1MM(epscell,i2,j2,k2,i,j,k)
      i = id
      j = 1
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M11(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M11(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M11(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M11(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M11(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M11(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_M11(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M11(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_M11(epscell,i2,j2,k2,i,j,k)
      i = id
      j = 1
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M1M(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M1M(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M1M(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M1M(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_M1M(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M1M(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_M1M(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_M1M(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_M1M(epscell,i2,j2,k2,i,j,k)
      i = id
      j = jd
      k = 1
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MM1(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MM1(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MM1(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MM1(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MM1(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MM1(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_MM1(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MM1(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_MM1(epscell,i2,j2,k2,i,j,k)
      i = id
      j = jd
      k = kd
      FLOW p(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MMM(pcell,i2,j2,k2,i,j,k)
      FLOW r(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MMM(rcell,i2,j2,k2,i,j,k)
      FLOW vz(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MMM(vzcell,i2,j2,k2,i,j,k)
      FLOW vr(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MMM(vrcell,i2,j2,k2,i,j,k)
      FLOW vt(i+(j-1)*id+(k-1)*id*jd)  = AVGV4_MMM(vtcell,i2,j2,k2,i,j,k)
      FLOW e(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MMM(recell,i2,j2,k2,i,j,k)
      FLOW mut(i+(j-1)*id+(k-1)*id*jd) = AVGV4_MMM(dmutcell,i2,j2,k2,i,j,k)
      FLOW k(i+(j-1)*id+(k-1)*id*jd)   = AVGV4_MMM(akcell,i2,j2,k2,i,j,k)
      FLOW eps(i+(j-1)*id+(k-1)*id*jd) = AVGV4_MMM(epscell,i2,j2,k2,i,j,k)


    DEALLOCATE(pcell,rcell,vzcell,vrcell,vtcell,recell,dmutcell,akcell,epscell)
    NULLIFY(pcell,rcell,vzcell,vrcell,vtcell,recell,dmutcell,akcell,epscell)
    WRITE(*,*) '  Flow successfully converted to vertices'

!=======================================================================
! Convert from conservation to primitive
    DO i=1,mvmax
      FLOW vz(i) = FLOW vz(i)/FLOW r(i)
      FLOW vr(i) = FLOW vr(i)/FLOW r(i)
      FLOW vt(i) = FLOW vt(i)/FLOW r(i)
      FLOW e(i)  = FLOW e(i)/FLOW r(i)
    ENDDO

  END SUBROUTINE Rd_Vstg_Flw

!#######################################################################
  SUBROUTINE Rd_Vstg_Dum(n,c)
!#######################################################################
    USE common_area, ONLY: pgr, pfw
    INTEGER,INTENT(IN) :: n,c
    INTEGER i, mvmax   

    mvmax = GRID li(1) * GRID lj(1) * GRID lk(1)
    ALLOCATE(FLOW p(mvmax))
    DO i = 1, mvmax
      FLOW p(i) = 1.
    ENDDO
    FLOW r =>   FLOW p
    FLOW vz =>  FLOW p
    FLOW vr =>  FLOW p
    FLOW vt =>  FLOW p
    FLOW e =>   FLOW p
    FLOW k =>   FLOW p
    FLOW eps => FLOW p
    FLOW mut => FLOW p

  END SUBROUTINE Rd_Vstg_Dum

!#######################################################################
  SUBROUTINE Rd_FV_Unf_Comb(fnam,n,c)
!#######################################################################
    USE general, ONLY: Fatal_Error
    USE common_area, ONLY: pgr, pfw, ngrid
    USE solver_uns
    USE readgrid
    IMPLICIT NONE

    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
    INTEGER, INTENT(IN) :: n, c

    CHARACTER*80 :: tmptxt
    INTEGER :: majver, minver
    INTEGER :: i, j, k, cnt, npts, hdr

    INTEGER nbcs, nvars, idum

    REAL, POINTER, DIMENSION(:,:) :: dw

    OPEN(UNIT=QUNIT,FILE=fnam,FORM='unformatted')
    READ(QUNIT) fv_magic
    SELECT CASE(fv_magic)
      CASE(66051)
        READ(QUNIT) tmptxt
        READ(QUNIT) majver, minver
        READ(QUNIT) FLOW tyme, FLOW fsmach, FLOW re, FLOW alpha
        READ(QUNIT) ngrid
        IF(ngrid.GT.1 .AND. n.EQ.1)THEN
          WRITE(*,*) ' This ain`t gonna work y`all...'
          CALL Fatal_Error(.TRUE., 'ngrid GT 1 in fv')
          DEALLOCATE(pgr)
          ALLOCATE(pgr(ngrid))
          CALL Init_Pgr()
          ALLOCATE(fvd(ngrid))
        ELSEIF(n.EQ.1)THEN
          ALLOCATE(fvd(1))
        ENDIF

        READ(QUNIT) nbcs
        fvd(n)%nbcs = nbcs
        ALLOCATE(facebased(nbcs), clockcons(nbcs), fvd(n)%bcnam(nbcs))
        DO i = 1, nbcs
          READ(QUNIT) facebased(i), clockcons(i), fvd(n)%bcnam(i)
        ENDDO

        READ(QUNIT) nvars
        fvd(n)%nvars = nvars
        ALLOCATE(fvd(n)%varnames(nvars))
        DO i = 1, nvars
          READ(QUNIT) fvd(n)%varnames(i)
        ENDDO
        READ(QUNIT) fvd(n)%nbvars

        READ(QUNIT) nodesflag, npts
PRINT*,' nodesflag,npts=',nodesflag,npts
        GRID tpts = npts
        ALLOCATE(GRID x(npts), GRID y(npts), GRID z(npts))
        READ(QUNIT) (GRID x(i),i=1,npts),(GRID y(i),i=1,npts),(GRID z(i),i=1,npts)

        ALLOCATE(fvd(n)%nbfaces(nbcs))
        ALLOCATE(fvd(n)%f(nbcs))
        DO i = 1, nbcs
          READ(QUNIT) fv_faces, k, fvd(n)%nbfaces(k)
PRINT*,' fv_faces = ', fv_faces, k, fvd(n)%nbfaces(k)
          ALLOCATE(fvd(n)%f(k)%f(4,fvd(n)%nbfaces(k)))
          READ(QUNIT) (fvd(n)%f(k)%f(1,j),fvd(n)%f(k)%f(2,j),fvd(n)%f(k)%f(3,j), &
                    fvd(n)%f(k)%f(4,j),j=1,fvd(n)%nbfaces(k))
        ENDDO

        READ(QUNIT) fv_elements, GRID ntet, GRID nhex, GRID npris, GRID npyr
PRINT*,' fv_elements = ', fv_elements, GRID ntet, GRID nhex, GRID npris, GRID npyr
        GRID tcells = GRID ntet + GRID nhex + GRID npris + GRID npyr
        ALLOCATE(fvd(n)%c(8,GRID tcells), fvd(n)%typ(GRID tcells))
        READ(QUNIT) (hdr, (fvd(n)%c(i,j), i=1,elpts(hdr,j,n)), j=1,GRID tcells)

!-----------------------------------------------------------------------
! Flip the 3/4 and 7/8 values for the hexahedral elements because FV
! programmers are jerks and just wanted extra work to do!!!
        DO i = 1, GRID tcells
          IF(fvd(n)%typ(i).EQ.4)THEN
            idum = fvd(n)%c(4,i)
            fvd(n)%c(4,i) = fvd(n)%c(3,i)
            fvd(n)%c(3,i) = idum
            idum = fvd(n)%c(8,i)
            fvd(n)%c(8,i) = fvd(n)%c(7,i)
            fvd(n)%c(7,i) = idum
          ENDIF
        ENDDO
!-----------------------------------------------------------------------

        READ(QUNIT) fv_variables
PRINT*,' fv_variables = ', fv_variables
        ALLOCATE(dw(nvars,npts))
        dw = -42.
        READ(QUNIT) ((dw(j,i), i=1,npts),j=1,nvars)
        READ(QUNIT) fv_bndry_vars
PRINT*,' fv_bndry_vars = ', fv_bndry_vars

        CLOSE(QUNIT)

      CASE DEFAULT
        CALL Fatal_Error(.TRUE.,' Fieldview improper Magic Number: check endianness')
    END SELECT

    CLOSE(QUNIT)

    ALLOCATE(FLOW p(npts), FLOW vx(npts), FLOW vy(npts), &
             FLOW vz(npts), FLOW r(npts))

    DO i = 1, npts
      FLOW r(i)  = dw(1,i)
      FLOW vx(i) = dw(2,i)
      FLOW vy(i) = dw(3,i)
      FLOW vz(i) = dw(4,i)
      FLOW p(i)  = dw(5,i)
    ENDDO

    DEALLOCATE(dw)

    WRITE(*,*) ' Fieldview combined file read successfully'

    RETURN

  END SUBROUTINE Rd_FV_Unf_Comb

!#######################################################################
  INTEGER FUNCTION elpts(hdr,j,n)
!#######################################################################
    USE solver_uns
    INTEGER, INTENT(IN) :: hdr, j, n

    fvd(n)%typ(j) = ISHFT(hdr, -fv_elem_type_byte_shift)
    SELECT CASE(fvd(n)%typ(j))
      CASE(1)  !tet
        elpts = 4
      CASE(2) !pyramid
        elpts = 5
      CASE(3) !prism
        elpts = 6
      CASE(4) !hex
        elpts = 8
    END SELECT
  END FUNCTION elpts

!#######################################################################
  SUBROUTINE Rd_AVUS_Flow(fnam,n)
!#######################################################################
    USE common_area, ONLY: ffmt, griddouble, pfw, npnts, nzones, fftype
    USE MOD_AVUSGRID, ONLY: ReadAVUSPixFile
    USE MOD_GridVars, ONLY : FP
    USE solver_uns, ONLY : avus_ndm

    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
    INTEGER, INTENT(IN) :: n

    REAL(kind=FP), POINTER, DIMENSION(:,:) :: dw
    INTEGER grdformat, npts
    INTEGER i, itwod
    
    itwod = 1
    npts = SUM(npnts)
    griddouble = .true.  !This is hardcoded, to read double precision files


    SELECT CASE(fftype)
      CASE(1,4)
        grdformat = 0
      CASE(2,5)
        grdformat = 1
      CASE(3,6)
        grdformat = 2
    END SELECT
    CALL ReadAVUSPixFile(fnam,grdformat,griddouble,nzones,npnts,dw)

    IF(avus_ndm.eq.2) THEN
      itwod = 2
    ENDIF

    ALLOCATE(FLOW p(npts*itwod), FLOW vx(npts*itwod), FLOW vy(npts*itwod), &
             FLOW vz(npts*itwod), FLOW r(npts*itwod))


    FLOW tyme = REAL(n)

    DO i = 1, npts
      FLOW r(i) = dw(1,i)
      FLOW vx(i) = dw(2,i)
      FLOW vy(i) = dw(3,i)
      FLOW vz(i) = dw(4,i)
      FLOW p(i) = dw(5,i)
    ENDDO


    IF(avus_ndm.eq.2)THEN
      DO i = 1, npts
        FLOW r(i+npts) = dw(1,i)
        FLOW vx(i+npts) = dw(2,i)
        FLOW vy(i+npts) = dw(3,i)
        FLOW vz(i+npts) = dw(4,i)
        FLOW p(i+npts) = dw(5,i)
      ENDDO
    ENDIF


    DEALLOCATE(dw)


  END SUBROUTINE Rd_AVUS_Flow

!#######################################################################
  SUBROUTINE Rd_FDL_Flow(fnam,n,c)
!#######################################################################
  USE common_area, ONLY: pgr, pfw, ffmt, nclients, griddouble, cid, fftype
  USE MOD_P3DGrid, ONLY : FP, DP, ReadP3DRestart, dtau, nfilesect, numrstrtvar, rstrtvarsct
  USE MOD_P3D_ParallelIO, ONLY : ReadParallelFDLRestartFile, tau
  CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
  INTEGER, INTENT(IN) :: n, c
  INTEGER :: i, nb, mvmax, nblks, rtype, grdformat, nstep
  INTEGER, POINTER, DIMENSION(:) :: ieng,jeng,keng,iblank
  REAL(kind=FP), POINTER, DIMENSION(:,:) :: dw, dx
  INTEGER, DIMENSION(:),POINTER :: npnts
  LOGICAL :: multib = .true.

!FDLF90 file format
  rtype = 0

  nblks = GRID nblks

  ieng => GRID li
  jeng => GRID lj
  keng => GRID lk

  ALLOCATE(npnts(1))
  
  npnts = 0
  DO nb = 1, nblks
     npnts = npnts + ieng(nb)*jeng(nb)*keng(nb)
  ENDDO

    SELECT CASE(fftype)
      CASE(1,4)
        grdformat = 0
      CASE(2,5)
        grdformat = 1
      CASE(3,6)
        grdformat = 2
    END SELECT

  nfilesect = 1
  numrstrtvar(1) = 5
  rstrtvarsct(1,1) = 1
  rstrtvarsct(1,2) = 5

    IF(nclients .eq. 1) THEN 
       CALL ReadP3DRestart(fnam,grdformat,griddouble,nblks,npnts,dw,dx,ieng,jeng,keng,rtype,multib)
       FLOW tyme = dtau
    ELSE
       IF(nclients .lt. nblks) THEN
         !Read in the inpartitioned grid and try to evenly distribure them
       ELSE
          CALL ReadParallelFDLRestartFile ( nstep, cid, dw, grdformat, griddouble, fnam )
          FLOW tyme = tau
       END IF
    END IF

    mvmax = GRID tpts


    ALLOCATE(FLOW p(mvmax), FLOW vx(mvmax), FLOW vy(mvmax), &
             FLOW vz(mvmax), FLOW r(mvmax))


    DO i = 1, mvmax
      FLOW vx(i) = dw(1,i)
      FLOW vy(i) = dw(2,i)
      FLOW vz(i) = dw(3,i)
      FLOW p(i)  = dw(4,i)
      FLOW r(i)  = dw(5,i)
    ENDDO


    DEALLOCATE(dw,npnts)

  END SUBROUTINE Rd_FDL_Flow

!#######################################################################
  SUBROUTINE Rd_FDL_Flow_Dum(n,c)
!#######################################################################
    USE common_area, ONLY: pgr, pfw

    INTEGER,INTENT(IN) :: n,c
    INTEGER mvmax, i

    mvmax = GRID tpts
    ALLOCATE(FLOW p(mvmax))
    FLOW tyme = 0.0
    FLOW ncyc = 0

    DO i = 1, mvmax
      FLOW p(i) = GRID iblank(i)
    ENDDO
    FLOW vx => FLOW p
    FLOW vy => FLOW p
    FLOW vz => FLOW p
    FLOW r  => FLOW p

  END SUBROUTINE Rd_FDL_Flow_Dum

!#######################################################################
  SUBROUTINE Rd_P3D_Flow(fnam,n,c)
!#######################################################################
  USE common_area, ONLY: pgr, pfw, ffmt, nclients, griddouble, cid, fftype
  USE MOD_P3DGrid, ONLY : FP, DP, ReadP3DRestart, dtau, nfilesect, numrstrtvar, rstrtvarsct
  USE MOD_P3D_ParallelIO, ONLY : ReadParallelP3DRestartFile, tau

  CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
  INTEGER, INTENT(IN) :: n,c
  INTEGER :: i, nb, mvmax, nblks, rtype, grdformat, nstep
  INTEGER, POINTER, DIMENSION(:) :: ieng,jeng,keng,iblank
  REAL(kind=FP), POINTER, DIMENSION(:,:) :: dw, dx
  INTEGER, DIMENSION(:),POINTER :: npnts
  LOGICAL :: multib = .true.

  IF(fnam.EQ.'xxx')THEN
    CALL Rd_P3DT_Dum(n,c)
    RETURN
  ENDIF

!Q-File Read
  rtype = 2

  nblks = GRID nblks

  ieng => GRID li
  jeng => GRID lj
  keng => GRID lk

  ALLOCATE(npnts(1))
  
  npnts = 0
  DO nb = 1, nblks
     npnts = npnts + ieng(nb)*jeng(nb)*keng(nb)
  ENDDO

    SELECT CASE(fftype)
      CASE(1,4)
        grdformat = 0
      CASE(2,5)
        grdformat = 1
      CASE(3,6)
        grdformat = 2
    END SELECT

    nfilesect = 1
    numrstrtvar(1) = 5
    rstrtvarsct(1,1) = 1
    rstrtvarsct(1,2) = 5

    IF(nclients .eq. 1) THEN 
       CALL ReadP3DRestart(fnam,grdformat,griddouble,nblks,npnts,dw,dx,ieng,jeng,keng,rtype,multib)
       FLOW tyme = dtau
    ELSE
       IF(nclients .lt. nblks) THEN
         !Read in the inpartitioned grid and try to evenly distribure them
       ELSE
          CALL ReadParallelP3DRestartFile ( nstep, cid, dw, grdformat, griddouble, fnam , rtype )
          FLOW tyme = tau
       END IF
    END IF

    mvmax = GRID tpts


    ALLOCATE(FLOW r(mvmax), FLOW vx(mvmax), FLOW vy(mvmax), &
             FLOW vz(mvmax), FLOW e(mvmax))

    FLOW tyme = dtau

    DO i = 1, mvmax
      FLOW r(i) = dw(1,i)
      FLOW vx(i) = dw(2,i)/dw(1,i)
      FLOW vy(i) = dw(3,i)/dw(1,i)
      FLOW vz(i)  = dw(4,i)/dw(1,i)
      FLOW e(i)  = dw(5,i)/dw(1,i)
    ENDDO


    DEALLOCATE(dw,npnts)

  END SUBROUTINE Rd_P3D_Flow

!#######################################################################
  SUBROUTINE Rd_P3D_Flw_Ray(fname,n,c,rntime)
!#######################################################################
    USE general, ONLY: Fatal_Error, Check_File_Exist
    USE common_area, ONLY: pfw, cur, pgr, igrid, griddouble, ntmax
    USE mod_clear, ONLY: ClearVar
    USE MOD_P3DRay, ONLY: FP, ReadP3DFlow, ReadP3DTime

    integer, intent(in) :: n, c
    integer, intent(out) :: rntime
    integer, pointer, dimension(:) :: ieng, jeng, keng
    integer :: nblgs = 1, ngrd = 1, ng = 1
    integer :: npts, mvmax, i, j, k

    real (kind=FP), pointer, dimension(:,:,:) :: q
    character(LEN=NSTRLEN), intent(in) :: fname

    logical :: ios

    ieng => GRID li
    jeng => GRID lj
    keng => GRID lk
    npts = ieng(ng) * jeng(ng) * keng(ng)

    if( n.EQ.0 ) then
      if( fname.EQ.'xxx' ) then
        rntime = 1
      else
        CALL ReadP3DTime(fname,0,griddouble,rntime)
      endif
      return
    endif

    if( fname.EQ.'xxx' ) then
        ! apply zero flow field
        ALLOCATE(FLOW r(npts), FLOW vx(npts))
        DO i = 1, npts
            FLOW r(i) = 1.
            FLOW vx(i) = 0.
        ENDDO
        FLOW vy => FLOW vx
        FLOW vz => FLOW vx
        FLOW vr => FLOW vx
        FLOW vt => FLOW vx
        FLOW e =>  FLOW r
        FLOW k =>  FLOW vx
        FLOW eps => FLOW vx
        FLOW mut => FLOW vx
        return
    endif
!    PRINT*, n, fname
    CALL Check_File_Exist(fname,ios)
    CALL Fatal_Error(ios,'*** Flow file does not exist ***')
    CALL ReadP3DTime(fname,0,griddouble,rntime)
!    write(*,*) '  Flow file scanned, number of timesteps: ',rntime
    IF((n+rntime-1).GT.ntmax) rntime = ntmax-n+1
    CALL ReadP3DFlow(fname,0,griddouble,q,ieng,jeng,keng,rntime)
!    WRITE(*,*) '  Flow file successfully read'

    do j=1,rntime
        k = j+n-1
        ! Reprocess cartesian coordinates to cylindrical
        ALLOCATE( pfw(k)%r(npts),  &
                  pfw(k)%vx(npts), &
                  pfw(k)%vy(npts), &
                  pfw(k)%vz(npts), &
                  pfw(k)%e(npts),  &
                  pfw(k)%vr(npts), &
                  pfw(k)%vt(npts), &
                  pfw(k)%k(npts) )
        DO i=1,npts
            pfw(k)%r(i) = q(j,1,i)
            pfw(k)%vx(i) = q(j,4,i)/q(j,1,i)
            pfw(k)%vy(i) = q(j,3,i)/q(j,1,i)
            pfw(k)%vz(i) = q(j,2,i)/q(j,1,i)
            pfw(k)%e(i) = q(j,5,i)/q(j,1,i)
            pfw(k)%k(i) = 0.
            pfw(k)%vr(i) = pfw(k)%vx(i)*COS(-GRID t0(i)) + pfw(k)%vy(i)*SIN(-GRID t0(i))
            pfw(k)%vt(i) = pfw(k)%vx(i)*SIN(-GRID t0(i)) - pfw(k)%vy(i)*COS(-GRID t0(i))
            pfw(k)%e(i)  = pfw(k)%e(i) - (pfw(k)%vr(i)**2 + pfw(k)%vt(i)**2 + pfw(k)%vz(i)**2)/2.
        ENDDO
        pfw(j)%tyme = FLOAT(j)
        CALL ClearVar(pfw(j)%vx)
        CALL ClearVar(pfw(j)%vy)
    enddo
    deallocate(q)
!    WRITE(*,*) '  Flow file successfully converted to cylindrical coordinates'

  END SUBROUTINE Rd_P3D_Flw_Ray

END MODULE readflow
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
