#include "parameters.h"
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE readgrid
  IMPLICIT NONE

CONTAINS

!#######################################################################
  SUBROUTINE Init_Pgr()
!#######################################################################
    USE common_area, ONLY: ngrid, pgr
    USE solver_uns
    
    INTEGER i

    NULLIFY(fvd)

    DO i=1,ngrid
      NULLIFY(pgr(i)%x)
      NULLIFY(pgr(i)%y)
      NULLIFY(pgr(i)%z)
      NULLIFY(pgr(i)%em)
      NULLIFY(pgr(i)%r)
      NULLIFY(pgr(i)%t)
      NULLIFY(pgr(i)%t0)
      NULLIFY(pgr(i)%ph)
      NULLIFY(pgr(i)%i)
      NULLIFY(pgr(i)%j)
      NULLIFY(pgr(i)%k)
      NULLIFY(pgr(i)%w)

      NULLIFY(pgr(i)%iblank)

      NULLIFY(pgr(i)%li)
      NULLIFY(pgr(i)%lj)
      NULLIFY(pgr(i)%lk)
      NULLIFY(pgr(i)%gis)
      NULLIFY(pgr(i)%gjs)
      NULLIFY(pgr(i)%gks)
      NULLIFY(pgr(i)%mvmax)
    ENDDO

  END SUBROUTINE Init_Pgr


!#######################################################################
  SUBROUTINE Get_Grids()
!#######################################################################
    USE solver_uns
    USE common_area, ONLY: pgr, ngrid, runmode, gridarr, ffmt, cid
    USE general, ONLY: int_to_string, Prompt_Get
    CHARACTER(LEN=NSTRLEN) :: fnam
    INTEGER i

! For now the grid structure is unchanging and there is no "moving mesh"
! allowed.  The grid type variable is set up for that case, but the input
! may be a pain in the @$$.  CFD++ is the exception, dumping a nodesout.bin
! file every ### iterations.
    ALLOCATE(pgr(ngrid))
    CALL Init_Pgr()

    SELECT CASE(runmode)
!=======================================================================
! This is for reading VSTAGE files in batch, after a merge for instance.
! It is also useful in setting up odd stuff???
      CASE (1,11)
        SELECT CASE(ffmt)
          CASE(1)
            !CALL Rd_Vstg_Grd(gridarr(cid),1)
            CALL Rd_Vstg_Grd(gridarr(1),1)
          CASE(2)
            CALL Rd_P3DT_Grd(gridarr(1),1)
          CASE(3)
            CALL Rd_P3D_Grd_Ray(gridarr(1),1)
        END SELECT

!=======================================================================
! Runmode 2 will be dedicated to reading in the AST.B#.X files from a
! local harddisk or a global harddisk.  These files are simply plot3D
! or vstage files
      CASE (2)
        WRITE(fnam,*) 'ast.b' // TRIM(ADJUSTL(int_to_string(cid))) // &
&                         '.' // TRIM(ADJUSTL(gridarr(1))) // '.x'
        SELECT CASE(ffmt)
          CASE(1)
            CALL Rd_Vstg_Grd(fnam,1)
          CASE(2)
            CALL Rd_P3DT_Grd(fnam,1)
          CASE(3)
            CALL Rd_P3D_Grd_Ray(fnam,1)
        END SELECT

!=======================================================================
! Runmode 3 will be the normal plot3d reads
      CASE(3,6)
        CALL Rd_P3D_Grd(gridarr(1),1)

!=======================================================================
! Runmode 4 will be VERY general stuff
      CASE(4)
        SELECT CASE(ffmt)
          CASE(1) ! Formatted assuming that there is split format
            DO i=1,ngrid
!              CALL Rd_FV_Frm_Grd(gridarr(i),i)
            ENDDO
          CASE(2) ! Unformatted assuming split
            DO i=1,ngrid
!              CALL Rd_FV_Unf_Grd(gridarr(i),i)
            ENDDO
        END SELECT

!=======================================================================
! AVUS
      CASE(5)
        ALLOCATE(fvd(ngrid))
        DO i=1,ngrid
          CALL Rd_AVUS_Grd(gridarr(i),i)
        ENDDO

!=======================================================================
! The default case is interactive mode.  Currently, support is for VSTAGE
! as can be seen by the single call and lack of a switch for the file format.
      CASE DEFAULT
        WRITE(*,*) '  Enter file format ==>'
        WRITE(*,*) ' 1  VStage'
        WRITE(*,*) ' 2  Plot3D'
        READ*,ffmt
        CALL Prompt_Get(fnam,' Enter grid file ==>')
        SELECT CASE(ffmt)
          CASE(1)
            CALL Rd_Vstg_Grd(fnam,1)
          CASE(2)
            CALL Rd_P3DT_Grd(fnam,1)
          CASE(3)
            CALL Rd_P3D_Grd_Ray(fnam,1)
        END SELECT

    END SELECT

  END SUBROUTINE Get_Grids

!#######################################################################
  SUBROUTINE ReloadGrid(newgrid)
!#######################################################################
    USE keydata, ONLY: RSTATE
    CHARACTER(LEN=NSTRLEN) :: newgrid
 
    RSTATE = RSTATE_READY

  END SUBROUTINE ReloadGrid

!#######################################################################
  SUBROUTINE Rd_P3D_Grd(fnam,n)
!#######################################################################
    USE general, ONLY: Check_File_Exist, Fatal_Error
    USE MOD_P3DGrid, ONLY : FP, ReadP3DGrid
    USE MOD_P3D_ParallelIO, ONLY : ReadGridIndexFile, ReadParallelP3DFile
    USE common_area, ONLY : nclients, gridindexfile, ffmt, coordsys, griddouble, pgr, liblank, cid, fftype
    LOGICAL :: ios
    CHARACTER(LEN=NSTRLEN) :: fnam
    INTEGER :: i,j,k,s,id,jd,kd,m,n,mvmax,nblks,ndm
    INTEGER grdformat
    INTEGER, POINTER, DIMENSION(:) :: ieng,jeng,keng
    INTEGER, POINTER, DIMENSION(:,:) :: ipnt,icell
    REAL(kind=FP), POINTER, DIMENSION(:,:) :: x
    LOGICAL :: multib = .true.

    coordsys = CARTESIAN

    CALL Check_File_Exist(fnam,ios)
    CALL Fatal_Error(ios,'*** Grid file does not exist ***')

    SELECT CASE(fftype)
      CASE(1,4)
        grdformat = 0
        OPEN(unit=1,file=fnam,form='unformatted',status='old')
        READ(1) nblks
        CLOSE(1)
      CASE(2,5)
        grdformat = 1
        OPEN(unit=1,file=fnam,form='formatted',status='old')
        READ(1,*) nblks
        CLOSE(1)
      CASE(3,6)
        grdformat = 2
        OPEN(unit=1,file=fnam,form='binary',status='old')
        READ(1) nblks
        CLOSE(1)
    END SELECT

    IF(nclients .eq. 1) THEN 
! pgr(n)%iblank will be allocated in the read call
       CALL ReadP3DGrid(fnam,grdformat,griddouble,nblks,ndm,x,ieng,jeng,keng,pgr(n)%iblank,liblank,multib)
       
       mvmax = 0
       ALLOCATE(pgr(n)%gis(nblks), pgr(n)%gjs(nblks), &
                pgr(n)%gks(nblks), pgr(n)%mvmax(nblks))
       DO m = 1, nblks
          pgr(n)%gis(m)=0
          pgr(n)%gjs(m)=0
          pgr(n)%gks(m)=0
          pgr(n)%mvmax(m)=ieng(m)*jeng(m)*keng(m)
          mvmax=mvmax+pgr(n)%mvmax(m)
       ENDDO
       pgr(n)%tpts = mvmax

       ALLOCATE(pgr(n)%x(mvmax),pgr(n)%y(mvmax),pgr(n)%z(mvmax))

       DO i = 1, mvmax
          pgr(n)%x(i) = x(1,i)
          pgr(n)%y(i) = x(2,i)
          pgr(n)%z(i) = x(3,i)
       ENDDO

       DEALLOCATE(x)

    ELSE
       IF(nclients .lt. nblks) THEN
         WRITE(*,*) 'ERROR: pv3_num_clients must be greater than the i'
         WRITE(*,*) '       number of grids in the grid file.'
         !Read in the inpartitioned grid and try to evenly distribure them
       ELSE
          ALLOCATE(ieng(1),jeng(1),keng(1),pgr(n)%mvmax(1))
          nblks = 1 
          WRITE(*,*) 'Reading and Partitioning the grid file'
         !Read in the partitioning information
          CALL ReadGridIndexFile ( nclients, gridindexfile )
         !Read a block from the plot3d grid
          CALL ReadParallelP3DFile ( cid, pgr(n)%x, pgr(n)%y, pgr(n)%z, pgr(n)%iblank, ieng(1), jeng(1), keng(1), grdformat, griddouble, fnam, liblank )
          pgr(n)%tpts = ieng(1)*jeng(1)*keng(1)
          pgr(n)%mvmax(1) = ieng(1)*jeng(1)*keng(1)
       END IF
    END IF

    ALLOCATE(pgr(n)%li(nblks),pgr(n)%lj(nblks),pgr(n)%lk(nblks))
    pgr(n)%nblks = nblks
    pgr(n)%li = ieng
    pgr(n)%lj = jeng
    pgr(n)%lk = keng


    DEALLOCATE(ieng,jeng,keng)

  END SUBROUTINE Rd_P3D_Grd

!#######################################################################
  SUBROUTINE Rd_P3DT_Grd(fnam,n)
!#######################################################################
    USE general, ONLY: Check_File_Exist, Fatal_Error
    USE common_area, ONLY: coordsys, pgr, liblank, nbld
    USE MOD_P3DGrid, ONLY : P3DSetDebug

    LOGICAL :: ios
    CHARACTER(LEN=NSTRLEN),INTENT(IN) :: fnam
    INTEGER :: i,j,k,s,id,jd,kd,m,n,mvmax,nblks

    CALL P3DSetDebug(.FALSE.)
    CALL Rd_P3D_Grd(fnam, n)
    IF(.NOT. ASSOCIATED(pgr(n)%x))PRINT*,'X not associated!!!'
    IF(.NOT. ASSOCIATED(pgr(n)%y))PRINT*,'Y not associated!!!'
    IF(.NOT. ASSOCIATED(pgr(n)%z))PRINT*,'Z not associated!!!'
    PRINT*,'MVMAX=',pgr(n)%mvmax(1)
    PRINT*,'TPTS =',pgr(n)%tpts
    coordsys = RHCYLINDRICAL

    mvmax = pgr(n)%tpts
    ALLOCATE(pgr(n)%r(mvmax),pgr(n)%t0(mvmax))
    DO m=1,mvmax
      pgr(n)%r(m) = SQRT(pgr(n)%x(m)**2 + pgr(n)%y(m)**2)
      pgr(n)%t0(m) = ATAN2(pgr(n)%y(m),pgr(n)%x(m))
      pgr(n)%y(m) = -pgr(n)%y(m)
    ENDDO

  END SUBROUTINE Rd_P3DT_Grd

!#######################################################################
  SUBROUTINE Rd_P3D_Grd_Ray(fname,n)
!#######################################################################
    USE general, ONLY: Check_File_Exist, Fatal_Error
    USE common_area, ONLY: coordsys, pgr, liblank, nbld, griddouble, &
                             cid, trot, omega, grd_translt
    USE MOD_P3DRay, ONLY: FP, ReadP3DGrid

    LOGICAL :: ios, exists, shift
    LOGICAL :: multiblock = .TRUE.
    INTEGER :: i, j, k, m, n, mvmax, nblks
    INTEGER grdformat
    CHARACTER(LEN=NSTRLEN),INTENT(IN) :: fname
    INTEGER, POINTER, DIMENSION(:) :: ieng, jeng, keng
    REAL(kind=FP), POINTER, DIMENSION(:,:) :: x
    INTEGER :: bid = 0

    CALL Check_File_Exist(fname,ios)
    CALL Fatal_Error(ios,'*** Grid file does not exist ***')

!    SELECT CASE(fftype)
!      CASE(1,4)
!       unformatted, single precision
        grdformat = 0
        OPEN(unit=1,file=fname,form='unformatted',status='old')
        READ(1) nblks
        CLOSE(1)
!      CASE(2,5)
!        grdformat = 1
!        OPEN(unit=1,file=fnam,form='formatted',status='old')
!        READ(1,*) nblks
!        CLOSE(1)
!      CASE(3,6)
!        grdformat = 2
!        OPEN(unit=1,file=fnam,form='binary',status='old')
!        READ(1) nblks
!        CLOSE(1)
!    END SELECT

    INQUIRE(FILE='rot.in',EXIST=exists)
    IF(exists)THEN
      OPEN(UNIT=2,FILE='rot.in',FORM='formatted')
      DO WHILE(1)
        READ(2,*,END=1110,ERR=1110) bid
        IF(bid.EQ.cid)THEN
          EXIT
        ENDIF
      ENDDO
      trot = .TRUE.
      GOTO 1111
1110  CONTINUE
      trot = .FALSE.
      grd_translt = 0.
1111  CONTINUE
      CLOSE(2)
    ENDIF

    INQUIRE(FILE='shift.in',EXIST=exists)
    IF(exists)THEN
      OPEN(UNIT=2,FILE='shift.in',FORM='formatted')
      DO WHILE(1)
        READ(2,*,END=1120,ERR=1120) bid
        IF(bid.EQ.cid)THEN
          EXIT
        ENDIF
      ENDDO
      shift = .TRUE.
      GOTO 1121
1120  CONTINUE
      shift = .FALSE.
1121  CONTINUE
      CLOSE(2)
    ENDIF

    PRINT*,'NBLKS=',nblks
    coordsys = CARTESIAN

    CALL ReadP3DGrid(fname,grdformat,griddouble,x,ieng,jeng,keng)

    ALLOCATE( pgr(n)%gis(nblks), &
              pgr(n)%gjs(nblks), &
              pgr(n)%gks(nblks), &
              pgr(n)%mvmax(nblks) )
    DO m = 1, nblks
        pgr(n)%gis(m) = 0
        pgr(n)%gjs(m) = 0
        pgr(n)%gks(m) = 0
        pgr(n)%mvmax(m) = ieng(m) * jeng(m) * keng(m)
        mvmax = mvmax + pgr(n)%mvmax(m)
    ENDDO
    pgr(n)%tpts = mvmax

    ALLOCATE( pgr(n)%x(mvmax), &
              pgr(n)%y(mvmax), &
              pgr(n)%z(mvmax)  )
    DO i = 1, mvmax
        pgr(n)%x(i) = x(1,i)
        pgr(n)%y(i) = x(2,i)
        pgr(n)%z(i) = x(3,i)
    ENDDO
    DEALLOCATE(x)
    
    ALLOCATE( pgr(n)%li(nblks), &
              pgr(n)%lj(nblks), &
              pgr(n)%lk(nblks)  )
    pgr(n)%nblks = nblks
    pgr(n)%li = ieng
    pgr(n)%lj = jeng
    pgr(n)%lk = keng

    DEALLOCATE(ieng, jeng, keng)

    IF(.NOT. ASSOCIATED(pgr(n)%x))PRINT*,'X not associated!!!'
    IF(.NOT. ASSOCIATED(pgr(n)%y))PRINT*,'Y not associated!!!'
    IF(.NOT. ASSOCIATED(pgr(n)%z))PRINT*,'Z not associated!!!'
    PRINT*,'MVMAX=',pgr(n)%mvmax(1)
    PRINT*,'TPTS =',pgr(n)%tpts

    coordsys = RHCYLINDRICAL
    mvmax = pgr(n)%tpts
    ALLOCATE(pgr(n)%r(mvmax),pgr(n)%t0(mvmax))
    DO m=1,mvmax
        pgr(n)%r(m) = SQRT(pgr(n)%x(m)**2 + pgr(n)%y(m)**2)
        pgr(n)%t0(m) = -ATAN2(pgr(n)%y(m),pgr(n)%x(m))
        if (shift) then
            if (pgr(n)%t0(m) < 0) pgr(n)%t0(m) = pgr(n)%t0(m) + 2*PI
        endif
        pgr(n)%y(m) = -pgr(n)%y(m)
    ENDDO

END SUBROUTINE Rd_P3D_Grd_Ray

!#######################################################################
  SUBROUTINE Rd_Vstg_Grd(fnam,n)
!#######################################################################
    USE general, ONLY: Check_File_Exist, Fatal_Error
    USE common_area, ONLY: coordsys, pgr
! fnam is the file name to open
! n is the number of the array of grids being filled (usually 1)
!
! The grid read stores both the points and the structure

    LOGICAL :: ios
    REAL :: minv,maxv,dtheta
    CHARACTER(LEN=NSTRLEN),INTENT(IN) :: fnam
    INTEGER :: i,j,k,id,jd,kd,n,mvmax

    coordsys = CYLINDRICAL

    WRITE(*,*) ' Reading grid file',n,fnam
    CALL Check_File_Exist(fnam,ios)
    CALL Fatal_Error(ios,'*** Grid file does not exist ***')

    OPEN(UNIT=GUNIT,FILE=fnam,FORM='unformatted')

    ALLOCATE(pgr(n)%li(1),pgr(n)%lj(1),pgr(n)%lk(1),pgr(n)%mvmax(1))
    READ(GUNIT) pgr(n)%lk(1),pgr(n)%lj(1),pgr(n)%li(1), &
             pgr(n)%mle,pgr(n)%mte,pgr(n)%mtp,pgr(n)%nbld
    pgr(n)%mhb=1
    pgr(n)%mvmax(1) = pgr(n)%lk(1)*pgr(n)%lj(1)*pgr(n)%li(1)
    pgr(n)%tpts = pgr(n)%mvmax(1)

    id = pgr(n)%li(1)
    jd = pgr(n)%lj(1)
    kd = pgr(n)%lk(1)
    mvmax = pgr(n)%tpts
    ALLOCATE(pgr(n)%r(mvmax),pgr(n)%t0(mvmax),pgr(n)%z(mvmax))
    WRITE(*,*) '  lk,lj,li,mvmax = ',kd,jd,id,mvmax
    WRITE(*,*) '  mle,mte,mtp,nbld = ',pgr(n)%mle,pgr(n)%mte,pgr(n)%mtp,pgr(n)%nbld

    pgr(n)%nblks = 1 ! Single block VSTAGE
    DO j=1,jd
      READ(GUNIT) &
        ((pgr(n)%z(i+(j-1)*id+(k-1)*id*jd),k=1,kd),i=1,id), &
        ((pgr(n)%r(i+(j-1)*id+(k-1)*id*jd),k=1,kd),i=1,id), &
        ((pgr(n)%t0(i+(j-1)*id+(k-1)*id*jd),k=1,kd),i=1,id)
    ENDDO
    WRITE(*,*) '  The grid file was read successfully'

    ALLOCATE(pgr(n)%gis(1),pgr(n)%gjs(1),pgr(n)%gks(1))
    pgr(n)%gis(1)=0
    pgr(n)%gjs(1)=0
    pgr(n)%gks(1)=0

    CLOSE(GUNIT)

    dtheta = 2.*PI/FLOAT(pgr(n)%nbld)
    minv = pgr(n)%t0(1)
    maxv = pgr(n)%t0(1)
    DO i=2,mvmax
      minv = MIN(minv,pgr(n)%t0(i))
      maxv = MAX(maxv,pgr(n)%t0(i))
    ENDDO
    IF(minv.LE.(-2.*PI+dtheta) .AND. maxv.GE.-dtheta)THEN
      PRINT*,' ADJUSTING ROGUE THETA VALUES'
      DO i=1,mvmax
        IF(pgr(n)%t0(i).LT.-dtheta)THEN
          pgr(n)%t0(i) = pgr(n)%t0(i) + 2.*PI
        ENDIF
      ENDDO
    ELSE
      PRINT*,' MINV=',minv,' MAXV=',maxv
    ENDIF

  END SUBROUTINE Rd_Vstg_Grd

!#######################################################################
  SUBROUTINE Rd_FV_Unf_Grd(fnam,n)
!#######################################################################
    USE general, ONLY: Check_File_Exist, Fatal_Error
    USE common_area, ONLY: coordsys

    LOGICAL :: ios
    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
    INTEGER, INTENT(IN) :: n

    INTEGER :: fv_magic, majver, minver

    coordsys = CARTESIAN

    WRITE(*,*) ' Reading grid file',n,fnam
    CALL Check_File_Exist(fnam,ios)
    CALL Fatal_Error(ios,'*** Grid file does not exist ***')

    OPEN(UNIT=GUNIT,FILE=fnam,FORM='unformatted')

    READ(GUNIT) fv_magic
    SELECT CASE(fv_magic)
      CASE(66051)
        READ(GUNIT) majver, minver
      CASE DEFAULT
        WRITE(*,*) ' Sorry `bout your luck...'
        STOP
    END SELECT

    CLOSE(GUNIT)

    RETURN

  END SUBROUTINE Rd_FV_Unf_Grd

!#######################################################################
  SUBROUTINE Rd_AVUS_Grd(fnam,n)
!#######################################################################
    USE MOD_AVUSGrid
    USE MOD_GridVars, ONLY : FP
    USE solver_uns
    USE common_area, ONLY: ffmt, coordsys, griddouble, nzones, npnts, fftype

    CHARACTER(LEN=NSTRLEN), INTENT(IN) :: fnam
    INTEGER, INTENT(IN) :: n

    INTEGER grdformat
    INTEGER, POINTER, DIMENSION(:) :: mxppf,mxfpc,numppf,nfaces,ncells
    INTEGER, POINTER, DIMENSION(:,:) :: ipnt,icell
    INTEGER :: npatch
    LOGICAL :: readcnct

    REAL(kind=FP), POINTER, DIMENSION(:,:) :: x

    coordsys = CARTESIAN

    SELECT CASE(fftype)
      CASE(1,4)
        grdformat = 0
      CASE(2,5)
        grdformat = 1
      CASE(3,6)
        grdformat = 2
    END SELECT
    readcnct = .TRUE.

    CALL ReadAVUSGrid(fnam,grdformat,griddouble,avus_ndm,nzones, &
                      npatch,x,npnts,nfaces,ncells,mxppf,mxfpc,numppf, &
                      ipnt,icell,readcnct)

    SELECT CASE(avus_ndm)
      CASE(2)
        CALL Store_AVUS_2D(x,npatch,ipnt,npnts,nfaces,icell,ncells,numppf,n)
      CASE(3)
        CALL Store_AVUS_3D(x,npatch,ipnt,npnts,nfaces,icell,ncells,numppf,n)
    END SELECT

    DEALLOCATE(mxppf,mxfpc,numppf,nfaces,ncells,ipnt,icell)

  END SUBROUTINE Rd_AVUS_Grd
 
END MODULE readgrid
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
