#include "parameters.h"
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
MODULE solver_turbo
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
  USE typedefs

  IMPLICIT NONE

  INTEGER :: nturbobcs = 0       ! The number of bcs found in bc.in
  INTEGER :: nturbobcs2 = 0      ! Adjusted number of bcs
  INTEGER :: nblds = 0           ! The number of blades in the grid
  INTEGER :: bladerow = 0        ! The blade row containing the grid
  INTEGER :: bladepass = 0       ! The blade passage containing the grid
  INTEGER :: num_b2b = 0         ! Number of local block connectivities
  INTEGER :: num_special_b2b = 0 ! Number of local special connectivities
  INTEGER :: ksym = 0            ! Symmetry factor

  INTEGER,POINTER,DIMENSION(:,:) :: turbobcs

  REAL,POINTER,DIMENSION(:) :: turbobcgr

  TYPE(dmap),POINTER,DIMENSION(:) :: b2b  ! block to block face info
  TYPE(dmap),POINTER,DIMENSION(:) :: b2bs ! special b2b pairs


  TYPE patchdata
    INTEGER,POINTER,DIMENSION(:,:) :: ptch ! Patch number, number of integer stored
    INTEGER :: nptch
    LOGICAL,POINTER,DIMENSION(:) :: used
  END TYPE patchdata

  LOGICAL,POINTER,DIMENSION(:) :: turbobcplot
  CHARACTER(LEN=20),POINTER,DIMENSION(:) :: turbobcnam
  TYPE(patchdata),POINTER,DIMENSION(:) :: patches

CONTAINS

!#######################################################################
  SUBROUTINE read_extract()
!#######################################################################
    USE common_area, ONLY: csnm, pgr, igrid, cid
    USE general, ONLY: int_to_string
    IMPLICIT NONE

    INTEGER :: bid,br,is,js,ks,ie,je,ke,dum1,dum2,dum3,bp
    LOGICAL :: exists=.FALSE.
    CHARACTER(LEN=NSTRLEN) :: newextract

    newextract='extract.in'
    INQUIRE(FILE=newextract,EXIST=exists)
    IF(.NOT. exists)THEN
      newextract=TRIM(csnm) // '.extract'
      INQUIRE(FILE=newextract,EXIST=exists)
      IF(.NOT. exists)RETURN
    ENDIF
    OPEN(UNIT=BCSUNIT,FILE=newextract,FORM='formatted',STATUS='old')
    bid=9999
    DO WHILE(bid.GT.0)
      READ(BCSUNIT,*,END=9001,ERR=9001) &
           bid,dum1,is,ie,js,je,ks,ke,dum2,br,bp
      IF(bid.EQ.cid)THEN
        EXIT
      ENDIF
    ENDDO

    bladerow = br
    bladepass = bp
    pgr(igrid)%gis(1)=ks-1
    pgr(igrid)%gjs(1)=js-1
    pgr(igrid)%gks(1)=is-1
    CLOSE(BCSUNIT)
    WRITE(*,*) ' My blade row is ',br

    RETURN

9001 CONTINUE
    WRITE(*,*) ' Error, did not find cid in extract.in'
    CLOSE(BCSUNIT)
    STOP

  END SUBROUTINE read_extract

!#######################################################################
  SUBROUTINE read_dmap()
!#######################################################################
    USE common_area, ONLY: cid
    IMPLICIT NONE

    INTEGER :: num_b2bl,id,dum,dummy(21)
    INTEGER :: i,ctr(2),cnt

    LOGICAL :: exists

    ctr=0

    INQUIRE(FILE='dmap.in',EXIST=exists)
    IF(.NOT. exists)RETURN
    OPEN(UNIT=BCSUNIT,FILE='dmap.in',FORM='formatted',STATUS='old')
    READ(BCSUNIT,*) num_b2bl ! Normal b2b
    DO i=1,num_b2bl
      READ(BCSUNIT,*) id, dum,dum,dum,dum,dum,dum,  &
 &      dum, dum,dum,dum, dum,dum,dum, dum,dum,dum, &
 &      dum,dum,dum,dum
      IF(id.EQ.cid)ctr(1)=ctr(1)+1
    ENDDO

    READ(BCSUNIT,*) num_b2bl ! Special b2bs
    DO i=1,num_b2bl
      READ(BCSUNIT,*) id, dum,dum,dum,dum,dum,dum,  &
 &      dum, dum,dum,dum, dum,dum,dum, dum,dum,dum, &
 &      dum,dum,dum,dum, dum
      IF(id.EQ.cid)ctr(2)=ctr(2)+1
    ENDDO

    REWIND(BCSUNIT)
    ALLOCATE(b2b(ctr(1)),b2bs(ctr(2)))
    num_b2b=ctr(1)
    num_special_b2b=ctr(2)

    cnt=1
    READ(BCSUNIT,*) num_b2bl ! Normal b2b
    DO i=1,num_b2bl
      READ(BCSUNIT,*) b2b(cnt)%id,             &
 &      b2b(cnt)%is,b2b(cnt)%ie,               &
 &      b2b(cnt)%js,b2b(cnt)%je,               &
 &      b2b(cnt)%ks,b2b(cnt)%ke,               &
 &      b2b(cnt)%blkb,                         &
 &      b2b(cnt)%d1,b2b(cnt)%d1s,b2b(cnt)%d1e, &
 &      b2b(cnt)%d2,b2b(cnt)%d2s,b2b(cnt)%d2e, &
 &      b2b(cnt)%d3,b2b(cnt)%d3s,b2b(cnt)%d3e, &
 &      b2b(cnt)%dir1,b2b(cnt)%dir2,           &
 &      b2b(cnt)%lor1,b2b(cnt)%lor2
      IF(id.EQ.cid)cnt=cnt+1
    ENDDO

    cnt=1
    READ(BCSUNIT,*) num_b2bl ! Normal b2b
    DO i=1,num_b2bl
      READ(BCSUNIT,*) b2bs(cnt)%id,               &
 &      b2bs(cnt)%is,b2bs(cnt)%ie,                &
 &      b2bs(cnt)%js,b2bs(cnt)%je,                &
 &      b2bs(cnt)%ks,b2bs(cnt)%ke,                &
 &      b2bs(cnt)%blkb,                           &
 &      b2bs(cnt)%d1,b2bs(cnt)%d1s,b2bs(cnt)%d1e, &
 &      b2bs(cnt)%d2,b2bs(cnt)%d2s,b2bs(cnt)%d2e, &
 &      b2bs(cnt)%d3,b2bs(cnt)%d3s,b2bs(cnt)%d3e, &
 &      b2bs(cnt)%dir1,b2bs(cnt)%dir2,            &
 &      b2bs(cnt)%lor1,b2bs(cnt)%lor2,b2bs(cnt)%bc
      IF(id.EQ.cid)cnt=cnt+1
    ENDDO
    
    CLOSE(BCSUNIT)

  END SUBROUTINE read_dmap

!#######################################################################
  SUBROUTINE read_rowinput()
!#######################################################################
    USE common_area, ONLY: vref_m, omega, diam
    USE general, ONLY: int_to_string
    IMPLICIT NONE

    CHARACTER*10 fnam
    INTEGER brn

      integer            :: num_blades 
      integer            :: blade_row_ksym
      real*8             :: omega_bld 
      integer            :: num_time_steps_stored 
      integer            :: num_adjacent_blades 
      logical            :: upstream_freq_only= .false.
      integer            :: num_sources
      integer            :: suction_surface 
      integer            :: use_wall_func_i
      integer            :: use_wall_func_j 
      integer            :: use_wall_func_k 
      real*8             :: transition 
      real*8             :: x_start_viscous_hub 
      real*8             :: x_start_viscous_case 
      real*8             :: x_start_hub_rotation 
      real*8             :: x_end_hub_rotation 
      real*8             :: x_start_case_rotation 
      real*8             :: x_end_case_rotation 
!xw_animantion
      integer            ::cut_plane_dir
      integer            ::num_cut_plane
      integer            ::cut_plane_global_posi(10)
      integer            ::ll

      NAMELIST/BLADE_ROW_PARAMETERS/                            &
      num_blades,             omega_bld,                        &
      num_time_steps_stored,  num_adjacent_blades,              &
      upstream_freq_only,                                       &
      suction_surface,        use_wall_func_i, use_wall_func_j, &
      use_wall_func_k,        transition,                       &
      x_start_viscous_hub,    x_start_viscous_case,             &
      x_start_hub_rotation,   x_end_hub_rotation,               &
      x_start_case_rotation,  x_end_case_rotation,              &
      num_sources,            blade_row_ksym,                   &
      cut_plane_dir,          num_cut_plane,                    &
      cut_plane_global_posi


    brn=bladerow
    SELECT CASE(brn)
      CASE(1:9)
        WRITE(fnam,*) 'input0' // ADJUSTL(TRIM(int_to_string(brn)))
      CASE(10:99)
        WRITE(fnam,*) 'input' // ADJUSTL(TRIM(int_to_string(brn)))
    END SELECT

!-----------------------------------------------------------------------
! Read the namelist to get the omega for the bladerow
    blade_row_ksym = ksym
    num_blades = -1
    omega_bld = 0.0
    OPEN(UNIT=11,FILE=fnam,FORM='formatted')
    READ(11,NML=BLADE_ROW_PARAMETERS)
    CLOSE(11)

    omega=omega_bld
    omega = PI *diam*omega/vref_m/30.
    nblds = num_blades
    ksym = blade_row_ksym

  END SUBROUTINE read_rowinput

!#######################################################################
  SUBROUTINE read_bc()
!#######################################################################
  USE common_area, ONLY: cid, bcavail
  IMPLICIT NONE

  INTEGER :: i,j,id,gid,bc2,ctr
  INTEGER :: starti,startj,startk,endi,endj,endk
  CHARACTER*20 :: gnam
  REAL    :: bc

  INTEGER,ALLOCATABLE,DIMENSION(:,:) :: tmp1,tmp2
  REAL,ALLOCATABLE,DIMENSION(:) :: tmp3,tmp4

  ctr=0
  ALLOCATE(tmp1(1,8),tmp3(1),tmp2(1,8),tmp4(1))
  OPEN(UNIT=10,FILE='bc.in',FORM='formatted',STATUS='old',ERR=9002)
  id = 9999
! Hot swap and reallocate as each line is read
  DO WHILE(id.GT.0)
    READ(10,*) id,bc,starti,startj,startk,endi,endj,endk
    IF(id.EQ.cid)THEN
      DEALLOCATE(tmp2,tmp4)
      ctr=ctr+1
      tmp1(ctr,:)=(/id,INT(bc),starti,startj,startk,endi,endj,endk/)
      tmp3(ctr)=bc
      ALLOCATE(tmp2(ctr,8),tmp4(ctr))
      DO i=1,ctr,1
        tmp2(i,:) = tmp1(i,:)
        tmp4(i) = tmp3(i)
      ENDDO
      DEALLOCATE(tmp1,tmp3)
      ALLOCATE(tmp1(ctr+1,8),tmp3(ctr+1))
      DO i=1,ctr,1
        tmp1(i,:) = tmp2(i,:)
        tmp3(i) = tmp4(i)
      ENDDO
    ENDIF
  ENDDO
  ALLOCATE(turbobcs(ctr,8),turbobcgr(ctr))
  nturbobcs=ctr
! Assign values to the final form.  These need to be sorted by boundary condition
! so that all of the same type can be consecutive???
  DO i=1,ctr
    turbobcs(i,:)=tmp2(i,:)
    turbobcgr(i)=tmp4(i)
  ENDDO
  DEALLOCATE(tmp1,tmp2,tmp3,tmp4)
  DO WHILE(1)
    READ(10,*,END=9001,ERR=9001) gid,gnam
  ENDDO
9001 CONTINUE
  CLOSE(10)
  WRITE(*,*) '   bc.in read successfully!'
  bcavail = .TRUE.
9002 CONTINUE

  RETURN

  END SUBROUTINE read_bc

!#######################################################################
  SUBROUTINE read_ref_nml
!#######################################################################
    USE common_area, ONLY: pref_m, tref_m, vref_m, diam, gammabase, &
                           gamtable, tempgam, suthc1_m, suthc2_m,   &
                           visref_m, href_m, rgasref_m, cpref_m,    &
                           rhoref_m
    IMPLICIT NONE

    CHARACTER(LEN=40)  :: fname                ! namelist input file name
    LOGICAL            :: fname_exists         ! logical telling whether namelist input file exists
    LOGICAL            :: file_end = .false.

    REAL(KIND=8)       :: ref_pressure
    REAL(KIND=8)       :: ref_temperature
    REAL(KIND=8)       :: ref_velocity
    REAL(KIND=8)       :: ref_length
    REAL(KIND=8)       :: ref_gamma
    REAL(KIND=8)       :: gamma_table(2)
    REAL(KIND=8)       :: temp_gam_table(2)
    REAL(KIND=8)       :: sutherland_const1
    REAL(KIND=8)       :: sutherland_const2
    REAL(KIND=8)       :: gamref_t1

    NAMELIST/REFERENCE_CONDITIONS/  &
    ref_pressure,            ref_temperature, &
    ref_velocity,            ref_gamma, &
    ref_length,              gamref_t1, &
    gamma_table,             temp_gam_table, &
    sutherland_const1,       sutherland_const2

    fname='input00'
    ref_pressure = pref_m
    ref_temperature = tref_m
    ref_velocity = vref_m
    ref_gamma = gammabase
    ref_length = diam
    gamma_table = gamtable
    temp_gam_table = tempgam
    sutherland_const1 = suthc1_m
    sutherland_const2 = suthc2_m

    INQUIRE(FILE=fname,EXIST=fname_exists)
    IF(.NOT. fname_exists)THEN
      WRITE(*,*) 'File input00 not found, using default reference quantities'
      RETURN
    ENDIF
    OPEN(UNIT=13,FILE=fname,FORM='formatted')
    REWIND(13)

    DO WHILE (.not. file_end)
      READ(13,NML=REFERENCE_CONDITIONS,ERR=2001,END=2003)
      CLOSE(13)

      pref_m = ref_pressure 
      tref_m = ref_temperature 
      vref_m = ref_velocity
      diam = ref_length
      gammabase = ref_gamma 
      gamtable = gamma_table 
      tempgam = temp_gam_table
      suthc1_m = sutherland_const1 
      suthc2_m = sutherland_const2 

      rgasref_m = vref_m**2/(gammabase*tref_m)
      rhoref_m = pref_m/vref_m**2
      visref_m = suthc1_m*tref_m**1.5/(tref_m+suthc2_m)
      href_m = (gammabase*rgasref_m)/(gammabase-1.)*tref_m
      cpref_m = (gammabase*rgasref_m)/(gammabase-1.)
      
      EXIT

2001  CONTINUE
    ENDDO

    RETURN

2003 CLOSE(13)
    WRITE(*,*) ' Did not find Reference Conditions'
    RETURN

  END SUBROUTINE read_ref_nml

!#######################################################################
  SUBROUTINE read_tacoma_p3d_bc(filnm)
!#######################################################################
    USE common_area, ONLY: bcavail, pgr, igrid, nclients, cid
    USE general, ONLY: MakeUpper, KeywordFunc
    IMPLICIT NONE

    CHARACTER(LEN=NSTRLEN),INTENT(IN) :: filnm

    INTEGER :: i, j, iblck, ibld, isurf, idummy, ibp
    INTEGER :: bp, cntr, linenumber, mylength
    INTEGER :: ib, npatch
    INTEGER :: int1, int2, int3
    REAL :: ribld
    CHARACTER(LEN=80) :: strin, inter

    OPEN(UNIT=BCSUNIT,FILE=filnm,FORM='formatted')

    linenumber=0
10  READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
    linenumber=linenumber+1

    strin=TRIM(ADJUSTL(strin))
    IF(strin.EQ.'# Topology Data')THEN
      READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
      READ(BCSUNIT,*,ERR=7000,END=9000) ribld
      linenumber=linenumber+2
    ELSE
      goto 10
    ENDIF

    ibld=FLOOR(ribld)

    pgr(igrid)%nbld = ibld

20  READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
    linenumber=linenumber+1

    strin=TRIM(ADJUSTL(strin))

    IF(strin.EQ.'# Block names / shafts')THEN
      READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
      READ(BCSUNIT,*,ERR=7000,END=9000) iblck
      linenumber=linenumber+2
    ELSE
      goto 20
    ENDIF

30  READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
    linenumber=linenumber+1

    IF(strin.EQ.'# Surface/Group definitions')THEN
      READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
      READ(BCSUNIT,"(A80)",ERR=7000,END=9000) strin
      READ(BCSUNIT,*,ERR=7000,END=9000) isurf
      linenumber=linenumber+3
    ELSE
      goto 30
    ENDIF

    ALLOCATE(turbobcplot(isurf), turbobcnam(isurf))
    bp=0
    inter='INTERFACE'
    DO i=1,isurf
      turbobcplot(i) = .TRUE.
      READ(BCSUNIT,"(I2,A80)",ERR=7000,END=9000) idummy, strin
      linenumber=linenumber+1
  
      strin=TRIM(ADJUSTL(strin))
      mylength = LEN_TRIM(strin)
      CALL MakeUpper(strin,mylength)
      IF(KeywordFunc(strin(1:20),inter(1:20)).GT.0)THEN
        bp=bp+1
        turbobcplot(i) = .FALSE.
      ENDIF
      turbobcnam(i)=strin
    ENDDO
    nturbobcs = isurf
    nturbobcs2 = isurf - bp

    ALLOCATE(patches(iblck))
    ib=-1
    DO WHILE(ib.NE.0)
7100  CONTINUE
      READ(BCSUNIT,*,ERR=7100,END=9000) ib, npatch
      IF(ib.EQ.0)EXIT
      ALLOCATE(patches(ib)%ptch(npatch,10))
      patches(ib)%nptch = npatch
      DO i=1,npatch
7200    CONTINUE
        READ(BCSUNIT,*,ERR=7200,END=9000) int1, int2, int3
        patches(ib)%ptch(i,1)=int1
        patches(ib)%ptch(i,2)=int2
        patches(ib)%ptch(i,3)=int3
        READ(BCSUNIT,*,ERR=7300,END=9000) int1, int2, int3
        patches(ib)%ptch(i,4)=int1
        patches(ib)%ptch(i,5)=int2
        patches(ib)%ptch(i,6)=int3
        READ(BCSUNIT,*,ERR=7300,END=9000) int1, int2
        patches(ib)%ptch(i,7)=int1
        patches(ib)%ptch(i,8)=int2
        READ(BCSUNIT,"(I4,I5)",ERR=7300,END=9000) int1, int2
        patches(ib)%ptch(i,9)=int1
        patches(ib)%ptch(i,10)=int2
        CYCLE
7300    CONTINUE
      ENDDO
    ENDDO
    bcavail = .TRUE.

!    IF(nclients.GT.1)THEN
!      DO i = 1, isurf
!        IF(.NOT. turbobcplot(i))CYCLE
!        DO j = 1, patches(cid)%nptch
!          IF(patches(cid)%ptch(j,7).NE.i)THEN
!            turbobcplot(i)=.FALSE.
!          ELSE
!            turbobcplot(i)=.TRUE.
!            EXIT
!          ENDIF
!        ENDDO
!      ENDDO
!    ENDIF

    RETURN

7000 CONTINUE
     WRITE(*,*) '  ERROR reading from bc file '
9000 CONTINUE
     WRITE(*,*) '  EOF reading from bc file '
     CLOSE(BCSUNIT)

  END SUBROUTINE read_tacoma_p3d_bc

!#######################################################################
  SUBROUTINE TacomaPVSurface(nsurf,scon,scel,tsurf)
!#######################################################################
    USE general, ONLY: KeywordFunc
    USE common_area, ONLY: bcavail, pgr, igrid, cid, nclients
    IMPLICIT NONE

    INTEGER :: ijk2nod, ijk2nodm, ijk2cell, ijk2cellm
!
!	routine that passes the surface info to pV3
!
!	where:	nsurf	- pointers to the end of each surface group and
!			  flag to indicate if surface is on
!		scon	- connection data for multi-cilent cases
!		scel    - node numbers for surfaces
!		tsurf	- surface group title (optional)
!
    INTEGER :: nsurf(3,*),scon(*),scel(4,*),ksurf,i,j,k, i1,j1,k1, s, ctr
    INTEGER :: il,jl,kl,m,mle,mte,mtp,mhb,sur,t, ii,jj,kk
    INTEGER :: j1min, j1max, sconval, nsurfval
    INTEGER :: nbcg, nbcc, currbc, ngpts, nbcs, nbcx, nb, nx, is, ie, js, je, ks, ke
    CHARACTER*20 :: tmpstr1,tmpstr2,tmpstr3,tmpstr4,tmpstr5
    CHARACTER*20 :: tsurf(*)
    CHARACTER*80 :: tmpstr
    LOGICAL :: surfasgn

      IF(.NOT. bcavail)RETURN
      WRITE(*,*) ' Generating boundaries...'
      tmpstr1='periodic'
      tmpstr2='cas'
      tmpstr3='inlet'
      tmpstr4='exit'
      tmpstr5='interf'
      ksurf = 0
      ctr = 0
      SELECT CASE(nclients)
        CASE(1)
          j1min = 1
          j1max = pgr(igrid)%nblks
        CASE DEFAULT
          j1min = cid
          j1max = cid 
      END SELECT
      DO i1 = 1, nturbobcs
        IF(.NOT. turbobcplot(i1)) CYCLE
        SELECT CASE(nclients)
          CASE(1)
            ctr = ctr + 1
          CASE DEFAULT
            DO j1 = j1min, j1max
              DO k1 = 1, patches(j1)%nptch
                IF(patches(j1)%ptch(k1,7).EQ.i1)THEN
                  ctr = ctr + 1
                  GOTO 1001
                ENDIF
              ENDDO
            ENDDO
            CYCLE
        END SELECT
!---/
1001    CONTINUE
        s = 0
        DO j1 = j1min, j1max
          SELECT CASE(nclients)
            CASE(1)
              il = pgr(igrid)%li(j1)
              jl = pgr(igrid)%lj(j1)
              kl = pgr(igrid)%lk(j1)
            CASE DEFAULT
              il = pgr(igrid)%li(1)
              jl = pgr(igrid)%lj(1)
              kl = pgr(igrid)%lk(1)
          END SELECT

          DO k1 = 1, patches(j1)%nptch
            SELECT CASE(nclients)
              CASE(1)
                nsurfval = i1
              CASE DEFAULT
                nsurfval = -patches(j1)%ptch(k1,9)
            END SELECT
            IF(patches(j1)%ptch(k1,7).EQ.i1)THEN

              IF(patches(j1)%ptch(k1,1).EQ.patches(j1)%ptch(k1,4))THEN
                i = patches(j1)%ptch(k1,1)
                DO j = patches(j1)%ptch(k1,2), patches(j1)%ptch(k1,5)-1
                  DO k = patches(j1)%ptch(k1,3), patches(j1)%ptch(k1,6)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i, j+1, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i, j+1, k+1, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j, k+1, il, jl) + s
                    scon(ksurf) = 0
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
! This set of code is useful for a periodic face, which in Tacoma
! has patch data on both sides... What to do, what to do?
!                    SELECT CASE(nclients)
!                      CASE(1)
!                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
!                          CALL IJKFaceMatching(i, j, k, j1, k1, ii,jj,kk)
!                          scon(ksurf) = ijk2cellm(ii,jj,kk,patches(j1)%ptch(k1,9))
!                        ELSE
!                          scon(ksurf) = 0
!                        ENDIF
!                      CASE DEFAULT
!                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
!                          scon(ksurf) = -1
!                        ELSE
!                          scon(ksurf) = 0
!                        ENDIF
!                    END SELECT
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                  ENDDO
                ENDDO
              ELSEIF(patches(j1)%ptch(k1,2).EQ.patches(j1)%ptch(k1,5))THEN
                j = patches(j1)%ptch(k1,2)
                DO i = patches(j1)%ptch(k1,1), patches(j1)%ptch(k1,4)-1
                  DO k = patches(j1)%ptch(k1,3), patches(j1)%ptch(k1,6)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i+1, j, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i+1, j, k+1, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j, k+1, il, jl) + s
                    scon(ksurf) = 0
                  ENDDO
                ENDDO
              ELSEIF(patches(j1)%ptch(k1,3).EQ.patches(j1)%ptch(k1,6))THEN
                k = patches(j1)%ptch(k1,3)
                DO i = patches(j1)%ptch(k1,1), patches(j1)%ptch(k1,4)-1
                  DO j = patches(j1)%ptch(k1,2), patches(j1)%ptch(k1,5)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i+1, j, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i+1, j+1, k, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j+1, k, il, jl) + s
                    scon(ksurf) = 0
                  ENDDO
                ENDDO
              ENDIF
              nsurf(1,ctr) = ksurf
              nsurf(2,ctr) = 0 + 256
              IF(KeywordFunc(turbobcnam(i1),tmpstr1).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr2).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr2).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr4).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr5).LT.0) nsurf(2,ctr) = 1 + nsurf(2,ctr)
              IF(patches(j1)%ptch(k1,9).NE.0)THEN
                nsurf(3,ctr) = nsurfval
              ELSE
                nsurf(3,ctr) = i1
              ENDIF
              tsurf(ctr) = turbobcnam(i1)
            ENDIF
          ENDDO
          s = s + pgr(igrid)%mvmax(j1)
        ENDDO
      ENDDO
!=======================================================================
! Collect all interfaces into a single surface
      WRITE(*,*) ' Generating internal interfaces...'
      ctr = ctr + 1
      DO i1 = 1, nturbobcs
        IF(turbobcplot(i1)) CYCLE
        DO j1 = j1min, j1max
          DO k1 = 1, patches(j1)%nptch
            IF(patches(j1)%ptch(k1,7).EQ.i1)GOTO 1002
          ENDDO
        ENDDO
        CYCLE
!---/
1002    CONTINUE
        s = 0
        DO j1 = j1min, j1max
          SELECT CASE(nclients)
            CASE(1)
              il = pgr(igrid)%li(j1)
              jl = pgr(igrid)%lj(j1)
              kl = pgr(igrid)%lk(j1)
            CASE DEFAULT
              il = pgr(igrid)%li(1)
              jl = pgr(igrid)%lj(1)
              kl = pgr(igrid)%lk(1)
          END SELECT

          DO k1 = 1, patches(j1)%nptch
            SELECT CASE(nclients)
              CASE(1)
                nsurfval = i1
              CASE DEFAULT
                nsurfval = -patches(j1)%ptch(k1,9)
            END SELECT
            IF(patches(j1)%ptch(k1,7).EQ.i1)THEN

              IF(patches(j1)%ptch(k1,1).EQ.patches(j1)%ptch(k1,4))THEN
                i = patches(j1)%ptch(k1,1)
                DO j = patches(j1)%ptch(k1,2), patches(j1)%ptch(k1,5)-1
                  DO k = patches(j1)%ptch(k1,3), patches(j1)%ptch(k1,6)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i, j+1, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i, j+1, k+1, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j, k+1, il, jl) + s
#ifndef PV3SERIAL
                    SELECT CASE(nclients)
                      CASE(1)
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          CALL IJKFaceMatching(i, j, k, j1, k1, ii,jj,kk)
                          scon(ksurf) = ijk2cellm(ii,jj,kk,patches(j1)%ptch(k1,9))
                        ENDIF
                      CASE DEFAULT
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          scon(ksurf) = -1
                        ENDIF
                    END SELECT
#endif
                  ENDDO
                ENDDO
              ELSEIF(patches(j1)%ptch(k1,2).EQ.patches(j1)%ptch(k1,5))THEN
                j = patches(j1)%ptch(k1,2)
                DO i = patches(j1)%ptch(k1,1), patches(j1)%ptch(k1,4)-1
                  DO k = patches(j1)%ptch(k1,3), patches(j1)%ptch(k1,6)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i+1, j, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i+1, j, k+1, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j, k+1, il, jl) + s
#ifndef PV3SERIAL
                    SELECT CASE(nclients)
                      CASE(1)
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          CALL IJKFaceMatching(i, j, k, j1, k1, ii,jj,kk)
                          scon(ksurf) = ijk2cellm(ii,jj,kk,patches(j1)%ptch(k1,9))
                        ENDIF
                      CASE DEFAULT
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          scon(ksurf) = -1
                        ENDIF
                    END SELECT
#endif
                  ENDDO
                ENDDO
              ELSEIF(patches(j1)%ptch(k1,3).EQ.patches(j1)%ptch(k1,6))THEN
                k = patches(j1)%ptch(k1,3)
                DO i = patches(j1)%ptch(k1,1), patches(j1)%ptch(k1,4)-1
                  DO j = patches(j1)%ptch(k1,2), patches(j1)%ptch(k1,5)-1
                    ksurf = ksurf + 1
                    scel(1,ksurf) = ijk2nod(i, j, k, il, jl) + s
                    scel(2,ksurf) = ijk2nod(i+1, j, k, il, jl) + s
                    scel(3,ksurf) = ijk2nod(i+1, j+1, k, il, jl) + s
                    scel(4,ksurf) = ijk2nod(i, j+1, k, il, jl) + s
#ifndef PV3SERIAL
                    SELECT CASE(nclients)
                      CASE(1)
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          CALL IJKFaceMatching(i, j, k, j1, k1, ii,jj,kk)
                          scon(ksurf) = ijk2cellm(ii,jj,kk,patches(j1)%ptch(k1,9))
                        ENDIF
                      CASE DEFAULT
                        IF(patches(j1)%ptch(k1,9).NE.0)THEN
                          scon(ksurf) = -1
                        ENDIF
                    END SELECT
#endif
                  ENDDO
                ENDDO
              ENDIF
              nsurf(1,ctr) = ksurf
              nsurf(2,ctr) = 0 + 256
              IF(KeywordFunc(turbobcnam(i1),tmpstr1).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr2).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr2).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr4).LT.0 .AND. &
                 KeywordFunc(turbobcnam(i1),tmpstr5).LT.0) nsurf(2,ctr) = 1 + nsurf(2,ctr)
              IF(patches(j1)%ptch(k1,9).NE.0)THEN
                nsurf(3,ctr) = nsurfval
              ELSE
                nsurf(3,ctr) = i1
PRINT*,' NSURF Not Matching where Im supposed to and this really is a problem...'
              ENDIF
              tsurf(ctr) = 'INTERFACES'
            ENDIF
          ENDDO
          s = s + pgr(igrid)%mvmax(j1)
        ENDDO
      ENDDO

  END SUBROUTINE TacomaPVSurface

END MODULE solver_turbo
