#include "parameters.h"
!#######################################################################
CHARACTER*20 FUNCTION GetTecName(fwctr,pref,suff)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: fwctr
  CHARACTER(LEN=3), INTENT(IN) :: pref, suff

  IF(fwctr.GT.999)THEN
    WRITE(GetTecName,"(2A,I4.4,2A)") pref,'out',fwctr,'.',suff
  ELSE
    WRITE(GetTecName,"(2A,I3.3,2A)") pref,'out',fwctr,'.',suff
  ENDIF

END FUNCTION GetTecName

!#######################################################################
SUBROUTINE WriteDataHeader(fwctr)
!#######################################################################
! This routine can be called from a V3 app because there are no
! SetState invocations.
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: fwctr
# ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: ivec(11)
# else
    INTEGER :: ivec(11)
# endif
  INTEGER :: i,j
  REAL :: rvec(12)
  CHARACTER*20  :: nam, GetTecName
  CHARACTER*32  :: tstr

! Open files and write header
!  OPEN(UNIT=TECUNIT,FILE=GetTecName(fwctr,'tec','dat'),FORM='formatted')
!  WRITE(TECUNIT,*) 'TITLE = "Tecplot Output"'
!  WRITE(TECUNIT,*) 'VARIABLES = "X","Y","Z","S","Vx","Vy","Vz"'
  OPEN(UNIT=POVUNIT,FILE=GetTecName(fwctr,'pov','pvb'),FORM='unformatted')

! Write POV header for current version
  WRITE(POVUNIT) 1000
  WRITE(POVUNIT) 1,0

! Write POV header for general data
  WRITE(POVUNIT) 4200
  CALL GETSTATE(-2, ivec, rvec, nam)
  WRITE(POVUNIT) rvec(1)

! Write POV header for camera location 4201
  WRITE(POVUNIT) 4201
  ivec(1) = 0
  CALL GETSTATE( 1, ivec, rvec, nam)
  WRITE(POVUNIT) (rvec(i), i=1,12),0.0,0.0,0.0,1.0

! Write POV header for lights
  WRITE(POVUNIT) 4202
  CALL GETSTATE( 4, ivec, rvec, nam)
# ifdef PV3SERIAL
    WRITE(POVUNIT) 1
    WRITE(POVUNIT) (rvec(j), j = 1, 6)
# else
    WRITE(POVUNIT) ivec(2)
    DO i=1,ivec(2)
      CALL GETSTATE(9, i, rvec, nam)
      WRITE(POVUNIT) (rvec(j), j = 4, 6), (rvec(j), j = 1, 3)
    ENDDO
# endif

! Write POV header for the current displayed scalars
  WRITE(POVUNIT) 4203
  CALL GETSTATE( 5, ivec, rvec, nam)
  WRITE(POVUNIT) (rvec(i), i = 1, 2)

! Write POV header for the radius/width of streamtubes/ribbons/bubbles
  WRITE(POVUNIT) 4204
  CALL GETSTATE(11, ivec, rvec, nam)
  WRITE(POVUNIT) rvec(2)

! Write POV header for the threshold values
  WRITE(POVUNIT) 4205
  CALL GETSTATE( 7, ivec, rvec, nam)
  WRITE(POVUNIT) (rvec(i), i = 1, 2)

! Write POV header for the surface start
  WRITE(POVUNIT) 4242

END SUBROUTINE WriteDataHeader

#ifndef PV3SERIAL

!#######################################################################
SUBROUTINE GetExtractData()
!#######################################################################
  USE keydata, ONLY: fwctr
  IMPLICIT NONE

  INTEGER*BSIZE :: ex
  INTEGER*BSIZE :: nextex
  INTEGER :: ivec(11), i, j
  INTEGER :: typ,num,numcs

  REAL :: rvec(12)

  CHARACTER*20  :: nam, GetTecName


  ivec(1) = 1
  CALL pV_GetState(-3, ivec, rvec, nam)
  numcs = ivec(4)

  fwctr = fwctr + 1
  WRITE(*,"(A,I4,A)") '*** Generating POV/TEC ',fwctr,' Output ***'

  CALL WriteDataHeader(fwctr)

! Get extracts
  ex = 0
  CALL pV_GetExtract(ex,typ,num,ivec,rvec,nam,nextex)
  WRITE(*,*) nam,typ,num, ivec(1)
  CALL CheckExtractData(ex,typ,nam,numcs,ivec)

  DO WHILE(nextex.NE.0)
    ex = nextex
    CALL pV_GetExtract(ex,typ,num,ivec,rvec,nam,nextex)
    WRITE(*,*) nam,typ,num, ivec(1)
    CALL CheckExtractData(ex,typ,nam,numcs,ivec)
  ENDDO

!  CLOSE(TECUNIT)
  CLOSE(POVUNIT)

  WRITE(*,"(A,I4,A)") '*** Generation ',fwctr,' Complete ***'

END SUBROUTINE GetExtractData

!#######################################################################
SUBROUTINE CheckExtractData(ex,typ,nam,numcs,geivec)
!#######################################################################
  USE general, ONLY: int_to_string, BitsOn
  USE mod_clear, ONLY: ClearVar
  USE server_common, ONLY: povteccombined
  IMPLICIT NONE

  INTEGER*BSIZE, INTENT(IN) :: ex
  INTEGER, INTENT(IN) :: typ, numcs, geivec(11)
  CHARACTER*20, INTENT(IN) :: nam

  INTEGER :: ptr0(1)    ! Sub-extract 0   tris
  INTEGER :: ptr1(1)    ! Sub-extract 1   quads
  INTEGER :: ptr3(1)    ! Sub-extract 3   mesh
  INTEGER :: ptr4(1)    ! Sub-extract 4   outline
  INTEGER :: ptr10s(1)  ! Sub-extract 10* planar clipping

  INTEGER :: pptr0(1)   ! Particle Sub-extract 0  number
  INTEGER :: pptr10(1)  ! Particle Sub-extract 10 group index

  REAL    :: ptr2(1)    ! Sub-extract 2   xyz
  REAL    :: ptr5(1)    ! Sub-extract 5   scalar
  REAL    :: ptr6(1)    ! Sub-extract 6   vector
  REAL    :: ptr7(1)    ! Sub-extract 7   threshold
  REAL    :: ptr8(1)    ! Sub-extract 8   2d mapping
  REAL    :: ptr9(1)    ! Sub-extract 9   transformed 2d
  REAL    :: ptr10(1)   ! Sub-extract 10  tri normals
  REAL    :: ptr11(1)   ! Sub-extract 11  quad normals

  REAL    :: pptr1(1)   ! Particle   Sub-extract 1  time
  REAL    :: pptr2(1)   ! Particle   Sub-extract 2  xyz
  REAL    :: pptr3(1)   ! Particle   Sub-extract 3  div
  REAL    :: pptr4(1)   ! Streamline Sub-extract 4  angle
  REAL    :: pptr5(1)   ! Particle   Sub-extract 5  scalar
  REAL    :: pptr6(1)   ! Particle   Sub-extract 6  scalar
  REAL    :: pptr7(1)   ! Particle   Sub-extract 7  scalar


  INTEGER :: slen0, slen1, slen2, slen3, slen4, slen5, slen6, &
             slen7, slen8, slen9, slen10, slen11, slen10s

  INTEGER :: ctr0, ctr1, ctr2, ctr3, ctr4, ctr5, ctr6, &
             ctr7, ctr8, ctr9, ctr10, ctr11, ctr10s

  POINTER (pntr0, ptr0), (pntr1, ptr1), (pntr2, ptr2), (pntr3, ptr3), &
          (pntr4, ptr4), (pntr5, ptr5), (pntr6, ptr6), (pntr7, ptr7), &
          (pntr8, ptr8), (pntr9, ptr9), (pntr10, ptr10),              &
          (pntr11, ptr11), (pntr10s, ptr10s),                         &
          (ppntr0, pptr0), (ppntr1, pptr1), (ppntr2, pptr2),          &
          (ppntr3, pptr3), (ppntr5, pptr5), (ppntr6, pptr6),          &
          (ppntr7, pptr7), (ppntr10, pptr10),                         &
          (ppntr4, pptr4)

  INTEGER :: istat, ecid
  INTEGER :: pV_GetSub
  INTEGER :: i, j, n
  INTEGER :: ivec(6)

  REAL :: rvec(12)
  REAL, ALLOCATABLE, DIMENSION(:) :: scals, angls, tymes

  CHARACTER*20 strng

  LOGICAL :: lptr(2,numcs)
  LOGICAL,POINTER,DIMENSION(:) :: lbits

  slen0=0
  slen1=0
  slen2=0
  slen3=0
  slen4=0
  slen5=0
  slen6=0
  slen7=0
  slen8=0
  slen9=0
  slen10=0
  slen11=0
  slen10s=0

  ctr0=0
  ctr1=0
  ctr2=0
  ctr3=0
  ctr4=0
  ctr5=0
  ctr6=0
  ctr7=0
  ctr8=0
  ctr9=0
  ctr10=0
  ctr11=0
  ctr10s=0

  ivec  = 0
  rvec  = 0
  strng = ' '

  CALL BitsOn( lbits, 16, geivec(1) )
  IF(lbits(0))THEN
    SELECT CASE(typ)
      CASE( EXT_SURF_PLANAR, EXT_SURF_BLOCK, EXT_SURF_GEOMETRIC, &
&           EXT_SURF_DOMAIN, EXT_SURF_ISO ) 

        IF(povteccombined)THEN
#         include "povteccombined.f90"
        ELSE
#         include "povtecindiv.f90"
        ENDIF

      CASE( EXT_STREAMLINE ) 
#       include "povtecstream.f90"

      CASE( EXT_PARTICLES ) 
#       include "povtecbubbles.f90"

      CASE( EXT_VCLOUDS ) 
      CASE( EXT_POINTS ) 

    END SELECT
  ENDIF
  CALL ClearVar(lbits)

END SUBROUTINE CheckExtractData

#endif
