#include "parameters.h"
!#######################################################################
PROGRAM pV3Client
!#######################################################################
  USE typedefs
  USE readgrid, ONLY: Get_Grids
  USE readflow, ONLY: Get_Flow
  USE readfunc, ONLY: Get_Func
  USE readbcs
  USE keydata
  USE common_area
  USE iter, ONLY: primary
  IMPLICIT NONE

  INTEGER nargs,i,iargc
  CHARACTER*20 :: comarg
  LOGICAL :: ios

#ifdef PV3SERIAL
  INTEGER :: cfilopt
  INTEGER :: knode, kequiv, kcel1, kcel2, kcel3, kcel4, knptet, kptet
  INTEGER :: knblck, ksurf, knsurf, khedra, hint
  INTEGER, ALLOCATABLE, DIMENSION(:,:) :: blocks
  INTEGER :: cmunit = 42
  LOGICAL :: win3d = .TRUE.
  CHARACTER*80 :: cmfile = 'spec.col'
#endif

!-----------------------------------------------------------------------
  itime = 1    ! Start the timing counter
  timedisp = 0 ! Set default to time-step display
  igrid = 1    ! Start the griding counter

  WRITE(*,*) '##########################################################'
  WRITE(*,*) ttl
  CALL PrintTimeStamp
  CALL PrintPV3Version
  WRITE(*,*) '##########################################################'

!-----------------------------------------------------------------------
! Command Line Arguments
  cid = 0 ! This is essential for checking what the client number should
          ! be.  Later it will be overwritten, or the user will have to
          ! provide the value via stdin.
  nargs=iargc()+1
  usrinput='pv3.in'
  SELECT CASE(nargs)
    CASE(2)
      CALL getarg(1,comarg)
      CALL Str2Int(comarg,cid)
    CASE(3)
      CALL getarg(1,comarg)
      CALL Str2Int(comarg,cid)
      CALL getarg(2,comarg)
      usrinput=comarg
  END SELECT

!-----------------------------------------------------------------------
  CALL Get_PV3_Input
  CALL Get_Grids
  CALL Get_Solver_Input
  CALL Get_Flow
  CALL Get_Func

  CALL Set_FX_Arrays

!-----------------------------------------------------------------------
  CALL Replication
  CALL GetNblocks
  CALL SetIopt
  CALL SetFLims
  CALL SetNames

  CALL KeyAllocInit   ! Init top-level key allocation flags

  CALL PreGrid

  CALL PreCalcVars

!-----------------------------------------------------------------------
#ifdef PV3SERIAL
  primary = .TRUE.
  cfilopt=1
  INQUIRE(FILE=cmfile,EXIST=ios)
  IF(.NOT.ios)CALL WriteColFile(cfilopt)
  mirror = 0
  ALLOCATE(blocks(3,maxblk))
  CALL pVStruc(knode, kequiv, kcel1, kcel2, kcel3, kcel4, &
                     knptet, kptet, knblck, blocks, khedra, &
                     ksurf, knsurf, hint)
  CALL V3_Init(ttl, iopt, cmfile, cmunit, win3d, nkeys, ikeys, tkeys, &
               keytyp, flims, mirror, knode, kequiv, kcel1, kcel2, &
               kcel3, kcel4, knptet, kptet, knblck, blocks, ksurf, &
               knsurf)
#else
1000 CONTINUE ! This is for restarting the client as a pvm process
  IF(iopt.LT.1)PRINT*,' IOPT=',iopt
  CALL pV_Init(ttl, cid, cname, dname, iopt, ncuts, tcuts, &
          nkeys, ikeys, tkeys, keytyp, flims, mirror, rep, &
          maxblk, istat)
  CALL ClientEventLoop
  CALL pV_Termin
#endif
!-----------------------------------------------------------------------

  CALL ClearAllFlow
  STOP

END PROGRAM pV3Client

#ifdef PV3SERIAL

!#######################################################################
SUBROUTINE V3Update(timeo)
!#######################################################################
  USE common_area
  USE iter, ONLY: SetTime
  USE keydata, ONLY: myfps, timedisp, timesav, dt
  IMPLICIT NONE

  REAL, INTENT(OUT) :: timeo

  CALL SetTime
  SELECT CASE(timedisp)
    CASE(0)
      timeo = FLOAT(itime)
    CASE(1)
      timeo = pfw(itime)%tyme
    CASE(2)
      timeo = timesav
      timesav = timesav + dt
  END SELECT

END SUBROUTINE V3Update

#else

!#######################################################################
SUBROUTINE ClientEventLoop()
!#######################################################################
  USE common_area
  USE readgrid, ONLY: ReloadGrid
  USE readflow, ONLY: ReloadFlow
  USE keydata, ONLY: timedisp, timesav, dt, keyfilename, RSTATE
  USE iter
  IMPLICIT NONE

  INTEGER :: istate, ccount, i, j
  REAL :: timeo
  LOGICAL :: serverrunning

  WRITE(*,*) 'istat from pV_Init = ',istat

  serverrunning = .FALSE.

  IF(istat.NE.0 .OR. iopt.EQ.0)STOP

  DO WHILE(1)
!=======================================================================
! Check server status
    CALL pV_Stat(istate)
    SELECT CASE(istate)
      CASE(0)
        !Client is not initialized
      CASE(1)
        IF(serverrunning)EXIT ! Server has closed
      CASE(2)
        !The server is active and we will continue
        IF (.NOT.serverrunning) THEN  ! setup timing master
            CALL pvmfmytid(i)
            CALL pvmfgetinst( 'pV3Server', i, j)
            IF (j.EQ.0) primary = .TRUE.
        ENDIF
        serverrunning = .TRUE.
      CASE DEFAULT
        !Error
    END SELECT

!=======================================================================
! Manage RSTATE values
      SELECT CASE(RSTATE)
        CASE(RSTATE_RELOAD_GRID)
          CALL ReloadGrid(keyfilename)
        CASE(RSTATE_RELOAD_FLOW)
          CALL ReloadFlow(keyfilename)
      END SELECT

!=======================================================================
! Set time step
    CALL SetTime

    SELECT CASE(timedisp)
      CASE(0)
        timeo = FLOAT(itime)
      CASE(1)
        timeo = pfw(itime)%tyme
      CASE(2)
        timeo = timesav
        timesav = timesav + dt
    END SELECT

!=======================================================================
! Execute the pV3 updates
    DEBUG1(' \ pV_Update')
    CALL pV_Update(timeo)
    IF(RSTATE.NE.RSTATE_READY) CALL pV_Update(timeo)
    DEBUG1(' / pV_Update')

  ENDDO

END SUBROUTINE ClientEventLoop

#endif
