!#######################################################################
SUBROUTINE pVDraw3D (istat)
!#######################################################################
  USE server_common, ONLY: cftoggle
  IMPLICIT NONE

  REAL,PARAMETER :: MAGN=.1
  REAL,PARAMETER :: START=-.9
  REAL,PARAMETER :: ZLOC=.9
  INTEGER,PARAMETER :: NFACE=6

  INTEGER, INTENT(IN) :: istat

  INTEGER(KIND=BSIZE) :: ivec(1)
  INTEGER :: np, i, j

  REAL :: rvec(12), xyzp(3,6)=0., bxyzp(3,NFACE*4)
  REAL :: radii(6)=0., bradii(NFACE*4)
  REAL :: col(6)=1., bcol(NFACE*4)=1.
  REAL :: xvec(3)=0., yvec(3)=0., zvec(3)=0., xmag=0., ymag=0., zmag=0.
  REAL :: cfrm(18)=(/ START,START,ZLOC, START,START,ZLOC, &
 &                    START,START,ZLOC, START,START+MAGN,ZLOC, &
 &                    START,START,ZLOC, START,START,ZLOC-MAGN/)
  REAL :: fsize=MAGN/4., fwidth=1., fxyz(3)=0.
  REAL :: fnorx(3)=0., fnory(3)=0., fcol(3) = 1.

  CHARACTER(LEN=3) :: str

  SELECT CASE(istat)
!=======================================================================
! Rendering of static objects
    CASE(0)
      IF(cftoggle)THEN
      CALL pV_GetState(1, ivec, rvec, str)

      xmag = 0.
      ymag = 0.
      zmag = 0.
      DO i = 1, 3
        xvec( i ) = rvec( ( i - 1 ) * 4 + 1 )
        xmag = xmag + xvec( i )**2
        yvec( i ) = rvec( ( i - 1 ) * 4 + 2 )
        ymag = ymag + yvec( i )**2
        zvec( i ) = rvec( ( i - 1 ) * 4 + 3 )
        zmag = zmag + zvec( i )**2
      ENDDO
      DO j = 1, 6
        DO i = 1, 3
          xyzp( i, j ) = cfrm( ( j - 1 ) * 3 + i )
        ENDDO
      ENDDO
      DO i = 1, 3
        xyzp(i,2) = xyzp(i,1) + MAGN*xvec(i)/SQRT(zmag)
        xyzp(i,4) = xyzp(i,3) + MAGN*yvec(i)/SQRT(zmag)
        xyzp(i,6) = xyzp(i,5) + MAGN*zvec(i)/SQRT(zmag)
      ENDDO
      np = 6
      CALL pV_Object3D(3,-1,xyzp,radii,col,-np)
      fnorx = (/0.,0.,0./)
      fnory = (/0.,0.,0./)
      str = 'XYZ'
      DO i = 1, 3
        DO j = 1, 3
          fxyz(j) = xyzp(j,i*2)
        ENDDO
        CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, str(i:i))
      ENDDO

      np = NFACE*4
      DO i = 1, 3
        xyzp(i,2) = xyzp(i,1) + MAGN*xvec(i)/(2.*SQRT(xmag))
        xyzp(i,4) = xyzp(i,3) + MAGN*yvec(i)/(2.*SQRT(ymag))
        xyzp(i,6) = xyzp(i,5) + MAGN*zvec(i)/(2.*SQRT(zmag))
      ENDDO
      xmag=0.
      ymag=0.
      zmag=0.
      DO i = 1, 3
        xvec( i ) = rvec((i-1)*4+1) + rvec((i-1)*4+2)
        xmag = xmag + xvec( i )**2
        yvec( i ) = rvec((i-1)*4+2) + rvec((i-1)*4+3)
        ymag = ymag + yvec( i )**2
        zvec( i ) = rvec((i-1)*4+1) + rvec((i-1)*4+3)
        zmag = zmag + zvec( i )**2
      ENDDO
    ! X-Y Quad
      bxyzp(1,1) = xyzp(1,1)
      bxyzp(2,1) = xyzp(2,1)
      bxyzp(3,1) = xyzp(3,1)
      bxyzp(1,2) = xyzp(1,2)
      bxyzp(2,2) = xyzp(2,2)
      bxyzp(3,2) = xyzp(3,2)
      bxyzp(1,3) = xyzp(1,1) + MAGN*xvec(1)/(SQRT(2.*xmag))
      bxyzp(2,3) = xyzp(2,1) + MAGN*xvec(2)/(SQRT(2.*xmag))
      bxyzp(3,3) = xyzp(3,1) + MAGN*xvec(3)/(SQRT(2.*xmag))
      bxyzp(1,4) = xyzp(1,4)
      bxyzp(2,4) = xyzp(2,4)
      bxyzp(3,4) = xyzp(3,4)
    ! Y-Z Quad
      bxyzp(1,5) = xyzp(1,1)
      bxyzp(2,5) = xyzp(2,1)
      bxyzp(3,5) = xyzp(3,1)
      bxyzp(1,6) = xyzp(1,4)
      bxyzp(2,6) = xyzp(2,4)
      bxyzp(3,6) = xyzp(3,4)
      bxyzp(1,7) = xyzp(1,1) + MAGN*yvec(1)/(SQRT(2.*ymag))
      bxyzp(2,7) = xyzp(2,1) + MAGN*yvec(2)/(SQRT(2.*ymag))
      bxyzp(3,7) = xyzp(3,1) + MAGN*yvec(3)/(SQRT(2.*ymag))
      bxyzp(1,8) = xyzp(1,6)
      bxyzp(2,8) = xyzp(2,6)
      bxyzp(3,8) = xyzp(3,6)
    ! X-Z Quad
      bxyzp(1,9) = xyzp(1,1)
      bxyzp(2,9) = xyzp(2,1)
      bxyzp(3,9) = xyzp(3,1)
      bxyzp(1,10) = xyzp(1,6)
      bxyzp(2,10) = xyzp(2,6)
      bxyzp(3,10) = xyzp(3,6)
      bxyzp(1,11) = xyzp(1,1) + MAGN*zvec(1)/(SQRT(2.*zmag))
      bxyzp(2,11) = xyzp(2,1) + MAGN*zvec(2)/(SQRT(2.*zmag))
      bxyzp(3,11) = xyzp(3,1) + MAGN*zvec(3)/(SQRT(2.*zmag))
      bxyzp(1,12) = xyzp(1,2)
      bxyzp(2,12) = xyzp(2,2)
      bxyzp(3,12) = xyzp(3,2)

      xmag=0.
      DO i = 1, 3
        xvec( i ) = rvec((i-1)*4+1) + rvec((i-1)*4+2) + rvec((i-1)*4+3)
        xmag = xmag + xvec( i )**2
      ENDDO
    ! X-Y : Z Quad
      bxyzp(1,13) = bxyzp(1,8)
      bxyzp(2,13) = bxyzp(2,8)
      bxyzp(3,13) = bxyzp(3,8)
      bxyzp(1,14) = bxyzp(1,7)
      bxyzp(2,14) = bxyzp(2,7)
      bxyzp(3,14) = bxyzp(3,7)
      bxyzp(1,15) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
      bxyzp(2,15) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
      bxyzp(3,15) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
      bxyzp(1,16) = bxyzp(1,11)
      bxyzp(2,16) = bxyzp(2,11)
      bxyzp(3,16) = bxyzp(3,11)
    ! Y-Z : X Quad
      bxyzp(1,17) = bxyzp(1,2)
      bxyzp(2,17) = bxyzp(2,2)
      bxyzp(3,17) = bxyzp(3,2)
      bxyzp(1,18) = bxyzp(1,11)
      bxyzp(2,18) = bxyzp(2,11)
      bxyzp(3,18) = bxyzp(3,11)
      bxyzp(1,19) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
      bxyzp(2,19) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
      bxyzp(3,19) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
      bxyzp(1,20) = bxyzp(1,3)
      bxyzp(2,20) = bxyzp(2,3)
      bxyzp(3,20) = bxyzp(3,3)
    ! X-Z : Y Quad
      bxyzp(1,21) = bxyzp(1,4)
      bxyzp(2,21) = bxyzp(2,4)
      bxyzp(3,21) = bxyzp(3,4)
      bxyzp(1,22) = bxyzp(1,3)
      bxyzp(2,22) = bxyzp(2,3)
      bxyzp(3,22) = bxyzp(3,3)
      bxyzp(1,23) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
      bxyzp(2,23) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
      bxyzp(3,23) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
      bxyzp(1,24) = bxyzp(1,7)
      bxyzp(2,24) = bxyzp(2,7)
      bxyzp(3,24) = bxyzp(3,7)
      CALL pV_Object3D(0,0,bxyzp,bradii,bcol,-np)
      ENDIF

      RETURN

!=======================================================================
! Rendering of dynamic objects
    CASE(1)
      RETURN

!=======================================================================
! Rendering of translucent objects
    CASE(2)
      RETURN

  END SELECT

  RETURN

END SUBROUTINE pVDraw3D

!#######################################################################
SUBROUTINE pVDraw2D (istat)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: istat


END SUBROUTINE pVDraw2D

!#######################################################################
SUBROUTINE pVDraw1D (istat)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: istat


END SUBROUTINE pVDraw1D
