#include "parameters.h"
#define MAGN 0.1

!#######################################################################
SUBROUTINE pVDraw3D (istat)
!#######################################################################
  USE server_common, ONLY: cftoggle, cltoggle, ddtoggle
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: istat

  SELECT CASE(istat)
    CASE(0) ! Rendering of static objects
!      IF(cftoggle) CALL Draw3DCoordSys
!      IF(cltoggle) CALL Draw3DContourLegend
!      IF(ddtoggle) CALL Draw3DDataDisplay

    CASE(1) ! Rendering of dynamic objects

    CASE(2) ! Rendering of translucent objects

  END SELECT

END SUBROUTINE pVDraw3D

!#######################################################################
SUBROUTINE Draw3DCoordSys()
!#######################################################################
  USE server_common, ONLY: winid
  IMPLICIT NONE

  INTERFACE
    SUBROUTINE DrawArrayEquate(arr1,ind1, arr2,ind2)
      REAL,INTENT(OUT) :: arr1(3,*)
      REAL,INTENT(IN) :: arr2(3,*)
      INTEGER,INTENT(IN) :: ind1,ind2
    END SUBROUTINE DrawArrayEquate
  END INTERFACE

  REAL,PARAMETER :: START=-.9
  REAL,PARAMETER :: ZLOC=-.9
  INTEGER,PARAMETER :: NFACE=6

# ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: ivec(1)
# else
    INTEGER :: ivec(1)
# endif
  INTEGER :: np, i, j, nc(24)

  REAL :: rvec(12), xyzp(3,6), bxyzp(3,NFACE*4)
  REAL :: radii(6), bradii(NFACE*4)
  REAL :: col(3,6), bcol(NFACE*4)
  REAL :: xvec(3), yvec(3), zvec(3), xmag, ymag, zmag
  REAL :: cfrm(18)
  REAL :: fsize, fwidth, fxyz(3)
  REAL :: fnorx(3), fnory(3), fcol(3)

  CHARACTER(LEN=3) :: str

  cfrm=            (/ START,START,ZLOC, START+MAGN,START,ZLOC, &
 &                    START,START,ZLOC, START,START+MAGN,ZLOC, &
 &                    START,START,ZLOC, START,START,ZLOC-MAGN/)
  xvec = 0.
  yvec = 0.
  zvec = 0.
  xmag = 0.
  ymag = 0.
  zmag = 0.
  fsize = MAGN/4.
  fwidth = 1.
  fxyz = 0.
  fnorx = 0.
  fnory = 0.
  col = 1.
  bcol = 1.
  xyzp = 0.
  bxyzp = 0.
  radii = 0.
  bradii = 0.
  nc = 0

  ivec(1) = -2  
  fcol = 1.

! First draw the 3 lines
  ivec(1) = 0
  CALL GETSTATE(1, ivec, rvec, str)
  xmag = 0.
  ymag = 0.
  zmag = 0.
  DO i = 1, 3
    xvec( i ) = rvec( ( i - 1 ) * 4 + 1 )
    xmag = xmag + xvec( i )**2
    yvec( i ) = rvec( ( i - 1 ) * 4 + 2 )
    ymag = ymag + yvec( i )**2
    zvec( i ) = -rvec( ( i - 1 ) * 4 + 3 )
    zmag = zmag + zvec( i )**2
  ENDDO
  DO j = 1, 6
    DO i = 1, 3
      xyzp( i, j ) = cfrm( ( j - 1 ) * 3 + i )
    ENDDO
  ENDDO
  DO i = 1, 3
    xyzp(i,2) = xyzp(i,1) + MAGN*xvec(i)/SQRT(zmag)
    xyzp(i,4) = xyzp(i,3) + MAGN*yvec(i)/SQRT(zmag)
    xyzp(i,6) = xyzp(i,5) + MAGN*zvec(i)/SQRT(zmag)
  ENDDO
  np = -6
#ifdef PV3SERIAL
  CALL V3_Object3D(3,1,xyzp,radii,col,nc,np)
#else
  CALL pV_Object3D(3,4,xyzp,radii,col,np)
#endif

! Draw the X, Y, and Z
  fnorx = (/0.,0.,0./)
  fnory = (/0.,0.,0./)
  str = 'XYZ'
  DO i = 1, 3
    DO j = 1, 3
      fxyz(j) = xyzp(j,i*2)
    ENDDO
#   ifdef PV3SERIAL
      CALL XftnString(winid(3), 2, fcol, fxyz, str(i:i))
#   else
      CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, str(i:i))
#   endif
  ENDDO
  
! Draw the 6 faces of the cube
  np = -NFACE*4
  DO i = 1, 3
    xyzp(i,2) = xyzp(i,1) + MAGN*xvec(i)/(2.*SQRT(xmag))
    xyzp(i,4) = xyzp(i,3) + MAGN*yvec(i)/(2.*SQRT(ymag))
    xyzp(i,6) = xyzp(i,5) + MAGN*zvec(i)/(2.*SQRT(zmag))
  ENDDO
  xmag=0.
  ymag=0.
  zmag=0.
  DO i = 1, 3
    xvec( i ) = rvec((i-1)*4+1) + rvec((i-1)*4+2)
    xmag = xmag + xvec( i )**2
    yvec( i ) = rvec((i-1)*4+2) - rvec((i-1)*4+3)
    ymag = ymag + yvec( i )**2
    zvec( i ) = rvec((i-1)*4+1) - rvec((i-1)*4+3)
    zmag = zmag + zvec( i )**2
  ENDDO
! X-Y Quad
  CALL DrawArrayEquate(bxyzp,1,xyzp,1)
  CALL DrawArrayEquate(bxyzp,2,xyzp,2)
  bxyzp(1,3) = xyzp(1,1) + MAGN*xvec(1)/(SQRT(2.*xmag))
  bxyzp(2,3) = xyzp(2,1) + MAGN*xvec(2)/(SQRT(2.*xmag))
  bxyzp(3,3) = xyzp(3,1) + MAGN*xvec(3)/(SQRT(2.*xmag))
  CALL DrawArrayEquate(bxyzp,4,xyzp,4)
! Y-Z Quad
  CALL DrawArrayEquate(bxyzp,5,xyzp,1)
  CALL DrawArrayEquate(bxyzp,6,xyzp,4)
  bxyzp(1,7) = xyzp(1,1) + MAGN*yvec(1)/(SQRT(2.*ymag))
  bxyzp(2,7) = xyzp(2,1) + MAGN*yvec(2)/(SQRT(2.*ymag))
  bxyzp(3,7) = xyzp(3,1) + MAGN*yvec(3)/(SQRT(2.*ymag))
  CALL DrawArrayEquate(bxyzp,8,xyzp,6)
! X-Z Quad
  CALL DrawArrayEquate(bxyzp,9,xyzp,1)
  CALL DrawArrayEquate(bxyzp,10,xyzp,6)
  bxyzp(1,11) = xyzp(1,1) + MAGN*zvec(1)/(SQRT(2.*zmag))
  bxyzp(2,11) = xyzp(2,1) + MAGN*zvec(2)/(SQRT(2.*zmag))
  bxyzp(3,11) = xyzp(3,1) + MAGN*zvec(3)/(SQRT(2.*zmag))
  CALL DrawArrayEquate(bxyzp,12,xyzp,2)
  
  xmag=0.
  DO i = 1, 3
    xvec( i ) = rvec((i-1)*4+1) + rvec((i-1)*4+2) - rvec((i-1)*4+3)
    xmag = xmag + xvec( i )**2
  ENDDO
! X-Y : Z Quad
  CALL DrawArrayEquate(bxyzp,13,bxyzp,8)
  CALL DrawArrayEquate(bxyzp,14,bxyzp,7)
  bxyzp(1,15) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
  bxyzp(2,15) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
  bxyzp(3,15) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
  CALL DrawArrayEquate(bxyzp,16,bxyzp,11)
! Y-Z : X Quad
  CALL DrawArrayEquate(bxyzp,17,bxyzp,2)
  CALL DrawArrayEquate(bxyzp,18,bxyzp,11)
  bxyzp(1,19) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
  bxyzp(2,19) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
  bxyzp(3,19) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
  CALL DrawArrayEquate(bxyzp,20,bxyzp,3)
! X-Z : Y Quad
  CALL DrawArrayEquate(bxyzp,21,bxyzp,4)
  CALL DrawArrayEquate(bxyzp,22,bxyzp,3)
  bxyzp(1,23) =  xyzp(1,1) + MAGN*xvec(1)/(2.*SQRT(xmag/3.))
  bxyzp(2,23) =  xyzp(2,1) + MAGN*xvec(2)/(2.*SQRT(xmag/3.))
  bxyzp(3,23) =  xyzp(3,1) + MAGN*xvec(3)/(2.*SQRT(xmag/3.))
  CALL DrawArrayEquate(bxyzp,24,bxyzp,7)

#ifdef PV3SERIAL
  CALL V3_Object3D(0,0,bxyzp,bradii,bcol,nc,np)
#else
  CALL pV_Object3D(0,0,bxyzp,bradii,bcol,np)
#endif

END SUBROUTINE Draw3DCoordSys

!#######################################################################
SUBROUTINE Draw3DContourLegend()
!#######################################################################
  USE server_common, ONLY: clnlevels, cllocation, clsize, winid, cltitle
  IMPLICIT NONE

  INTERFACE
    SUBROUTINE DrawArrayEquate(arr1,ind1, arr2,ind2)
      REAL,INTENT(OUT) :: arr1(3,*)
      REAL,INTENT(IN) :: arr2(3,*)
      INTEGER,INTENT(IN) :: ind1,ind2
    END SUBROUTINE DrawArrayEquate
  END INTERFACE

# ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: ivec(1)
# else
    INTEGER :: ivec(1)
# endif
  INTEGER :: i, j, k, np, icolor, itype, hind

  REAL :: rvec(5)
  REAL :: radii(1)=0., startx, starty, startz, dx, dy, sizex, sizey, fx, fy, tx, ty
  REAL :: olxyzp(3,8), fcol(3), fsize, fwidth, fxyz(3), fnorx(3), fnory(3)
  REAL,ALLOCATABLE :: xyzp(:,:), col(:), col2(:,:), nc(:)

  CHARACTER(LEN=32) :: str

  np = -clnlevels*4
  ALLOCATE(xyzp(3,4*clnlevels), col(4*clnlevels), &
           col2(3,4*clnlevels), nc(4*clnlevels))

  ivec(1) = -2  

  CALL GETSTATE( 5, ivec, rvec, str ) ! returns the scalar title
! This call should be replaced by the variable being set in getkeys...
  str = cltitle

  fxyz  = 0.
  fnorx = 0.
  fnory = 0.
  fsize = MAGN/4.
  fwidth = 1.
  itype = 0
  icolor = 3
  fcol = 1.
  col2 = 1.

  hind = 1
  IF(cllocation.LT.0) hind = 2
  SELECT CASE(cllocation)
    CASE(1)
      tx = .65
      ty = .95
      dx = 0.
      dy = clsize/FLOAT(clnlevels)
      sizex = 2.*clsize/FLOAT(clnlevels)
      sizey = clsize/FLOAT(clnlevels)
      startx = .65
      starty = .90
      startz = -.99
      fx = -0.025
      fy = sizey/4.
    CASE(-1)
      tx = .95-clsize
      ty = .95
      dx = 2.*clsize/FLOAT(clnlevels)
      dy = 0.
      sizex = 2.*clsize/FLOAT(clnlevels)
      sizey = clsize/FLOAT(clnlevels)
      startx = .95
      starty = .90
      startz = -.99
      fx = -sizex/3.
      fy = -.05
    CASE(2)
      tx = .65
      ty = -.95+clsize+.05
      dx = 0.
      dy = clsize/FLOAT(clnlevels)
      sizex = 2.*clsize/FLOAT(clnlevels)
      sizey = clsize/FLOAT(clnlevels)
      startx = .65
      starty = -.95+clsize
      startz = -.99
      fx = -0.025
      fy = sizey/4.
    CASE(-2)
    CASE(3)
      tx = -.95
      ty = -.95+clsize+.05
      dx = 0.
      dy = clsize/FLOAT(clnlevels)
      sizex = 2.*clsize/FLOAT(clnlevels)
      sizey = clsize/FLOAT(clnlevels)
      startx = -.85
      starty = -.95+clsize
      startz = -.99
      fx = -0.025
      fy = sizey/4.
    CASE(-3)
    CASE(4)
      tx = -.95
      ty = .95
      dx = 0.
      dy = clsize/FLOAT(clnlevels)
      sizex = 2.*clsize/FLOAT(clnlevels)
      sizey = clsize/FLOAT(clnlevels)
      startx = -.85
      starty = .90
      startz = -.99
      fx = -0.025
      fy = sizey/4.
    CASE(-4)
  END SELECT
    
  fxyz(1) = tx
  fxyz(2) = ty
  fxyz(3) = startz
# ifdef PV3SERIAL
    CALL XftnString(winid(3), 2, fcol, fxyz, TRIM(ADJUSTL(str)))
# else
    CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, TRIM(ADJUSTL(str)))
# endif
  DO i = 1, clnlevels
    DO j = 1, 4
      xyzp(1,(i-1)*4+j) = startx
      xyzp(2,(i-1)*4+j) = starty
      xyzp(3,(i-1)*4+j) = startz
    ENDDO

    xyzp(2,(i-1)*4+2) = starty - sizey
    xyzp(1,(i-1)*4+3) = startx - sizex
    xyzp(2,(i-1)*4+3) = starty - sizey
    xyzp(1,(i-1)*4+4) = startx - sizex

    DO j = 1, 4
      col((i-1)*4+j) = FLOAT(clnlevels-i)/FLOAT(clnlevels-1)
    ENDDO

    fxyz(1) = xyzp(1,(i-1)*4+hind)-fx
    fxyz(2) = xyzp(2,(i-1)*4+hind)-fy
    fxyz(3) = xyzp(3,(i-1)*4+hind)

    rvec(5) = (rvec(2)-rvec(1))/FLOAT(clnlevels)*FLOAT(clnlevels-i+1) + rvec(1)
    WRITE(str,*) rvec(5)
#   ifdef PV3SERIAL
      CALL XftnString(winid(3), 2, fcol, fxyz, TRIM(ADJUSTL(str)))
#   else
      CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, TRIM(ADJUSTL(str)))
#   endif
    
    startx = startx - dx
    starty = starty - dy
  ENDDO
  fxyz(1) = xyzp(1,(clnlevels-1)*4+hind+1)-fx
  fxyz(2) = xyzp(2,(clnlevels-1)*4+hind+1)-fy
  fxyz(3) = xyzp(3,(clnlevels-1)*4+hind+1)
  WRITE(str,*) rvec(1)
# ifdef PV3SERIAL
    icolor = 1
    CALL XftnString(winid(3), 2, fcol, fxyz, TRIM(ADJUSTL(str)))
    CALL V3_Object3D(itype, icolor, xyzp, radii, col, nc, np)
# else
    icolor = 3 ! same as V3 but no lighting
    CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, TRIM(ADJUSTL(str)))
    CALL pV_Object3D(itype, icolor, xyzp, radii, col, np)
# endif

  itype = 3
  icolor = 4
  np = -8
  DO i = 1, clnlevels
    CALL DrawArrayEquate(olxyzp,1,xyzp,(i-1)*4+1)
    DO j = 1, 3
      CALL DrawArrayEquate(olxyzp,j*2,xyzp,(i-1)*4+j+1)
      CALL DrawArrayEquate(olxyzp,j*2+1,xyzp,(i-1)*4+j+1)
    ENDDO
    CALL DrawArrayEquate(olxyzp,8,xyzp,(i-1)*4+1)
    olxyzp(3,:) = olxyzp(3,:) - 0.01
#   ifdef PV3SERIAL
#   else
      CALL pV_Object3D(itype, icolor, olxyzp, radii, col2, np)
#   endif
  ENDDO

  DEALLOCATE(xyzp, col, col2, nc)

END SUBROUTINE Draw3DContourLegend

!#######################################################################
SUBROUTINE Draw3DDataDisplay()
!#######################################################################
  USE server_common, ONLY: winid
  IMPLICIT NONE
# ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: ivec(8)
# else
    INTEGER :: ivec(8)
# endif
  REAL :: rvec(8)
  REAL :: fcol(3), fsize, fwidth, fxyz(3), fnorx(3), fnory(3)
  CHARACTER :: str
  CHARACTER(LEN=32) :: strng

  fxyz  = (/.5,-1.,-.9/)
  fnorx = 0.
  fnory = 0.
  fsize = MAGN/4.
  fwidth = 1.
  fcol = 1.

  CALL GETSTATE( 3, ivec, rvec, str )
  SELECT CASE(ivec(1))
    CASE(2)
      WRITE(strng,"(A)") 'Planar Cut z='
    CASE(4)
#     ifdef PV3SERIAL
        WRITE(strng,"(A,F8.4)") 'Prog Cut z=',rvec(1)
#     else
      SELECT CASE(ivec(2))
        CASE(1)
          WRITE(strng,"(A,F8.4)") 'M`-T Cut z=',rvec(1)
        CASE(2)
          WRITE(strng,"(A,F8.4)") 'R-Z Cut z=',rvec(1)
        CASE(3)
          WRITE(strng,"(A,F8.4)") 'R Cut z=',rvec(1)
        CASE(4)
          WRITE(strng,"(A,F8.4)") 'T Cut z=',rvec(1)
        CASE(5)
          WRITE(strng,"(A,F8.4)") 'Z Cut z=',rvec(1)
        CASE(6)
          WRITE(strng,"(A,F8.4)") 'X Cut z=',rvec(1)
        CASE(7)
          WRITE(strng,"(A,F8.4)") 'Y Cut z=',rvec(1)
      END SELECT
#     endif
    CASE(5)
      WRITE(strng,"(A)") 'Mapped Surface'
    CASE(6)
      WRITE(strng,"(A)") 'Mapped Surf with Function'
    CASE(7)
      WRITE(strng,"(A,F8.4)") 'Iso z=',rvec(1)
    CASE DEFAULT
      RETURN
  END SELECT
# ifdef PV3SERIAL
    CALL XftnString(winid(3), 2, fcol, fxyz, TRIM(ADJUSTL(strng)))
# else
    CALL pV_Annotate3D(0, fcol, fsize, fwidth, fxyz, fnorx, fnory, TRIM(ADJUSTL(strng)))
# endif
      
END SUBROUTINE Draw3DDataDisplay

!#######################################################################
SUBROUTINE DrawArrayEquate(arr1,ind1, arr2,ind2)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: ind1, ind2
  REAL,INTENT(OUT) :: arr1(3,*)
  REAL,INTENT(IN) :: arr2(3,*)

  INTEGER :: i

  DO i = 1, 3
    arr1(i,ind1) = arr2(i,ind2)
  ENDDO

END SUBROUTINE DrawArrayEquate
