#include "parameters.h"

!#######################################################################
SUBROUTINE EVENTS(iwin, itype, ie, je, istate)
!#######################################################################
  USE keydata
  USE mod_events
# ifdef PV3SERIAL
    USE server_common, ONLY: winid, extracton, cltitle
    USE common_area, ONLY: nfunc
# else
    USE server_common, ONLY: winid, nfunc, extracton, cltitle
!    USE journal_common, ONLY: OpenJournalFile,  &
!                            FinalizeJournal,    &
!                            RecordJournalEntry, &
!                            ExecuteJournalEntry
# endif
  IMPLICIT NONE
!
!       Programmed Events Loop - Check and use appropriate events
!
!         called right after the event is pulled of the queue,
!         therefore if this routine changes the event info, pV3
!         uses the changed data.  events can be ignored by resetting
!         the window id to -1/0.  use pv_getpointer(...) to get
!         window data.
!
!       where:  iwin    - X window id of event
!                         -1 - no event and break out of event collection
!                          0 - no event but continue in tight loop
!               itype   - type of X event
!                          2 - KeyPress
!                          3 - KeyRelease
!                          4 - ButtonPress
!                          5 - ButtonRelease
!               ie      - x pixel number
!               je      - y pixel number
!               istate  - state info (key stroke or button #)
!
  INTEGER(KIND=BSIZE), INTENT(INOUT) :: iwin
  INTEGER, INTENT(INOUT) :: itype, istate, ie, je

# ifdef PV3SERIAL
    INTEGER(KIND=BSIZE) :: POVnode2d(3,1)
    INTEGER(KIND=BSIZE) :: POVcell(3,1)
    INTEGER(KIND=BSIZE) :: POVnsurf(2,1)
    INTEGER(KIND=BSIZE) :: POVscel(4,1)
    INTEGER(KIND=BSIZE) :: POVcellsl(1)

    REAL :: POVxyz(3,1)
    REAL :: POVs(1)
    REAL :: POVt(1)
    REAL :: POVdterp(1)
    REAL :: POVweightsl(8,1)
    REAL :: POVtimesl(1)
    REAL :: POVanglesl(1)

    POINTER (pPOVxyz,POVxyz)
    POINTER (pPOVs,POVs)
    POINTER (pPOVt,POVt)
    POINTER (pPOVdterp,POVdterp)
    POINTER (pPOVnode2d,POVnode2d)
    POINTER (pPOVcell,POVcell)
    POINTER (pPOVnsurf,POVnsurf)
    POINTER (pPOVscel,POVscel)
    POINTER (pPOVcellsl,POVcellsl)
    POINTER (pPOVweightsl,POVweightsl)
    POINTER (pPOVtimesl,POVtimesl)
    POINTER (pPOVanglesl,POVanglesl)

    INTEGER :: nnode, knode2, kcell, knsurf, ksurf, kstream
    INTEGER :: plotmask, typ, maxnsurf, domsurfctr, domstart, ctr

    CHARACTER(LEN=20) :: nam

    REAL :: POVtmpx,POVtmpy,POVtmpz,POVtmps,POVtmpt, nodeweights(8)
# endif

  INTEGER key1, key2, key3, key4
  INTEGER :: i, j, k
  INTEGER :: isstat

  REAL :: rdum(12)
  REAL :: newplaneloc

  CHARACTER*80 cstr
  CHARACTER*32 :: strs ! String for scalars
  CHARACTER*32, SAVE :: varstr=''
  CHARACTER strt
  CHARACTER(LEN=4), SAVE :: modchar = '0000'

  LOGICAL :: ret = .FALSE.

!=======================================================================
! The first time in - get window ids and special key bindings
!=======================================================================
  IF(first)THEN
#ifdef PV3SERIAL
    CALL V3_GetState(0, winid, rdum, cstr)
    CALL V3_GetState(-1, keybin, rvec, cstr)
    CALL V3_GetState(-2, ivec,   rvec, cstr)
    iopt  = ivec(1)
#else
    DO i = 1, 9
      CALL pV_GetPointer(i, winid(i))
    ENDDO
    CALL pV_GetState(-1, keybin, rvec, cstr)
    CALL pV_GetState(-1,ivec,rvec,cstr)
    CALL pV_GetState(-2,ivec,rvec,cstr)
    iopt = ivec(1)
!    CALL OpenJournalFile
#endif
    first = .FALSE.
    rdum = 0.
    iwin=-1
    autotec = .FALSE.
    CALL event_set_meta_false
  ENDIF
  cltitle = varstr

!IF(iwin.EQ.winid(3) .AND. itype.EQ.4 .AND. istate.EQ.1) CALL G_Thread_Init(1)

!=======================================================================
!  Setup basic event data
!=======================================================================

  IF (.NOT.event_set_meta(istate,iwin,itype)) RETURN
  IF (.NOT.event_set_window(iwin)) RETURN

!=======================================================================
!  Fork event data processing
!=======================================================================

  SELECT CASE(window)
    CASE(0)
      CALL event_process_batch(itype,istate,iwin,ie,je)
    CASE(1)
    CASE(2)
      CALL event_process_2d(itype,istate,iwin,ie,je)
    CASE(3)
      CALL event_process_3d(itype,istate,iwin,ie,je)
    CASE(4)
      CALL event_process_dials(itype,istate,iwin,ie,je)
    CASE(5)
    CASE(6)
    CASE(7)
    CASE(8)
    CASE(9)
  END SELECT

  IF (auto) CALL event_process_auto(itype,istate,iwin,ie,je)
  IF (autotec) CALL event_process_autotec(itype,istate,iwin,ie,je)
  IF (autoimg) CALL event_process_autoimg(itype,istate,iwin,ie,je)

END SUBROUTINE EVENTS



!#######################################################################
SUBROUTINE event_process_auto(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin

  IF (istate.EQ.ICHAR('S')) THEN
      CALL GETSTATE(3,ivect,rvect,strt)
      SELECT CASE(ivect(1))
        CASE(4,7) ! Programmed cut or iso-surface
          CALL CURSOR(.TRUE.)
1001      CONTINUE
          WRITE(*,*) '  Enter number of intervals'
          WRITE(*,*) ' < 0  Scan will reverse automatically'
          WRITE(*,*) ' Max: ', rvect(7)
          WRITE(*,*) ' Min: ', rvect(6)
          READ(*,*,ERR=1001) nintervals,uzmin,uzmax
          ds = (uzmax - uzmin)/ABS(FLOAT(nintervals))
          WRITE(*,*) ' Scanning is On!'
          CALL CURSOR(.FALSE.)
      END SELECT
  ELSE
      CALL GETSTATE(3, ivect, rvect, strt)
      SELECT CASE(ivect(1))
        CASE(4,7) ! Programmed cut or isosurface
          newplaneloc = rvect(1) + ds
          IF(newplaneloc.GT.uzmax .OR. newplaneloc.LT.uzmin)THEN
              IF(nintervals.LT.0)THEN
                  ds = -ds
                  newplaneloc = rvect(1) + ds
              ELSE
                  newplaneloc = uzmin
              ENDIF
          ENDIF
          rvect(1) = newplaneloc
#ifdef PV3SERIAL
          ivect=-1
#endif
          CALL SETSTATE(3, ivect, rvect, strt)
      END SELECT
  ENDIF

END SUBROUTINE event_process_auto


!#######################################################################
SUBROUTINE event_process_autotec(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin

  tesent = .NOT. tesent
  IF(tesent)THEN
      PRINT*,' AUTODUMPING*********************************************************************'
#     ifdef PV3SERIAL
#       include "postv3.f90"
#     else
        CALL GetExtractData()
#     endif
  ENDIF
END SUBROUTINE event_process_autotec



!#######################################################################
SUBROUTINE event_process_autoimg(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr
  INTEGER, INTENT(inout) :: itype, istate, iwin, ie, je

  imsent = .NOT. imsent
  IF(imsent)THEN
      cstr = ' '
      ivec(1) = 0
      CALL SETSTATE(8, ivec, rdum, cstr)
      iwin = winid(aiwin)
      itype = 2
      ie = 1
      je = 1
      istate = 126
  ENDIF
END SUBROUTINE event_process_autoimg



!#######################################################################
SUBROUTINE event_process_batch(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin

  IF (isprocess) THEN
      ispbsent = .NOT. ispbsent
      IF(ispbsent)THEN
          CALL ImageScriptProcessBatch(iwin, itype, ie, je, istate, winid,isstat)
#         ifdef PV3SERIAL
            SELECT CASE(isstat)
              CASE(1)
                CALL V3_SetState(8, ivec, rvec, '')
              CASE(2)
#               include "postv3.f90"
            END SELECT
#         endif
          IF(iwin.EQ.-1) CALL SETSTATE(-1, ivec, rvec, cstr)
      ENDIF
  ENDIF
END SUBROUTINE event_process_batch



!#######################################################################
SUBROUTINE event_process_2d(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin

  SELECT CASE (itype)
!-----------------------------------------------------------------------
!  key press
    CASE(2)
      SELECT CASE(istate)
!!! "?" for help
        CASE(ICHAR('?'))
          CALL event_help_2d

!!! "I" for Auto-image saving
        CASE(ICHAR('I'))
          autoimg = .NOT. autoimg
          IF(autoimg)THEN
              imsent = .FALSE.
              aiwin = 2
          ENDIF
          WRITE(*,*) ' '

! 'M' for domain surface mapping
        CASE(ICHAR('M'))
          CALL event_opt_domainmap(estr)

! Replace middle mouse with "s"
        CASE(ICHAR('s'))
          itype  = 4
          istate = 2
          RETURN ! Must return or it will interfere with pseudo-dials ???

      END SELECT
!-----------------------------------------------------------------------
!  key release
    CASE(3)

!-----------------------------------------------------------------------
!  mouse press
    CASE(4)
      IF (istate.EQ.2) THEN
          CALL XFtnSDialFun(iwin,ie,je,2,lineout)
          iwin = 0
      ENDIF

!-----------------------------------------------------------------------
!  mouse release
    CASE(5)
      IF(istate.EQ.2)THEN
          CALL XFtnSDialFun(iwin,ie,je,0,lineout)
          iwin = 0
      ENDIF
  END SELECT


END SUBROUTINE event_process_2d



!#######################################################################
SUBROUTINE event_process_3d(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  USE keydata
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin
  INTEGER :: isstat
  LOGICAL :: ret

  SELECT CASE (itype)
!-----------------------------------------------------------------------
! key press
    CASE(2)
!!----------------------------------------------------------------------
!! alt
      IF (event_get_meta_alt()) THEN
          SELECT CASE (istate)
            CASE(ICHAR('c'))
              CALL CURSOR(.TRUE.)
              CALL ControlPanel(isstat)
              CALL CURSOR(.FALSE.)
#             ifdef PV3SERIAL
                  IF (isstat) CALL SETSTATE(-1,ivec,rvec,cstr)
#             endif
!!! "alt f" for display rate
            CASE(ICHAR('f'))
              CALL event_opt_disprate(eventstr,istate)
            CASE(ICHAR('p'))
              autotec = .NOT. autotec
              IF (autotec) THEN
                  WRITE(*,*) ' Starting tecplot dumping'
                  tesent = .FALSE.
              ELSE
                  WRITE(*,*) ' Ending tecplot dumping'
              END IF
!!! "alt -" for decreasing display rate
            CASE(65451)
              CALL event_opt_disprate(eventstr,istate)
!!! "alt +" for increasing display rate
            CASE(65453)
              CALL event_opt_disprate(eventstr,istate)
!!! "alt pause" for pausing updates
            CASE(65299)
              CALL event_opt_disprate(eventstr,istate)
!!! "alt *" for changing rotation direction
            CASE(65450)
              CALL event_opt_disprate(eventstr,istate)
          END SELECT
          istate = 0
!!----------------------------------------------------------------------
!! no meta-keys
      ELSE
          SELECT CASE (istate)
#ifdef      PV3SERIAL
            CASE(ICHAR('/'))
              CALL V3_GetState(14,ivec,rvec,cstr)
              ivec(1) = 1 - ivec(1)
              IF(ivec(1).EQ.0)THEN
                  WRITE(*,*) ' Edge outlines off!'
              ELSE
                  WRITE(*,*) ' Edge outlines on!'
              ENDIF
              CALL V3_GetState(14,ivec,rvec,cstr)
#endif
            CASE(27) ! ...?
!!! scroll lock to lock frame update with iteration stepping
            CASE(65300)
              CALL event_opt_disprate(eventstr,istate)
#ifdef WIIMOTE
!!! wiimote interface
            CASE(ICHAR('u'))
              CALL fwii_interface(1)
            CASE(ICHAR('U'))
              CALL fwii_interface(-1)
#endif
!!! "?" for help
            CASE(ICHAR('?'))
              CALL event_help_3d
!!! "A" for averaging
            CASE(ICHAR('A'))
              IF (.NOT.event_var_avg(istate)) iwin = -1
!!! "D" for dimensionality
            CASE(ICHAR('D'))
              dimdisp = dimdisp + 1
              IF (dimdisp.GT.2) dimdisp = 0
              SELECT CASE(dimdisp)
                CASE(0)
                  WRITE(*,*) ' Dimension set to: non-dimensional'
                CASE(1)
                  WRITE(*,*) ' Dimension set to: Metric'
                CASE(2)
                  WRITE(*,*) ' Dimension set to: English'
              END SELECT
!!! "F" for time mode toggles
            CASE(ICHAR('F'))
              timedisp = timedisp + 1
              IF(timedisp.GT.2)timedisp = 0
              SELECT CASE(timedisp)
                CASE(0)
                  WRITE(*,*) ' Time mode set to flow file number'
                CASE(1)
                  WRITE(*,*) ' Time mode set to simulation time'
                CASE(2)
                  timesav = 0.
                  WRITE(*,*) 'What is physical delta t of each frame? (sec-in/ft)'
                  CALL CURSOR(.TRUE.)
                  READ(*,*) dt
                  CALL CURSOR(.FALSE.)
                  WRITE(*,*) ' Time mode set to phase-lag time'
            END SELECT
!!! "I" for auto-image dumps
            CASE(ICHAR('I'))
              autoimg = .NOT. autoimg
              IF(autoimg)THEN
                WRITE(*,*) ' Starting image dumping'
                imsent = .FALSE.
                aiwin = 3
              ELSE
                WRITE(*,*) ' Ending image dumping'
              ENDIF
              WRITE(*,*) ' '
!!! "M" for vizmath
            CASE(ICHAR('M'))
              IF (.NOT.event_var_vizmath(istate)) iwin = -1
#ifndef PV3SERIAL
!!! "N" for replication modification
            CASE(ICHAR('N'))
              CALL ServerReplication
#endif
!!! "P" for post-processing
            CASE(ICHAR('P'))
#ifdef PV3SERIAL
#             include "postv3.f90"
#else
              CALL GetExtractData()
#endif           
!!! "T" for thresholds
            CASE(ICHAR('T'))
              IF (.NOT.event_var_threshold(istate)) iwin = -1
!!! "V" for vectors
            CASE(ICHAR('V'))
              IF (.NOT.event_var_vector(istate)) iwin = -1
!!! "X" for feature extraction
            CASE(ICHAR('X'))
              CALL event_opt_featureext(eventstr)
!!! Up and Down arrows for adjusting display rate
            CASE(65362,65364)
              CALL event_opt_disprate(eventstr,istate)
!!! Function keys to disable "auto"
            CASE(374,371)
              auto = .FALSE.
!!! "a" for angles
            CASE(ICHAR('a'))
              IF (.NOT.event_var_angles(istate)) iwin = -1
!!! "c" for coordinates
            CASE(ICHAR('c'))
              IF (.NOT.event_var_coord(istate)) iwin = -1
!!! "d" for density
            CASE(ICHAR('d'))
              istate = 11500
!!! "e" for internal energy
            CASE(ICHAR('e'))
              IF (.NOT.event_var_intengy(istate)) iwin = -1
!!! "h" for enthalpy
            CASE(ICHAR('h'))
              IF (.NOT.event_var_enthalpy(istate)) iwin = -1
!!! "i" for rothalpy
            CASE(ICHAR('i'))
              IF (.NOT.event_var_rothalpy(istate)) iwin = -1
!!! "m" for mach
            CASE(ICHAR('m'))
              IF (.NOT.event_var_mach(istate)) iwin = -1
!!! "o" for other
            CASE(ICHAR('o'))
              IF (.NOT.event_var_other(istate)) iwin = -1
!!! "p" for pressure
            CASE(ICHAR('p'))
              IF (.NOT.event_var_pressure(istate)) iwin = -1
!!! "q" for turbulence
            CASE(ICHAR('q'))
              IF (.NOT.event_var_turbulence(istate)) iwin = -1
!!! "r" for rho-v
            CASE(ICHAR('r'))
              IF (.NOT.event_var_rhov(istate)) iwin = -1
!!! "s" for entropy
            CASE(ICHAR('s'))
              IF (.NOT.event_var_entropy(istate)) iwin = -1
!!! "t" for temperature
            CASE(ICHAR('t'))
              IF (.NOT.event_var_temperature(istate)) iwin = -1
!!! "v" for velocity
            CASE(ICHAR('v'))
              IF (.NOT.event_var_velocity(istate)) iwin = -1
!!! "w" for vorticity
            CASE(ICHAR('w'))
              IF (.NOT.event_var_vorticity(istate)) iwin = -1
!!! "f" for functions
            CASE(ICHAR('f'))
              IF (.NOT.event_var_func(istate)) iwin = -1
!!! "S" for cut plane scanning
            CASE(ICHAR('S'))
              auto = .NOT.auto
              IF (.NOT.auto) WRITE(*,*) ' Scanning is off!'
          END SELECT
      END IF

!-----------------------------------------------------------------------
!  mouse press
    CASE(4)
      IF (istate.EQ.2) THEN
          CALL XFtnSDialFun(iwin,ie,je,3,lineout)
          iwin = 0
      ENDIF
!-----------------------------------------------------------------------
!  mouse release
    CASE(5)
      IF (istate.EQ.2) THEN
          CALL XFtnSDialFun(iwin,ie,je,0,lineout)
          iwin = 0
      ENDIF
  END SELECT

END SUBROUTINE event_process_3d



!#######################################################################
SUBROUTINE event_process_dials(estr, itype, istate, iwin, ie, je)
!#######################################################################
  USE mod_events
  INTEGER, INTENT(in) :: estr, ie, je
  INTEGER, INTENT(inout) :: itype, istate, iwin

  SELECT CASE (itype)
!-----------------------------------------------------------------------
!  key press
    CASE(2)
      SELECT CASE(istate)
!!! "O" for outline
        CASE(ICHAR('O'))
          lineout = lineout + 1
          IF(lineout.GT.1) THEN
              lineout = 0
              WRITE(*,*) ' Outline mode: Off!'
          ELSE
              WRITE(*,*) ' Outline mode: On!'
          ENDIF
          WRITE(*,*) ' '
!!! "I" for auto-image dumps
        CASE(ICHAR('I'))
          autoimg = .NOT. autoimg
          IF(autoimg)THEN
              WRITE(*,*) ' Starting image dumping'
              imsent = .FALSE.
              aiwin = 3
          ELSE
              WRITE(*,*) ' Ending image dumping'
          ENDIF
          WRITE(*,*) ' '
      END SELECT
!-----------------------------------------------------------------------
!  key release
    CASE(3)

!-----------------------------------------------------------------------
!  mouse press
    CASE(4)

!-----------------------------------------------------------------------
!  mouse release
    CASE(5)

  END SELECT

END SUBROUTINE event_process_dials
