#include "parameters.h"
!#######################################################################
SUBROUTINE pVExtract(indx, exnum, reqmask, ival, rvec)
!#######################################################################
  USE keydata, ONLY: RSTATE
  USE common_area, ONLY: runmode, coordsys, mirror, itime, vcstor, &
                         sastor, blstor, rtstor,                   &
                         FXnbc, FXsepind, FXattind, nfunc
  INTERFACE
    SUBROUTINE FXpV3Shock(indx, exnum, minode, mindp)
      INTEGER, INTENT(IN) :: indx, exnum, minode
      REAL, INTENT(IN) :: mindp
    END SUBROUTINE FXpV3Shock
  END INTERFACE

  INTEGER, INTENT(IN) :: indx, exnum, reqmask, ival
  REAL, INTENT(IN) :: rvec(9)

  INTEGER :: sub0(5)
  INTEGER :: segs(1), bc(1), blcell(1), nseg, i, ibc
  INTEGER :: nblpts, nbltris
  INTEGER :: idummy(3) = (/1,1,1/)

  REAL :: vcoor(1), stren(1), slxyz(1), alxyz(1)
  REAL :: blxyz(1), bld(1), bltris(1)
  REAL :: dummy(3) = (/1.,1.,1./)

  POINTER (psegs,segs), (pvcoor,vcoor), (pstren,stren)
  POINTER (pslxyz,slxyz), (palxyz,alxyz), (pbc,bc)
  POINTER (pblxyz,blxyz), (pbld,bld), (pbltris,bltris), (pblcell,blcell)

  DEBUG2('   \ pVExtract')
  SELECT CASE(indx)
#ifdef FXPV3CLIENT
    CASE( EXT_VORTEX )

      IF(vcstor(itime)%nseg.EQ.-1)THEN
        WRITE(*,*) ' Calculating vortex cores'
        CALL FX_VortexCore(ival, nseg, psegs, pvcoor, pstren)
        vcstor(itime)%nseg = nseg
        ALLOCATE(vcstor(itime)%segs(nseg), vcstor(itime)%vcoor(3,segs(nseg)), vcstor(itime)%stren(segs(nseg)))
        DO i = 1, nseg
          vcstor(itime)%segs(i) = segs(i)
        ENDDO
        DO i = 1, segs(nseg)
          vcstor(itime)%vcoor(1,i) = vcoor(3*(i-1)+1)
          vcstor(itime)%vcoor(2,i) = vcoor(3*(i-1)+2)
          vcstor(itime)%vcoor(3,i) = vcoor(3*(i-1)+3)
        ENDDO
        DO i = 1, segs(nseg)
          vcstor(itime)%stren(i) = stren(i)
        ENDDO
        CALL FX_Free(psegs)
        CALL FX_Free(pvcoor)
        CALL FX_Free(pstren)
        WRITE(*,*) ' Done Calculating vortex cores'
      ENDIF

      IF(vcstor(itime)%nseg.GT.0)THEN
        CALL pV_SendXR(indx, exnum, 0, 3, vcstor(itime)%segs(vcstor(itime)%nseg), vcstor(itime)%vcoor)
        CALL pV_SendXI(indx, exnum, 1, 1, vcstor(itime)%nseg, vcstor(itime)%segs)
        IF(reqmask.EQ.7) CALL pV_SendXR(indx, exnum, 2, 1, vcstor(itime)%segs(vcstor(itime)%nseg), vcstor(itime)%stren)
      ELSE
        CALL pV_SendXR(indx, exnum, 0, 3, 1, dummy)
        CALL pV_SendXR(indx, exnum, 1, 1, 1, 1)
        IF(reqmask.EQ.7) CALL pV_SendXR(indx, exnum, 2, 1, 1, 1.)
      ENDIF

    CASE( EXT_SHOCKS )

      CALL FXpV3Shock(indx, exnum, ival, rvec(1))

    CASE( EXT_SEPATT )

      IF(sastor(itime)%calc.EQ.0)THEN
        WRITE(*,*) ' Calculating sep & att lines'
        CALL FX_SepNLine(FXsepind, pslxyz, FXattind, palxyz)
        sastor(itime)%calc = 1
        sastor(itime)%sepadd = pslxyz
        sastor(itime)%attadd = palxyz
        ALLOCATE(sastor(itime)%sepind(FXnbc), sastor(itime)%attind(FXnbc))
        DO i = 1, FXnbc
          sastor(itime)%sepind(i) = FXsepind(i)
          sastor(itime)%attind(i) = FXattind(i)
        ENDDO
!        PRINT*,' FXsepind = ',FXsepind
!        PRINT*,' FXattind = ',FXattind
        ALLOCATE(sastor(itime)%sepxyz(3,FXsepind(FXnbc)), sastor(itime)%attxyz(3,FXsepind(FXnbc)))
        DO i = 1, FXsepind(FXnbc)
          sastor(itime)%sepxyz(1,i) = slxyz(3*(i-1)+1)
          sastor(itime)%sepxyz(2,i) = slxyz(3*(i-1)+2)
          sastor(itime)%sepxyz(3,i) = slxyz(3*(i-1)+3)
        ENDDO
        DO i = 1, FXattind(FXnbc)
          sastor(itime)%attxyz(1,i) = alxyz(3*(i-1)+1)
          sastor(itime)%attxyz(2,i) = alxyz(3*(i-1)+2)
          sastor(itime)%attxyz(3,i) = alxyz(3*(i-1)+3)
        ENDDO
        CALL FX_Free(pslxyz)
        CALL FX_Free(palxyz)
        WRITE(*,*) ' Done Calculating sep & att lines'
      ENDIF

      IF(sastor(itime)%sepadd.EQ.0)THEN
        WRITE(*,*) ' Error, sepadd == 0'
        CALL pV_SendXR(indx, exnum, 0, 3, 1, dummy)
        CALL pV_SendXR(indx, exnum, 1, 3, 1, dummy)
        RETURN
      ENDIF

      CALL pV_GetStruc(321, pbc, ibc)
      ibc = -1
      DO i = 1, FXnbc
        IF(bc(3*(i-1)+3).EQ.ival) ibc = i
      ENDDO

      IF(ibc.EQ.-1)THEN
        WRITE(*,*) ' pVExtract: Global Surface #',ival,' NOT Found!'
        CALL pV_SendXR(indx, exnum, 0, 3, 1, dummy)
        CALL pV_SendXR(indx, exnum, 1, 3, 1, dummy)
        RETURN
      ENDIF

      i = 0
      IF(ibc.GT.0) i = sastor(itime)%sepind(ibc)
      nseg = sastor(itime)%sepind(ibc+1)-1
      IF(nseg.GT.1)THEN
        CALL pV_SendXR(indx, exnum, 0, 3, sastor(itime)%sepind(FXnbc), sastor(itime)%sepxyz)
      ELSE
        WRITE(*,*) ' Sep Dum vals'
        CALL pV_SendXR(indx, exnum, 0, 3, 1, dummy)
      ENDIF

      i = 0
      IF(ibc.GT.0) i = sastor(itime)%attind(ibc)
      nseg = sastor(itime)%attind(ibc+1)-1
      IF(nseg.GT.1)THEN
        CALL pV_SendXR(indx, exnum, 1, 3, sastor(itime)%attind(FXnbc), sastor(itime)%attxyz)
      ELSE
        WRITE(*,*) ' Att Dum vals'
        CALL pV_SendXR(indx, exnum, 1, 3, 1, dummy)
      ENDIF

    CASE( EXT_RESIDENCE )
      IF(rtstor(itime)%nrtpt.EQ.-1)THEN
      ENDIF

    CASE( EXT_BLWAKE )
! We are still waiting on Boundary layer stuff in the FX library
#  ifdef NEVERDEFINE
      IF(blstor(itime)%nblpts.EQ.-1)THEN
        WRITE(*,*) ' Calculating BL/Wake'
        CALL FX_BLSurface( nblpts, pblxyz, pbld, nbltris, pbltris, pblcell )
        IF(nblpts.GT.0)THEN
          blstor(itime)%nblpts = nblpts
          blstor(itime)%nbltris = nbltris
          ALLOCATE( blstor(itime)%blxyz(3,nblpts), blstor(itime)%bld(nblpts) )
          DO i = 1, nblpts
            blstor(itime)%blxyz( 1, i ) = blxyz( ( i - 1 ) * 3 + 1 )
            blstor(itime)%blxyz( 2, i ) = blxyz( ( i - 1 ) * 3 + 2 )
            blstor(itime)%blxyz( 3, i ) = blxyz( ( i - 1 ) * 3 + 3 )
            blstor(itime)%bld( i ) = bld( i )
          ENDDO
          CALL FX_Free( pblxyz )
          CALL FX_Free( pbld   )
          ALLOCATE( blstor(itime)%bltris(3,nbltris), blstor(itime)%blcell(nbltris) )
          DO i = 1, nbltris
            blstor(itime)%bltris( 1, i ) = bltris( ( i - 1 ) * 3 + 1 )
            blstor(itime)%bltris( 2, i ) = bltris( ( i - 1 ) * 3 + 2 )
            blstor(itime)%bltris( 3, i ) = bltris( ( i - 1 ) * 3 + 3 )
            blstor(itime)%blcell( i ) = blcell( i )
          ENDDO
          CALL FX_Free( pbltris )
          CALL FX_Free( pblcell )
        ELSE
          blstor(itime)%nblpts = 1
          blstor(itime)%nbltris = 1
          ALLOCATE(blstor(itime)%blxyz(3,1),blstor(itime)%bld(1))
          blstor(itime)%blxyz(1,1) = dummy(1)
          blstor(itime)%blxyz(2,1) = dummy(2)
          blstor(itime)%blxyz(3,1) = dummy(3)
          blstor(itime)%bld(1) = 0.
          ALLOCATE(blstor(itime)%bltris(3,1), blstor(itime)%blcell(1))
          blstor(itime)%bltris(1,1) = 1
          blstor(itime)%bltris(2,1) = 1
          blstor(itime)%bltris(3,1) = 1
          blstor(itime)%blcell(1) = 0
        ENDIF
        WRITE(*,*) ' Done calculating BL/Wake'
      ENDIF

      IF(blstor(itime)%nblpts.GT.0)THEN
        CALL pV_SendXI(indx, exnum, 0, 3, blstor(itime)%nbltris, blstor(itime)%bltris)
        CALL pV_SendXR(indx, exnum, 2, 3, blstor(itime)%nblpts, blstor(itime)%blxyz)
      ELSE
        CALL pV_SendXR(indx, exnum, 0, 3, 1, dummy)
        CALL pV_SendXR(indx, exnum, 1, 3, 1, idummy)
      ENDIF
#  endif
#endif
    CASE( EXT_CLIENTDATA )
      sub0 = (/runmode, coordsys, mirror, RSTATE, nfunc/)
      CALL pV_SendXi(indx, exnum, 0, 1, 5, sub0)

  END SELECT

  DEBUG2('   / pVExtract')

END SUBROUTINE pVExtract
