#include "parameters.h"
!#######################################################################
SUBROUTINE pVGrid(xyz)
!#######################################################################
  USE common_area, ONLY: coordsys, runmode, pgr, igrid, ntime,       &
&                        grd_offsetx, grd_offsety, grd_offsetz,      &
&                        grd_offsetr, grd_offsett, trot, pfw, itime, &
&                        grd_tadjust
  IMPLICIT NONE

  REAL :: xyz(3,*)
  REAL :: offs(5), rtmp, ttmp, vxtmp, vytmp
  INTEGER :: i,j,m

  DEBUG1(' \ pVGrid')

  offs(1) = grd_offsetx * FLOAT(igrid-1) ! THESE SHOULD ALL BE 0 UNTIL THERE IS A LOOP
  offs(2) = grd_offsety * FLOAT(igrid-1)
  offs(3) = grd_offsetz * FLOAT(igrid-1)
  offs(4) = grd_offsetr * FLOAT(igrid-1)
  offs(5) = grd_offsett * FLOAT(igrid-1)
!=======================================================================
! If we are in grid unsteady mode then we need to set the theta values
! to the right time step
  SELECT CASE(coordsys)
    CASE(CARTESIAN)
        DO i = 1, pgr(igrid)%tpts
          xyz(1,i) = pgr(igrid)%x(i) + offs(1)
          xyz(2,i) = pgr(igrid)%y(i) + offs(2)
          xyz(3,i) = pgr(igrid)%z(i) + offs(3)
        ENDDO

    CASE(CYLINDRICAL)
      IF(trot)CALL GridSetTheta(igrid)
      IF(grd_tadjust)CALL GridThetaAdjust(igrid)
      DO i=1,pgr(igrid)%tpts
        xyz(1,i) = (pgr(igrid)%r(i) + offs(4))*SIN(pgr(igrid)%t(i) + offs(5))
        xyz(2,i) = (pgr(igrid)%r(i) + offs(4))*COS(pgr(igrid)%t(i) + offs(5))
        xyz(3,i) = -pgr(igrid)%z(i) + offs(3)
      ENDDO

    CASE(RHCYLINDRICAL)
      IF(trot)CALL GridSetTheta(igrid)
      IF(grd_tadjust)CALL GridThetaAdjust(igrid)
      DO i=1,pgr(igrid)%tpts
        xyz(1,i) = (pgr(igrid)%r(i) + offs(4))*COS(pgr(igrid)%t(i) + offs(5))
        xyz(2,i) = -(pgr(igrid)%r(i) + offs(4))*SIN(pgr(igrid)%t(i) + offs(5))
        xyz(3,i) = -pgr(igrid)%z(i) + offs(3)
      ENDDO

  END SELECT
  DEBUG1(' / pVGrid')

END SUBROUTINE pVGrid

!#######################################################################
SUBROUTINE GridThetaAdjust(c)
!#######################################################################
  USE common_area, ONLY: pgr, pfw, itime, omega
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: c
  INTEGER :: m
  REAL :: tmp, pi2

  pi2 = 2. * PI
  tmp = MOD(omega * pfw(itime)%tyme, pi2)
  DO m = 1, pgr(c)%tpts
!    pgr(c)%t0(m) = MOD(pgr(c)%t0(m) + tmp, pi2)
    pgr(c)%t0(m) = pgr(c)%t0(m) + tmp
  ENDDO

END SUBROUTINE GridThetaAdjust

!#######################################################################
SUBROUTINE GridSetTheta(c)
!#######################################################################
  USE common_area, ONLY: pgr, pfw, itime, omega
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: c
  INTEGER :: m
  REAL :: tmp, pi2

  pi2 = 2. * PI
  tmp = MOD(omega * pfw(itime)%tyme, pi2)
  DO m=1,pgr(c)%tpts
!    pgr(c)%t(m) = MOD(pgr(c)%t0(m) + tmp, pi2)
    pgr(c)%t(m) = pgr(c)%t0(m) + tmp
  ENDDO

END SUBROUTINE GridSetTheta

!#######################################################################
SUBROUTINE PreGrid()
!#######################################################################
  USE common_area, ONLY: coordsys, pgr, igrid, ntime,       &
&                        pfw, itime, &
&                        grd_translx, grd_transly,      &
&                        grd_translz, grd_translr, grd_translt
  IMPLICIT NONE

  REAL :: rtmp, ttmp, vxtmp, vytmp
  INTEGER :: i,j,m

  DEBUG1(' \ PreGrid')

!=======================================================================
! If we are in grid unsteady mode then we need to set the theta values
! to the right time step
  SELECT CASE(coordsys)
    CASE(CARTESIAN)
      IF(grd_translt.EQ.0. .AND. grd_translr.EQ.0.)THEN
        DO i = 1, pgr(igrid)%tpts
          pgr(igrid)%x(i) = pgr(igrid)%x(i) + grd_translx
          pgr(igrid)%y(i) = pgr(igrid)%y(i) + grd_transly
          pgr(igrid)%z(i) = pgr(igrid)%z(i) + grd_translz
        ENDDO
      ELSE
        DO i = 1, pgr(igrid)%tpts
          rtmp = SQRT((pgr(igrid)%x(i) + grd_translx)**2 + &
&                     (pgr(igrid)%y(i) + grd_transly)**2) + grd_translr
          ttmp = ATAN2(pgr(igrid)%y(i) + grd_transly,      &
&                      pgr(igrid)%x(i) + grd_translx) - grd_translt
          pgr(igrid)%x(i) = rtmp * COS(ttmp)
          pgr(igrid)%y(i) = rtmp * SIN(ttmp)
          pgr(igrid)%z(i) = pgr(igrid)%z(i) + grd_translz
          DO j = 1, ntime
            vxtmp = pfw(j)%vx(i) * COS(grd_translt) + pfw(j)%vy(i) * SIN(grd_translt)
            vytmp = -pfw(j)%vx(i) * SIN(grd_translt) + pfw(j)%vy(i) * COS(grd_translt)
            pfw(j)%vx(i) = vxtmp
            pfw(j)%vy(i) = vytmp
          ENDDO
        ENDDO
      ENDIF

    CASE(CYLINDRICAL,RHCYLINDRICAL)
      DO i = 1, pgr(igrid)%tpts
        pgr(igrid)%r(i) = pgr(igrid)%r(i)+grd_translr
        pgr(igrid)%t0(i) = pgr(igrid)%t0(i)+grd_translt
        pgr(igrid)%z(i) = pgr(igrid)%z(i)+grd_translz
      ENDDO

  END SELECT
  DEBUG1(' / PreGrid')

END SUBROUTINE PreGrid
