#include "parameters.h"
!#######################################################################
SUBROUTINE pVNotify(handle,messid)
!#######################################################################
  USE keydata
  USE readgrid, ONLY: ReloadGrid
  USE readflow, ONLY: ReloadFlow
  USE common_area, ONLY: set_ref_quantities, dimen
  USE iter, ONLY: stime
 
  INTEGER,INTENT(IN) :: handle,messid

  INTEGER :: numint,nfloat,nchar

  INTEGER :: ints(1)
  REAL :: floats(1)
  CHARACTER(LEN=80) :: chars
  POINTER  (pints,ints),(pfloats,floats),(pchars,chars)

  DEBUG2('   \ pVNotify')
  SELECT CASE(messid)
!=======================================================================
! Update of keys for scalar, threshold, and vector calculations
    CASE(SIG_STV_UPDATE)
      CALL pV_GetServerData(handle,messid,numint,pints, &
                            nfloat,pfloats,nchar,pchars)
      IF(pints.NE.0)THEN
        timedisp  = ints(1)
        dimdisp   = ints(2)
        stime     = ints(3)
        IF(dimdisp.NE.dimen)THEN
          dimen = dimdisp
          CALL set_ref_quantities(dimen)
        ENDIF
        CALL freepV(pints)
      ENDIF

      IF(pfloats.NE.0)CALL freepV(pfloats)
      IF(pchars.NE.0)CALL freepV(pchars)

!=======================================================================
! Update of time data
    CASE(SIG_TIM_UPDATE)
      CALL pV_GetServerData(handle,messid,numint,pints, &
                            nfloat,pfloats,nchar,pchars)
      IF(pints.NE.0)CALL freepV(pints)
      IF(pfloats.NE.0)THEN
        dt      = floats(1)
        timesav = floats(2)
        CALL freepV(pfloats)
      ENDIF
      IF(pchars.NE.0)CALL freepV(pchars)

!=======================================================================
! Reload grid
    CASE(SIG_RELOAD_GRID)
      CALL pV_GetServerData(handle,messid,numint,pints, &
                            nfloat,pfloats,nchar,pchars)
      PRINT*,' Replacing grid with ',chars(1:NSTRLEN)
      keyfilename = chars(1:NSTRLEN)
      RSTATE = RSTATE_RELOAD_GRID
      IF(pints.NE.0) CALL freepV(pints)
      IF(pfloats.NE.0) CALL freepV(pfloats)
      IF(pchars.NE.0) CALL freepV(pchars)

!=======================================================================
! Reload flow
    CASE(SIG_RELOAD_FLOW)
      CALL pV_GetServerData(handle,messid,numint,pints, &
                            nfloat,pfloats,nchar,pchars)
      PRINT*,' Replacing flow with ',chars(1:NSTRLEN)
      keyfilename = chars(1:NSTRLEN)
      RSTATE = RSTATE_RELOAD_FLOW
      IF(pints.NE.0) CALL freepV(pints)
      IF(pfloats.NE.0) CALL freepV(pfloats)
      IF(pchars.NE.0) CALL freepV(pchars)

!=======================================================================
! Clear status
    CASE(SIG_RSTATE_CLEAR)
      RSTATE = RSTATE_READY
      IF(pints.NE.0) CALL freepV(pints)
      IF(pfloats.NE.0) CALL freepV(pfloats)
      IF(pchars.NE.0) CALL freepV(pchars)

  END SELECT
  DEBUG2('   / pVNotify')

END SUBROUTINE pVNotify
