#include "parameters.h"
!#######################################################################
SUBROUTINE pVScal(key, v)
!#######################################################################
  USE calc_dens,     ONLY: ScalDensity
  USE calc_angles,   ONLY: ScalAngles
  USE calc_coord,    ONLY: ScalCoord
  USE calc_enth,     ONLY: ScalEnthalpy
  USE calc_roth,     ONLY: ScalRothalpy
  USE calc_mach,     ONLY: ScalMach
  USE calc_pres,     ONLY: ScalPressure
  USE calc_rhov,     ONLY: ScalRhov
  USE calc_entr,     ONLY: ScalEntropy
  USE calc_temp,     ONLY: ScalTemperature
  USE calc_vel,      ONLY: ScalVelocity
  USE calc_vrtc,     ONLY: ScalVorticity
  USE calc_turb,     ONLY: ScalTurbulence
  USE calc_ener,     ONLY: ScalEnergy
  USE calc_othr,     ONLY: ScalOther
  USE calc_tavg,     ONLY: ScalTimeaverage
  USE calc_vmth,     ONLY: ScalVizmath
  USE calc_func,     ONLY: ScalFunction
  USE common_area,   ONLY: pgr, igrid, itime, ntime, novort,      &
                           Clear_Mem_Flow
  USE keydata,       ONLY: key2_1, key2_2, key2_3, uistat, ikeys, &
                           nkeys, nscalkeys, KeyAllocStat
  USE mod_events,    ONLY: event_get_primary
  IMPLICIT NONE

  REAL, DIMENSION(*) :: v

  INTEGER :: key, skey ! Key for main switch statement
  INTEGER :: k, j, s, mvmax, ctr
  INTEGER, SAVE :: itimelast = 0
  LOGICAL :: ios

  IF(itimelast.EQ.itime .AND. ntime.GT.1)THEN
    RETURN
  ELSE
    itimelast = itime
  ENDIF
  DEBUG2('   \ pVScal')

!=======================================================================
! Startup of the calculations.
! Some changes on the switch routine were necessary, since FX will want
! direct access to calculations (and maybe more than 1 type of pressure,
! so setting it as the default for the "p" hits would not work).  These
! extra values are added as 10000 - 19999.

  igrid=1 ! Force grid to be structure-steady for now.  In fact, it may
          ! be forever and I sort of wish I hadn't ever come up with it.
  mvmax = pgr(igrid)%tpts ! Get the total number of points

  SELECT CASE (event_get_primary(key))
    CASE(100) ! input from FX functions
      SELECT CASE (key) 
        CASE(10000)
          CALL ScalPressure(v, 1, mvmax)
        CASE(10001)
          CALL ScalMach(v, 4, mvmax)
        CASE(10002)
          CALL ScalTurbulence(v, 7, mvmax)
        CASE(10003)
          CALL ScalEnthalpy(v, 1, mvmax)
      END SELECT
    CASE(110)
      CALL ScalTimeaverage(v,key,mvmax)
    CASE(111,112,113,114)
      CALL ScalVizmath(v,key,mvmax)
    CASE(115)
      CALL ScalDensity(v,mvmax)
    CASE(116)
      CALL ScalAngles(v,key,mvmax)
    CASE(117)
      CALL ScalCoord(v,key,mvmax)
    CASE(118)
      CALL ScalEnergy(v,key,mvmax)
    CASE(119)
      CALL ScalEnthalpy(v,key,mvmax)
    CASE(120)
      CALL ScalRothalpy(v,key,mvmax)
    CASE(121)
      CALL ScalMach(v,key,mvmax)
    CASE(122)
      CALL ScalOther(v,key,mvmax)
    CASE(123)
      CALL ScalPressure(v,key,mvmax)
    CASE(124)
      CALL ScalTurbulence(v,key,mvmax)
    CASE(125)
      CALL ScalRhov(v,key,mvmax)
    CASE(126)
      CALL ScalEntropy(v,key,mvmax)
    CASE(127)
      CALL ScalTemperature(v,key,mvmax)
    CASE(128)
      CALL ScalVelocity(v,key,mvmax)
    CASE(129)
      IF (novort) RETURN
      CALL ScalVorticity(v,key,mvmax)
    CASE(130)
      CALL ScalFunction(v,key,mvmax)
    CASE(132) ! vectors, do nothing
    CASE(131) ! thresholds, do nothing
    CASE DEFAULT ! test for standard keypress
      SELECT CASE (skey)
        CASE DEFAULT

      END SELECT
  END SELECT


  DEBUG2('   / pVScal')

END SUBROUTINE pVScal
