#include "parameters.h"
!#######################################################################
SUBROUTINE pVStruc(knode, kequiv, kcel1, kcel2, kcel3, kcel4, &
                           knptet, kptet, knblck, kblk, khedra, &
                           ksurf, knsurf, hint)
!#######################################################################
  USE common_area
  USE readbcs
  USE MOD_ReadPegFile, ONLY : nbc, ibcs, ibce, jbcs, jbce, kbcs, kbce, ibtyp
  IMPLICIT NONE

  INTEGER EquivCount

  INTEGER :: knode, kequiv, kcel1, kcel2, kcel3, kcel4, knptet, kptet
  INTEGER :: knblck, kblk(3,*), khedra, ksurf, knsurf, hint
  INTEGER :: i,j,k, m, sur, ctr, mbb, msp, mfl, il, jl, kl, j1min,j1max
  INTEGER :: nbcx, nb, nx, is, ie, js, je, ks, ke
  INTEGER, ALLOCATABLE, DIMENSION(:) :: tmpbc

  LOGICAL :: surfasgn, intrfc

  DEBUG1(' \ pVStruc')
 
  SELECT CASE(runmode)
!=======================================================================
    CASE(-1,1,3)
! This is for the default interactive VSTAGE and batch VSTAGE modes
      knode  = 0
      kequiv = 0
      kcel1  = 0
      kcel2  = 0
      kcel3  = 0
      kcel4  = 0
      knptet = 0
      kptet  = 0
      knblck = pgr(igrid)%nblks
      khedra = 0
      knsurf = 8
      hint   = 0
      ksurf = 0
      DO i=1,pgr(igrid)%nblks
        mbb = pgr(igrid)%li(i)
        msp = pgr(igrid)%lj(i)
        mfl = pgr(igrid)%lk(i)
        kblk(1,i) = mbb
        kblk(2,i) = msp
        kblk(3,i) = mfl
        ksurf  = ksurf+2*((mbb-1)*(msp-1)+(mbb-1)*(mfl-1)+(msp-1)*(mfl-1))
      ENDDO

!=======================================================================
    CASE(2)
! AST files are displayed via the data in bc.in.  In local form, it can
! give the locations of the surfaces and their relative sizes.  We also
! know what the name is based on the number.
      knode  = 0
      kequiv = 0
      kcel1  = 0
      kcel2  = 0
      kcel3  = 0
      kcel4  = 0
      knptet = 0
      kptet  = 0
      knblck = pgr(1)%nblks ! This should be 1
      khedra = 0
      hint   = 0
! To calculate ksurf, we need to cycle through each bc or just give all
! the exterior cell walls...
      ksurf=0
      DO i=1,nturbobcs
        mbb=turbobcs(i,8)-turbobcs(i,5)+1
        msp=turbobcs(i,7)-turbobcs(i,4)+1
        mfl=turbobcs(i,6)-turbobcs(i,3)+1
        ksurf=ksurf+mbb*msp*mfl
      ENDDO
! Calculate the dimensions of each block
      DO i=1,pgr(igrid)%nblks
        mbb = pgr(igrid)%li(i)
        msp = pgr(igrid)%lj(i)
        mfl = pgr(igrid)%lk(i)
        kblk(1,i) = mbb
        kblk(2,i) = msp
        kblk(3,i) = mfl
      ENDDO
!-----------------------------------------------------------------------
! Figure out how many bcs there really are
      il=pgr(igrid)%li(1) ! This is for single block only
      jl=pgr(igrid)%lj(1)
      kl=pgr(igrid)%lk(1)
      ctr=0
!     Hub and case check
      DO sur=1,jl,jl-1
        surfasgn=.FALSE.
        DO m=1,nturbobcs
          IF(turbobcs(m,4).EQ.sur .AND. turbobcs(m,7).EQ.sur)surfasgn=.TRUE.
        ENDDO
        IF(surfasgn)ctr=ctr+1
      ENDDO
!     Blade check
      DO sur=1,il,il-1
        surfasgn=.FALSE.
        DO m=1,nturbobcs
          IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
             turbobcs(m,2).EQ.2)surfasgn=.TRUE.
        ENDDO
        IF(surfasgn)ctr=ctr+1
      ENDDO
!     Periodic and clearance conditions
      DO sur=1,il,il-1
        surfasgn=.FALSE.
        DO m=1,nturbobcs
          IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
             turbobcs(m,2).NE.2)surfasgn=.TRUE.
        ENDDO
        IF(surfasgn)ctr=ctr+1
      ENDDO
!     Inlet and outlet
      DO sur=1,kl,kl-1
        surfasgn=.FALSE.
        DO m=1,nturbobcs
          IF(turbobcs(m,3).EQ.sur .AND. turbobcs(m,6).EQ.sur)surfasgn=.TRUE.
        ENDDO
        IF(surfasgn)ctr=ctr+1
      ENDDO
      knsurf = ctr ! Gathered in bc.in

!=======================================================================
! Fieldview and AVUS
    CASE(4,5)
      knode  = pgr(igrid)%tpts
      kequiv = 0
      kcel1  = pgr(igrid)%ntet
      kcel2  = pgr(igrid)%npyr
      kcel3  = pgr(igrid)%npris
      kcel4  = pgr(igrid)%nhex
      knptet = 0
      kptet  = 0
      knblck = 0
      khedra = 0
      knsurf = fvd(igrid)%nbcs
      hint   = 1
      ksurf = 0
      DO i = 1, fvd(igrid)%nbcs
        ksurf = ksurf + fvd(igrid)%nbfaces(i)
      ENDDO     
!=======================================================================
! FDLF90
    CASE(6)
     ALLOCATE(tmpbc(SUM(nbc)))

      tmpbc = 0
      knsurf = 0
      DO nb = 1, pgr(igrid)%nblks
        DO nx = 1, nbc(nb)
          IF (.not. ANY(tmpbc(1:knsurf).eq.ibtyp(nx,nb))) THEN
            knsurf = knsurf + 1
            tmpbc(knsurf) = ibtyp(nx,nb)
          ENDIF
        ENDDO
      ENDDO

      DEALLOCATE(tmpbc)
      knode  = 0
      kequiv = 0
      kcel1  = 0
      kcel2  = 0
      kcel3  = 0
      kcel4  = 0
      knptet = 0
      kptet  = 0
      knblck = pgr(igrid)%nblks
      khedra = 0
      hint   = 0
      ksurf  = 0
      DO nb =1, pgr(igrid)%nblks
         kblk(1,nb) = pgr(igrid)%li(nb)
         kblk(2,nb) = pgr(igrid)%lj(nb)
         kblk(3,nb) = pgr(igrid)%lk(nb)

         nbcx = nbc(nb)
         DO nx = 1, nbcx
   
            is = ibcs(nx,nb)
            ie = ibce(nx,nb)
            js = jbcs(nx,nb)
            je = jbce(nx,nb)
            ks = kbcs(nx,nb)
            ke = kbce(nx,nb)

            IF (is .eq. ie) THEN
               js = MAX(1, js - 1)
               ks = MAX(1, ks - 1)
               msp = (je - js)
               mfl = (ke - ks)
               ksurf  = ksurf + msp*mfl
            ENDIF

            IF (js .eq. je) THEN
               is = MAX(1, is - 1)
               ks = MAX(1, ks - 1)
               mbb = (ie - is)
               mfl = (ke - ks)
               ksurf  = ksurf + mbb*mfl
            ENDIF

            IF (ks .eq. ke) THEN
               is = MAX(1, is - 1)
               js = MAX(1, js - 1)
               mbb = (ie - is)
               msp = (je - js)
               ksurf  = ksurf + mbb*msp
            ENDIF

         ENDDO
      ENDDO

!=======================================================================
! Tacoma
    CASE(11)
      knode  = 0
      kequiv = 0
      CALL ReadEquivFile(kequiv)
      kcel1  = 0
      kcel2  = 0
      kcel3  = 0
      kcel4  = 0
      knptet = 0
      kptet  = 0
      knblck = pgr(igrid)%nblks
      khedra = 0
      knsurf = 0
      hint   = 0
      ksurf = 0
      DO i=1,pgr(igrid)%nblks
        kblk(1,i) = pgr(igrid)%li(i)
        kblk(2,i) = pgr(igrid)%lj(i)
        kblk(3,i) = pgr(igrid)%lk(i)
      ENDDO
      SELECT CASE(nclients)
        CASE(1)
          j1min = 1
          j1max = pgr(igrid)%nblks
        CASE DEFAULT
          j1min = cid
          j1max = cid 
      END SELECT
      DO i = 1, nturbobcs
! Commenting the following section makes all of the surfaces come up, even the internal boundaries
        IF(.NOT. turbobcplot(i)) CYCLE
        SELECT CASE(nclients)
          CASE(1)
            knsurf = knsurf+1
          CASE DEFAULT
            DO j = j1min, j1max
              DO k = 1, patches(j)%nptch
                IF(patches(j)%ptch(k,7).EQ.i)THEN
                  knsurf = knsurf + 1
                  GOTO 1001
                ENDIF
              ENDDO
            ENDDO
            CYCLE
        END SELECT
!---/
1001    CONTINUE
        DO j = j1min, j1max
          SELECT CASE(nclients)
            CASE(1)
              il = pgr(igrid)%li(j)
              jl = pgr(igrid)%lj(j)
              kl = pgr(igrid)%lk(j)
            CASE DEFAULT
              il = pgr(igrid)%li(1)
              jl = pgr(igrid)%lj(1)
              kl = pgr(igrid)%lk(1)
          END SELECT

          DO k = 1, patches(j)%nptch
            IF(patches(j)%ptch(k,7).EQ.i)THEN
              IF(patches(j)%ptch(k,1).EQ.patches(j)%ptch(k,4))THEN
                DO jl = patches(j)%ptch(k,2), patches(j)%ptch(k,5)-1
                  DO kl = patches(j)%ptch(k,3), patches(j)%ptch(k,6)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ELSEIF(patches(j)%ptch(k,2).EQ.patches(j)%ptch(k,5))THEN
                DO il = patches(j)%ptch(k,1), patches(j)%ptch(k,4)-1
                  DO kl = patches(j)%ptch(k,3), patches(j)%ptch(k,6)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ELSEIF(patches(j)%ptch(k,3).EQ.patches(j)%ptch(k,6))THEN
                DO il = patches(j)%ptch(k,1), patches(j)%ptch(k,4)-1
                  DO jl = patches(j)%ptch(k,2), patches(j)%ptch(k,5)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ENDIF
            ENDIF
          ENDDO
        ENDDO
      ENDDO
      intrfc = .FALSE.
      DO i = 1, nturbobcs
        IF(turbobcplot(i))CYCLE
        intrfc = .TRUE.
        DO j = j1min, j1max
          SELECT CASE(nclients)
            CASE(1)
              il = pgr(igrid)%li(j)
              jl = pgr(igrid)%lj(j)
              kl = pgr(igrid)%lk(j)
            CASE DEFAULT
              il = pgr(igrid)%li(1)
              jl = pgr(igrid)%lj(1)
              kl = pgr(igrid)%lk(1)
          END SELECT

          DO k = 1, patches(j)%nptch
            IF(patches(j)%ptch(k,7).EQ.i)THEN
              IF(patches(j)%ptch(k,1).EQ.patches(j)%ptch(k,4))THEN
                DO jl = patches(j)%ptch(k,2), patches(j)%ptch(k,5)-1
                  DO kl = patches(j)%ptch(k,3), patches(j)%ptch(k,6)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ELSEIF(patches(j)%ptch(k,2).EQ.patches(j)%ptch(k,5))THEN
                DO il = patches(j)%ptch(k,1), patches(j)%ptch(k,4)-1
                  DO kl = patches(j)%ptch(k,3), patches(j)%ptch(k,6)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ELSEIF(patches(j)%ptch(k,3).EQ.patches(j)%ptch(k,6))THEN
                DO il = patches(j)%ptch(k,1), patches(j)%ptch(k,4)-1
                  DO jl = patches(j)%ptch(k,2), patches(j)%ptch(k,5)-1
                    ksurf = ksurf + 1
                  ENDDO
                ENDDO
              ENDIF
            ENDIF
          ENDDO
        ENDDO
      ENDDO
      IF(intrfc)knsurf = knsurf + 1

  END SELECT

  gksurf = knsurf ! Required for serial version in the v3surface call to this routine!!!
  gsurf = ksurf ! Required for serial version in the v3surface call to this routine!!!

  IF(liblank)THEN
    knblck = -knblck
  ENDIF
  IF(.NOT. bcavail)THEN
    ksurf = 0
    knsurf = 0
  ENDIF

  DEBUG1(' / pVStruc')

END SUBROUTINE pVStruc
