#include "parameters.h"
!#######################################################################
SUBROUTINE pVSurface(nsurf, scon, scel, tsurf)
!#######################################################################
  USE general, ONLY: KeywordFunc
  USE common_area
  USE readbcs
  USE MOD_BCDefs
  USE MOD_ReadPegFile, ONLY : nbc, ibtyp, ibdir, ibcs, ibce, jbcs, jbce, kbcs, kbce
  IMPLICIT NONE

!
!	routine that passes the surface info to pV3
!
!	where:	nsurf	- pointers to the end of each surface group and
!			  flag to indicate if surface is on
!		scon	- connection data for multi-cilent cases
!		scel    - node numbers for surfaces
!		tsurf	- surface group title (optional)
!
  INTEGER :: nsurf(3,*),scon(*),scel(4,*),ksurf,i,j,k, i1,j1,k1, s, ctr
  INTEGER :: il,jl,kl,m,mle,mte,mtp,mhb,sur,t
  INTEGER :: j1min, j1max
  INTEGER :: nbcg, nbcc, currbc, ngpts, nbcs, nbcx, nb, nx, is, ie, js, je, ks, ke
  INTEGER, ALLOCATABLE, DIMENSION(:) :: tmpbc
  INTEGER :: ijk2nod
  CHARACTER*20 :: tmpstr1,tmpstr2,tmpstr3,tmpstr4,tmpstr5
  CHARACTER*20 :: tsurf(*)
  CHARACTER*80 :: tmpstr
  LOGICAL :: surfasgn

DEBUG1(' \ pVSurface')

  SELECT CASE(runmode)
!=======================================================================
! First the default and batch vstage modes are presented for handling
!=======================================================================
    CASE(-1,1)
        ksurf = 0
        il  = pgr(igrid)%li(1)
        jl  = pgr(igrid)%lj(1)
        kl  = pgr(igrid)%lk(1)
        mle = pgr(igrid)%mle
        mte = pgr(igrid)%mte
        mtp = pgr(igrid)%mtp
        mhb = pgr(igrid)%mhb

!	surface 1 - inflow
        DO j = 1, jl-1
          DO i = 1, il-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(i,j,1,il,jl)
            scel(2,ksurf) = ijk2nod(i+1,j,1,il,jl)
            scel(3,ksurf) = ijk2nod(i+1,j+1,1,il,jl)
            scel(4,ksurf) = ijk2nod(i,j+1,1,il,jl)
            scon(ksurf) = 0
          ENDDO
        ENDDO
        nsurf(1,1) = ksurf
        nsurf(2,1) = 4 + 256
        nsurf(3,1) = 1
        tsurf(1) = 'InFlow'

!	surface 2 - outflow
        DO j = 1, jl-1
          DO i = 1, il-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(i,j,kl,il,jl)
            scel(2,ksurf) = ijk2nod(i+1,j,kl,il,jl)
            scel(3,ksurf) = ijk2nod(i+1,j+1,kl,il,jl)
            scel(4,ksurf) = ijk2nod(i,j+1,kl,il,jl)
            scon(ksurf) = 0
          enddo
        enddo
        nsurf(1,2) = ksurf
        nsurf(2,2) = 4 + 256
        nsurf(3,2) = 2
        tsurf(2) = 'OutFlow'

!	surface 3 - hub
        DO k = 1, kl-1
          DO i = 1, il-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(i,1,k,il,jl)
            scel(2,ksurf) = ijk2nod(i+1,1,k,il,jl)
            scel(3,ksurf) = ijk2nod(i+1,1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(i,1,k+1,il,jl)
            scon(ksurf) = 0
          enddo
        enddo
        nsurf(1,3) = ksurf
        nsurf(2,3) = 0 + 1 + 256
        nsurf(3,3) = 3
        tsurf(3) = 'Hub'

!	surface 4 - tip
        DO k = 1, kl-1
          DO i = 1, il-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(i,jl,k,il,jl)
            scel(2,ksurf) = ijk2nod(i+1,jl,k,il,jl)
            scel(3,ksurf) = ijk2nod(i+1,jl,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(i,jl,k+1,il,jl)
            scon(ksurf) = 0
          enddo
        enddo
        nsurf(1,4) = ksurf
        nsurf(2,4) = 0 + 4 + 256
        nsurf(3,4) = 4
        tsurf(4) = 'Tip'

!	surface 5 - blade
        DO k = mle,mte-1
          DO j=mhb,mtp-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(1,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(1,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(1,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(1,j+1,k,il,jl)
            scon(ksurf) = 0
          ENDDO
        ENDDO
        nsurf(1,5) = ksurf
        nsurf(2,5) = 1 + 256
        nsurf(3,5) = 5
        tsurf(5) = 'Blade Surface'

!	surface 6 - blade
        do k = mle,mte-1
          do j=mhb,mtp-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(il,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(il,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(il,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(il,j+1,k,il,jl)
            scon(ksurf) = 0
          enddo
        enddo
        nsurf(1,6) = ksurf
        nsurf(2,6) = 1 + 256
        nsurf(3,6) = 6
        tsurf(6) = 'Blade Surface'

!	surface 7 - periodic
        do k = 1, mle-1
          do j = 1, jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(1,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(1,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(1,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(1,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        do k=mle,mte-1
          do j=mtp,jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(1,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(1,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(1,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(1,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        do k = mte,kl-1
          do j = 1,jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(1,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(1,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(1,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(1,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        nsurf(1,7) = ksurf
        nsurf(2,7) = 4 + 256
        nsurf(3,7) = 7
        tsurf(7) = 'Periodic'

!	surface 8 - periodic
        do k = 1, mle-1
          do j = 1, jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(il,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(il,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(il,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(il,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        do k=mle,mte-1
          do j=mtp,jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(il,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(il,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(il,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(il,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        do k = mte,kl-1
          do j = 1,jl-1
            ksurf = ksurf + 1
            scel(1,ksurf) = ijk2nod(il,j,k,il,jl)
            scel(2,ksurf) = ijk2nod(il,j,k+1,il,jl)
            scel(3,ksurf) = ijk2nod(il,j+1,k+1,il,jl)
            scel(4,ksurf) = ijk2nod(il,j+1,k,il,jl)
            scon(ksurf) = -1
          enddo
        enddo
        nsurf(1,8) = ksurf
        nsurf(2,8) = 4 + 256
        nsurf(3,8) = 8
        tsurf(8) = 'Periodic'
!=======================================================================
! AST.B#.Q##
! These files are mostly the same as vstage, we just need to use the bc
! from bc.in to determine where we actually want a surface
!=======================================================================
    CASE(2)
      ksurf=0
      t=0
      il=pgr(igrid)%li(1)
      jl=pgr(igrid)%lj(1)
      kl=pgr(igrid)%lk(1)
!-----------------------------------------------------------------------
! Hub, jmin
      DO m=1,nturbobcs
        sur=1
        IF(turbobcs(m,4).EQ.sur .AND. turbobcs(m,7).EQ.sur)THEN
          j=1
          DO k=turbobcs(m,3),turbobcs(m,6)-1
            DO i=turbobcs(m,5),turbobcs(m,8)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i  ,j,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i+1,j,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i+1,j,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i  ,j,k+1,il,jl)
              scon(ksurf) = 0
            ENDDO
          ENDDO
          t=t+1
          nsurf(1,t) = ksurf
          nsurf(2,t) = 1 + 256
          nsurf(3,t) = 3
          tsurf(t) = 'Hub'
        ENDIF
      ENDDO
!-----------------------------------------------------------------------
! Case, jmax
      DO m=1,nturbobcs
        sur=jl
        IF(turbobcs(m,4).EQ.sur .AND. turbobcs(m,7).EQ.sur)THEN
          j=jl
          DO k=turbobcs(m,3),turbobcs(m,6)-1
            DO i=turbobcs(m,5),turbobcs(m,8)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i  ,j,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i+1,j,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i+1,j,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i  ,j,k+1,il,jl)
              scon(ksurf) = 0
            ENDDO
          ENDDO
          t=t+1
          nsurf(1,t) = ksurf
          nsurf(2,t) = 0 + 256
          nsurf(3,t) = 4
          tsurf(t) = 'Case'
        ENDIF
      ENDDO
!-----------------------------------------------------------------------
! Blade imin
      surfasgn=.FALSE.
      DO m=1,nturbobcs
        sur=1
        IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
           turbobcs(m,2).EQ.2)THEN
          i=1
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO k=turbobcs(m,3),turbobcs(m,6)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i,j  ,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i,j+1,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i,j+1,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i,j  ,k+1,il,jl)
              scon(ksurf) = 0
            ENDDO
          ENDDO
          surfasgn=.TRUE.
        ENDIF
      ENDDO
      IF(surfasgn)THEN
        t=t+1
        nsurf(1,t) = ksurf
        nsurf(2,t) = 1 + 256
        nsurf(3,t) = 5
        tsurf(t) = 'Wall -- imin'
      ENDIF
!-----------------------------------------------------------------------
! Periodic imin
      surfasgn=.FALSE.
      DO m=1,nturbobcs
        sur=1
        IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
           turbobcs(m,2).NE.2)THEN
          i=1
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO k=turbobcs(m,3),turbobcs(m,6)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i,j  ,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i,j+1,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i,j+1,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i,j  ,k+1,il,jl)
              scon(ksurf) = -1
            ENDDO
          ENDDO
          surfasgn=.TRUE.
        ENDIF
      ENDDO
      IF(surfasgn)THEN
        t=t+1
        nsurf(1,t) = ksurf
        nsurf(2,t) = 0 + 256
        nsurf(3,t) = 7
        tsurf(t) = 'Periodic -- imin'
      ENDIF
!-----------------------------------------------------------------------
! Blade imax
      surfasgn=.FALSE.
      DO m=1,nturbobcs
        sur=il
        IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
           turbobcs(m,2).EQ.2)THEN
          i=il
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO k=turbobcs(m,3),turbobcs(m,6)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i,j  ,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i,j+1,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i,j+1,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i,j  ,k+1,il,jl)
              scon(ksurf) = 0
            ENDDO
          ENDDO
          surfasgn=.TRUE.
        ENDIF
      ENDDO
      IF(surfasgn)THEN
        t=t+1
        nsurf(1,t) = ksurf
        nsurf(2,t) = 1 + 256
        nsurf(3,t) = 6
        tsurf(t) = 'Wall -- imax'
      ENDIF
!-----------------------------------------------------------------------
! Periodic imax
      surfasgn=.FALSE.
      DO m=1,nturbobcs
        sur=il
        IF(turbobcs(m,5).EQ.sur .AND. turbobcs(m,8).EQ.sur .AND. &
           turbobcs(m,2).NE.2)THEN
          i=il
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO k=turbobcs(m,3),turbobcs(m,6)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i,j  ,k  ,il,jl)
              scel(2,ksurf) = ijk2nod(i,j+1,k  ,il,jl)
              scel(3,ksurf) = ijk2nod(i,j+1,k+1,il,jl)
              scel(4,ksurf) = ijk2nod(i,j  ,k+1,il,jl)
              scon(ksurf) = -1
            ENDDO
          ENDDO
          surfasgn=.TRUE.
        ENDIF
      ENDDO
      IF(surfasgn)THEN
        t=t+1
        nsurf(1,t) = ksurf
        nsurf(2,t) = 0 + 256
        nsurf(3,t) = 8
        tsurf(t) = 'Periodic -- imax'
      ENDIF
!-----------------------------------------------------------------------
! Inlet, kmin
      DO m=1,nturbobcs
        sur=1
        IF(turbobcs(m,3).EQ.sur .AND. turbobcs(m,6).EQ.sur)THEN
          k=1
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO i=turbobcs(m,5),turbobcs(m,8)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i  ,j  ,k,il,jl)
              scel(2,ksurf) = ijk2nod(i+1,j  ,k,il,jl)
              scel(3,ksurf) = ijk2nod(i+1,j+1,k,il,jl)
              scel(4,ksurf) = ijk2nod(i  ,j+1,k,il,jl)
              scon(ksurf) = -1
            ENDDO
          ENDDO
          t=t+1
          nsurf(1,t) = ksurf
          nsurf(2,t) = 0 + 256
          nsurf(3,t) = 1
          tsurf(t) = 'InFlow'
        ENDIF
      ENDDO
!-----------------------------------------------------------------------
! Outlet, kmax
      DO m=1,nturbobcs
        sur=kl
        IF(turbobcs(m,3).EQ.sur .AND. turbobcs(m,6).EQ.sur)THEN
          k=kl
          DO j=turbobcs(m,4),turbobcs(m,7)-1
            DO i=turbobcs(m,5),turbobcs(m,8)-1
              ksurf = ksurf + 1
              scel(1,ksurf) = ijk2nod(i  ,j  ,k,il,jl)
              scel(2,ksurf) = ijk2nod(i+1,j  ,k,il,jl)
              scel(3,ksurf) = ijk2nod(i+1,j+1,k,il,jl)
              scel(4,ksurf) = ijk2nod(i  ,j+1,k,il,jl)
              scon(ksurf) = -1
            ENDDO
          ENDDO
          t=t+1
          nsurf(1,t) = ksurf
          nsurf(2,t) = 0 + 256
          nsurf(3,t) = 2
          tsurf(t) = 'OutFlow'
        ENDIF
      ENDDO

!=======================================================================
! Fieldview and AVUS unstructured files
!=======================================================================
    CASE(4,5)
      ksurf = 0
      DO i = 1, fvd(igrid)%nbcs
        DO j = 1, fvd(igrid)%nbfaces(i)
          ksurf = ksurf + 1
          DO k = 1, 4
            scel(k,ksurf) = fvd(igrid)%f(i)%f(k,j)
          ENDDO
          scon(ksurf) = 0
        ENDDO
        nsurf(1,i) = ksurf
        nsurf(2,i) = 1
        nsurf(3,i) = i
        tmpstr = TRIM(fvd(igrid)%bcnam(i))
        tsurf(i) = tmpstr(1:20)
      ENDDO
!=======================================================================
! FDL3DI/FDL3F90
!=======================================================================
    CASE(6)
      IF(.NOT. bcavail)RETURN

      ksurf = 0
      nbcs = 0
 
      ALLOCATE(tmpbc(SUM(nbc)))
      tmpbc = 0 

      nbcg = 0
      DO nb = 1, pgr(igrid)%nblks
        DO nx = 1, nbc(nb)
          IF (.not. ANY(tmpbc(1:nbcg).eq.ibtyp(nx,nb))) THEN
            nbcg = nbcg + 1
            tmpbc(nbcg) = ibtyp(nx,nb)
          ENDIF
        ENDDO
      ENDDO

      DO nbcc = 1, nbcg

         ngpts = 0
         currbc = tmpbc(nbcc)

      DO nb =1, pgr(igrid)%nblks
         il=pgr(igrid)%li(nb)
         jl=pgr(igrid)%lj(nb)
         kl=pgr(igrid)%lk(nb)
         nbcx = nbc(nb)
         DO nx = 1, nbcx
           
            IF (ibtyp(nx,nb) .ne. currbc) CYCLE 
        
            is = ibcs(nx,nb)
            ie = ibce(nx,nb)
            js = jbcs(nx,nb)
            je = jbce(nx,nb)
            ks = kbcs(nx,nb)
            ke = kbce(nx,nb)

            IF (is .eq. ie) THEN
               i = is
               js = MAX(1, js - 1)
               ks = MAX(1, ks - 1)
               DO k = ks, ke-1
                  DO j = js, je-1
                     ksurf = ksurf + 1
                     scel(1,ksurf) = ijk2nod(i,j  ,k  ,il,jl) + ngpts
                     scel(2,ksurf) = ijk2nod(i,j+1,k  ,il,jl) + ngpts
                     scel(3,ksurf) = ijk2nod(i,j+1,k+1,il,jl) + ngpts
                     scel(4,ksurf) = ijk2nod(i,j  ,k+1,il,jl) + ngpts
                     scon(ksurf) = 0
                  ENDDO
               ENDDO
            ENDIF

            IF (js .eq. je) THEN
               is = MAX(1, is - 1)
               j = js
               ks = MAX(1, ks - 1)
               DO k = ks, ke-1
                  DO i = is, ie-1
                     ksurf = ksurf + 1
                     scel(1,ksurf) = ijk2nod(i  ,j,k  ,il,jl) + ngpts
                     scel(2,ksurf) = ijk2nod(i+1,j,k  ,il,jl) + ngpts
                     scel(3,ksurf) = ijk2nod(i+1,j,k+1,il,jl) + ngpts
                     scel(4,ksurf) = ijk2nod(i  ,j,k+1,il,jl) + ngpts
                     scon(ksurf) = 0
                  ENDDO
               ENDDO
            ENDIF

            IF (ks .eq. ke) THEN
               is = MAX(1, is - 1)
               js = MAX(1, js - 1)
               k = ks
               DO j = js, je-1
                  DO i = is, ie-1
                     ksurf = ksurf + 1
                     scel(1,ksurf) = ijk2nod(i  ,j  ,k,il,jl) + ngpts
                     scel(2,ksurf) = ijk2nod(i+1,j  ,k,il,jl) + ngpts
                     scel(3,ksurf) = ijk2nod(i+1,j+1,k,il,jl) + ngpts
                     scel(4,ksurf) = ijk2nod(i  ,j+1,k,il,jl) + ngpts
                     scon(ksurf) = 0
                  ENDDO
               ENDDO
            ENDIF
            
         ENDDO
         ngpts = ngpts + il*jl*kl
      ENDDO
         nbcs = nbcs + 1
         nsurf(1,nbcs) = ksurf
         nsurf(2,nbcs) = 1 + 256
         nsurf(3,nbcs) = currbc
         SELECT CASE(currbc)
         CASE(1:nCFDBC)
            tsurf(  nbcs) = TRIM(CFDbcname(currbc))
         CASE(200:299)
            WRITE(tmpstr,"(I6)") currbc
            tsurf(  nbcs) = 'Periodic ' // TRIM(ADJUSTL(tmpstr))
         CASE DEFAULT
            WRITE(tmpstr,"(I6)") currbc
            tsurf(  nbcs) = 'Unknown BC ' // TRIM(ADJUSTL(tmpstr))
         END SELECT
      ENDDO
      
      DEALLOCATE(tmpbc)

!=======================================================================
! TACOMA
!=======================================================================
    CASE(11)
      CALL TacomaPVSurface(nsurf,scon,scel,tsurf)

    CASE DEFAULT
      WRITE(*,*) ' No Surfaces to display for runmode=',runmode
  END SELECT

DEBUG1(' / pVSurface')

END SUBROUTINE pVSurface
