#include "parameters.h"
!#######################################################################
SUBROUTINE pVXYPrime(zprime, kn, xyz, n, xyp)
!#######################################################################
  USE common_area, ONLY: pgr, igrid, coordsys, diam
  IMPLICIT NONE
!
!	return x-prime and y-prime for programmer defined cutting surface
!
!		this routine is called during the data-collection phase
!		of the program defined surfaces.  this routine may be
!		called many times during this phase.
!		there is no ordering of the nodes in kn, infact kn
!		may contain the same node more than once in a single call.
!
!	where:	zprime	- the current z-prime value
!		kn	- a vector of indices for the nodes to be
!			  transformed
!		xyz	- coordinate triads for 3-D nodes.
!			  this is the same data as set in V3GRID
!		n	- the number of nodes to be transformed.
!
!		note: zprime, kn, xyz and n must not be modified!
!
!		the following paramter is returned by this routine:
!
!		xyp	- the calculated x-prime and y-prime values for 
!                         the 3-D nodes
!
  INTEGER,INTENT(IN) :: n, kn(n)
  REAL,INTENT(IN)    :: zprime, xyz(3,*)
  REAL,INTENT(OUT)   :: xyp(2,n)

  REAL :: rave, tave, zave, xave, yave, rdev, tdev, zdev, xdev, ydev

  INTEGER :: itype, i, k, nnode

  COMMON /prime/ itype, rave, tave, zave, xave, yave, rdev, tdev, zdev, xdev, ydev, nnode

!=======================================================================
! M Prime-Theta or Radial cut
  IF(itype.EQ.1 .OR. itype.EQ.3)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = xyz(1,k)
          !xyp(2,i) = -pgr(igrid)%t0(k)*pgr(igrid)%r(k)
          xyp(2,i) = -pgr(igrid)%t(k)
        ENDDO
      CASE DEFAULT
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = -xyz(3,k)
!          xyp(2,i) = pgr(igrid)%t0(k)*pgr(igrid)%r(k)
          xyp(2,i) = pgr(igrid)%t(k)*diam
        ENDDO
    END SELECT
  ENDIF

!=======================================================================
! R-Z
  IF(itype.EQ.2)THEN
    DO i = 1, n
      k = kn(i)
      xyp(1,i) = -xyz(3,k)
      xyp(2,i) = pgr(igrid)%r(k)
    ENDDO
  ENDIF

!=======================================================================
! Constant Theta
  IF(itype.EQ.4)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = xyz(1,k)
          xyp(2,i) = xyz(2,k)
        ENDDO
      CASE DEFAULT
        DO i = 1, n
          k = kn(i)
!          xyp(1,i) = pgr(igrid)%z(k)
          xyp(1,i) = -xyz(3,k)
          xyp(2,i) = pgr(igrid)%r(k)
        ENDDO
    END SELECT
  ENDIF

!=======================================================================
! Constant Z
  IF(itype.EQ.5)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = xyz(1,k)
          xyp(2,i) = xyz(2,k)
        ENDDO
      CASE DEFAULT
        DO i = 1, n
          k = kn(i)
!          xyp(1,i) = -pgr(igrid)%r(k)*SIN(pgr(igrid)%t0(k))
!          xyp(2,i) = pgr(igrid)%r(k)*COS(pgr(igrid)%t0(k))
          xyp(1,i) = xyz(1,k)
          xyp(2,i) = xyz(2,k)
        ENDDO
    END SELECT
  ENDIF

!=======================================================================
! Constant X
  IF(itype.EQ.6)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = xyz(2,k)
          xyp(2,i) = xyz(3,k)
        ENDDO
      CASE DEFAULT
        DO i = 1, n
          k = kn(i)
!          xyp(1,i) = pgr(igrid)%z(k)
          xyp(1,i) = -xyz(3,k)
!          xyp(2,i) = pgr(igrid)%r(k)*COS(pgr(igrid)%t0(k))
          xyp(2,i) = xyz(2,k)
        ENDDO
    END SELECT
  ENDIF

!=======================================================================
! Constant Y
  IF(itype.EQ.7)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i = 1, n
          k = kn(i)
          xyp(1,i) = xyz(1,k)
          xyp(2,i) = xyz(3,k)
        ENDDO
      CASE DEFAULT
        DO i = 1, n
          k = kn(i)
!          xyp(1,i) = pgr(igrid)%z(k)
          xyp(1,i) = -xyz(3,k)
!          xyp(2,i) = pgr(igrid)%r(k)*SIN(pgr(igrid)%t0(k))
          xyp(2,i) = xyz(1,k)
        ENDDO
    END SELECT
  ENDIF

  RETURN

END SUBROUTINE pVXYPrime
