#include "parameters.h"
!#######################################################################
SUBROUTINE pVZPrime(id, xyz, knode, zp, zprime, xpc, ypc, halfw)
!#######################################################################
  USE common_area, ONLY: pgr, igrid, coordsys, diam
  USE keydata, ONLY: keycut
  USE calc_coord, ONLY: Calc_CoordR, Calc_CoordT, Calc_CoordJ, Calc_CoordI
  IMPLICIT NONE
!
!	return z-prime for programmer defined cutting surfaces
!
!		this routine is called when the programmed cutting plane
!		function key is hit to set up the surface data.
!		the operator may be queried in this routine.
!	        Note: this is call is NOT locked out in unsteady usage!
!
!	where:	id	- (+) first call
!			  (-) grid change (unsteady - iopt > 1)
!		xyz	- coordinate triads for all the 3-D nodes.
!			  this is the same data as set in V3GRID
!		knode   - the number of 3-D nodes.   this is the same
!			  value as set in the call to V3_INIT
!
!		note: id, xyz and knode must not be modified!
!
!		the following paramters are returned by this routine:
!
!		zp	- the calculated z-prime values for the 3-D nodes.
!			  this vector must be completely filled
!		zprime	- the starting z-prime value
!		xpc	- the starting x-prime center value for 
!			  the 2-D window
!		ypc	- the starting y-prime center value for 
!			  the 2-D window
!		halfw	- the starting 1/2 width for the 2-D window in
!			  x-prime/y-prime space.
!
  INTEGER,INTENT(IN) :: id, knode

  REAL,INTENT(IN)    :: xyz(3,knode)
  REAL,INTENT(OUT)   :: zp(knode), zprime, xpc, ypc, halfw

  INTEGER :: i
  INTEGER :: itype,nnode

  REAL :: rave,tave,xave,yave,zave
  REAL :: rdev,tdev,xdev,ydev,zdev

  COMMON /prime/ itype, rave, tave, zave, xave, yave, rdev, tdev, zdev, xdev, ydev, nnode


!
!	R-Theta Cuts
!
  IF(id.GT.0)THEN
#   ifdef PV3SERIAL
1000  CONTINUE
      CALL V3_Cursor(.TRUE.)
      WRITE(*,*) ' Enter cutting plane option:'
      WRITE(*,*) '1  M-Prime Theta'
      WRITE(*,*) '2  R-Z'
      WRITE(*,*) '3  Constant R'
      WRITE(*,*) '4  Constant Theta'
      WRITE(*,*) '5  Constant Z'
      WRITE(*,*) '6  Constant X'
      WRITE(*,*) '7  Constant Y'
      READ(*,*) keycut
      IF(keycut.LT.1 .OR. keycut.GT.6)GOTO 1000
      CALL V3_Cursor(.FALSE.)
      itype = keycut
#   else
      itype = id
#   endif
    CALL MappingCalculations(knode,xyz)
  ENDIF

!=======================================================================
! M Prime-Theta
! This should be a % span cut...
  IF(itype.EQ.1)THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%j))CALL Calc_CoordJ()
    DO i=1,knode
      zp(i) = pgr(igrid)%j(i)
    ENDDO
    xpc=-zave
    ypc=tave*diam
    halfw=(zdev+tdev*diam)/FLOAT(knode)
  ENDIF

!=======================================================================
! R-Z
! This needs to be adjusted to be a %theta cut....
  IF(itype.EQ.2)THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%i))CALL Calc_CoordI()
    DO i=1,knode
      zp(i) = pgr(igrid)%i(i)
    ENDDO
    zprime=(pgr(igrid)%li(1)+pgr(igrid)%gis(1))/2.
    xpc=-zave
    ypc=rave
    halfw=(zdev+rdev)/FLOAT(knode)
  ENDIF

!=======================================================================
! Constant R
  IF(itype.EQ.3)THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%r))CALL Calc_CoordR()
    DO i=1,knode
      zp(i) = pgr(igrid)%r(i)
    ENDDO
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        xpc=xave
        ypc=tave*rave
        halfw=(xdev+rdev*tdev)/FLOAT(knode)
      CASE DEFAULT
        xpc=-zave
        ypc=tave*diam
        halfw=(zdev+tdev*diam)/FLOAT(knode)
    END SELECT
  ENDIF

!=======================================================================
! Constant Theta
  IF(itype.EQ.4)THEN
    IF(.NOT. ASSOCIATED(pgr(igrid)%t0))CALL Calc_CoordT()
    DO i=1,knode
      zp(i) = pgr(igrid)%t0(i)
    ENDDO
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        xpc=zave
      CASE DEFAULT
        xpc=-zave
    END SELECT
    ypc=rave
    halfw=(zdev+rdev)/FLOAT(knode)
  ENDIF

!=======================================================================
! Constant Z
  IF(itype.EQ.5)THEN
    SELECT CASE(coordsys)
      CASE(CARTESIAN)
        DO i=1,knode
          zp(i) = xyz(3,i)
        ENDDO
      CASE DEFAULT
        DO i=1,knode
          zp(i) = -xyz(3,i)
        ENDDO
    END SELECT
    xpc=xave
    ypc=yave
    halfw=(xdev+ydev)/FLOAT(knode)
  ENDIF

!=======================================================================
! Constant X
  IF(itype.EQ.6)THEN
    DO i=1,knode
      zp(i) = xyz(1,i)
    ENDDO
    xpc=zave
    ypc=yave
    halfw=(ydev+zdev)/FLOAT(knode)
  ENDIF

!=======================================================================
! Constant Y
  IF(itype.EQ.7)THEN
    DO i=1,knode
      zp(i) = xyz(2,i)
    ENDDO
    xpc=xave
    ypc=zave
    halfw=(xdev+zdev)/FLOAT(knode)
  ENDIF

  IF(id.GT.0)THEN
#   ifdef PV3SERIAL
      zprime=(MAXVAL(zp)+MINVAL(zp))/2.
#   else
      zprime=0.0
#   endif
  ENDIF

END SUBROUTINE pVZPrime

!#######################################################################
SUBROUTINE MappingCalculations(knode,xyz)
!#######################################################################
  USE common_area, ONLY: pgr, igrid, coordsys
  USE calc_coord, ONLY: Calc_CoordR, Calc_CoordT, Calc_CoordJ, Calc_CoordI
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: knode
  REAL,INTENT(IN) :: xyz(3,*)

  INTEGER :: i
  INTEGER :: itype,nnode

  REAL :: rave,tave,xave,yave,zave
  REAL :: rdev,tdev,xdev,ydev,zdev

  COMMON /prime/ itype, rave, tave, zave, xave, yave, rdev, tdev, zdev, xdev, ydev, nnode

  nnode = knode
  rave = 0.0
  tave = 0.0
  xave = 0.0
  yave = 0.0
  zave = 0.0
  SELECT CASE(coordsys)
    CASE(CARTESIAN)
      IF(.NOT. ASSOCIATED(pgr(igrid)%r))CALL Calc_CoordR
      IF(.NOT. ASSOCIATED(pgr(igrid)%t0))CALL Calc_CoordT
      DO i = 1, knode
        rave = rave + pgr(igrid)%r(i)
        tave = tave + pgr(igrid)%t0(i)
        xave = xave + xyz(1,i)
        yave = yave + xyz(2,i)
        zave = zave + xyz(3,i)
      ENDDO
    CASE DEFAULT
      DO i = 1, knode
        rave = rave + pgr(igrid)%r(i)
        tave = tave + pgr(igrid)%t0(i)
        xave = xave + pgr(igrid)%r(i)*SIN(pgr(igrid)%t0(i))
        yave = yave + -pgr(igrid)%r(i)*COS(pgr(igrid)%t0(i))
        zave = zave + xyz(3,i)
      ENDDO
  END SELECT
  rave = rave/FLOAT(knode)
  tave = tave/FLOAT(knode)
  xave = xave/FLOAT(knode)
  yave = yave/FLOAT(knode)
  zave = zave/FLOAT(knode)

  rdev = 0.0
  tdev = 0.0
  xdev = 0.0
  ydev = 0.0
  zdev = 0.0
  SELECT CASE(coordsys)
    CASE(CARTESIAN)
      DO i = 1, knode
        xdev = xdev + ABS(xyz(1,i)-xave)
        ydev = ydev + ABS(xyz(2,i)-yave)
        zdev = zdev + ABS(xyz(3,i)-zave)
      ENDDO
    CASE DEFAULT
      DO i = 1, knode
        rdev = rdev + ABS(pgr(igrid)%r(i)-rave)
        tdev = tdev + ABS(pgr(igrid)%t0(i)-tave)
        xdev = xdev + ABS(pgr(igrid)%r(i)*SIN(pgr(igrid)%t0(i))-xave)
        ydev = ydev + ABS(pgr(igrid)%r(i)*COS(pgr(igrid)%t0(i))-yave)
        zdev = zdev + ABS(xyz(3,i)-zave)
      ENDDO
  END SELECT

END SUBROUTINE MappingCalculations
