program AVUSINTERP
USE MOD_AVUSINTERP
USE MOD_P3DGRID
USE MOD_OCTREE

IMPLICIT NONE

integer ::  count_rate, count_max, countbeg, countend
real(kind=SP) :: totsec

!Allways display all information on the screen
prints =.true.
!Always attempt to order points
orderpnts = .true.

Call system_clock(countbeg,count_rate,count_max)

Call AVUSINTERP_logo()

Call Inintialize()

Call ReadInput()

Call ReadInGridFiles()

if(interpolate) Call InterpolateRestart()

if(createrstrt) Call WriteInterpolatedSolution()

if(.not.calculatewalldist.and.rstrttypec.eq.1) then
  write(*,*)
  write(*,*) 'Please restart AVUS with 3 as the START OPTION'
  write(*,*) 'in order to recalculate wall distances.'
endif

Call system_clock(countend,count_rate,count_max)
totsec = real(countend - countbeg) / real(count_rate)

write(*,*)
write(*,'(" Total CPU seconds: ")', advance='NO')
write(*,1000) totsec
write(*,*)
write(*,*) 'Done!'

1000	format(1x, es10.3)

end program AVUSInterp


!********************************************
subroutine ReadInGridFiles()
USE MOD_AVUSINTERP
USE MOD_AVUSGrid
USE MOD_P3DGrid
USE MOD_P3DGridRefiner
USE MOD_OCTREE
USE MOD_VSTAGEGrid
USE MOD_P3D2AVUS

IMPLICIT NONE

INTERFACE 
  subroutine InitOctree(narraypnts,dx,nzones,npnts,ndim)
    USE MOD_OCTREE
    integer,intent(in) :: narraypnts,nzones,ndim
    real(kind=FP), dimension(:,:), pointer :: dx
    integer, dimension(:), pointer :: npnts
  end subroutine
END INTERFACE

INTERFACE
  recursive subroutine WriteBuckets(bucket)
    USE MOD_OCTREE
    type(buckettype), pointer :: bucket
  end subroutine WriteBuckets
END INTERFACE

INTEGER :: ierr
integer :: n,np,ndim,nz,ngrds,totpnts,totcells
logical :: readcnct

!type(buckettype), pointer :: bucket


if(interpolate.or..not.interpolate.and.gridtypec.le.0) then
!Read in the parent grid
select case(gridtypep)
case(1)
  Call ReadAvusGrid(parentgridfile,grdformatp,griddoublep,ndmp,nzonesp,npatchp, &
                    dxp,npntsp,nfacesp,ncellsp,mxppfp,mxfpcp,numppfp,ipntp,icellp,.true.)
case(2)
  Call ReadP3DGrid (parentgridfile,grdformatp,griddoublep,nzonesp,ndmp,dxp,iengp,jengp,kengp,iblankp,iblankedp,multiblockp)
  if(gridtypec.gt.0) then
    Call P3D2AVUS(nzonesp,ndmp,npatchp,npntsp,nfacesp,ncellsp,mxppfp,mxfpcp,numppfp,dxp,ipntp,icellp,iengp,jengp,kengp)
  endif

  if(all(kengp.eq.1).and.abs(gridtypec).eq.1) ndmp = 2

  if(iblankedp) then
!Make sure the iblanking is nothing but 0's and 1's
    do n=1,sum(npntsp)
      if(iblankp(n).lt.0) iblankp(n) = 1
      if(iblankp(n).gt.1) iblankp(n) = 1
    enddo
  endif

  select case(p3drstrttypep)
  case(fdlmin:fdlmax)
    nzonesp = 1
    Call ReadP3DRestart(parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp, &
                        dxp,iengp,jengp,kengp,p3drstrttypep,multiblockp)
    if(all(kengp.eq.1)) ndmp = 2
    Call P3D2AVUS(nzonesp,ndmp,npatchp,npntsp,nfacesp,ncellsp,mxppfp,mxfpcp, &
                  numppfp,dxp,ipntp,icellp,iengp,jengp,kengp)
  end select

case(3)
  Call ReadVSTAGEGrid(parentgridfile,grdformatp,griddoublep,nzonesp,ndmp,dxp,iengp,jengp,kengp)
  Call P3D2AVUS(nzonesp,ndmp,npatchp,npntsp,nfacesp,ncellsp,mxppfp,mxfpcp,numppfp,dxp,ipntp,icellp,iengp,jengp,kengp)
end select


else

  select case(gridtypep)
  case(1)
    nullify(npntsp,ncellsp)
    Call ReadAVUSGridHeader(parentgridfile,grdformatp,ndmp,nzonesp,npatchp,npntsp,ncellsp)
!    write(*,*) 'This is an error for now'
!    stop
  case(2)
    Call ReadP3DGridHeader(parentgridfile,grdformatp,ngrds,ndim,multiblockp,totpnts)
    ALLOCATE(iengp(ngrds),jengp(ngrds),kengp(ngrds),stat=ierr)
    ALLOCATE(iengc(ngrds),jengc(ngrds),kengc(ngrds),stat=ierr)
    CALL CheckAllocationError(ierr)
    open(unit=1,file=parentgridfile,form='unformatted')
    read(1) nzonesp
    read(1) (iengp(nz),jengp(nz),kengp(nz),nz=1,nzonesp)
    close(1)
    iengc(:) = iengp(:)
    jengc(:) = jengp(:)
    kengc(:) = kengp(:)
  case(3)
    write(*,*) 'This is an error for now'
    stop
  end select

endif


!Genereate the child grid if it is a refinement of the parent grid
Call GenChildGrid()




!Initialise all the searching arrays
if(interpolate.and.(order.eq.1.or.order.eq.2)) then

  select case(gridtypec)
  case(1)
    Call ReadAVUSGridHeader(childgridfile,grdformatc,ndmc,nzonesc,npatchc,npntsc,ncellsc)
    deallocate(npntsc,ncellsc)
    nullify(npntsc,ncellsc)
  case(2)
    ndmc = 3
    if(gridtypep.eq.2) then
      if(all(kengp.eq.1)) ndmp = 2
    endif
  case(3)
    ndmc = 3
  end select
  ndim = min(ndmp,ndmc)

  Call CreateCellArray()

  Call InitOctree(sum(npntsp),dxp,nzonesp,npntsp,ndim)     !Do this before reading in the child grid, because it will free up memory after completed
  !Now that the buckets are created, sort the parent grid points
  !such that they exist contineously within a bucket
!  Call SortParentGridPoints()
!  Call CreateCellArray()
!  Call ConvertBucketPnts2Cells()
  Call InitHoOctree()
!  Call ConvertBucketCells2Pnts()
endif

!open(unit=1,file='junk.dat')
!write(1,*) 'VARIABLES = X,Y,Z'
!do nz=1,aczones
!  bucket => mainbucket(nz)%bucket
!  Call WriteBuckets(bucket)
!enddo
!close(1)
!stop

if(createrstrt) then
select case(rstrttypep)
case(0)
  Call ReadAVUSPixFile(parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp)
  numrstrtvar(1) = 6+nfree
case(1)
!  allocate(dwp(numrstrtvar,sum(ncellsp)))
!  dwp(:,:) = 1.0_DP
  Call ReadAVUSRstrtFile(parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,ncellsp,dwp)
  if(readallvars) then
    Call Push2Pnt()
  endif
case(2)
!  allocate(dwp(numrstrtvar,sum(ncellsp)))
!  dwp(:,:) = 1.0_DP
  select case(p3drstrttypep)
  case(fdlmin:fdlmax)
  case default
    Call ReadP3DRestart(parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp, &
                        dwp,dxp,iengp,jengp,kengp,p3drstrttypep,multiblockp)
  end select
  if(interpolate) then
!    Call SortParentVariables()
!    if(readallvars) deallocate(sortmap)
  endif
case(3)
  select case(ReadAllVars)
  case(.true.)
    Call ReadVSTAGERestart(parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp,iengp,jengp,kengp)
    Call PushV2Pt(dwp,iengp,jengp,kengp,sum(numrstrtvar))
  case(.false.)
    !The variables will be read in in the interpolation subroutine
  end select
end select
endif


if(order.eq.3) then
  select case(gridtypec)
  case(1)
    Call ReadAVUSGridHeader(childgridfile,grdformatc,ndmc,nzonesc,npatchc,npntsc,ncellsc)
    deallocate(npntsc,ncellsc)
    nullify(npntsc,ncellsc)
  case(2)
    ndmc = 3
    if(gridtypep.eq.2) then
      if(all(kengp.eq.1)) ndmp = 2
    endif
  end select
  ndim = min(ndmp,ndmc)
  Call InitOctree(sum(npntsp),dxp,nzonesp,npntsp,ndim) 
else

  if(interpolate) then
    deallocate(nfacesp,numppfp,ipntp,icellp)
!  Call ConvertBucketPnts2Cells()
  endif
endif



!do n = 1,sum(npntsp(:))
!   dwp(1,n) = n
!enddo
!do n = 2,numrstrtvar
!  dwp(n,:) = dwp(1,:)
!enddo

!Call WriteP3DRestart(childrstrtfile,rstrtformatp,rstrtdoublep,ndmp,nzonesp,npntsp,dwp)
!stop


! Read in thie child grid
if((.not.interpolate.and.gridtypec.gt.0).or.interpolate) then
select case(gridtypec)
case(1)
  readcnct = .true.
  if(rstrttypep.eq.0.and.avusvisfile.ne.2)then
    readcnct = .false.
  end if

  Call ReadAvusGrid(childgridfile,grdformatc,griddoublec,ndmc,nzonesc,npatchc, &
                    dxc,npntsc,nfacesc,ncellsc,mxppfc,mxfpcc,numppfc,ipntc,icellc,readcnct)
case(2)
  Call ReadP3DGrid (childgridfile,grdformatc,griddoublec,nzonesc,ndmc, &
                    dxc,iengc,jengc,kengc,iblankc,iblankedc,multiblockc)


!I don't need all the connectivity informaton for the child grid, just the points and number of points
  allocate(npntsc(nzonesc),stat = ierr)
  Call CheckAllocationError(ierr)
  DO n = 1, nzonesc
    npntsc(n) = iengc(n) * jengc(n) * kengc(n)
  END DO
case(3)
  Call ReadVSTAGEGrid (childgridfile,grdformatc,griddoublec,nzonesc,ndmc,dxc,iengc,jengc,kengc)

!I don't need all the connectivity informaton for the child grid, just the points and number of points
  allocate(npntsc(nzonesc),stat = ierr)
  Call CheckAllocationError(ierr)
  DO n = 1, nzonesc
    npntsc(n) = iengc(n) * jengc(n) * kengc(n)
  END DO

end select
endif

!Normalize the child grid
!do n =1,ndmc
!  dxc(n,:) = (dxc(n,:)-minval(dxc(n,:)))/(maxval(dxc(n,:))-minval(dxc(n,:)))*100
!enddo

if(createrstrt) then
!Allocate memory for the child variables
  select case(readallvars)
  case(.true.)
    nrstrtvarread = sum(numrstrtvar)
    allocate( dwc(nrstrtvarread,sum(npntsc)),stat = ierr)
    Call CheckAllocationError(ierr)
  case(.false.)
    allocate( dwc(1,sum(npntsc)),stat = ierr)
    Call CheckAllocationError(ierr)
  end select  

  if(interpolate) then
    dwc(:,:) = 0.0_DP
  elseif(.not.interpolate) then
    do np=1,sum(npntsc)
      dwc(:,np) = dwp(:,np)
    enddo
  endif
endif




end subroutine ReadInGridFiles


!********************************************
subroutine WriteInterpolatedSolution()
USE MOD_AVUSINTERP
USE MOD_AVUSGrid
USE MOD_P3DGrid
USE MOD_VSTAGEGrid

IMPLICIT NONE

INTERFACE 

  subroutine WriteFVUns(cobc,fvdat,npatch,ndim,npts,nfaces,ncells, &
                        dxyz,ipnt,icell,numppf,griddouble,pixformat,dw)
! cobc - AVUS boundary condition file name
! fvdat - FieldView file name
! ndim - Dimension of the grid (2D or 3D)
! npts - Number of points in the grid
! nfaces - Number of faces in the grid
! ncells - Number of cells in the grid
! dxyz - node coordinates
  use MOD_AVUSGrid
  use fvconv
  logical,intent(in) :: griddouble
  integer, intent(in) :: npatch,ndim,npts,nfaces,ncells,pixformat
  real (kind=FP),dimension(:,:),pointer :: dxyz,dw
  integer, dimension(:,:), pointer :: ipnt,icell
  integer, dimension(:), pointer :: numppf
  character(len=chrlen) :: cobc,fvdat
  end subroutine WriteFVUns

END INTERFACE



character(len=chrlen) :: pixfilename
!integer :: n

if(readallvars) then

  Call ConvertRestart()
  
  select case(rstrttypec)
  case(1)


    !Write out a pix file of the interpolated solution
    Call GetPixFilename(pixfilename)
    select case(avusvisfile)
    case(0) !Don't write an AVUS vizualization file
    case(1)
      Call WriteAVUSPixFile(pixfilename,avusvisformat,rstrtdoublec,nzonesc,npntsc,dwc)
    case(2)
      Call WriteFVUns(childbcfile,pixfilename,npatchc,ndmc,npntsc(1),nfacesc(1),ncellsc(1), &
                      dxc,ipntc,icellc,numppfc,griddoublec,avusvisformat,dwc)
    end select
      
    select case(rstrttypep)
    case(0)
      select case(avusvisfile)
      case(2)
        deallocate(npntsc,nfacesc,mxppfc,mxfpcc,numppfc,ipntc,icellc)
        nullify(npntsc,nfacesc,mxppfc,mxfpcc,numppfc,ipntc,icellc)
      case default
        !nothing should be deallocated because it was never read in
      end select    
      return 
    end select
    
!    write(pixfilename,"(A,A)") trim(childrstrtfile),'.dat'
!    Call WriteTecplotFile(pixfilename,ndmc,nzonesc,dxc,npntsc,nfacesc,ncellsc,mxppfc,mxfpcc,numppfc,ipntc,icellc,dwc)
!    return
  
   if(interpolate) then
     Call CreateChildiface()
     Call Push2Center()
   endif
   if(calculatewalldist) Call CalcWallDist()

   if(associated(dxc)) then
      deallocate(dxc)
      nullify(dxc)
   endif
   deallocate(npntsc,nfacesc,mxppfc,mxfpcc,numppfc,ipntc,icellc)
   nullify(npntsc,nfacesc,mxppfc,mxfpcc,numppfc,ipntc,icellc)
   Call WriteAVUSRstrtFile(childrstrtfile,rstrtformatc,rstrtdoublec,nzonesc,ncellsc,dwc)

  case(2)
!    n=1
!    do while(.not.(childrstrtfile(n:n).eq.'.'.or.childrstrtfile(n:n).eq.''))
!      n=n+1
!    enddo
!    if(childrstrtfile(n:n).eq.'.') n = n-1
!    write(pixfilename,"(A,A)") trim(childrstrtfile(1:n)),'.dat'
!    Call WriteTecplotFile_P3D(pixfilename,nzonesc,dxc,dwc,iengc,jengc,kengc,iblankc,iblankedc)
    Call WriteP3DRestart(childrstrtfile,rstrtformatc,rstrtdoublec,nzonesp, &
                         nzonesc,dwc,dxc,iengc,jengc,kengc,p3drstrttypec,multiblockc)
  case(3)
    Call PushV2Center(dwc,iengc,jengc,kengc,sum(numrstrtvar))
    Call WriteVSTAGERestart(childrstrtfile,rstrtformatc,rstrtdoublec,nzonesc,dwc,iengc,jengc,kengc)
  end select
endif

end subroutine WriteInterpolatedSolution


!********************************************
subroutine InterpolateRestart()
USE MOD_AVUSINTERP
USE MOD_AVUSGrid
USE MOD_P3DGrid
USE MOD_VSTAGEGrid
USE MOD_OCTREE
IMPLICIT NONE

character(len=chrlen) :: pixfilename
INTEGER :: npc,totpnts 
integer ::  count_rate, count_max, countbeg, countend,nvar,IAM,COUNTER
real (kind=SP) :: totsec,prcnt,incr !,prcnt2,incr2
!$ integer :: OMP_GET_THREAD_NUM

IAM = 0
incr = 0.1_SP
!incr2 = 0.01_SP
COUNTER = 0
totpnts = sum(npntsc(:))

Call system_clock(countbeg,count_rate,count_max)


select case(readallvars)
case(.true.)

  write(*,*)
  write(*,*) 'Interpolating All Variables...'
!$OMP PARALLEL PRIVATE(npc,prcnt,IAM) SHARED(dijpnts,linpnts,hopnts,failpnts,COUNTER)
!$ IAM = OMP_GET_THREAD_NUM()
  prcnt = incr
!  prcnt2 = incr2
!$OMP DO
  do npc = 1,totpnts
    Call OctreeSearchInterp(npc)
!$OMP ATOMIC
    COUNTER = COUNTER + 1
!    select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt2.and.IAM.eq.0)
!    case(.true.)
!       write(*,"($A)") '.'
!       prcnt2 = prcnt2 + incr2
!    end select
    select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt.and.IAM.eq.0)
    case(.true.)
      write(*,"(I3,A)") int(prcnt*100),'% Complete...'
!      write(*,"(I3,A)", advance='NO') int(prcnt*100),'% Complete...'
!      write(*,"(SS,3(A,es7.1))") ' DI Pts: ',real(dijpnts), ', LI Pts: ', real(linpnts), ', HO Pts: ',real(hopnts)
      prcnt = prcnt + incr
    end select

  enddo
!$OMP END DO NOWAIT
  do while(COUNTER.lt.totpnts*0.901.and.IAM.eq.0)
!    select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt2.and.IAM.eq.0)
!    case(.true.)
!       write(*,"($A)") '.'
!       prcnt2 = prcnt2 + incr2
!    end select
    select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt.and.IAM.eq.0)
    case(.true.)
      write(*,"(I3,A)") int(prcnt*100),'% Complete...'
!      write(*,"(I3,A)", advance='NO') int(prcnt*100),'% Complete...'
!      write(*,"(SS,3(A,es7.1))") ' DI Pts: ',real(dijpnts), ', LI Pts: ', real(linpnts), ', HO Pts: ',real(hopnts)
      prcnt = prcnt + incr
    end select
  enddo
!$OMP BARRIER
!$OMP END PARALLEL
  write(*,"(I4,A)") 100,'% Complete... '

case(.false.)

  Call GetPixFilename(pixfilename)


  nvar = 1
  open(unit=lu_tmpint,file=tmpinterpfile,form='unformatted')
  invdpnts = 0
  dijpnts = 0
  linpnts = 0
  hopnts = 0
  nptwander = 0

  select case(gridtypep)
  case(1)
!    Call ReadNextAVUSPixVar(nvar,parentpixfile,pixformatp,pixdoublep,nzonesp,npntsp,dwp)
     Call ReadNextAVUSRstrtVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,ncellsp,dwp)
     Call Push2Pnt()
  case(2)
    Call ReadNextP3DVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp, &
                        iengp,jengp,kengp,p3drstrttypep,multiblockp)
  case(3)
    Call ReadNextVSTAGEVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp, &
                        iengp,jengp,kengp)
    Call PushV2Pt(dwp,iengp,jengp,kengp,1)
  end select

!  Call SortParentVariables()

  write(*,*)
  write(*,*) 'Interpolating Variable ',nvar,'...'
!$OMP PARALLEL PRIVATE(npc,prcnt,IAM) SHARED(dijpnts,linpnts,hopnts,failpnts,COUNTER)
!$ IAM = OMP_GET_THREAD_NUM()
  prcnt = incr
  COUNTER = 0
 !$OMP DO
  do npc = 1,totpnts
    Call OctreeSearchInterp(npc)
!$OMP ATOMIC
    COUNTER = COUNTER + 1
    select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt.and.IAM.eq.0)
    case(.true.)
    write(*,"(I3,A)") int(prcnt*100),'% Complete...'
!      write(*,"(I3,A)", advance='NO') int(prcnt*100),'% Complete...'
!      write(*,"(SS,3(A,es7.1))") ' DI Pts: ',real(dijpnts), ', LI Pts: ', real(linpnts), ', HO Pts: ',real(hopnts)
      prcnt = prcnt + incr
    end select
  enddo
!$OMP END DO
!$OMP END PARALLEL
  write(*,"(I4,A)") 100,'% Complete... '

  close(lu_tmpint)
  open(unit=lu_tmpint,file=tmpinterpfile,form='unformatted')


  select case(abs(gridtypec))
  case(1)
     select case(nvar)
     case(1:6)
       Call WriteNextAVUSPixVar(nvar,pixfilename,rstrtformatc,rstrtdoublec,nzonesc,npntsc,dwc)
     end select
     if(nvar.eq.1)  Call CreateChildiface()
     Call Push2Center()
     Call ConvertRestart()
     Call WriteNextAVUSRstrtVar(nvar,childrstrtfile,rstrtformatc,rstrtdoublec,nzonesc,ncellsc,dwc)
  case(2)
    Call ConvertRestart()
    Call WriteNextP3DVar(nvar,childrstrtfile,rstrtdoublec,nzonesc,npntsc,dwc,iengc,jengc,kengc)
  case(3)
    Call WriteNextVSTAGEVar(nvar,childrstrtfile,rstrtdoublec,nzonesc,npntsc,dwc,iengc,jengc,kengc)
  end select
    

  do nvar = 2,sum(numrstrtvar)
    invdpnts = 0
    dijpnts = 0
    linpnts = 0
    hopnts = 0
    nptwander = 0

    select case(gridtypep)
    case(1)
!      Call ReadNextAVUSPixVar(nvar,parentpixfile,pixformatp,pixdoublep,nzonesp,npntsp,dwp)
       Call ReadNextAVUSRstrtVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,ncellsp,dwp)
       Call Push2Pnt()
    case(2)
      Call ReadNextP3DVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp, &
                          iengp,jengp,kengp,p3drstrttypep,multiblockp)
    case(3)
      Call ReadNextVSTAGEVar(nvar,parentrstrtfile,rstrtformatp,rstrtdoublep,nzonesp,npntsp,dwp, &
                          iengp,jengp,kengp)
      Call PushV2Pt(dwp,iengp,jengp,kengp,1)
    end select

!    Call SortParentVariables()

    write(*,*)
    write(*,*) 'Interpolating Variable ',nvar,'...'
!$OMP PARALLEL PRIVATE(npc,prcnt,IAM) SHARED(dijpnts,linpnts,hopnts,failpnts,COUNTER)
!$ IAM = OMP_GET_THREAD_NUM()
   prcnt = incr
   COUNTER = 0
 !$OMP DO
    do npc = 1,totpnts
      Call TmpFileInterp(npc)
!$OMP ATOMIC
      COUNTER = COUNTER + 1
      select case(real(COUNTER,kind=SP)/real(totpnts,kind=SP).ge.prcnt.and.IAM.eq.0)
      case(.true.)
      write(*,"(I3,A)") int(prcnt*100),'% Complete...'
!        write(*,"(I3,A)", advance='NO') int(prcnt*100),'% Complete...'
!        write(*,"(SS,3(A,es7.1))") ' DI Pts: ',real(dijpnts), ', LI Pts: ', real(linpnts), ', HO Pts: ',real(hopnts)
        prcnt = prcnt + incr
      end select
    enddo
!$OMP END DO
!$OMP END PARALLEL
    write(*,"(I4,A)") 100,'% Complete... '

    rewind(lu_tmpint)

    select case(abs(gridtypec))
    case(1)
       select case(nvar)
       case(1:6)
         Call WriteNextAVUSPixVar(nvar,pixfilename,rstrtformatc,rstrtdoublec,nzonesc,npntsc,dwc)
       end select
       if(nvar.eq.1)  Call CreateChildiface()
       Call Push2Center()
       Call ConvertRestart()
       Call WriteNextAVUSRstrtVar(nvar,childrstrtfile,rstrtformatc,rstrtdoublec,nzonesc,ncellsc,dwc)
    case(2)
      Call ConvertRestart()
      Call WriteNextP3DVar(nvar,childrstrtfile,rstrtdoublec,nzonesc,npntsc,dwc,iengc,jengc,kengc)
    case(3)
      Call WriteNextVSTAGEVar(nvar,childrstrtfile,rstrtdoublec,nzonesc,npntsc,dwc,iengc,jengc,kengc)
    end select
  enddo
  
  close(lu_tmpint,STATUS='DELETE')

  select case(gridtypep)
  case(1)
    !The file is already compined for AVUS files
  case(2)
    deallocate(dwc)
    nullify(dwc)
    Call CombineP3DVarFiles(dxc,childrstrtfile,rstrtformatc,rstrtdoublec,nzonesp, &
                            nzonesc,iengc,jengc,kengc,p3drstrttypec,multiblockc)
  case(3)
    deallocate(dwc)
    nullify(dwc)
    Call CombineVSTAGEVarFiles(childrstrtfile,rstrtformatc,rstrtdoublec,nzonesp,nzonesc,iengc,jengc,kengc)
  end select

end select


Call system_clock(countend,count_rate,count_max)
totsec = real(countend - countbeg) / real(count_rate)
write(*,*) 'Interpolation Complete.'
write(*,*)
write(*,'(" Interpolation CPU seconds: ")', advance='NO')
write(*,1000) totsec
write(*,*)
!write(*,1001) ' Number of Failed Search Points          : ',failpnts,'   ',real(failpnts,kind=SP)/real(totpnts,kind=SP)*100,'%'
write(*,1001) ' Number of Inverse Distance Points       : ',invdpnts,'   ',real(invdpnts,kind=SP)/real(totpnts,kind=SP)*100,'%'
write(*,1001) ' Number of Extrapolated Points           : ',extpnts,'   ',real(extpnts,kind=SP)/real(totpnts,kind=SP)*100,'%'
write(*,1001) ' Number of Direct Injection Points       : ',dijpnts,'   ',real(dijpnts,kind=SP)/real(totpnts,kind=SP)*100,'%'
if(order.eq.1.or.order.eq.2) then
write(*,1001) ' Number of Linearly Interpolated Points  : ',linpnts,'   ',real(linpnts,kind=SP)/real(totpnts,kind=SP)*100,'%'
endif
if(order.eq.2) then
write(*,1001) ' Number of Quadratic Interpolated Points : ',hopnts ,'   ',real(hopnts,kind=SP)/real(totpnts,kind=SP)*100 ,'%'
endif
write(*,*)
write(*,1001) ' Number of Points In Neighboring Buckets : ',nptwander,'   ',real(nptwander,kind=SP)/real(totpnts,kind=SP)*100,'%'
!write(*,*) 'Total Distance Wandered :',nwander
write(*,*)

1000	format(1x, es10.3)
1001    format(A,i9,A,f5.1,A)

!Erase all the buckets
Call CleanupOctree()

!Clean up the parent grid, for it is no longer needed
!if(.not.calculatewalldist.and.p3drstrttypec.ne.100) then
!   deallocate(dxc)
!   nullify(dxc)
!endif
deallocate(npntsp, ncellsp, mxppfp, mxfpcp,dxp,dwp)
nullify(npntsp,dxp,dwp)
nullify(nfacesp, ncellsp, mxppfp, mxfpcp)
nullify(numppfp,ipntp,icellp)

!Triangulated cells are no longer needed
select case(order)
case(1,2)
  deallocate(cellnodes,cellneighbors,cnbsptr)
end select


end subroutine InterpolateRestart



