MODULE MOD_AVUSINTERP

USE MOD_GRIDVARS
!Global variables used to interpolate values from a parent grid to a child grid
! variables ending with 'p' are associated with the parant grid
! variables ending with 'c' are associated with the child grid

!Determines if information should be printed to the screen
!only removes statements for use as a library
logical :: prints,orderpnts

!Visualization file format
integer :: avusvisfile,avusvisformat

!Defines the grid type to be read in
! 1 - AVUS Grid
! 2 - Plot3D Grid
integer :: gridtypep,gridtypec
integer :: rstrttypep,rstrttypec

!The order of the interpolation
integer :: order

!Values are interpolated from the parent grid to the child grid
character(len=chrlen) :: parentgridfile, childgridfile,ggbcfile
character(len=chrlen) :: parentrstrtfile, childrstrtfile,childbcfile

!The parent and child grid coordinates in double precision
real (kind=FP),dimension(:,:)    ,pointer :: dxp, dxc, dxcc
real (kind=FP),dimension(:,:)    ,pointer :: dwp, dwc

integer :: ndmp,nzonesp,npatchp
integer, dimension(:), pointer :: npntsp
integer, dimension(:), pointer :: nfacesp, ncellsp, mxppfp, mxfpcp
integer, dimension(:), pointer :: numppfp
integer, dimension(:,:), pointer :: ipntp,icellp
integer :: grdformatp,rstrtformatp
logical :: griddoublep,rstrtdoublep,iblankedp
integer, dimension(:), pointer :: iengp,jengp,kengp,iblankp

integer :: ndmc,nzonesc,npatchc
integer, dimension(:), pointer :: npntsc, nfacesc, ncellsc, mxppfc,  mxfpcc
integer, dimension(:), pointer :: numppfc
integer, dimension(:,:), pointer :: ipntc,icellc
integer :: grdformatc,rstrtformatc
logical :: griddoublec,rstrtdoublec,iblankedc
integer, dimension(:), pointer :: iengc,jengc,kengc,iblankc

integer, dimension(:,:), allocatable :: cellnodes
integer, dimension(:),   allocatable :: cellneighbors,cnbsptr
!integer, dimension(:,:), allocatable :: singpnts
!integer, parameter :: maxsingpnts = 50
integer  :: maxcellnb

!The total number of cells after the parent grid has been triangulated
integer :: ncells

integer, dimension(:,:), allocatable :: iface
integer, dimension(:), allocatable :: numfpcp,numfpcc

!Input error handeling constants
INTEGER, PARAMETER :: err_readerror  = 1
INTEGER, PARAMETER :: err_gridtype = 2
INTEGER, PARAMETER :: err_rstrttype = 3
INTEGER, PARAMETER :: err_filenotexist = 4
INTEGER, PARAMETER :: err_grdformat = 5
INTEGER, PARAMETER :: err_rstrtformat = 6
INTEGER, PARAMETER :: err_p3drstrttype = 7
INTEGER, PARAMETER :: err_childgen = 8
INTEGER, PARAMETER :: err_order = 9
INTEGER, PARAMETER :: err_avusvisfile = 10
INTEGER, PARAMETER :: err_avusvisformat = 11

INTEGER, PARAMETER :: err_parent = 1
INTEGER, PARAMETER :: err_child = 2


!Determans if interpolation should be performed
logical :: interpolate

!Determans if walldistances should be calculated
logical :: calculatewalldist

!The restart file type for P3D files
integer :: p3drstrttypep, p3drstrttypec 
logical :: multiblockp, multiblockc

!Number to check of two double are equal
real(kind=FP) :: DirectInjectMaxRadius = 1e-5
!Number to fudge the check to see if a point is inside a triangular element
real(kind=FP) :: PNTINCELLNUMERR = 1e-5

!Number to check if two double are equal
real(kind=FP),parameter :: SMALL = 1e-32

!The input file unit
integer, parameter :: lu_inp = 5

!The temporary interpolation file
integer, parameter :: lu_tmpint = 11
character(len=chrlen) :: tmpinterpfile = 'tmpinterp.dat'

integer, parameter :: lu_intdij = 1
integer, parameter :: lu_intlin = 2
integer, parameter :: lu_intquad = 3
integer, parameter :: lu_intquadfail = 4
integer, parameter :: lu_intivd = 5

!Gamma, Mach Number, Reynolds Number, Alpha
real(kind=FP) :: Gamma,Mach,Re,Alpha,Pstat,Tstat,Rgas

!Variables related to reading and AVUS BC file
!The maximum number of patches that can exist
integer, parameter :: MAXPAX = 50,SZQ = 8
integer :: NPATCH,NTURB
integer, dimension(MAXPAX) :: IDPATCH
real(kind=FP) :: C0,R0,R0C02,SMU0

!Detremines if a restart file should be read in or not
logical :: createrstrt

END MODULE MOD_AVUSINTERP

!*******************************************
INTEGER FUNCTION ConvertIJKtoIT (igs, i, j, k, idm, jdm)
!*******************************************
!
IMPLICIT NONE

INTEGER, INTENT(IN) :: igs, i, j, k, idm, jdm

 ConvertIJKtoIT = (igs-1) + jdm*idm*(k-1) + idm*(j-1) + i

RETURN
END FUNCTION ConvertIJKtoIT


!*****************************************************************************
SUBROUTINE CheckAllocationError(ierr)
!*****************************************************************************

IMPLICIT NONE

INTEGER, INTENT(IN) :: ierr

if(ierr /= 0) then 
  write(*,*) 'Error allocating Memory'
  write(*,*)
  write(*,*) '  --- CODE HALTED ---'
  stop
end if

END SUBROUTINE CheckAllocationError





