MODULE MOD_AVUSGrid
USE MOD_GRIDVARS
implicit none


real(kind=DP) :: dtime
real(kind=DP) :: dvar, dgamma, drgas
real(kind=DP) :: dr0,drm0,dp0

integer :: nfree
integer :: lastcyc,rstype
character*25 :: string,strfree(20)


integer, dimension(:), allocatable, private :: mfpc  !used in the tecplot writing subroutine

integer, parameter :: icbpixread = 13
integer, parameter :: icbpixwrite = 14
integer, parameter :: icbrstrtread = 15
integer, parameter :: icbrstrtwrite = 16


CONTAINS


!**************************************************************************************************
SUBROUTINE ReadAVUSGrid(gridfilename,grdformat,griddouble,ndm,nzones,npatch,x, &
                        npnts,nfaces,ncells,mxppf,mxfpc,numppf,ipnt,icell,readcnct)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

real (kind=FP),dimension(:,:),pointer :: x

character(len=chrlen),intent(in) :: gridfilename

integer,intent(out) :: ndm,nzones,npatch
integer, dimension(:),pointer :: npnts, nfaces, ncells
integer, dimension(:),pointer :: numppf
integer, dimension(:,:),pointer :: ipnt,icell
integer, dimension(:), pointer :: mxppf,  mxfpc


integer,intent(in) :: grdformat
logical,intent(in) :: griddouble,readcnct


real (kind=SP),dimension(:,:)    ,allocatable ::  sx
real (kind=DP),dimension(:,:)    ,allocatable ::  dx


integer :: icbgrd
integer :: i,j,np,nz,nf,ijunk,nppfjunk,lstpnt,lstface
integer :: ierr
character(len=chrlen) :: fileformat
lstpnt = 1
lstface = 1

Call DispReadWrite(dsp_read,dsp_grid,dsp_avusfile,gridfilename,grdformat,griddouble)

!
! This is where we read the grid. 
!
!
    icbgrd = 13


select case(grdformat)
case(0,2)
  select case(grdformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select

  open(icbgrd,file=gridfilename,form=fileformat,status='old')

  read(icbgrd,ERR=101,END=101) ndm,nzones,npatch

  nz = nzones
  allocate(npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),stat = ierr)
  CALL CheckAllocationError(ierr)

  read(icbgrd,ERR=101,END=101) (npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),nz = 1,nzones)

  if(readcnct) then
    allocate(numppf(sum(nfaces)),ipnt(maxval(mxppf),sum(nfaces)), &
    icell(2,sum(nfaces)),stat=ierr)
    CALL CheckAllocationError(ierr)
  endif

  if(griddouble) then  !Read double precission

    select case(FP)
    case(SP)
      allocate( dx(ndm,sum(npnts)),stat = ierr)
    case(DP)
      allocate( x(ndm,sum(npnts)),stat = ierr)
    end select

    CALL CheckAllocationError(ierr)
    do nz = 1,nzones
      write(*,*) 'Reading Grid Points in Zone ',nz,'...'
      select case(FP)
      case(SP)
        do np=lstpnt,npnts(nz)+lstpnt-1
          read(icbgrd,ERR=101,END=101) (dx(i,np),i=1,ndm)
        enddo
      case(DP)
        do np=lstpnt,npnts(nz)+lstpnt-1
          read(icbgrd,ERR=101,END=101) (x(i,np),i=1,ndm)
        enddo
      end select

      select case(readcnct)
      case(.true.)
        write(*,*) 'Reading Connectivity Information in Zone ',nz,'...'
        do nf=lstface,nfaces(nz) + lstface - 1
          read(icbgrd,ERR=101,END=101) numppf(nf),(ipnt(i,nf),i=1,numppf(nf)),(icell(j,nf),j=1,2)
        enddo
      case(.false.)
        write(*,*) 'Skipping Connectivity Information in Zone ',nz,'...'
        if(nz.eq.1) exit !No need to read it in if there are no more zones
        do nf=1,nfaces(nz)
          read(icbgrd,ERR=101,END=101) nppfjunk,(ijunk,i=1,nppfjunk),(ijunk,j=1,2)
        enddo
      end select
      lstpnt = lstpnt + npnts(nz)
      lstface = lstface + nfaces(nz)
    enddo
	
    select case(FP)
    case(SP)
      allocate( x(ndm,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      x(:,:) = dx(:,:)
      DEALLOCATE(dx)
    end select



  else !Read single precission





    select case(FP)
    case(SP)
      allocate( x(ndm,sum(npnts)),stat = ierr)
    case(DP)
      allocate( sx(ndm,sum(npnts)),stat = ierr)
    end select
    CALL CheckAllocationError(ierr)

    do nz = 1,nzones
      write(*,*) 'Reading Grid Points in Zone ',nz,'...'
      
      select case(FP)
      case(SP)
        do np=lstpnt,npnts(nz)+lstpnt-1
          read(icbgrd,ERR=101,END=101) (x(i,np),i=1,ndm)
        enddo
      case(DP)
        do np=lstpnt,npnts(nz)+lstpnt-1
          read(icbgrd,ERR=101,END=101) (sx(i,np),i=1,ndm)
        enddo
      end select
  
      select case(readcnct)
      case(.true.)
        write(*,*) 'Reading Connectivity Information in Zone ',nz,'...'
        do nf=lstface,nfaces(nz) + lstface - 1
          read(icbgrd,ERR=101,END=101) numppf(nf),(ipnt(i,nf),i=1,numppf(nf)),(icell(j,nf),j=1,2)
        enddo
      case(.false.)
        write(*,*) 'Skipping Connectivity Information in Zone ',nz,'...'
        if(nz.eq.1) exit !No need to read it in if there are no more zones
        do nf=1,nfaces(nz) 
          read(icbgrd,ERR=101,END=101) nppfjunk,(ijunk,i=1,nppfjunk),(ijunk,j=1,2)
        enddo
      end select
      lstpnt = lstpnt + npnts(nz)
      lstface = lstface + nfaces(nz)
    enddo

    select case(FP)
    case(DP)
      allocate( x(ndm,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      x(:,:) = sx(:,:)
      DEALLOCATE(sx)
    end select



  endif
case(1) ! Read a formatted grid

  open(icbgrd,file=gridfilename,form='formatted',status='old')
  read(icbgrd,*,ERR=101,END=101) ndm,nzones,npatch

  nz = nzones
  allocate(npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),stat = ierr)
  CALL CheckAllocationError(ierr)

  do nz = 1,nzones
    read(icbgrd,*,ERR=101,END=101) npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz)
  enddo

  allocate(numppf(sum(nfaces)),ipnt(maxval(mxppf),sum(nfaces)), &
  icell(2,sum(nfaces)),stat=ierr)
  CALL CheckAllocationError(ierr)

  allocate( x(ndm,sum(npnts)),stat = ierr)
  CALL CheckAllocationError(ierr)

  do nz = 1,nzones
    write(*,*) 'Reading Grid Points in Zone ',nz,'...'
    do np=lstpnt,npnts(nz)+lstpnt-1
      read(icbgrd,*,ERR=101,END=101) (x(i,np),i=1,ndm)
    enddo

    select case(readcnct)
    case(.true.)
      write(*,*) 'Reading Connectivity Information in Zone ',nz,'...'
      do nf=lstface,nfaces(nz) + lstface - 1
        read(icbgrd,*,ERR=101,END=101) numppf(nf),(ipnt(i,nf),i=1,numppf(nf)),(icell(j,nf),j=1,2)
      enddo
    case(.false.)
      write(*,*) 'Skipping Connectivity Information in Zone ',nz,'...'
      if(nz.eq.1) exit !No need to read it in if there are no more zones
      do nf=1,nfaces(nz)
        read(icbgrd,*,ERR=101,END=101) nppfjunk,(ijunk,i=1,nppfjunk),(ijunk,j=1,2)
      enddo
    end select
    lstpnt = lstpnt + npnts(nz)
    lstface = lstface + nfaces(nz)
  enddo
end select


close(icbgrd)

if(.not.readcnct) deallocate(nfaces,ncells,mxppf,mxfpc)

return

101 continue

Call DispFileReadError()


END SUBROUTINE ReadAVUSGrid

!**************************************************************************************************
SUBROUTINE ReadAVUSGridHeader(gridfilename,grdformat,ndm,nzones,npatch,npnts,ncells)
!**************************************************************************************************

character(len=chrlen),intent(in) :: gridfilename

integer,intent(out) :: ndm,nzones,npatch
integer,intent(in) :: grdformat
!integer,intent(out),optional :: totpnts,totcells

integer, dimension(:), pointer :: npnts,ncells
integer, dimension(:), allocatable :: nfaces,mxppf,mxfpc

integer :: icbgrd,nz,ierr
character(len=chrlen) :: fileformat

icbgrd = 14
select case(grdformat)
case(0,2)

  select case(grdformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select

  open(icbgrd,file=gridfilename,form=fileformat,status='old')

  read(icbgrd,ERR=101,END=101) ndm,nzones,npatch

case(1)
  open(icbgrd,file=gridfilename,form='formatted',status='old')
  read(icbgrd,*,ERR=101,END=101) ndm,nzones,npatch

end select


nz = nzones
allocate(npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),stat = ierr)
CALL CheckAllocationError(ierr)

select case(grdformat)
case(0,2)
  read(icbgrd,ERR=101,END=101) (npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),nz = 1,nzones)
case(1)
  read(icbgrd,*,ERR=101,END=101) (npnts(nz),nfaces(nz),ncells(nz),mxppf(nz),mxfpc(nz),nz = 1,nzones)
end select

deallocate(nfaces,mxppf,mxfpc)

close(icbgrd)

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadAVUSGridHeader
!**************************************************************************************************
SUBROUTINE ReadAVUSRstrtFile(rstrtfilename,rstrtformat,rstrtdouble,nzones,ncells,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: ncells

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble


real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=SP) :: stime
real(kind=SP) :: sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0

INTEGER :: ierr,lstcell
INTEGER :: i,nz,nc
character(len=chrlen) :: fileformat
ierr=0

Call DispReadWrite(dsp_read,dsp_rstrt,dsp_avusfile,rstrtfilename,rstrtformat,rstrtdouble)

select case(rstrtformat)
case(0,2)
  select case(rstrtformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select

  open(icbrstrtread,file=rstrtfilename,form=fileformat,status='old')

  if(rstrtdouble) then

    read (icbrstrtread,ERR=101,END=101) dtime,lastcyc,rstype
    read (icbrstrtread,ERR=101,END=101) dgamma,drgas
    read (icbrstrtread,ERR=101,END=101) dr0,drm0,dp0
    
    select case(readallvars)
    case(.true.)
      allocate( w(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)
    case(.false.)
      allocate( w(1,sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)
      return
    end select

    select case(FP)
    case(SP)
      allocate( dw(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Reading Variable ',i,' in Zone ',nz,'...'
          read (icbrstrtread,ERR=101,END=101) (dw(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo 
      
      do nc = 1,sum(ncells)
        w(:,nc) = dw(:,nc)
      enddo    
      deallocate(dw)

    case(DP)
      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Reading Variable ',i,' in Zone ',nz,'...'
          read (icbrstrtread,ERR=101,END=101) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo        
    end select



  
  else
  
    read (icbrstrtread,ERR=101,END=101) stime,lastcyc,rstype
    read (icbrstrtread,ERR=101,END=101) sgamma,srgas
    read (icbrstrtread,ERR=101,END=101) sr0,srm0,sp0
    
    dtime    = stime
    dgamma   = sgamma
    drgas    = srgas
    dr0      = sr0
    drm0     = srm0
    dp0      = sp0

    select case(readallvars)
    case(.true.)
      allocate( w(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)
    case(.false.)
      allocate( w(1,sum(ncells)),stat = ierr)  !This should be dw
      CALL CheckAllocationError(ierr)
      return
    end select
    
    select case(FP)
    case(SP)
      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Reading Variable ',i,' in Zone ',nz,'...'
          read (icbrstrtread,ERR=101,END=101) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo        
    case(DP)
      allocate( dw(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)
   
      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Reading Variable ',i,' in Zone ',nz,'...'
          read (icbrstrtread,ERR=101,END=101) (dw(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo
      
      do nc = 1,sum(ncells)   
        w(:,nc) = dw(:,nc)
      enddo
      DEALLOCATE(dw)
    end select

  endif
  
case(1)
  open(icbrstrtread,file=rstrtfilename,form='formatted',status='old')

  read (icbrstrtread,*,ERR=101,END=101) dtime,lastcyc,rstype
  read (icbrstrtread,*,ERR=101,END=101) dgamma,drgas
  read (icbrstrtread,*,ERR=101,END=101) dr0,drm0,dp0
 
  select case(readallvars)
  case(.true.)
    allocate( w(numrstrtvar(1),sum(ncells)),stat = ierr)
    CALL CheckAllocationError(ierr)
  case(.false.)
    allocate( w(1,sum(ncells)),stat = ierr)
    CALL CheckAllocationError(ierr)
    return
  end select

  do i=1,numrstrtvar(1)
    lstcell = 1
    do nz=1,nzones
      write(*,*) 'Reading Variable ',i,' in Zone ',nz,'...'
      read (icbrstrtread,*,ERR=101,END=101) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
      if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
    enddo
  enddo        

end select

42 close(icbrstrtread)

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadAVUSRstrtFile

!**************************************************************************************************
SUBROUTINE ReadNextAVUSRstrtVar(nvar,rstrtfilename,rstrtformat,rstrtdouble,nzones,ncells,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones,nvar
integer, dimension(:),pointer :: ncells

real (kind=FP),dimension(:,:)    ,pointer :: w

integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble


real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

INTEGER :: ierr,lstcell
INTEGER :: nz,nc

ierr=0
lstcell = 1

Call DispReadWrite(dsp_read,dsp_rstrt,dsp_avusfile,rstrtfilename,rstrtformat,rstrtdouble,nvar=nvar)

deallocate(dw)
allocate( dw(1,sum(ncells)),stat = ierr)
CALL CheckAllocationError(ierr)

if (rstrtformat == 0) then

  if(rstrtdouble) then
    select case(FP)
    case(SP)

      allocate( dw(1,sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do  nz=1,nzones
        read (icbrstrtread,ERR=101,END=101) (dw(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo

      do nc= 1,sum(ncells)
        w(:,nc) = dw(:,nc)
      enddo
      deallocate(dw)

    case(DP)
      do  nz=1,nzones
        read (icbrstrtread,ERR=101,END=101) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
    end select
  else
  
    select case(FP)
    case(SP)
      do  nz=1,nzones
        read (icbrstrtread,ERR=101,END=101) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
    case(DP)
      allocate( sw(1,sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do  nz=1,nzones
        read (icbrstrtread,ERR=101,END=101) (sw(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
     
      do nc = 1,sum(ncells)
        w(:,nc) = sw(:,nc)
      enddo
      deallocate(sw)  
    end select

  endif
  
else

  do  nz=1,nzones
    read (icbrstrtread,*,ERR=101,END=101) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
    if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
  enddo

endif

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadNextAVUSRstrtVar


!**************************************************************************************************
SUBROUTINE WriteAVUSRstrtFile(rstrtfilename,rstrtformat,rstrtdouble,nzones,ncells,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: ncells

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble


real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=SP) :: stime
real(kind=SP) :: sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0
INTEGER :: ierr,lstcell
INTEGER :: i,nz,nc
character(len=chrlen) :: fileformat

ierr=0

Call DispReadWrite(dsp_write,dsp_rstrt,dsp_avusfile,rstrtfilename,rstrtformat,rstrtdouble)


select case(rstrtformat)
case(0,2)
  select case(rstrtformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select

  open(icbrstrtwrite,file=rstrtfilename,form=fileformat,status='unknown')

  if(rstrtdouble) then

    write (icbrstrtwrite) dtime,lastcyc,rstype
    write (icbrstrtwrite) dgamma,drgas
    write (icbrstrtwrite) dr0,drm0,dp0
    
    select case(FP)
    case(SP)
      allocate( dw(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do nc= 1,sum(ncells)
        dw(:,nc) = w(:,nc)
      enddo


      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Writing Variable ',i,' in Zone ',nz,'...'
          write (icbrstrtwrite) (dw(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo       
       
      deallocate(dw)

    case(DP)
      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Writing Variable ',i,' in Zone ',nz,'...'
          write (icbrstrtwrite) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo        
    end select
  
  else

    stime    = dtime
    sgamma   = dgamma
    srgas    = drgas
    sr0      = dr0
    srm0     = drm0
    sp0      = dp0

    write (icbrstrtwrite) stime,lastcyc,rstype
    write (icbrstrtwrite) sgamma,srgas
    write (icbrstrtwrite) sr0,srm0,sp0


    select case(FP)
    case(SP)
      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Writing Variable ',i,' in Zone ',nz,'...'
          write (icbrstrtwrite) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo        
    case(DP)
      allocate( sw(numrstrtvar(1),sum(ncells)),stat = ierr)
      CALL CheckAllocationError(ierr)
  
      do nc = 1,sum(ncells)
        sw(:,nc) = w(:,nc)
      enddo

      do i=1,numrstrtvar(1)
        lstcell = 1
        do nz=1,nzones
          write(*,*) 'Writing Variable ',i,' in Zone ',nz,'...'
          write (icbrstrtwrite) (sw(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
          if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
        enddo
      enddo        

      deallocate(sw)
    end select
  endif
  
case(1)
  open(icbrstrtwrite,file=rstrtfilename,form='formatted',status='unknown')

  write (icbrstrtwrite,*) dtime,lastcyc,rstype
  write (icbrstrtwrite,*) dgamma,drgas
  write (icbrstrtwrite,*) dr0,drm0,dp0
 
  do i=1,numrstrtvar(1)
    lstcell = 1
    do nz=1,nzones
      write(*,*) 'Writing Variable ',i,' in Zone ',nz,'...'
      write (icbrstrtwrite,*) (w(i,nc),nc=lstcell,ncells(nz)+lstcell-1)
      if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
    enddo
  enddo        

end select

42 close(icbrstrtwrite)


END SUBROUTINE WriteAVUSRstrtFile

!**************************************************************************************************
SUBROUTINE WriteNextAVUSRstrtVar(nvar,rstrtfilename,rstrtformat,rstrtdouble,nzones,ncells,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

integer, intent(in) :: nvar
character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: ncells

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble

real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=SP) :: stime
real(kind=SP) :: sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0
character(len=chrlen) :: fileformat
INTEGER :: ierr
INTEGER :: nz,nc,lstcell


Call DispReadWrite(dsp_write,dsp_rstrt,dsp_avusfile,rstrtfilename,rstrtformat,rstrtdouble,nvar=nvar)

ierr=0
lstcell = 1

select case(rstrtformat) 
case(0,2)

  if(nvar.eq.1) then
    select case(rstrtformat)
    case(0)
      fileformat = 'unformatted'
    case(2)
      fileformat = 'binary'
    end select
  endif
  
  open(icbrstrtwrite,file=rstrtfilename,form=fileformat,status='unknown')

  if(rstrtdouble) then

    if(nvar.eq.1) then
      write (icbrstrtwrite) dtime,lastcyc,rstype
      write (icbrstrtwrite) dgamma,drgas
      write (icbrstrtwrite) dr0,drm0,dp0
    endif
  
    select case(FP)
    case(SP)
      allocate(dw(1,sum(ncells)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do nc=1,sum(ncells)
         dw(:,nc) = w(:,nc)
      enddo

      do nz=1,nzones
        write (icbrstrtwrite) (dw(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo

      deallocate(dw)
    case(DP)
      do nz=1,nzones
        write (icbrstrtwrite) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
    end select
  
  else

    if(nvar.eq.1) then
      stime   = dtime
      sgamma   = dgamma
      srgas    = drgas
      sr0      = dr0
      srm0     = drm0
      sp0      = dp0
    
      write (icbrstrtwrite) stime,lastcyc,rstype
      write (icbrstrtwrite) sgamma,srgas
      write (icbrstrtwrite) sr0,srm0,sp0
    endif

    select case(FP)
    case(SP)
      do nz=1,nzones
        write (icbrstrtwrite) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
    case(DP)
      allocate(sw(1,sum(ncells)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do nc=1,sum(ncells)
        sw(:,nc) = w(:,nc)
      enddo

      do nz=1,nzones
        write (icbrstrtwrite) (sw(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
        if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
      enddo
    
      deallocate(sw)
    end select
  endif

case(1)
  if(nvar.eq.1) then
  
    open(icbrstrtwrite,file=rstrtfilename,form='formatted',status='unknown')

    write (icbrstrtwrite,*) dtime,lastcyc,rstype
    write (icbrstrtwrite,*) dgamma,drgas
    write (icbrstrtwrite,*) dr0,drm0,dp0
  endif

  lstcell = 1
  do nz=1,nzones
    write (icbrstrtwrite,*) (w(1,nc),nc=lstcell,ncells(nz)+lstcell-1)
    if(nz.gt.1) lstcell = sum(ncells(1:nz-1)) + 1
  enddo

end select


END SUBROUTINE WriteNextAVUSRstrtVar




!**************************************************************************************************
SUBROUTINE ReadAVUSPixFile(pixfilename,pixformat,pixdouble,nzones,npnts,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: pixfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: pixformat
logical,intent(in) :: pixdouble


real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=DP) :: djunk
real(kind=SP) :: sjunk
real(kind=SP) :: svar, sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0
INTEGER :: ierr,lstpnt
INTEGER :: i,n,nz,np
character(len=chrlen) :: fileformat
ierr=0

Call DispReadWrite(dsp_read,dsp_pix,dsp_avusfile,pixfilename,pixformat,pixdouble)


select case(pixformat) 
case(0,2)
  select case(pixformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select
  
  open(icbpixread,file=pixfilename,form=fileformat,status='old')
  
  if(pixdouble) then

    read (icbpixread,ERR=101,END=101) dvar,dgamma,drgas
    read (icbpixread,ERR=101,END=101) dr0,drm0,dp0
    read (icbpixread,ERR=101,END=101) string
    do  i=1,6
      do  nz=1,nzones
        read (icbpixread,ERR=101,END=101) (djunk,np=1,npnts(nz))
      enddo
    enddo
    read (icbpixread ,ERR=8,END=8) nfree

    GOTO 18
8   continue
    nfree = 0

18  continue
    
    npixvar = nfree+6

    select case(readallvars)
    case(.true.)
      allocate( w(npixvar,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      rewind(icbpixread)
    case(.false.)
      allocate( w(1,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      rewind(icbpixread)
      return
    end select
    
    
    read (icbpixread,ERR=101,END=101) dvar,dgamma,drgas
    read (icbpixread,ERR=101,END=101) dr0,drm0,dp0
    read (icbpixread,ERR=101,END=101) string

    if(.not.readallvars) return


    select case(FP)
    case(SP)
      allocate( dw(npixvar,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (dw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo

      read (icbpixread,ERR=40,END=40) nfree

      do n=1,nfree
        read (icbpixread,ERR=101,END=101) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (dw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo

40    continue

      do np=1,sum(npnts)
        w(:,np) = dw(:,np)
      enddo

      deallocate(dw)
    case(DP)
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo

      read (icbpixread,ERR=42,END=42) nfree

      do n=1,nfree
        read (icbpixread,ERR=101,END=101) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
    end select





  else
  
    read (icbpixread,ERR=101,END=101) svar,sgamma,srgas
    read (icbpixread,ERR=101,END=101) sr0,srm0,sp0
    read (icbpixread,ERR=101,END=101) string
    do  i=1,6
      do  nz=1,nzones
        read (icbpixread,ERR=101,END=101) (sjunk,np=1,npnts(nz))
      enddo
    enddo
    read (icbpixread ,ERR=9,END=9) nfree

    GOTO 19
9   continue
    nfree = 0

19  continue
    
    npixvar = nfree+6
    select case(readallvars)
    case(.true.)
      allocate( w(npixvar,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      rewind(icbpixread)
    case(.false.)
      allocate( w(1,sum(npnts)),stat = ierr)  !This should be dw
      CALL CheckAllocationError(ierr)
      rewind(icbpixread)
      return
    end select
    
    read (icbpixread,ERR=101,END=101) svar,sgamma,srgas
    read (icbpixread,ERR=101,END=101) sr0,srm0,sp0
    read (icbpixread,ERR=101,END=101) string

    dvar   = svar
    dgamma = sgamma
    drgas  = srgas
    dr0    = sr0
    drm0   = srm0
    dp0    = sp0

    if(.not.readallvars) return


    select case(FP)
    case(SP)
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo


      read (icbpixread,ERR=42,END=42) nfree

      do n=1,nfree
        read (icbpixread,ERR=101,END=101) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
    case(DP)

      allocate( sw(npixvar,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (sw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo


      read (icbpixread,ERR=41,END=41) nfree

      do n=1,nfree
        read (icbpixread,ERR=101,END=101) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do  nz=1,nzones
          write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
          read (icbpixread,ERR=101,END=101) (sw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo

41    continue
    
      do np=1,sum(npnts)
        w(:,np) = sw(:,np)
      enddo
      DEALLOCATE(sw)
    end select

  endif
  
case(1)
  open(icbpixread,file=pixfilename,form='formatted',status='old')

  read (icbpixread,*,ERR=101,END=101) dvar,dgamma,drgas
  read (icbpixread,*,ERR=101,END=101) dr0,drm0,dp0
  read (icbpixread,'(a25)',ERR=101,END=101) string
  do  i=1,6
    do  nz=1,nzones
      read (icbpixread,*,ERR=101,END=101) (djunk,np=1,npnts(nz))
    enddo
  enddo
  read (icbpixread,* ,ERR=10,END=10) nfree

  GOTO 20
10 continue
  nfree = 0

20 continue
  
  npixvar = nfree+6
  select case(readallvars)
  case(.true.)
    allocate( w(npixvar,sum(npnts)),stat = ierr)
    CALL CheckAllocationError(ierr)
    rewind(icbpixread)
  case(.false.)
    allocate( w(1,sum(npnts)),stat = ierr)
    CALL CheckAllocationError(ierr)
    rewind(icbpixread)
  end select

  read (icbpixread,*,ERR=101,END=101) dvar,dgamma,drgas
  read (icbpixread,*,ERR=101,END=101) dr0,drm0,dp0
  read (icbpixread,'(a25)',ERR=101,END=101) string

  if(.not.readallvars) return

  do i=1,6
    lstpnt = 1
    do  nz=1,nzones
      write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
      read (icbpixread,*,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
      if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
    enddo
  enddo
  
  read (icbpixread,*,ERR=42,END=42) nfree
  do n=1,nfree
    read (icbpixread,'(a25)',ERR=101,END=101) strfree(n)
  enddo

  do i=7,nfree+6
    lstpnt = 1
    do  nz=1,nzones
      write(*,*) 'Reading Pix Variable ',i,' in Zone ',nz,'...'
      read (icbpixread,*,ERR=101,END=101) (w(i,np),np=lstpnt,npnts(nz))
      if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
    enddo
  enddo

end select

42 close(icbpixread)

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadAVUSPixFile

!**************************************************************************************************
SUBROUTINE ReadNextAVUSPixVar(nvar,pixfilename,pixformat,pixdouble,nzones,npnts,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: pixfilename
integer,intent(in) :: nzones,nvar
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: pixformat
logical,intent(in) :: pixdouble


real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

INTEGER :: ierr
INTEGER :: n,nz,np,lstpnt

ierr=0
lstpnt = 1

Call DispReadWrite(dsp_read,dsp_pix,dsp_avusfile,pixfilename,pixformat,pixdouble,nvar=nvar)

if (pixformat.eq.0.or.pixformat.eq.2) then

  if(nvar.eq.7) then
    read (icbpixread,ERR=101,END=101) nfree
    do n=1,nfree
      read (icbpixread,ERR=101,END=101) strfree(n)
    enddo 
  endif

  if(pixdouble) then


    select case(FP)
    case(SP)
      allocate( sw(1,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do nz=1,nzones
        read (icbpixread,ERR=101,END=101) (sw(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo
      do np=1,sum(npnts)
        w(:,np) = sw(:,np)
      enddo
      deallocate(sw)
    case(DP)

      do nz=1,nzones
        read (icbpixread,ERR=101,END=101) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo

    end select
  
  else
  
    select case(FP)
    case(SP)
      do nz=1,nzones
        read (icbpixread,ERR=101,END=101) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo
    case(DP)
      allocate( dw(1,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)

      do nz=1,nzones
        read (icbpixread,ERR=101,END=101) (dw(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
      enddo

      do np=1,sum(npnts)
        w(:,np) = dw(:,np)
      enddo
      deallocate(dw)
    end select
  endif
  
else


  if(nvar.eq.7) then
    read (icbpixread,*,ERR=101,END=101) nfree
    do n=1,nfree
      read (icbpixread,'(a25)',ERR=101,END=101) strfree(n)
    enddo
  endif

  do nz=1,nzones
    read (icbpixread,*,ERR=101,END=101) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
    if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
  enddo

endif

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadNextAVUSPixVar

!**************************************************************************************************
SUBROUTINE WriteAVUSPixFile(pixfilename,pixformat,pixdouble,nzones,npnts,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: pixfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: pixformat
logical,intent(in) :: pixdouble

real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=SP) :: svar, sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0
INTEGER :: ierr
INTEGER :: i,n,nz,np,lstpnt
character(len=chrlen) :: fileformat


Call DispReadWrite(dsp_write,dsp_pix,dsp_avusfile,pixfilename,pixformat,pixdouble)

ierr=0
select case(pixformat)
case(0,2)
  select case(pixformat)
  case(0)
    fileformat = 'unformatted'
  case(2)
    fileformat = 'binary'
  end select

  open(icbpixwrite,file=pixfilename,form=fileformat,status='unknown')

  if(pixdouble) then
    write (icbpixwrite) dvar,dgamma,drgas
    write (icbpixwrite) dr0,drm0,dp0
    write (icbpixwrite) string

    select case(FP)
    case(SP)
      
      allocate(dw(npixvar,sum(npnts)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do np=1,sum(npnts)
        dw(:,np) = dble(w(:,np))
      enddo
      
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write (icbpixwrite) (dw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
 
      if(nfree.gt.0) write (icbpixwrite) nfree
 
      do n=1,nfree
        write (icbpixwrite) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do nz=1,nzones
          write (icbpixwrite) (dw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo

      deallocate(dw)
    case(DP)
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write (icbpixwrite) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
 
      if(nfree.gt.0) write (icbpixwrite) nfree
 
      do n=1,nfree
        write (icbpixwrite) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do nz=1,nzones
          write (icbpixwrite) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
    end select
  
  else
  
    svar = dvar
    sgamma = dgamma
    srgas = drgas
    sr0 = dr0
    srm0 = drm0
    sp0 = dp0
    
    write (icbpixwrite) svar,sgamma,srgas
    write (icbpixwrite) sr0,srm0,sp0
    write (icbpixwrite) string

    select case(FP)
    case(SP)
      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write (icbpixwrite) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
 
      if(nfree.gt.0) write (icbpixwrite) nfree
 
      do n=1,nfree
        write (icbpixwrite) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do nz=1,nzones
          write (icbpixwrite) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
    case(DP)
      allocate(sw(npixvar,sum(npnts)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do np = 1,sum(npnts)
        sw(:,np) = w(:,np)
      enddo

      do  i=1,6
        lstpnt = 1
        do  nz=1,nzones
          write (icbpixwrite) (sw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
        enddo
      enddo
      if(nfree.gt.0) write (icbpixwrite) nfree
      do n=1,nfree
        write (icbpixwrite) strfree(n)
      enddo

      do i=7,nfree+6
        lstpnt = 1
        do nz=1,nzones
          write (icbpixwrite) (sw(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
          if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
       enddo
      enddo
    
      deallocate(sw)
    end select
  endif

case(1)
  open(icbpixwrite,file=pixfilename,form='formatted',status='unknown')

  write (icbpixwrite,*) dvar,dgamma,drgas
  write (icbpixwrite,*) dr0,drm0,dp0
  write (icbpixwrite,'(a25)') string
  do  i=1,6
    lstpnt = 1
    do  nz=1,nzones
      write (icbpixwrite,*) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
      if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
    enddo
  enddo
  if(nfree.gt.0) write (icbpixwrite,*) nfree
  do n=1,nfree
    write (icbpixwrite,'(a25)') strfree(n)
  enddo

  do i=7,nfree+6
    lstpnt = 1
    do nz=1,nzones
      write (icbpixwrite,*) (w(i,np),np=lstpnt,npnts(nz)+lstpnt-1)
      if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
    enddo
  enddo

end select

close(icbpixwrite)


END SUBROUTINE WriteAVUSPixFile

!**************************************************************************************************
SUBROUTINE WriteNextAVUSPixVar(nvar,pixfilename,pixformat,pixdouble,nzones,npnts,w)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

integer, intent(in) :: nvar
character(len=chrlen),intent(in) :: pixfilename
integer,intent(in) :: nzones
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: pixformat
logical,intent(in) :: pixdouble

real (kind=SP),dimension(:,:)    ,allocatable :: sw
real (kind=DP),dimension(:,:)    ,allocatable :: dw

real(kind=SP) :: svar, sgamma, srgas
real(kind=SP) :: sr0,srm0,sp0
INTEGER :: ierr
INTEGER :: n,nz,np,lstpnt
character(len=chrlen) :: fileformat


Call DispReadWrite(dsp_write,dsp_pix,dsp_avusfile,pixfilename,pixformat,pixdouble,nvar=nvar)

ierr=0
select case(pixformat) 
case(0,2)
  if(nvar.eq.1) then
    select case(pixformat)
    case(0)
      fileformat = 'unformatted'
    case(2)
      fileformat = 'binary'
    end select
  endif

  open(icbpixwrite,file=pixfilename,form=fileformat,status='unknown')

  if(nvar.eq.7) then 
    write (icbpixwrite) nfree
    do n=1,nfree
      write (icbpixwrite) strfree(n)
    enddo
  endif

  if(pixdouble) then

    if(nvar.eq.1) then
      write (icbpixwrite) dvar,dgamma,drgas
      write (icbpixwrite) dr0,drm0,dp0
      write (icbpixwrite) string
    endif

    select case(FP)
    case(SP)
      allocate(dw(1,sum(npnts)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do np=1,sum(npnts)
        dw(:,np) = w(:,np)
      enddo
      
      lstpnt = 1
      do  nz=1,nzones
        write (icbpixwrite) (dw(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo

      deallocate(dw)
    case(DP)
      lstpnt = 1
      do  nz=1,nzones
        write (icbpixwrite) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo
    end select
  
  else
  
    if(nvar.eq.1) then
      svar = dvar
      sgamma = dgamma
      srgas = drgas
      sr0 = dr0
      srm0 = drm0
      sp0 = dp0
    
      write (icbpixwrite) svar,sgamma,srgas
      write (icbpixwrite) sr0,srm0,sp0
      write (icbpixwrite) string
    endif

    select case(FP)
    case(SP)
      lstpnt = 1
      do  nz=1,nzones
        write (icbpixwrite) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo
    case(DP)
      allocate(sw(1,sum(npnts)),stat=ierr)
      CALL CheckAllocationError(ierr)

      do np=1,sum(npnts)
        sw(:,np) = w(:,np)
      enddo
      
      lstpnt = 1
      do  nz=1,nzones
        write (icbpixwrite) (sw(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
        if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
      enddo

      deallocate(sw)
    end select
  endif

case(1)
  if(nvar.eq.1) then
  
    open(icbpixwrite,file=pixfilename,form='formatted',status='unknown')

    write (icbpixwrite,*) dvar,dgamma,drgas
    write (icbpixwrite,*) dr0,drm0,dp0
    write (icbpixwrite,'(a25)') string
  endif

  if(nvar.eq.7) then
    write (icbpixwrite,*) nfree
    do n=1,nfree
      write (icbpixwrite,'(a25)') strfree(n)
    enddo
  endif

  lstpnt = 1
  do  nz=1,nzones
    write (icbpixwrite,*) (w(1,np),np=lstpnt,npnts(nz)+lstpnt-1)
    if(nz.gt.1) lstpnt = sum(npnts(1:nz-1)) + 1
  enddo

endselect


END SUBROUTINE WriteNextAVUSPixVar



!*****************************************************************************
SUBROUTINE WriteAVUSGrid(gridfilename,grdformat,griddouble,ndm,nzones,npatch,x,npnts,nfaces,ncells,mxppf,mxfpc,numppf,ipnt,icell)
!*****************************************************************************

!USE MOD_AVUSGrid
IMPLICIT NONE

real (kind=FP),dimension(:,:),pointer :: x
real (kind=SP),dimension(:,:),allocatable :: sx
real (kind=DP),dimension(:,:),allocatable :: dx

character(len=chrlen),intent(in) :: gridfilename

integer,intent(in) :: ndm,nzones,npatch
integer, dimension(:),pointer :: npnts, nfaces, ncells
integer, dimension(:),pointer :: numppf
integer, dimension(:,:),pointer :: ipnt,icell
integer, dimension(:), pointer :: mxppf,  mxfpc


integer,intent(in) :: grdformat
logical,intent(in) :: griddouble

integer :: nz,np,nf,i,j,ierr,lstpnt,lstface
character(len=chrlen) :: fileformat


lstpnt = 1
lstface = 1

Call DispReadWrite(dsp_write,dsp_grid,dsp_avusfile,gridfilename,grdformat,griddouble)


select case(grdformat)
case(0,2)

   if(griddouble) then

     select case(FP)
     case(SP)
       allocate( dx(ndm,sum(npnts)),stat = ierr)
       CALL CheckAllocationError(ierr)
       do np=1,sum(npnts)
         dx(:,np) = x(:,np)
       enddo
     case(DP)
     end select
   else
     select case(FP)
     case(SP)
     case(DP)
       allocate( sx(ndm,sum(npnts)),stat = ierr)
       CALL CheckAllocationError(ierr)
       do np=1,sum(npnts)
         sx(:,np) = x(:,np)
       enddo
     end select
   endif

   select case(grdformat)
   case(0)
     fileformat = 'unformatted'
   case(2)
     fileformat = 'binary'
   end select
   
   open(unit=4,file=gridfilename,form=fileformat,status='unknown')

   write(4)ndm,nzones,npatch
   write(4)(npnts(nz),nfaces(nz),ncells(nz), &
&             maxval(numppf),mxfpc(nz),nz=1,nzones)

   do nz=1,nzones
     write(*,*) 'Writing Grid Points in Zone ',nz,'...'
     
     if(griddouble) then
       select case(FP)
       case(SP)
         do np=lstpnt,npnts(nz)+lstpnt-1
           write(4) (dx(i,np),i=1,ndm)
         enddo
       case(DP)
         do np=lstpnt,npnts(nz)+lstpnt-1
           write(4) (x(i,np),i=1,ndm)
         enddo
       end select
     else
       select case(FP)
       case(SP)
         do np=lstpnt,npnts(nz)+lstpnt-1
           write(4) (x(i,np),i=1,ndm)
         enddo
       case(DP)
         do np=lstpnt,npnts(nz)+lstpnt-1
           write(4) (sx(i,np),i=1,ndm)
         enddo
       end select
     endif

     write(*,*) 'Writing Connectivity in Zone ',nz,'...'
     do nf=lstface,nfaces(nz)+lstface-1
       write(4) numppf(nf),(ipnt(i,nf),i=1,numppf(nf)),(icell(j,nf),j=1,2)
     enddo
     lstpnt = lstpnt + npnts(nz)
     lstface = lstface + nfaces(nz)
   enddo
   

case(1)
   open(unit=4,file=gridfilename,form='formatted',status='unknown')

   write(4,*)ndm,nzones,npatch
   write(4,*)(npnts(nz),nfaces(nz),ncells(nz), &
&             mxppf(nz),mxfpc(nz),nz=1,nzones)
   do nz=1,nzones
     write(*,*) 'Writing Grid Points in Zone ',nz,'...'
     do np=lstpnt,npnts(nz)+lstpnt-1
       write(4,*) (x(i,np),i=1,ndm)
     enddo
     write(*,*) 'Writing Connectivity in Zone ',nz,'...'
     do nf=lstface,nfaces(nz)+lstface-1
       write(4,*) numppf(nf),(ipnt(i,nf),i=1,numppf(nf)),(icell(j,nf),j=1,2)
     enddo
     lstpnt = lstpnt + npnts(nz)
     lstface = lstface + nfaces(nz)
   enddo
end select

END SUBROUTINE WriteAVUSGrid





!********************************************************************************
! Tecplot file writing subroutines
!********************************************************************************

!********************************************************************************
SUBROUTINE WriteTecplotFile(pixfilename,ndm,nzones,dx,npnts,nfaces,ncells,mxppf,mxfpc,numppf,ipnt,icell,dw)
!********************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts,
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!     &              ndots
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

real (kind=FP),dimension(:,:),pointer :: dx,dw

character(len=chrlen),intent(in) :: pixfilename

integer,intent(in) :: ndm,nzones
integer, dimension(:),pointer :: npnts, nfaces, ncells,mxppf,mxfpc
integer, dimension(:),pointer :: numppf
integer, dimension(:,:),pointer :: ipnt,icell

integer, dimension(:)  , pointer :: ibf,ipf,mfpc
integer, dimension(:,:), pointer :: iface

INTEGER :: k,iz,nf,nc,ierr, lu_solfile

integer :: maxfbd
integer :: jbcp,i1,i2,i3,jf,jp,if1,ip1

iopt = 2
nout = 1
nbc = 3  !Write all boundaries
jbcp = 0
ipfmax = maxval(mxppf)

lu_solfile = 10

write(*,*) 'Writing to Tecplot file...'

OPEN (UNIT=lu_solfile, FILE = pixfilename)

!WRITE(lu_solfile,*) 'VARIABLES = "X", "Y", "Z", "U", "V", "W", "P", "RHO"' 

maxfbd = count(icell(1,:).lt.0)
maxfbd = count(icell(2,:).lt.0) + maxfbd

allocate(mfpc(sum(ncells)),ibf(maxfbd),ipf(sum(npnts)),iface(sum(ncells),maxval(mxfpc)),mfpc(sum(ncells)),stat=ierr)
Call CheckAllocationError(ierr)

mfpc(:) = 0
ipf(:) = 0

do iz = 1,nzones

      write(6,*) ' '
      write(6,910) iz
910   format(1x,'generating face/cell relationships for zone: ',i3)
      ifbnd = 0
      ifmax = 0
      do k=1,2
        do nf=1,nfaces(iz)
          nc = icell(k,nf)
          if (nc.gt.0) then
            mfpc(nc) = mfpc(nc) + 1
            iface(nc,mfpc(nc)) = nf
            ifmax = max(ifmax,mfpc(nc))
!...flag boundary faces (where b.c. less than 0)
!...note that boundary faces are triangles (only) in 3-D
      else if (nc.lt.0) then
!          else if (nbc.gt.1.and.nc.lt.0) then
!            if (.not.(nbc.eq.2.and.abs(nc).ne.jbcp)) then
            ifbnd = ifbnd + 1
!             if (ifbnd.gt.maxfbd) then
!                 write(6,*) '******* ERROR: no. bndry faces greater than maxfbd'
!                 write(6,*) '        no. bndry faces = ',ifbnd
!                 write(6,*) '        max allowed     = ',maxfbd
!                 write(6,*) 'Adjust parameter(s), recompile and try again'
!                 stop
!              end if
              ibf(ifbnd) = nf
              i1 = ipnt(1,nf) 
              i2 = ipnt(2,nf) 
              if (ndm.eq.3) i3 = ipnt(3,nf) 
              ipf(i1) = -1
              ipf(i2) = -1
              if (ndm.eq.3) ipf(i3) = -1
!            endif
          endif
!...zero out the icell array to use as a working array later...
          icell(k,nf) = 0
        enddo
      enddo
      write(6,*) 'done with face/cell relationships'

      if (ndm.eq.3) then
        nttot = 0
        ntpts  = 0
        nptot = 0
        nppts  = 0
        do nc = 1,ncells(iz)
          select case (mfpc(nc))
          case(4)
            nttot = nttot + 1
!...set working array (icell) for tet's...
!...icell(1) sets a flag - 1 - for the absolute cell number...
!...icell(2) sets a flag - 1 - for the absolute node number...
            icell(1,nc) = 1
            do jf = 1,4
              if1 = iface(nc,jf)
              do jp = 1,numppf(if1)
                ip1 = ipnt(jp,if1)
                if (icell(2,ip1).eq.0) then
                  icell(2,ip1) = 1
                  ntpts = ntpts + 1
                end if
              enddo
            enddo
          case(5)
            nptot = nptot + 1
!...set working array (icell) triangular prisms...
!...icell(1) sets a flag - 2 - for the absolute cell number...
!...icell(2) sets a flag - 2 or 3 - for the absolute node number...
!...note that icell(2) is 3 for nodes on shared (tet-prism) tri faces
            icell(1,nc) = 2
            do jf = 1,5
              if1 = iface(nc,jf)
              do jp = 1,numppf(if1)
                ip1 = ipnt(jp,if1)
                if (icell(2,ip1).eq.0) then
                  icell(2,ip1) = 2
                  nppts = nppts + 1
                end if
                if (icell(2,ip1).eq.1) then
                  icell(2,ip1) = 3
                  nppts = nppts + 1
                end if
              enddo
            enddo
          case(6) 
            nptot = nptot + 1
!...set working array (icell) rectangular prisms...
!...icell(1) sets a flag - 2 - for the absolute cell number...
!...icell(2) sets a flag - 2 - for the absolute node number...
            icell(1,nc) = 2
            do jf = 1,5
              if1 = iface(nc,jf)
              do jp = 1,numppf(if1)
                ip1 = ipnt(jp,if1)
                if (icell(2,ip1).eq.0) then
                  icell(2,ip1) = 2
                  nppts = nppts + 1
                end if
!                if (icell(2,ip1).eq.1) then
!                  icell(2,ip1) = 3
!                  nppts = nppts + 1
!                end if
              enddo
            enddo
          end select
        enddo
 
        write(6,*) ' '
        write(6,*) 'total number of tetrahedra,pts: ',nttot,ntpts
        if (nptot.gt.0) write(6,*) 'total number of prisms,pts: ',nptot,nppts
      end if

    call writer(iz,ndm,mxfpc,npnts,ncells,ipf,dx,dw,ipnt,iface,numppf,icell,ibf)

enddo









close(lu_solfile)


deallocate(mfpc)

    
 
END SUBROUTINE WriteTecplotFile



!***********************************************************************
subroutine writer(izone,ndm,mxfpc,npnts,ncells,ipf,dx,dw,ipnt,iface,numppf,icell,ibf)
!***********************************************************************
implicit none


!      parameter ( maxpt = 400000, maxcell = 2500000,
!     &            maxface = 5000000, mxppf = 4, mxfpc = 5,
!     &            maxfbd =  500000, maxpbd = 100000, maxzon = 64 )
!
!...writes out a mesh and variable file in Tecplot finite-
!...element form 
!
!...comment out/or uncomment the next line depending on precision...
!      implicit double precision (a-h,o-z)
!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts,
!     &              ndots
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

integer, intent(in) :: izone,ndm

integer, dimension(:),pointer :: npnts, ncells,mxfpc,numppf
integer, dimension(:),pointer :: ibf,ipf
integer, dimension(:,:),pointer :: ipnt,icell
integer, dimension(:,:), pointer :: iface

integer, dimension(:),pointer :: ipbf

INTEGER :: i,j,ierr

integer :: ipt,ibfc
integer :: i1,i2,i3,i1w,i2w,i3w

real (kind=FP),dimension(:,:),pointer :: dx,dw

allocate(ipbf(sum(npnts)),stat=ierr)
Call CheckAllocationError(ierr)

      if (izone.eq.1) then
        write(10,*) 'TITLE = "COBALT Unstructured Grid Result" '
        if (ndm.eq.3) then
!         if (iopt.eq.2) then
!            write(10,*) 'VARIABLES = "X","Y","Z","RHO","U","V","W","P","FREE","TT","LL","KK","JJ","HH","GG","YY","II","OO","PP"'
            write(10,*) 'VARIABLES = "X","Y","Z","RHO","U","V","W","P","FREE"'
!         else
!           write(10,*) 'VARIABLES = "X","Y","Z"'
!          end if
        else
!         if (iopt.eq.2) then
!            write(10,*) 'VARIABLES = "X","Y","RHO","U","V","W","P","FREE","TT","LL","KK","JJ","HH","GG","YY","II","OO","PP"'
            write(10,*) 'VARIABLES = "X","Y","RHO","U","V","W","P","FREE"'
!         else
!           write(10,*) 'VARIABLES = "X","Y"'
!          end if
        end if
      end if

      write(6,*) ' '
      write(6,900) izone
900  format(1x,'writing info for zone: ',i3)
!
!...node info for 2-D grids...
      if (ndm.eq.2) then
        if ((nout.eq.1).and.(mxfpc(izone).eq.3)) write (10,903) izone,npnts(izone),ncells(izone)
        if ((nout.eq.1).and.(mxfpc(izone).eq.4)) write (10,902) izone,npnts(izone),ncells(izone)

!...field data for tri's only...
      ipbnd = 0
      if (iopt.eq.2) then
        do i=1,npnts(izone)
          if (ipf(i).lt.0) then
            ipbnd = ipbnd + 1
!      if (ipbnd.gt.maxpbd) then
!      write(6,*) '******* ERROR: no. bndry nodes greater than maxpbd'
!      write(6,*) '        no. bndry nodes = ',ipbnd
!      write(6,*) '        max allowed     = ',maxpbd
!      write(6,*) 'Adjust parameter(s), recompile and try again'
!      stop
!      end if
            ipbf(ipbnd) = i
          end if
          if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm),(dw(j,i),j=1,npixvar)
        enddo
      else
        do i=1,npnts(izone)
          if (ipf(i).lt.0) then
           ipbnd = ipbnd + 1
!      if (ipbnd.gt.maxpbd) then
!      write(6,*) '******* ERROR: no. bndry nodes greater than maxpbd'
!      write(6,*) '        no. bndry nodes = ',ipbnd
!      write(6,*) '        max allowed     = ',maxpbd
!      write(6,*) 'Adjust parameter(s), recompile and try again'
!      stop
!      end if
           ipbf(ipbnd) = i
         end if
         if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm)
        enddo
      end if
      
      if (nout.eq.1) write(6,*) 'done writing title and node info.......'
!...node/cell connectivity for 2-D grids...
        write(*,*) 'NDOTS',mxfpc(izone)
        if(mxfpc(izone).eq.4) call quapts(izone,ncells,iface,ipnt)
        if(mxfpc(izone).eq.3) call tripts(izone,ncells,iface,ipnt)
      end if

!...field data and node/cell connectivity for 3-D grids...
      if (ndm.eq.3) then
        call tetpts(izone,ndm,ncells,iface,ipnt,ipf,ipbf,dx,dw,npnts,icell)
        if (ipfmax.eq.4) call prsmpts(izone,ndm,ncells,iface,ipnt,ipf,ipbf,dx,dw,numppf,npnts,icell)
      end if

      if (nbc.gt.1) then
        write(6,*) ' '
        write(6,*) 'writing info for body surfaces.......'
        write(6,*) ' # of boundary faces: ',ifbnd
        write(6,*) ' # of boundary nodes for those faces: ',ipbnd

!...boundary faces only go into "zone 2"...
        if (ndm.eq.3) then
          if(mxfpc(izone).eq.4) then
            write (10,*) 'ZONE  T= "Body ',izone,'"  N= ',ipbnd,' , E= ',ifbnd,' ,',' F=FEPOINT, ET=QUADRILATERAL'
          else if(mxfpc(izone).eq.3) then
            write (10,*) 'ZONE  T= "Body ',izone,'"  N= ',ipbnd,' , E= ',ifbnd,' ,',' F=FEPOINT, ET=TRIANGLE'
          endif
        else
          write (10,*) 'ZONE  T= "Body ',izone,'" F=POINT, I= ',ipbnd
        end if

!...flow variables on boundary surfaces...
        if (iopt.eq.2) then
          do i=1,ipbnd
            ipt = ipbf(i)
            ipf(ipt) = i
            write (10,*) (dx(j,ipt),j=1,ndm),(dw(j,ipt),j=1,npixvar)
          enddo
        else
          do i=1,ipbnd
            ipt = ipbf(i)
            ipf(ipt) = i
            write (10,*) (dx(j,ipt),j=1,ndm)
          enddo
        end if

        if (ndm.eq.3) then
!...node/surface triangle connectivity...
          do i=1,ifbnd
            ibfc = ibf(i)
            i1 = ipnt(1,ibfc)
            i2 = ipnt(2,ibfc)
!      i1w = i1
!      i2w = i2
            i1w = ipf(i1)
            i2w = ipf(i2)
            if (ndm.eq.3) then
              i3 = ipnt(3,ibfc)
!      i3w = i3
              i3w = ipf(i3)
            end if
!...determine correct node pointers (reordered on boundaries)...
!      do 103 j=1,ipbnd
!      if (i1.eq.ipbf(j)) i1w = j
!      if (i2.eq.ipbf(j)) i2w = j
!      if (ndm.eq.3) then
!      if (i3.eq.ipbf(j)) i3w = j
!      end if
! 103  continue
            if (ndm.eq.2) write (10,*) i1w,i2w
            if (ndm.eq.3) write (10,*) i1w,i2w,i3w
          enddo
        end if

      write(6,*) 'done writing info for body surfaces......'
      end if


903  format(1x,'ZONE  T= "Zone ',i3,'"  N= ',i9,' , E= ',i9,' ,',' F=FEPOINT, ET=TRIANGLE')
902  format(1x,'ZONE  T= "Zone ',i3,'"  N= ',i9,' , E= ',i9,' ,',' F=FEPOINT, ET=QUADRILATERAL')

      end subroutine writer
!***********************************************************************
subroutine tripts(izone,ncells,iface,ipnt)
!***********************************************************************
implicit none


!
!      parameter ( maxpt = 400000, maxcell = 2500000,
!     &            maxface = 5000000, mxppf = 4, mxfpc = 5,
!     &            maxfbd =  500000, maxpbd = 100000, maxzon = 64 )

!...comment out/or uncomment the next line depending on precision...
!      implicit double precision (a-h,o-z)
!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

integer, intent(in) :: izone
integer, dimension(:),pointer :: ncells
integer, dimension(:,:),pointer :: ipnt

integer :: ic,if1,ifc1,ifc2,ifc3,i,j,if2
integer, dimension(:,:), pointer :: iface

      if (nout.eq.1) write(6,*) 'writing cell/node connectivity info for triangles'

      do ic = 1,ncells(izone)

        if1 = iface(ic,1)
        ifc1 = ipnt(1,if1)
        ifc2 = ipnt(2,if1)
        ifc3 = 0

!...have two points, need third point...
        do j = 2,3
          if2 = iface(ic,j)
          do i = 1,2
            if (ipnt(i,if2).eq.ifc1.or. &
                ipnt(i,if2).eq.ifc2.or. &
                ipnt(i,if2).eq.ifc3) then
            else 
              if (ifc3.eq.0) ifc3 = ipnt(i,if2)
            endif
          enddo
        enddo
        if (ifc3.eq.0) then
          write(6,*) ' error: in subroutine tripts: ifc3.eq.0 '
          stop
        else
          if (nout.eq.1) write (10,*) ifc1,ifc2,ifc3
        end if

      enddo
      if (nout.eq.1) write(6,*) 'done writing tri cell/node info.......'

      end subroutine tripts
!***********************************************************************
subroutine quapts(izone,ncells,iface,ipnt)
!***********************************************************************
implicit none

!      parameter ( maxpt = 400000, maxcell = 2500000,
!     &            maxface = 5000000, mxppf = 4, mxfpc = 5,
!     &            maxfbd =  500000, maxpbd = 100000, maxzon = 64 )

!...comment out/or uncomment the next line depending on precision...
!      implicit double precision (a-h,o-z)
!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

integer, intent(in) :: izone
integer, dimension(:),pointer :: ncells
integer, dimension(:,:),pointer :: ipnt

integer :: ic,if1,ifc1,ifc2,ifc3,ifc4,j,if2
integer, dimension(:,:), pointer :: iface

      if (nout.eq.1) write(6,*) 'writing cell/node connectivity info for quads'

      do ic = 1,ncells(izone)
 
        if1 = iface(ic,1)
        ifc1 = ipnt(1,if1)
        ifc2 = ipnt(2,if1)
        ifc3 = 0
        ifc4 = 0

!...have two points, need third point
        do j = 2,4
          if2 = iface(ic,j)
!     do 110 i = 1,2
!        if (ipnt(if2,i).eq.ifc1.or.
!    &       ipnt(if2,i).eq.ifc2.or.
!    &       ipnt(if2,i).eq.ifc3.or.
!    &       ipnt(if2,i).eq.ifc4) goto 110 
!        if ((ifc3.ne.0).and.(ifc4.eq.0)) ifc4 = ipnt(if2,i)
!        if ((ifc3.eq.0).and.(ifc4.eq.0)) ifc3 = ipnt(if2,i)
!110  continue
          if (ipnt(1,if2).eq.ifc1) ifc4=ipnt(2,if2)
          if (ipnt(2,if2).eq.ifc1) ifc4=ipnt(1,if2)
          if (ipnt(1,if2).eq.ifc2) ifc3=ipnt(2,if2)
          if (ipnt(2,if2).eq.ifc2) ifc3=ipnt(1,if2)
        enddo
        if (ifc3.eq.0) then
          write(6,*) ' error: in subroutine tripts: ifc3.eq.0 '
          stop
        else
          if (nout.eq.1) write (10,'(4(1x,i6))') ifc1,ifc2,ifc3,ifc4
        end if

      enddo
      if (nout.eq.1) write(6,*) 'done writing quad cell/node info.......'

      end subroutine quapts
!***********************************************************************
subroutine tetpts(izone,ndm,ncells,iface,ipnt,ipf,ipbf,dx,dw,npnts,icell)
!***********************************************************************
implicit none

!      parameter ( maxpt = 400000, maxcell = 2500000,
!     &            maxface = 5000000, mxppf = 4, mxfpc = 5,
!     &            maxfbd =  500000, maxpbd = 100000, maxzon = 64 )

!...comment out/or uncomment the next line depending on precision...
!      implicit double precision (a-h,o-z)
!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

integer, intent(in) :: izone,ndm
integer, dimension(:),pointer :: npnts, ncells
integer, dimension(:,:),pointer :: ipnt,icell
integer, dimension(:),pointer :: ipf
integer, dimension(:),pointer :: ipbf
real (kind=FP),dimension(:,:),pointer :: dx,dw

integer :: ic,if1,ifc1,ifc2,ifc3,ifc4,i,j,if2,iknt,ifc1a,ifc2a,ifc3a,ifc4a
integer, dimension(:,:), pointer :: iface

      if (nout.eq.1) write (10,900) izone,ntpts,nttot
900  format(1x,'ZONE  T= "Tet-Zone ',i3,'"  N= ',i9,' , E= ',i9,' ,',' F=FEPOINT, ET=TETRAHEDRON')

      if (nout.eq.1) write(6,*) 'writing node & connectivity info for tetrahedra'

      ipbnd = 0
      iknt = 0

      do i = 1,npnts(izone)

        if (icell(2,i).eq.1.or.icell(2,i).eq.3) then

          if (ipf(i).lt.0) then
            ipbnd = ipbnd + 1
!      if (ipbnd.gt.maxpbd) then
!      write(6,*) '******* ERROR: no. bndry nodes greater than maxpbd'
!      write(6,*) '        no. bndry nodes = ',ipbnd
!      write(6,*) '        max allowed     = ',maxpbd
!      write(6,*) 'Adjust parameter(s), recompile and try again'
!      stop
!      end if
            ipbf(ipbnd) = i
          end if

          iknt = iknt + 1
          ipf(i) = iknt

          if (iopt.eq.2) then
            if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm),(dw(j,i),j=1,numrstrtvar(1))
          else
            if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm)
          end if
        end if

      enddo

      if (nout.eq.1) write(6,*) 'done writing tet node info.......'
!...debug
      write(6,*) ' number of nodes (known,counted): ',ntpts,iknt

      iknt = 0

      do ic = 1,ncells(izone)

        if (icell(1,ic).eq.1) then

          iknt = iknt + 1
          if1 = iface(ic,1)
          ifc1 = ipnt(1,if1)
          ifc2 = ipnt(2,if1)
          ifc3 = ipnt(3,if1)
          ifc4 = 0

!...have three points, need fourth point...
          do j = 2,4
            if2 = iface(ic,j)
            do i = 1,3
              if (ipnt(i,if2).eq.ifc1.or. &
                  ipnt(i,if2).eq.ifc2.or. &
                  ipnt(i,if2).eq.ifc3) then
              else 
                ifc4 = ipnt(i,if2)
              endif
            enddo
          enddo


          if (ifc4.eq.0) then
            write(6,*) ' error: in subroutine tetpts: ifc4.eq.0 '
            stop
          else
            ifc1a = ipf(ifc1)
            ifc2a = ipf(ifc2)
            ifc3a = ipf(ifc3)
            ifc4a = ipf(ifc4)
            if (nout.eq.1) write (10,*) ifc1a,ifc2a,ifc3a,ifc4a
          end if
        end if
      enddo

      if (nout.eq.1) write(6,*) 'done writing tet connectivity info......'
!...debug
      write(6,*) ' number of tets (known,counted):  ',nttot,iknt

      end subroutine tetpts
!***********************************************************************
subroutine prsmpts(izone,ndm,ncells,iface,ipnt,ipf,ipbf,dx,dw,numppf,npnts,icell)
!***********************************************************************
implicit none


!      parameter ( maxpt = 400000, maxcell = 2500000,
!     &            maxface = 5000000, mxppf = 4, mxfpc = 5,
!     &            maxfbd =  500000, maxpbd = 100000, maxzon = 64 )

!...comment out/or uncomment the next line depending on precision...
!      implicit double precision (a-h,o-z)
!      common/adres/ ipnt(maxface,mxppf),icell(2,maxface),mppf(maxface),
!     &              iface(maxcell,mxfpc),mfpc(maxcell),ibf(maxfbd),
!     &              ipf(maxpt),ipbf(maxpbd)
!      common/const/ ndm,npts(maxzon),nfaces(maxzon),ncells(maxzon),
!     &              ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      common/const/ ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
      integer :: ipbnd,ifbnd,ipfmax,ifmax,nttot,nptot,ntpts,nppts
!      common/geom/  x(3,maxpt)
!      common/flow/  q(maxpt,6)
      common/formt/ igform,ipform,iopt,nout,nbc
      integer :: igform,ipform,iopt,nout,nbc

integer, intent(in) :: izone,ndm
integer, dimension(:),pointer :: npnts, ncells,numppf
integer, dimension(:,:),pointer :: ipnt,icell
integer, dimension(:),pointer :: ipf
integer, dimension(:),pointer :: ipbf
real (kind=FP),dimension(:,:),pointer :: dx,dw

integer :: ic,if1,i,j,iknt,inum,jf,ipf1
integer :: itmp1,itmp2,itmp3,itmp4,itmp5,itmp6,itmp7,itmp8,ip,inxt,ilast,k
integer, dimension(:,:), pointer :: iface


      integer :: itmp(8),nfscr(5),npscr(6)

      if (nout.eq.1) write (10,900) izone,nppts,nptot
900  format(1x,'ZONE  T= "Prism-Zone ',i3,'"  N= ',i9,' , E= ',i9,' ,',' F=FEPOINT, ET=BRICK')

      if (nout.eq.1) write(6,*) 'writing node & connectivity info for prisms'

      ipbnd = 0
      iknt = 0

      do i = 1,npnts(izone)

        if (icell(2,i).eq.2.or.icell(2,i).eq.3) then

          if (ipf(i).lt.0) then
            ipbnd = ipbnd + 1
!      if (ipbnd.gt.maxpbd) then
!      write(6,*) '******* ERROR: no. bndry nodes greater than maxpbd'
!      write(6,*) '        no. bndry nodes = ',ipbnd
!      write(6,*) '        max allowed     = ',maxpbd
!      write(6,*) 'Adjust parameter(s), recompile and try again'
!      stop
!      end if
            ipbf(ipbnd) = i
          end if

          iknt = iknt + 1
          ipf(i) = iknt

          if (iopt.eq.2) then
            if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm),(dw(j,i),j=1,numrstrtvar(1))
          else
            if (nout.eq.1) write (10,*) (dx(j,i),j=1,ndm)
          end if

        end if

      enddo
      if (nout.eq.1) write(6,*) 'done writing prism node info.......'
!...debug
      write(6,*) ' number of nodes (known,counted): ',nppts,iknt

      inum = 0

      do ic = 1,ncells(izone)

!...use flagged prisms here...
        if (icell(1,ic).eq.2) then
          inum = inum + 1

!...flag two triangular faces and identify their nodes...
          iknt = 0
          do jf = 1,5
            if1 = iface(ic,jf)
            ipf1 = numppf(if1)
            nfscr(jf) = -if1
            if (ipf1.eq.3) then
              npscr(iknt+1) = ipnt(if1,1)
              npscr(iknt+2) = ipnt(if1,2)
              npscr(iknt+3) = ipnt(if1,3)
              iknt = iknt + 3
              nfscr(jf) = if1
            end if
          enddo

!...pick a quad face and order the nodes...
          jf = 0
110       jf = jf + 1
          if (nfscr(jf).lt.0) then
            if1 = -nfscr(jf)
            itmp1 = 0
            itmp2 = 0
!...pick first and second nodes on the face...
            do ip = 1,6
              if (ipnt(1,if1).eq.npscr(ip)) itmp1 = ip
              if (ipnt(2,if1).eq.npscr(ip)) itmp2 = ip
            enddo
!      if (itmp1.eq.0) write(6,*) ' ERROR: something wrong with prism'
!      if (itmp2.eq.0) write(6,*) ' ERROR: something wrong with prism'
!...see how things match up...
            if (itmp1.lt.4.and.itmp2.lt.4) then
!...both nodes on tri face #1...
              itmp5 = ipnt(1,if1)
              itmp6 = ipnt(2,if1)
              itmp7 = ipnt(3,if1)
              itmp8 = ipnt(4,if1)
            else if (itmp1.gt.3.and.itmp2.gt.3) then
!...both nodes on tri face #2...
              itmp5 = ipnt(1,if1)
              itmp6 = ipnt(2,if1)
              itmp7 = ipnt(3,if1)
              itmp8 = ipnt(4,if1)
            else
!...one node on tri face #1, other node on tri face #2...
!...so, reorder...
              itmp5 = ipnt(1,if1)
              itmp6 = ipnt(4,if1)
              itmp7 = ipnt(3,if1)
              itmp8 = ipnt(2,if1)
            end if
!...now find 1,2 and 3,4 nodes to assign...
            itmp1 = 0
            itmp2 = 0
            itmp3 = 0
            itmp4 = 0
            do ip = 1,3
              inxt = mod(ip,3) + 1
              ilast = mod(inxt,3) + 1
              if ((itmp5.eq.npscr(ip).and.itmp6.eq.npscr(inxt)).or. &
                  (itmp5.eq.npscr(inxt).and.itmp6.eq.npscr(ip))) then
                itmp1 = npscr(ilast)
                itmp2 = npscr(ilast)
              end if
              if ((itmp5.eq.npscr(ip).and.itmp6.eq.npscr(ilast)).or. &
                  (itmp5.eq.npscr(ilast).and.itmp6.eq.npscr(ip))) then
                itmp1 = npscr(inxt)
                itmp2 = npscr(inxt)
              end if
              if ((itmp7.eq.npscr(ip).and.itmp8.eq.npscr(inxt)).or. &
                  (itmp7.eq.npscr(inxt).and.itmp8.eq.npscr(ip))) then
                itmp3 = npscr(ilast)
                itmp4 = npscr(ilast)
              end if
              if ((itmp7.eq.npscr(ip).and.itmp8.eq.npscr(ilast)).or. &
                  (itmp7.eq.npscr(ilast).and.itmp8.eq.npscr(ip))) then
                itmp3 = npscr(inxt)
                itmp4 = npscr(inxt)
              end if
            enddo
            do ip = 4,6
              inxt = mod(ip,6) + 1
              ilast = mod(inxt,6) + 1
              if ((itmp5.eq.npscr(ip).and.itmp6.eq.npscr(inxt)).or. &
                  (itmp5.eq.npscr(inxt).and.itmp6.eq.npscr(ip))) then
                itmp1 = npscr(ilast)
                itmp2 = npscr(ilast)
              end if
              if ((itmp5.eq.npscr(ip).and.itmp6.eq.npscr(ilast)).or. &
                  (itmp5.eq.npscr(ilast).and.itmp6.eq.npscr(ip))) then
                itmp1 = npscr(inxt)
                itmp2 = npscr(inxt)
              end if
              if ((itmp7.eq.npscr(ip).and.itmp8.eq.npscr(inxt)).or. &
                  (itmp7.eq.npscr(inxt).and.itmp8.eq.npscr(ip))) then
                itmp3 = npscr(ilast)
                itmp4 = npscr(ilast)
              end if
              if ((itmp7.eq.npscr(ip).and.itmp8.eq.npscr(ilast)).or. &
                  (itmp7.eq.npscr(ilast).and.itmp8.eq.npscr(ip))) then
                itmp3 = npscr(inxt)
                itmp4 = npscr(inxt)
              end if
            enddo
!      if (itmp1.eq.0) write(6,*) ' ERROR: something wrong with prism'
!      if (itmp2.eq.0) write(6,*) ' ERROR: something wrong with prism'
!      if (itmp3.eq.0) write(6,*) ' ERROR: something wrong with prism'
!      if (itmp4.eq.0) write(6,*) ' ERROR: something wrong with prism'
            go to 120
          else
            go to 110
      end if
120     continue
!
!...write out the prism connectivity data...
        itmp(1) = ipf(itmp1)
        itmp(2) = ipf(itmp2)
        itmp(3) = ipf(itmp3)
        itmp(4) = ipf(itmp4)
        itmp(5) = ipf(itmp5)
        itmp(6) = ipf(itmp6)
        itmp(7) = ipf(itmp7)
        itmp(8) = ipf(itmp8)
        if (nout.eq.1)  write(10,*) (itmp(k),k=1,8)

      end if

      enddo
      if (nout.eq.1) write(6,*) 'done writing prism connectivity info......'
!...debug
      write(6,*) ' number of prisms (known,counted):',nptot,inum

      end subroutine prsmpts

!********************************************************************************
! END Tecplot file writing subroutines
!********************************************************************************



END MODULE MOD_AVUSGrid
