MODULE MOD_FDL3DI_ParallelIO

USE MOD_GRIDVARS, ONLY : chrlen, FP, DP, SP


IMPLICIT NONE

INTEGER, PARAMETER :: CFDID = 1
INTEGER, PARAMETER :: CEMID = 2
INTEGER, PARAMETER :: CEMPPID = 3


INTEGER :: ngrd, is, ie, js, je, ks, ke

INTEGER :: nvar

INTEGER :: ngorig
INTEGER, DIMENSION(:), ALLOCATABLE :: iengorig, jengorig, kengorig
INTEGER, DIMENSION(:,:), ALLOCATABLE :: isa, iea, jsa, jea, ksa, kea
INTEGER, DIMENSION(:), ALLOCATABLE :: nbpg, nbsptr, nbeptr

INTEGER :: fVersion, fEqID, fitimeint
INTEGER, PARAMETER :: fWriteVersion = 1

INTEGER, DIMENSION(:), ALLOCATABLE :: ieng, jeng, keng

INTEGER :: isgsmodel
REAL(KIND=DP) :: re, xm1, alfa, beta, gam, s1, pr, prt


!Make these input parameters
CHARACTER(LEN=30) :: prefix= 'fdl3di.out.',suffix= ''
INTEGER :: ilen = 6
CHARACTER(LEN=30) :: fdl3difilefmt !Variable to store the fdl3di filename format

CONTAINS
!
!==================================================================================================
SUBROUTINE ReadParallelP3DFile ( ng, x, y, z, iblank, idm, jdm, kdm, gridformat, griddouble, gridfile )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
REAL(KIND=FP), DIMENSION(:), POINTER :: x,y,z
INTEGER, DIMENSION(:), POINTER :: iblank
INTEGER, INTENT(OUT) :: idm, jdm, kdm
CHARACTER(LEN=chrlen), INTENT(IN) :: gridfile
INTEGER, INTENT(IN) :: gridformat
LOGICAL, INTENT(IN) :: griddouble
!
! LOCAL VARIABLES
!
INTEGER :: nogr, nogg, fnvar, ierror, fibeta
CHARACTER(LEN=chrlen) :: filename, chreqset
LOGICAL :: tof

INQUIRE(file=gridfile, EXIST=tof)
IF (.not.tof) THEN
   WRITE(*,*) 'COULD NOT FIND GRID FILE "',TRIM(gridfile),'"'
   WRITE(*,*)
   STOP
END IF

OPEN(UNIT=2,FILE=gridfile,FORM='UNFORMATTED',STATUS='OLD')
!
! Read number of grids in grid file
!
READ(2) nogg
ngorig = nogg
!
! Read general solver header information
!

IF ( nogg .GT. ngrd ) THEN
   WRITE(*,*) 'NUMBER IF GRIDS IN GRID FILE GREATER THAN NUMBER OF BLOCKS IN GRIDINDEXFILE'
   WRITE(*,*)
   WRITE(*,*) 'STOPPING'
   STOP
ELSE IF ( nogg .LT. ngrd ) THEN
  CALL ReadCombinedGrid ( ng, x, y, z, iblank, gridformat, griddouble )
END IF

CLOSE(2)

idm = ieng(ng)
jdm = jeng(ng)
kdm = keng(ng)

END SUBROUTINE ReadParallelP3DFile
!
!==================================================================================================
SUBROUTINE ReadParallelFDLRestartFile ( nstepi, tau, ng, var, rstrtformat, rstrtdouble, restartinfile )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
INTEGER, INTENT(OUT) :: nstepi
REAL(KIND=DP), INTENT(OUT) :: tau
REAL(KIND=FP), DIMENSION(:,:), POINTER :: var
INTEGER, INTENT(IN) :: rstrtformat
LOGICAL, INTENT(IN) :: rstrtdouble
CHARACTER(LEN=chrlen), INTENT(IN) :: restartinfile
!
! LOCAL VARIABLES
!
INTEGER :: nogr, fibeta


WRITE(*,*) ''
WRITE(*,*) 'Reading restart file: ', TRIM(restartinfile)
!
! Read general solver header information
!
OPEN(UNIT=1,FILE=restartinfile,FORM='UNFORMATTED',STATUS='OLD')
READ(1) fVersion, fEqID
READ(1) nogr, nvar, fitimeint, fibeta, nstepi, tau
   
IF ( nogr .NE. ngorig ) THEN
   IF ( ng .EQ. 1 ) THEN
      WRITE(*,*) 'GRID COUNT MISMATCH'
      WRITE(*,*) 'NUMBER OF GRIDS IN GRID FILE   = ', ngorig
      WRITE(*,*) 'NUMBER OF GRID IN RESTART FILE = ', nogr
      WRITE(*,*) 'STOPPING'
   ENDIF
   STOP
END IF

!
! Read information pertinent to the equation set
!
SELECT CASE(fEqID)
CASE(CFDID)
   READ(1) isgsmodel, re, xm1, alfa, beta, gam, s1, pr, prt
CASE(CEMID)

CASE(CEMPPID)

END SELECT


IF ( nogr .GT. ngrd ) THEN
   IF ( ng .eq. 1 ) THEN
      WRITE(*,*) 'NUMBER OF GRIDS IN RESTART FILE EXCEEDS NGRD PARAMETER'
      WRITE(*,*) 'FROM RESTART FILE = ',nogr
      WRITE(*,*) 'NGRD PARAMETER = ',ngrd
      WRITE(*,*) 'GRID CONGLOMERATION CAPABILITY NOT CURRENTLY IMPLEMENTED'
      WRITE(*,*) 'STOPPING'
   END IF
   STOP
ELSE IF ( nogr .LT. ngrd ) THEN
   CALL ReadCombinedRestart ( ng, var, rstrtformat, rstrtdouble, restartinfile)
END IF

CLOSE(1)

END SUBROUTINE ReadParallelFDLRestartFile
!
!==================================================================================================
SUBROUTINE ReadGridIndexFile ( nclient, gridindexfile )
!==================================================================================================
!

INTEGER, INTENT(IN) :: nclient
CHARACTER(LEN=chrlen) :: gridindexfile
!
! LOCAL VARIABLES
!
INTEGER :: nb, ngx, ngo, junki, nbglobal, nblk, nbl, nblocal, ngpmax

OPEN(UNIT=9,FILE=gridindexfile,FORM='FORMATTED',STATUS='OLD',ACTION='READ',ERR=101)
ALLOCATE(nbpg(nclient))
nbpg(:) = 0
nblk = 0
ngo = 0
GO TO 201

101 WRITE(*,*) 'ERROR OPENING FILE ',TRIM(gridindexfile),' IN SUBROUTINE ReadRestartFile...STOPPING.'
STOP

201 READ(9,*,END=301) nbglobal, ngx, nblocal, junki, junki, junki, junki, junki, junki
IF ( nbglobal .GT. nblk ) nblk = nbglobal
IF ( ngx .GT. ngo ) ngo = ngx
IF ( nblocal .GT. nbpg(ngx) ) nbpg(ngx) = nblocal
GO TO 201


301 IF (nblk .NE. nclient) THEN
   WRITE(*,*) 'Mismatch grid no. between ',trim(gridindexfile),' and NGRD parameter'
   WRITE(*,*) nblk,' grids in ',trim(gridindexfile),'...stopping'
   STOP
END IF

REWIND (9)
ALLOCATE(nbsptr(ngo))
ALLOCATE(nbeptr(ngo))
DO ngx = 1, ngo
   SELECT CASE (ngx)
   CASE (1)
      nbsptr(ngx) = 1
   CASE DEFAULT
      nbsptr(ngx) = nbeptr(ngx-1) + 1
   END SELECT
   nbeptr(ngx) = nbsptr(ngx) + nbpg(ngx) - 1
END DO

ngpmax = MAXVAL(nbpg)
ALLOCATE(isa(ngo,ngpmax))
ALLOCATE(iea(ngo,ngpmax))
ALLOCATE(jsa(ngo,ngpmax))
ALLOCATE(jea(ngo,ngpmax))
ALLOCATE(ksa(ngo,ngpmax))
ALLOCATE(kea(ngo,ngpmax))
ALLOCATE(iengorig(ngo))
ALLOCATE(jengorig(ngo))
ALLOCATE(kengorig(ngo))

ALLOCATE(ieng(nblk),jeng(nblk),keng(nblk))

iea(:,:) = 0
jea(:,:) = 0
kea(:,:) = 0

WRITE(*,*) 'Partitioned grid dimensions:'
DO nb = 1, nclient
   READ(9,*) junki, ngx, nbl, isa(ngx,nbl), iea(ngx,nbl), jsa(ngx,nbl), jea(ngx,nbl), ksa(ngx,nbl), kea(ngx,nbl)
   ieng(nb) = iea(ngx,nbl) - isa(ngx,nbl) + 1
   jeng(nb) = jea(ngx,nbl) - jsa(ngx,nbl) + 1
   keng(nb) = kea(ngx,nbl) - ksa(ngx,nbl) + 1
   WRITE(*,*) nb, ieng(nb),jeng(nb),keng(nb)
END DO
CLOSE(9)

ngrd = nclient


END SUBROUTINE ReadGridIndexFile
!
!==================================================================================================
SUBROUTINE ReadPartitionedGrid ( ng, x, y, z, iblank, gridfile )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
REAL(KIND=FP), DIMENSION(:), POINTER :: x,y,z
INTEGER, DIMENSION(:), POINTER :: iblank
!
! LOCAL VARIABLES
!
INTEGER :: i, j, k, iv, n, ngx
INTEGER :: junki
REAL :: junkr
CHARACTER(LEN=chrlen) :: gridfile

READ(2) (ieng(ngx), jeng(ngx), keng(ngx), ngx = 1, ngrd)


IF ( ng .EQ. 1 .AND. .NOT. ALLOCATED(isa) ) THEN
   WRITE(*,*) 'NUMBER OF GRIDS IN RESTART FILE EQUAL TO NGRD PARAMETER'
   WRITE(*,*) 'FROM RESTART FILE = NGRD PARAMETER = ',ngrd
   WRITE(*,*) 'RESTART FILE ALREADY DECOMPOSED...WILL READ AS IS'
      
   ngorig = ngrd
   ALLOCATE(isa(ngrd,1))
   ALLOCATE(iea(ngrd,1))
   ALLOCATE(jsa(ngrd,1))
   ALLOCATE(jea(ngrd,1))
   ALLOCATE(ksa(ngrd,1))
   ALLOCATE(kea(ngrd,1))
   ALLOCATE(nbsptr(ngrd))
   ALLOCATE(nbeptr(ngrd))
   ALLOCATE(iengorig(ngrd))
   ALLOCATE(jengorig(ngrd))
   ALLOCATE(kengorig(ngrd))
   isa(:,1) = 1
   iea(:,1) = ieng(:)
   jsa(:,1) = 1
   jea(:,1) = jeng(:)
   ksa(:,1) = 1
   kea(:,1) = keng(:)
   iengorig = ieng
   jengorig = jeng
   kengorig = keng
   DO n = 1,ngrd
      nbsptr(n) = n
      nbeptr(n) = n
   END DO
END IF

DO ngx = 1, ng-1
   READ(2)
END DO

CALL GetGridIndices (ng)
ALLOCATE(x(ieng(ng)*jeng(ng)*keng(ng)), &
         y(ieng(ng)*jeng(ng)*keng(ng)), &
         z(ieng(ng)*jeng(ng)*keng(ng)), &
         iblank(ieng(ng)*jeng(ng)*keng(ng)))

READ(2) (((x(i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke), &
&       (((y(i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke), &
&       (((z(i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke), &
        (((iblank(i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke)


END SUBROUTINE ReadPartitionedGrid
!
!==================================================================================================
SUBROUTINE ReadPartitionedRestart ( ng, var, rstrtformat, rstrtdouble )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
INTEGER, INTENT(IN) :: rstrtformat
LOGICAL, INTENT(IN) :: rstrtdouble
REAL(KIND=FP), DIMENSION(:,:), POINTER :: var
!
! LOCAL VARIABLES
!
INTEGER :: i, j, k, iv, n, ngx
INTEGER :: junki
REAL(KIND=DP), DIMENSION(:,:), POINTER :: dvar
REAL(KIND=SP), DIMENSION(:,:), POINTER :: svar
REAL(KIND=DP) :: djunkr
REAL(KIND=SP) :: sjunkr

READ(2) (ieng(ngx), jeng(ngx), keng(ngx), ngx = 1, ngrd)


IF ( ng .EQ. 1 .AND. .NOT. ALLOCATED(isa) ) THEN
   WRITE(*,*) 'NUMBER OF GRIDS IN RESTART FILE EQUAL TO NGRD PARAMETER'
   WRITE(*,*) 'FROM RESTART FILE = NGRD PARAMETER = ',ngrd
   WRITE(*,*) 'RESTART FILE ALREADY DECOMPOSED...WILL READ AS IS'
      
   ngorig = ngrd
   ALLOCATE(isa(ngrd,1))
   ALLOCATE(iea(ngrd,1))
   ALLOCATE(jsa(ngrd,1))
   ALLOCATE(jea(ngrd,1))
   ALLOCATE(ksa(ngrd,1))
   ALLOCATE(kea(ngrd,1))
   ALLOCATE(nbsptr(ngrd))
   ALLOCATE(nbeptr(ngrd))
   ALLOCATE(iengorig(ngrd))
   ALLOCATE(jengorig(ngrd))
   ALLOCATE(kengorig(ngrd))
   isa(:,1) = 1
   iea(:,1) = ieng(:)
   jsa(:,1) = 1
   jea(:,1) = jeng(:)
   ksa(:,1) = 1
   kea(:,1) = keng(:)
   iengorig = ieng
   jengorig = jeng
   kengorig = keng
   DO n = 1,ngrd
      nbsptr(n) = n
      nbeptr(n) = n
   END DO
END IF

DO ngx = 1, ng-1
   READ(1)
   IF ( fitimeint .GE. 2 ) THEN
      READ(1) 
   END IF
   IF ( fitimeint .EQ. 3 ) THEN
      READ(1) 
   END IF  
END DO

CALL GetGridIndices (ng)
ALLOCATE(var(nvar,ieng(ng)*jeng(ng)*keng(ng)))


SELECT CASE(rstrtformat)
CASE(0,2)
   SELECT CASE(FP)
   CASE(SP)
      SELECT CASE(rstrtdouble)
      CASE(.true.)
         ALLOCATE(dvar(nvar,ieng(ng)*jeng(ng)*keng(ng)))
         READ(1) ((((dvar(iv,i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke),iv=1,NVAR)
         var = dvar
         DEALLOCATE(dvar)
      CASE(.false.)
         READ(1) ((((var(iv,i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke),iv=1,NVAR)
      END SELECT
   CASE(DP)
      SELECT CASE(rstrtdouble)
      CASE(.true.)
         READ(1) ((((var(iv,i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke),iv=1,NVAR)
      CASE(.false.)
         ALLOCATE(svar(nvar,ieng(ng)*jeng(ng)*keng(ng)))
         READ(1) ((((svar(iv,i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke),iv=1,NVAR)
         var = svar
         DEALLOCATE(svar)
      END SELECT
   END SELECT
CASE(1)

   READ(1,*) ((((var(iv,i+(j-1)*ieng(ng)+(k-1)*ieng(ng)*jeng(ng)),i=is,ie),j=js,je),k=ks,ke),iv=1,NVAR)

END SELECT

IF ( fitimeint .GE. 2 ) THEN
   READ(1) 
END IF
IF ( fitimeint .EQ. 3 ) THEN
   READ(1) 
END IF  

END SUBROUTINE ReadPartitionedRestart
!
!==================================================================================================
SUBROUTINE ReadCombinedGrid ( ng, x, y, z, iblank, gridformat, griddouble )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
REAL(KIND=FP), DIMENSION(:), POINTER :: x,y,z
INTEGER, DIMENSION(:), POINTER :: iblank
INTEGER, INTENT(IN) :: gridformat
LOGICAL, INTENT(IN) :: griddouble
!
! LOCAL VARIABLES
!
INTEGER, DIMENSION(ngorig) :: idim, jdim, kdim
INTEGER :: i, j, k, iv, n, nbl, iengo, jengo, kengo, ngox, igs, ige, ngx, istop, ierror
INTEGER :: isi, iei, jsi, jei, ksi, kei
INTEGER :: isks, iske, jsks, jske, ksks, kske
CHARACTER(LEN=4) :: chardir
INTEGER :: junki
REAL(KIND=DP), DIMENSION(:), POINTER :: dx,dy,dz
REAL(KIND=SP), DIMENSION(:), POINTER :: sx,sy,sz
REAL(KIND=DP) :: djunkr
REAL(KIND=SP) :: sjunkr


SELECT CASE(gridformat)
CASE(0,2)
   READ(2) (iengorig(ngx), jengorig(ngx), kengorig(ngx), ngx = 1, ngorig)
CASE(1)
   READ(2,*) (iengorig(ngx), jengorig(ngx), kengorig(ngx), ngx = 1, ngorig)
END SELECT

DO n = 1, ngorig
   IF ( ng .GE. nbsptr(n) .AND. ng .LE. nbeptr(n) ) THEN
      nbl = ng - nbsptr(n) + 1
      iengo = iengorig(n)
      jengo = jengorig(n)
      kengo = kengorig(n)
      ngox = n
   END IF
END DO

isi = isa(ngox,nbl)
iei = iea(ngox,nbl)
jsi = jsa(ngox,nbl)
jei = jea(ngox,nbl)
ksi = ksa(ngox,nbl)
kei = kea(ngox,nbl)

isks = isi - 1
iske = iengo - iei
jsks = iengo * (jsi - 1)
jske = iengo * (jengo - jei)
ksks = iengo * jengo * (ksi - 1)
kske = iengo * jengo * (kengo - kei)

WRITE(*,*)'Skipping over ', ngox-1,' Grids...'
DO n = 1, ngox-1
   READ(2) 
END DO

WRITE(*,*)'Reading grid size ',ieng(ng),jeng(ng),keng(ng)
CALL GetGridIndices (ng)
ALLOCATE(x(ieng(ng)*jeng(ng)*keng(ng)), &
         y(ieng(ng)*jeng(ng)*keng(ng)), &
         z(ieng(ng)*jeng(ng)*keng(ng)), &
         iblank(ieng(ng)*jeng(ng)*keng(ng)))


SELECT CASE(gridformat)
CASE(0,2)

   SELECT CASE(FP)
   CASE(SP)
      SELECT CASE(griddouble)
      CASE(.true.)
         WRITE(*,*)'Reading Double Precission Grid'
         ALLOCATE(dx(ieng(ng)*jeng(ng)*keng(ng)), &
                  dy(ieng(ng)*jeng(ng)*keng(ng)), &
                  dz(ieng(ng)*jeng(ng)*keng(ng)))
         READ(2) (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(dx((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(dy((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(dz((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (junki,k=1,ksks),((junki,j=1,jsks),((junki,i=1,isks),(iblank((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (junki,i=1,iske),j=jsi,jei),(junki,j=1,jske),k=ksi,kei),(junki,k=1,kske)
         x = dx
         y = dy
         z = dz

         DEALLOCATE(dx,dy,dz)
      CASE(.false.)
         READ(2) (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(x((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(y((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(z((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (junki,k=1,ksks),((junki,j=1,jsks),((junki,i=1,isks),(iblank((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (junki,i=1,iske),j=jsi,jei),(junki,j=1,jske),k=ksi,kei),(junki,k=1,kske)
      END SELECT
   CASE(DP)

      SELECT CASE(griddouble)
      CASE(.true.)
         READ(2) (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(x((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(y((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(z((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske), &
&                (junki,k=1,ksks),((junki,j=1,jsks),((junki,i=1,isks),(iblank((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (junki,i=1,iske),j=jsi,jei),(junki,j=1,jske),k=ksi,kei),(junki,k=1,kske)
      CASE(.false.)
         ALLOCATE(sx(ieng(ng)*jeng(ng)*keng(ng)), &
                  sy(ieng(ng)*jeng(ng)*keng(ng)), &
                  sz(ieng(ng)*jeng(ng)*keng(ng)))

         READ(2) (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(sx((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(sy((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(sz((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&                (junki,k=1,ksks),((junki,j=1,jsks),((junki,i=1,isks),(iblank((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                (junki,i=1,iske),j=jsi,jei),(junki,j=1,jske),k=ksi,kei),(junki,k=1,kske)
         x = sx
         y = sy
         z = sz

         DEALLOCATE(sx,sy,sz)

      END SELECT


   END SELECT
CASE(1)

   READ(2,*) (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(x((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&            (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&            (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(y((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&            (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&            (sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(z((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&            (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske), &
&            (junki,k=1,ksks),((junki,j=1,jsks),((junki,i=1,isks),(iblank((i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&            (junki,i=1,iske),j=jsi,jei),(junki,j=1,jske),k=ksi,kei),(junki,k=1,kske)

END SELECT


END SUBROUTINE ReadCombinedGrid
!
!==================================================================================================
SUBROUTINE ReadCombinedRestart ( ng, var, rstrtformat, rstrtdouble, restartinfile )
!==================================================================================================
!

INTEGER, INTENT(IN) :: ng
REAL(KIND=FP), DIMENSION(:,:), POINTER :: var
INTEGER, INTENT(IN) :: rstrtformat
LOGICAL, INTENT(IN) :: rstrtdouble
CHARACTER(LEN=chrlen) :: restartinfile
!
! LOCAL VARIABLES
!
INTEGER, DIMENSION(ngorig) :: idim, jdim, kdim
INTEGER :: i, j, k, iv, n, nbl, iengo, jengo, kengo, ngox, igs, ige, ngx, istop, ierror
INTEGER :: isi, iei, jsi, jei, ksi, kei
INTEGER :: isks, iske, jsks, jske, ksks, kske
CHARACTER(LEN=4) :: chardir
INTEGER :: junki
REAL(KIND=DP), DIMENSION(:,:), POINTER :: dvar
REAL(KIND=SP), DIMENSION(:,:), POINTER :: svar
REAL(KIND=DP) :: djunkr
REAL(KIND=SP) :: sjunkr

DO n = 1, ngorig
   READ(1) idim(n), jdim(n), kdim(n)
   IF ( ng .GE. nbsptr(n) .AND. ng .LE. nbeptr(n) ) THEN
      nbl = ng - nbsptr(n) + 1
      iengo = iengorig(n)
      jengo = jengorig(n)
      kengo = kengorig(n)
      ngox = n
   END IF
END DO

DO ngx = 1, ngorig
   istop = 0
   IF ( idim(ngx) .NE. iengorig(ngx) ) THEN
      chardir = 'ieng'
      WRITE(*,876) ngx, chardir
      WRITE(*,878) iengorig(ngx)
      WRITE(*,877) TRIM(restartinfile), idim(ngx)
      istop = 1
   END IF
   IF ( jdim(ngx) .NE. jengorig(ngx) ) THEN
      chardir = 'jeng'
      WRITE(*,876) ngx, chardir
      WRITE(*,878) jengorig(ngx)
      WRITE(*,877) TRIM(restartinfile), jdim(ngx)
      istop = 1
   END IF
   IF (kdim(ngx) .NE. kengorig(ngx)) THEN
      chardir = 'keng'
      WRITE(*,876) ngx, chardir
      WRITE(*,878) kengorig(ngx)
      WRITE(*,877) TRIM(restartinfile), kdim(ngx)
      istop = 1
   END IF
   IF (istop .EQ. 1) THEN
      WRITE(*,*) 'Stopping.'
      STOP
   END IF
END DO

isi = isa(ngox,nbl)
iei = iea(ngox,nbl)
jsi = jsa(ngox,nbl)
jei = jea(ngox,nbl)
ksi = ksa(ngox,nbl)
kei = kea(ngox,nbl)

isks = isi - 1
iske = iengo - iei
jsks = iengo * (jsi - 1)
jske = iengo * (jengo - jei)
ksks = iengo * jengo * (ksi - 1)
kske = iengo * jengo * (kengo - kei)
 
DO n = 1, ngox-1
   READ(1) 
   IF ( fitimeint .GE. 2 ) THEN
      READ(1) 
   END IF
   IF ( fitimeint .EQ. 3 ) THEN
      READ(1) 
   END IF
END DO

ALLOCATE(var(nvar,ieng(ng)*jeng(ng)*keng(ng)))

CALL GetGridIndices (ng)

SELECT CASE(rstrtformat)
CASE(0,2)
   SELECT CASE(FP)
   CASE(SP)
      SELECT CASE(rstrtdouble)
      CASE(.true.)
         ALLOCATE(dvar(nvar,ieng(ng)*jeng(ng)*keng(ng)))
         READ(1) ((djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(dvar(iv,(i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                 (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske),iv=1,NVAR)
         var = dvar
         DEALLOCATE(dvar)
      CASE(.false.)
         READ(1) ((sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(var(iv,(i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                 (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske),iv=1,NVAR)
      END SELECT
   CASE(DP)
      SELECT CASE(rstrtdouble)
      CASE(.true.)
         READ(1) ((djunkr,k=1,ksks),((djunkr,j=1,jsks),((djunkr,i=1,isks),(var(iv,(i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                 (djunkr,i=1,iske),j=jsi,jei),(djunkr,j=1,jske),k=ksi,kei),(djunkr,k=1,kske),iv=1,NVAR)
      CASE(.false.)
         ALLOCATE(svar(nvar,ieng(ng)*jeng(ng)*keng(ng)))
         READ(1) ((sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(svar(iv,(i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&                 (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske),iv=1,NVAR)
         var = svar
         DEALLOCATE(svar)
      END SELECT
   END SELECT
CASE(1)

   READ(1,*) ((sjunkr,k=1,ksks),((sjunkr,j=1,jsks),((sjunkr,i=1,isks),(var(iv,(i-isi+1) + (j-jsi)*ieng(ng) + (k-ksi)*ieng(ng)*jeng(ng)),i=isi,iei), &
&             (sjunkr,i=1,iske),j=jsi,jei),(sjunkr,j=1,jske),k=ksi,kei),(sjunkr,k=1,kske),iv=1,NVAR)

END SELECT


876 FORMAT('Grid ',I3,1X,A4,' values do not match between restart and grid files...')
877 FORMAT('   From restart file ',A,', value is ',I4)
878 FORMAT('   From grid file value is ',I4)

END SUBROUTINE ReadCombinedRestart

!
!==================================================================================================
SUBROUTINE ClanParallelP3DRead ()
!==================================================================================================
!

IMPLICIT NONE

IF (ALLOCATED(isa)) DEALLOCATE(isa)
IF (ALLOCATED(iea)) DEALLOCATE(iea)
IF (ALLOCATED(jsa)) DEALLOCATE(jsa)
IF (ALLOCATED(jea)) DEALLOCATE(jea)
IF (ALLOCATED(ksa)) DEALLOCATE(ksa)
IF (ALLOCATED(kea)) DEALLOCATE(kea)
IF (ALLOCATED(iengorig)) DEALLOCATE(iengorig)
IF (ALLOCATED(jengorig)) DEALLOCATE(jengorig)
IF (ALLOCATED(kengorig)) DEALLOCATE(kengorig)
IF (ALLOCATED(nbsptr)) DEALLOCATE(nbsptr)
IF (ALLOCATED(nbeptr)) DEALLOCATE(nbeptr)
   
IF (ALLOCATED(nbpg)) DEALLOCATE(nbpg)

END SUBROUTINE ClanParallelP3DRead
!
!==================================================================================================
SUBROUTINE GetGridIndices (ngl)
!==================================================================================================
!

IMPLICIT NONE

INTEGER, INTENT(IN) :: ngl

is = 1
ie = ieng(ngl)
js = 1
je = jeng(ngl)
ks = 1
ke = keng(ngl)

END SUBROUTINE GetGridIndices

!
!==================================================================================================
SUBROUTINE ReadDecompBCFile(ng, bcdecompfile)
!==================================================================================================
!
USE MOD_ReadPegFile, ONLY : nbc, ibtyp, ibdir, ibcs, ibce, jbcs, jbce, kbcs, kbce

INTEGER, INTENT(IN) :: ng
CHARACTER(LEN=chrlen), INTENT(IN) :: bcdecompfile
!
! LOCAL VARIABLES
!
INTEGER :: junki, nbcall, ngcur, nglast, ngr, nbcs, bcng, nbcmax
CHARACTER(len=60) :: filename

OPEN(UNIT=55,FILE=bcdecompfile,FORM='FORMATTED',STATUS='OLD',ACTION='READ')

nbcall = 0
150 READ(55,*,END=200) junki, junki, junki, junki, junki, junki, junki, junki, junki
nbcall = nbcall + 1
GO TO 150
200 CONTINUE

WRITE(*,*) 'Total Number of Boundary Conditions = ',nbcall

REWIND(55)
IF (ALLOCATED(nbc)) DEALLOCATE(nbc)
ALLOCATE(nbc(ngrd))
nbc = 0
!
! Set BC pointer values between blocks read in
!
ngcur = 1
DO nbcs = 1, nbcall
   READ(55,*) bcng, junki, junki, junki, junki, junki, junki, junki, junki
   IF (bcng .EQ. ngcur) THEN
      nbc(ngcur) = nbc(ngcur) + 1
   ELSE
      ngcur = ngcur+1 
      nbc(ngcur) = 1
   END IF
END DO

nbcmax = nbc(ng)
IF (ALLOCATED(ibtyp)) DEALLOCATE(ibtyp)
IF (ALLOCATED(ibdir)) DEALLOCATE(ibdir)
IF (ALLOCATED(ibcs)) DEALLOCATE(ibcs)
IF (ALLOCATED(ibce)) DEALLOCATE(ibce)
IF (ALLOCATED(jbcs)) DEALLOCATE(jbcs)
IF (ALLOCATED(jbce)) DEALLOCATE(jbce)
IF (ALLOCATED(kbcs)) DEALLOCATE(kbcs)
IF (ALLOCATED(kbce)) DEALLOCATE(kbce)
ALLOCATE(ibtyp(nbcmax,1))
ALLOCATE(ibdir(nbcmax,1))
ALLOCATE(ibcs(nbcmax,1))
ALLOCATE(ibce(nbcmax,1))
ALLOCATE(jbcs(nbcmax,1))
ALLOCATE(jbce(nbcmax,1))
ALLOCATE(kbcs(nbcmax,1))
ALLOCATE(kbce(nbcmax,1))

!
! Set BC pointer values between blocks read in
!
REWIND(55)
bcng = 0
DO ngr = 1, ngrd
   bcng = bcng + 1
   IF(bcng.eq.ng) EXIT
   
   DO nbcs = 1, nbc(ngr)
      READ(55,*) junki, junki, junki, junki, junki, junki, junki, junki, junki
   END DO
END DO

DEALLOCATE(nbc)
ALLOCATE(nbc(1))
nbc = nbcmax

DO nbcs = 1, nbc(1)
   READ(55,*) bcng, &
&             ibtyp(nbcs,1), &
&             ibdir(nbcs,1), &
&             ibcs(nbcs,1), &
&             ibce(nbcs,1), &
&             jbcs(nbcs,1), &
&             jbce(nbcs,1), &
&             kbcs(nbcs,1), &
&             kbce(nbcs,1)
END DO


END SUBROUTINE ReadDecompBCFile

END MODULE MOD_FDL3DI_ParallelIO
