MODULE MOD_GRIDVARS

INTEGER, PARAMETER :: SP=selected_real_kind(6)
INTEGER, PARAMETER :: DP=selected_real_kind(12)
INTEGER, PARAMETER :: FP=SP

!True if all variables should be read, false to only read in one at a time
logical :: readallvars = .true.

!The number of variables in a pix file and restart file
integer :: npixvar
integer :: nfilesect = 1

integer, dimension(30) :: numrstrtvar = 0
integer, dimension(30,2) :: rstrtvarsct = 0

!FDL3DI File formats
INTEGER, PARAMETER :: fdlmin = 100
INTEGER, PARAMETER :: fdl_1 = 100
INTEGER, PARAMETER :: fdl_2 = 101
INTEGER, PARAMETER :: fdl_3 = 102
INTEGER, PARAMETER :: fdlmax = 102

!Length of file name strings
integer, parameter :: chrlen = 40

integer, parameter :: dsp_read = 1
integer, parameter :: dsp_write = 2
integer, parameter :: dsp_parent = 3
integer, parameter :: dsp_child = 4

integer, parameter :: dsp_grid = 10
integer, parameter :: dsp_rstrt = 20
integer, parameter :: dsp_pix = 30
integer, parameter :: dsp_p3dQfile = 40
integer, parameter :: dsp_p3dFunfile = 50
integer, parameter :: dsp_flow = 60

integer, parameter :: dsp_avusfile = 100
integer, parameter :: dsp_p3dfile = 200
integer, parameter :: dsp_fdl3difile = 300
integer, parameter :: dsp_vstagefile = 400

CONTAINS

!*****************************************************************************
SUBROUTINE DispReadWrite(ReadWrite,GrdOrRst,filetype,filename,fileformat,double,nvar,iblanked,multib,totpnts,totcells)
!*****************************************************************************

IMPLICIT NONE

integer, intent(in) :: ReadWrite,GrdOrRst,filetype,fileformat
logical, intent(in) :: double
integer, optional, intent(in) :: nvar,totpnts,totcells
logical, optional, intent(in) :: iblanked,multib
character(len=chrlen), intent(in) :: filename

character(len=80) :: sentance

write(*,*)

select case(ReadWrite)
case(dsp_read)  
  sentance = 'Reading '
case(dsp_write)  
  sentance = 'Writing '
case(dsp_parent)
  sentance = 'Donor '
case(dsp_child)
  sentance = 'Receiver '
end select

select case(filetype)
case(dsp_avusfile)
  sentance = trim(sentance) // ' AVUS'
case(dsp_p3dfile)
  sentance = trim(sentance) // ' Plot3D'
case(dsp_fdl3difile)
  sentance = trim(sentance) // ' FDL3DI'
case(dsp_vstagefile)
  sentance = trim(sentance) // ' VSTAGE'
end select

select case(GrdOrRst)
case(dsp_grid)
  sentance = trim(sentance) // ' Grid'
case(dsp_rstrt)
  sentance = trim(sentance) // ' Restart'
case(dsp_pix)
  sentance = trim(sentance) // ' Pix'
case(dsp_p3dQfile)
  sentance = trim(sentance) // ' Q-File'
case(dsp_p3dfunfile)
  sentance = trim(sentance) // ' Fuction'
case(dsp_flow)
  sentance = trim(sentance) // ' Flow'
end select

select case(present(nvar))
case(.false.)
  sentance = trim(sentance) // ' Data'
case(.true.)
  select case(ReadWrite)
  case(dsp_read)
    write(sentance,"(A,A,4I,A)") trim(sentance), ' Variable', nvar
  case(dsp_write)
    write(sentance,"(A,A,4I,A)") trim(sentance), ' Interpolated Variable', nvar
  end select
end select

select case(ReadWrite)
case(dsp_read)
  sentance = trim(sentance) // ' from File ' // trim(filename) // ' ...'
case(dsp_write)
  sentance = trim(sentance) // ' to File ' // trim(filename) // ' ...'
case(dsp_parent,dsp_child)
  sentance = trim(sentance) // ' File ' // trim(filename)
end select

write(*,*) trim(sentance)



select case(present(totpnts))
case(.true.)
  write(*,*) 'Total Number of Points : ',totpnts
end select

select case(present(totcells))
case(.true.)
  write(*,*) 'Total Number of Cellc  : ',totcells
end select



write(*,*) 'File Properties:'

select case(present(multib))
case(.true.)
  select case(multib)
  case(.true.)
    write(*,200) 'Multi Block'
  case(.false.)
    write(*,200) 'Single Block'
  end select
end select

select case(present(iblanked))
case(.true.)
  select case(iblanked)
  case(.true.)
    write(*,200) 'IBlanked'
  case(.false.)
    write(*,200) 'No IBlanking'
  end select
end select

select case(fileformat)
case(0)
  write(*,200) 'Unformatted'
case(1)
  write(*,200) 'Formatted'
case(2)
  write(*,200) 'Binary'
end select


select case(fileformat)
case(0,2)
  select case(double)
  case(.true.)
    write(*,200) 'Double Precision'
  case(.false.) 
    write(*,200) 'Single Precision'
  end select 
end select

write(*,*)

200 format('                   ',A)


END SUBROUTINE DispReadWrite



!*****************************************************************************
SUBROUTINE DispFileReadError()
!*****************************************************************************

IMPLICIT NONE

write(*,*) 
write(*,*) 'An Error Occured While Attempting to Read this File'
write(*,*) 
write(*,*) 'Please Check the File Properties in the Input File.'
write(*,*)
write(*,*) '  --- CODE HALTED ---'
stop

END SUBROUTINE DispFileReadError

!*****************************************************************************
SUBROUTINE CheckAllocationError(ierr)
!*****************************************************************************

IMPLICIT NONE

INTEGER, INTENT(IN) :: ierr

if(ierr /= 0) then 
  write(*,*) 'Error allocating Memory'
  write(*,*)
  write(*,*) '  --- CODE HALTED ---'
  stop
end if

END SUBROUTINE CheckAllocationError



END MODULE MOD_GRIDVARS
