
MODULE MOD_P3DGrid
USE MOD_GRIDVARS
implicit none

LOGICAL :: DEBUGPRINT=.TRUE.
!Number to check if two double are equal
!real(kind=DP),parameter :: SMALL = 1e-20


!The Q-Variables
real(kind=DP),dimension(:,:), allocatable :: DQVar
real(kind=SP),dimension(:,:), allocatable :: SQVar

!Plot3D function files
integer,dimension(:), allocatable :: NFVAR

!FDL3DI restart flag
integer,dimension(3) :: iflag

!--------------------------------------------------------------------
! Solver Related Quanteties
INTEGER :: fileversion, eqID
INTEGER :: fdlnvar, itimeint, ibeta
REAL(kind=DP) :: DTAU
REAL(KIND=DP) :: p3dre, p3dxm1, p3dalfa, p3dtau
INTEGER :: NSTEP

INTEGER, PARAMETER :: CFDID = 1
INTEGER, PARAMETER :: CEMID = 2
INTEGER, PARAMETER :: CEMPPID = 3
!--------------------------------------------------------------------

!--------------------------------------------------------------------
! FDL3DI CFD Related Quanteties
INTEGER :: isgsmodel
REAL(KIND=DP) :: fdlre, fdlxm1, fdlalfa, fdlbeta, fdlgam, fdls1, fdlpr, fdlprt
!--------------------------------------------------------------------

integer, parameter :: icbfdlread  = 201  !201 through 201+numrstrtvar are used when interpolating one variable at a time
integer, parameter :: icbfdlwrite = 301  !301 through 301+numrstrtvar are used when interpolating one variable at a time
character(len=chrlen) :: tmpfile(15)     !This array size must greater or equal to the number of variables in a FDL3DI restart file

type dbpntr
  real(kind=DP), dimension(:,:), pointer :: dw
end type

type sgpntr
  real(kind=SP), dimension(:,:), pointer :: w
end type

type intpntr
  integer, pointer :: i
end type

CONTAINS

!**************************************************************************************************
SUBROUTINE P3DSetDebug(lgcl)
!**************************************************************************************************
  LOGICAL,INTENT(IN) :: lgcl
  DEBUGPRINT=lgcl
END SUBROUTINE P3DSetDebug

!**************************************************************************************************
SUBROUTINE ReadP3DGrid(gridfilename,grdformat,griddouble,nzones,ndm,x,ieng,jeng,keng,iblank,iblanked,multib)
!**************************************************************************************************
!Reads in a Plot3D grid file and converts it to an AVUS grid format
!
!Input:
!	gridfilename - Plot3d file read in
!
!Outputs and Allocates:
!
!	dx(3,:) - Grid coodrinates in double precision
!
!	igsptr - Grid starting indexes in xt,yt,zt,iblankt
!	igeptr - Grid ending indexes in xt,yt,zt,iblankt
!	igpnts - Grid sizes

IMPLICIT NONE


real (kind=FP),dimension(:,:),pointer :: x

character(len=chrlen),intent(in) :: gridfilename

integer,intent(out) :: ndm,nzones
integer, dimension(:),pointer :: ieng,jeng,keng,iblank
integer, dimension(:), allocatable :: igpnts, igsptr, igeptr

integer,intent(in) :: grdformat
logical,intent(in) :: griddouble,iblanked,multib

real (kind=SP),dimension(:,:),allocatable :: sx
real (kind=DP),dimension(:,:),allocatable :: dx
character(len=chrlen) :: fileformat


INTEGER :: ngrd, ng, i1, i2, igall, iengx, jengx, kengx, icbgrd
INTEGER :: i,n,np
INTEGER :: ierr
!INTEGER :: iprec,ib


Call DispReadWrite(dsp_read,dsp_grid,dsp_p3dfile,gridfilename,grdformat,griddouble,iblanked=iblanked,multib=multib)


ngrd = 1 !This is so that a single block grid can be read
ndm = 3  !All Plot3D grids are all 3-D to the extent of my knoledge

icbgrd = 13
!call p3dfor(gridfilename,iprec,ib)

!if(griddouble.and.iprec.eq.1) then
!  write(*,*) 

!elseif(.not.griddouble.and.iprec.eq.2) then

!endif


select case (grdformat)
case(0)
   open(icbgrd,file=gridfilename,form='unformatted',status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
case(1)
   open(icbgrd,file=gridfilename,form='formatted',status='old')
   if(multib) READ(icbgrd,*,ERR=101,END=101) ngrd
case(2)
   fileformat = 'binary'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
end select

WRITE(*,*) 'Number of grids is ',ngrd
nzones = ngrd

ALLOCATE(ieng(ngrd),jeng(ngrd),keng(ngrd),igpnts(ngrd),igsptr(ngrd),igeptr(ngrd),stat=ierr)
CALL CheckAllocationError(ierr)


select case (grdformat)
case(0,2)
  READ(icbgrd,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
case(1)
  READ(icbgrd,*,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
end select

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   IF(DEBUGPRINT)WRITE(*,200) ' Dimensions of grid ',ng,' is ',iengx, jengx, kengx
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)

   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO

200 format(A,i5.1,A,3i5.1)

if(iblanked) then
  allocate(iblank(igall),stat=ierr)
  CALL CheckAllocationError(ierr)
  iblank(:) = 1
endif

allocate( x(ndm,igall),stat = ierr)
CALL CheckAllocationError(ierr)

select case (grdformat)
case(0,2)
  if(griddouble)then
    select case(FP)
    case(SP)
      allocate( dx(ndm,igall),stat = ierr)
      CALL CheckAllocationError(ierr)


      DO ng = 1, ngrd
        IF(DEBUGPRINT)WRITE(*,*) 'Reading grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          READ(icbgrd,ERR=101,END=101) ((dx(n,i),i=i1,i2),n=1,ndm), &
                                        (iblank(i),i=i1,i2)
        case(.false.)
          READ(icbgrd,ERR=101,END=101) ((dx(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO

      do np =1,igall
        x(:,np) = dx(:,np)
      enddo

      deallocate(dx)
    case(DP)
      DO ng = 1, ngrd
        IF(DEBUGPRINT)WRITE(*,*) 'Reading grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          READ(icbgrd,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                        (iblank(i),i=i1,i2)
        case(.false.)
          READ(icbgrd,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO
    end select
  else

    select case(FP)
    case(SP)
      DO ng = 1, ngrd
        IF(DEBUGPRINT)WRITE(*,*) 'Reading grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          READ(icbgrd,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                        (iblank(i),i=i1,i2)
        case(.false.)
          READ(icbgrd,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO
    case(DP)
      allocate( sx(ndm,igall),stat = ierr)
      CALL CheckAllocationError(ierr)


      DO ng = 1, ngrd
        IF(DEBUGPRINT)WRITE(*,*) 'Reading grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          READ(icbgrd,ERR=101,END=101) ((sx(n,i),i=i1,i2),n=1,ndm), &
                                        (iblank(i),i=i1,i2)
        case(.false.)
          READ(icbgrd,ERR=101,END=101) ((sx(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO

      do np =1,igall
        x(:,np) = sx(:,np)
      enddo

      deallocate(sx)
    end select
  endif
case(1)

    DO ng = 1, ngrd
      IF(DEBUGPRINT)WRITE(*,*) 'Reading grid point data for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      select case(iblanked)
      case(.true.)
        READ(icbgrd,*,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                        (iblank(i),i=i1,i2)
      case(.false.)
        READ(icbgrd,*,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm)
      end select
    END DO
end select

CLOSE(icbgrd)


deallocate(igpnts, igsptr, igeptr)

return

101 continue

Call DispFileReadError()

end subroutine ReadP3DGrid

!**************************************************************************************************
SUBROUTINE ReadP3DGridHeader(gridfilename,grdformat,nzones,ndm,&
                             multib,totpnts)
!**************************************************************************************************

IMPLICIT NONE

character(len=chrlen),intent(in) :: gridfilename

integer,intent(out) :: ndm,nzones
integer,intent(out),optional :: totpnts
integer, dimension(:),allocatable :: ieng,jeng,keng

integer,intent(in) :: grdformat
logical,intent(in) :: multib
character(len=chrlen) :: fileformat

integer :: icbgrd,ngrd,ng,ierr

icbgrd = 22
ndm = 3
ngrd = 1

select case (grdformat)
case(0)
   fileformat = 'unformatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
case(1)
   fileformat = 'formatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,*,ERR=101,END=101) ngrd
case(2)
   fileformat = 'binary'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
end select


nzones = ngrd

select case(present(totpnts))
case(.true.)
  ALLOCATE(ieng(ngrd),jeng(ngrd),keng(ngrd),stat=ierr)
  CALL CheckAllocationError(ierr)

  select case (grdformat)
  case(0,2)
    READ(icbgrd,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
  case(1)
    READ(icbgrd,*,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
  end select

  totpnts = 0
  do ng = 1,ngrd
    totpnts = totpnts + ieng(ng)*jeng(ng)*keng(ng)
  enddo
end select

close(icbgrd)

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadP3DGridHeader

!**************************************************************************************************
SUBROUTINE CheckP3DRstrtHeader(gridfilename,grdformat, & 
                                     rstrtfilename,rstrtformat,rstrtdouble,rtype,multib)
!**************************************************************************************************

IMPLICIT NONE

character(len=chrlen),intent(in) :: gridfilename,rstrtfilename

integer, dimension(:),allocatable :: ieng,jeng,keng
integer, dimension(:),pointer :: iengrst,jengrst,kengrst

integer,intent(in) :: grdformat,rstrtformat,rtype
logical,intent(in) :: rstrtdouble,multib
character(len=chrlen) :: fileformat

integer :: icbrst,ngrd
integer :: icbgrd,ng,ierr
integer :: ngrdrst
logical :: lstop

icbgrd = 22
icbrst = 23
ngrd = 1
lstop = .false.
nullify(iengrst,jengrst,kengrst)

select case (grdformat)
case(0)
   fileformat = 'unformatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
case(1)
   fileformat = 'formatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,*,ERR=101,END=101) ngrd
case(2)
   fileformat = 'binary'
   open(icbgrd,file=gridfilename,form=fileformat,status='old')
   if(multib) READ(icbgrd,ERR=101,END=101) ngrd
end select

Call ReadRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrdrst, &
                       rtype,multib,iengrst,jengrst,kengrst)
close(icbfdlread)

if(multib) then
  if(ngrd.ne.ngrdrst) then
    write(*,*)
    write(*,*) 'Number of blocks in Grid do not match the Restart file'
    write(*,*) 'Number of Grid blocks   : ',ngrd
    write(*,*) 'Number of Restart blocks: ',ngrdrst
    
    write(*,*)
    write(*,*) '  --- CODE HALTED ---'
    stop
  endif
endif


ALLOCATE(ieng(ngrd),jeng(ngrd),keng(ngrd),stat=ierr)
CALL CheckAllocationError(ierr)

select case (grdformat)
case(0,2)
  READ(icbgrd,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
case(1)
  READ(icbgrd,*,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
end select

do ng = 1, ngrd
  if(ieng(ng).ne.iengrst(ng) .or. &
     jeng(ng).ne.jengrst(ng) .or. &
     keng(ng).ne.kengrst(ng) ) then

    write(*,*) 
    write(*,*) 'Indecies for block ',ng,' in grid and restart file do not match.'
    write(*,*) '              i     j     k  '
    write(*,100)' Grid   : ',ieng(ng),jeng(ng),keng(ng)
    write(*,100)' Restart: ',iengrst(ng),jengrst(ng),kengrst(ng)
    lstop = .true.
  endif
enddo


deallocate(ieng,jeng,keng)
deallocate(iengrst,jengrst,kengrst)

close(icbgrd)
close(icbrst)

if(lstop) then
    write(*,*)
    write(*,*) '  --- CODE HALTED ---'
    stop
endif

return

100 format(A,i6,i6,i6)
101 continue

Call DispFileReadError()

END SUBROUTINE CheckP3DRstrtHeader



!**************************************************************************************************
SUBROUTINE WriteP3DGrid(gridfilename,grdformat,griddouble,nzones,ndm,x,ieng,jeng,keng,iblank,iblanked,multib)
!**************************************************************************************************
!Reads in a Plot3D grid file and converts it to an AVUS grid format
!
!Input:
!	gridfilename - Plot3d file read out
!
!Outputs and Allocates:
!
!	dx(3,:) - Grid coodrinates in double precision
!
!	igsptr - Grid starting indexes in xt,yt,zt,iblankt
!	igeptr - Grid ending indexes in xt,yt,zt,iblankt
!	igpnts - Grid sizes
!USE MOD_AVUSGrid

IMPLICIT NONE


real (kind=FP),dimension(:,:),pointer :: x

character(len=chrlen),intent(in) :: gridfilename

integer,intent(in) :: nzones,ndm
integer, dimension(:),pointer :: ieng,jeng,keng,iblank
integer, dimension(nzones)    :: igpnts, igsptr, igeptr

integer,intent(in) :: grdformat
logical,intent(in) :: griddouble,iblanked,multib

real (kind=SP),dimension(:,:)    ,allocatable :: sx
real (kind=DP),dimension(:,:)    ,allocatable :: dx
character(len=chrlen) :: fileformat


INTEGER :: ngrd, ng, i1, i2, igall, iengx, jengx, kengx, icbgrd
INTEGER :: i,n,np
INTEGER :: ierr
!INTEGER :: iprec,ib

Call DispReadWrite(dsp_write,dsp_grid,dsp_p3dfile,gridfilename,grdformat,griddouble,iblanked=iblanked,multib=multib)


icbgrd = 13
ngrd = nzones
!call p3dfor(gridfilename,iprec,ib)

!if(griddouble.and.iprec.eq.1) then
!  write(*,*) 

!elseif(.not.griddouble.and.iprec.eq.2) then

!endif


select case (grdformat)
case(0)
   fileformat = 'unformatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='unknown')
   if(multib) WRITE(icbgrd) ngrd
case(1)
   fileformat = 'formatted'
   open(icbgrd,file=gridfilename,form=fileformat,status='unknown')
   if(multib) WRITE(icbgrd,*) ngrd
case(2)
   fileformat = 'binary'
   open(icbgrd,file=gridfilename,form=fileformat,status='unknown')
   if(multib) WRITE(icbgrd) ngrd
end select



select case (grdformat)
case(0,2)
  WRITE(icbgrd) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
case(1)
  WRITE(icbgrd,*) (ieng(ng), jeng(ng), keng(ng), ng=1,ngrd)
end select

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)

   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO

200 format(A,i5.1,A,3i5.1)

select case (grdformat)
case(0,2)
  if(griddouble)then
    
    select case(FP)
    case(SP)
      allocate( dx(3,igall),stat = ierr)
      CALL CheckAllocationError(ierr)

      do np=1,igall
        dx(:,np) = x(:,np)
      enddo

      DO ng = 1, ngrd
        WRITE(*,*) 'Writing grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          WRITE(icbgrd) ((dx(n,i),i=i1,i2),n=1,ndm), &
                         (iblank(i),i=i1,i2)
        case(.false.)
          WRITE(icbgrd) ((dx(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO

      deallocate(dx)
    case(DP)
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          WRITE(icbgrd) ((x(n,i),i=i1,i2),n=1,ndm), &
                         (iblank(i),i=i1,i2)
        case(.false.)
          WRITE(icbgrd) ((x(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO
    end select
  else

    select case(FP)
    case(SP)
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          WRITE(icbgrd) ((x(n,i),i=i1,i2),n=1,ndm), &
                         (iblank(i),i=i1,i2)
        case(.false.)
          WRITE(icbgrd) ((x(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO
    case(DP)
      allocate( sx(3,igall),stat = ierr)
      CALL CheckAllocationError(ierr)

      do np=1,igall
        sx(:,np) = x(:,np)
      enddo

      DO ng = 1, ngrd
        WRITE(*,*) 'Writing grid point data for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        select case(iblanked)
        case(.true.)
          WRITE(icbgrd) ((sx(n,i),i=i1,i2),n=1,ndm), &
                         (iblank(i),i=i1,i2)
        case(.false.)
          WRITE(icbgrd) ((sx(n,i),i=i1,i2),n=1,ndm)
        end select
      END DO

      deallocate(sx)
    end select
  endif
case(1)

    DO ng = 1, ngrd
      WRITE(*,*) 'Writing grid point data for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      select case(iblanked)
      case(.true.)
        WRITE(icbgrd,*) ((x(n,i),i=i1,i2),n=1,ndm), &
                         (iblank(i),i=i1,i2)
      case(.false.)
        WRITE(icbgrd,*) ((x(n,i),i=i1,i2),n=1,ndm)
      end select
    END DO
end select

CLOSE(icbgrd)


end subroutine  WriteP3DGrid
!**************************************************************************************************
SUBROUTINE ReadP3DRestart(rstrtfilename,rstrtformat,rstrtdouble,nzones,npnts,w,x,ieng,jeng,keng,rtype,multib)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones,rtype
integer, dimension(:),pointer :: npnts
integer, dimension(:),pointer :: ieng,jeng,keng
integer, dimension(nzones) :: igpnts, igsptr, igeptr

real (kind=FP),dimension(:,:)    ,pointer :: w,x
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib

real (kind=SP),dimension(:,:)    ,pointer :: sw,sx,swtmp
real (kind=DP),dimension(:,:)    ,pointer :: dw,dx,dwtmp

real (kind=FP),dimension(:,:)    ,pointer :: wtmp

INTEGER :: ngrd, ng, i1, i2, igall, iengx, jengx, kengx
INTEGER :: i,j,ierr,n,ifl,ndm,nvar1,nvar2,np
INTEGER :: totnumrstrtvar

REAL (kind=DP) :: djunk
REAL (kind=SP) :: sjunk

ndm = 3


select case(rtype)
case(0,1,fdlmin:fdlmax)
  i = dsp_fdl3difile
  n = dsp_rstrt
case(2)
  i = dsp_p3dfile
  n = dsp_p3dQfile
case(3)
  i = dsp_p3dfile
  n = dsp_p3dFunfile
end select

Call DispReadWrite(dsp_read,n,i,rstrtfilename,rstrtformat,rstrtdouble,multib=multib)

Call ReadRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,rtype,multib)

!There is always 5 vaiables in a q file
if(rtype.eq.2) then
  numrstrtvar(1) = 5
  if(rstrtdouble.or.rstrtformat.eq.1) then
     allocate(DQVar(NGRD,4))
     DQVar(:,:) = 0.0_DP
  else
     allocate(SQVar(NGRD,4))
     SQVar(:,:) = 0.0_SP
  endif
endif

igall = sum(npnts)
select case(readallvars)
case(.false.)
  allocate( w(1,igall),stat = ierr)
  CALL CheckAllocationError(ierr)
  close(icbfdlread)
  return
end select

WRITE(*,*) 'Number of grids is ',ngrd

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)
   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO


if(rtype.eq.3) numrstrtvar(1) = maxval(NFVAR)




select case (rstrtformat)
case(0,2)
  if(rstrtdouble)then

    select case(FP)
    case(SP)
      allocate( dw(sum(numrstrtvar),igall),stat = ierr)
    case(DP)
      allocate( w(sum(numrstrtvar),igall),stat = ierr)
    end select
    CALL CheckAllocationError(ierr)

    select case(rtype)
    case(fdl_1,fdl_2) !Grid is contined within the restart file
      totnumrstrtvar = sum(numrstrtvar)
      numrstrtvar(1) = 5
      ifl = 1
      iflag(:) = 0
      allocate( x(ndm,igall),stat = ierr)
      CALL CheckAllocationError(ierr)
      select case(FP)
      case(SP)
        allocate( dx(ndm,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
      case(DP)
      end select
    end select

    DO ng = 1, ngrd
      IF(DEBUGPRINT)WRITE(*,*) 'Reading flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2)then
        READ(icbfdlread,ERR=101,END=101) DQvar(ng,:)
        p3dxm1  = DQvar(ng,1)
        p3dalfa = DQvar(ng,2)
        p3dre   = DQvar(ng,3)
        p3dtau  = DQvar(ng,4)
      endif
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(FP)
      case(SP)
        select case(rtype)
        case(fdl_1,fdl_2)
          READ(icbfdlread,ERR=101,END=101) ((dx(n,i),i=i1,i2),n=1,ndm), &
                                           ((dw(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
          IF (itimeint .ge. 2) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
          IF (itimeint .eq. 3) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
        end select
      case(DP)
        select case(rtype)
        case(fdl_1,fdl_2)
          READ(icbfdlread,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                           ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
          IF (itimeint .ge. 2) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
          IF (itimeint .eq. 3) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
        end select
      end select

    END DO

    select case(rtype)
    case(fdl_2)
      read(icbfdlread,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          if(totnumrstrtvar.lt.10) then
            select case(FP)
            case(SP)
              allocate( dwtmp(10,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              dwtmp(1:5,1:igeptr(ngrd)) = dw(1:5,1:igeptr(ngrd))
              deallocate(dw)
              dw => dwtmp
              nullify(dwtmp)
            case(DP)
              allocate( wtmp(10,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
              deallocate(w)
              w => wtmp
              nullify(wtmp)
            end select
            totnumrstrtvar = 10
          endif
          numrstrtvar(2) = 10-5
          select case(FP)
          case(SP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((dw(n,i),i=i1,i2),n=6,10)
            ENDDO
          case(DP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=6,10)
            ENDDO
          end select
        case(41)
          select case(ifl)
          case(1)
            if(totnumrstrtvar.lt.8) then
              select case(FP)
              case(SP)
                allocate( dwtmp(8,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                dwtmp(1:5,1:igeptr(ngrd)) = dw(1:5,1:igeptr(ngrd))
                deallocate(dw)
                dw => dwtmp
                nullify(dwtmp)
              case(DP)
                allocate( wtmp(8,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
                deallocate(w)
                w => wtmp
                nullify(wtmp)
              end select
              totnumrstrtvar = 8
              numrstrtvar(2) = 8-5
            endif
            nvar1 = 6
            nvar2 = 8
          case(2)
            if(totnumrstrtvar.lt.13) then
              select case(FP)
              case(SP)
                allocate( dwtmp(13,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                dwtmp(1:10,1:igeptr(ngrd)) = dw(1:10,1:igeptr(ngrd))
                deallocate(dw)
                dw => dwtmp
                nullify(dwtmp)
              case(DP)
                allocate( wtmp(13,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                wtmp(1:10,1:igeptr(ngrd)) = w(1:10,1:igeptr(ngrd))
                deallocate(w)
                w => wtmp
                nullify(wtmp)
              end select
              totnumrstrtvar = 13
              numrstrtvar(2) = 13-5
            endif
            nvar1 = 11
            nvar2 = 13
          end select
          select case(FP)
          case(SP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((dw(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          case(DP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          end select
        end select
        ifl = ifl + 1
        read(icbfdlread,ERR=101,END=101) iflag(ifl)
      enddo

      select case(FP)
      case(SP)
        allocate( x(ndm,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        x(1:ndm,1:igall) = dx(1:ndm,1:igall)
        DEALLOCATE(dx)
      case(DP)
      end select
    end select

    select case(FP)
    case(SP)
      allocate( w(sum(numrstrtvar),igall),stat = ierr)
      CALL CheckAllocationError(ierr)
      do np=1,igall
        w(:,np) = dw(:,np)
      enddo
      deallocate(dw)
    case(DP)
    end select

  else

    select case(FP)
    case(SP)
      allocate( w(sum(numrstrtvar),igall),stat = ierr)
    case(DP)
      allocate( sw(sum(numrstrtvar),igall),stat = ierr)
    end select
    CALL CheckAllocationError(ierr)

    select case(rtype)
    case(fdl_1,fdl_2) !Grid is contined within the restart file
      totnumrstrtvar = sum(numrstrtvar)
      numrstrtvar(1) = 5
      ifl = 1
      iflag(:) = 0
      select case(FP)
      case(SP)
        allocate( x(ndm,igall),stat = ierr)
      case(DP)
        allocate( sx(ndm,igall),stat = ierr)
      end select
      CALL CheckAllocationError(ierr)
    end select

    DO ng = 1, ngrd
      IF(DEBUGPRINT)WRITE(*,*) 'Reading flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2)then
        READ(icbfdlread,ERR=101,END=101) SQvar(ng,:)
        p3dxm1  = DBLE(SQvar(ng,1))
        p3dalfa = DBLE(SQvar(ng,2))
        p3dre   = DBLE(SQvar(ng,3))
        p3dtau  = DBLE(SQvar(ng,4))
      endif
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(FP)
      case(SP)
        select case(rtype)
        case(fdl_2)
          READ(icbfdlread,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                           ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
          IF (itimeint .ge. 2) READ(icbfdlread,ERR=101,END=101) ((sjunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
          IF (itimeint .eq. 3) READ(icbfdlread,ERR=101,END=101) ((sjunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
        end select
      case(DP)
        select case(rtype)
        case(fdl_2)
          READ(icbfdlread,ERR=101,END=101) ((sx(n,i),i=i1,i2),n=1,ndm), &
                                           ((sw(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            READ(icbfdlread,ERR=101,END=101) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
          IF (itimeint .ge. 2) READ(icbfdlread,ERR=101,END=101) ((sjunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
          IF (itimeint .eq. 3) READ(icbfdlread,ERR=101,END=101) ((sjunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
        end select
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      read(icbfdlread,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          if(totnumrstrtvar.lt.10) then
            select case(FP)
            case(SP)
              allocate( wtmp(10,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
              deallocate(w)
              w => wtmp
              nullify(wtmp)
            case(DP)
              allocate( swtmp(10,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              swtmp(1:5,1:igeptr(ngrd)) = sw(1:5,1:igeptr(ngrd))
              deallocate(sw)
              sw => swtmp
              nullify(swtmp)
            end select
            totnumrstrtvar = 10
          endif
          numrstrtvar(2) = 10-5
          select case(FP)
          case(SP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=6,10)
            ENDDO
          case(DP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((sw(n,i),i=i1,i2),n=6,10)
            ENDDO
          end select
        case(41)
          select case(ifl)
          case(1)
            if(totnumrstrtvar.lt.8) then
              select case(FP)
              case(SP)
                allocate( wtmp(8,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
                deallocate(w)
                w => wtmp
                nullify(wtmp)
              case(DP)
                allocate( swtmp(8,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                swtmp(1:5,1:igeptr(ngrd)) = sw(1:5,1:igeptr(ngrd))
                deallocate(sw)
                sw => swtmp
                nullify(swtmp)
              end select
              totnumrstrtvar = 8
              numrstrtvar(2) = 8-5
            endif
            nvar1 = 6
            nvar2 = 8
          case(2)
            if(totnumrstrtvar.lt.13) then
              select case(FP)
              case(SP)
                allocate( wtmp(13,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                wtmp(1:10,1:igeptr(ngrd)) = w(1:10,1:igeptr(ngrd))
                deallocate(w)
                w => wtmp
                nullify(wtmp)
              case(DP)
                allocate( swtmp(13,igall),stat = ierr)
                CALL CheckAllocationError(ierr)
                swtmp(1:10,1:igeptr(ngrd)) = sw(1:10,1:igeptr(ngrd))
                deallocate(sw)
                sw => swtmp
                nullify(swtmp)
              end select
              totnumrstrtvar = 13
              numrstrtvar(2) = 13-5
            endif
            nvar1 = 11
            nvar2 = 13
          end select
          select case(FP)
          case(SP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          case(DP)
            DO ng = 1, ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              READ(icbfdlread,ERR=101,END=101) ((sw(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          end select
        end select
        ifl = ifl + 1
        read(icbfdlread,ERR=101,END=101) iflag(ifl)
      enddo
 
      select case(FP)
      case(SP)
      case(DP)
        allocate( x(ndm,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        x(1:ndm,1:igall) = sx(1:ndm,1:igall)
        DEALLOCATE(sx)
      end select
   end select

    select case(FP)
    case(SP)
    case(DP)
      allocate( w(sum(numrstrtvar),igall),stat = ierr)
      CALL CheckAllocationError(ierr)
      do np=1,igall
        w(:,np) = sw(:,np)
      enddo
      deallocate(sw)
    end select
  endif
case(1)

    allocate( w(sum(numrstrtvar),igall),stat = ierr)
    CALL CheckAllocationError(ierr)

    select case(rtype)
    case(fdl_1,fdl_2) !Grid is contined within the restart file
      totnumrstrtvar = sum(numrstrtvar)
      numrstrtvar(1) = 5
      ifl = 1
      iflag(:) = 0
      allocate( x(ndm,igall),stat = ierr)
      CALL CheckAllocationError(ierr)
    end select

    DO ng = 1, ngrd
      IF(DEBUGPRINT)WRITE(*,*) 'Reading flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) READ(icbfdlread,*,ERR=101,END=101) DQvar(ng,:)
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(rtype)
      case(fdl_1,fdl_2)
        READ(icbfdlread,*,ERR=101,END=101) ((x(n,i),i=i1,i2),n=1,ndm), &
                                           ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          READ(icbfdlread,*,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          READ(icbfdlread,*,ERR=101,END=101) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
        IF (itimeint .ge. 2) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
        IF (itimeint .eq. 3) READ(icbfdlread,ERR=101,END=101) ((djunk,i=i1,i2),n=rstrtvarsct(1,1),rstrtvarsct(1,2))
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      read(icbfdlread,*,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          if(totnumrstrtvar.lt.10) then
            allocate( wtmp(10,igall),stat = ierr)
            CALL CheckAllocationError(ierr)
            wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
            deallocate(w)
            w => wtmp
            nullify(wtmp)
            totnumrstrtvar = 10
          endif
          numrstrtvar(2) = 10-5
          DO ng = 1, ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,*,ERR=101,END=101) ((w(n,i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            if(totnumrstrtvar.lt.8) then
              allocate( wtmp(8,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              wtmp(1:5,1:igeptr(ngrd)) = w(1:5,1:igeptr(ngrd))
              deallocate(w)
              w => wtmp
              nullify(dwtmp)
              totnumrstrtvar = 8
              numrstrtvar(2) = 8-5
            endif
            nvar1 = 6
            nvar2 = 8
          case(2)
            if(totnumrstrtvar.lt.13) then
              allocate( wtmp(13,igall),stat = ierr)
              CALL CheckAllocationError(ierr)
              wtmp(1:10,1:igeptr(ngrd)) = w(1:10,1:igeptr(ngrd))
              deallocate(w)
              w => wtmp
              nullify(dwtmp)
              totnumrstrtvar = 13
              numrstrtvar(2) = 13-5
            endif
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1, ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,*,ERR=101,END=101) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        read(icbfdlread,*,ERR=101,END=101) iflag(ifl)
      enddo
    end select


end select

CLOSE(icbfdlread)

if(rtype.eq.3) numrstrtvar(1) = maxval(NFVAR)

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadP3DRestart

!**************************************************************************************************
SUBROUTINE ReadRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,rtype,multib, &
                             iengrst,jengrst,kengrst)
!**************************************************************************************************
IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: rtype
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib
integer, intent(out) :: ngrd
integer,optional, dimension(:),pointer :: iengrst,jengrst,kengrst
integer, dimension(:),pointer :: ieng,jeng,keng
real(kind=SP) :: STAU,STAUAVG
integer :: ng,iengx, jengx, kengx,ierr
character(len=chrlen) :: fileformat

ngrd = 1 !This is for singel blocks
itimeint = 0

select case (rstrtformat)
case(0,2)
   select case(rstrtformat)
   case(0)
     fileformat = 'unformatted'
   case(2)
     fileformat = 'binary'
   end select
   
   open(icbfdlread,file=rstrtfilename,form=fileformat,status='old')
   
   if(rstrtdouble) then
     select case(rtype)
     case(0)
       READ(icbfdlread,ERR=101,END=101) fileversion, eqID
       READ(icbfdlread,ERR=101,END=101) NGRD, fdlnvar, itimeint, ibeta, nstep, dtau
       SELECT CASE(eqID)
       CASE(CFDID)
         READ(icbfdlread,ERR=101,END=101) isgsmodel, fdlre, fdlxm1, fdlalfa, fdlbeta, fdlgam, fdls1, fdlpr, fdlprt
       CASE(CEMID)

       CASE(CEMPPID)

       END SELECT

     case(1,fdl_1)
       READ(icbfdlread,ERR=101,END=101) NSTEP,DTAU,NGRD
     case(fdl_2)
       NGRD = 1
       allocate(ieng(1),jeng(1),keng(1),stat=ierr)
       CALL CheckAllocationError(ierr)
       READ(icbfdlread,ERR=101,END=101) ieng(1),jeng(1),keng(1),NSTEP,DTAU
     case(2,3) !Plot3D Q-File, Function File
       if(multib) READ(icbfdlread,ERR=101,END=101) NGRD
     end select
   else
     select case(rtype)
     case(0)
       READ(icbfdlread,ERR=101,END=101) fileversion, eqID
       READ(icbfdlread,ERR=101,END=101) NGRD, fdlnvar, itimeint, ibeta, nstep, stau
       SELECT CASE(eqID)
       CASE(CFDID)
         WRITE(*,*) 'Single precision FDLF90 reads not working!'
         STOP
         READ(icbfdlread,ERR=101,END=101) isgsmodel, fdlre, fdlxm1, fdlalfa, fdlbeta, fdlgam, fdls1, fdlpr, fdlprt
       CASE(CEMID)

       CASE(CEMPPID)

       END SELECT
       DTAU = STAU
     case(1,fdl_1)
       READ(icbfdlread,ERR=101,END=101) NSTEP,STAU,NGRD
       DTAU = STAU
     case(fdl_2)
       NGRD = 1
       allocate(ieng(1),jeng(1),keng(1),stat=ierr)
       CALL CheckAllocationError(ierr)
       READ(icbfdlread,ERR=101,END=101) ieng(1),jeng(1),keng(1),NSTEP,STAU
       DTAU = STAU
    case(2,3) !Plot3D Q-File, Function File
       if(multib) READ(icbfdlread,ERR=101,END=101) NGRD
     end select
   endif
case(1)
   open(icbfdlread,file=rstrtfilename,form='formatted',status='old')
   select case(rtype)
   case(0)
     READ(icbfdlread,*,ERR=101,END=101) fileversion, eqID
     READ(icbfdlread,*,ERR=101,END=101) NGRD, fdlnvar, itimeint, ibeta, nstep, dtau
     SELECT CASE(eqID)
     CASE(CFDID)
       READ(icbfdlread,*,ERR=101,END=101) isgsmodel, fdlre, fdlxm1, fdlalfa, fdlbeta, fdlgam, fdls1, fdlpr, fdlprt
     CASE(CEMID)

     CASE(CEMPPID)

     END SELECT
   case(1,fdl_1)
     READ(icbfdlread,*,ERR=101,END=101) NSTEP,DTAU,NGRD
   case(fdl_2)
     NGRD = 1
     allocate(ieng(1),jeng(1),keng(1),stat=ierr)
     CALL CheckAllocationError(ierr)
     READ(icbfdlread,*,ERR=101,END=101) ieng(1),jeng(1),keng(1),NSTEP,DTAU
   case(2,3) !Plot3D Q-File, Function File
     if(multib) READ(icbfdlread,*,ERR=101,END=101) NGRD
   end select
end select

allocate(ieng(ngrd),jeng(ngrd),keng(ngrd),stat=ierr)
CALL CheckAllocationError(ierr)


select case(rtype)
case(0,1)!FDL3DI

  select case (rstrtformat)
  case(0,2)
    do ng = 1,ngrd
      READ(icbfdlread,ERR=101,END=101) ieng(ng), jeng(ng), keng(ng)
    enddo
  case(1)
    do ng = 1,ngrd
      READ(icbfdlread,*,ERR=101,END=101) ieng(ng), jeng(ng), keng(ng)
    enddo
  end select
case(2)!Plot3D Q-File

  select case (rstrtformat)
  case(0,2)
    READ(icbfdlread,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng),ng=1,ngrd)
  case(1)
    READ(icbfdlread,*,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng),ng=1,ngrd)
  end select
case(3)!Plot3D Function File
  if(.not.allocated(NFVAR)) then
    allocate(NFVAR(ngrd),stat = ierr)
    CALL CheckAllocationError(ierr)
  endif

  select case (rstrtformat)
  case(0,2)
    READ(icbfdlread,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng),NFVAR(ng),ng=1,ngrd)
  case(1)
    READ(icbfdlread,*,ERR=101,END=101) (ieng(ng), jeng(ng), keng(ng),NFVAR(ng),ng=1,ngrd)
  end select
end select

if(present(iengrst)) then
  iengrst => ieng
  jengrst => jeng
  kengrst => keng
else
  deallocate(ieng,jeng,keng)
endif

return

101 continue

Call DispFileReadError()

END SUBROUTINE ReadRestartHeader

!**************************************************************************************************
SUBROUTINE WriteP3DRestart(rstrtfilename,rstrtformat,rstrtdouble,ngp,nzones,w,x,ieng,jeng,keng,rtype,multib)
!**************************************************************************************************
USE MOD_GridVars

IMPLICIT NONE


character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones,rtype,ngp
integer, dimension(:),pointer :: ieng,jeng,keng
integer, dimension(nzones) :: igpnts, igsptr, igeptr

real (kind=FP),dimension(:,:)    ,pointer :: w,x

integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib

real (kind=SP),dimension(:,:)    ,allocatable ::  sw,sx
real (kind=DP),dimension(:,:)    ,allocatable ::  dw,dx

INTEGER :: ngrd, ng, i1, i2, igall, iengx, jengx, kengx
INTEGER :: i,j,ierr,n,ifl,nvar1,nvar2,np




select case(rtype)
case(0,1,fdlmin:fdlmax)
  i = dsp_fdl3difile
  n = dsp_rstrt
case(2)
  i = dsp_p3dfile
  n = dsp_p3dQfile
case(3)
  i = dsp_p3dfile
  n = dsp_p3dFunfile
end select


Call DispReadWrite(dsp_write,n,i,rstrtfilename,rstrtformat,rstrtdouble,multib=multib)

ngrd = nzones

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)
   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO
  

Call WriteRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,ieng,jeng,keng,rtype,multib)

select case(rtype)
case(3)
  numrstrtvar(1) = maxval(NFVAR)
case(fdl_2)
  numrstrtvar(1) = 5
end select

select case (rstrtformat)
case(0,2)
  if(rstrtdouble)then

    select case(FP)
    case(SP)
      allocate( dw(sum(numrstrtvar),igall),stat = ierr)
      CALL CheckAllocationError(ierr)
      do np=1,igall
        dw(:,np) = w(:,np)
      enddo
    case(DP)
    end select

    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
        allocate( dx(3,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        dx(1:3,1:igall) = x(1:3,1:igall)
      case(DP)
      end select
    end select
  
    if(.not.allocated(DQvar).and.rtype.eq.2) then
      allocate(DQvar(ngrd,4),stat=ierr)
      CALL CheckAllocationError(ierr)
      do ng = 1, ngrd
        DQvar(ng,:) = SQVar(min(ng,ngp),:)
      enddo
    endif

    
    DO ng = 1, ngrd
      WRITE(*,*) 'Writing flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) WRITE(icbfdlwrite) DQvar(min(ng,ngp),:)
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(FP)
      case(SP)
        select case(rtype)
        case(fdl_1,fdl_2)
          WRITE(icbfdlwrite) ((dx(n,i),i=i1,i2),n=1,3), &
                             ((dw(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        end select
      case(DP)
        select case(rtype)
        case(fdl_1,fdl_2)
          WRITE(icbfdlwrite) ((x(n,i),i=i1,i2),n=1,3), &
                             ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        end select
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          select case(FP)
          case(SP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=6,10)
            ENDDO
          case(DP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=6,10)
            ENDDO
          end select
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          select case(FP)
          case(SP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          case(DP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          end select
        end select
        ifl = ifl + 1
        write(icbfdlwrite) iflag(ifl)
      end do
    end select
  else

    select case(FP)
    case(SP)
    case(DP)
      allocate( sw(sum(numrstrtvar),igall),stat = ierr)
      CALL CheckAllocationError(ierr)
      do np=1,igall
        sw(:,np) = w(:,np)
      enddo
    end select


    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
      case(DP)
        allocate( sx(3,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        sx(1:3,1:igall) = x(1:3,1:igall)
      end select
    end select

    if(.not.allocated(SQvar).and.rtype.eq.2) then
      allocate(SQvar(ngrd,4),stat=ierr)
      CALL CheckAllocationError(ierr)
      do ng = 1, ngrd
        SQvar(ng,:) = DQVar(min(ng,ngp),:)
      enddo
    endif

   
    DO ng = 1, ngrd
      WRITE(*,*) 'Writing flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) WRITE(icbfdlwrite) SQvar(min(ng,ngp),:)
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(FP)
      case(SP)
        select case(rtype)
        case(fdl_1,fdl_2)
          WRITE(icbfdlwrite) ((x(n,i),i=i1,i2),n=1,3), &
                             ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        end select
      case(DP)
        select case(rtype)
        case(fdl_1,fdl_2)
          WRITE(icbfdlwrite) ((sx(n,i),i=i1,i2),n=1,3), &
                             ((sw(n,i),i=i1,i2),n=1,numrstrtvar(1))
          do j=2,nfilesect
            WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        case default
          do j=1,nfilesect
            WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
          enddo
        end select
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          select case(FP)
          case(SP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=6,10)
            ENDDO
          case(DP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=6,10)
            ENDDO
          end select
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          select case(FP)
          case(SP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          case(DP)
            DO ng = 1,ngrd
              i1 = igsptr(ng)
              i2 = igeptr(ng)
              WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=nvar1,nvar2)
            ENDDO
          end select
        end select
        ifl = ifl + 1
        write(icbfdlwrite) iflag(ifl)
      end do
    end select
  endif
case(1)

    if(.not.allocated(DQvar).and.rtype.eq.2) then
      allocate(DQvar(ngrd,4),stat=ierr)
      CALL CheckAllocationError(ierr)
      do ng = 1, ngrd
        DQvar(ng,:) = SQVar(min(ng,ngp),:)
      enddo
    endif

    DO ng = 1, ngrd
      WRITE(*,*) 'Writing flow variables for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) WRITE(icbfdlwrite,*) DQvar(min(ng,ngp),:)
      if(rtype.eq.3) numrstrtvar(1) = NFVAR(NG)
      select case(rtype)
      case(fdl_1,fdl_2)
        WRITE(icbfdlwrite,*) ((x(n,i),i=i1,i2),n=1,3), &
                             ((w(n,i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          WRITE(icbfdlwrite,*) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          WRITE(icbfdlwrite,*) ((w(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite,*) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            WRITE(icbfdlwrite,*) ((w(n,i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            WRITE(icbfdlwrite,*) ((w(n,i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        write(icbfdlwrite,*) iflag(ifl)
      end do
    end select
end select

CLOSE(icbfdlread)


END SUBROUTINE WriteP3DRestart

!**************************************************************************************************
SUBROUTINE WriteRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,ieng,jeng,keng,rtype,multib)
!**************************************************************************************************
IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: rtype
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib
integer, dimension(:),pointer :: ieng,jeng,keng
integer, intent(in) :: ngrd

real(kind=SP) :: STAU,STAUAVG
integer :: ng,ierr
character(len=chrlen) :: fileformat

select case (rstrtformat)
case(0,2)
   select case(rstrtformat)
   case(0)
     fileformat = 'unformatted'
   case(2)
     fileformat = 'binary'
   end select
   
   open(icbfdlwrite,file=rstrtfilename,form=fileformat,status='unknown')
   
   if(rstrtdouble) then
     select case(rtype)
     case(0)
       WRITE(*,*) 'FDLF90 Writer not implimented!'
       STOP
     case(1,fdl_1)
       WRITE(icbfdlwrite) NSTEP,DTAU,NGRD
     case(fdl_2)
       WRITE(icbfdlwrite) ieng(1),jeng(1),keng(1),NSTEP,DTAU
     case(2,3) !Plot3D Q-File, Function File
       if(multib) WRITE(icbfdlwrite) NGRD
     end select
  
   else
     select case(rtype)
     case(0)
       STAU = DTAU
       WRITE(*,*) 'FDLF90 Writer not implimented!'
       STOP
     case(1,fdl_1)
       STAU = DTAU
       WRITE(icbfdlwrite) NSTEP,STAU,NGRD
     case(fdl_2)
       STAU = DTAU
       WRITE(icbfdlwrite) ieng(1),jeng(1),keng(1),NSTEP,STAU
     case(2,3) !Plot3D Q-File, Function File
       if(multib) WRITE(icbfdlwrite) NGRD
     end select
 
   endif
case(1)
   open(icbfdlwrite,file=rstrtfilename,form='formatted',status='unknown')
   select case(rtype)
   case(0)
       WRITE(*,*) 'FDLF90 Writer not implimented!'
       STOP
   case(1,fdl_1)
     WRITE(icbfdlwrite,*) NSTEP,DTAU,NGRD
   case(fdl_2)
     WRITE(icbfdlwrite,*) ieng(1),jeng(1),keng(1),NSTEP,DTAU
   case(2,3) !Plot3D Q-File, Function File
     if(multib) WRITE(icbfdlwrite,*) NGRD
   end select

end select


select case(rtype)
case(0,1)
  select case (rstrtformat)
  case(0,2)
    do ng = 1,ngrd
      WRITE(icbfdlwrite) ieng(ng), jeng(ng), keng(ng)
    enddo
  case(1)
    do ng = 1,ngrd
      WRITE(icbfdlwrite,*) ieng(ng), jeng(ng), keng(ng)
    enddo
  end select
case(2) !Plot3D Q-File
  select case (rstrtformat)
  case(0,2)
    WRITE(icbfdlwrite) (ieng(ng), jeng(ng), keng(ng),ng=1,ngrd)
  case(1)
    WRITE(icbfdlwrite,*) (ieng(ng), jeng(ng), keng(ng),ng=1,ngrd)
  end select
case(3) !Plot3D Function File
  if(.not.allocated(NFVAR)) then
    allocate(NFVAR(ngrd),stat=ierr)
    Call CheckAllocationError(ierr)
    NFVAR(:) = numrstrtvar(1)
  endif
  select case (rstrtformat)
  case(0,2)
    WRITE(icbfdlwrite) (ieng(ng), jeng(ng), keng(ng),NFVAR(ng),ng=1,ngrd)
  case(1)
    WRITE(icbfdlwrite,*) (ieng(ng), jeng(ng), keng(ng),NFVAR(ng),ng=1,ngrd)
  end select
end select

END SUBROUTINE WriteRestartHeader

!*****************************************************************************
SUBROUTINE ReadNextP3DVar(nvar,rstrtfilename,rstrtformat,rstrtdouble,nzones,npnts,w,ieng,jeng,keng,rtype,multib)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nvar,nzones,rtype
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib
integer, dimension(:),pointer :: ieng,jeng,keng
integer, dimension(nzones) :: igpnts, igsptr, igeptr

real(kind=SP) ,dimension(:,:),pointer ::  sw
real(kind=DP) ,dimension(:,:),pointer ::  dw

!INTEGER, dimension(:,:), allocatable :: readi
type(dbpntr), dimension(:), pointer :: dpntr
type(sgpntr), dimension(:), pointer :: spntr
type(intpntr),dimension(:), pointer :: readii

real(kind=SP) :: srjunk
real(kind=DP) :: drjunk

INTEGER :: j, ngrd, ng, i1, i2, igall, iengx, jengx, kengx, ifl
INTEGER :: ierr,n,nvar1,nvar2

INTEGER, TARGET :: i,dumy

REAL(kind=DP),target :: djunk(1,1)
REAL(kind=SP),target :: sjunk(1,1)

select case(rtype)
case(0,1,fdlmin:fdlmax)
  i = dsp_fdl3difile
  n = dsp_rstrt
case(2)
  i = dsp_p3dfile
  n = dsp_p3dQfile
case(3)
  i = dsp_p3dfile
  n = dsp_p3dfunfile
end select

Call DispReadWrite(dsp_read,n,i,rstrtfilename,rstrtformat,rstrtdouble,nvar=nvar,multib=multib)

ngrd = nzones

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)
   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO

Call ReadRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,rtype,multib)

select case (rstrtformat)
case(0,2)
  if(rstrtdouble)then
    allocate( dw(1,sum(npnts)),stat = ierr)
    CALL CheckAllocationError(ierr)
    allocate(dpntr(sum(numrstrtvar)),readii(sum(numrstrtvar)),stat=ierr)
    CALL CheckAllocationError(ierr)

    dw(:,:) = 0.0_DP
    do n=1,sum(numrstrtvar)
      dpntr(n)%dw => djunk
    enddo
    dpntr(nvar)%dw => dw

    dumy = 1
    do n=1,sum(numrstrtvar)
      readii(n)%i => dumy
    enddo
    readii(nvar)%i => i


    DO ng = 1, ngrd
      WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) READ(icbfdlread,ERR=101,END=101) DQvar(ng,:)
      select case(rtype)
      case(fdl_1,fdl_2)
        READ(icbfdlread,ERR=101,END=101) ((drjunk,i=i1,i2),n=1,3), &
                                         ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          READ(icbfdlread,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          READ(icbfdlread,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    if(nvar.ge.6.and.rtype.eq.fdl_2) then
      ifl = 1
      read(icbfdlread,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        read(icbfdlread,ERR=101,END=101) iflag(ifl)
      end do
    end if

    w(1,1:sum(npnts)) = dw(1,1:sum(npnts))
    deallocate(dpntr,readii,dw)
  else
    allocate( sw(1,sum(npnts)),stat = ierr)
    CALL CheckAllocationError(ierr)
    allocate(spntr(sum(numrstrtvar)),readii(sum(numrstrtvar)),stat=ierr)
    CALL CheckAllocationError(ierr)

    sw(:,:) = 0.0_DP
    do n=1,sum(numrstrtvar)
      spntr(n)%w => sjunk
    enddo
    spntr(nvar)%w => sw

    dumy = 1
    do n=1,sum(numrstrtvar)
      readii(n)%i => dumy
    enddo
    readii(nvar)%i => i

    DO ng = 1, ngrd
      WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) READ(icbfdlread,ERR=101,END=101) SQvar(ng,:)
       select case(rtype)
      case(fdl_1,fdl_2)
        READ(icbfdlread,ERR=101,END=101) ((srjunk,i=i1,i2),n=1,3), &
                                         ((spntr(n)%w(1,readii(n)%i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          READ(icbfdlread,ERR=101,END=101) ((spntr(n)%w(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          READ(icbfdlread,ERR=101,END=101) ((spntr(n)%w(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    if(nvar.ge.6.and.rtype.eq.fdl_2) then
      ifl = 1
      read(icbfdlread,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,ERR=101,END=101) ((spntr(n)%w(1,readii(n)%i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,ERR=101,END=101) ((spntr(n)%w(1,readii(n)%i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        read(icbfdlread,ERR=101,END=101) iflag(ifl)
      end do
    end if


    w(1,1:sum(npnts)) = sw(1,1:sum(npnts))
    deallocate(spntr,readii,sw)
  endif

case(1)
    allocate( dw(1,sum(npnts)),stat = ierr)
    CALL CheckAllocationError(ierr)
    allocate(dpntr(sum(numrstrtvar)),readii(sum(numrstrtvar)),stat=ierr)
    CALL CheckAllocationError(ierr)

    dw(:,:) = 0.0_DP
    do n=1,sum(numrstrtvar)
      dpntr(n)%dw => djunk
    enddo
    dpntr(nvar)%dw => dw

    dumy = 1
    do n=1,sum(numrstrtvar)
      readii(n)%i => dumy
    enddo
    readii(nvar)%i => i

    DO ng = 1, ngrd
      WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
      i1 = igsptr(ng)
      i2 = igeptr(ng)
      if(rtype.eq.2) READ(icbfdlread,*,ERR=101,END=101) DQvar(ng,:)
      select case(rtype)
      case(fdl_1,fdl_2)
        READ(icbfdlread,*,ERR=101,END=101) ((drjunk,i=i1,i2),n=1,3), &
                                           ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          READ(icbfdlread,*,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          READ(icbfdlread,*,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    if(nvar.ge.6.and.rtype.eq.fdl_2) then
      ifl = 1
      read(icbfdlread,*,ERR=101,END=101) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,*,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = igsptr(ng)
            i2 = igeptr(ng)
            READ(icbfdlread,*,ERR=101,END=101) ((dpntr(n)%dw(1,readii(n)%i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        read(icbfdlread,*,ERR=101,END=101) iflag(ifl)
      end do
    end if

    
    w(1,1:sum(npnts)) = dw(1,1:sum(npnts))
    deallocate(dpntr,readii,dw)
end select

close(icbfdlread)

return

101 continue

Call DispFileReadError()

end subroutine ReadNextP3DVar
!*****************************************************************************
SUBROUTINE WriteNextP3DVar(nvar,rstrtfilename,rstrtdouble,nzones,npnts,w,ieng,jeng,keng)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nvar,nzones
integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:)    ,pointer :: w
logical,intent(in) :: rstrtdouble
integer, dimension(:),pointer :: ieng,jeng,keng
integer, dimension(nzones) :: igpnts, igsptr, igeptr

real(kind=SP) ,dimension(:,:),pointer :: sw
real(kind=DP) ,dimension(:,:),pointer :: dw

INTEGER :: ngrd, ng, i1, i2, igall, iengx, jengx, kengx
INTEGER :: i,ierr
logical :: tof


WRITE(tmpfile(nvar), "(A,i6.6,'.tmp')") trim(rstrtfilename),nvar

INQUIRE(file = tmpfile(nvar), EXIST=tof)
do while(tof)
  tmpfile(nvar) = trim(tmpfile(nvar)) // 't'
  INQUIRE(file = tmpfile(nvar), EXIST=tof)
enddo

WRITE(*,*)
WRITE(*,*) 'Writing Interpolated Variable ',nvar,' to Temporary File ', TRIM(ADJUSTL(tmpfile(nvar)))

ngrd = nzones

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)
   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO

open(unit=icbfdlwrite,file=tmpfile(nvar),form='unformatted',status='new')


if(rstrtdouble)then

    select case(FP)
    case(SP)
      allocate( dw(1,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      do i=1,igall
        dw(:,i) = w(:,i)
      enddo
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing flow variable ',nvar,' for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        WRITE(icbfdlwrite) (dw(1,i),i=i1,i2)
      END DO
      deallocate(dw)
    case(DP)
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing flow variable ',nvar,' for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        WRITE(icbfdlwrite) (w(1,i),i=i1,i2)
      END DO
    end select

  else

    select case(FP)
    case(SP)
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing flow variable ',nvar,' for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        WRITE(icbfdlwrite) (w(1,i),i=i1,i2)
      END DO
    case(DP)
      allocate( sw(1,sum(npnts)),stat = ierr)
      CALL CheckAllocationError(ierr)
      do i=1,sum(npnts)
        sw(1,i) = w(1,i)
      enddo
      DO ng = 1, ngrd
        WRITE(*,*) 'Writing flow variable ',nvar,' for grid ', ng
        i1 = igsptr(ng)
        i2 = igeptr(ng)
        WRITE(icbfdlwrite) (sw(1,i),i=i1,i2)
      END DO
      deallocate(sw)
    end select
  endif

close(icbfdlwrite)


end subroutine WriteNextP3DVar



!*****************************************************************************
SUBROUTINE CombineP3DVarFiles(x,rstrtfilename,rstrtformat,rstrtdouble,ngp,nzones,ieng,jeng,keng,rtype,multib)
!**************************************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen),intent(in) :: rstrtfilename
integer,intent(in) :: nzones,rtype,ngp
!integer, dimension(:),pointer :: npnts

real (kind=FP),dimension(:,:) ,pointer :: x

integer,intent(in) :: rstrtformat
logical,intent(in) :: rstrtdouble,multib
integer, dimension(:),pointer :: ieng,jeng,keng
integer, dimension(nzones) :: igpnts, igsptr, igeptr

real(kind=SP),dimension(:,:),allocatable ::  sw,sx
real(kind=DP),dimension(:,:),allocatable  :: dw,dx

INTEGER :: j, ngrd, ng, i1, i2, igall, iengx, jengx, kengx
INTEGER :: i,ierr,nvar,n,nvar1,nvar2,ifl
integer,dimension(:),allocatable :: varfileunit

allocate(varfileunit(sum(numrstrtvar)),stat=ierr)
Call CheckAllocationError(ierr)


WRITE(*,*)
WRITE(*,*) 'Combining the Temporary Variable Files into ', TRIM(ADJUSTL(rstrtfilename))

do nvar = 1,sum(numrstrtvar)
  WRITE(*,*) 'Reading variable ', nvar, ' from file', TRIM(ADJUSTL(tmpfile(nvar)))
  varfileunit(nvar) = icbfdlwrite+nvar
  open(unit=varfileunit(nvar),file=tmpfile(nvar),form='unformatted',status='old')
enddo

ngrd = nzones

igall = 0
DO ng = 1, ngrd
   iengx = ieng(ng)
   jengx = jeng(ng)
   kengx = keng(ng)
   igpnts(ng) = iengx * jengx * kengx
   igall = igall + igpnts(ng)
   SELECT CASE (ng)
      CASE (1)
         igsptr(ng) = 1
      CASE DEFAULT
         igsptr(ng) = igeptr(ng-1) + 1
   END SELECT
   igeptr(ng) = igall
END DO


Call WriteRestartHeader(rstrtfilename,rstrtformat,rstrtdouble,ngrd,ieng,jeng,keng,rtype,multib)


select case (rstrtformat)
case(0,2)
  if(rstrtdouble)then
    allocate( dw(maxval(numrstrtvar),maxval(igpnts(:))),stat = ierr)
    CALL CheckAllocationError(ierr)

    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
        allocate( dx(3,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        dx(1:3,1:igall) = x(1:3,1:igall)
      case(DP)
      end select
    end select
      

    DO ng = 1, ngrd
      i1 = 1
      i2 = igpnts(ng)
      do nvar = 1,numrstrtvar(1)
        WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
        READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar,i),i=i1,i2)
      enddo
      WRITE(*,*) 'Writing flow variables for grid ', ng
      if(rtype.eq.2) WRITE(icbfdlwrite) DQVar(min(ng,ngp),:)
      select case(rtype)
      case(fdl_1,fdl_2)
        select case(FP)
        case(SP)
          WRITE(icbfdlwrite) ((dx(n,i),i=i1,i2),n=1,3), &
                             ((dw(n,i),i=i1,i2),n=1,numrstrtvar(1))
        case(DP)
          WRITE(icbfdlwrite) (( x(n,i),i=i1,i2),n=1,3), &
                             ((dw(n,i),i=i1,i2),n=1,numrstrtvar(1))
        end select
        do j=2,nfilesect
          WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          WRITE(icbfdlwrite) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = 6,10
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar-5,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite) ((dw(n-5,i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = nvar1,nvar2
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar-nvar1+1,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite) ((dw(n-nvar1+1,i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        write(icbfdlwrite) iflag(ifl)
      end do
    end select

    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
        deallocate( dx)
      case(DP)
      end select
    end select
    deallocate(dw)

  else

    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
      case(DP)
        allocate( sx(3,igall),stat = ierr)
        CALL CheckAllocationError(ierr)
        sx(1:3,1:igall) = x(1:3,1:igall)
      end select
    end select

    allocate( sw(sum(numrstrtvar),maxval(igpnts(:))),stat = ierr)
    CALL CheckAllocationError(ierr)

    DO ng = 1, ngrd
      i1 = 1
      i2 = igpnts(ng)
      do nvar = 1,numrstrtvar(1)
        WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
        READ(varfileunit(nvar),ERR=101,END=101) (sw(nvar,i),i=i1,i2)
      enddo
      WRITE(*,*) 'Writing flow variables for grid ', ng
      if(rtype.eq.2) WRITE(icbfdlwrite) SQVar(min(ng,ngp),:)
      select case(rtype)
      case(fdl_1,fdl_2)
        select case(FP)
        case(SP)
          WRITE(icbfdlwrite) (( x(n,i),i=i1,i2),n=1,3), &
                             ((sw(n,i),i=i1,i2),n=1,numrstrtvar(1))
        case(DP)
          WRITE(icbfdlwrite) ((sx(n,i),i=i1,i2),n=1,3), &
                             ((sw(n,i),i=i1,i2),n=1,numrstrtvar(1))
        end select

        do j=2,nfilesect
          WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          WRITE(icbfdlwrite) ((sw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = 6,10
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (sw(nvar-5,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite) ((sw(n-5,i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = nvar1,nvar2
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (sw(nvar-nvar1+1,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite) ((sw(n-nvar1+1,i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        write(icbfdlwrite) iflag(ifl)
      end do
    end select

    select case(rtype)
    case(fdl_1,fdl_2)
      select case(FP)
      case(SP)
      case(DP)
        deallocate( sx)
      end select
    end select

    deallocate(sw)
  endif
case(1)
    allocate( dw(sum(numrstrtvar),maxval(igpnts(:))),stat = ierr)
    CALL CheckAllocationError(ierr)

    dw(:,:) = 0.0_DP
    DO ng = 1, ngrd
      i1 = 1
      i2 = igpnts(ng)
      do nvar = 1,numrstrtvar(1)
        WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
        READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar,i),i=i1,i2) !This file is always unformatted
      enddo
      WRITE(*,*) 'Writing flow variables for grid ', ng
      if(rtype.eq.2) WRITE(icbfdlwrite,*) DQVar(min(ng,ngp),:)
      select case(rtype)
      case(fdl_1,fdl_2)
        WRITE(icbfdlwrite,*) (( x(n,i),i=i1,i2),n=1,3), &
                             ((dw(n,i),i=i1,i2),n=1,numrstrtvar(1))
        do j=2,nfilesect
          WRITE(icbfdlwrite,*) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      case default
        do j=1,nfilesect
          WRITE(icbfdlwrite,*) ((dw(n,i),i=i1,i2),n=rstrtvarsct(j,1),rstrtvarsct(j,2))
        enddo
      end select
    END DO

    select case(rtype)
    case(fdl_2)
      ifl = 1
      write(icbfdlwrite,*) iflag(ifl)
      do while(iflag(ifl).ne.999)
        select case(iflag(ifl))
        case(1)
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = 6,10
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar-5,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite,*) ((dw(n-5,i),i=i1,i2),n=6,10)
          ENDDO
        case(41)
          select case(ifl)
          case(1)
            nvar1 = 6
            nvar2 = 8
          case(2)
            nvar1 = 11
            nvar2 = 13
          end select
          DO ng = 1,ngrd
            i1 = 1
            i2 = igpnts(ng)
            do nvar = nvar1,nvar2
              WRITE(*,*) 'Reading flow variable ',nvar,' for grid ', ng
              READ(varfileunit(nvar),ERR=101,END=101) (dw(nvar-nvar1+1,i),i=i1,i2)
            enddo
            WRITE(icbfdlwrite,*) ((dw(n-nvar1+1,i),i=i1,i2),n=nvar1,nvar2)
          ENDDO
        end select
        ifl = ifl + 1
        write(icbfdlwrite,*) iflag(ifl)
      end do
    end select

    
    deallocate(dw)
end select

do nvar = 1,sum(numrstrtvar)
  close(varfileunit(nvar),STATUS='DELETE')
enddo

close(icbfdlwrite)

deallocate(varfileunit)

return

101 continue

do nvar = 1,sum(numrstrtvar)
  close(varfileunit(nvar),STATUS='DELETE')
enddo

Call DispFileReadError()

end subroutine CombineP3DVarFiles


!********************************************************************************
! Tecplot file writing subroutines
!********************************************************************************

!********************************************************************************
SUBROUTINE WriteTecplotFile_P3D(rstrtfilename,NGRD,dx,dw,ieng,jeng,keng,iblank,iblanked)
!********************************************************************************
!USE MOD_AVUSGrid

IMPLICIT NONE

character(len=chrlen), intent(in) :: rstrtfilename
real (kind=DP),dimension(:,:),pointer :: dx,dw
integer ,dimension(:),pointer :: ieng,jeng,keng,iblank
integer, intent(in) :: NGRD
logical, intent(in) :: iblanked

INTEGER :: NG, I, J, K,lu_solfile,IP,lastip

lu_solfile = 594

write(*,*) 'Writing to Tecplot file ',trim(adjustl(rstrtfilename)),'...'

OPEN (UNIT=lu_solfile, FILE = rstrtfilename)

select case(iblanked)
case(.true.)
  WRITE(lu_solfile,*) 'VARIABLES = "X", "Y", "Z", "U", "V", "W", "P", "RHO","IBLANK"' 
case(.false.)
  WRITE(lu_solfile,*) 'VARIABLES = "X", "Y", "Z", "U", "V", "W", "P", "RHO"' 
end select
lastip = 0

do NG=1,NGRD
     
  WRITE(lu_solfile,"(A , i6.6,A,i4.4,' J=', i4.4,' K=',i4.4)") &
& 'ZONE T="ZONE-',NG,'" I=',IENG(NG),JENG(NG),KENG(NG)

if(ng.gt.1) lastip = lastip + ieng(ng-1)*jeng(ng-1)*keng(ng-1)

select case(iblanked)
case(.true.)
DO K=1,KENG(NG)
 DO J=1,JENG(NG)
   DO I=1,IENG(NG)
    IP = i+  ieng(ng)*(j-1)+ieng(ng)*jeng(ng)*(k-1)+lastip
    WRITE(lu_solfile,*) dx(:,IP),  dw(:,IP), IBLANK(IP)

   ENDDO
  ENDDO
 ENDDO
case(.false.)
DO K=1,KENG(NG)
 DO J=1,JENG(NG)
   DO I=1,IENG(NG)
    IP = i+  ieng(ng)*(j-1)+ieng(ng)*jeng(ng)*(k-1)+lastip
    WRITE(lu_solfile,*) dx(:,IP),  dw(:,IP)
   ENDDO
  ENDDO
 ENDDO
end select
ENDDO
 
END SUBROUTINE WriteTecplotFile_P3D

!********************************************************************************
! END Tecplot file writing subroutine
!********************************************************************************

END MODULE MOD_P3DGrid
