MODULE MOD_ReadPegFile

IMPLICIT NONE

INTEGER, PARAMETER :: mbc = 50  !Maximum number of boundary conditions per mesh
INTEGER, PARAMETER :: nchar = 40  ! maximum number of characters in a name
INTEGER :: numerr
INTEGER, DIMENSION(:,:), ALLOCATABLE :: ibtyp, ibdir, ibcs, ibce, jbcs, jbce, kbcs, kbce
INTEGER, DIMENSION(:), ALLOCATABLE :: nbc

CONTAINS
!
!==================================================================================================
SUBROUTINE ReadPegFile ( ngrd, idm_grid, jdm_grid, kdm_grid, pegasusfile)
!==================================================================================================
!

INTEGER, INTENT(IN) :: ngrd
CHARACTER(len=nchar), INTENT(IN) :: pegasusfile
INTEGER, DIMENSION(:), POINTER :: idm_grid, jdm_grid, kdm_grid

INTEGER :: ng, nfound
INTEGER, DIMENSION(:,:), ALLOCATABLE :: ibtyp_tmp, ibcs_tmp, ibce_tmp, jbcs_tmp, jbce_tmp, kbcs_tmp, kbce_tmp, ibdir_tmp
INTEGER :: mesh_no, n, nbcmax

CHARACTER(LEN=10000) :: abc
CHARACTER(LEN=10000) :: amesh
CHARACTER(LEN=20)    :: var1, var2
LOGICAL :: endfile, phantom
CHARACTER (LEN=nchar), DIMENSION(:), ALLOCATABLE :: name
CHARACTER (LEN=nchar) :: tmpstring

ALLOCATE(ibtyp_tmp(mbc,ngrd))
ALLOCATE(ibcs_tmp(mbc,ngrd))
ALLOCATE(ibce_tmp(mbc,ngrd))
ALLOCATE(jbcs_tmp(mbc,ngrd))
ALLOCATE(jbce_tmp(mbc,ngrd))
ALLOCATE(kbcs_tmp(mbc,ngrd))
ALLOCATE(kbce_tmp(mbc,ngrd))
ALLOCATE(ibdir_tmp(mbc,ngrd))
ALLOCATE(name(ngrd))
ALLOCATE(nbc(ngrd))

WRITE(*,*)
WRITE(*,*) 'Obtaining boundary condition information from file ', TRIM(ADJUSTL(pegasusfile))
!CALL OpenUp (pegasusfile, 17, 'FORMATTED')
OPEN(unit=17,file=pegasusfile,form='FORMATTED',STATUS='OLD',ACTION='READ')

OPEN(UNIT=18,FILE='bcinp_errlog.dat',FORM='FORMATTED',STATUS='UNKNOWN')
!
!   READ IN THE USER SPECIFIED MESH PARAMETERS
!
ng = 0
numerr = 0
nbc = 0
REWIND(UNIT=17)
var1 = '$MESH'
var2 = '$mesh'

DO 
   CALL get_namelist_group( amesh,var1,var2,endfile )
   IF( endfile ) EXIT
   phantom = .false.
   CALL readnml( amesh,'PHANTOM    ',phantom )
   if(.not.phantom) then
     ng = ng + 1
     CALL readnmc( amesh,'NAME    ',name(ng), nchar )
   endif
END DO
   
nbcmax = 0
REWIND(UNIT=17)
var1 = '$BCINP'
var2 = '$bcinp'

33  CALL get_namelist_group( abc,var1,var2,endfile )

ng = 0

IF( .NOT. endfile ) THEN

   CALL readnmc( abc,'ISPARTOF',tmpstring,nchar )
   
   !Only read meshes that are not phantom meshes
   DO ng = 1,ngrd
     IF(TRIM(name(ng)).eq.TRIM(tmpstring)) GOTO 32
   ENDDO
   GOTO 33
 
32 CONTINUE
   CALL meshnum(tmpstring, nchar, ngrd, name, mesh_no)
!   ng = mesh_no( tmpstring, nchar, ngrd, name )
   ng = mesh_no
   CALL readnmit( abc,'IBTYP   ',ibtyp_tmp, mbc, ngrd, ng, nfound )
   nbc(ng) = nfound

   CALL readnmit( abc,'IBDIR   ',ibdir_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
       WRITE(18,*)'ERROR: inconsistent number of BCs for IBDIR'
       numerr = numerr + 1
   END IF

   CALL readnmit( abc,'JBCS    ',ibcs_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for JBCS'
      numerr = numerr + 1
   END IF
   
   CALL readnmit( abc,'JBCE    ',ibce_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for JBCE'
      numerr = numerr + 1
   END IF
   
   CALL readnmit( abc,'KBCS    ',jbcs_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for KBCS'
      numerr = numerr + 1
   END IF

   CALL readnmit( abc,'KBCE    ',jbce_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for KBCE'
      numerr = numerr + 1
   END IF

   CALL readnmit( abc,'LBCS    ',kbcs_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for LBCS'
      numerr = numerr + 1
   END IF

   CALL readnmit( abc,'LBCE    ',kbce_tmp, mbc, ngrd, ng, nfound )
   IF (nfound .ne. nbc(ng)) THEN
      WRITE(18,*)'ERROR: inconsistent number of BCs for LBCE'
      numerr = numerr + 1
   END IF
!
!  Convert negative indices to positive indices
!
   DO n = 1, nbc(ng)
      IF( ibcs_tmp(n,ng).LT.0 )ibcs_tmp(n,ng) = idm_grid(ng)+ibcs_tmp(n,ng)+1 
      IF( ibce_tmp(n,ng).LT.0 )ibce_tmp(n,ng) = idm_grid(ng)+ibce_tmp(n,ng)+1 
      IF( jbcs_tmp(n,ng).LT.0 )jbcs_tmp(n,ng) = jdm_grid(ng)+jbcs_tmp(n,ng)+1 
      IF( jbce_tmp(n,ng).LT.0 )jbce_tmp(n,ng) = jdm_grid(ng)+jbce_tmp(n,ng)+1 
      IF( kbcs_tmp(n,ng).LT.0 )kbcs_tmp(n,ng) = kdm_grid(ng)+kbcs_tmp(n,ng)+1 
      IF( kbce_tmp(n,ng).LT.0 )kbce_tmp(n,ng) = kdm_grid(ng)+kbce_tmp(n,ng)+1 
   END DO
!
!  Check for valid indices and valid ibdir
!
   DO n = 1, nbc(ng)
      IF ( ibcs_tmp(n,ng).le.0 .or. ibcs_tmp(n,ng).gt.idm_grid(ng) .or. &
           jbcs_tmp(n,ng).le.0 .or. jbcs_tmp(n,ng).gt.jdm_grid(ng) .or. &
           kbcs_tmp(n,ng).le.0 .or. kbcs_tmp(n,ng).gt.kdm_grid(ng) .or. &
           ibcs_tmp(n,ng) .gt. ibce_tmp(n,ng) .or. &
           jbcs_tmp(n,ng) .gt. jbce_tmp(n,ng) .or. &
           kbcs_tmp(n,ng) .gt. kbce_tmp(n,ng) ) THEN
         numerr = numerr + 1
         WRITE(18,*) ' '
         WRITE(18,*) 'ERROR: invalid indices for mesh ',ng
         WRITE(18,*) '      ibcs,ibce = ',ibcs_tmp(n,ng),ibce_tmp(n,ng)
         WRITE(18,*) '      jbcs,jbce = ',jbcs_tmp(n,ng),jbce_tmp(n,ng)
         WRITE(18,*) '      kbcs,kbce = ',kbcs_tmp(n,ng),kbce_tmp(n,ng)
      END IF
      
      IF ( ibdir_tmp(n,ng).ne.1 .and. ibdir_tmp(n,ng).ne.-1 .and. &
           ibdir_tmp(n,ng).ne.2 .and. ibdir_tmp(n,ng).ne.-2 .and. &
           ibdir_tmp(n,ng).ne.3 .and. ibdir_tmp(n,ng).ne.-3) THEN
         numerr = numerr + 1
         WRITE(18,*) ' '
         WRITE(18,*) 'ERROR: invalid value of IBDIR = ',ibdir_tmp(n,ng),&
                  '   found in mesh ',ng
         WRITE(18,*) '   IBDIR must be one of 1,2,3,-1,-2,-3'
      END IF
   END DO
!
!  Go back and read BCs for the next mesh
!       
   IF (nbc(ng) .GT. nbcmax) THEN
       nbcmax = nbc(ng)
   END IF
   
   GOTO 33

END IF

IF (numerr .NE. 0) THEN
   WRITE(*,*) 'Errors detected while reading in data from PEGASUS input file.'
   WRITE(*,*) 'See file bcinp_errorlog.dat for details...Stopping execution'
   STOP
ELSE
   CLOSE(UNIT=18,STATUS='DELETE')  !XXXX CHECK FORM OF THIS STATEMENT
END IF
!
! Allocate real bc arrays and move data from tmp arrays
!
ALLOCATE(ibcs(nbcmax,ngrd))
ALLOCATE(ibce(nbcmax,ngrd))
ALLOCATE(jbcs(nbcmax,ngrd))
ALLOCATE(jbce(nbcmax,ngrd))
ALLOCATE(kbcs(nbcmax,ngrd))
ALLOCATE(kbce(nbcmax,ngrd))
ALLOCATE(ibtyp(nbcmax,ngrd))
ALLOCATE(ibdir(nbcmax,ngrd))

ibcs(1:nbcmax,:) = ibcs_tmp(1:nbcmax,:)
ibce(1:nbcmax,:) = ibce_tmp(1:nbcmax,:)
jbcs(1:nbcmax,:) = jbcs_tmp(1:nbcmax,:)
jbce(1:nbcmax,:) = jbce_tmp(1:nbcmax,:)
kbcs(1:nbcmax,:) = kbcs_tmp(1:nbcmax,:)
kbce(1:nbcmax,:) = kbce_tmp(1:nbcmax,:)
ibtyp(1:nbcmax,:) = ibtyp_tmp(1:nbcmax,:)
ibdir(1:nbcmax,:) = ibdir_tmp(1:nbcmax,:)

DEALLOCATE(ibcs_tmp)
DEALLOCATE(ibce_tmp)
DEALLOCATE(jbcs_tmp)
DEALLOCATE(jbce_tmp)
DEALLOCATE(kbcs_tmp)
DEALLOCATE(kbce_tmp)
DEALLOCATE(ibtyp_tmp)
DEALLOCATE(ibdir_tmp)
DEALLOCATE(name)

CLOSE(17)

END SUBROUTINE ReadPegFile
!
!==================================================================================================
!==================================================================================================
!
! All the procedures below were taken directly from PEGASUS5
!
!==================================================================================================

SUBROUTINE get_namelist_group( string, test1, test2, endfile )

!***********************************************************************
!
!
!***********************************************************************
  IMPLICIT NONE


    CHARACTER(LEN=10000) :: string
    CHARACTER(LEN=132)   :: ain
    CHARACTER(LEN=20)    :: test1, test2
    INTEGER            :: ich, istart, iend
    LOGICAL              :: endfile

  string = ' '
  endfile = .FALSE.

10 READ (17,99,END=100) ain
99 FORMAT (A132)

  CALL rmspcs(ain,ich)

  IF (INDEX (ain,TRIM(test1)) .NE. 0 .OR. &
      INDEX (ain,TRIM(test2)) .NE. 0) THEN

    istart      = 1
    iend        = ich

20 CONTINUE

   string(istart:iend) = ain(1:ich)

   IF (INDEX (ain,'$END') .NE. 0 .OR.             &
       INDEX (ain,'$end') .NE. 0) GOTO 21

   READ (17,99,END=21) ain

   CALL rmspcs(ain,ich)
   istart      = iend + 1
   iend        = istart + ich - 1

   IF( iend.GT.10000 ) THEN

     numerr      = numerr + 1
     WRITE(18,902)TRIM(test1)
902  FORMAT(/,'***ERROR*** ',A20,' INPUT GREATER',    &
              '            THAN 10000 CHARACTERS')
     iend        = 10000

    END IF

    GOTO 20

  ENDIF
  GO TO 10

21 CONTINUE

!***********************************************************************

  RETURN

100 endfile = .TRUE.

  RETURN
END SUBROUTINE get_namelist_group


      SUBROUTINE RMSPCS(ALINE,K)

!***********************************************************************
!
!     REMOVES THE SPACES AND TABS FROM NAMELIST LIKE USER INPUT.
!
!***********************************************************************

      IMPLICIT NONE

      CHARACTER*132  ALINE,    ALIN1

      CHARACTER(LEN=1) ::     TAB
      LOGICAL       FLAG
      INTEGER :: j, k

!***********************************************************************

      TAB         = CHAR(9)
      K           = 0
      FLAG        = .FALSE.

      ALIN1       = ' '


      DO J = 1,132

        IF(ALINE(J:J).EQ.''''.AND.FLAG) THEN

!   ENDING ' FOUND

            K           = K + 1
            ALIN1(K:K)  = ALINE(J:J)
            FLAG        = .FALSE.

        ELSE IF(ALINE(J:J).EQ.'''') THEN

!   STARTING ' FOUND

            K           = K + 1
            ALIN1(K:K)  = ALINE(J:J)
            FLAG        = .TRUE.

        ELSE IF((ALINE(J:J).EQ.' '.OR.ALINE(J:J).EQ.TAB) &
     &          .AND.(.NOT.FLAG)) THEN

!   BLANK FOUND AND REMOVED

            CONTINUE

        ELSE IF((ALINE(J:J).EQ.' '.OR.ALINE(J:J).EQ.TAB) &
     &          .AND.FLAG) THEN

!   BLANK FOUND AND NOT REMOVED

            K           = K + 1
            ALIN1(K:K)  = ALINE(J:J)

        ELSE

!   OTHER CHARACTER FOUND AND NOT REMOVED

            K           = K + 1
            ALIN1(K:K)  = ALINE(J:J)

        ENDIF

      ENDDO

      ALINE       = ALIN1

      RETURN
      END SUBROUTINE RMSPCS

SUBROUTINE check_input(nmlist, num_keywords, keyword )

!***********************************************************************
!
!     Check the keywords for the user inputs.
!
!***********************************************************************

  IMPLICIT NONE

!***********************************************************************

  CHARACTER(LEN=*)              :: nmlist
  INTEGER                     :: num_keywords
  CHARACTER(LEN=*),DIMENSION(*) :: keyword

!***********************************************************************

  CHARACTER(LEN=100) :: sym

  INTEGER          :: intot, istart, iend, i, n, inc, key_len

!***********************************************************************

      intot       = LEN_TRIM( nmlist )

      istart      = 1

50 DO i = istart,intot

!   Find an '=' Sign

     IF( nmlist(i:i).EQ.'=' ) THEN
         GO TO 100
     END IF

   ENDDO

   RETURN

100 iend        = i

!   LOOP FOR EACH KEY WORD

    DO n = 1,num_keywords

      key_len     = LEN_TRIM(keyword(n))
      sym         = keyword(n)
      inc         = 0

      istart      = iend - key_len
      IF( istart.LT.1 ) istart      = 1

      inc         = INDEX(nmlist(istart:iend),sym(1:key_len))

      IF( inc.NE.0 ) THEN
          GO TO 120
      END IF

    ENDDO

    numerr      = numerr + 1

    istart = iend - 8
    WRITE(18,900)nmlist(istart:iend+1)
900 FORMAT(/,'***ERROR*** Bad Keyword:',A9)

120 istart      = iend + 1

    IF( iend.LT.intot ) THEN
        GO TO 50
    END IF

    RETURN

END SUBROUTINE check_input

      SUBROUTINE READNMIT(NMLIST,SYMVAR,IVARS,IDIM,JDIM,MESHNO,NVAL)

!***********************************************************************
!
!     NAMELIST LIKE READ FOR VARIABLE LENGTH  INTEGER
!     ARRAY.
!
!     READS IN A CHARACTER STRING CALLED NMLIST WHICH IS 10000
!     CHARACTERS LONG AND SEARCHS THIS ARRAY FOR UP TO IDIM NUMBER
!     OF INTEGERS. THE ACTUAL NUMBER OF INTEGERS FOUND IS
!     RETURNED IN NVAL.  IF NVAL = -1 THEN THE ARRAY IVAR WILL BE
!     ASSUMED TO CONTAIN ALL POSSIBLE VALUES.
!
!     Inputs:
!       NMLIST - string used as buffer to be parsed
!       SYMVAR - name of $ var defining namelist group
!       IDIM,JDIM   - dimensions of IVAR
!       MESHNO - element of JDIM we are at - so we can fill IVAR
!                properly
!
!     Outputs:
!       IVAR   - array of found values
!       NVAL   - number of integers found and placed in IVAR   
!***********************************************************************
!
! NOTE: Changed local IS and IE arrays from original to ISX and IEX to prevent conflict with
!  use-associated arrays is and ie
!

      IMPLICIT NONE
 
      INTEGER :: idim, nval, jdim, meshno

      CHARACTER*10000 NMLIST
      CHARACTER*8    SYMVAR
      CHARACTER*15   TEMPA
      INTEGER :: ivars(idim,jdim)

      CHARACTER*10000 NMTEMP

      INTEGER :: intot, i, len, iprime, iprim2 
      INTEGER :: inc, istart, iend, id, ic, nv, jj, ii, n1

      INTEGER :: isX(idim),ieX(idim)

!***********************************************************************

      INTOT       = 10000
      NVAL        = 0
      ISTART      = 0

!***********************************************************************


!     DETERMINE THE ACTUAL LENGTH OF THE SYMVAR ARRAY
!     AND SET THE VARIABLE LEN EQUAL TO THIS LENGTH

      len         = 0

      DO i = 1,8

        IF(symvar(i:i).NE.' ') THEN
            len         = len + 1
        ENDIF

      ENDDO


!     DETERMINE THE POSITION OF SYMVAR IN THE NMLIST ARRAY

      nmtemp      = nmlist

    4 CONTINUE

      iprime      = INDEX(nmtemp,'''')
      IF( iprime.NE.0 ) THEN
          iprim2      = INDEX(nmtemp(iprime+1:),'''')
          IF( iprim2.NE.0 ) THEN
              iprim2      = iprim2 + iprime
              nmtemp(iprime:iprim2) = ' '
          ELSE
              nmtemp(iprime:)       = ' '
          ENDIF
          GOTO 4
      ENDIF


      inc         = INDEX(nmtemp,symvar(1:len))

      IF( inc.NE.0 ) THEN

!     SEARCH FOR THE EQUAL SIGN. PRINT ERROR MESSAGE
!     AND RETURN TO THE MAIN CODE IF IT IS NOT FOUND.

          i           = inc + len

          IF( nmlist(i:i).NE.'=' )THEN

              numerr      = numerr + 1

              WRITE(*,901) symvar
              WRITE(*,902)
  901         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  902         FORMAT(  '            NO ''='' SIGN FOUND')
              WRITE(*,916) (nmlist(jj:jj),jj=inc,i)
 916          FORMAT(5X,75A1)



              RETURN

          END IF

!     IF FIRST CHARACTER AFTER '=' SIGN IS A COMMA, SET
!     NVAL = -1 AND RETURN.

          IF( nmlist(i+1:i+1).EQ.',' ) THEN

              nval        = -1
              RETURN

          ENDIF

!     Search for a case where we cannot write the string
!     to an integer - that signals we are at a character,
!     and are done with this variable.

          NV = 1
          
 12       ISTART = I + 1

          ISX(NV)      = ISTART
          DO 5 I = ISTART,INTOT

             IF( NMLIST(I:I).EQ.',' )THEN
                IEND = I - 1
                GO TO 9
             ELSE IF( NMLIST(I:I).EQ.'=' ) THEN
                GO TO 6
             ELSE IF( NMLIST(I:I).EQ.'-' ) THEN
                GO TO 5
             ELSE
                READ(NMLIST(I:I),400,ERR=500)N1
                GO TO 5
 400            FORMAT(I1)
!
!  Read a string, so we're through with this variable
!
 500            IEND = I - 1
                IEX(NV)      = IEND
                NVAL        = NV - 1
                GO TO 10
             END IF
              
 5        CONTINUE

 6        NVAL = 0

          RETURN

 9        IEX(NV)      = IEND
          NVAL        = NV 
          NV          = NV + 1
          GO TO 12

 10       CONTINUE

!     FEED THE CHARCTER STRING FROM NMLIST INTO
!     THE VAR ARRAY.

!     BLANKING OUT THE TEMPA ARRAY

          DO 20 NV = 1,NVAL

            DO 17 II = 1,15
              TEMPA(II:II) = ' '
   17       CONTINUE

            ID          = 0

            DO 18 IC = ISX(NV),IEX(NV)

              ID           = ID + 1
              TEMPA(ID:ID) = NMLIST(IC:IC)

   18       CONTINUE

            READ(TEMPA,2000) IVARS(NV,MESHNO)
 2000       FORMAT(I15)

   20     CONTINUE

      ELSE

!     SYMVAR NOT FOUND IN THE NMLIST ARRAY.

      ENDIF


      RETURN
      END SUBROUTINE READNMIT

!*************************************************************************

  SUBROUTINE meshnum( mesh, nchar, ngrd, name, mesh_no )

!    USE mod_parameters
!    USE mod_mem_alloc

    IMPLICIT NONE

    LOGICAL                           :: flag

    INTEGER                         :: ng, nchar, ngrd
    INTEGER                         :: mesh_no

    CHARACTER(LEN=nchar)              :: mesh
    CHARACTER(LEN=nchar), DIMENSION(*) :: name
    
    DO ng = 1,ngrd

       flag       = .TRUE.

       IF( name(ng) .NE. mesh )THEN
          flag       = .FALSE.
       END IF

       IF( flag )THEN
          mesh_no     = ng
          RETURN
       END IF

    ENDDO

    mesh_no     = 0

    RETURN

  END SUBROUTINE meshnum

!*************************************************************************
      SUBROUTINE READNMC(NMLIST,SYMVAR,IVAR,ICDIM)

!***********************************************************************
!
!     NAMELIST LIKE READ FOR A SINGLE VARIABLE NAME CHARACTER ARRAY.
!
!     READS IN A CHARACTER STRING CALLED NMLIST WHICH IS 2000
!     CHARACTERS LONG AND SEARCHS THIS ARRAY FOR A CHARACTER STRING
!     ICDIM LONG.
!
!***********************************************************************

!      USE mod_parameters
!      USE mod_error

      IMPLICIT NONE
!

      INTEGER :: icdim

      CHARACTER*10000 NMLIST
      CHARACTER*8     SYMVAR
      CHARACTER(LEN=icdim) :: IVAR
      CHARACTER(LEN=icdim) :: blank_string
      CHARACTER*10000 NMTEMP

      INTEGER :: intot, i, len, iprime, iprim2 
      INTEGER :: inc, istart, iend, id, ic

!***********************************************************************

      INTOT       = 10000
      blank_string = " "
!***********************************************************************


!     DETERMINE THE ACTUAL LENGTH OF THE SYMVAR ARRAY
!     AND SET THE VARIABLE LEN EQUAL TO THIS LENGTH

      LEN         = 0

      DO I = 1,8

        IF(SYMVAR(I:I).NE.' ') THEN
            LEN         = LEN + 1
        ENDIF

      ENDDO

!     DETERMINE THE POSITION OF SYMVAR IN THE NMLIST ARRAY

      NMTEMP      = NMLIST

    4 CONTINUE

      IPRIME      = INDEX(NMTEMP,'''')
      IF( IPRIME.NE.0 ) THEN
          IPRIM2      = INDEX(NMTEMP(IPRIME+1:),'''')
          IF( IPRIM2.NE.0 ) THEN
              IPRIM2      = IPRIM2 + IPRIME
              NMTEMP(IPRIME:IPRIM2) = ' '
          ELSE
              NMTEMP(IPRIME:)       = ' '
          ENDIF
          GOTO 4
      ENDIF


      INC         = INDEX(NMTEMP,SYMVAR(1:LEN))

      IF( INC.NE.0 ) THEN

!     SEARCH FOR THE EQUAL SIGN. PRINT ERROR MESSAGE
!     AND RETURN TO THE MAIN CODE IF IT IS NOT FOUND.

          I           = INC + LEN

          IF( NMLIST(I:I).NE.'=' )THEN

              NUMERR      = NUMERR + 1

              WRITE(*,901) SYMVAR
              WRITE(*,902)
  901         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  902         FORMAT(  '            NO ''='' SIGN FOUND')
              WRITE(*,916) (NMLIST(I:I),I=INC,INC+ICDIM+3)

              RETURN

          END IF

!     CHECK FOR OPENING QUOTATION MARK OF THE STRING.
!     RETURN TO THE MAIN CODE IF IT IS NOT FOUND.

          I           = I + 1

          IF( NMLIST(I:I).NE.'''' )THEN

              NUMERR      = NUMERR + 1

              WRITE(*,903) SYMVAR
              WRITE(*,904)
  903         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  904         FORMAT(  '            NO OPENING QUOTATION MARK FOUND')
              WRITE(*,916) (NMLIST(I:I),I=INC,INC+ICDIM+3)

              RETURN

          END IF

          istart      = i + 1

!     CHECK FOR CLOSING QUOTATION MARK OF THE STRING.
!     RETURN TO THE MAIN CODE IF IT IS NOT FOUND.

          DO i = istart+1,istart+icdim


            IF( nmlist(i:i).EQ.'''' )THEN

              iend      = i - 1
              GO TO 300

            END IF

          ENDDO


          NUMERR      = NUMERR + 1

          WRITE(*,905) SYMVAR
          WRITE(*,906)
  905     FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  906     FORMAT(  '            NO CLOSING QUOTATION MARK FOUND')
          WRITE(*,916) (NMLIST(I:I),I=INC,INC+ICDIM+3)
  916     FORMAT(5X,75A1)
          RETURN


!     FEED THE CHARCTER STRING FROM NMLIST INTO
!     THE IVAR ARRAY.

  300       id          = 0

            ivar = blank_string

            DO ic = istart,iend

              id          = id + 1
              ivar(id:id) = nmlist(ic:ic)

            ENDDO

      ELSE

!     SYMVAR NOT FOUND IN THE NMLIST ARRAY.

      ENDIF

      RETURN
      END SUBROUTINE READNMC

!*************************************************************************
      SUBROUTINE READNML(NMLIST,SYMVAR,LVAR)

!***********************************************************************
!
!     NAMELIST LIKE READ FOR A LOGICAL VALUE.
!
!     READS IN A CHARACTER STRING CALLED NMLIST WHICH IS 2000
!     CHARACTERS LONG AND SEARCHS THIS ARRAY FOR A LOGICAL VARIABLE
!     WHICH IS RETURNED IN LVAR.
!
!***********************************************************************

!      USE mod_error

      IMPLICIT NONE


      CHARACTER*10000 NMLIST
      CHARACTER*8    SYMVAR
      LOGICAL        LVAR

      CHARACTER*10000 NMTEMP

      INTEGER :: intot, i, len, iprime, iprim2 
      INTEGER :: inc

!***********************************************************************

      INTOT       = 10000

!***********************************************************************

!     DETERMINE THE ACTUAL LENGTH OF THE SYMVAR ARRAY
!     AND SET THE VARIABLE LEN EQUAL TO THIS LENGTH

      LEN        = 0

      DO 3 I = 1,8

        IF(SYMVAR(I:I).NE.' ') THEN
            LEN         = LEN + 1
        ENDIF

    3 CONTINUE


!     DETERMINE THE POSITION OF SYMVAR IN THE NMLIST ARRAY

      NMTEMP      = NMLIST

    4 CONTINUE

      IPRIME      = INDEX(NMTEMP,'''')
      IF( IPRIME.NE.0 ) THEN
          IPRIM2      = INDEX(NMTEMP(IPRIME+1:),'''')
          IF( IPRIM2.NE.0 ) THEN
              IPRIM2      = IPRIM2 + IPRIME
              NMTEMP(IPRIME:IPRIM2) = ' '
          ELSE
              NMTEMP(IPRIME:)       = ' '
          ENDIF
          GOTO 4
      ENDIF


      INC         = INDEX(NMTEMP,SYMVAR(1:LEN))

      IF( INC.NE.0 ) THEN

!     SEARCH FOR THE EQUAL SIGN. PRINT ERROR MESSAGE
!     AND RETURN TO THE MAIN CODE IF IT IS NOT FOUND.

          I           = INC + LEN

          IF( NMLIST(I:I).NE.'=' )THEN

              NUMERR      = NUMERR + 1

              WRITE(*,901) SYMVAR
              WRITE(*,902)
  901         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  902         FORMAT(  '            NO ''='' SIGN FOUND')

              RETURN

          END IF


!     CHECK TO SEE IF THE FIRST NON-BLANK CHARACTER
!     IS THE OPENING PERIOD OF THE STRING. RETURN TO
!     THE MAIN CODE IF IT IS NOT FOUND.

          I           = I + 1

          IF( NMLIST(I:I).NE.'.' )THEN

              NUMERR      = NUMERR + 1

              WRITE(*,903) SYMVAR
              WRITE(*,904)
  903         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  904         FORMAT(  '            NO OPENING ''.'' FOUND')

              RETURN

          END IF


!     CHECK TO SEE IF THE CHARACTER AFTER THE
!     OPENING PERIOD IS A 'T' OR AN 'F'. IF
!     IT IS A 'T' SET LVAR =.TRUE. IF IT IS
!     A 'F' SET LVAR =.FALSE. PRINT AN ERROR
!     MESSAGE OTHERWISE.

          I           = I + 1

          IF( NMLIST(I:I).EQ.'T' .OR. NMLIST(I:I).EQ.'t' ) THEN

              LVAR        = .TRUE.

          ELSE IF( NMLIST(I:I).EQ.'F' .OR. NMLIST(I:I).EQ.'f' ) THEN

              LVAR        = .FALSE.

          ELSE

              NUMERR      = NUMERR + 1

              WRITE(*,905) SYMVAR
              WRITE(*,906)
  905         FORMAT(/,'***ERROR*** FOR INPUT - ',A8)
  906         FORMAT(  '            ILLEGAL VALUE FOUND')

              RETURN


          ENDIF

      ELSE

!     SYMVAR NOT FOUND IN THE NMLIST ARRAY.

      ENDIF


      RETURN
      END SUBROUTINE READNML
      
END MODULE MOD_ReadPegFile
