!#######################################################################
SUBROUTINE Replication()
!#######################################################################
  USE common_area, ONLY: runmode, rep, dtheta, offsetx, offsety, offsetz
  IMPLICIT NONE

  INTEGER i

  rep = 0.
  SELECT CASE(runmode)
    CASE(-1,1,2,11)
      rep(1) = COS(dtheta)
      rep(2) = SIN(dtheta)
      rep(5) = -SIN(dtheta)
      rep(6) = COS(dtheta)
      rep(11) = 1.
      rep(16) = 1.
      rep(13) = offsetx
      rep(14) = offsety
      rep(15) = offsetz
    CASE DEFAULT
      rep(1) = 1.
      rep(6) = 1.
      rep(11) = 1.
      rep(16) = 1.
      rep(13) = offsetx
      rep(14) = offsety
      rep(15) = offsetz
  END SELECT

  RETURN

END SUBROUTINE Replication

!#######################################################################
SUBROUTINE GetNBlocks()
!#######################################################################
  USE common_area, ONLY: maxblk, ngrid, pgr, runmode
  IMPLICIT NONE

  INTEGER i

  SELECT CASE(runmode)
    CASE(-1,1,2,3,6,11)
      maxblk = pgr(1)%nblks
      DO i=2,ngrid
        maxblk = MAX(maxblk,pgr(i)%nblks) ! For now...
      ENDDO
    CASE(4,5)
      maxblk = 0
  END SELECT

END SUBROUTINE GetNBlocks

!#######################################################################
SUBROUTINE SetIopt()
!#######################################################################
  USE common_area, ONLY: istat, iopt, trot, ngrid, pgr, runmode, ntime
  IMPLICIT NONE

  INTEGER i,m
  
  istat = 1
# ifdef PV3SERIAL
    iopt = 0
# else
    iopt = 1   ! pV3 must have unsteady to transfer key data
# endif
  IF(ntime.GT.1)iopt=1
  SELECT CASE(runmode)
    CASE(-1,1,2,11)
      IF(trot)THEN
        iopt=2
        DO i=1,ngrid
          ALLOCATE(pgr(i)%t(pgr(i)%tpts))
          DO m=1,pgr(i)%tpts
            pgr(i)%t(m) = pgr(i)%t0(m)
          ENDDO
        ENDDO
      ELSE
        DO i = 1, ngrid
          pgr(i)%t => pgr(i)%t0
        ENDDO
      ENDIF
  END SELECT

END SUBROUTINE SetIopt

!#######################################################################
SUBROUTINE SetFLims()
!#######################################################################
  USE keydata, ONLY: nkeys, flims, myfps
  IMPLICIT NONE

  INTEGER i, ctr

  INTEGER, PARAMETER :: nscalkeys = 18
  INTEGER, PARAMETER :: nvectkeys = 1
  INTEGER, PARAMETER :: nsscalkeys = 1
  INTEGER, PARAMETER :: nsvectkeys = 0
  INTEGER, PARAMETER :: nthrekeys = 1
  myfps = 99999

  ctr = 0
  DO i = 1, nscalkeys, 1
    flims(1,i+ctr) = 0.
    flims(2,i+ctr) = 0.
  ENDDO
  ctr = ctr + nscalkeys
  DO i = 1, nvectkeys, 1
    flims(1,i+ctr) = 0.01
  ENDDO
  ctr = ctr + nvectkeys
  DO i = 1, nsscalkeys, 1
    flims(1,i+ctr) = 0.
    flims(2,i+ctr) = 1.
  ENDDO
  ctr = ctr + nsscalkeys
  DO i = 1, nsvectkeys, 1
    flims(1,i+ctr) = 0.01
  ENDDO
  ctr = ctr + nsvectkeys
  DO i = 1, nthrekeys, 1
    flims(1,i+ctr) = 0.
    flims(2,i+ctr) = 0.
  ENDDO

END SUBROUTINE SetFLims

!#######################################################################
SUBROUTINE SetNames()
!#######################################################################
  USE general, ONLY: int_to_string
  USE common_area, ONLY: cname, cid, runmode
  USE solver_turbo, ONLY: bladerow, bladepass

  SELECT CASE(runmode)
    CASE(-1,1,11)
      IF(bladerow.NE.0 .AND. bladepass.NE.0)THEN
        WRITE(cname,*) 'R' // TRIM(ADJUSTL(int_to_string(bladerow))) // 'B' // TRIM(ADJUSTL(int_to_string(bladepass)))
      ELSE
        WRITE(cname,*) 'Client ' // TRIM(ADJUSTL(int_to_string(cid)))
      ENDIF

    CASE(2)
      WRITE(cname,*) 'R' // TRIM(ADJUSTL(int_to_string(bladerow))) // 'B' // TRIM(ADJUSTL(int_to_string(bladepass)))
  END SELECT
  
END SUBROUTINE SetNames

!#######################################################################
SUBROUTINE Set_FX_Arrays()
!#######################################################################
  USE mod_clear, ONLY: ClearVar
  USE common_area, ONLY: ntime, vcstor, sfstor, sastor, blstor, rtstor
  IMPLICIT NONE

  INTEGER :: i

  ALLOCATE(vcstor(ntime), sfstor(ntime), sastor(ntime), &
           blstor(ntime), rtstor(ntime))
  DO i = 1, ntime
    vcstor(i)%nseg = -1
    NULLIFY(vcstor(i)%segs)
    NULLIFY(vcstor(i)%vcoor)
    NULLIFY(vcstor(i)%stren)
    sastor(i)%calc = 0
    NULLIFY(sastor(i)%sepind)
    NULLIFY(sastor(i)%attind)
    NULLIFY(sastor(i)%sepxyz)
    NULLIFY(sastor(i)%attxyz)
    sfstor(i)%npts = -1
    NULLIFY(sfstor(i)%pts)
    blstor(i)%nblpts = -1
    blstor(i)%nbltris = -1
    NULLIFY(blstor(i)%blxyz)
    NULLIFY(blstor(i)%bld)
    NULLIFY(blstor(i)%bltris)
    NULLIFY(blstor(i)%blcell)
    rtstor(i)%nrtpt = -1
    rtstor(i)%nrttri = -1
    NULLIFY(rtstor(i)%rtxyz)
    NULLIFY(rtstor(i)%rttri)
    NULLIFY(rtstor(i)%rtcell)
  ENDDO

END SUBROUTINE Set_FX_Arrays
