!#######################################################################
SUBROUTINE V3Cell(cel1, cel2, cel3, cel4, nptet, ptet)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(OUT) :: cel1(4,*), cel2(5,*), cel3(6,*), cel4(8,*)
  INTEGER, INTENT(OUT) :: nptet(8,*), ptet(*)

  CALL pVCell(cel1, cel2, cel3, cel4, nptet, ptet)

END SUBROUTINE V3Cell

!#######################################################################
SUBROUTINE V3Grid(xyz)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(OUT) :: xyz(3,*)

  CALL pVGrid(xyz)

END SUBROUTINE V3Grid

!#######################################################################
SUBROUTINE V3Scal(jkey,s)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: jkey
  REAL, INTENT(OUT) :: s(*)

  CALL pVScal(jkey,s)

END SUBROUTINE V3Scal

!#######################################################################
SUBROUTINE V3Vect(jkey,v)
!#######################################################################
  IMPLICIT NONE

  INTEGER, INTENT(IN) :: jkey
  REAL, INTENT(OUT) :: v(*)

  CALL pVVect(jkey,v)

END SUBROUTINE V3Vect

!#######################################################################
SUBROUTINE V3ZPrime(flag, xyz, nnode, zp, zprime, xpc, ypc, halfw)
!#######################################################################
  IMPLICIT NONE

  LOGICAL flag
  REAL, INTENT(IN) :: xyz(3,*), zprime
  REAL, INTENT(OUT) :: zp(*), xpc, ypc, halfw
  INTEGER, INTENT(IN) :: nnode

  IF(flag)THEN
    CALL pVZPrime(1, xyz, nnode, zp, zprime, xpc, ypc, halfw)
  ELSE
    CALL pVZPrime(-1, xyz, nnode, zp, zprime, xpc, ypc, halfw)
  ENDIF

END SUBROUTINE V3ZPrime

!#######################################################################
SUBROUTINE V3XYPrime(zprime, kn, xyz, n, xyp)
!#######################################################################
  IMPLICIT NONE

  REAL, INTENT(IN) :: zprime, xyz(3,*)
  REAL, INTENT(OUT) :: xyp(2,*)
  INTEGER, INTENT(IN) :: kn(*), n

  CALL pVXYPrime(zprime, kn, xyz, n, xyp)

END SUBROUTINE V3XYPrime

!#######################################################################
SUBROUTINE V3Surface(nsurf, surf, scel, tsurf)
!#######################################################################
  USE common_area, ONLY: gksurf, gsurf
  IMPLICIT NONE

  INTEGER, INTENT(INOUT) :: nsurf(2,*), surf(*), scel(4,*)
  CHARACTER*20, INTENT(INOUT) :: tsurf(*)

  INTEGER, ALLOCATABLE, DIMENSION(:,:) :: pnsurf
  INTEGER, ALLOCATABLE, DIMENSION(:) :: psurf
  INTEGER i

  ALLOCATE(pnsurf(3,gksurf),psurf(gsurf))
  CALL pVSurface(pnsurf, psurf, scel, tsurf)

  DO i = 1, gksurf
    nsurf(1,i) = pnsurf(1,i)
    nsurf(2,i) = pnsurf(2,i)
  ENDDO

  DEALLOCATE(pnsurf,psurf)

END SUBROUTINE V3Surface

!#######################################################################
SUBROUTINE V3Blank(iblank)
!#######################################################################
  USE common_area, ONLY: pgr, igrid
  IMPLICIT NONE

  INTEGER, INTENT(INOUT) :: iblank(*)
  INTEGER, POINTER, DIMENSION(:) :: tbcon

  ALLOCATE(tbcon(pgr(igrid)%tpts))
  CALL pVBlank(iblank, tbcon)
  DEALLOCATE(tbcon)
  NULLIFY(tbcon)

END SUBROUTINE V3Blank

!#######################################################################
SUBROUTINE V3Surf(isurf, xpc, ypc, halfw)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: isurf
  REAL,INTENT(OUT) :: xpc, ypc, halfw

  CALL pVSurf(isurf, xpc, ypc, halfw)

END SUBROUTINE V3Surf

!#######################################################################
SUBROUTINE V3XYSurf(kn, xyz, n, xyp)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: n
  INTEGER,INTENT(IN) :: kn(n)
  REAL,INTENT(IN) :: xyz(3,*)
  REAL,INTENT(OUT) :: xyp(2,n)

  CALL pVXYSurf(kn, xyz, n, xyp)

END SUBROUTINE V3XYSurf

!#######################################################################
SUBROUTINE V3SSurf(jkey, kn, xyz, n, s)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: jkey, n
  INTEGER,INTENT(IN) :: kn(n)
  REAL,INTENT(IN) :: xyz(3,*)
  REAL,INTENT(OUT) :: s(n)

  CALL pVSSurf(jkey, kn, xyz, n, s)

END SUBROUTINE V3SSurf

!#######################################################################
SUBROUTINE V3VSurf(jkey, kn, xyz, n, v)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: jkey, n
  INTEGER,INTENT(IN) :: kn(n)
  REAL,INTENT(IN) :: xyz(3,*)
  REAL,INTENT(OUT) :: v(3,n)

  CALL pVVSurf(jkey, kn, xyz, n, v)

END SUBROUTINE V3VSurf

!#######################################################################
SUBROUTINE V3Draw3D(istat)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: istat

  CALL pVDraw3D(istat)

END SUBROUTINE V3Draw3D

!#######################################################################
SUBROUTINE V3Draw2D(istat)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(IN) :: istat

  CALL pVDraw2D(istat)

END SUBROUTINE V3Draw2D

!#######################################################################
SUBROUTINE V3Equiv(listeq)
!#######################################################################
  IMPLICIT NONE

  INTEGER,INTENT(INOUT) :: listeq(2,*)

  CALL pVEquiv(listeq)

END SUBROUTINE V3Equiv
