
/* ~~~~~~~~~~~~~~~~~~~~~~~~ Pre-Processor ~~~~~~~~~~~~~~~~~~~~~~*/

# include <stdio.h>
# include <stdlib.h>
# include <unistd.h>
# include <sys/select.h>
# include <bluetooth/bluetooth.h>
# include "wiimote.h"
# include "wiimote_internal.h"
# include <math.h>
# include <ctype.h>
# include <string.h>
# include <X11/Xlib.h>
# include <time.h>
# include "wiiv3.h"

#define MOVEMENT 256
#define TRUE 1
#define FALSE 0

#ifdef WIN32
#  define PREFIX __stdcall
#else
#  define PREFIX
#  ifdef UNDERSCORE
#    define FWII_INTERFACE fwii_interface_
#  else
#    define FWII_INTERFACE fwii_interface
#  endif
#endif

wiimote_t *wiimote = NULL;
struct acc acc_zero, acc_one;
struct wiimote_ir_mesg ir_data;
struct acc nc_acc_zero, nc_acc_one;
struct stick nc_stick;
struct WiiMotion remote, nunchuk;
struct WIIStatus WV3Status;

extern Display *XFtnDpy;
int DspHeight,DspWidth;
Window root;
extern int XTestFakeButtonEvent(Display *display, unsigned int button, Bool is_press, unsigned long delay);


/* Wiimote Message Handler Functions */
wiimote_mesg_callback_t wiimote_callback;
void WM_btn(struct wiimote_btn_mesg *);
void WM_acc(struct wiimote_acc_mesg *);
void WM_ir(struct wiimote_ir_mesg *);
void WM_nunchuk(struct wiimote_nunchuk_mesg *);


/* Main interface to pTV3/Fortran */

void PREFIX
FWII_INTERFACE (int *sw)
{
  int ierr = 0;

  switch (*sw) {
     case WIIMOTEDISCONNECT:
       printf("Disconnecting from Wiimote...\n");
       ierr = WMDisconnect();
       break;
     case WIIMOTECONNECT:
       printf("Connecting to Wiimote...\n");
       ierr = WMConnect();
       break;
     default:
       break;
  }

}

WV3Status_t WMGetState(){
      
      struct WIIStatus WV3Pass;
      
      int icnt, increment;

      double x, y, difftyme;

      WV3Pass.mode = WV3Status.mode;

      WV3Status.tyme[1] = WV3Status.tyme[0];
      WV3Status.tyme[0] = clock();
      difftyme = ((double)(WV3Status.tyme[0]-WV3Status.tyme[1]))/CLOCKS_PER_SEC;


      switch (WV3Status.mode) {

         case WIIMOTE_BTN:
            break;
         case WIIMOTE_ACC_REL:
            increment = (difftyme/sensitivity)*DIALSTEPS;

            //printf("(WM/NC) pitch: %f %f roll: %f %f yaw: %f %f inc: %d \n",remote.pitch[0]*(180/PI),nunchuk.pitch[0]*(180/PI),remote.roll[0]*(180/PI),nunchuk.roll[0]*(180/PI),remote.yaw[0]*(180/PI),nunchuk.yaw[0]*(180/PI),increment);
            
            if (WV3Status.nc_buttons&0x01){
                  WV3Status.dails[0] = (int)((-nunchuk.pitch[0]*(180.0/PI)/10)/2.0*increment);
                  WV3Status.dails[1] = (int)((nunchuk.roll[0]*(180.0/PI)/10)/5.0*increment);
                  WV3Status.dails[4] = (-nc_stick.x/9)/2.0*increment;
            } else if (WV3Status.nc_buttons&0x02){ 
                  WV3Status.dails[6] = (int)((nunchuk.pitch[0]*(180.0/PI)/10)/2.0*increment);
                  WV3Status.dails[2] = (int)((nunchuk.roll[0]*(180.0/PI)/10)/5.0*increment);
                  /*WV3Status.dails[2] = (int)((nc_stick.x/9)/5.0*increment);
                  WV3Status.dails[6] = (int)((-nc_stick.y/9)/5.0*increment);*/
            } else {
                  for (icnt = 0; icnt < 8; icnt++) WV3Status.dails[icnt] = 0;
		  WV3Status.dails[3] = (int)((nc_stick.y/9)/5.0*increment);
                  WV3Status.dails[5] = (int)((-nc_stick.x/9)/5.0*increment);
		  if (abs(WV3Status.dails[3]) > abs(WV3Status.dails[5])) WV3Status.dails[5] = 0;
		  if (abs(WV3Status.dails[5]) > abs(WV3Status.dails[3])) WV3Status.dails[3] = 0;
            }
                        
            break;
         case WIIMOTE_ACC_ABS:
            //printf("(WM/NC) pitch: %f %f roll: %f %f yaw: %f %f \n",remote.pitch[0]*(180/PI),nunchuk.pitch[0]*(180/PI),remote.roll[0]*(180/PI),nunchuk.roll[0]*(180/PI),remote.yaw[0]*(180/PI),nunchuk.yaw[0]*(180/PI));
            WV3Status.dails[3] = nc_stick.y / 15;
            if (WV3Status.nc_buttons&0x01){
                  WV3Status.dails[0] = -((nunchuk.pitch[5]*(180/PI)+180)-(nunchuk.pitch[0]*(180/PI)+180)*(1024/983));
                  WV3Status.dails[1] = -((nunchuk.yaw[5]*(180/PI)+180)-(nunchuk.yaw[0]*(180/PI)+180)*(1024/983));
                  WV3Status.dails[4] = (nunchuk.roll[5]*(180/PI)+180)-(nunchuk.roll[0]*(180/PI)+180)*(1024/983);
                  if (WV3Status.dails[0] > 128 || WV3Status.dails[0] <-128)
                        WV3Status.dails[0] = 0;
                  if (WV3Status.dails[1] > 128 || WV3Status.dails[1] <-128)
                        WV3Status.dails[1] = 0;
                  if (WV3Status.dails[4] > 128 || WV3Status.dails[4] <-128)
                        WV3Status.dails[4] = 0;
                  //printf("dial0: %d  dial1: %d  dial4: %d\n", WV3Status.dails[0],WV3Status.dails[1],WV3Status.dails[4]);
                  
            } else {
                  WV3Status.dails[0] = 0;
                  WV3Status.dails[1] = 0;
                  WV3Status.dails[4] = 0;
            }
            if (WV3Status.buttons&0x000c){
                  WV3Status.dails[2] = remote.roll[0]*(180/PI)/15;
                  WV3Status.dails[6] = remote.pitch[0]*(180/PI)/15;
            } else {
                  WV3Status.dails[2] = 0;
                  WV3Status.dails[6] = 0;
            }
            
            break;

         case WIIMOTE_IR:
            /*x = -(WV3Status.IRLocX/WIIMOTE_IR_X_MAX)*DspWidth + DspWidth;
            y = (WV3Status.IRLocY/WIIMOTE_IR_Y_MAX)*DspHeight;
            printf("x: %d y: %d \n",x,y);*/
            x = -((double)WV3Status.IRLocX)/1000*1280+1280;
            y = ((double)WV3Status.IRLocY)/760*1024;
            XWarpPointer(XFtnDpy, None, DefaultRootWindow(XFtnDpy), 0, 0, 0, 0, x, y);
            break;
         /*case FISHING:
            printf("(WM/NC) pitch: %f %f roll: %f %f \n",remote.pitch[0],nunchuk.pitch[0],remote.roll[0],nunchuk.roll[0]);
            break;*/
         default:
            break;
      }

      //printf("Dails %d %d %d %d %d %d %d %d\n", WV3Status.dails[0],WV3Status.dails[1],WV3Status.dails[2],WV3Status.dails[3],WV3Status.dails[4],WV3Status.dails[5],WV3Status.dails[6],WV3Status.dails[7]);
      
      WV3Pass.update = 0;
      for (icnt = 0; icnt < 8; icnt++) {
         if (WV3Status.dails[icnt] != 0){
            WV3Pass.dails[icnt] = WV3Status.dails[icnt];
            WV3Pass.update = 1 | WV3Pass.update;
         } else {
            WV3Pass.dails[icnt] = 0;
         }
      }

      return WV3Pass;
}

/* Begin Wiimote Interface */

/* Call to connect to first discovered Wiimote: returns wiimote data structure if successful, NULL if unsuccessful */
int WMConnect()
{
      bdaddr_t bdaddr;
      bdaddr = *BDADDR_ANY;
      unsigned char read_buf[WIIMOTE_MAX_READ_LEN];
      int read_offset, read_len;
      int i;

      wiimote = NULL;
      
      printf("Put Wiimote in discoverable mode now (press 1+2)...\n");
      sleep(5);
      printf("Attempting to connect!\n");
      if ((wiimote = wiimote_connect(bdaddr, &wiimote_callback)) == NULL) {
             printf("Unable to connect...\n");
             return 1;
      }

      /* Get Calibration Data */
      if (wiimote_read(wiimote, WIIMOTE_RW_EEPROM, 0x16, 7, read_buf)) {
            printf("Unable to retrieve accelerometer calibration");
      } else {
            acc_zero.x = read_buf[0];
            acc_zero.y = read_buf[1];
            acc_zero.z = read_buf[2];
            acc_one.x  = read_buf[4];
            acc_one.y  = read_buf[5];
            acc_one.z  = read_buf[6];
      }

      WV3Status.mode = WIIMOTE_BTN;
      WMSetReport();
      
      wiimote_command(wiimote, WIIMOTE_CMD_STATUS, 0);

      printf("Wiimote Connected!\n");

      XInitThreads();
      
      return 0;
}

/* Call to disconnect current wiimote: returns 0 if there are no wiimotes connected, 1 if unsuccessful */
int WMDisconnect()
{
      if (wiimote) {
            if (wiimote_disconnect(wiimote)) {
                  printf("error on disconnect\n");
                  return 1;
            }
      } else {
            printf("There is no wiimote currently connected...\n");
            return 0;
      }
      wiimote = NULL;
      return 0;
}

int WMSetReport()
{
	unsigned char rpt_mode;
	
	rpt_mode = WIIMOTE_RPT_STATUS | WIIMOTE_RPT_BTN;

      // Force Relative Acceleration mode...
      WV3Status.mode = WIIMOTE_ACC_REL;

      switch (WV3Status.mode){
      case WIIMOTE_ACC_REL:
            rpt_mode |= WIIMOTE_RPT_EXT;
		rpt_mode |= WIIMOTE_RPT_ACC;
            printf("Setting Rel-ACC Mode...\n");
            sensitivity = 4.0;
            break;
      case WIIMOTE_ACC_ABS:
		rpt_mode |= WIIMOTE_RPT_EXT;
		rpt_mode |= WIIMOTE_RPT_ACC;
            printf("Setting Abs-ACC Mode...\n");
            sensitivity = 1.0;
            break;
      case WIIMOTE_IR:
		rpt_mode |= WIIMOTE_RPT_IR;
		rpt_mode |= WIIMOTE_RPT_ACC;
            printf("Setting IR Mode...\n");
            DspWidth = XDisplayWidth(XFtnDpy,XDefaultScreen(XFtnDpy));
            DspHeight = XDisplayHeight(XFtnDpy,XDefaultScreen(XFtnDpy));
            break;
	/*case FISHING:
		rpt_mode |= WIIMOTE_RPT_EXT;
            rpt_mode |= WIIMOTE_RPT_ACC;
            printf("Going fishing...\n");
            break;*/
      default:
            break;
      }

      if (wiimote_command(wiimote, WIIMOTE_CMD_RPT_MODE, rpt_mode)) {
           printf("error setting report mode");
            return 1;
      }
      
      return 0;

}

int WMSetLED(unsigned char flags)
{
      if (wiimote_command(wiimote, WIIMOTE_CMD_LED, flags)) {
            printf("error setting LEDs\n");
            return 1;
      }
      
      return 0;
}

int WMTogRumble()
{
      unsigned char flags;
      
      flags = (wiimote->led_rumble_state == 1) ? 0 : 1;
      if (wiimote_command(wiimote, WIIMOTE_CMD_RUMBLE, flags)) {
            printf("error rumbling\n");
            return 1;
      }
      
      return 0;
}

void WMClrIRData()
{
      int i;

      for (i=0; i < WIIMOTE_IR_SRC_COUNT; i++) {
            ir_data.src[i].x = -1;
            ir_data.src[i].y = -1;
            ir_data.src[i].size = -1;
      }
}

void wiimote_callback(union wiimote_mesg *mesg)
{

      char *ext_str;
      unsigned char buf[7];
      static enum wiimote_ext_type ext_type = WIIMOTE_EXT_NONE;

      switch (mesg->type) {
      case WIIMOTE_MESG_STATUS:
            switch (mesg->status_mesg.extension) {
            case WIIMOTE_EXT_NONE:
                  ext_str = "No extension";
                  break;
            case WIIMOTE_EXT_NUNCHUK:
                  ext_str = "Nunchuk";
                  if (ext_type != WIIMOTE_EXT_NUNCHUK) {
                        if (wiimote_read(wiimote, WIIMOTE_RW_REG | WIIMOTE_RW_DECODE,0xA40020, 7, buf)) {
                              printf("Unable to retrieve nunchuk calibration");
                        }
                        else {
                              nc_acc_zero.x = buf[0];
                              nc_acc_zero.y = buf[1];
                              nc_acc_zero.z = buf[2];
                              nc_acc_one.x  = buf[4];
                              nc_acc_one.y  = buf[5];
                              nc_acc_one.z  = buf[6];
                        }
                  }
                  break;
            case WIIMOTE_EXT_CLASSIC:
                  ext_str = "Classic controller";
                  break;
            case WIIMOTE_EXT_UNKNOWN:
                  ext_str = "Unknown extension";
                  break;
            }
            ext_type = mesg->status_mesg.extension;
            break;
      case WIIMOTE_MESG_BTN:
            WM_btn(&mesg->btn_mesg);
            break;
      case WIIMOTE_MESG_ACC:
            WM_acc(&mesg->acc_mesg);
            break;
      case WIIMOTE_MESG_IR:
            WM_ir(&mesg->ir_mesg);
            break;
      case WIIMOTE_MESG_NUNCHUK:
            WM_nunchuk(&mesg->nunchuk_mesg);
            break;
      case WIIMOTE_MESG_CLASSIC:
            break;
      default:
            break;
      }

}

void WM_btn(struct wiimote_btn_mesg *mesg)
{      

      int i;

      WV3Status.buttons = mesg->buttons;

      switch (mesg->buttons) {
      case WIIMOTE_BTN_2:
            break;
      case WIIMOTE_BTN_1:
            break;
      case 0x0003: /* 1 and 2 pressed */
            break;
      case WIIMOTE_BTN_B:
            break;
      case 0x0005: /* 2 and B pressed */
            break;
      case 0x0006: /* 1 and B pressed */
            break;
      case 0x0007: /* 1 2 and B pressed */
            break;
      case WIIMOTE_BTN_A:
            break;
      case 0x0009: /* 2 and A pressed */
            break;
      case 0x000a: /* 1 and A pressed */
      case 0x000b: /* 1 2 and A pressed */
            break;
      case 0x000c: /* A and B pressed */
            
            break;
      case 0x000d: /* 2 A and B pressed */
            break;
      case 0x000e: /* 1 A and B pressed */
            break;
      case 0x000f: /* 1 2 A and B pressed */
            break;
      case WIIMOTE_BTN_HOME:
            WV3Status.mode++;
            if (WV3Status.mode >= LASTMODE){
                  WV3Status.mode = NOMODE;
                  WV3Status.mode++;
            }
            WMSetReport();
            break;
      case WIIMOTE_BTN_LEFT:
            break;
      case WIIMOTE_BTN_RIGHT:
             break;
      case WIIMOTE_BTN_DOWN:
            break;
      case WIIMOTE_BTN_UP:
            break;
      case WIIMOTE_BTN_PLUS:
            if (WV3Status.mode == WIIMOTE_ACC_REL) {
                  sensitivity = sensitivity / 2.0;
                  if (sensitivity <=0.25) sensitivity = 0.25;
                  printf("Rotation Rate is %2.2f\n",-(log(sensitivity)/log(2))+8);
            }
            break;
      case WIIMOTE_BTN_MINUS:
            if (WV3Status.mode == WIIMOTE_ACC_REL) {
                  sensitivity = sensitivity * 2.0;
                  if (sensitivity >=128.0) sensitivity = 128.0;
                  printf("Rotation Rate is %2.2f\n",-(log(sensitivity)/log(2))+8);
            }
            break;
      default:
             break;
      }
      
}

void WM_acc(struct wiimote_acc_mesg *mesg)
{

      double a_x, a_y, a_z, a;
      double roll, pitch, yaw;
      int i;
      
      a_x = ((double)mesg->x - acc_zero.x) / (acc_one.x - acc_zero.x);
      a_y = ((double)mesg->y - acc_zero.y) / (acc_one.y - acc_zero.y);
      a_z = ((double)mesg->z - acc_zero.z) / (acc_one.z - acc_zero.z);
      
      a = sqrt(pow(a_x,2)+pow(a_y,2)+pow(a_z,2));
      /*roll = atan(a_x/a_z);
      if (a_z <= 0.0) {
            roll += PI * ((a_x > 0.0) ? 1 : -1);
      }
      roll *= -1;

      pitch = atan(a_y/a_z*cos(roll));

      yaw = atan(a_z/1);*/

      roll = atan(a_x/a_z)*(1-abs(a_y));
      if (a_z <= 0.0) {
            roll += PI * ((a_x > 0.0) ? 1 : -1);
      }
      if (isnan(roll)){
            roll = 0.0;
      }else if (roll*(180/PI) == 270) {
            roll = PI/2;
      }else if (roll*(180/PI) == -270) {
            roll = -PI/2;
      }

      if (WV3Status.mode == WIIMOTE_ACC_REL){
            pitch = atan(a_y/a_z);
            if (a_z <= 0.0) {
                  pitch += PI * ((a_y > 0.0) ? 1 : -1);
            }
      }else{
            pitch = atan(a_z/a_y)*(1-abs(a_x));
            if (a_y <= 0.0) {
                  pitch += PI * ((a_z > 0.0) ? 1 : -1);
            }
      }

      if (isnan(pitch)){
            pitch = 0.0;
      }else if (pitch*(180/PI) == 270){
            pitch = PI/2;
      }else if (pitch*(180/PI) == -270) {
            pitch = -PI/2;
      }

      yaw = atan(a_y/a_x)*(1-abs(a_z));
      if (a_x <= 0.0) {
            yaw += PI * ((a_y > 0.0) ? 1 : -1);
      }
      if (isnan(yaw)){
            yaw = 0.0;
      }else if (yaw*(180/PI) == 270){
            yaw = PI/2;
      }else if (yaw*(180/PI) == -270) {
            yaw = -PI/2;
      }

      //printf("a_x: %f a_y: %f a_z: %f\n roll: %f pitch: %f yaw: %f\n", a_x,a_y,a_z,roll,pitch,yaw);

      /*shift and store motion array values*/
      for ( i = MtnStrSiz-1 ; i > 0 ; i-- ) 
      {
            remote.a_x[i] = remote.a_x[i-1];
            remote.a_y[i] = remote.a_y[i-1];
            remote.a_z[i] = remote.a_z[i-1];
            remote.roll[i] = remote.roll[i-1];
            remote.pitch[i] = remote.pitch[i-1];
            remote.yaw[i] = remote.yaw[i-1];
      }
      remote.a_x[0] = a_x;
      remote.a_y[0] = a_y;
      remote.a_z[0] = a_z;
      remote.roll[0] = roll;
      remote.pitch[0] = pitch;
      remote.yaw[0] = yaw;

}

void WM_ir(struct wiimote_ir_mesg *mesg)
{

      int i, size;
      int cntsrc;
      int srcx[WIIMOTE_IR_SRC_COUNT],srcy[WIIMOTE_IR_SRC_COUNT];

      ir_data = *mesg;

      cntsrc = 0;
      WV3Status.IRLocX = 0;
      WV3Status.IRLocY = 0;

      for (i=0; i < WIIMOTE_IR_SRC_COUNT; i++) {
            if (ir_data.src[i].valid) {
                  
                  srcx[i] = ir_data.src[i].x;
                  srcy[i] = ir_data.src[i].y;

                  //printf("IR Source: %d X: %u Y: %u Size: %c \n", i, ir_data.src[i].x, ir_data.src[i].y, ir_data.src[i].size);
                 
                  /*filter out stray data*/
                  if (abs(srcx[i]-srcx[0]) > 200 || abs(srcy[i]-srcy[0]) > 150)
                     continue;

                  WV3Status.IRLocX = WV3Status.IRLocX + srcx[i];
                  WV3Status.IRLocY = WV3Status.IRLocY + srcy[i];
                  cntsrc++; 
            }
      }
      
      if (cntsrc > 0) {
            WV3Status.IRLocX = WV3Status.IRLocX/cntsrc;
            WV3Status.IRLocY = WV3Status.IRLocY/cntsrc;
            //printf("IRLocX: %d IRLocY: %d cntsrc:%d \n",WV3Status.IRLocX,WV3Status.IRLocY,cntsrc);
      }

}

void WM_nunchuk(struct wiimote_nunchuk_mesg *mesg)
{

      double a_x, a_y, a_z, a;
      double roll, pitch, yaw;
      int i;
      int x, y;

      WV3Status.nc_buttons=mesg->buttons;
      
      switch (mesg->buttons){
      case WIIMOTE_NUNCHUK_BTN_Z:
            break;
      case WIIMOTE_NUNCHUK_BTN_C:
            break;
      default:
            break;
      }

      nc_stick.valid = 1;
      nc_stick.x = (int)mesg->stick_x - 130;
      nc_stick.y = (int)mesg->stick_y - 125;

      if (abs(nc_stick.x) <= DEADBAND) 
        nc_stick.x = 0;
      if (abs(nc_stick.y) <= DEADBAND) 
        nc_stick.y = 0;

      //printf("stick values: %d %d ~ %d %d\n",(int)mesg->stick_x,(int)mesg->stick_y,nc_stick.x,nc_stick.y);

      /* TODO: get nunchuk calibration */
      a_x = ((double)mesg->acc_x - nc_acc_zero.x) / (nc_acc_one.x - nc_acc_zero.x);
      a_y = ((double)mesg->acc_y - nc_acc_zero.y) / (nc_acc_one.y - nc_acc_zero.y);
      a_z = ((double)mesg->acc_z - nc_acc_zero.z) / (nc_acc_one.z - nc_acc_zero.z);
      a = sqrt(pow(a_x,2)+pow(a_y,2)+pow(a_z,2));
      /*roll = atan(a_x/a_z);
      if (a_z <= 0.0) {
            roll += PI * ((a_x > 0.0) ? 1 : -1);
      }
      roll *= -1;

      pitch = atan(a_y/a_z*cos(roll));*/

      roll = atan(a_x/a_z);
      if (a_z <= 0.0) {
            roll += PI * ((a_x > 0.0) ? 1 : -1);
      }
      if (isnan(roll)){
            roll = 0.0;
      }else if (roll*(180/PI) == 270) {
            roll = PI/2;
      }else if (roll*(180/PI) == -270) {
            roll = -PI/2;
      }

      if (WV3Status.mode == WIIMOTE_ACC_REL){
            pitch = atan(a_y/a_z);
            if (a_z <= 0.0) {
                  pitch += PI * ((a_y > 0.0) ? 1 : -1);
            }
      }else{
            pitch = atan(a_z/a_y);
            if (a_y <= 0.0) {
                  pitch += PI * ((a_z > 0.0) ? 1 : -1);
            }
      }
      if (isnan(pitch)){
            pitch = 0.0;
      }else if (pitch*(180/PI) == 270){
            pitch = PI/2;
      }else if (pitch*(180/PI) == -270) {
            pitch = -PI/2;
      }

      yaw = atan(a_y/a_x);
      if (a_x <= 0.0) {
            yaw += PI * ((a_y > 0.0) ? 1 : -1);
      }
      if (isnan(yaw)){
            yaw = 0.0;
      }else if (yaw*(180/PI) == 270){
            yaw = PI/2;
      }else if (yaw*(180/PI) == -270) {
            yaw = -PI/2;
      }

      /*shift and store motion array values*/
      for ( i = MtnStrSiz-1 ; i > 0 ; i-- ) 
      {
            nunchuk.a_x[i] = nunchuk.a_x[i-1];
            nunchuk.a_y[i] = nunchuk.a_y[i-1];
            nunchuk.a_z[i] = nunchuk.a_z[i-1];
            nunchuk.roll[i] = nunchuk.roll[i-1];
            nunchuk.pitch[i] = nunchuk.pitch[i-1];
            nunchuk.yaw[i] = nunchuk.yaw[i-1];
      }
      nunchuk.a_x[0] = a_x;
      nunchuk.a_y[0] = a_y;
      nunchuk.a_z[0] = a_z;
      nunchuk.roll[0] = roll;
      nunchuk.pitch[0] = pitch;
      nunchuk.yaw[0] = yaw;


}
