/*
 * xftnStart3D.c
 *
 * Allows for custom Visual3 3D window backgrounds.
 *
 * Copyright 1990 - 2005, Massachusetts Institute of Technology.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef WIN32
#  include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <tiffio.h>


extern float XFtn3Dsize[4];
extern int   XFtn3Dtype;


/* when called:
 *      1) the view matrix is the identity
 *      2) the context is to the 3D window
 */

void
xftnStart3D(int xsize, int ysize)
{
 char    *env;
 int     i, j, k;
 float   sx, sy, dx, dy, x[3];
 size_t  npixels;
 GLubyte u[3];
 TIFF    *tif;
 static uint32 w, h, *raster = NULL;

 env = getenv("Visual3_BG");
 if (env == NULL) return;

 if (raster == NULL) {
   tif = TIFFOpen(env, "r");
   if (tif == NULL) return;

   TIFFGetField(tif, TIFFTAG_IMAGEWIDTH,  &w);
   TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
   npixels = w * h;
   raster  = (uint32 *) _TIFFmalloc(npixels * sizeof(uint32));
   if (raster == NULL) {
     _TIFFfree(raster);
     return;
   }
   if (TIFFReadRGBAImage(tif, w, h, raster, 0) == 0) {
     printf(" TIFFReadRGBAImage error!\n");
     _TIFFfree(raster);
     raster = NULL;
     return;
   }
   TIFFClose(tif);
 }

 dx = (XFtn3Dsize[1] - XFtn3Dsize[0])/xsize;
 dy = (XFtn3Dsize[3] - XFtn3Dsize[2])/ysize;
                      x[2] = -0.999;
 if (XFtn3Dtype == 3) x[2] =  0.999;
 if (xsize >= w) {
   sx = XFtn3Dsize[0] + 0.5*(xsize-w)*dx;
 } else {
   sx = XFtn3Dsize[0] - 0.5*(w-xsize)*dx;
 }
 if (ysize >= h) {
   sy = XFtn3Dsize[2] + 0.5*(ysize-h)*dy;
 } else {
   sy = XFtn3Dsize[2] - 0.5*(h-ysize)*dy;
 }

 glBegin(GL_POINTS);
 for (k = j = 0; j < h; j++) {
   x[1] = sy + dy*j;
   for (i = 0; i < w; i++, k++) {
     x[0] = sx + dx*i;
     u[0] = TIFFGetR(raster[k]);
     u[1] = TIFFGetG(raster[k]);
     u[2] = TIFFGetB(raster[k]);
     glColor3ubv(u);
     glVertex3fv(x);
   }
 }
 glEnd();
/*  _TIFFfree(raster);  */
}
