!#######################################################################
      module common_area
!#######################################################################
      implicit none
      integer :: ncycs=1
      integer :: ncyce=1
      integer :: nstep=1
      integer :: iout=3
      integer :: mq
      integer :: cut_dir=2
      integer :: num_cut=0
      integer :: my_id
      integer :: cut_posi(10)
      integer,dimension(30,10):: id_planes

      integer :: msu_or_apnasa=1
      INTEGER :: merge_output=1 ! same values as iout
                                ! 1 Vstage
                                ! 3 Plot3D
      integer :: relative_or_abs=1
      integer :: my_row=0
      integer :: my_psge=0
      INTEGER :: vstg_hdr(7)=(/1,1,1,1,1,1,1/)
      INTEGER :: lu_vstg=42
      integer :: lu_ext=10
      integer :: lu_ast=11
      integer :: lu_g=12
      integer :: lu_q=13
      integer :: lu_out=14
      integer :: ni=0
      integer :: nj=0
      integer :: nk=0

      real*8  :: dtdt,dt_lr(2),pi2
      real*4,allocatable,dimension(:,:,:) ::x,y,z,ym,zm
      real*4,allocatable,dimension(:,:,:,:) ::grd,vstg
      real*4,allocatable,dimension(:,:,:) ::r,ru,rv,rw,e,rv_rad,rv_tan
      real*4  ::  fsmach,alpha,rel,time

      real*8 :: gam,gm1,sqrgam

      character(len=256) :: error_message(10)
      character(len=40)  :: qchar_ext='?'
      character :: casename*40

      logical :: has_qchar_ext=.false.
      logical :: motion_grid=.false.
      end module common_area

!#######################################################################
      program p_merge
!#######################################################################
      use common_area
      implicit none
#ifndef SERIAL
      include "mpif.h"
#endif

      integer ncyct,ierr,ierror,numnodes,n,dummy
      INTEGER,ALLOCATABLE,DIMENSION(:) :: mems_id,is,ie,js,je,ks,ke,ig
      INTEGER,DIMENSION(30) :: mems_cut_id ! THIS IS HARDCODED
      integer :: ip,isd,ied,jsd,jed,ksd,ked,igd,irow,ipsge
      integer :: i,j,k,i1,j1,k1,i2,j2,k2,l1,l2,l3,count
      integer :: norm1,norm2,norm3,ll,m,num_members,ni1,nj1,nk1
      integer :: mems=0
      integer :: mems_cut=0
      integer :: mi=0
      integer :: mj=0
      integer :: mk=0
      integer :: my_job=-1
      integer :: id=9999
      integer :: blk_num_cuts=0

      REAL :: cput(2)

      character :: fname*40, hnam*40

!=======================================================================
! First, establish id and determine which blade passage needs to be
! merged.  Even in the serial version, it will read the namelist
! file pmerge.in.
#ifdef SERIAL
      numnodes=1
      WRITE(*,*) 'Enter global passage number to merge:'
      READ(*,*) my_id
      my_id=my_id-1
#else
      call mpi_init( ierr )
      call mpi_comm_rank( MPI_COMM_WORLD, my_id, ierr )
      call mpi_comm_size( MPI_COMM_WORLD, numnodes, ierr )
#endif
      IF(my_id.EQ.0)CALL CPU_TIME(cput(1))
      WRITE(*,*) 'I am ',my_id,' of ',numnodes

      pi2=2*cos(-1.0)
      gm1=0.4
      mems_cut_id=0
      

! Establish output files for stdout. The serial version will write
! simply to the screen while the parallel version will open .out
! files as in TURBO.
      ll=my_id+1
      call get_MERGE_out_name(ll,fname)
#ifdef SERIAL
      lu_out=6
#else
      open(lu_out,file=fname,status='unknown',form='formatted')
#endif

!=======================================================================
! Read the namelist from pmerge.in and open and read the extract file
      call read_pmerge_nml
      call get_extract_name(fname)
      open(lu_ext,file=fname,form='formatted')

! Get the number of entries for this passage to dynamically allocate
! arrays.
      DO WHILE(id.NE.0)       
        READ(lu_ext,*)id,ip,isd,ied,jsd,jed,ksd,ked,igd,irow,ipsge
        IF(ip.EQ.my_id+1)THEN
          mems=mems+1
        ENDIF
      ENDDO
      ALLOCATE(mems_id(mems),is(mems),ie(mems),
     &         js(mems),je(mems),ks(mems),ke(mems),ig(mems))
      REWIND(lu_ext)
      mems_id=0
      is=0
      ie=0
      js=0
      je=0
      ks=0
      ke=0
      ig=0

      mems=0
      id=9999
      do while(id .ne. 0)       
         read(lu_ext,*)id,ip,isd,ied,jsd,jed,ksd,ked,igd,irow,ipsge
! ip==global passage id, ipsge==passage id in each row
! check for different passage data
         if(ip==my_id+1)then
            mems=mems+1
            my_row=irow
            my_psge=ipsge
            mems_id(mems)=id
            is(mems)=isd
            ie(mems)=ied
            js(mems)=jsd
            je(mems)=jed
            ks(mems)=ksd
            ke(mems)=ked
            ig(mems)=igd
            ni=max(ni,ied)
            nj=max(nj,jed)
            nk=max(nk,ked)
          endif
      enddo
      close(lu_ext)  

! Read namelists from input0?
      call read_output_nml
      call read_bldrow_nml(my_row)
      call nondim_dtdt
      id_planes=0

! Operate on cut planes
      if(iout==5)then
        if(num_cut >0)then
           norm1=1/cut_dir
           norm2=(2/cut_dir)*(cut_dir/2)
           norm3=(3/cut_dir)*(cut_dir/3)
           do n=1,mems
              count=0
              l1=norm1*is(n)+norm2*js(n)+norm3*ks(n)
              l2=norm1*ie(n)+norm2*je(n)+norm3*ke(n)
              ll=norm1*ni+norm2*nj+norm3*nk
              do i=1,num_cut
                 if(cut_posi(i).lt.1 .or. cut_posi(i).gt.ll)then
              write(lu_out,*)'cut_plane_global_posi out of range'
                    goto 99
                 endif
                 if(cut_posi(i).ge.l1.and.cut_posi(i).le.l2)then
                     count=count+1
                     id_planes(n,count)=i
                 endif 
              enddo
              if(id_planes(n,1) /=0) then
                 mems_cut=mems_cut+1
                 mems_cut_id(mems_cut)=mems_id(n)
              endif
           enddo 
         endif

         select case(cut_dir)
           case(1) 
               ni=num_cut
           case(2) 
               nj=num_cut
           case(3) 
               nk=num_cut
         end select
      endif
               ni1=ni+1
               nj1=nj+1
               nk1=nk+1

!=======================================================================
! Allocate Plot3D variables.  This should be totally unnecessary for Vstage files
      IF(merge_output.NE.1)THEN
        allocate(x(ni,nj,nk))
        allocate(y(ni,nj,nk))
        allocate(z(ni,nj,nk))
        allocate(ym(ni,nj,nk))
        allocate(zm(ni,nj,nk))
        allocate(r(ni,nj,nk))
        allocate(ru(ni,nj,nk))
        allocate(rv(ni,nj,nk))
        allocate(rw(ni,nj,nk))
        allocate(e(ni,nj,nk))
        allocate(rv_rad(ni,nj,nk))
        allocate(rv_tan(ni,nj,nk))
      ENDIF


! Allocate the Vstage data based on the maximum indicies from extract.in
      allocate(vstg(ni1,nj1,nk1,9))
      allocate(grd(ni,nj,nk,3))
      CALL GetVstageHeader

!=======================================================================
! Read in all of the ast grid files and store them into grd
      num_members=mems
      if(iout==5) num_members=mems_cut 
      do 10 n=1,num_members
         id=mems_id(n)
         if(iout==5) id=mems_cut_id(n)
         i1=is(n)
         i2=ie(n)
         j1=js(n)
         j2=je(n)
         k1=ks(n)
         k2=ke(n)
         call read_ast_g(n,id,iout,i1,i2,j1,j2,k1,k2)
 10   continue  

         if(iout==1) then
!           Cylindrical grid (grd) to Cartersian grid (xyz)
            IF(merge_output.NE.1)call grd_to_xyz 
            if(msu_or_apnasa==1)call xyz_toggle_LR
         elseif(iout==3)then
              call cal_grd
              if(msu_or_apnasa==2)call xyz_toggle_LR
         endif 

      if(.not. motion_grid)call write_merged_grid
      if(motion_grid .and. dtdt==0)call write_merged_grid
      WRITE(*,*) 'Grid merged on ',my_id

!=======================================================================
!---loop over iteration number for solution---------------------
          do ncyct=ncycs,ncyce,nstep
             do n=1,num_members
               id=mems_id(n)
               if(iout==5) id=mems_cut_id(n)
                i1=is(n)
                i2=ie(n)
                j1=js(n)
                j2=je(n)
                k1=ks(n)
                k2=ke(n)
                call read_ast_q(n,id,ncyct,i1,i2,j1,j2,k1,k2)
              enddo !n=1,num_members

                if(iout==1)call q_ctr_cyl_to_node_car
                if(iout==3.and.msu_or_apnasa==2)call q_msu_to_apnasa
                if(iout==1.and.msu_or_apnasa==1)call q_apnasa_to_msu

! vel_abs_to_rel must be called after q has been convtered 
! because q(msu, or apnasa) has to be consistent with dt_lr(1,or,2)

                if(relative_or_abs==1)call vel_abs_to_rel

                if(motion_grid)call write_merged_motion_grid(ncyct)
                call write_merged_solu(ncyct)
          enddo !ncyct
!--------end of loops -----------------------------------
      WRITE(*,*) 'Flows merged on ',my_id
      WRITE(lu_out,*) 'Entering the MPI_BARRIER'

!=======================================================================
! Finalize program
#ifdef SERIAL
  99  CONTINUE
#else
  99  CLOSE(lu_out)
      CALL mpi_barrier(MPI_COMM_WORLD,ierror)
      IF(my_id==0)THEN
        CALL CPU_TIME(cput(2))
        WRITE(*,*) 'All Complete'
        WRITE(*,*) 'Total CPU time: ',cput(2)-cput(1),' s'
      ENDIF
      CALL mpi_finalize(ierr)
#endif
        STOP
      end

!#######################################################################
      subroutine get_extract_name(fname)
!#######################################################################
      use common_area
      implicit none
      integer :: ll
      character :: xch*8
      character,intent(out):: fname*40

       xch='.extract'
       ll=len_trim(casename)
       fname=casename(1:ll)//xch
       WRITE(lu_out,*) 'Extract Name is ',fname

       end subroutine get_extract_name

!#######################################################################
      subroutine get_MERGE_out_name(npsge,name)
!#######################################################################
      use common_area
      implicit none
      character   :: xch1*5,xch2*20,xch3*4
      character,intent(out):: name*40
      integer,intent(in)::npsge
      integer :: l1
        xch1='MERGE.'
        xch3='.out'

       call int_to_char(npsge,xch2,l1)
       name=xch1//xch2(1:l1)//xch3
      end subroutine get_MERGE_out_name
!#######################################################################
      subroutine get_ast_gridname(nblk,fname)
!#######################################################################
      use common_area
      implicit none
      character   :: xch1*5,xch2*20,xch3*2
      character,intent(out) :: fname*40
      integer,intent(in) :: nblk
      integer :: l1

        xch1='ast.b'
        xch3='.x'

       call int_to_char(nblk,xch2,l1)
       fname=xch1//xch2(1:l1)//xch3
       end subroutine get_ast_gridname

!#######################################################################
      subroutine get_ast_qname(nblk,iter,fname)
!#######################################################################
      use common_area
      implicit none

      character   :: xch1*5,xch2*20,xch3*2,xch4*20,xch5
      character,intent(out) :: fname*40
      integer,intent(in)    :: nblk,iter
      integer               :: l1,l2,lext

           xch1='ast.b'
           xch3='.q'
           xch5='.'
           lext=len_trim(qchar_ext)

           call int_to_char(nblk,xch2,l1)
           call int_to_char(iter,xch4,l2)

       if(.not. has_qchar_ext)then
           fname=xch1//xch2(1:l1)//xch3//xch4(1:l2)
       else if(has_qchar_ext)then
           fname=xch1//xch2(1:l1)//xch3//qchar_ext(1:lext)
       endif
       end subroutine get_ast_qname

!#######################################################################
      subroutine get_merged_gridname5(lrow,lpsge,ldir,lpln,iter,fname)
!#######################################################################
      use common_area
      implicit none
      character   :: xch1*2,xch2,xch3*5,xdir,xch4*2
      character(len=20):: xrow,xpsge,xplane,xiter
      character,intent(out):: fname*40
      integer,intent(in):: lrow,lpsge,ldir,lpln,iter
      integer     :: ll,l1,l2,l3,l4

        xch1='.r'
        xch2='b'
        xch3='.p3d.'
        xch4='.g'
        ll=len_trim(casename)
        if(ldir==1)xdir='i'
        if(ldir==2)xdir='j'
        if(ldir==3)xdir='k'
       
       call int_to_char(lrow,xrow,l1)
       call int_to_char(lpsge,xpsge,l2)
       call int_to_char(lpln,xplane,l3)

       fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     .                     //xch3//xdir//xch4
       if(iter>=0)then
         call int_to_char(iter,xiter,l4)
         fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     .                 //xch3//xdir//xch4//xiter(1:l4)
       endif
       end subroutine get_merged_gridname5

!#######################################################################
      subroutine read_pmerge_nml
!#######################################################################
! Read the input file for the program, pmerge.in.
      use common_area
      implicit none
      character(len=40)  :: fname                ! namelist input file name
      logical            :: fname_exists         ! logical telling whether namelist input file exists
      logical            :: file_end = .false.
      integer            :: ierr
      integer, parameter :: ioutt=6              ! output location
      integer, parameter :: lunml=39             ! logical units to open namelist file on
      character(len=40)  :: case_name
      integer            :: iter_start
      integer            :: iter_end
      integer            :: iter_step
     

      namelist/PMERGE_PARAMETERS/
     . case_name,iter_start,iter_end,iter_step,
     . msu_or_apnasa,motion_grid,qchar_ext,
     . relative_or_abs,merge_output

      fname='pmerge.in'
      open(lunml,file=fname,form='formatted')
      rewind(lunml)

        read(lunml,nml=PMERGE_PARAMETERS,end=203)
 203    continue
        close(lunml)

        casename=case_name
        ncycs=iter_start
        ncyce=iter_end
        nstep=iter_step
        if(qchar_ext .ne. '?')then
          has_qchar_ext=.true.
          ncycs=1
          ncyce=1
        endif

      return
      end subroutine read_pmerge_nml
!#######################################################################
      subroutine read_bldrow_nml(nbr)
!     read inital condition namelist
!#######################################################################
      use common_area
      implicit none
      character(len=40)  :: fname                ! namelist input file name
      logical            :: fname_exists         ! logical telling whether namelist input file exists
      logical            :: file_end = .false.
      logical            :: upstream_freq_only

      integer,intent(in) :: nbr
      integer, parameter :: ioutt=6              ! output location
      integer, parameter :: lunml=39             ! logical units to open namelist file on
      integer            :: num_blades=0
      integer            :: blade_row_ksym=-1
      integer            :: num_time_steps_stored=-1
      integer            :: num_adjacent_blades=0
      integer            :: suction_surface=1
      integer            :: use_wall_func_i=1
      integer            :: use_wall_func_j=1
      integer            :: use_wall_func_k=1
      integer            :: cut_plane_dir=1
      integer            :: num_cut_plane=0
      integer            :: cut_plane_global_posi(10)
      integer            :: num_sources=0
      integer            :: l=0
      real*8             :: omega_bld=0.0
      real*8             :: transition=.1
      real*8             :: x_start_viscous_hub=-9999
      real*8             :: x_start_viscous_case=-9999
      real*8             :: x_start_hub_rotation=9999
      real*8             :: x_end_hub_rotation=9999
      real*8             :: x_start_case_rotation=9999
      real*8             :: x_end_case_rotation=9999

      namelist/BLADE_ROW_PARAMETERS/
     .num_blades,   blade_row_ksym,
     .omega_bld, num_time_steps_stored,
     .num_adjacent_blades,suction_surface,
     .use_wall_func_i,use_wall_func_j,use_wall_func_k,
     .transition,x_start_viscous_hub,
     .x_start_viscous_case,x_start_hub_rotation,
     .x_end_hub_rotation,x_start_case_rotation,
     .x_end_case_rotation,cut_plane_dir,
     .num_cut_plane,cut_plane_global_posi,
     .num_sources,upstream_freq_only

      fname='input'
      select case(nbr)
      case(1:9)
       write(fname(6:6),'(i1)') 0
       write(fname(7:7),'(i1)') nbr
      case(10:99)
       write(fname(6:7),'(i2)') nbr
      end select
      inquire(file=fname,exist=fname_exists)

  510   format(' Reading namelist BLADE_ROW_PARAMETERS for row ',i2)
!        write(ioutt,510) nbr

        open(lunml,file=fname,form='formatted')
        rewind(lunml)

          read(lunml,nml=BLADE_ROW_PARAMETERS,end=503)
  503     continue
          close(lunml)

! input dtdt is in APNASA's convention
          dtdt=omega_bld
          cut_dir=cut_plane_dir
          num_cut=num_cut_plane
          do l=1,num_cut_plane
             cut_posi(l)=cut_plane_global_posi(l)
          enddo

      return
      end subroutine read_bldrow_nml

!#######################################################################
      subroutine nondim_dtdt
!     determine non-dimensional dtheta/dt
!#######################################################################
      use common_area
      implicit none
      character(len=40)  :: fname                ! namelist input file name
      logical            :: fname_exists         ! logical telling whether namelist input file exists
      logical            :: file_end = .false.
      integer, parameter :: ioutt=6              ! output location
      integer, parameter :: lunml=39             ! logical units to open namelist file on
      real*8             :: ref_pressure
      real*8             :: ref_temperature
      real*8             :: ref_velocity
      real*8             :: ref_length
      real*8             :: ref_gamma
      real*8             :: gamma_table(2)
      real*8             :: temp_gam_table(2)
      real*8             :: sutherland_const1
      real*8             :: sutherland_const2
      real*8             :: gamref_t1
      real*8             :: pi

      namelist/REFERENCE_CONDITIONS/
     . ref_pressure,            ref_temperature,
     . ref_velocity,            ref_gamma,
     . ref_length,              gamref_t1,
     . gamma_table,             temp_gam_table,
     . sutherland_const1,       sutherland_const2

      fname='input00'

!      write(ioutt,*) 'Reading namelist REFERENCE_CONDITIONS'
      open(lunml,file=fname,form='formatted')
      rewind(lunml)

        read(lunml,nml=REFERENCE_CONDITIONS,end=2003)
 2003   continue
        close(lunml)

        pi=acos(-1.0)

        gam=ref_gamma
        gm1=gam-1.
        sqrgam=sqrt(gam)
! dt_lr(2) is for APNASA, which takes the value direct from input00
! dt_lr(1) is for TURBO, which is the revers direction
! attention, they use different reference values for nondimensionliztion

        dt_lr(1)=-dtdt*ref_length*pi/(30.0*ref_velocity*sqrt(ref_gamma))
        dt_lr(2)=dtdt*ref_length*pi/(30.0*ref_velocity)
      return
      end subroutine nondim_dtdt

!#######################################################################
      subroutine read_output_nml
!     read output namelist from input00
!#######################################################################
      use common_area
      implicit none
      character(len=40)  :: fname                ! namelist input file name
      logical            :: fname_exists         ! logical telling whether namelist input file exists
      logical            :: file_end = .false.
      integer, parameter :: ioutt=6              ! output location
      integer, parameter :: lunml=39             ! logical units to open namelist file on
      integer            :: num_soln_per_flow_file
      integer            :: num_iter_per_soln_dump
      integer            :: output_format
      integer            :: num_parallel_writes
      integer            :: itime_start_anim
      logical            :: tmavg_output
      logical            :: steady_output

      namelist/OUTPUT/
     . output_format, num_iter_per_soln_dump,
     . num_soln_per_flow_file,num_parallel_writes,
     . itime_start_anim,tmavg_output,steady_output

      iout=3
      mq=1
      fname='input00'
!      write(ioutt,*) 'Reading namelist OUTPUT'
      open(lunml,file=fname,form='formatted')
      rewind(lunml)

      do while (.not. file_end)
        read(lunml,nml=OUTPUT,err=201,end=203)
        close(lunml)
        iout=output_format
        mq=num_soln_per_flow_file
        goto 202

 201    continue
      enddo

 203  write(6,*)'error while reading output_nml '
      stop
 202  continue

      return
      end subroutine read_output_nml

!#######################################################################
      subroutine int_to_char(num,xch,l1)
!     convert integers to characters
!#######################################################################
      implicit none
      integer,intent(in) :: num
      integer:: l1,inum,iord
      character,intent(out):: xch*20

       l1=0
       inum=num
       do iord=1,10
         inum=inum/10
         if(inum.gt.0)l1=l1+1
       enddo
       l1=l1+1
       if(l1.eq.10)write(xch,'(i10)') num
       if(l1.eq.9)write(xch,'(i9)') num
       if(l1.eq.8)write(xch,'(i8)') num
       if(l1.eq.7)write(xch,'(i7)') num
       if(l1.eq.6)write(xch,'(i6)') num
       if(l1.eq.5)write(xch,'(i5)') num
       if(l1.eq.4)write(xch,'(i4)') num
       if(l1.eq.3)write(xch,'(i3)') num
       if(l1.eq.2)write(xch,'(i2)') num
       if(l1.eq.1)write(xch,'(i1)') num
      return
      end subroutine int_to_char

!#######################################################################
      subroutine read_ast_g(nd,id,iformat,i1,i2,j1,j2,k1,k2)
!#######################################################################
      use common_area
      implicit none
      integer,intent(in) :: nd,id,iformat,i1,i2,j1,j2,k1,k2
      integer :: ll,i,j,k,m
      integer :: l1,l2,l3,dummy
      character:: fname*40

      call get_ast_gridname(id,fname)
      open(lu_ast,file=fname,status='unknown',form='unformatted')
      write(lu_out,*)my_id+1,"Reading block grids ",fname

! vstage
       if (iformat==1)then
          rewind(lu_ast)
          read(lu_ast)l1,l2,l3,dummy,dummy,dummy,dummy,
     &                        dummy,dummy,dummy 
          do j=j1,j2
             read(lu_ast)
     &        ((grd(i,j,k,1),i=i1,i2),k=k1,k2),
     &        ((grd(i,j,k,2),i=i1,i2),k=k1,k2),
     &        ((grd(i,j,k,3),i=i1,i2),k=k1,k2)
          enddo
! plot3d
       elseif(iformat==3)then
          rewind(lu_ast)
          read (lu_ast) ll   ! ast files always has one block
          read (lu_ast) (l1,l2,l3,i=1,ll)
          read (lu_ast)
     &            (((x(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            (((y(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            (((z(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2)
          
! cutting plane
       elseif(iformat==5 .and. num_cut>0)then
          read (lu_ast) ll   ! ast files always has one block
          read (lu_ast) (l1,l2,l3,i=1,ll)
          do m=1,ll
          select case(cut_dir)
             case(2) 
               read(lu_ast)
     &         (((x(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         (((y(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         (((z(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2)

             case(1)
               read(lu_ast)
     &         (((x(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         (((y(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         (((z(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2)

             case(3)
               read(lu_ast)
     &         (((x(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         (((y(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         (((z(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m)
          end select
          enddo
       endif
      close(lu_ast)
      return
      end subroutine read_ast_g

!#######################################################################
      subroutine read_ast_q(nd,id,ncyct,i1,i2,j1,j2,k1,k2)
!#######################################################################     
      use common_area
      implicit none
      integer,intent(in) :: nd,id,ncyct,i1,i2,j1,j2,k1,k2
      integer :: m,ll,l1,l2,l3,i,j,k,ncyct_sngl,i2p,j2p,k2p
      character ::fname*40
      REAL,ALLOCATABLE,DIMENSION(:,:,:,:) :: tmpvstg

          call get_ast_qname(id,ncyct,fname)
          write(lu_out,*)my_id+1,"Reading block solution ",fname
          open(lu_ast,file=fname,status='unknown',form='unformatted')
          rewind(lu_ast)

! cutting planes
      if(iout==5.and.num_cut>0)then
          read (lu_ast) ll !single solu of multiple cutting planes
          read (lu_ast) (l1,l2,l3,i=1,ll)
          do m=1,ll
          read (lu_ast) fsmach,alpha,rel,time
          select case(cut_dir)
             case(1)
               read (lu_ast)
     &         ((( r(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         (((ru(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         (((rv(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         (((rw(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2),
     &         ((( e(id_planes(nd,i),j,k),i=m,m),j=j1,j2),k=k1,k2)
             case(2)
               read (lu_ast)
     &         ((( r(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         (((ru(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         (((rv(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         (((rw(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2),
     &         ((( e(i,id_planes(nd,j),k),i=i1,i2),j=m,m),k=k1,k2)
             case(3)
               read (lu_ast)
     &         ((( r(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         (((ru(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         (((rv(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         (((rw(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m),
     &         ((( e(i,j,id_planes(nd,k)),i=i1,i2),j=j1,j2),k=m,m)
           end select
           enddo
! plot3d
       else if(iout==3)then
          read (lu_ast) ll
          if(ll.ne.mq)write(lu_out,*)"num_soln_per_flow_file does 
     .       not match the solution file ",fname,mq,ll
          read (lu_ast) (l1,l2,l3,i=1,ll)
          read (lu_ast) fsmach,alpha,rel,time
          read (lu_ast)
     &            ((( r(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            (((ru(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            (((rv(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            (((rw(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2),
     &            ((( e(i,j,k),i=i1,i2),j=j1,j2),k=k1,k2)
! Vstage
       else if(iout==1)then
          i2p=i2+1
          j2p=j2+1
          k2p=k2+1
          ALLOCATE(tmpvstg(i2p-i1+1,j2p-j1+1,k2p-k1+1,9))
! This should allocate a temporary array and put the data into
! the correct indices....
          read(lu_ast)ncyct_sngl,time,
     &   (((vstg(i,j,k,6), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,1), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,2), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,3), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,4), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,5), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,7), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,8), i = i1,i2p), j =j1,j2p), k = k1,k2p),
     &   (((vstg(i,j,k,9), i = i1,i2p), j =j1,j2p), k = k1,k2p)
         
      endif
      close(lu_ast)
      end subroutine read_ast_q

!#######################################################################
      subroutine write_merged_grid
!#######################################################################
! Added support for vstage grids.
      use common_area
      implicit none
      integer :: i,j,k,dymmy
      character:: fname*40

      if(iout==5 .and. num_cut==0)then
         write(lu_out,*)"blade row:",my_id+1
         write(6,*)"num_cut_plane must not be zero when output_format=5"
         call dummy_flush(6)
         stop
      endif

          if(iout==3 .or. iout==1)then
            call get_merged_gridname3(my_row,my_psge,-1,fname)
          endif

          if(iout==5 .and. num_cut >0)then
               call get_merged_gridname5
     .              (my_row,my_psge,cut_dir,1,-1,fname)
          endif

          write(lu_out,*)my_id+1,"writing passage grid ",fname
          write(lu_out,*)
          call dummy_flush(6)
          open(lu_g,file=fname,status='unknown',form='unformatted')

          SELECT CASE(merge_output)
            CASE(1)
              WRITE(lu_g) vstg_hdr
              DO j=1,nj
                WRITE(lu_g)
     &            ((grd(i,j,k,1),i=1,ni),k=1,nk),
     &            ((grd(i,j,k,2),i=1,ni),k=1,nk),
     &            ((grd(i,j,k,3),i=1,ni),k=1,nk)
              ENDDO

            CASE DEFAULT
          write (lu_g) 1
          write (lu_g) ni,nj,nk
          write (lu_g)
     &            (((x(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((y(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((z(i,j,k),i=1,ni),j=1,nj),k=1,nk)
          END SELECT
           
       close(lu_g)
      end subroutine write_merged_grid

!#######################################################################
      subroutine write_merged_motion_grid(iter)
!#######################################################################
      use common_area
      implicit none
      integer,intent(in) :: iter
      integer :: i,j,k,dymmy
      character:: fname*40
      real*8 :: roll,angsft,tmp,thta,rad
! this routine is controled by motion_grid,dtdt,iter

      if(dtdt==0)return
     
      if(iout==3)then
         roll=time*dt_lr(1)
         if(msu_or_apnasa==2)roll=-roll
      else if(iout==1)then
         roll=time*dt_lr(2)
         if(msu_or_apnasa==1)roll=-roll
      endif

      angsft=mod(roll,pi2)

         do i=1,ni
           do j=1,nj
             do k=1,nk
                tmp=atan2(y(i,j,k),z(i,j,k))
                thta=mod(tmp+angsft,pi2)
                rad=sqrt(y(i,j,k)**2+z(i,j,k)**2)
                ym(i,j,k)=rad*sin(thta)
                zm(i,j,k)=rad*cos(thta)
             enddo
           enddo
         enddo

      if(iout==5 .and. num_cut==0)then
         write(lu_out,*)"blade row:",my_id+1
         write(6,*)"num_cut_plane must not be zero when output_format=5"
         call dummy_flush(6)
         stop
      endif

          if(iout==3 .or. iout==1)then
            call get_merged_gridname3(my_row,my_psge,iter,fname)
          endif

          if(iout==5 .and. num_cut >0)then
               call get_merged_gridname5
     .              (my_row,my_psge,cut_dir,1,iter,fname)
          endif

          write(lu_out,*)my_id+1,"writing passage grid",fname
          call dummy_flush(6)
          open(lu_g,file=fname,status='unknown',form='unformatted')

          write (lu_g) 1
          write (lu_g) ni,nj,nk
          write (lu_g)
     &            (((x(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((ym(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((zm(i,j,k),i=1,ni),j=1,nj),k=1,nk)

       close(lu_g)
      end subroutine write_merged_motion_grid
!#######################################################################
      subroutine cal_grd
!#######################################################################
      use common_area
      implicit none
      integer :: i,j,k
      real*4 :: thta

      do i=1,ni
        do j=1,nj
          do k=1,nk
             thta=atan2(y(i,j,k),z(i,j,k))
             grd(i,j,k,3)=thta
          enddo
        enddo
      enddo
      return
      end subroutine cal_grd 
!#######################################################################
      subroutine grd_to_xyz
!#######################################################################
      use common_area
      implicit none
      integer :: i,j,k
      real*4 :: rad,thta

! convert the cylindrial grid (grd) into Cartesian grid (xyz)
      do i=1,ni
        do j=1,nj
          do k=1,nk
             thta=grd(i,j,k,3)
             rad=grd(i,j,k,2)
             x(i,j,k)=grd(i,j,k,1)
             y(i,j,k)=sin(thta)*rad
             z(i,j,k)=cos(thta)*rad
          enddo
        enddo
      enddo
      return
      end subroutine grd_to_xyz

!#######################################################################
      subroutine xyz_toggle_LR
!#######################################################################
      use common_area
      implicit none
      integer :: i,j,k
      real*4 :: rad,thta

      do i=1,ni
        do j=1,nj
          do k=1,nk
             thta=atan2(y(i,j,k),z(i,j,k))
             rad=sqrt(y(i,j,k)**2+z(i,j,k)**2)
             y(i,j,k)=sin(-thta)*rad
             z(i,j,k)=cos(-thta)*rad
          enddo
        enddo
      enddo
      return
      end subroutine xyz_toggle_LR

!#######################################################################
      subroutine q_msu_to_apnasa
!#######################################################################
      use common_area
      implicit none
      integer :: i,j,k
      real*4 :: rqdq,p,yr,zr,rvcar,rwcar,rvr,rvth,thta

             ru(:,:,:)=ru(:,:,:)*sqrgam
             rv(:,:,:)=rv(:,:,:)*sqrgam
             rw(:,:,:)=rw(:,:,:)*sqrgam
              e(:,:,:)= e(:,:,:)*gam
      do i=1,ni
        do j=1,nj
          do k=1,nk
             rvcar=rv(i,j,k)
             rwcar=rw(i,j,k)
             thta=grd(i,j,k,3)
             rvr=rvcar*sin(thta)+rwcar*cos(thta)
             rvth=-(rvcar*cos(thta)-rwcar*sin(thta))
             thta=-grd(i,j,k,3)
             rvcar=rvr*sin(thta)+rvth*cos(thta)
             rwcar=rvr*cos(thta)-rvth*sin(thta)
             rv(i,j,k)=rvcar
             rw(i,j,k)=rwcar
          enddo
        enddo
      enddo
      return 
      end subroutine q_msu_to_apnasa

!#######################################################################
      subroutine q_apnasa_to_msu
!#######################################################################
      use common_area
      implicit none
      integer :: i,j,k
      real*4 :: rvr,rvth,thta

             ru(:,:,:)=ru(:,:,:)/sqrgam
             rv_rad(:,:,:)=rv_rad(:,:,:)/sqrgam
             rv_tan(:,:,:)=rv_tan(:,:,:)/sqrgam
              e(:,:,:)= e(:,:,:)/gam
      do i=1,ni
        do j=1,nj
          do k=1,nk
              rvr=rv_rad(i,j,k)
             rvth=-rv_tan(i,j,k)
             thta=-grd(i,j,k,3)
             rv(i,j,k)=rvr*sin(thta)+rvth*cos(thta)
             rw(i,j,k)=rvr*cos(thta)-rvth*sin(thta)
          enddo
        enddo
      enddo
      return
      end subroutine q_apnasa_to_msu

!#######################################################################
      subroutine vel_abs_to_rel
!#######################################################################
      use common_area
      implicit none
      character(len=40)::fname
      integer :: i,j,k
      real*4 :: rqdq,p,yr,zr

      do i=1,ni
        do j=1,nj
          do k=1,nk
            rqdq=0.5*(ru(i,j,k)**2+rv(i,j,k)**2+rw(i,j,k)**2)
     .            /r(i,j,k)
             p=gm1*(e(i,j,k)-rqdq)
             yr=r(i,j,k)*z(i,j,k)*dt_lr(msu_or_apnasa)
             zr=-r(i,j,k)*y(i,j,k)*dt_lr(msu_or_apnasa)
             rv(i,j,k)=rv(i,j,k)-yr
             rw(i,j,k)=rw(i,j,k)-zr
             rqdq=0.5*(ru(i,j,k)**2+rv(i,j,k)**2+rw(i,j,k)**2)
     .               /r(i,j,k)
             e(i,j,k)=(p/gm1)+rqdq
          enddo
        enddo
      enddo
      end subroutine vel_abs_to_rel

!#######################################################################
      subroutine write_merged_solu(ncyct)
!#######################################################################
! Write the merged solution for the flow file.  Vstage has been added,
! but it does not support rotating grid.  Obviously since the solution
! is in cylindrical coordinates, this is not a problem for the flow file.
      use common_area
      implicit none
      character(len=40)::fname
      integer,intent(in)::ncyct
      integer :: i,j,k
      real*8 :: thta,qnor,qtan
      
      call get_merged_qname(my_row,my_psge,ncyct,fname)
      write(lu_out,*)my_id+1,"writing passage solution ",fname
      write(lu_out,*)
      call dummy_flush(6)
      open(lu_q,file=fname,status='unknown',form='unformatted')
      rewind(lu_q) 

      SELECT CASE(merge_output)
        CASE(1) ! Vstage does not support moving grid yet/ever
          WRITE(lu_q) ncyct, time,
     &            ((( vstg(i,j,k,6),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,1),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,2),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,3),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,4),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,5),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,7),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,8),i=1,ni+1),j=1,nj+1),k=1,nk+1),
     &            ((( vstg(i,j,k,9),i=1,ni+1),j=1,nj+1),k=1,nk+1)

        CASE DEFAULT
          if(motion_grid .and. dtdt/=0.0)then
            do i=1,ni
              do j=1,nj
                do k=1,nk
                  thta=atan2(y(i,j,k),z(i,j,k))
                  qnor=rv(i,j,k)*sin(thta)+rw(i,j,k)*cos(thta)
                  qtan=rv(i,j,k)*cos(thta)-rw(i,j,k)*sin(thta)
                  thta=atan2(ym(i,j,k),zm(i,j,k))
                  rv(i,j,k)=qnor*sin(thta)+qtan*cos(thta)
                  rw(i,j,k)=qnor*cos(thta)-qtan*sin(thta)
                enddo
              enddo
            enddo
          endif

          write (lu_q) 1
          write (lu_q) ni,nj,nk
          write (lu_q) fsmach,alpha,rel,time
          write (lu_q)
     &            ((( r(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((ru(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((rv(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            (((rw(i,j,k),i=1,ni),j=1,nj),k=1,nk),
     &            ((( e(i,j,k),i=1,ni),j=1,nj),k=1,nk)
        END SELECT
          WRITE(lu_out,*) ' Completed file write'
          close(lu_q)
      end subroutine write_merged_solu

!#######################################################################
      subroutine get_merged_gridname3(lrow,lpsge,iter,fname)
!#######################################################################
      use common_area
      implicit none
      character   :: xch1*2,xch2,xch3*6,xch3v*13
      character(len=20)::xrow,xpsge,xiter
      character,intent(out):: fname*40
      integer,intent(in):: lrow,lpsge,iter
      integer     :: inum,ll,l1,l2,l3

        xch1='.r'
        xch2='b'
        xch3='.p3d.g'
        xch3v='.p3dv.g.r4'
        ll=len_trim(casename)

       call int_to_char(lrow,xrow,l1)
       call int_to_char(lpsge,xpsge,l2)
       SELECT CASE(merge_output)
         CASE(1)
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//
     &           xpsge(1:l2)//xch3v
         CASE DEFAULT
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//
     &           xpsge(1:l2)//xch3
       END SELECT

       if(iter>=0)then
           call int_to_char(iter,xiter,l3)        
         SELECT CASE(merge_output)
           CASE(1)
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     &                         //xch3v//xiter(1:l3)
           CASE DEFAULT
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     &                         //xch3//xiter(1:l3)
         END SELECT
       endif
       end subroutine get_merged_gridname3

!#######################################################################
      subroutine get_merged_qname(lrow,lpsge,iter,fname)
!#######################################################################
      use common_area
      implicit none
      character   :: xch1*2,xch2,xch3*6,xch3v*7,xch5
      character(len=20)::xrow,xpsge,xiter
      character,intent(out):: fname*40
      integer,intent(in):: lrow,lpsge,iter
      integer     :: inum,ll,l1,l2,l3,lext

        xch1='.r'
        xch2='b'
        xch3='.p3d.q'
        xch3v='.p3dv.q'
        xch5='.'
        ll=len_trim(casename)
        lext=len_trim(qchar_ext)
        call int_to_char(lrow,xrow,l1)
        call int_to_char(lpsge,xpsge,l2)
        call int_to_char(iter,xiter,l3)
 
       if(.not. has_qchar_ext) then
         SELECT CASE(merge_output)
           CASE(1)
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2
     .                     //xpsge(1:l2)//xch3v//xiter(1:l3)//'.1.r4'
           CASE DEFAULT
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2
     .                     //xpsge(1:l2)//xch3//xiter(1:l3)
         END SELECT
       else if(has_qchar_ext)then
         SELECT CASE(merge_output)
           CASE(1)
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     .                         //xch3v//qchar_ext(1:lext)//'.1.r4'
           CASE DEFAULT
           fname=casename(1:ll)//xch1//xrow(1:l1)//xch2//xpsge(1:l2)
     .                         //xch3//qchar_ext(1:lext)
         END SELECT
       endif

       end subroutine get_merged_qname

!#######################################################################
      subroutine dummy_flush(unit)
!#######################################################################
      implicit none

      integer unit

#ifndef IBM
      call flush(unit)
#endif

#ifdef IBM
      call flush_(unit)
#endif
      return
      end subroutine dummy_flush

!#######################################################################
      subroutine q_ctr_cyl_to_node_car
!#######################################################################
      use common_area
      implicit none

! map q from cell center to node
! convert vector from cylindric to Cartensian
 
      integer i,j,k,i1,j1,k1,ni1,nj1,nk1
      integer jst,kst,jnd,knd,l
      !real*4,dimension(ni,nj,nk) :: ra,rua,rva,rwa,ea
      !real*4,dimension(4,ni+1,nj+1,nk+1) :: q
      !real*4,dimension(ni+1,nj+1,nk+1) :: p
      real*4,pointer,dimension(:,:,:) :: ra,rua,rva,rwa,ea
      real*4,pointer,dimension(:,:,:,:) :: q
      real*4,pointer,dimension(:,:,:) :: p
      
      real*8 :: fake_rgm1,rqdq,pa,rvr,rvth,thta

      SELECT CASE(merge_output)
        CASE(1)
          RETURN
      END SELECT

      ALLOCATE(ra(ni,nj,nk),rua(ni,nj,nk),rva(ni,nj,nk),rwa(ni,nj,nk),
     .         ea(ni,nj,nk))
      ALLOCATE(q(4,ni+1,nj+1,nk+1),p(ni+1,nj+1,nk+1))
      
      fake_rgm1=2.5

      ni1=ni+1
      nj1=nj+1
      nk1=nk+1
      do 10 i=1,ni1
       do 10 j=1,nj1
        do 10 k=1,nk1
                p(i,j,k)=vstg(i,j,k,6)
              q(1,i,j,k)=vstg(i,j,k,1)
              q(2,i,j,k)=vstg(i,j,k,2)
              q(3,i,j,k)=vstg(i,j,k,3)
              q(4,i,j,k)=vstg(i,j,k,4)
 10   continue

            do 312  i = 1, ni
         i1  = i+1
         jst = 1
         jnd = nj
         if ( i .eq. 1 .or. i .eq. ni ) then
            jst = 2
            jnd = nj-1
         endif
         do 318  j = jst, jnd
            j1  = j+1
            kst = 1
            knd = nk
            if (i.eq.1.or.i.eq.ni.or.j.eq.1.or.j.eq.nj) then
               kst = 2
               knd = nk -1
            endif
            do 311  k = kst, knd
               k1 = k+1
               ra(i,j,k)=(q(1,i,j,k)+q(1,i1,j,k)+q(1,i,j1,k)+
     &            q(1,i1,j1,k)+q(1,i,j,k1)+q(1,i1,j,k1)+q(1,i,j1,k1)+
     &            q(1,i1,j1,k1))*.125
               rua(i,j,k)=(q(2,i,j,k)+q(2,i1,j,k)+q(2,i,j1,k)+
     &            q(2,i1,j1,k)+q(2,i,j,k1)+q(2,i1,j,k1)+q(2,i,j1,k1)+
     &            q(2,i1,j1,k1))*.125
               rva(i,j,k)=(q(3,i,j,k)+q(3,i1,j,k)+q(3,i,j1,k)+
     &            q(3,i1,j1,k)+q(3,i,j,k1)+q(3,i1,j,k1)+q(3,i,j1,k1)+
     &            q(3,i1,j1,k1))*.125
               rwa(i,j,k)=(q(4,i,j,k)+q(4,i1,j,k)+q(4,i,j1,k)+
     &            q(4,i1,j1,k)+q(4,i,j,k1)+q(4,i1,j,k1)+q(4,i,j1,k1)+
     &            q(4,i1,j1,k1))*.125
               pa=(p(i,j,k)+p(i1,j,k)+p(i,j1,k)+p(i1,j1,k)
     &            +p(i,j,k1)+p(i1,j,k1)+p(i,j1,k1)+p(i1,j1,k1))*.125
               rqdq=0.5*(rua(i,j,k)**2+rva(i,j,k)**2+rwa(i,j,k)**2)/
     &            ra(i,j,k)
               ea(i,j,k)=fake_rgm1*pa+rqdq
 311  continue
 318  continue
 312  continue

c...edges
c....k edges
      do 313  k = 2, nk-1
         k1 = k+1
c.....edge (i=1,j=1)
         ra(1,1,k)=(q(1,2,1,k)+q(1,1,2,k)+q(1,2,2,k)
     &      +q(1,2,1,k1)+q(1,1,2,k1)+q(1,2,2,k1))/6.
         rua(1,1,k)=(q(2,2,1,k)+q(2,1,2,k)+q(2,2,2,k)
     &      +q(2,2,1,k1)+q(2,1,2,k1)+q(2,2,2,k1))/6.
         rva(1,1,k)=(q(3,2,1,k)+q(3,1,2,k)+q(3,2,2,k)
     &      +q(3,2,1,k1)+q(3,1,2,k1)+q(3,2,2,k1))/6.
         rwa(1,1,k)=(q(4,2,1,k)+q(4,1,2,k)+q(4,2,2,k)
     &      +q(4,2,1,k1)+q(4,1,2,k1)+q(4,2,2,k1))/6.
          pa=(p(2,1,k)+p(1,2,k)+p(2,2,k)
     &      +p(2,1,k1)+p(1,2,k1)+p(2,2,k1))/6.
         rqdq=0.5*(rua(1,1,k)**2+rva(1,1,k)**2+rwa(1,1,k)**2)/
     &            ra(1,1,k)
          ea(1,1,k)=fake_rgm1*pa+rqdq
c.....edge (i=1,j=nj)
         ra(1,nj,k)=(q(1,1,nj,k)+q(1,2,nj,k)+q(1,2,nj1,k)
     &      +q(1,1,nj,k1)+q(1,2,nj,k1)+q(1,2,nj1,k1))/6.
         rua(1,nj,k)=(q(2,1,nj,k)+q(2,2,nj,k)+q(2,2,nj1,k)
     &      +q(2,1,nj,k1)+q(2,2,nj,k1)+q(2,2,nj1,k1))/6.
         rva(1,nj,k)=(q(3,1,nj,k)+q(3,2,nj,k)+q(3,2,nj1,k)
     &      +q(3,1,nj,k1)+q(3,2,nj,k1)+q(3,2,nj1,k1))/6.
         rwa(1,nj,k)=(q(4,1,nj,k)+q(4,2,nj,k)+q(4,2,nj1,k)
     &      +q(4,1,nj,k1)+q(4,2,nj,k1)+q(4,2,nj1,k1))/6.
          pa=(p(1,nj,k)+p(2,nj,k)+p(2,nj1,k)
     &      +p(1,nj,k1)+p(2,nj,k1)+p(2,nj1,k1))/6.
         rqdq=0.5*(rua(1,nj,k)**2+rva(1,nj,k)**2+rwa(1,nj,k)**2)/
     &            ra(1,nj,k)
          ea(1,nj,k)=fake_rgm1*pa+rqdq
c.....edge (i=ni,j=1)
         ra(ni,1,k)=(q(1,ni,1,k)+q(1,ni,2,k)+q(1,ni1,2,k)
     &      +q(1,ni,1,k1)+q(1,ni,2,k1)+q(1,ni1,2,k1))/6.
         rua(ni,1,k)=(q(2,ni,1,k)+q(2,ni,2,k)+q(2,ni1,2,k)
     &      +q(2,ni,1,k1)+q(2,ni,2,k1)+q(2,ni1,2,k1))/6.
         rva(ni,1,k)=(q(3,ni,1,k)+q(3,ni,2,k)+q(3,ni1,2,k)
     &      +q(3,ni,1,k1)+q(3,ni,2,k1)+q(3,ni1,2,k1))/6.
         rwa(ni,1,k)=(q(4,ni,1,k)+q(4,ni,2,k)+q(4,ni1,2,k)
     &      +q(4,ni,1,k1)+q(4,ni,2,k1)+q(4,ni1,2,k1))/6.
          pa=(p(ni,1,k)+p(ni,2,k)+p(ni1,2,k)
     &      +p(ni,1,k1)+p(ni,2,k1)+p(ni1,2,k1))/6.
         rqdq=0.5*(rua(ni,1,k)**2+rva(ni,1,k)**2+rwa(ni,1,k)**2)/
     &            ra(ni,1,k)
          ea(ni,1,k)=fake_rgm1*pa+rqdq
c.....edge (i=ni,j=nj)
         ra(ni,nj,k)=(q(1,ni,nj,k)+q(1,ni,nj1,k)+q(1,ni1,nj,k)
     &      +q(1,ni,nj,k1)+q(1,ni,nj1,k1)+q(1,ni1,nj,k1))/6.
         rua(ni,nj,k)=(q(2,ni,nj,k)+q(2,ni,nj1,k)+q(2,ni1,nj,k)
     &      +q(2,ni,nj,k1)+q(2,ni,nj1,k1)+q(2,ni1,nj,k1))/6.
         rva(ni,nj,k)=(q(3,ni,nj,k)+q(3,ni,nj1,k)+q(3,ni1,nj,k)
     &      +q(3,ni,nj,k1)+q(3,ni,nj1,k1)+q(3,ni1,nj,k1))/6.
         rwa(ni,nj,k)=(q(4,ni,nj,k)+q(4,ni,nj1,k)+q(4,ni1,nj,k)
     &      +q(4,ni,nj,k1)+q(4,ni,nj1,k1)+q(4,ni1,nj,k1))/6.
          pa=(p(ni,nj,k)+p(ni,nj1,k)+p(ni1,nj,k)
     &      +p(ni,nj,k1)+p(ni,nj1,k1)+p(ni1,nj,k1))/6.
         rqdq=0.5*(rua(ni,nj,k)**2+rva(ni,nj,k)**2+rwa(ni,nj,k)**2)/
     &            ra(ni,nj,k)
          ea(ni,nj,k)=fake_rgm1*pa+rqdq
  313 continue
c
c....j edges
c....j edges
      do 314  j = 2, nj-1
         j1 = j+1
c.....edge (i=1,k=1)
         ra(1,j,1)=(q(1,1,j,2)+q(1,2,j,1)+q(1,2,j,2)
     &      +q(1,1,j1,2)+q(1,2,j1,1)+q(1,2,j1,2))/6.
         rua(1,j,1)=(q(2,1,j,2)+q(2,2,j,1)+q(2,2,j,2)
     &      +q(2,1,j1,2)+q(2,2,j1,1)+q(2,2,j1,2))/6.
         rva(1,j,1)=(q(3,1,j,2)+q(3,2,j,1)+q(3,2,j,2)
     &      +q(3,1,j1,2)+q(3,2,j1,1)+q(3,2,j1,2))/6.
         rwa(1,j,1)=(q(4,1,j,2)+q(4,2,j,1)+q(4,2,j,2)
     &      +q(4,1,j1,2)+q(4,2,j1,1)+q(4,2,j1,2))/6.
          pa=(p(1,j,2)+p(2,j,1)+p(2,j,2)
     &      +p(1,j1,2)+p(2,j1,1)+p(2,j1,2))/6.
         rqdq=0.5*(rua(1,j,1)**2+rva(1,j,1)**2+rwa(1,j,1)**2)/
     &            ra(1,j,1)
          ea(1,j,1)=fake_rgm1*pa+rqdq
c.....edge (i=1,k=nk)
         ra(1,j,nk)=(q(1,1,j,nk)+q(1,2,j,nk)+q(1,2,j,nk1)
     &      +q(1,1,j1,nk)+q(1,2,j1,nk)+q(1,2,j1,nk1))/6.
         rua(1,j,nk)=(q(2,1,j,nk)+q(2,2,j,nk)+q(2,2,j,nk1)
     &      +q(2,1,j1,nk)+q(2,2,j1,nk)+q(2,2,j1,nk1))/6.
         rva(1,j,nk)=(q(3,1,j,nk)+q(3,2,j,nk)+q(3,2,j,nk1)
     &      +q(3,1,j1,nk)+q(3,2,j1,nk)+q(3,2,j1,nk1))/6.
         rwa(1,j,nk)=(q(4,1,j,nk)+q(4,2,j,nk)+q(4,2,j,nk1)
     &      +q(4,1,j1,nk)+q(4,2,j1,nk)+q(4,2,j1,nk1))/6.
          pa=(p(1,j,nk)+p(2,j,nk)+p(2,j,nk1)
     &      +p(1,j1,nk)+p(2,j1,nk)+p(2,j1,nk1))/6.
         rqdq=0.5*(rua(1,j,nk)**2+rva(1,j,nk)**2+rwa(1,j,nk)**2)/
     &            ra(1,j,nk)
          ea(1,j,nk)=fake_rgm1*pa+rqdq
c.....edge (i=ni,k=1)
         ra(ni,j,1)=(q(1,ni,j,1)+q(1,ni,j,2)+q(1,ni1,j,2)
     &      +q(1,ni,j1,1)+q(1,ni,j1,2)+q(1,ni1,j1,2))/6.
         rua(ni,j,1)=(q(2,ni,j,1)+q(2,ni,j,2)+q(2,ni1,j,2)
     &      +q(2,ni,j1,1)+q(2,ni,j1,2)+q(2,ni1,j1,2))/6.
         rva(ni,j,1)=(q(3,ni,j,1)+q(3,ni,j,2)+q(3,ni1,j,2)
     &      +q(3,ni,j1,1)+q(3,ni,j1,2)+q(3,ni1,j1,2))/6.
         rwa(ni,j,1)=(q(4,ni,j,1)+q(4,ni,j,2)+q(4,ni1,j,2)
     &      +q(4,ni,j1,1)+q(4,ni,j1,2)+q(4,ni1,j1,2))/6.
          pa=(p(ni,j,1)+p(ni,j,2)+p(ni1,j,2)
     &      +p(ni,j1,1)+p(ni,j1,2)+p(ni1,j1,2))/6.
         rqdq=0.5*(rua(ni,j,1)**2+rva(ni,j,1)**2+rwa(ni,j,1)**2)/
     &            ra(ni,j,1)
          ea(ni,j,1)=fake_rgm1*pa+rqdq

c.....edge (i=ni1,k=nk)
         ra(ni,j,nk)=(q(1,ni,j,nk)+q(1,ni,j,nk1)+q(1,ni1,j,nk)
     &      +q(1,ni,j1,nk)+q(1,ni,j1,nk1)+q(1,ni1,j1,nk))/6.
         rua(ni,j,nk)=(q(2,ni,j,nk)+q(2,ni,j,nk1)+q(2,ni1,j,nk)
     &      +q(2,ni,j1,nk)+q(2,ni,j1,nk1)+q(2,ni1,j1,nk))/6.
         rva(ni,j,nk)=(q(3,ni,j,nk)+q(3,ni,j,nk1)+q(3,ni1,j,nk)
     &      +q(3,ni,j1,nk)+q(3,ni,j1,nk1)+q(3,ni1,j1,nk))/6.
         rwa(ni,j,nk)=(q(4,ni,j,nk)+q(4,ni,j,nk1)+q(4,ni1,j,nk)
     &      +q(4,ni,j1,nk)+q(4,ni,j1,nk1)+q(4,ni1,j1,nk))/6.
          pa=(p(ni,j,nk)+p(ni,j,nk1)+p(ni1,j,nk)
     &      +p(ni,j1,nk)+p(ni,j1,nk1)+p(ni1,j1,nk))/6.
         rqdq=0.5*(rua(ni,j,nk)**2+rva(ni,j,nk)**2+rwa(ni,j,nk)**2)/
     &            ra(ni,j,nk)
          ea(ni,j,nk)=fake_rgm1*pa+rqdq
  314 continue
c
c....i edges
      do 315  i = 2, ni-1
         i1 = i+1
c.....edge (j=1,k=1)
         ra(i,1,1)=(q(1,i,1,2)+q(1,i,2,1)+q(1,i,2,2)
     &      +q(1,i1,1,2)+q(1,i1,2,1)+q(1,i1,2,2))/6.
         rua(i,1,1)=(q(2,i,1,2)+q(2,i,2,1)+q(2,i,2,2)
     &      +q(2,i1,1,2)+q(2,i1,2,1)+q(2,i1,2,2))/6.
         rva(i,1,1)=(q(3,i,1,2)+q(3,i,2,1)+q(3,i,2,2)
     &      +q(3,i1,1,2)+q(3,i1,2,1)+q(3,i1,2,2))/6.
         rwa(i,1,1)=(q(4,i,1,2)+q(4,i,2,1)+q(4,i,2,2)
     &      +q(4,i1,1,2)+q(4,i1,2,1)+q(4,i1,2,2))/6.
          pa=(p(i,1,2)+p(i,2,1)+p(i,2,2)
     &      +p(i1,1,2)+p(i1,2,1)+p(i1,2,2))/6.
         rqdq=0.5*(rua(i,1,1)**2+rva(i,1,1)**2+rwa(i,1,1)**2)/
     &            ra(i,1,1)
          ea(i,1,1)=fake_rgm1*pa+rqdq
!.....edge (j=1,k=nk)
         ra(i,1,nk)=(q(1,i,1,nk)+q(1,i,2,nk)+q(1,i,2,nk1)
     &      +q(1,i1,1,nk)+q(1,i1,2,nk)+q(1,i1,2,nk1))/6.
         rua(i,1,nk)=(q(2,i,1,nk)+q(2,i,2,nk)+q(2,i,2,nk1)
     &      +q(2,i1,1,nk)+q(2,i1,2,nk)+q(2,i1,2,nk1))/6.
         rva(i,1,nk)=(q(3,i,1,nk)+q(3,i,2,nk)+q(3,i,2,nk1)
     &      +q(3,i1,1,nk)+q(3,i1,2,nk)+q(3,i1,2,nk1))/6.
         rwa(i,1,nk)=(q(4,i,1,nk)+q(4,i,2,nk)+q(4,i,2,nk1)
     &      +q(4,i1,1,nk)+q(4,i1,2,nk)+q(4,i1,2,nk1))/6.
          pa=(p(i,1,nk)+p(i,2,nk)+p(i,2,nk1)
     &      +p(i1,1,nk)+p(i1,2,nk)+p(i1,2,nk1))/6.
         rqdq=0.5*(rua(i,1,nk)**2+rva(i,1,nk)**2+rwa(i,1,nk)**2)/
     &            ra(i,1,nk)
          ea(i,1,nk)=fake_rgm1*pa+rqdq
!.....edge (j=nj,k=1)
         ra(i,nj,1)=(q(1,i,nj,1)+q(1,i,nj,2)+q(1,i,nj1,2)
     &      +q(1,i1,nj,1)+q(1,i1,nj,2)+q(1,i1,nj1,2))/6.
         rua(i,nj,1)=(q(2,i,nj,1)+q(2,i,nj,2)+q(2,i,nj1,2)
     &      +q(2,i1,nj,1)+q(2,i1,nj,2)+q(2,i1,nj1,2))/6.
         rva(i,nj,1)=(q(3,i,nj,1)+q(3,i,nj,2)+q(3,i,nj1,2)
     &      +q(3,i1,nj,1)+q(3,i1,nj,2)+q(3,i1,nj1,2))/6.
         rwa(i,nj,1)=(q(4,i,nj,1)+q(4,i,nj,2)+q(4,i,nj1,2)
     &      +q(4,i1,nj,1)+q(4,i1,nj,2)+q(4,i1,nj1,2))/6.
         pa=(p(i,nj,1)+p(i,nj,2)+p(i,nj1,2)
     &      +p(i1,nj,1)+p(i1,nj,2)+p(i1,nj1,2))/6.
         rqdq=0.5*(rua(i,nj,1)**2+rva(i,nj,1)**2+rwa(i,nj,1)**2)/
     &            ra(i,nj,1)
          ea(i,nj,1)=fake_rgm1*pa+rqdq
!.....edge (j=nj,k=nk)

         ra(i,nj,nk)=(q(1,i,nj,nk)+q(1,i,nj,nk1)+q(1,i,nj1,nk)
     &      +q(1,i1,nj,nk)+q(1,i1,nj,nk1)+q(1,i1,nj1,nk))/6.
         rua(i,nj,nk)=(q(2,i,nj,nk)+q(2,i,nj,nk1)+q(2,i,nj1,nk)
     &      +q(2,i1,nj,nk)+q(2,i1,nj,nk1)+q(2,i1,nj1,nk))/6.
         rva(i,nj,nk)=(q(3,i,nj,nk)+q(3,i,nj,nk1)+q(3,i,nj1,nk)
     &      +q(3,i1,nj,nk)+q(3,i1,nj,nk1)+q(3,i1,nj1,nk))/6.
         rwa(i,nj,nk)=(q(4,i,nj,nk)+q(4,i,nj,nk1)+q(4,i,nj1,nk)
     &      +q(4,i1,nj,nk)+q(4,i1,nj,nk1)+q(4,i1,nj1,nk))/6.
         pa=(p(i,nj,nk)+p(i,nj,nk1)+p(i,nj1,nk)
     &      +p(i1,nj,nk)+p(i1,nj,nk1)+p(i1,nj1,nk))/6.
         rqdq=0.5*(rua(i,nj,nk)**2+rva(i,nj,nk)**2+rwa(i,nj,nk)**2)/
     &            ra(i,nj,nk)
          ea(i,nj,nk)=fake_rgm1*pa+rqdq

  315 continue
c...corners
c.....corner ( 1, 1, 1)
      ra( 1, 1, 1)=(q(1, 1, 2, 2)+q(1, 2, 2, 1)
     &   +q(1, 2, 1, 2)+q(1, 2, 2, 2))*.25
      rua( 1, 1, 1)=(q(2, 1, 2, 2)+q(2, 2, 2, 1)
     &   +q(2, 2, 1, 2)+q(2, 2, 2, 2))*.25
      rva( 1, 1, 1)=(q(3, 1, 2, 2)+q(3, 2, 2, 1)
     &   +q(3, 2, 1, 2)+q(3, 2, 2, 2))*.25
      rwa( 1, 1, 1)=(q(4, 1, 2, 2)+q(4, 2, 2, 1)
     &   +q(4, 2, 1, 2)+q(4, 2, 2, 2))*.25
      pa=(p( 1, 2, 2)+p( 2, 2, 1)
     &   +p( 2, 1, 2)+p( 2, 2, 2))*.25
      rqdq=0.5*(rua(1,1,1)**2+rva(1,1,1)**2+rwa(1,1,1)**2)/
     &           ra(1,1,1)
      ea( 1, 1, 1)=fake_rgm1*pa+rqdq
c.....corner ( 1, 1,nk)
      ra( 1, 1,nk)=(q(1, 1, 2,nk)+q(1, 2, 2,nk)
     &   +q(1, 2, 1,nk)+q(1, 2, 2,nk1))*.25
      rua( 1, 1,nk)=(q(2, 1, 2,nk)+q(2, 2, 2,nk)
     &   +q(2, 2, 1,nk)+q(2, 2, 2,nk1))*.25
      rva( 1, 1,nk)=(q(3, 1, 2,nk)+q(3, 2, 2,nk)
     &   +q(3, 2, 1,nk)+q(3, 2, 2,nk1))*.25
      rwa( 1, 1,nk)=(q(4, 1, 2,nk)+q(4, 2, 2,nk)
     &   +q(4, 2, 1,nk)+q(4, 2, 2,nk1))*.25
      pa=(p( 1, 2,nk)+p( 2, 2,nk)
     &   +p( 2, 1,nk)+p( 2, 2,nk1))*.25
      rqdq=0.5*(rua(1,1,nk)**2+rva(1,1,nk)**2+rwa(1,1,nk)**2)/
     &           ra(1,1,nk)
      ea( 1, 1,nk)=fake_rgm1*pa+rqdq
c.....corner ( 1,nj, 1)
      ra( 1,nj, 1)=(q(1, 1,nj, 2)+q(1, 2,nj, 1)
     &   +q(1, 2,nj, 2)+q(1, 2,nj1, 2))*.25
      rua( 1,nj, 1)=(q(2, 1,nj, 2)+q(2, 2,nj, 1)
     &   +q(2, 2,nj, 2)+q(2, 2,nj1, 2))*.25
      rva( 1,nj, 1)=(q(3, 1,nj, 2)+q(3, 2,nj, 1)
     &   +q(3, 2,nj, 2)+q(3, 2,nj1, 2))*.25
      rwa( 1,nj, 1)=(q(4, 1,nj, 2)+q(4, 2,nj, 1)
     &   +q(4, 2,nj, 2)+q(4, 2,nj1, 2))*.25
      pa=(p( 1,nj, 2)+p( 2,nj, 1)
     &   +p( 2,nj, 2)+p( 2,nj1, 2))*.25
      rqdq=0.5*(rua(1,nj,1)**2+rva(1,nj,1)**2+rwa(1,nj,1)**2)/
     &           ra(1,nj,1)
      ea( 1,nj, 1)=fake_rgm1*pa+rqdq

c.....corner ( 1,nj,nk)
      ra( 1,nj,nk)=(q(1, 1,nj,nk)+q(1, 2,nj,nk)
     &   +q(1, 2,nj,nk1)+q(1, 2,nj1,nk))*.25
      rua( 1,nj,nk)=(q(2, 1,nj,nk)+q(2, 2,nj,nk)
     &   +q(2, 2,nj,nk1)+q(2, 2,nj1,nk))*.25
      rva( 1,nj,nk)=(q(3, 1,nj,nk)+q(3, 2,nj,nk)
     &   +q(3, 2,nj,nk1)+q(3, 2,nj1,nk))*.25
      rwa( 1,nj,nk)=(q(4, 1,nj,nk)+q(4, 2,nj,nk)
     &   +q(4, 2,nj,nk1)+q(4, 2,nj1,nk))*.25
      pa=(p( 1,nj,nk)+p( 2,nj,nk)
     &   +p( 2,nj,nk1)+p( 2,nj1,nk))*.25
      rqdq=0.5*(rua(1,nj,nk)**2+rva(1,nj,nk)**2+rwa(1,nj,nk)**2)/
     &           ra(1,nj,nk)
      ea( 1,nj,nk)=fake_rgm1*pa+rqdq
c.....corner (ni, 1, 1)
      ra(ni, 1, 1)=(q(1,ni, 2, 1)+q(1,ni, 1, 2)
     &   +q(1,ni, 2, 2)+q(1,ni1, 2, 2))*.25
      rua(ni, 1, 1)=(q(2,ni, 2, 1)+q(2,ni, 1, 2)
     &   +q(2,ni, 2, 2)+q(2,ni1, 2, 2))*.25
      rva(ni, 1, 1)=(q(3,ni, 2, 1)+q(3,ni, 1, 2)
     &   +q(3,ni, 2, 2)+q(3,ni1, 2, 2))*.25
      rwa(ni, 1, 1)=(q(4,ni, 2, 1)+q(4,ni, 1, 2)
     &   +q(4,ni, 2, 2)+q(4,ni1, 2, 2))*.25
      pa=(p(ni, 2, 1)+p(ni, 1, 2)
     &   +p(ni, 2, 2)+p(ni1, 2, 2))*.25
      rqdq=0.5*(rua(ni,1,1)**2+rva(ni,1,1)**2+rwa(ni,1,1)**2)/
     &           ra(ni,1,1)
      ea(ni, 1, 1)=fake_rgm1*pa+rqdq
c.....corner (ni, 1,nk)
      ra(ni, 1,nk)=(q(1,ni, 1,nk)+q(1,ni, 2,nk)
     &   +q(1,ni, 2,nk1)+q(1,ni1, 2,nk))*.25
      rua(ni, 1,nk)=(q(2,ni, 1,nk)+q(2,ni, 2,nk)
     &   +q(2,ni, 2,nk1)+q(2,ni1, 2,nk))*.25
      rva(ni, 1,nk)=(q(3,ni, 1,nk)+q(3,ni, 2,nk)
     &   +q(3,ni, 2,nk1)+q(3,ni1, 2,nk))*.25
      rwa(ni, 1,nk)=(q(4,ni, 1,nk)+q(4,ni, 2,nk)
     &   +q(4,ni, 2,nk1)+q(4,ni1, 2,nk))*.25
      pa=(p(ni, 1,nk)+p(ni, 2,nk)
     &   +p(ni, 2,nk1)+p(ni1, 2,nk))*.25
      rqdq=0.5*(rua(ni,1,nk)**2+rva(ni,1,nk)**2+rwa(ni,1,nk)**2)/
     &           ra(ni,1,nk)
      ea(ni, 1,nk)=fake_rgm1*pa+rqdq
c.....corner (ni,nj, 1)
      ra(ni,nj, 1)=(q(1,ni,nj, 1)+q(1,ni,nj, 2)
     &   +q(1,ni,nj1, 2)+q(1,ni1,nj, 2))*.25
      rua(ni,nj, 1)=(q(2,ni,nj, 1)+q(2,ni,nj, 2)
     &   +q(2,ni,nj1, 2)+q(2,ni1,nj, 2))*.25
      rva(ni,nj, 1)=(q(3,ni,nj, 1)+q(3,ni,nj, 2)
     &   +q(3,ni,nj1, 2)+q(3,ni1,nj, 2))*.25
      rwa(ni,nj, 1)=(q(4,ni,nj, 1)+q(4,ni,nj, 2)
     &   +q(4,ni,nj1, 2)+q(4,ni1,nj, 2))*.25
      pa=(p(ni,nj, 1)+p(ni,nj, 2)
     &   +p(ni,nj1, 2)+p(ni1,nj, 2))*.25
      rqdq=0.5*(rua(ni,nj,1)**2+rva(ni,nj,1)**2+rwa(ni,nj,1)**2)/
     &           ra(ni,nj,1)
      ea(ni,nj, 1)=fake_rgm1*pa+rqdq
c.....corner (ni,nj,nk)
      ra(ni,nj,nk)=(q(1,ni,nj,nk)+q(1,ni,nj1,nk)
     &   +q(1,ni1,nj,nk)+q(1,ni,nj,nk1))*.25
      rua(ni,nj,nk)=(q(2,ni,nj,nk)+q(2,ni,nj1,nk)
     &   +q(2,ni1,nj,nk)+q(2,ni,nj,nk1))*.25
      rva(ni,nj,nk)=(q(3,ni,nj,nk)+q(3,ni,nj1,nk)
     &   +q(3,ni1,nj,nk)+q(3,ni,nj,nk1))*.25
      rwa(ni,nj,nk)=(q(4,ni,nj,nk)+q(4,ni,nj1,nk)
     &   +q(4,ni1,nj,nk)+q(4,ni,nj,nk1))*.25
      pa=(p(ni,nj,nk)+p(ni,nj1,nk)
     &   +p(ni1,nj,nk)+p(ni,nj,nk1))*.25
      rqdq=0.5*(rua(ni,nj,nk)**2+rva(ni,nj,nk)**2+rwa(ni,nj,nk)**2)/
     &           ra(ni,nj,nk)
      ea(ni,nj,nk)=fake_rgm1*pa+rqdq

      do 13 i=1,ni
         do 13 j=1,nj
          do 13 k=1,nk
             r(i,j,k)= ra(i,j,k)
             ru(i,j,k)=rua(i,j,k)

!solu still in Left-handed grid, do not use y,z to cal. thta 
!because y,z could be toggled 
              thta=grd(i,j,k,3) !do not use y,z to cal thta
              rvr=rva(i,j,k)
              rvth=rwa(i,j,k)
              rv(i,j,k)=rvr*sin(thta)+rvth*cos(thta)
              rw(i,j,k)=rvr*cos(thta)-rvth*sin(thta)
              e(i,j,k)= ea(i,j,k)

              rv_rad(i,j,k)=rvr
              rv_tan(i,j,k)=rvth
 13   continue

      DEALLOCATE(ra,rua,rva,rwa,ea,p,q)
         
      return
      end subroutine q_ctr_cyl_to_node_car

!#######################################################################
      SUBROUTINE GetVstageHeader()
!#######################################################################
! This subroutine is used to read the header from the original Vstage
! grid.  This allows a quick determination of the vstage values without
! the need to parse a bunch of files in C.
        USE common_area
        IMPLICIT NONE
        INTEGER :: ll,l1
        CHARACTER(LEN=20) :: xrow
        CHARACTER(LEN=40) :: fname

        SELECT CASE(merge_output)
          CASE(1)
            ll=LEN_TRIM(casename)
            CALL int_to_char(my_row,xrow,l1)
            fname=casename(1:ll)//'.r'//xrow(1:l1)//'b1.vstage.g.r4'
            OPEN(UNIT=lu_vstg,FILE=fname,FORM='unformatted')
            READ(lu_vstg) vstg_hdr
            WRITE(lu_out,*) ' VSTG_HDR= ',vstg_hdr
            CLOSE(lu_vstg)
        END SELECT

        RETURN
      END SUBROUTINE
