MODULE general
  IMPLICIT NONE

  REAL, PARAMETER :: pi=3.141592653589793

  INTEGER :: iwinsizes(9)
  INTEGER :: iblck, isurf

  LOGICAL, ALLOCATABLE, DIMENSION(:) :: trueint
  INTEGER,POINTER,DIMENSION(:,:) :: eqpairnodes
  INTEGER :: eqpairs

  CHARACTER(LEN=20), ALLOCATABLE, DIMENSION(:) :: surfnames

  TYPE patchdata
    INTEGER,POINTER,DIMENSION(:,:) :: ptch ! Patch number, number of integer stored
    CHARACTER(LEN=20),POINTER,DIMENSION(:) :: nam
    INTEGER :: nptch
    LOGICAL,POINTER,DIMENSION(:) :: used
  END TYPE patchdata

  TYPE(patchdata),POINTER,DIMENSION(:) :: patches


CONTAINS

!#######################################################################
  REAL FUNCTION AVGV(X,imax,jmax,kmax,i,j,k)
!#######################################################################
! This function finds the average of the vertex array X for the 8 vertices
! defined by the coordinates (i,j,k)-(i+1,j+1,k+1).

  INTEGER, INTENT(IN) :: imax,jmax,kmax,i,j,k
  REAL, INTENT(IN) :: X(imax,jmax,kmax)

  AVGV = X(i,j,k)
  AVGV = AVGV + X(i+1,j  ,k)
  AVGV = AVGV + X(i+1,j+1,k)
  AVGV = AVGV + X(i  ,j+1,k)
  AVGV = AVGV + X(i  ,j  ,k+1)
  AVGV = AVGV + X(i+1,j  ,k+1)
  AVGV = AVGV + X(i+1,j+1,k+1)
  AVGV = AVGV + X(i  ,j+1,k+1)

  AVGV = AVGV/8.

  RETURN
  END FUNCTION AVGV

!#######################################################################
  SUBROUTINE Str2Int(str,val)
!#######################################################################
! Zero is ascii code 48, i.e. ichar('0') == 48
    CHARACTER(*),INTENT(IN) :: str
    INTEGER,INTENT(OUT) :: val
    INTEGER :: tmp
    INTEGER :: length, i, ctr

    length=LEN(TRIM(str))
    ctr = 0
    tmp = 0
    DO i=length,1,-1
      tmp = tmp + (10**ctr)*(ICHAR(str(i:i))-48)
      ctr=ctr+1
    ENDDO
    val = tmp

  END SUBROUTINE Str2Int

!#######################################################################
  SUBROUTINE Prompt_Get(var,msg)
!#######################################################################
    CHARACTER(*),INTENT(OUT) :: var
    CHARACTER(*),INTENT(IN) :: msg

    WRITE(*,*) msg
    READ*,var
    RETURN

  END SUBROUTINE Prompt_Get

!#######################################################################
  SUBROUTINE Check_File_Exist(fnam,ios)
!#######################################################################
    LOGICAL,INTENT(OUT) :: ios
    CHARACTER(*),INTENT(IN) :: fnam

    INQUIRE(FILE=fnam,EXIST=ios)
    RETURN

  END SUBROUTINE Check_File_Exist

!#######################################################################
  SUBROUTINE Fatal_Error(ios,msg)
!#######################################################################
    LOGICAL,INTENT(IN) :: ios
    CHARACTER(*),INTENT(IN) :: msg

    IF(ios)THEN
      RETURN
    ELSE
      WRITE(*,*) msg
      STOP
    ENDIF

  END SUBROUTINE Fatal_Error

!#######################################################################
  FUNCTION int_to_string(i)
!#######################################################################
    INTEGER :: i
    CHARACTER(LEN=50) :: int_to_string

    IF(i.LE.9)THEN
      WRITE(int_to_string,'(i1)') i
    ELSEIF(i.GE.10 .AND. i.LE.99)THEN
      WRITE(int_to_string,'(i2)') i
    ELSEIF(i.GE.100 .AND. i.LE.999)THEN
      WRITE(int_to_string,'(i3)') i
    ELSEIF(i.GE.1000 .AND. i.LE.9999)THEN
      WRITE(int_to_string,'(i4)') i
    ELSEIF(i.GE.10000 .AND. i.LE.99999)THEN
      WRITE(int_to_string,'(i5)') i
    ELSEIF(i.GE.100000 .AND. i.LE.999999)THEN
      WRITE(int_to_string,'(i6)') i
    ENDIF

    RETURN
  END FUNCTION int_to_string

!#######################################################################
  SUBROUTINE MakeUpper(str,slen)
!#######################################################################
    IMPLICIT NONE

    INTEGER, INTENT(IN) :: slen
    CHARACTER(LEN=slen),INTENT(INOUT) :: str

    INTEGER :: shift=0, ilen=0, i=0, ich=0

    shift = ICHAR('a')-ICHAR('A')
    ilen = LEN_TRIM(str)

    DO i = 1, ilen
      ich = ICHAR(str(i:i))
      IF(ich.GE.ICHAR('a') .AND. ich.LE.ICHAR('z')) str(i:i) = CHAR(ich-shift)
    ENDDO

    RETURN
 
  END SUBROUTINE MakeUpper

END MODULE general
